/*
 * Copyright (C) 2013-2014 Paul Davis <paul@linuxaudiosystems.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef __canvas_canvas_fwd_h__
#define __canvas_canvas_fwd_h__

namespace ArdourCanvas {
	class Arrow;
	class Line;
	class LineSet;
	class Rectangle;
	class Ruler;
	class Polygon;
	class PolyLine;
	class GtkCanvas;
	class GtkCanvasViewport;
	class Text;
	class Curve;
	class ScrollGroup;
}

#endif /* __canvas_canvas_fwd_h__ */
