/*
 * Copyright (C) 2021 Paul Davis <paul@linuxaudiosystems.com>
 * Copyright (C) 2021 Ben Loftis <ben@harrisonconsoles.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef _gtk_ardour_midi_trigger_properties_box_h_
#define _gtk_ardour_midi_trigger_properties_box_h_

#include <ytkmm/label.h>

#include "ardour/ardour.h"
#include "ardour/triggerbox.h"

#include "audio_trigger_properties_box.h"

class MidiTriggerPropertiesBox : public TriggerPropertiesBox
{
public:
	MidiTriggerPropertiesBox ();
	~MidiTriggerPropertiesBox ();

protected:
	virtual void on_trigger_changed (const PBD::PropertyChange& what_changed);

private:
	ArdourWidgets::ArdourButton _patch_enable_button;
	ArdourWidgets::ArdourButton _cc_enable_button;

};

#endif
