
/* This file was generated by tools/process-metadata. DO NOT EDIT THIS FILE. EVER! */
void
UIConfiguration::build_metadata ()
{

#define VAR_META(name,...)  { char const * _x[] { __VA_ARGS__ }; all_metadata.insert (std::make_pair<std::string,Metadata> ((name), PBD::upcase (_x))); }

	VAR_META (X_("all-floating-windows-are-dialogs"), _("KDE"), _("quirks"), _("dialogs"), _("window"), _("hidden"), _("behind"),  NULL);
	VAR_META (X_("automation-edit-cancels-auto-hide"), _("automation"), _("spill"), _("touch"),  NULL);
	VAR_META (X_("autoplay-clips"), _("audition"), _("clips"), _("slots"), _("triggers"), _("autoplay"), _("automatically"), _("play"),  NULL);
	VAR_META (X_("autoplay-files"), _("files"), _("audition"), _("autoplay"), _("automatically"), _("play"),  NULL);
	VAR_META (X_("autoscroll-editor"), _("scroll"), _("editor"),  NULL);
	VAR_META (X_("blink-alert-indicators"), _("blink"), _("alert"), _("indicators"),  NULL);
	VAR_META (X_("blink-rec-arm"), _("appearance"), _("blink"), _("record"), _("rec"), _("enable"), _("rec-enable"), _("record-enable"),  NULL);
	VAR_META (X_("boxy-buttons"), _("appearance"), _("style"), _("boxy"), _("buttons"), _("theme"),  NULL);
	VAR_META (X_("buggy-gradients"), _("appearance"), _("bugs"), _("tweaks"), _("kwirks"),  NULL);
	VAR_META (X_("check-announcements"), _("check"), _("announcements"), _("phone"), _("home"),  NULL);
	VAR_META (X_("clock-display-limit"), _("clock"), _("display"), _("limit"), _("length"), _("maximum"), _("duration"),  NULL);
	VAR_META (X_("color-file"), _("theme"), _("colors"), _("appearance"), _("style"), _("theming"),  NULL);
	VAR_META (X_("color-regions-using-track-color"), _("theme"), _("colors"), _("appearance"), _("style"), _("theming"),  NULL);
	VAR_META (X_("default-bindings"), _("shortcuts"), _("keys"), _("keybindings"), _("bindings"),  NULL);
	VAR_META (X_("default-lower-midi-note"), _("MIDI"), _("low"), _("lowest"), _("lower"),  NULL);
	VAR_META (X_("default-narrow_ms"), _("appearance"), _("width"), _("mixer"),  NULL);
	VAR_META (X_("default-upper-midi-note"), _("MIDI"), _("upper"), _("high"), _("highest"),  NULL);
	VAR_META (X_("draggable-playhead"), _("playhead"), _("dragging"),  NULL);
	VAR_META (X_("draggable-playhead-speed"), _("playhead"), _("dragging"),  NULL);
	VAR_META (X_("editor-stereo-only-meters"), _("editor"), _("stereo"), _("meters"),  NULL);
	VAR_META (X_("flat-buttons"), _("appearance"), _("style"), _("flat"), _("buttons"), _("theme"), _("ableton"),  NULL);
	VAR_META (X_("floating-monitor-section"), _("appearance"), _("monitor"), _("floating"), _("separate"), _("window"),  NULL);
	VAR_META (X_("font-scale"), _("fonts"), _("font"), _("size"), _("scaling"), _("readable"), _("readability"),  NULL);
	VAR_META (X_("freesound-dir"), _("freesound"), _("folder"), _("folders"), _("directory"), _("directories"), _("download"),  NULL);
	VAR_META (X_("grid-follows-internal"), _("grid"), _("switch"), _("automatic"), _("tools"), _("mode"), _("selection"), _("internal"), _("edit"),  NULL);
	VAR_META (X_("hide-splash-screen"), _("appearance"), _("hide"), _("splash"), _("screen"),  NULL);
	VAR_META (X_("input-meter-layout"), _("meter"), _("recorder"), _("layout"),  NULL);
	VAR_META (X_("input-meter-scopes"), _("scopes"), _("recorder"), _("layout"),  NULL);
	VAR_META (X_("insert-at-position"), _("import"), _("embed"), _("position"), _("insert"),  NULL);
	VAR_META (X_("lock-gui-after-seconds"), _("lock"), _("gui"), _("time"), _("seconds"),  NULL);
	VAR_META (X_("max-note-height"), _("maximum"), _("note"), _("height"), _("theme"), _("appearance"), _("size"), _("visible"), _("range"),  NULL);
	VAR_META (X_("max-inline-controls"), _("plugins"), _("mixer-strip"), _("controls"), _("inline"),  NULL);
	VAR_META (X_("max-plugin-chart"), _("plugins"), _("top"), _("favorite"), _("sidebar"),  NULL);
	VAR_META (X_("max-plugin-recent"), _("plugins"), _("recent"), _("favorite"), _("sidebar"),  NULL);
	VAR_META (X_("meter-hold"), _("metering"), _("standard"), _("hold"), _("falloff"), _("level"), _("dbfs"),  NULL);
	VAR_META (X_("meter-line-up-din"), _("metering"), _("standard"), _("din"), _("line-up"), _("level"), _("dbfs"),  NULL);
	VAR_META (X_("meter-line-up-level"), _("metering"), _("standard"), _("line-up"), _("level"), _("dbfs"),  NULL);
	VAR_META (X_("meter-peak"), _("metering"), _("peak"), _("line-up"), _("level"), _("dbfs"),  NULL);
	VAR_META (X_("meter-style-led"), _("metering"), _("style"), _("led"),  NULL);
	VAR_META (X_("meter-vu-standard"), _("metering"), _("standard"), _("vu"), _("line-up"), _("level"), _("dbfs"),  NULL);
	VAR_META (X_("mixer-element-visibility"), _("mixer"), _("visibility"), _("comment"), _("solo"), _("mute"), _("phase"), _("polarity"), _("invert"), _("rec"), _("solo"), _("locked"), _("vca"), _("trigger"),  NULL);
	VAR_META (X_("name-new-markers"), _("markers"), _("name"), _("dialog"), _("naming"),  NULL);
	VAR_META (X_("never-display-periodic-midi"), _("sysex"), _("mtc"), _("midi"), _("regions"), _("messages"), _("events"),  NULL);
	VAR_META (X_("new-automation-points-on-lane"), _("automation"), _("y-axis"), _("click"), _("draw"), _("lane"), _("points"),  NULL);
	VAR_META (X_("note-name-display"), _("notes"), _("names"), _("midnam"), _("display"), _("show"),  NULL);
	VAR_META (X_("one-plugin-window-only"), _("plugins"), _("window"), _("single"), _("close"),  NULL);
	VAR_META (X_("only-copy-imported-files"), _("drag"), _("drop"), _("import"), _("copy"), _("link"), _("files"),  NULL);
	VAR_META (X_("open-gui-after-adding-plugin"), _("plugins"), _("gui"), _("open"), _("window"),  NULL);
	VAR_META (X_("plugin-gui-behavior"), _("plugins"), _("gui"), _("resources"), _("vst"), _("destroy"), _("window"),  NULL);
	VAR_META (X_("prefer-inline-over-gui"), _("plugins"), _("inline"), _("mixer"), _("strip"),  NULL);
	VAR_META (X_("prefer-tap-tempo"), _("tap"), _("tempo"), _("preference"),  NULL);
	VAR_META (X_("primary-clock-delta-mode"), _("clocks"), _("mode"), _("primary"), _("delta"), _("difference"), _("edit"), _("cursor"), _("playhead"),  NULL);
	VAR_META (X_("rubberbanding-snaps-to-grid"), _("rubberband"), _("snap"), _("grid"), _("time"), _("stretch"),  NULL);
	VAR_META (X_("ruler-granularity"), _("ruler"), _("spacing"), _("grid"), _("snap"), _("distance"),  NULL);
	VAR_META (X_("rulers-follow-grid"), _("ruler"), _("grid"), _("auto"), _("show"), _("visibility"),  NULL);
	VAR_META (X_("save-export-analysis-image"), _("export"), _("analysis"), _("save"), _("image"),  NULL);
	VAR_META (X_("save-export-mixer-screenshot"), _("export"), _("save"), _("image"), _("mixer"), _("screenshot"),  NULL);
	VAR_META (X_("screen-saver-mode"), _("screensaver"), _("screen-saver"), _("prevent"), _("lock"),  NULL);
	VAR_META (X_("secondary-clock-delta-mode"), _("clocks"), _("mode"), _("secondary"), _("delta"), _("difference"), _("edit"), _("cursor"), _("playhead"),  NULL);
	VAR_META (X_("show-editor-meter"), _("toolbar"), _("editor"), _("meter"), _("master"),  NULL);
	VAR_META (X_("show-inline-display-by-default"), _("plugins"), _("inline"), _("controls"), _("mixer"),  NULL);
	VAR_META (X_("show-manager-if-plugins-are-missing"), _("plugins"), _("manager"), _("missing"),  NULL);
	VAR_META (X_("show-mini-timeline"), _("minitimeline"), _("mini"), _("timeline"), _("toolbar"),  NULL);
	VAR_META (X_("show-name-highlight"), _("editor"), _("region"), _("name"), _("bar"), _("highlight"), _("layout"),  NULL);
	VAR_META (X_("show-plugin-scan-window"), _("plugins"), _("popup"), _("progress"), _("scan"), _("indexing"),  NULL);
	VAR_META (X_("show-region-cue-markers"), _("cue"), _("markers"), _("regions"), _("editor"),  NULL);
	VAR_META (X_("show-region-gain"), _("regions"), _("gain"), _("envelope"), _("volume"), _("modes"),  NULL);
	VAR_META (X_("show-region-name"), _("regions"), _("name"), _("appearance"), _("editor"),  NULL);
	VAR_META (X_("show-selection-marker"), _("selection"), _("ruler"), _("marker"), _("range"),  NULL);
	VAR_META (X_("show-region-xrun-markers"), _("regions"), _("xruns"), _("markers"), _("dropouts"),  NULL);
	VAR_META (X_("show-secondary-clock"), _("clock"), _("toolbar"), _("visibility"), _("secondary"),  NULL);
	VAR_META (X_("show-snapped-cursor"), _("cursor"), _("snap"), _("edit"), _("point"), _("line"),  NULL);
	VAR_META (X_("show-toolbar-cuectrl"), _("toolbar"), _("cue"),  NULL);
	VAR_META (X_("show-toolbar-latency"), _("toolbar"), _("latency"), _("information"),  NULL);
	VAR_META (X_("show-toolbar-monitor-info"), _("toolbar"), _("monitoring"), _("information"), _("section"),  NULL);
	VAR_META (X_("show-toolbar-monitoring"), _("toolbar"), _("monitoring"), _("information"), _("section"),  NULL);
	VAR_META (X_("show-toolbar-recpunch"), _("toolbar"), _("punch"), _("recording"), _("layered"), _("sound"), _("transparent"),  NULL);
	VAR_META (X_("show-toolbar-selclock"), _("toolbar"), _("clock"), _("selection"), _("time"), _("display"),  NULL);
	VAR_META (X_("show-track-meters"), _("editor"), _("meter"), _("small"), _("track"), _("header"),  NULL);
	VAR_META (X_("show-waveform-clipping"), _("waveform"), _("clipping"), _("red"), _("peak"),  NULL);
	VAR_META (X_("show-waveforms"), _("waveform"), _("display"), _("box"), _("empty"), _("invisible"), _("hide"), _("show"),  NULL);
	VAR_META (X_("show-waveforms-while-recording"), _("waveform"), _("display"), _("recording"),  NULL);
	VAR_META (X_("snap-threshold"), _("snap"), _("threshold"), _("distance"), _("mouse"),  NULL);
	VAR_META (X_("snap-target"), _("snap"), _("grid"), _("target"),  NULL);
	VAR_META (X_("snap-to-grid"), _("snap"), _("grid"), _("editor"), _("mouse"),  NULL);
	VAR_META (X_("snap-to-marks"), _("snap"), _("grid"), _("editor"), _("mouse"), _("markers"), _("marks"),  NULL);
	VAR_META (X_("snap-to-playhead"), _("snap"), _("grid"), _("editor"), _("mouse"), _("playhead"),  NULL);
	VAR_META (X_("snap-to-region-end"), _("snap"), _("grid"), _("editor"), _("mouse"), _("regions"), _("boundaries"), _("boundary"),  NULL);
	VAR_META (X_("snap-to-region-start"), _("snap"), _("grid"), _("editor"), _("mouse"), _("regions"), _("boundaries"), _("boundary"),  NULL);
	VAR_META (X_("snap-to-region-sync"), _("snap"), _("grid"), _("editor"), _("mouse"), _("regions"), _("sync"),  NULL);
	VAR_META (X_("sound-midi-notes"), _("play"), _("notes"), _("selection"), _("midi"), _("sound"), _("hear"),  NULL);
	VAR_META (X_("super-rapid-clock-update"), _("clocks"), _("fast"), _("update"), _("display"), _("timecode"), _("video"), _("slow"), _("updates"),  NULL);
	VAR_META (X_("time-axis-name-ellipsize-mode"), _("editor"), _("width"), _("name"), _("track"), _("header"), _("ellipsize"), _("size"),  NULL);
	VAR_META (X_("timeline-item-gradient-depth"), _("editor"), _("gradient"), _("blur"), _("contrast"), _("style"),  NULL);
	VAR_META (X_("transients-follow-front"), _("KDE"), _("quirks"), _("dialogs"), _("window"), _("hidden"), _("behind"), _("transient"),  NULL);
	VAR_META (X_("update-editor-during-summary-drag"), _("editor"), _("summary"), _("scroll"), _("drag"), _("pan"), _("follow"),  NULL);
	VAR_META (X_("use-double-click-to-zoom-to-selection"), _("editor"), _("zoom"), _("selection"), _("mouse"), _("click"),  NULL);
	VAR_META (X_("use-mouse-position-as-zoom-focus-on-scroll"), _("editor"), _("zoom"), _("selection"), _("mouse"), _("wheel"), _("focus"), _("scroll"),  NULL);
	VAR_META (X_("use-note-bars-for-velocity"), _("midi"), _("velocity"), _("bars"), _("box"),  NULL);
	VAR_META (X_("default-midi-note-color-mode"), _("midi"), _("velocity"), _("colors"), _("pitches"), _("channels"),  NULL);
	VAR_META (X_("use-route-color-widely"), _("mixer"), _("colors"), _("faders"),  NULL);
	VAR_META (X_("use-time-rulers-to-zoom-with-vertical-drag"), _("zoom"), _("drag"), _("mouse"),  NULL);
	VAR_META (X_("use-tooltips"), _("tooltips"), _("help"), _("popup"), _("information"),  NULL);
	VAR_META (X_("use-wm-visibility"), _("quirks"), _("visibility"), _("switch"), _("windows"),  NULL);
	VAR_META (X_("vertical-region-gap"), _("visual"), _("gap"), _("tracks"), _("spacing"),  NULL);
	VAR_META (X_("vertical-summary-uses-track-colors"), _("summary"), _("scrollbar"), _("tracks"), _("colors"),  NULL);
	VAR_META (X_("vkeybd-layout"), _("virtual"), _("keyboard"), _("layout"), _("qwerty"), _("midi"),  NULL);
	VAR_META (X_("waveform-cache-size"), _("memory"), _("cache"), _("performance"), _("optimization"), _("image"),  NULL);
	VAR_META (X_("waveform-clip-level"), _("clip"), _("level"), _("dbfs"), _("waveform"), _("peaking"),  NULL);
	VAR_META (X_("waveform-gradient-depth"), _("editor"), _("gradient"), _("blur"), _("contrast"), _("style"), _("waveform"),  NULL);
	VAR_META (X_("waveform-scale"), _("waveform"), _("logarithmic"), _("logscale"), _("linear"),  NULL);
	VAR_META (X_("waveform-shape"), _("waveform"), _("rectified"), _("half"), _("shape"), _("display"),  NULL);
	VAR_META (X_("widget-prelight"), _("highlight"), _("mouse"), _("prelight"), _("hover"), _("indicate"),  NULL);
	VAR_META (X_("drag-sensitivity"), _("drag"), _("sensitivity"), _("sensitive"), _("move"), _("threshold"),  NULL);
	VAR_META (X_("auto-play"), _("transport"), _("autoplay"), _("automatically"), _("play"),  NULL);
	VAR_META (X_("auto-return"), _("transport"), _("autoreturn"), _("automatically"), _("return"), _("go"), _("back"),  NULL);
	VAR_META (X_("count-in"), _("transport"), _("count"), _("in"), _("count-in"),  NULL);
	VAR_META (X_("default-time-domain"), _("time"), _("domain"), _("musical"), _("audio"), _("markers"),  NULL);
	VAR_META (X_("external-sync"), _("transport"), _("use"), _("external"), _("synchronization"), _("master"),  NULL);
	VAR_META (X_("jack-time-master"), _("transport"), _("synchronization"), _("jack"), _("time"), _("master"),  NULL);
	VAR_META (X_("meterbridge-label-height"), _("meter"), _("meterbridge"), _("bridge"), _("label"), _("height"),  NULL);
	VAR_META (X_("midi-copy-is-fork"), _("midi"), _("copy"), _("clone"), _("duplicate"), _("fork"), _("linked"),  NULL);
	VAR_META (X_("native-file-data-format"), _("audio"), _("files"), _("audiofiles"), _("native"), _("data"), _("format"),  NULL);
	VAR_META (X_("native-file-header-format"), _("audio"), _("files"), _("audiofiles"), _("native"), _("header"), _("format"), _("container"), _("envelope"),  NULL);
	VAR_META (X_("realtime-export"), _("exporting"), _("realtime"), _("real"), _("time"), _("nonrealtime"), _("nonreal"),  NULL);
	VAR_META (X_("record-mode"), _("recording"), _("capture"), _("layers"), _("layering"), _("sound"), _("on"), _("soundonsound"), _("nonlayered"), _("layered"),  NULL);
	VAR_META (X_("session-monitoring"), _("monitoring"), _("hearing"), _("listening"), _("signal"), _("monitors"),  NULL);
	VAR_META (X_("show-busses-on-meterbridge"), _("busses"), _("bus"), _("meterbridge"), _("meters"),  NULL);
	VAR_META (X_("show-fader-on-meterbridge"), _("faders"), _("meterbridge"),  NULL);
	VAR_META (X_("show-group-tabs"), _("group"), _("tabs"), _("show"), _("hide"), _("visible"), _("visibility"),  NULL);
	VAR_META (X_("show-master-on-meterbridge"), _("master"), _("meterbridge"),  NULL);
	VAR_META (X_("show-midi-on-meterbridge"), _("midi"), _("meterbridge"),  NULL);
	VAR_META (X_("show-monitor-on-meterbridge"), _("monitor"), _("meterbridge"),  NULL);
	VAR_META (X_("show-mute-on-meterbridge"), _("mute"), _("meterbridge"),  NULL);
	VAR_META (X_("show-name-on-meterbridge"), _("solo"), _("meterbridge"),  NULL);
	VAR_META (X_("show-rec-on-meterbridge"), _("record"), _("recenable"), _("meterbridge"),  NULL);
	VAR_META (X_("show-region-fades"), _("region"), _("gain"), _("fades"), _("show"), _("hide"), _("fadein"), _("fadeout"),  NULL);
	VAR_META (X_("show-solo-on-meterbridge"), _("solo"), _("meterbridge"),  NULL);
	VAR_META (X_("show-summary"), _("summary"), _("show"),  NULL);
	VAR_META (X_("show-vertical-summary"), _("vertical"), _("summary"), _("show"),  NULL);
	VAR_META (X_("slave-timecode-offset"), _("timecode"), _("synchronization"), _("offset"),  NULL);
	VAR_META (X_("subframes-per-frame"), _("timecode"), _("subframes"), _("sync"),  NULL);
	VAR_META (X_("take-name"), _("take"), _("names"),  NULL);
	VAR_META (X_("timecode-format"), _("timecode"), _("synchronization"), _("format"), _("tc"),  NULL);
	VAR_META (X_("timecode-generator-offset"), _("timecode"), _("synchronization"), _("offset"), _("generate"), _("send"),  NULL);
	VAR_META (X_("timecode-offset"), _("timecode"), _("synchronization"), _("offset"), _("receive"),  NULL);
	VAR_META (X_("timecode-offset-negative"), _("timecode"), _("synchronization"), _("offset"), _("negative"),  NULL);
	VAR_META (X_("track-name-number"), _("name"), _("naming"), _("numbered"), _("tracks"),  NULL);
	VAR_META (X_("track-name-take"), _("name"), _("naming"), _("tracks"), _("takes"),  NULL);
	VAR_META (X_("use-monitor-fades"), _("monitoring"), _("fades"), _("fadein"), _("fadeout"), _("enable"), _("disable"), _("use"),  NULL);
	VAR_META (X_("use-region-fades"), _("region"), _("fades"), _("fadein"), _("fadeout"), _("enable"), _("disable"), _("use"),  NULL);
	VAR_META (X_("use-transport-fades"), _("transport"), _("fades"), _("fadein"), _("fadeout"), _("enable"), _("disable"), _("use"),  NULL);
	VAR_META (X_("use-video-file-fps"), _("video"), _("use"), _("frames"), _("per"), _("second"), _("fps"),  NULL);
	VAR_META (X_("afl-position"), _("monitoring"), _("monitor"), _("afl"), _("pfl"), _("pre"), _("post"), _("position"),  NULL);
	VAR_META (X_("auto-analyse-audio"), _("automatic"), _("automated"), _("audio"), _("analysis"), _("transients"),  NULL);
	VAR_META (X_("automation-follows-regions"), _("automation"), _("follows"), _("regions"),  NULL);
	VAR_META (X_("automation-interval-msecs"), _("automation"),  NULL);
	VAR_META (X_("automation-thinning-factor"), _("automation"),  NULL);
	VAR_META (X_("click-emphasis-sound"), _("metronome"), _("click"), _("beat"), _("downbeat"), _("emphasis"), _("sample"), _("sound"),  NULL);
	VAR_META (X_("click-gain"), _("metronome"), _("click"), _("beat"), _("volume"), _("gain"), _("level"),  NULL);
	VAR_META (X_("click-sound"), _("metronome"), _("click"), _("beat"), _("sound"), _("sample"),  NULL);
	VAR_META (X_("clip-library-dir"), _("folder"), _("folders"), _("directory"), _("directories"), _("download"), _("clips"), _("library"),  NULL);
	VAR_META (X_("cpu-dma-latency"), _("cpu"), _("dma"), _("latency"), _("performance"), _("xrun"),  NULL);
	VAR_META (X_("create-xrun-marker"), _("xrun"), _("xmarker"),  NULL);
	VAR_META (X_("default-automation-time-domain"), _("automation"), _("time"), _("domain"), _("default"),  NULL);
	VAR_META (X_("default-fade-shape"), _("fades"), _("fadein"), _("fadeout"), _("shape"), _("type"), _("style"), _("default"),  NULL);
	VAR_META (X_("default-session-parent-dir"), _("folders"), _("directory"), _("directories"), _("session"), _("path"), _("default"),  NULL);
	VAR_META (X_("denormal-model"), _("denormal"), _("model"), _("handling"), _("cpu"), _("performance"), _("speed"), _("xruns"), _("dsp"), _("load"),  NULL);
	VAR_META (X_("denormal-protection"), _("denormal"), _("model"), _("handling"), _("cpu"), _("performance"), _("speed"), _("xruns"), _("dsp"), _("load"),  NULL);
	VAR_META (X_("discover-plugins-on-start"), _("plugins"), _("scan"), _("discover"), _("rescan"), _("reload"), _("startup"),  NULL);
	VAR_META (X_("history-depth"), _("history"), _("undo"), _("redo"), _("depth"), _("length"), _("size"),  NULL);
	VAR_META (X_("layer-model"), _("editing"), _("layering"), _("model"), _("style"), _("type"),  NULL);
	VAR_META (X_("link-send-and-route-panner"), _("mixing"), _("panning"), _("send"), _("panner"), _("link"), _("connect"), _("tie"),  NULL);
	VAR_META (X_("listen-position"), _("afl"), _("pfl"), _("listen"), _("monitoring"), _("position"),  NULL);
	VAR_META (X_("loop-fade-choice"), _("looping"), _("fades"), _("fadein"), _("fadeout"), _("choice"), _("type"), _("style"), _("model"),  NULL);
	VAR_META (X_("loop-is-mode"), _("looping"), _("mode"), _("behavior"),  NULL);
	VAR_META (X_("ltc-output-port"), _("timecode"), _("output"), _("port"), _("routing"),  NULL);
	VAR_META (X_("ltc-output-volume"), _("timecode"), _("output"), _("level"), _("gain"), _("volume"), _("ltc"),  NULL);
	VAR_META (X_("ltc-send-continuously"), _("timecode"), _("transmit"), _("send"), _("continuously"), _("always"), _("ltc"),  NULL);
	VAR_META (X_("max-gain"), _("volume"), _("gain"), _("fader"), _("maximum"), _("peak"),  NULL);
	VAR_META (X_("max-recent-sessions"), _("number"), _("maximum"), _("recent"), _("sessions"),  NULL);
	VAR_META (X_("max-recent-templates"), _("number"), _("maximum"), _("recent"), _("templates"),  NULL);
	VAR_META (X_("max-transport-speed"), _("maximum"), _("top"), _("fastest"), _("highest"), _("transport"), _("speed"),  NULL);
	VAR_META (X_("meter-falloff"), _("meter"), _("falloff"), _("fall"), _("off"),  NULL);
	VAR_META (X_("meter-type-bus"), _("meter"), _("type"), _("bus"),  NULL);
	VAR_META (X_("meter-type-master"), _("meter"), _("type"), _("master"),  NULL);
	VAR_META (X_("meter-type-track"), _("meter"), _("type"), _("track"),  NULL);
	VAR_META (X_("midi-audition-synth-uri"), _("midi"), _("audition"), _("synthesizer"), _("plugin"), _("name"), _("id"), _("uri"), _("location"),  NULL);
	VAR_META (X_("midi-chase"), _("midi"), _("chase"), _("notes"), _("locate"), _("play"),  NULL);
	VAR_META (X_("midi-clock-resolution"), _("midi"), _("clock"), _("resolution"), _("quantize"), _("limit"), _("precision"),  NULL);
	VAR_META (X_("midi-clock-sets-tempo"), _("midi"), _("clock"), _("mclk"), _("set"), _("tempo"),  NULL);
	VAR_META (X_("midi-feedback"), _("midi"), _("generic"), _("control"), _("surface"), _("feedback"),  NULL);
	VAR_META (X_("midi-input-follows-selection"), _("midi"), _("inputs"), _("port"), _("device"), _("follow"), _("selection"), _("tracks"),  NULL);
	VAR_META (X_("midi-panic-when-looping"), _("midi"), _("panic"), _("looping"), _("ports"), _("notes"), _("off"),  NULL);
	VAR_META (X_("minimum-disk-read-bytes"), _("disk"), _("disc"), _("i/o"), _("io"), _("chunk"), _("read"), _("size"), _("bytes"), _("buffering"),  NULL);
	VAR_META (X_("minimum-disk-write-bytes"), _("disk"), _("disc"), _("i/o"), _("io"), _("chunk"), _("write"), _("size"), _("bytes"), _("buffering"),  NULL);
	VAR_META (X_("mmc-receive-device-id"), _("midi"), _("machine"), _("control"), _("mmc"), _("receive"), _("receiving"), _("device"), _("id"),  NULL);
	VAR_META (X_("mmc-send-device-id"), _("midi"), _("machine"), _("control"), _("mmc"), _("send"), _("sending"), _("device"), _("id"),  NULL);
	VAR_META (X_("monitoring-model"), _("monitoring"), _("model"), _("style"), _("type"),  NULL);
	VAR_META (X_("mtc-qf-speed-tolerance"), _("transport"), _("synchronization"), _("MIDI"), _("timecode"), _("time"), _("code"), _("threshold"), _("tolerance"), _("sensitivity"), _("quarter"), _("frame"),  NULL);
	VAR_META (X_("mute-affects-control-outs"), _("mute"), _("muting"), _("monitor"), _("outs"), _("control"), _("outputs"),  NULL);
	VAR_META (X_("mute-affects-main-outs"), _("mute"), _("muting"), _("main"), _("outs"), _("outputs"), _("master"),  NULL);
	VAR_META (X_("mute-affects-post-fader"), _("mute"), _("muting"), _("postfader"), _("post"), _("fader"),  NULL);
	VAR_META (X_("mute-affects-pre-fader"), _("mute"), _("muting"), _("prefader"), _("pre"), _("fader"),  NULL);
	VAR_META (X_("new-plugins-active"), _("plugins"), _("newly"), _("added"), _("active"),  NULL);
	VAR_META (X_("osc-port"), _("OSC"), _("port"), _("number"),  NULL);
	VAR_META (X_("output-auto-connect"), _("outputs"), _("automatically"), _("connect"),  NULL);
	VAR_META (X_("periodic-safety-backup-interval"), _("backups"), _("autosave"),  NULL);
	VAR_META (X_("periodic-safety-backups"), _("backups"), _("autosave"),  NULL);
	VAR_META (X_("pfl-position"), _("solo"), _("listen"), _("prefader"), _("position"),  NULL);
	VAR_META (X_("playback-buffer-seconds"), _("performance"), _("disk"), _("disc"), _("buffering"), _("playback"), _("keep"), _("up"), _("size"), _("seconds"),  NULL);
	VAR_META (X_("plugin-path-lxvst"), _("plugins"), _("paths"), _("folders"), _("directory"), _("directories"), _("linux"), _("vst"), _("lxvst"), _("search"),  NULL);
	VAR_META (X_("plugin-path-vst"), _("plugins"), _("paths"), _("folders"), _("directory"), _("directories"), _("vst"), _("search"),  NULL);
	VAR_META (X_("plugin-path-vst3"), _("plugins"), _("paths"), _("folders"), _("directory"), _("directories"), _("vst3"), _("search"),  NULL);
	VAR_META (X_("plugin-scan-timeout"), _("plugins"), _("scan"), _("timeout"), _("fail"), _("wait"),  NULL);
	VAR_META (X_("plugins-stop-with-transport"), _("plugins"), _("stop"), _("transport"), _("tail"), _("reverb"), _("ringing"), _("reset"),  NULL);
	VAR_META (X_("port-resampler-quality"), _("resampling"), _("audioengine"), _("global"), _("quality"), _("level"),  NULL);
	VAR_META (X_("processor-usage"), _("cpu"), _("threads"), _("usage"), _("distribute"), _("parallel"),  NULL);
	VAR_META (X_("quieten-at-speed"), _("volume"), _("gain"), _("level"), _("reduce"), _("quieten"), _("cut"), _("speed"), _("fast"), _("forward"), _("rewind"), _("ffwd"),  NULL);
	VAR_META (X_("range-selection-after-split"), _("editing"), _("split"), _("selection"), _("range"), _("after"), _("split"),  NULL);
	VAR_META (X_("recording-resets-xrun-count"), _("recording"), _("resets"), _("xruns"),  NULL);
	VAR_META (X_("region-boundaries-from-onscreen_tracks"), _("grid"), _("region"), _("boundaries"), _("from"), _("visible"), _("onscreen"), _("tracks"), _("lanes"),  NULL);
	VAR_META (X_("region-boundaries-from-selected-tracks"), _("grid"), _("region"), _("boundaries"), _("from"), _("selected"), _("tracks"), _("lanes"),  NULL);
	VAR_META (X_("region-equivalency"), _("options"), _("equivalent"), _("equivalence"), _("regions"), _("comparison"), _("selection"),  NULL);
	VAR_META (X_("region-selection-after-split"), _("selection"), _("regions"), _("after"), _("split"), _("left"), _("right"), _("side"),  NULL);
	VAR_META (X_("replicate-missing-region-channels"), _("channels"), _("replicate"), _("missing"), _("region"),  NULL);
	VAR_META (X_("reset-default-speed-on-stop"), _("transport"), _("reset"), _("default"), _("speed"), _("stop"),  NULL);
	VAR_META (X_("ripple-mode"), _("ripple"), _("all"), _("interview"), _("selected"),  NULL);
	VAR_META (X_("run-all-transport-masters-always"), _("synchronization"), _("always"), _("run"), _("masters"), _("transport"),  NULL);
	VAR_META (X_("sample-lib-path"), _("files"), _("folders"), _("samples"), _("library"), _("path"),  NULL);
	VAR_META (X_("save-history"), _("history"), _("save"), _("disk"), _("serialize"), _("store"),  NULL);
	VAR_META (X_("save-history-depth"), _("history"), _("depth"), _("size"), _("length"),  NULL);
	VAR_META (X_("send-ltc"), _("synchronization"), _("send"), _("transmit"), _("deliver"), _("linear"), _("timecode"), _("ltc"),  NULL);
	VAR_META (X_("send-midi-clock"), _("synchronization"), _("send"), _("transmit"), _("deliver"), _("midi"), _("clock"), _("mclk"),  NULL);
	VAR_META (X_("send-mmc"), _("synchronization"), _("send"), _("transmit"), _("deliver"), _("midi"), _("machine"), _("code"), _("mmc"),  NULL);
	VAR_META (X_("send-mtc"), _("synchronization"), _("send"), _("transmit"), _("deliver"), _("midi"), _("time"), _("code"), _("timecode"), _("mtc"),  NULL);
	VAR_META (X_("show-solo-mutes"), _("appearance"), _("solo"), _("mute"),  NULL);
	VAR_META (X_("shuttle-max-speed"), _("transport"), _("shuttle"), _("varispeed"), _("control"), _("speed"), _("maximum"),  NULL);
	VAR_META (X_("shuttle-speed-factor"), _("transport"), _("shuttle"), _("varispeed"), _("control"), _("speed"),  NULL);
	VAR_META (X_("shuttle-speed-threshold"), _("transport"), _("shuttle"), _("varispeed"), _("control"), _("threshold"),  NULL);
	VAR_META (X_("shuttle-units"), _("transport"), _("shuttle"), _("varispeed"), _("units"), _("semitones"), _("percentage"),  NULL);
	VAR_META (X_("solo-mute-override"), _("solo"), _("mute"), _("override"), _("overrule"), _("priority"),  NULL);
	VAR_META (X_("stop-at-session-end"), _("transport"), _("stop"), _("end"), _("session"),  NULL);
	VAR_META (X_("stop-recording-on-xrun"), _("transport"), _("recording"), _("xrun"), _("stop"),  NULL);
	VAR_META (X_("strict-io"), _("plugins"), _("outputs"), _("number"), _("flexible"), _("variable"), _("fixed"),  NULL);
	VAR_META (X_("timecode-sync-frame-rate"), _("timecode"), _("synchronization"), _("frame"), _("rate"),  NULL);
	VAR_META (X_("tracks-auto-naming"), _("automatically"), _("name"), _("tracks"),  NULL);
	VAR_META (X_("transient-sensitivity"), _("analysis"), _("transients"), _("sensitive"), _("sensitivity"), _("vamp"),  NULL);
	VAR_META (X_("try-autostart-engine"), _("audioengine"), _("startup"),  NULL);
	VAR_META (X_("use-audio-units"), _("plugins"), _("audiounits"), _("audio-unit"), _("enable"), _("disable"),  NULL);
	VAR_META (X_("use-click-emphasis"), _("metronome"), _("beat"), _("click"), _("sound"), _("emphasis"),  NULL);
	VAR_META (X_("use-lxvst"), _("plugins"), _("vst2"), _("enable"), _("disable"),  NULL);
	VAR_META (X_("use-macvst"), _("plugins"), _("vst2"), _("enable"), _("disable"),  NULL);
	VAR_META (X_("use-monitor-bus"), _("monitoring"), _("bus"), _("optional"),  NULL);
	VAR_META (X_("use-osc"), _("osc"), _("open"), _("sound"), _("control"),  NULL);
	VAR_META (X_("use-plugin-own-gui"), _("plugins"), _("GUI"), _("editor"), _("prefer"), _("use"), _("own"),  NULL);
	VAR_META (X_("use-vst3"), _("plugins"), _("vst3"),  NULL);
	VAR_META (X_("verbose-plugin-scan"), _("plugins"), _("scanning"), _("verbose"),  NULL);
	VAR_META (X_("verify-remove-last-capture"), _("verify"), _("remove"), _("removal"), _("last"), _("most"), _("recent"), _("capture"), _("recording"), _("take"),  NULL);
	VAR_META (X_("video-advanced-setup"), _("video"), _("setup"),  NULL);
	VAR_META (X_("video-server-url"), _("video"), _("server"), _("location"), _("url"), _("file"),  NULL);

}
