/*
 * Copyright (C) 2018 Paul Davis <paul@linuxaudiosystems.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef __CANVAS_NOTE_H__
#define __CANVAS_NOTE_H__

#include "canvas/visibility.h"
#include "canvas/rectangle.h"

namespace ArdourCanvas
{

class LIBCANVAS_API Note : public Rectangle
{
public:
	Note (Canvas*);
	Note (Item*);

	void render (Rect const &, Cairo::RefPtr<Cairo::Context>) const;
	void set_velocity (double fract);
	void set_fill_color (Gtkmm2ext::Color);
	void set_outline_color (Gtkmm2ext::Color);

	double velocity() const { return _velocity; }

	static void set_show_velocity_bars (bool);

  private:
	static bool      _show_velocity_bars;
	double           _velocity;
	Gtkmm2ext::Color _velocity_color;
};

}

#endif /* __CANVAS_NOTE_H__ */
