/*
 * (C) 2026 EZ4Stephen <ez4stephen@tutamail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <chrono>

inline int gettimeofday(struct timeval* tv, struct timezone*) {
    if (!tv) return -1;

    using namespace std::chrono;

    auto now = system_clock::now();
    auto duration = now.time_since_epoch();

    auto secs  = duration_cast<seconds>(duration);
    auto usecs = duration_cast<microseconds>(duration - secs);

    tv->tv_sec  = static_cast<long>(secs.count());
    tv->tv_usec = static_cast<long>(usecs.count());

    return 0;
}
