//
//  main.m
//  CrashReporter
//
//  Created by Taybin Rutkin on 6/10/09.
//  Copyright Penguin Sounds 2009. All rights reserved.
//

#import <Cocoa/Cocoa.h>

int main(int argc, char *argv[])
{
    return NSApplicationMain(argc,  (const char **) argv);
}
