#ifndef __OPDefinition_h__
#define __OPDefinition_h__

#include "aaf/AAFTypes.h"

// AAF well-known OPDefinition instances
//

//{0d011201-0100-0000-060e-2b3404010105}
static const aafUID_t AAFOPDef_EditProtocol =
{0x0d011201, 0x0100, 0x0000, {0x06, 0x0e, 0x2b, 0x34, 0x04, 0x01, 0x01, 0x05}};


//{0d011201-0200-0000-060e-2b3404010109}
static const aafUID_t AAFOPDef_Unconstrained =
{0x0d011201, 0x0200, 0x0000, {0x06, 0x0e, 0x2b, 0x34, 0x04, 0x01, 0x01, 0x09}};


// AAF OPDefinition legacy aliases
//


#endif // ! __OPDefinition_h__
