#include "evoral/types.h"

PBD::DebugBits PBD::DEBUG::Sequence = PBD::new_debug_bit ("sequence");
PBD::DebugBits PBD::DEBUG::Note = PBD::new_debug_bit ("note");
PBD::DebugBits PBD::DEBUG::ControlList = PBD::new_debug_bit ("controllist");

