#ifndef _ATKMM_WRAP_INIT_H
#define _ATKMM_WRAP_INIT_H

/* wrap_init.h
 *
 * Copyright (C) 1998-2001 The Gtk-- Development Team
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

// wrap_init.cc is generated by tools/generate_wrap_init.pl

namespace Atk
{

void wrap_init();

} /* namespace Atk */

#endif // _ATKMM_WRAP_INIT_H
