/* always defined to indicate that i18n is enabled */
#undef ENABLE_NLS

/* Define the gettext package to be used */
#define GETTEXT_PACKAGE "atk10"

/* Define to 1 if you have the `bind_textdomain_codeset' function. */
#define HAVE_BIND_TEXTDOMAIN_CODESET 1

/* Version number of package */
#define VERSION "2.14.0"

/* defines how to decorate public symbols while building */
#ifdef _MSC_VER
#define _ATK_EXTERN __declspec (dllexport) extern
#endif
