@import "ardour-toolkit-theme/theme.css";

html {
    height: 100%;
}

body {
    background: rgb(62,61,61);
    color: rgb(248,248,242);
    font-family: Helvetica, Arial, sans-serif !important;
    height: 100%;
    width: 100%;
    position: fixed;
    margin: 0;
    overflow: none;
}

div {
    box-sizing: border-box;
}

#root {
    display: flex;
    flex-direction: column;
    width: 100%;
    height: 100%;
    /* keep these for dialogs */
    align-items: center;
    justify-content: center;
}

#fullscreen {
    position: fixed;
    z-index: 10;
    top: 16px;
    right: 12px;
    width: 24px;
    height: 24px;
    opacity: 0.5;
    outline: none;
    border: none;
    background: none;
    background-size: contain;
    background-repeat: no-repeat;
    background-position: center;
    background-image: url('expand.svg');
}

#mixer {
    display: flex;
    flex-direction: row;
    height: 100%;
    width: 100%;
    padding: 10vh 0;
    overflow-x: auto;
    overflow-y: hidden;
    justify-content: space-between;
}

.strip {
    display: flex;
    flex-direction: column;
    align-items: center;
    min-width: 200px;
    max-height: 768px;
}

.strip-meter-fader {
    display: flex;
    flex: 1;
    min-height: 300px;
}

.strip-label {
    padding: 1em 0;
}

.strip-mute {
    background: rgb(62,61,61);
    border: 1px solid rgb(36,36,36);
    width: 50%;
}

.strip-mute.toolkit-active {
    background: rgb(172,155,5);
}

.strip-mute > .toolkit-label {
    color: rgb(248,248,242);
    background-color: inherit;
}

.strip-mute.toolkit-active > .toolkit-label {
    color: rgb(36,36,36);
    background-color: inherit;
}

.strip-plugins {
    color: rgb(248,248,242);
    border: 1px solid rgb(36,36,36);
    width: 50%;
}

.strip-plugins.disabled {
    color: rgb(36,36,36);
}

#plugins-dialog {
    width: 60%;
    min-width: 666px;
    height: 80%;
    min-height: 500px;
    background: rgb(52,51,51);
    border: none;
    /* absolute positioning cannot be disabled for tk dialogs */
    left: inherit !important;
    top: inherit !important;
}

#plugins {
    width: 100%;
    height: 100%;
    padding: 1em;
    overflow-x: hidden;
    overflow-y: auto;
}

#plugins-title {
    width: 100%;
    text-align: center;
    font-size: 1.25em;
    margin-bottom: 1em;
}

#plugins-close {
    position: fixed;
    top: 0;
    right: 0;
    padding: 5px;
    opacity: 0.5;
    z-index: 10;
}

.plugin-parameters {
    display: flex;
    flex-direction: row;
    flex-wrap: wrap;
    justify-content: center;
    width: 100%;
    margin-bottom: 1em;
}

.plugin-parameters:not(:last-child) {
    border-bottom: solid 1px rgba(255,255,255,0.1);
}

.plugin-parameter {
    display: flex;
    flex-direction: column;
    align-items: center;
    justify-content: center;
    margin: 1em;
}

.plugin-parameter-widget {
    display: flex;
    flex-direction: column;
    flex-grow: 1;
    justify-content: center;
}
