// This is -*- C++ -*-

/*
 * gdkmm.h
 *
 * Copyright 2002 The gtkmm Development Team
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef _GDKMM_GDKMM_H_
#define _GDKMM_GDKMM_H_

#include <ydkmm/screen.h>
#include <ydkmm/types.h>
#include <ydkmm/visual.h>
#include <ydkmm/gc.h>
#include <ydkmm/drawable.h>
#include <ydkmm/window.h>
#include <ydkmm/pixbuf.h>
#include <ydkmm/pixbufanimation.h>
#include <ydkmm/pixbufformat.h>
#include <ydkmm/pixbufloader.h>
#include <ydkmm/pixmap.h>
#include <ydkmm/bitmap.h>
#include <ydkmm/image.h>
#include <ydkmm/cursor.h>
#include <ydkmm/rectangle.h>
#include <ydkmm/display.h>
#include <ydkmm/displaymanager.h>
#include <ydkmm/general.h>
#include <ydkmm/rgb.h>

#endif /* _GDKMM_GDKMM_H_ */
