.toolkit-notification {
  background: rgba(238, 238, 238, 0.9);
  border: 1px solid rgba(255, 255, 255, 0.9);
  margin: 4px;
  padding: 4px 8px;
  min-height: 32px;
  
  font-size: 1rem;
  text-align: left;
  
  -webkit-box-shadow: 0px 2px 9px 0px rgba(0, 0, 0, 0.25);
  -moz-box-shadow:    0px 2px 9px 0px rgba(0, 0, 0, 0.25);
  box-shadow:         0px 2px 9px 0px rgba(0, 0, 0, 0.25);
  
  -webkit-border-radius: 6px;
  -moz-border-radius:    6px;
  border-radius:         6px;
  
  display: block;
}

.toolkit-notification > .toolkit-button {
  background: transparent !important;
  border: none !important;
  position: absolute;
  top: 4px;
  right: 4px;
  width: 20px !important;
  height: 20px !important;
  margin: 0;
  padding: 0;
  box-shadow: none !important;
}
.toolkit-notification > .toolkit-button.toolkit-close .toolkit-icon {
  font-size: 20px !important;
  line-height: 20px !important;
  width: 20px !important;
  height: 20px !important;
}

.toolkit-notification > .toolkit-icon {
  position: absolute;
  font-size: 32px;
  line-height: 32px;
  color: #002A42;
  top: 8px;
  left: 8px;
}

.toolkit-notification.toolkit-has-icon {
  padding-left: 56px;
}

.toolkit-notification.toolkit-has-close {
  padding-right: 48px;
}


.toolkit-notification {
  opacity: 0;
  
  -webkit-animation: -webkit-fade-in 0.5s;
  -moz-animation: -moz-fade-in 0.5s;
  -ms-animation: -ms-fade-in 0.5s;
  -o-animation: -o-fade-in 0.5s;
  animation: fade-in 0.5s;
  
  -webkit-transform: scale(1);
  -moz-transform: scale(1);
  -ms-transform: scale(1);
  -o-transform: scale(1);
  transform: scale(1);
}

.toolkit-notification.toolkit-show {
  opacity: 1;
}
.toolkit-notification.toolkit-hiding {
  opacity: 0;
  
  -webkit-transition: opacity 0.5s, transform 0.5s;
  -moz-transition: opacity 0.5s, transform 0.5s;
  -ms-transition: opacity 0.5s, transform 0.5s;
  -o-transition: opacity 0.5s, transform 0.5s;
  transition: opacity 0.5s, transform 0.5s;
  
  -webkit-transform: scale(0);
  -moz-transform: scale(0);
  -ms-transform: scale(0);
  -o-transform: scale(0);
  transform: scale(0);
}

@-webkit-keyframes -webit-fade-in {
  0% { opacity: 0; -webit-transform: scale(1); }
  100% { opacity: 1; -webit-transform: scale(1);}
}
@-moz-keyframes -moz-fade-in {
  0% { opacity: 0; -moz-transform: scale(1); }
  100% { opacity: 1; -moz-transform: scale(1);}
}
@-ms-keyframes -ms-fade-in {
  0% { opacity: 0; -ms-transform: scale(1); }
  100% { opacity: 1; -ms-transform: scale(1);}
}
@-o-keyframes -o-fade-in {
  0% { opacity: 0; -o-transform: scale(1); }
  100% { opacity: 1; -o-transform: scale(1);}
}
@keyframes fade-in {
  0% { opacity: 0; transform: scale(1); }
  100% { opacity: 1; transform: scale(1);}
}
