#include "ardour/ardour.h"
namespace ARDOUR {
const char* const ardour_config_info = "\n\
Build documentation: False\n\
Debuggable build: True\n\
Export all symbols (backtrace): False\n\
Install prefix: /usr/local\n\
Strict compiler flags: []\n\
Internal Shared Libraries: True\n\
Use External Libraries: False\n\
Library exports hidden: True\n\
Free/Demo copy: False\n\
\n\
ALSA DBus Reservation: True\n\
Architecture flags: None\n\
ARM NEON support: False\n\
Aubio: True\n\
AudioUnits: False\n\
Build target: x86_64\n\
Canvas Test UI: False\n\
Beatbox test app: False\n\
CoreAudio: False\n\
Debug RT allocations: False\n\
Debug Symbols: True\n\
Denormal exceptions: False\n\
Dr. Mingw: False\n\
FLAC: True\n\
FPU optimization: True\n\
FPU AVX512F support: True\n\
FPU AVX/FMA support: True\n\
Futex Semaphore: True\n\
Freedesktop files: False\n\
G_ENABLE_DEBUG: True\n\
I/O Priority Set: True\n\
Libjack linking: link\n\
Libjack metadata: True\n\
Lua Binding Doc: False\n\
Lua Commandline Tool: True\n\
LV2 UI embedding: True\n\
LV2 support: True\n\
LV2 extensions: True\n\
LXVST support: True\n\
Mac VST support: False\n\
NI-Maschine: False\n\
OGG: True\n\
Phone home: True\n\
Program name: Ardour\n\
Samplerate: True\n\
PT format: False\n\
PTW32 Semaphore: False\n\
Threaded WaveViews: True\n\
Translation: True\n\
Unit tests: False\n\
Use LLD linker: False\n\
VST3 support: True\n\
Windows VST support: False\n\
Wiimote support: True\n\
Windows key: Mod4><Super\n\
\n\
PortAudio Backend: False\n\
CoreAudio/Midi Backend: False\n\
ALSA Backend: True\n\
Dummy backend: True\n\
JACK Backend: True\n\
PulseAudio Backend: True\n\
\n\
Buildstack: -system-\n\
Mac i386 Architecture: False\n\
Mac ppc Architecture: False\n\
Mac arm64 Architecture: False\n\
\n\
C compiler flags: ['-I/tmp/a/ardour', '-DHAVE_RF64_RIFF', '-DCOMPILER_INT128_SUPPORT', '-DWAF_BUILD', '-g', '-fshow-column', '-DG_ENABLE_DEBUG', '-DARCH_X86', '-mmmx', '-msse', '-mfpmath=sse', '-DUSE_XMMINTRIN', '-DBUILD_SSE_OPTIMIZATIONS', '-DLXVST_64BIT', '-Wall', '-Wpointer-arith', '-Wcast-qual', '-Wcast-align', '-Wno-unused-parameter', '-DBOOST_SYSTEM_NO_DEPRECATED', '-D_ISOC9X_SOURCE', '-D_LARGEFILE64_SOURCE', '-D_FILE_OFFSET_BITS=64', '-DPROGRAM_NAME=\"Ardour\"', '-DPROGRAM_VERSION=\"9\"', '-Wstrict-prototypes', '-Wmissing-prototypes', '-D_POSIX_C_SOURCE=200809L']\n\
C++ compiler flags: ['-I/tmp/a/ardour', '-DHAVE_RF64_RIFF', '-DCOMPILER_INT128_SUPPORT', '-DWAF_BUILD', '-g', '-fshow-column', '-DG_ENABLE_DEBUG', '-DARCH_X86', '-mmmx', '-msse', '-mfpmath=sse', '-DUSE_XMMINTRIN', '-DBUILD_SSE_OPTIMIZATIONS', '-DLXVST_64BIT', '-Wall', '-Wpointer-arith', '-Wcast-qual', '-Wcast-align', '-Wno-unused-parameter', '-DBOOST_SYSTEM_NO_DEPRECATED', '-D_ISOC9X_SOURCE', '-D_LARGEFILE64_SOURCE', '-D_FILE_OFFSET_BITS=64', '-DPROGRAM_NAME=\"Ardour\"', '-DPROGRAM_VERSION=\"9\"', '-Woverloaded-virtual', '-Wno-unused-local-typedefs', '-Wno-deprecated-copy', '-D__STDC_LIMIT_MACROS', '-D__STDC_FORMAT_MACROS', '-DCANVAS_DEBUG', '-DBOOST_ERROR_CODE_HEADER_ONLY']\n\
Linker flags: ['']\n\
";
}
