/*
 * Copyright (C) 2007-2011 David Robillard <d@drobilla.net>
 * Copyright (C) 2007-2015 Paul Davis <paul@linuxaudiosystems.com>
 * Copyright (C) 2007 Doug McLain <doug@nostar.net>
 * Copyright (C) 2008-2009 Hans Baier <hansfbaier@googlemail.com>
 * Copyright (C) 2010-2012 Carl Hetherington <carl@carlh.net>
 * Copyright (C) 2012 Ben Loftis <ben@harrisonconsoles.com>
 * Copyright (C) 2013-2014 Robin Gareus <robin@gareus.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

CANVAS_FONT_VARIABLE(VerySmallFont, "very small font")
CANVAS_FONT_VARIABLE(SmallerFont, "smaller font")
CANVAS_FONT_VARIABLE(SmallFont, "small font")
CANVAS_FONT_VARIABLE(NormalFont, "normal font")
CANVAS_FONT_VARIABLE(BigFont, "big font")
CANVAS_FONT_VARIABLE(LargeFont, "large font")
CANVAS_FONT_VARIABLE(LargerFont, "larger font")
CANVAS_FONT_VARIABLE(HugerFont, "huger font")
CANVAS_FONT_VARIABLE(MassiveFont, "massive font")
CANVAS_FONT_VARIABLE(VerySmallBoldFont, "very small bold font")
CANVAS_FONT_VARIABLE(SmallBoldFont, "small bold font")
CANVAS_FONT_VARIABLE(SmallerBoldFont, "smaller bold font")
CANVAS_FONT_VARIABLE(NormalBoldFont, "normal bold font")
CANVAS_FONT_VARIABLE(BigBoldFont, "big bold font")
CANVAS_FONT_VARIABLE(LargeBoldFont, "large bold font")
CANVAS_FONT_VARIABLE(LargerBoldFont, "larger bold font")
CANVAS_FONT_VARIABLE(HugerBoldFont, "huger bold font")
CANVAS_FONT_VARIABLE(MassiveBoldFont, "massive bold font")
CANVAS_FONT_VARIABLE(SmallItalicFont, "small italic font")
CANVAS_FONT_VARIABLE(SmallerItalicFont, "smaller italic font")
CANVAS_FONT_VARIABLE(NormalItalicFont, "normal italic font")
CANVAS_FONT_VARIABLE(BigItalicFont, "big italic font")
CANVAS_FONT_VARIABLE(LargeItalicFont, "large italic font")
CANVAS_FONT_VARIABLE(LargerItalicFont, "larger italic font")
CANVAS_FONT_VARIABLE(HugerItalicFont, "huger italic font")
CANVAS_FONT_VARIABLE(MassiveItalicFont, "massive italic font")
CANVAS_FONT_VARIABLE(SmallMonospaceFont, "small monospace font")
CANVAS_FONT_VARIABLE(SmallerMonospaceFont, "smaller monospace font")
CANVAS_FONT_VARIABLE(NormalMonospaceFont, "normal monospace font")
CANVAS_FONT_VARIABLE(BigMonospaceFont, "big monospace font")
CANVAS_FONT_VARIABLE(LargeMonospaceFont, "large monospace font")
CANVAS_FONT_VARIABLE(LargerMonospaceFont, "larger monospace font")
CANVAS_FONT_VARIABLE(HugerMonospaceFont, "huger monospace font")
CANVAS_FONT_VARIABLE(MassiveMonospaceFont, "massive monospace font")
CANVAS_FONT_VARIABLE(SmallBoldMonospaceFont, "small bold monospace font")
CANVAS_FONT_VARIABLE(SmallerBoldMonospaceFont, "smaller bold monospace font")
CANVAS_FONT_VARIABLE(NormalBoldMonospaceFont, "normal bold monospace font")
CANVAS_FONT_VARIABLE(BigBoldMonospaceFont, "big bold monospace font")
CANVAS_FONT_VARIABLE(LargeBoldMonospaceFont, "large bold monospace font")
CANVAS_FONT_VARIABLE(LargerBoldMonospaceFont, "larger bold monospace font")
CANVAS_FONT_VARIABLE(HugerBoldMonospaceFont, "huger bold monospace font")
CANVAS_FONT_VARIABLE(MassiveBoldMonospaceFont, "massive bold monospace font")

