/*
 * Copyright (C) 2016-2017 Paul Davis <paul@linuxaudiosystems.com>
 * Copyright (C) 2017 Robin Gareus <robin@gareus.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#pragma once

#include <memory>
#include <set>
#include <string>
#include <stdint.h>

#include "pbd/rwlock.h"
#include "pbd/signals.h"

#include "evoral/Parameter.h"

#include "ardour/types.h"
#include "ardour/libardour_visibility.h"

class XMLNode;

namespace ARDOUR {

class VCA;
class VCAManager;
class AutomationControl;
class SlavableAutomationControl;

class LIBARDOUR_API Slavable
{
public:
	Slavable ();
	virtual ~Slavable() {}

	XMLNode& get_state () const;
	int set_state (XMLNode const&, int);

	virtual void assign (std::shared_ptr<VCA>);
	void unassign (std::shared_ptr<VCA>);

	PBD::Signal<void(std::shared_ptr<VCA>,bool)> AssignmentChange;

	virtual std::shared_ptr<AutomationControl> automation_control (const Evoral::Parameter&) = 0;

	static std::string xml_node_name;

	/* signal sent VCAManager once assignment is possible */
	static PBD::Signal<void(VCAManager*)> Assign;

	std::vector<std::shared_ptr<VCA> > masters (VCAManager*) const;

	/** recursively test for master assignment to given VCA */
	bool assigned_to (VCAManager*, std::shared_ptr<VCA>) const;

protected:
	virtual SlavableAutomationControlList slavables () const = 0;

private:
	bool assign_controls (std::shared_ptr<VCA>);
	void unassign_controls (std::shared_ptr<VCA>);

	bool assign_control (std::shared_ptr<VCA>, std::shared_ptr<SlavableAutomationControl>);
	void unassign_control (std::shared_ptr<VCA>, std::shared_ptr<SlavableAutomationControl>);

	mutable PBD::RWLock master_lock;
	std::set<uint32_t> _masters;
	PBD::ScopedConnection assign_connection;
	PBD::ScopedConnectionList unassign_connections;

	int do_assign (VCAManager* s);
	void weak_unassign (std::weak_ptr<VCA>);
};

} // namespace ARDOUR

