//
//  AppDelegate.m
//  CrashReporter
//
//  Created by Taybin Rutkin on 6/10/09.
//  Copyright 2009 Penguin Sounds. All rights reserved.
//

#import "AppDelegate.h"


@implementation AppDelegate

- (IBAction)sendButton:(id)sender
{
    
}

@end
