/*
 * This file is part of Toolkit.
 *
 * Toolkit is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301  USA
 */
.toolkit-value > .toolkit-input {
/*
    border-top: 1px solid #ddd;
    border-bottom: 1px solid #fff;
    border-left: 1px solid #e8e8e8;
    border-right: 1px solid #f3f3f3;
*/
    border: none;
    color: #002f42;
    background: rgba(0, 0, 0, 0.05);
    font-size: 1.2em;
    font-family: inherit;
    line-height: 1.1em;
    padding: 0;
    margin: 0;
    outline: 0;
    text-align: center;
    overflow: hidden;
    width: 100%;
    cursor: pointer;
}

.toolkit-value.toolkit-active > .toolkit-input {
    background: #002a42;
    color: white;
/*
    border: 1px solid transparent;
*/
}
