/* Generated by makegtkalias.pl */

#ifndef DISABLE_VISIBILITY

#include <glib.h>

#ifdef G_HAVE_GNUC_VISIBILITY

#define IN_FILE(x) 1
#define IN_HEADER defined

#if IN_HEADER(__GTK_ABOUT_DIALOG_H__)
#if IN_FILE(__GTK_ABOUT_DIALOG_C__)
extern __typeof (gtk_show_about_dialog) IA__gtk_show_about_dialog __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_show_about_dialog IA__gtk_show_about_dialog

extern __typeof (gtk_about_dialog_get_artists) IA__gtk_about_dialog_get_artists __attribute((visibility("hidden")));
#define gtk_about_dialog_get_artists IA__gtk_about_dialog_get_artists

extern __typeof (gtk_about_dialog_get_authors) IA__gtk_about_dialog_get_authors __attribute((visibility("hidden")));
#define gtk_about_dialog_get_authors IA__gtk_about_dialog_get_authors

extern __typeof (gtk_about_dialog_get_comments) IA__gtk_about_dialog_get_comments __attribute((visibility("hidden")));
#define gtk_about_dialog_get_comments IA__gtk_about_dialog_get_comments

extern __typeof (gtk_about_dialog_get_copyright) IA__gtk_about_dialog_get_copyright __attribute((visibility("hidden")));
#define gtk_about_dialog_get_copyright IA__gtk_about_dialog_get_copyright

extern __typeof (gtk_about_dialog_get_documenters) IA__gtk_about_dialog_get_documenters __attribute((visibility("hidden")));
#define gtk_about_dialog_get_documenters IA__gtk_about_dialog_get_documenters

extern __typeof (gtk_about_dialog_get_license) IA__gtk_about_dialog_get_license __attribute((visibility("hidden")));
#define gtk_about_dialog_get_license IA__gtk_about_dialog_get_license

extern __typeof (gtk_about_dialog_get_logo) IA__gtk_about_dialog_get_logo __attribute((visibility("hidden")));
#define gtk_about_dialog_get_logo IA__gtk_about_dialog_get_logo

extern __typeof (gtk_about_dialog_get_logo_icon_name) IA__gtk_about_dialog_get_logo_icon_name __attribute((visibility("hidden")));
#define gtk_about_dialog_get_logo_icon_name IA__gtk_about_dialog_get_logo_icon_name

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_about_dialog_get_name) IA__gtk_about_dialog_get_name __attribute((visibility("hidden")));
#define gtk_about_dialog_get_name IA__gtk_about_dialog_get_name

extern __typeof (gtk_about_dialog_set_name) IA__gtk_about_dialog_set_name __attribute((visibility("hidden")));
#define gtk_about_dialog_set_name IA__gtk_about_dialog_set_name

#endif
extern __typeof (gtk_about_dialog_get_program_name) IA__gtk_about_dialog_get_program_name __attribute((visibility("hidden")));
#define gtk_about_dialog_get_program_name IA__gtk_about_dialog_get_program_name

extern __typeof (gtk_about_dialog_set_program_name) IA__gtk_about_dialog_set_program_name __attribute((visibility("hidden")));
#define gtk_about_dialog_set_program_name IA__gtk_about_dialog_set_program_name

extern __typeof (gtk_about_dialog_get_translator_credits) IA__gtk_about_dialog_get_translator_credits __attribute((visibility("hidden")));
#define gtk_about_dialog_get_translator_credits IA__gtk_about_dialog_get_translator_credits

extern __typeof (gtk_about_dialog_get_type) IA__gtk_about_dialog_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_about_dialog_get_type IA__gtk_about_dialog_get_type

extern __typeof (gtk_about_dialog_get_version) IA__gtk_about_dialog_get_version __attribute((visibility("hidden")));
#define gtk_about_dialog_get_version IA__gtk_about_dialog_get_version

extern __typeof (gtk_about_dialog_get_website) IA__gtk_about_dialog_get_website __attribute((visibility("hidden")));
#define gtk_about_dialog_get_website IA__gtk_about_dialog_get_website

extern __typeof (gtk_about_dialog_get_website_label) IA__gtk_about_dialog_get_website_label __attribute((visibility("hidden")));
#define gtk_about_dialog_get_website_label IA__gtk_about_dialog_get_website_label

extern __typeof (gtk_about_dialog_get_wrap_license) IA__gtk_about_dialog_get_wrap_license __attribute((visibility("hidden")));
#define gtk_about_dialog_get_wrap_license IA__gtk_about_dialog_get_wrap_license

extern __typeof (gtk_about_dialog_new) IA__gtk_about_dialog_new __attribute((visibility("hidden")));
#define gtk_about_dialog_new IA__gtk_about_dialog_new

extern __typeof (gtk_about_dialog_set_artists) IA__gtk_about_dialog_set_artists __attribute((visibility("hidden")));
#define gtk_about_dialog_set_artists IA__gtk_about_dialog_set_artists

extern __typeof (gtk_about_dialog_set_authors) IA__gtk_about_dialog_set_authors __attribute((visibility("hidden")));
#define gtk_about_dialog_set_authors IA__gtk_about_dialog_set_authors

extern __typeof (gtk_about_dialog_set_comments) IA__gtk_about_dialog_set_comments __attribute((visibility("hidden")));
#define gtk_about_dialog_set_comments IA__gtk_about_dialog_set_comments

extern __typeof (gtk_about_dialog_set_copyright) IA__gtk_about_dialog_set_copyright __attribute((visibility("hidden")));
#define gtk_about_dialog_set_copyright IA__gtk_about_dialog_set_copyright

extern __typeof (gtk_about_dialog_set_documenters) IA__gtk_about_dialog_set_documenters __attribute((visibility("hidden")));
#define gtk_about_dialog_set_documenters IA__gtk_about_dialog_set_documenters

extern __typeof (gtk_about_dialog_set_license) IA__gtk_about_dialog_set_license __attribute((visibility("hidden")));
#define gtk_about_dialog_set_license IA__gtk_about_dialog_set_license

extern __typeof (gtk_about_dialog_set_logo) IA__gtk_about_dialog_set_logo __attribute((visibility("hidden")));
#define gtk_about_dialog_set_logo IA__gtk_about_dialog_set_logo

extern __typeof (gtk_about_dialog_set_logo_icon_name) IA__gtk_about_dialog_set_logo_icon_name __attribute((visibility("hidden")));
#define gtk_about_dialog_set_logo_icon_name IA__gtk_about_dialog_set_logo_icon_name

extern __typeof (gtk_about_dialog_set_translator_credits) IA__gtk_about_dialog_set_translator_credits __attribute((visibility("hidden")));
#define gtk_about_dialog_set_translator_credits IA__gtk_about_dialog_set_translator_credits

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_about_dialog_set_email_hook) IA__gtk_about_dialog_set_email_hook __attribute((visibility("hidden")));
#define gtk_about_dialog_set_email_hook IA__gtk_about_dialog_set_email_hook

extern __typeof (gtk_about_dialog_set_url_hook) IA__gtk_about_dialog_set_url_hook __attribute((visibility("hidden")));
#define gtk_about_dialog_set_url_hook IA__gtk_about_dialog_set_url_hook

#endif
extern __typeof (gtk_about_dialog_set_version) IA__gtk_about_dialog_set_version __attribute((visibility("hidden")));
#define gtk_about_dialog_set_version IA__gtk_about_dialog_set_version

extern __typeof (gtk_about_dialog_set_website) IA__gtk_about_dialog_set_website __attribute((visibility("hidden")));
#define gtk_about_dialog_set_website IA__gtk_about_dialog_set_website

extern __typeof (gtk_about_dialog_set_website_label) IA__gtk_about_dialog_set_website_label __attribute((visibility("hidden")));
#define gtk_about_dialog_set_website_label IA__gtk_about_dialog_set_website_label

extern __typeof (gtk_about_dialog_set_wrap_license) IA__gtk_about_dialog_set_wrap_license __attribute((visibility("hidden")));
#define gtk_about_dialog_set_wrap_license IA__gtk_about_dialog_set_wrap_license

#endif
#endif
#if IN_HEADER(__GTK_ACCEL_GROUP_H__)
#if IN_FILE(__GTK_ACCEL_GROUP_C__)
extern __typeof (gtk_accelerator_get_default_mod_mask) IA__gtk_accelerator_get_default_mod_mask __attribute((visibility("hidden")));
#define gtk_accelerator_get_default_mod_mask IA__gtk_accelerator_get_default_mod_mask

extern __typeof (gtk_accelerator_get_label) IA__gtk_accelerator_get_label __attribute((visibility("hidden")));
#define gtk_accelerator_get_label IA__gtk_accelerator_get_label

extern __typeof (gtk_accelerator_name) IA__gtk_accelerator_name __attribute((visibility("hidden")));
#define gtk_accelerator_name IA__gtk_accelerator_name

extern __typeof (gtk_accelerator_parse) IA__gtk_accelerator_parse __attribute((visibility("hidden")));
#define gtk_accelerator_parse IA__gtk_accelerator_parse

extern __typeof (gtk_accelerator_set_default_mod_mask) IA__gtk_accelerator_set_default_mod_mask __attribute((visibility("hidden")));
#define gtk_accelerator_set_default_mod_mask IA__gtk_accelerator_set_default_mod_mask

extern __typeof (gtk_accelerator_valid) IA__gtk_accelerator_valid __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_accelerator_valid IA__gtk_accelerator_valid

extern __typeof (gtk_accel_group_activate) IA__gtk_accel_group_activate __attribute((visibility("hidden")));
#define gtk_accel_group_activate IA__gtk_accel_group_activate

extern __typeof (gtk_accel_group_connect) IA__gtk_accel_group_connect __attribute((visibility("hidden")));
#define gtk_accel_group_connect IA__gtk_accel_group_connect

extern __typeof (gtk_accel_group_connect_by_path) IA__gtk_accel_group_connect_by_path __attribute((visibility("hidden")));
#define gtk_accel_group_connect_by_path IA__gtk_accel_group_connect_by_path

extern __typeof (gtk_accel_group_disconnect) IA__gtk_accel_group_disconnect __attribute((visibility("hidden")));
#define gtk_accel_group_disconnect IA__gtk_accel_group_disconnect

extern __typeof (gtk_accel_group_disconnect_key) IA__gtk_accel_group_disconnect_key __attribute((visibility("hidden")));
#define gtk_accel_group_disconnect_key IA__gtk_accel_group_disconnect_key

extern __typeof (gtk_accel_group_find) IA__gtk_accel_group_find __attribute((visibility("hidden")));
#define gtk_accel_group_find IA__gtk_accel_group_find

extern __typeof (gtk_accel_group_from_accel_closure) IA__gtk_accel_group_from_accel_closure __attribute((visibility("hidden")));
#define gtk_accel_group_from_accel_closure IA__gtk_accel_group_from_accel_closure

extern __typeof (gtk_accel_group_get_type) IA__gtk_accel_group_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_accel_group_get_type IA__gtk_accel_group_get_type

extern __typeof (gtk_accel_group_get_is_locked) IA__gtk_accel_group_get_is_locked __attribute((visibility("hidden")));
#define gtk_accel_group_get_is_locked IA__gtk_accel_group_get_is_locked

extern __typeof (gtk_accel_group_get_modifier_mask) IA__gtk_accel_group_get_modifier_mask __attribute((visibility("hidden")));
#define gtk_accel_group_get_modifier_mask IA__gtk_accel_group_get_modifier_mask

extern __typeof (gtk_accel_group_lock) IA__gtk_accel_group_lock __attribute((visibility("hidden")));
#define gtk_accel_group_lock IA__gtk_accel_group_lock

extern __typeof (gtk_accel_group_new) IA__gtk_accel_group_new __attribute((visibility("hidden")));
#define gtk_accel_group_new IA__gtk_accel_group_new

extern __typeof (gtk_accel_group_query) IA__gtk_accel_group_query __attribute((visibility("hidden")));
#define gtk_accel_group_query IA__gtk_accel_group_query

extern __typeof (gtk_accel_groups_activate) IA__gtk_accel_groups_activate __attribute((visibility("hidden")));
#define gtk_accel_groups_activate IA__gtk_accel_groups_activate

extern __typeof (gtk_accel_groups_from_object) IA__gtk_accel_groups_from_object __attribute((visibility("hidden")));
#define gtk_accel_groups_from_object IA__gtk_accel_groups_from_object

extern __typeof (gtk_accel_group_unlock) IA__gtk_accel_group_unlock __attribute((visibility("hidden")));
#define gtk_accel_group_unlock IA__gtk_accel_group_unlock

#endif
#endif
#if IN_HEADER(__GTK_ACCEL_LABEL_H__)
#if IN_FILE(__GTK_ACCEL_LABEL_C__)
extern __typeof (gtk_accel_label_get_accel_widget) IA__gtk_accel_label_get_accel_widget __attribute((visibility("hidden")));
#define gtk_accel_label_get_accel_widget IA__gtk_accel_label_get_accel_widget

extern __typeof (gtk_accel_label_get_accel_width) IA__gtk_accel_label_get_accel_width __attribute((visibility("hidden")));
#define gtk_accel_label_get_accel_width IA__gtk_accel_label_get_accel_width

extern __typeof (gtk_accel_label_get_type) IA__gtk_accel_label_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_accel_label_get_type IA__gtk_accel_label_get_type

extern __typeof (gtk_accel_label_new) IA__gtk_accel_label_new __attribute((visibility("hidden")));
#define gtk_accel_label_new IA__gtk_accel_label_new

extern __typeof (gtk_accel_label_refetch) IA__gtk_accel_label_refetch __attribute((visibility("hidden")));
#define gtk_accel_label_refetch IA__gtk_accel_label_refetch

extern __typeof (gtk_accel_label_set_accel_closure) IA__gtk_accel_label_set_accel_closure __attribute((visibility("hidden")));
#define gtk_accel_label_set_accel_closure IA__gtk_accel_label_set_accel_closure

extern __typeof (gtk_accel_label_set_accel_widget) IA__gtk_accel_label_set_accel_widget __attribute((visibility("hidden")));
#define gtk_accel_label_set_accel_widget IA__gtk_accel_label_set_accel_widget

#endif
#endif
#if IN_HEADER(__GTK_ACCEL_MAP_H__)
#if IN_FILE(__GTK_ACCEL_MAP_C__)
extern __typeof (gtk_accel_map_add_entry) IA__gtk_accel_map_add_entry __attribute((visibility("hidden")));
#define gtk_accel_map_add_entry IA__gtk_accel_map_add_entry

extern __typeof (gtk_accel_map_add_filter) IA__gtk_accel_map_add_filter __attribute((visibility("hidden")));
#define gtk_accel_map_add_filter IA__gtk_accel_map_add_filter

extern __typeof (gtk_accel_map_change_entry) IA__gtk_accel_map_change_entry __attribute((visibility("hidden")));
#define gtk_accel_map_change_entry IA__gtk_accel_map_change_entry

extern __typeof (gtk_accel_map_foreach) IA__gtk_accel_map_foreach __attribute((visibility("hidden")));
#define gtk_accel_map_foreach IA__gtk_accel_map_foreach

extern __typeof (gtk_accel_map_foreach_unfiltered) IA__gtk_accel_map_foreach_unfiltered __attribute((visibility("hidden")));
#define gtk_accel_map_foreach_unfiltered IA__gtk_accel_map_foreach_unfiltered

extern __typeof (gtk_accel_map_get) IA__gtk_accel_map_get __attribute((visibility("hidden")));
#define gtk_accel_map_get IA__gtk_accel_map_get

extern __typeof (gtk_accel_map_get_type) IA__gtk_accel_map_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_accel_map_get_type IA__gtk_accel_map_get_type

#ifndef _WIN64
extern __typeof (gtk_accel_map_load) IA__gtk_accel_map_load __attribute((visibility("hidden")));
#define gtk_accel_map_load IA__gtk_accel_map_load

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_accel_map_load_utf8) IA__gtk_accel_map_load_utf8 __attribute((visibility("hidden")));
#define gtk_accel_map_load_utf8 IA__gtk_accel_map_load_utf8

#endif
extern __typeof (gtk_accel_map_load_fd) IA__gtk_accel_map_load_fd __attribute((visibility("hidden")));
#define gtk_accel_map_load_fd IA__gtk_accel_map_load_fd

extern __typeof (gtk_accel_map_load_scanner) IA__gtk_accel_map_load_scanner __attribute((visibility("hidden")));
#define gtk_accel_map_load_scanner IA__gtk_accel_map_load_scanner

extern __typeof (gtk_accel_map_lock_path) IA__gtk_accel_map_lock_path __attribute((visibility("hidden")));
#define gtk_accel_map_lock_path IA__gtk_accel_map_lock_path

extern __typeof (gtk_accel_map_lookup_entry) IA__gtk_accel_map_lookup_entry __attribute((visibility("hidden")));
#define gtk_accel_map_lookup_entry IA__gtk_accel_map_lookup_entry

#ifndef _WIN64
extern __typeof (gtk_accel_map_save) IA__gtk_accel_map_save __attribute((visibility("hidden")));
#define gtk_accel_map_save IA__gtk_accel_map_save

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_accel_map_save_utf8) IA__gtk_accel_map_save_utf8 __attribute((visibility("hidden")));
#define gtk_accel_map_save_utf8 IA__gtk_accel_map_save_utf8

#endif
extern __typeof (gtk_accel_map_save_fd) IA__gtk_accel_map_save_fd __attribute((visibility("hidden")));
#define gtk_accel_map_save_fd IA__gtk_accel_map_save_fd

extern __typeof (gtk_accel_map_unlock_path) IA__gtk_accel_map_unlock_path __attribute((visibility("hidden")));
#define gtk_accel_map_unlock_path IA__gtk_accel_map_unlock_path

#endif
#endif
#if IN_HEADER(__GTK_ACCESSIBLE_H__)
#if IN_FILE(__GTK_ACCESSIBLE_C__)
extern __typeof (gtk_accessible_connect_widget_destroyed) IA__gtk_accessible_connect_widget_destroyed __attribute((visibility("hidden")));
#define gtk_accessible_connect_widget_destroyed IA__gtk_accessible_connect_widget_destroyed

extern __typeof (gtk_accessible_get_type) IA__gtk_accessible_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_accessible_get_type IA__gtk_accessible_get_type

extern __typeof (gtk_accessible_get_widget) IA__gtk_accessible_get_widget __attribute((visibility("hidden")));
#define gtk_accessible_get_widget IA__gtk_accessible_get_widget

extern __typeof (gtk_accessible_set_widget) IA__gtk_accessible_set_widget __attribute((visibility("hidden")));
#define gtk_accessible_set_widget IA__gtk_accessible_set_widget

#endif
#endif
#if IN_HEADER(__GTK_ACTION_H__)
#if IN_FILE(__GTK_ACTION_C__)
extern __typeof (gtk_action_activate) IA__gtk_action_activate __attribute((visibility("hidden")));
#define gtk_action_activate IA__gtk_action_activate

extern __typeof (gtk_action_block_activate) IA__gtk_action_block_activate __attribute((visibility("hidden")));
#define gtk_action_block_activate IA__gtk_action_block_activate

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_action_block_activate_from) IA__gtk_action_block_activate_from __attribute((visibility("hidden")));
#define gtk_action_block_activate_from IA__gtk_action_block_activate_from

#endif
extern __typeof (gtk_action_connect_accelerator) IA__gtk_action_connect_accelerator __attribute((visibility("hidden")));
#define gtk_action_connect_accelerator IA__gtk_action_connect_accelerator

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_action_connect_proxy) IA__gtk_action_connect_proxy __attribute((visibility("hidden")));
#define gtk_action_connect_proxy IA__gtk_action_connect_proxy

#endif
extern __typeof (gtk_action_create_icon) IA__gtk_action_create_icon __attribute((visibility("hidden")));
#define gtk_action_create_icon IA__gtk_action_create_icon

extern __typeof (gtk_action_create_menu_item) IA__gtk_action_create_menu_item __attribute((visibility("hidden")));
#define gtk_action_create_menu_item IA__gtk_action_create_menu_item

extern __typeof (gtk_action_create_tool_item) IA__gtk_action_create_tool_item __attribute((visibility("hidden")));
#define gtk_action_create_tool_item IA__gtk_action_create_tool_item

extern __typeof (gtk_action_create_menu) IA__gtk_action_create_menu __attribute((visibility("hidden")));
#define gtk_action_create_menu IA__gtk_action_create_menu

extern __typeof (gtk_action_disconnect_accelerator) IA__gtk_action_disconnect_accelerator __attribute((visibility("hidden")));
#define gtk_action_disconnect_accelerator IA__gtk_action_disconnect_accelerator

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_action_disconnect_proxy) IA__gtk_action_disconnect_proxy __attribute((visibility("hidden")));
#define gtk_action_disconnect_proxy IA__gtk_action_disconnect_proxy

#endif
extern __typeof (gtk_action_get_always_show_image) IA__gtk_action_get_always_show_image __attribute((visibility("hidden")));
#define gtk_action_get_always_show_image IA__gtk_action_get_always_show_image

extern __typeof (gtk_action_get_accel_closure) IA__gtk_action_get_accel_closure __attribute((visibility("hidden")));
#define gtk_action_get_accel_closure IA__gtk_action_get_accel_closure

extern __typeof (gtk_action_get_accel_path) IA__gtk_action_get_accel_path __attribute((visibility("hidden")));
#define gtk_action_get_accel_path IA__gtk_action_get_accel_path

extern __typeof (gtk_action_get_gicon) IA__gtk_action_get_gicon __attribute((visibility("hidden")));
#define gtk_action_get_gicon IA__gtk_action_get_gicon

extern __typeof (gtk_action_get_label) IA__gtk_action_get_label __attribute((visibility("hidden")));
#define gtk_action_get_label IA__gtk_action_get_label

extern __typeof (gtk_action_get_name) IA__gtk_action_get_name __attribute((visibility("hidden")));
#define gtk_action_get_name IA__gtk_action_get_name

extern __typeof (gtk_action_get_proxies) IA__gtk_action_get_proxies __attribute((visibility("hidden")));
#define gtk_action_get_proxies IA__gtk_action_get_proxies

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_widget_get_action) IA__gtk_widget_get_action __attribute((visibility("hidden")));
#define gtk_widget_get_action IA__gtk_widget_get_action

#endif
extern __typeof (gtk_action_get_icon_name) IA__gtk_action_get_icon_name __attribute((visibility("hidden")));
#define gtk_action_get_icon_name IA__gtk_action_get_icon_name

extern __typeof (gtk_action_get_is_important) IA__gtk_action_get_is_important __attribute((visibility("hidden")));
#define gtk_action_get_is_important IA__gtk_action_get_is_important

extern __typeof (gtk_action_get_sensitive) IA__gtk_action_get_sensitive __attribute((visibility("hidden")));
#define gtk_action_get_sensitive IA__gtk_action_get_sensitive

extern __typeof (gtk_action_get_short_label) IA__gtk_action_get_short_label __attribute((visibility("hidden")));
#define gtk_action_get_short_label IA__gtk_action_get_short_label

extern __typeof (gtk_action_get_stock_id) IA__gtk_action_get_stock_id __attribute((visibility("hidden")));
#define gtk_action_get_stock_id IA__gtk_action_get_stock_id

extern __typeof (gtk_action_get_tooltip) IA__gtk_action_get_tooltip __attribute((visibility("hidden")));
#define gtk_action_get_tooltip IA__gtk_action_get_tooltip

extern __typeof (gtk_action_get_type) IA__gtk_action_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_action_get_type IA__gtk_action_get_type

extern __typeof (gtk_action_get_visible) IA__gtk_action_get_visible __attribute((visibility("hidden")));
#define gtk_action_get_visible IA__gtk_action_get_visible

extern __typeof (gtk_action_get_visible_horizontal) IA__gtk_action_get_visible_horizontal __attribute((visibility("hidden")));
#define gtk_action_get_visible_horizontal IA__gtk_action_get_visible_horizontal

extern __typeof (gtk_action_get_visible_vertical) IA__gtk_action_get_visible_vertical __attribute((visibility("hidden")));
#define gtk_action_get_visible_vertical IA__gtk_action_get_visible_vertical

extern __typeof (gtk_action_is_sensitive) IA__gtk_action_is_sensitive __attribute((visibility("hidden")));
#define gtk_action_is_sensitive IA__gtk_action_is_sensitive

extern __typeof (gtk_action_is_visible) IA__gtk_action_is_visible __attribute((visibility("hidden")));
#define gtk_action_is_visible IA__gtk_action_is_visible

extern __typeof (gtk_action_new) IA__gtk_action_new __attribute((visibility("hidden")));
#define gtk_action_new IA__gtk_action_new

extern __typeof (gtk_action_set_accel_group) IA__gtk_action_set_accel_group __attribute((visibility("hidden")));
#define gtk_action_set_accel_group IA__gtk_action_set_accel_group

extern __typeof (gtk_action_set_accel_path) IA__gtk_action_set_accel_path __attribute((visibility("hidden")));
#define gtk_action_set_accel_path IA__gtk_action_set_accel_path

extern __typeof (gtk_action_set_always_show_image) IA__gtk_action_set_always_show_image __attribute((visibility("hidden")));
#define gtk_action_set_always_show_image IA__gtk_action_set_always_show_image

extern __typeof (gtk_action_set_gicon) IA__gtk_action_set_gicon __attribute((visibility("hidden")));
#define gtk_action_set_gicon IA__gtk_action_set_gicon

extern __typeof (gtk_action_set_icon_name) IA__gtk_action_set_icon_name __attribute((visibility("hidden")));
#define gtk_action_set_icon_name IA__gtk_action_set_icon_name

extern __typeof (gtk_action_set_is_important) IA__gtk_action_set_is_important __attribute((visibility("hidden")));
#define gtk_action_set_is_important IA__gtk_action_set_is_important

extern __typeof (gtk_action_set_label) IA__gtk_action_set_label __attribute((visibility("hidden")));
#define gtk_action_set_label IA__gtk_action_set_label

extern __typeof (gtk_action_set_sensitive) IA__gtk_action_set_sensitive __attribute((visibility("hidden")));
#define gtk_action_set_sensitive IA__gtk_action_set_sensitive

extern __typeof (gtk_action_set_short_label) IA__gtk_action_set_short_label __attribute((visibility("hidden")));
#define gtk_action_set_short_label IA__gtk_action_set_short_label

extern __typeof (gtk_action_set_stock_id) IA__gtk_action_set_stock_id __attribute((visibility("hidden")));
#define gtk_action_set_stock_id IA__gtk_action_set_stock_id

extern __typeof (gtk_action_set_tooltip) IA__gtk_action_set_tooltip __attribute((visibility("hidden")));
#define gtk_action_set_tooltip IA__gtk_action_set_tooltip

extern __typeof (gtk_action_set_visible) IA__gtk_action_set_visible __attribute((visibility("hidden")));
#define gtk_action_set_visible IA__gtk_action_set_visible

extern __typeof (gtk_action_set_visible_horizontal) IA__gtk_action_set_visible_horizontal __attribute((visibility("hidden")));
#define gtk_action_set_visible_horizontal IA__gtk_action_set_visible_horizontal

extern __typeof (gtk_action_set_visible_vertical) IA__gtk_action_set_visible_vertical __attribute((visibility("hidden")));
#define gtk_action_set_visible_vertical IA__gtk_action_set_visible_vertical

extern __typeof (gtk_action_unblock_activate) IA__gtk_action_unblock_activate __attribute((visibility("hidden")));
#define gtk_action_unblock_activate IA__gtk_action_unblock_activate

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_action_unblock_activate_from) IA__gtk_action_unblock_activate_from __attribute((visibility("hidden")));
#define gtk_action_unblock_activate_from IA__gtk_action_unblock_activate_from

#endif
#endif
#endif
#if IN_HEADER(__GTK_ACTION_GROUP_H__)
#if IN_FILE(__GTK_ACTION_GROUP_C__)
extern __typeof (gtk_action_group_add_action) IA__gtk_action_group_add_action __attribute((visibility("hidden")));
#define gtk_action_group_add_action IA__gtk_action_group_add_action

extern __typeof (gtk_action_group_add_actions) IA__gtk_action_group_add_actions __attribute((visibility("hidden")));
#define gtk_action_group_add_actions IA__gtk_action_group_add_actions

extern __typeof (gtk_action_group_add_actions_full) IA__gtk_action_group_add_actions_full __attribute((visibility("hidden")));
#define gtk_action_group_add_actions_full IA__gtk_action_group_add_actions_full

extern __typeof (gtk_action_group_add_action_with_accel) IA__gtk_action_group_add_action_with_accel __attribute((visibility("hidden")));
#define gtk_action_group_add_action_with_accel IA__gtk_action_group_add_action_with_accel

extern __typeof (gtk_action_group_add_radio_actions) IA__gtk_action_group_add_radio_actions __attribute((visibility("hidden")));
#define gtk_action_group_add_radio_actions IA__gtk_action_group_add_radio_actions

extern __typeof (gtk_action_group_add_radio_actions_full) IA__gtk_action_group_add_radio_actions_full __attribute((visibility("hidden")));
#define gtk_action_group_add_radio_actions_full IA__gtk_action_group_add_radio_actions_full

extern __typeof (gtk_action_group_add_toggle_actions) IA__gtk_action_group_add_toggle_actions __attribute((visibility("hidden")));
#define gtk_action_group_add_toggle_actions IA__gtk_action_group_add_toggle_actions

extern __typeof (gtk_action_group_add_toggle_actions_full) IA__gtk_action_group_add_toggle_actions_full __attribute((visibility("hidden")));
#define gtk_action_group_add_toggle_actions_full IA__gtk_action_group_add_toggle_actions_full

extern __typeof (gtk_action_group_get_action) IA__gtk_action_group_get_action __attribute((visibility("hidden")));
#define gtk_action_group_get_action IA__gtk_action_group_get_action

extern __typeof (gtk_action_group_get_name) IA__gtk_action_group_get_name __attribute((visibility("hidden")));
#define gtk_action_group_get_name IA__gtk_action_group_get_name

extern __typeof (gtk_action_group_get_sensitive) IA__gtk_action_group_get_sensitive __attribute((visibility("hidden")));
#define gtk_action_group_get_sensitive IA__gtk_action_group_get_sensitive

extern __typeof (gtk_action_group_get_type) IA__gtk_action_group_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_action_group_get_type IA__gtk_action_group_get_type

extern __typeof (gtk_action_group_get_visible) IA__gtk_action_group_get_visible __attribute((visibility("hidden")));
#define gtk_action_group_get_visible IA__gtk_action_group_get_visible

extern __typeof (gtk_action_group_list_actions) IA__gtk_action_group_list_actions __attribute((visibility("hidden")));
#define gtk_action_group_list_actions IA__gtk_action_group_list_actions

extern __typeof (gtk_action_group_new) IA__gtk_action_group_new __attribute((visibility("hidden")));
#define gtk_action_group_new IA__gtk_action_group_new

extern __typeof (gtk_action_group_remove_action) IA__gtk_action_group_remove_action __attribute((visibility("hidden")));
#define gtk_action_group_remove_action IA__gtk_action_group_remove_action

extern __typeof (gtk_action_group_set_sensitive) IA__gtk_action_group_set_sensitive __attribute((visibility("hidden")));
#define gtk_action_group_set_sensitive IA__gtk_action_group_set_sensitive

extern __typeof (gtk_action_group_set_translate_func) IA__gtk_action_group_set_translate_func __attribute((visibility("hidden")));
#define gtk_action_group_set_translate_func IA__gtk_action_group_set_translate_func

extern __typeof (gtk_action_group_set_translation_domain) IA__gtk_action_group_set_translation_domain __attribute((visibility("hidden")));
#define gtk_action_group_set_translation_domain IA__gtk_action_group_set_translation_domain

extern __typeof (gtk_action_group_set_visible) IA__gtk_action_group_set_visible __attribute((visibility("hidden")));
#define gtk_action_group_set_visible IA__gtk_action_group_set_visible

extern __typeof (gtk_action_group_translate_string) IA__gtk_action_group_translate_string __attribute((visibility("hidden")));
#define gtk_action_group_translate_string IA__gtk_action_group_translate_string

#endif
#endif
#if IN_HEADER(__GTK_ACTIVATABLE_H__)
#if IN_FILE(__GTK_ACTIVATABLE_C__)
extern __typeof (gtk_activatable_do_set_related_action) IA__gtk_activatable_do_set_related_action __attribute((visibility("hidden")));
#define gtk_activatable_do_set_related_action IA__gtk_activatable_do_set_related_action

extern __typeof (gtk_activatable_get_related_action) IA__gtk_activatable_get_related_action __attribute((visibility("hidden")));
#define gtk_activatable_get_related_action IA__gtk_activatable_get_related_action

extern __typeof (gtk_activatable_get_type) IA__gtk_activatable_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_activatable_get_type IA__gtk_activatable_get_type

extern __typeof (gtk_activatable_get_use_action_appearance) IA__gtk_activatable_get_use_action_appearance __attribute((visibility("hidden")));
#define gtk_activatable_get_use_action_appearance IA__gtk_activatable_get_use_action_appearance

extern __typeof (gtk_activatable_sync_action_properties) IA__gtk_activatable_sync_action_properties __attribute((visibility("hidden")));
#define gtk_activatable_sync_action_properties IA__gtk_activatable_sync_action_properties

extern __typeof (gtk_activatable_set_related_action) IA__gtk_activatable_set_related_action __attribute((visibility("hidden")));
#define gtk_activatable_set_related_action IA__gtk_activatable_set_related_action

extern __typeof (gtk_activatable_set_use_action_appearance) IA__gtk_activatable_set_use_action_appearance __attribute((visibility("hidden")));
#define gtk_activatable_set_use_action_appearance IA__gtk_activatable_set_use_action_appearance

#endif
#endif
#if IN_HEADER(__GTK_ADJUSTMENT_H__)
#if IN_FILE(__GTK_ADJUSTMENT_C__)
extern __typeof (gtk_adjustment_changed) IA__gtk_adjustment_changed __attribute((visibility("hidden")));
#define gtk_adjustment_changed IA__gtk_adjustment_changed

extern __typeof (gtk_adjustment_clamp_page) IA__gtk_adjustment_clamp_page __attribute((visibility("hidden")));
#define gtk_adjustment_clamp_page IA__gtk_adjustment_clamp_page

extern __typeof (gtk_adjustment_configure) IA__gtk_adjustment_configure __attribute((visibility("hidden")));
#define gtk_adjustment_configure IA__gtk_adjustment_configure

extern __typeof (gtk_adjustment_get_type) IA__gtk_adjustment_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_adjustment_get_type IA__gtk_adjustment_get_type

extern __typeof (gtk_adjustment_get_value) IA__gtk_adjustment_get_value __attribute((visibility("hidden")));
#define gtk_adjustment_get_value IA__gtk_adjustment_get_value

extern __typeof (gtk_adjustment_get_lower) IA__gtk_adjustment_get_lower __attribute((visibility("hidden")));
#define gtk_adjustment_get_lower IA__gtk_adjustment_get_lower

extern __typeof (gtk_adjustment_get_upper) IA__gtk_adjustment_get_upper __attribute((visibility("hidden")));
#define gtk_adjustment_get_upper IA__gtk_adjustment_get_upper

extern __typeof (gtk_adjustment_get_step_increment) IA__gtk_adjustment_get_step_increment __attribute((visibility("hidden")));
#define gtk_adjustment_get_step_increment IA__gtk_adjustment_get_step_increment

extern __typeof (gtk_adjustment_get_page_increment) IA__gtk_adjustment_get_page_increment __attribute((visibility("hidden")));
#define gtk_adjustment_get_page_increment IA__gtk_adjustment_get_page_increment

extern __typeof (gtk_adjustment_get_page_size) IA__gtk_adjustment_get_page_size __attribute((visibility("hidden")));
#define gtk_adjustment_get_page_size IA__gtk_adjustment_get_page_size

extern __typeof (gtk_adjustment_new) IA__gtk_adjustment_new __attribute((visibility("hidden")));
#define gtk_adjustment_new IA__gtk_adjustment_new

extern __typeof (gtk_adjustment_set_value) IA__gtk_adjustment_set_value __attribute((visibility("hidden")));
#define gtk_adjustment_set_value IA__gtk_adjustment_set_value

extern __typeof (gtk_adjustment_set_lower) IA__gtk_adjustment_set_lower __attribute((visibility("hidden")));
#define gtk_adjustment_set_lower IA__gtk_adjustment_set_lower

extern __typeof (gtk_adjustment_set_upper) IA__gtk_adjustment_set_upper __attribute((visibility("hidden")));
#define gtk_adjustment_set_upper IA__gtk_adjustment_set_upper

extern __typeof (gtk_adjustment_set_step_increment) IA__gtk_adjustment_set_step_increment __attribute((visibility("hidden")));
#define gtk_adjustment_set_step_increment IA__gtk_adjustment_set_step_increment

extern __typeof (gtk_adjustment_set_page_increment) IA__gtk_adjustment_set_page_increment __attribute((visibility("hidden")));
#define gtk_adjustment_set_page_increment IA__gtk_adjustment_set_page_increment

extern __typeof (gtk_adjustment_set_page_size) IA__gtk_adjustment_set_page_size __attribute((visibility("hidden")));
#define gtk_adjustment_set_page_size IA__gtk_adjustment_set_page_size

extern __typeof (gtk_adjustment_value_changed) IA__gtk_adjustment_value_changed __attribute((visibility("hidden")));
#define gtk_adjustment_value_changed IA__gtk_adjustment_value_changed

#endif
#endif
#if IN_HEADER(__GTK_ALIGNMENT_H__)
#if IN_FILE(__GTK_ALIGNMENT_C__)
extern __typeof (gtk_alignment_get_padding) IA__gtk_alignment_get_padding __attribute((visibility("hidden")));
#define gtk_alignment_get_padding IA__gtk_alignment_get_padding

extern __typeof (gtk_alignment_get_type) IA__gtk_alignment_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_alignment_get_type IA__gtk_alignment_get_type

extern __typeof (gtk_alignment_new) IA__gtk_alignment_new __attribute((visibility("hidden")));
#define gtk_alignment_new IA__gtk_alignment_new

extern __typeof (gtk_alignment_set) IA__gtk_alignment_set __attribute((visibility("hidden")));
#define gtk_alignment_set IA__gtk_alignment_set

extern __typeof (gtk_alignment_set_padding) IA__gtk_alignment_set_padding __attribute((visibility("hidden")));
#define gtk_alignment_set_padding IA__gtk_alignment_set_padding

#endif
#endif
#if IN_HEADER(__GTK_ASSISTANT_H__)
#if IN_FILE(__GTK_ASSISTANT_C__)
extern __typeof (gtk_assistant_get_type) IA__gtk_assistant_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_assistant_get_type IA__gtk_assistant_get_type

extern __typeof (gtk_assistant_new) IA__gtk_assistant_new __attribute((visibility("hidden")));
#define gtk_assistant_new IA__gtk_assistant_new

extern __typeof (gtk_assistant_get_current_page) IA__gtk_assistant_get_current_page __attribute((visibility("hidden")));
#define gtk_assistant_get_current_page IA__gtk_assistant_get_current_page

extern __typeof (gtk_assistant_set_current_page) IA__gtk_assistant_set_current_page __attribute((visibility("hidden")));
#define gtk_assistant_set_current_page IA__gtk_assistant_set_current_page

extern __typeof (gtk_assistant_get_n_pages) IA__gtk_assistant_get_n_pages __attribute((visibility("hidden")));
#define gtk_assistant_get_n_pages IA__gtk_assistant_get_n_pages

extern __typeof (gtk_assistant_get_nth_page) IA__gtk_assistant_get_nth_page __attribute((visibility("hidden")));
#define gtk_assistant_get_nth_page IA__gtk_assistant_get_nth_page

extern __typeof (gtk_assistant_prepend_page) IA__gtk_assistant_prepend_page __attribute((visibility("hidden")));
#define gtk_assistant_prepend_page IA__gtk_assistant_prepend_page

extern __typeof (gtk_assistant_append_page) IA__gtk_assistant_append_page __attribute((visibility("hidden")));
#define gtk_assistant_append_page IA__gtk_assistant_append_page

extern __typeof (gtk_assistant_insert_page) IA__gtk_assistant_insert_page __attribute((visibility("hidden")));
#define gtk_assistant_insert_page IA__gtk_assistant_insert_page

extern __typeof (gtk_assistant_set_forward_page_func) IA__gtk_assistant_set_forward_page_func __attribute((visibility("hidden")));
#define gtk_assistant_set_forward_page_func IA__gtk_assistant_set_forward_page_func

extern __typeof (gtk_assistant_set_page_type) IA__gtk_assistant_set_page_type __attribute((visibility("hidden")));
#define gtk_assistant_set_page_type IA__gtk_assistant_set_page_type

extern __typeof (gtk_assistant_get_page_type) IA__gtk_assistant_get_page_type __attribute((visibility("hidden")));
#define gtk_assistant_get_page_type IA__gtk_assistant_get_page_type

extern __typeof (gtk_assistant_set_page_title) IA__gtk_assistant_set_page_title __attribute((visibility("hidden")));
#define gtk_assistant_set_page_title IA__gtk_assistant_set_page_title

extern __typeof (gtk_assistant_get_page_title) IA__gtk_assistant_get_page_title __attribute((visibility("hidden")));
#define gtk_assistant_get_page_title IA__gtk_assistant_get_page_title

extern __typeof (gtk_assistant_set_page_header_image) IA__gtk_assistant_set_page_header_image __attribute((visibility("hidden")));
#define gtk_assistant_set_page_header_image IA__gtk_assistant_set_page_header_image

extern __typeof (gtk_assistant_get_page_header_image) IA__gtk_assistant_get_page_header_image __attribute((visibility("hidden")));
#define gtk_assistant_get_page_header_image IA__gtk_assistant_get_page_header_image

extern __typeof (gtk_assistant_set_page_side_image) IA__gtk_assistant_set_page_side_image __attribute((visibility("hidden")));
#define gtk_assistant_set_page_side_image IA__gtk_assistant_set_page_side_image

extern __typeof (gtk_assistant_get_page_side_image) IA__gtk_assistant_get_page_side_image __attribute((visibility("hidden")));
#define gtk_assistant_get_page_side_image IA__gtk_assistant_get_page_side_image

extern __typeof (gtk_assistant_set_page_complete) IA__gtk_assistant_set_page_complete __attribute((visibility("hidden")));
#define gtk_assistant_set_page_complete IA__gtk_assistant_set_page_complete

extern __typeof (gtk_assistant_get_page_complete) IA__gtk_assistant_get_page_complete __attribute((visibility("hidden")));
#define gtk_assistant_get_page_complete IA__gtk_assistant_get_page_complete

extern __typeof (gtk_assistant_add_action_widget) IA__gtk_assistant_add_action_widget __attribute((visibility("hidden")));
#define gtk_assistant_add_action_widget IA__gtk_assistant_add_action_widget

extern __typeof (gtk_assistant_remove_action_widget) IA__gtk_assistant_remove_action_widget __attribute((visibility("hidden")));
#define gtk_assistant_remove_action_widget IA__gtk_assistant_remove_action_widget

extern __typeof (gtk_assistant_update_buttons_state) IA__gtk_assistant_update_buttons_state __attribute((visibility("hidden")));
#define gtk_assistant_update_buttons_state IA__gtk_assistant_update_buttons_state

extern __typeof (gtk_assistant_commit) IA__gtk_assistant_commit __attribute((visibility("hidden")));
#define gtk_assistant_commit IA__gtk_assistant_commit

#endif
#endif
#if IN_HEADER(__GTK_TYPE_BUILTINS_H__)
#if IN_FILE(__GTK_TYPE_BUILTINS_C__)
extern __typeof (gtk_pack_direction_get_type) IA__gtk_pack_direction_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_pack_direction_get_type IA__gtk_pack_direction_get_type

extern __typeof (gtk_ui_manager_item_type_get_type) IA__gtk_ui_manager_item_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_ui_manager_item_type_get_type IA__gtk_ui_manager_item_type_get_type

extern __typeof (gtk_spin_button_update_policy_get_type) IA__gtk_spin_button_update_policy_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_spin_button_update_policy_get_type IA__gtk_spin_button_update_policy_get_type

extern __typeof (gtk_notebook_tab_get_type) IA__gtk_notebook_tab_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_notebook_tab_get_type IA__gtk_notebook_tab_get_type

extern __typeof (gtk_number_up_layout_get_type) IA__gtk_number_up_layout_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_number_up_layout_get_type IA__gtk_number_up_layout_get_type

extern __typeof (gtk_anchor_type_get_type) IA__gtk_anchor_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_anchor_type_get_type IA__gtk_anchor_type_get_type

extern __typeof (gtk_arg_flags_get_type) IA__gtk_arg_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_arg_flags_get_type IA__gtk_arg_flags_get_type

extern __typeof (gtk_arrow_type_get_type) IA__gtk_arrow_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_arrow_type_get_type IA__gtk_arrow_type_get_type

extern __typeof (gtk_arrow_placement_get_type) IA__gtk_arrow_placement_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_arrow_placement_get_type IA__gtk_arrow_placement_get_type

extern __typeof (gtk_assistant_page_type_get_type) IA__gtk_assistant_page_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_assistant_page_type_get_type IA__gtk_assistant_page_type_get_type

extern __typeof (gtk_attach_options_get_type) IA__gtk_attach_options_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_attach_options_get_type IA__gtk_attach_options_get_type

extern __typeof (gtk_builder_error_get_type) IA__gtk_builder_error_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_builder_error_get_type IA__gtk_builder_error_get_type

extern __typeof (gtk_buttons_type_get_type) IA__gtk_buttons_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_buttons_type_get_type IA__gtk_buttons_type_get_type

extern __typeof (gtk_cell_renderer_mode_get_type) IA__gtk_cell_renderer_mode_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_renderer_mode_get_type IA__gtk_cell_renderer_mode_get_type

extern __typeof (gtk_cell_renderer_state_get_type) IA__gtk_cell_renderer_state_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_renderer_state_get_type IA__gtk_cell_renderer_state_get_type

extern __typeof (gtk_cell_renderer_accel_mode_get_type) IA__gtk_cell_renderer_accel_mode_get_type __attribute((visibility("hidden")));
#define gtk_cell_renderer_accel_mode_get_type IA__gtk_cell_renderer_accel_mode_get_type

extern __typeof (gtk_corner_type_get_type) IA__gtk_corner_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_corner_type_get_type IA__gtk_corner_type_get_type

extern __typeof (gtk_debug_flag_get_type) IA__gtk_debug_flag_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_debug_flag_get_type IA__gtk_debug_flag_get_type

extern __typeof (gtk_delete_type_get_type) IA__gtk_delete_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_delete_type_get_type IA__gtk_delete_type_get_type

extern __typeof (gtk_dest_defaults_get_type) IA__gtk_dest_defaults_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_dest_defaults_get_type IA__gtk_dest_defaults_get_type

extern __typeof (gtk_drag_result_get_type) IA__gtk_drag_result_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_drag_result_get_type IA__gtk_drag_result_get_type

extern __typeof (gtk_dialog_flags_get_type) IA__gtk_dialog_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_dialog_flags_get_type IA__gtk_dialog_flags_get_type

extern __typeof (gtk_direction_type_get_type) IA__gtk_direction_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_direction_type_get_type IA__gtk_direction_type_get_type

extern __typeof (gtk_entry_icon_position_get_type) IA__gtk_entry_icon_position_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_entry_icon_position_get_type IA__gtk_entry_icon_position_get_type

extern __typeof (gtk_expander_style_get_type) IA__gtk_expander_style_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_expander_style_get_type IA__gtk_expander_style_get_type

extern __typeof (gtk_file_chooser_action_get_type) IA__gtk_file_chooser_action_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_file_chooser_action_get_type IA__gtk_file_chooser_action_get_type

extern __typeof (gtk_file_chooser_confirmation_get_type) IA__gtk_file_chooser_confirmation_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_file_chooser_confirmation_get_type IA__gtk_file_chooser_confirmation_get_type

extern __typeof (gtk_file_filter_flags_get_type) IA__gtk_file_filter_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_file_filter_flags_get_type IA__gtk_file_filter_flags_get_type

extern __typeof (gtk_icon_view_drop_position_get_type) IA__gtk_icon_view_drop_position_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_icon_view_drop_position_get_type IA__gtk_icon_view_drop_position_get_type

extern __typeof (gtk_sensitivity_type_get_type) IA__gtk_sensitivity_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_sensitivity_type_get_type IA__gtk_sensitivity_type_get_type

extern __typeof (gtk_window_type_get_type) IA__gtk_window_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_window_type_get_type IA__gtk_window_type_get_type

extern __typeof (gtk_wrap_mode_get_type) IA__gtk_wrap_mode_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_wrap_mode_get_type IA__gtk_wrap_mode_get_type

extern __typeof (gtk_widget_flags_get_type) IA__gtk_widget_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_widget_flags_get_type IA__gtk_widget_flags_get_type

extern __typeof (gtk_widget_help_type_get_type) IA__gtk_widget_help_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_widget_help_type_get_type IA__gtk_widget_help_type_get_type

extern __typeof (gtk_window_position_get_type) IA__gtk_window_position_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_window_position_get_type IA__gtk_window_position_get_type

extern __typeof (gtk_tree_view_column_sizing_get_type) IA__gtk_tree_view_column_sizing_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_view_column_sizing_get_type IA__gtk_tree_view_column_sizing_get_type

extern __typeof (gtk_tree_view_drop_position_get_type) IA__gtk_tree_view_drop_position_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_view_drop_position_get_type IA__gtk_tree_view_drop_position_get_type

extern __typeof (gtk_button_box_style_get_type) IA__gtk_button_box_style_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_button_box_style_get_type IA__gtk_button_box_style_get_type

extern __typeof (gtk_curve_type_get_type) IA__gtk_curve_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_curve_type_get_type IA__gtk_curve_type_get_type

extern __typeof (gtk_icon_lookup_flags_get_type) IA__gtk_icon_lookup_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_icon_lookup_flags_get_type IA__gtk_icon_lookup_flags_get_type

extern __typeof (gtk_image_type_get_type) IA__gtk_image_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_image_type_get_type IA__gtk_image_type_get_type

extern __typeof (gtk_im_preedit_style_get_type) IA__gtk_im_preedit_style_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_im_preedit_style_get_type IA__gtk_im_preedit_style_get_type

extern __typeof (gtk_im_status_style_get_type) IA__gtk_im_status_style_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_im_status_style_get_type IA__gtk_im_status_style_get_type

extern __typeof (gtk_justification_get_type) IA__gtk_justification_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_justification_get_type IA__gtk_justification_get_type

extern __typeof (gtk_match_type_get_type) IA__gtk_match_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_match_type_get_type IA__gtk_match_type_get_type

extern __typeof (gtk_menu_direction_type_get_type) IA__gtk_menu_direction_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_menu_direction_type_get_type IA__gtk_menu_direction_type_get_type

extern __typeof (gtk_message_type_get_type) IA__gtk_message_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_message_type_get_type IA__gtk_message_type_get_type

extern __typeof (gtk_metric_type_get_type) IA__gtk_metric_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_metric_type_get_type IA__gtk_metric_type_get_type

extern __typeof (gtk_movement_step_get_type) IA__gtk_movement_step_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_movement_step_get_type IA__gtk_movement_step_get_type

extern __typeof (gtk_orientation_get_type) IA__gtk_orientation_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_orientation_get_type IA__gtk_orientation_get_type

extern __typeof (gtk_pack_type_get_type) IA__gtk_pack_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_pack_type_get_type IA__gtk_pack_type_get_type

extern __typeof (gtk_page_orientation_get_type) IA__gtk_page_orientation_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_page_orientation_get_type IA__gtk_page_orientation_get_type

extern __typeof (gtk_print_pages_get_type) IA__gtk_print_pages_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_print_pages_get_type IA__gtk_print_pages_get_type

extern __typeof (gtk_print_quality_get_type) IA__gtk_print_quality_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_print_quality_get_type IA__gtk_print_quality_get_type

extern __typeof (gtk_page_set_get_type) IA__gtk_page_set_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_page_set_get_type IA__gtk_page_set_get_type

extern __typeof (gtk_path_priority_type_get_type) IA__gtk_path_priority_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_path_priority_type_get_type IA__gtk_path_priority_type_get_type

extern __typeof (gtk_path_type_get_type) IA__gtk_path_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_path_type_get_type IA__gtk_path_type_get_type

extern __typeof (gtk_policy_type_get_type) IA__gtk_policy_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_policy_type_get_type IA__gtk_policy_type_get_type

extern __typeof (gtk_position_type_get_type) IA__gtk_position_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_position_type_get_type IA__gtk_position_type_get_type

extern __typeof (gtk_preview_type_get_type) IA__gtk_preview_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_preview_type_get_type IA__gtk_preview_type_get_type

extern __typeof (gtk_print_duplex_get_type) IA__gtk_print_duplex_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_print_duplex_get_type IA__gtk_print_duplex_get_type

extern __typeof (gtk_recent_filter_flags_get_type) IA__gtk_recent_filter_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_filter_flags_get_type IA__gtk_recent_filter_flags_get_type

extern __typeof (gtk_private_flags_get_type) IA__gtk_private_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_private_flags_get_type IA__gtk_private_flags_get_type

extern __typeof (gtk_progress_bar_orientation_get_type) IA__gtk_progress_bar_orientation_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_progress_bar_orientation_get_type IA__gtk_progress_bar_orientation_get_type

extern __typeof (gtk_progress_bar_style_get_type) IA__gtk_progress_bar_style_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_progress_bar_style_get_type IA__gtk_progress_bar_style_get_type

extern __typeof (gtk_rc_flags_get_type) IA__gtk_rc_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_rc_flags_get_type IA__gtk_rc_flags_get_type

extern __typeof (gtk_rc_token_type_get_type) IA__gtk_rc_token_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_rc_token_type_get_type IA__gtk_rc_token_type_get_type

extern __typeof (gtk_recent_chooser_error_get_type) IA__gtk_recent_chooser_error_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_chooser_error_get_type IA__gtk_recent_chooser_error_get_type

extern __typeof (gtk_recent_manager_error_get_type) IA__gtk_recent_manager_error_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_manager_error_get_type IA__gtk_recent_manager_error_get_type

extern __typeof (gtk_recent_sort_type_get_type) IA__gtk_recent_sort_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_sort_type_get_type IA__gtk_recent_sort_type_get_type

extern __typeof (gtk_unit_get_type) IA__gtk_unit_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_unit_get_type IA__gtk_unit_get_type

extern __typeof (gtk_relief_style_get_type) IA__gtk_relief_style_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_relief_style_get_type IA__gtk_relief_style_get_type

extern __typeof (gtk_resize_mode_get_type) IA__gtk_resize_mode_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_resize_mode_get_type IA__gtk_resize_mode_get_type

extern __typeof (gtk_response_type_get_type) IA__gtk_response_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_response_type_get_type IA__gtk_response_type_get_type

extern __typeof (gtk_scroll_step_get_type) IA__gtk_scroll_step_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_scroll_step_get_type IA__gtk_scroll_step_get_type

extern __typeof (gtk_scroll_type_get_type) IA__gtk_scroll_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_scroll_type_get_type IA__gtk_scroll_type_get_type

extern __typeof (gtk_selection_mode_get_type) IA__gtk_selection_mode_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_selection_mode_get_type IA__gtk_selection_mode_get_type

extern __typeof (gtk_shadow_type_get_type) IA__gtk_shadow_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_shadow_type_get_type IA__gtk_shadow_type_get_type

extern __typeof (gtk_side_type_get_type) IA__gtk_side_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_side_type_get_type IA__gtk_side_type_get_type

extern __typeof (gtk_signal_run_type_get_type) IA__gtk_signal_run_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_signal_run_type_get_type IA__gtk_signal_run_type_get_type

extern __typeof (gtk_size_group_mode_get_type) IA__gtk_size_group_mode_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_size_group_mode_get_type IA__gtk_size_group_mode_get_type

extern __typeof (gtk_sort_type_get_type) IA__gtk_sort_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_sort_type_get_type IA__gtk_sort_type_get_type

extern __typeof (gtk_spin_type_get_type) IA__gtk_spin_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_spin_type_get_type IA__gtk_spin_type_get_type

extern __typeof (gtk_state_type_get_type) IA__gtk_state_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_state_type_get_type IA__gtk_state_type_get_type

extern __typeof (gtk_submenu_direction_get_type) IA__gtk_submenu_direction_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_submenu_direction_get_type IA__gtk_submenu_direction_get_type

extern __typeof (gtk_submenu_placement_get_type) IA__gtk_submenu_placement_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_submenu_placement_get_type IA__gtk_submenu_placement_get_type

extern __typeof (gtk_target_flags_get_type) IA__gtk_target_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_target_flags_get_type IA__gtk_target_flags_get_type

extern __typeof (gtk_text_direction_get_type) IA__gtk_text_direction_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_direction_get_type IA__gtk_text_direction_get_type

extern __typeof (gtk_text_buffer_target_info_get_type) IA__gtk_text_buffer_target_info_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_buffer_target_info_get_type IA__gtk_text_buffer_target_info_get_type

extern __typeof (gtk_text_search_flags_get_type) IA__gtk_text_search_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_search_flags_get_type IA__gtk_text_search_flags_get_type

extern __typeof (gtk_text_window_type_get_type) IA__gtk_text_window_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_window_type_get_type IA__gtk_text_window_type_get_type

extern __typeof (gtk_tool_palette_drag_targets_get_type) IA__gtk_tool_palette_drag_targets_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tool_palette_drag_targets_get_type IA__gtk_tool_palette_drag_targets_get_type

extern __typeof (gtk_tree_model_flags_get_type) IA__gtk_tree_model_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_model_flags_get_type IA__gtk_tree_model_flags_get_type

extern __typeof (gtk_tree_view_grid_lines_get_type) IA__gtk_tree_view_grid_lines_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_view_grid_lines_get_type IA__gtk_tree_view_grid_lines_get_type

extern __typeof (gtk_update_type_get_type) IA__gtk_update_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_update_type_get_type IA__gtk_update_type_get_type

extern __typeof (gtk_visibility_get_type) IA__gtk_visibility_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_visibility_get_type IA__gtk_visibility_get_type

extern __typeof (gtk_object_flags_get_type) IA__gtk_object_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_object_flags_get_type IA__gtk_object_flags_get_type

extern __typeof (gtk_accel_flags_get_type) IA__gtk_accel_flags_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_accel_flags_get_type IA__gtk_accel_flags_get_type

extern __typeof (gtk_icon_size_get_type) IA__gtk_icon_size_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_icon_size_get_type IA__gtk_icon_size_get_type

extern __typeof (gtk_file_chooser_error_get_type) IA__gtk_file_chooser_error_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_file_chooser_error_get_type IA__gtk_file_chooser_error_get_type

extern __typeof (gtk_icon_theme_error_get_type) IA__gtk_icon_theme_error_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_icon_theme_error_get_type IA__gtk_icon_theme_error_get_type

extern __typeof (gtk_toolbar_child_type_get_type) IA__gtk_toolbar_child_type_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_toolbar_child_type_get_type IA__gtk_toolbar_child_type_get_type

extern __typeof (gtk_toolbar_space_style_get_type) IA__gtk_toolbar_space_style_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_toolbar_space_style_get_type IA__gtk_toolbar_space_style_get_type

extern __typeof (gtk_toolbar_style_get_type) IA__gtk_toolbar_style_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_toolbar_style_get_type IA__gtk_toolbar_style_get_type

#endif
#endif
#if IN_HEADER(__GTK_ARROW_H__)
#if IN_FILE(__GTK_ARROW_C__)
extern __typeof (gtk_arrow_get_type) IA__gtk_arrow_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_arrow_get_type IA__gtk_arrow_get_type

extern __typeof (gtk_arrow_new) IA__gtk_arrow_new __attribute((visibility("hidden")));
#define gtk_arrow_new IA__gtk_arrow_new

extern __typeof (gtk_arrow_set) IA__gtk_arrow_set __attribute((visibility("hidden")));
#define gtk_arrow_set IA__gtk_arrow_set

#endif
#endif
#if IN_HEADER(__GTK_ASPECT_FRAME_H__)
#if IN_FILE(__GTK_ASPECT_FRAME_C__)
extern __typeof (gtk_aspect_frame_get_type) IA__gtk_aspect_frame_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_aspect_frame_get_type IA__gtk_aspect_frame_get_type

extern __typeof (gtk_aspect_frame_new) IA__gtk_aspect_frame_new __attribute((visibility("hidden")));
#define gtk_aspect_frame_new IA__gtk_aspect_frame_new

extern __typeof (gtk_aspect_frame_set) IA__gtk_aspect_frame_set __attribute((visibility("hidden")));
#define gtk_aspect_frame_set IA__gtk_aspect_frame_set

#endif
#endif
#if IN_HEADER(__GTK_BINDINGS_H__)
#if IN_FILE(__GTK_BINDINGS_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_binding_entry_clear) IA__gtk_binding_entry_clear __attribute((visibility("hidden")));
#define gtk_binding_entry_clear IA__gtk_binding_entry_clear

extern __typeof (gtk_binding_parse_binding) IA__gtk_binding_parse_binding __attribute((visibility("hidden")));
#define gtk_binding_parse_binding IA__gtk_binding_parse_binding

#endif
extern __typeof (gtk_binding_entry_skip) IA__gtk_binding_entry_skip __attribute((visibility("hidden")));
#define gtk_binding_entry_skip IA__gtk_binding_entry_skip

extern __typeof (gtk_binding_entry_add_signal) IA__gtk_binding_entry_add_signal __attribute((visibility("hidden")));
#define gtk_binding_entry_add_signal IA__gtk_binding_entry_add_signal

extern __typeof (gtk_binding_entry_add_signall) IA__gtk_binding_entry_add_signall __attribute((visibility("hidden")));
#define gtk_binding_entry_add_signall IA__gtk_binding_entry_add_signall

extern __typeof (gtk_binding_entry_remove) IA__gtk_binding_entry_remove __attribute((visibility("hidden")));
#define gtk_binding_entry_remove IA__gtk_binding_entry_remove

extern __typeof (gtk_bindings_activate) IA__gtk_bindings_activate __attribute((visibility("hidden")));
#define gtk_bindings_activate IA__gtk_bindings_activate

extern __typeof (gtk_bindings_activate_event) IA__gtk_bindings_activate_event __attribute((visibility("hidden")));
#define gtk_bindings_activate_event IA__gtk_bindings_activate_event

extern __typeof (gtk_binding_set_activate) IA__gtk_binding_set_activate __attribute((visibility("hidden")));
#define gtk_binding_set_activate IA__gtk_binding_set_activate

extern __typeof (gtk_binding_set_add_path) IA__gtk_binding_set_add_path __attribute((visibility("hidden")));
#define gtk_binding_set_add_path IA__gtk_binding_set_add_path

extern __typeof (gtk_binding_set_by_class) IA__gtk_binding_set_by_class __attribute((visibility("hidden")));
#define gtk_binding_set_by_class IA__gtk_binding_set_by_class

extern __typeof (gtk_binding_set_find) IA__gtk_binding_set_find __attribute((visibility("hidden")));
#define gtk_binding_set_find IA__gtk_binding_set_find

extern __typeof (gtk_binding_set_new) IA__gtk_binding_set_new __attribute((visibility("hidden")));
#define gtk_binding_set_new IA__gtk_binding_set_new

#endif
#endif
#if IN_HEADER(__GTK_BIN_H__)
#if IN_FILE(__GTK_BIN_C__)
extern __typeof (gtk_bin_get_child) IA__gtk_bin_get_child __attribute((visibility("hidden")));
#define gtk_bin_get_child IA__gtk_bin_get_child

extern __typeof (gtk_bin_get_type) IA__gtk_bin_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_bin_get_type IA__gtk_bin_get_type

#endif
#endif
#if IN_HEADER(__GTK_BOX_H__)
#if IN_FILE(__GTK_BOX_C__)
extern __typeof (gtk_box_get_homogeneous) IA__gtk_box_get_homogeneous __attribute((visibility("hidden")));
#define gtk_box_get_homogeneous IA__gtk_box_get_homogeneous

extern __typeof (gtk_box_get_spacing) IA__gtk_box_get_spacing __attribute((visibility("hidden")));
#define gtk_box_get_spacing IA__gtk_box_get_spacing

extern __typeof (gtk_box_get_type) IA__gtk_box_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_box_get_type IA__gtk_box_get_type

extern __typeof (gtk_box_pack_end) IA__gtk_box_pack_end __attribute((visibility("hidden")));
#define gtk_box_pack_end IA__gtk_box_pack_end

extern __typeof (gtk_box_pack_start) IA__gtk_box_pack_start __attribute((visibility("hidden")));
#define gtk_box_pack_start IA__gtk_box_pack_start

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_box_pack_end_defaults) IA__gtk_box_pack_end_defaults __attribute((visibility("hidden")));
#define gtk_box_pack_end_defaults IA__gtk_box_pack_end_defaults

extern __typeof (gtk_box_pack_start_defaults) IA__gtk_box_pack_start_defaults __attribute((visibility("hidden")));
#define gtk_box_pack_start_defaults IA__gtk_box_pack_start_defaults

#endif
extern __typeof (gtk_box_query_child_packing) IA__gtk_box_query_child_packing __attribute((visibility("hidden")));
#define gtk_box_query_child_packing IA__gtk_box_query_child_packing

extern __typeof (gtk_box_reorder_child) IA__gtk_box_reorder_child __attribute((visibility("hidden")));
#define gtk_box_reorder_child IA__gtk_box_reorder_child

extern __typeof (gtk_box_set_child_packing) IA__gtk_box_set_child_packing __attribute((visibility("hidden")));
#define gtk_box_set_child_packing IA__gtk_box_set_child_packing

extern __typeof (gtk_box_set_homogeneous) IA__gtk_box_set_homogeneous __attribute((visibility("hidden")));
#define gtk_box_set_homogeneous IA__gtk_box_set_homogeneous

extern __typeof (gtk_box_set_spacing) IA__gtk_box_set_spacing __attribute((visibility("hidden")));
#define gtk_box_set_spacing IA__gtk_box_set_spacing

#endif
#endif
#if IN_HEADER(__GTK_BUILDABLE_H__)
#if IN_FILE(__GTK_BUILDABLE_C__)
extern __typeof (gtk_buildable_add_child) IA__gtk_buildable_add_child __attribute((visibility("hidden")));
#define gtk_buildable_add_child IA__gtk_buildable_add_child

extern __typeof (gtk_buildable_construct_child) IA__gtk_buildable_construct_child __attribute((visibility("hidden")));
#define gtk_buildable_construct_child IA__gtk_buildable_construct_child

extern __typeof (gtk_buildable_custom_tag_start) IA__gtk_buildable_custom_tag_start __attribute((visibility("hidden")));
#define gtk_buildable_custom_tag_start IA__gtk_buildable_custom_tag_start

extern __typeof (gtk_buildable_custom_tag_end) IA__gtk_buildable_custom_tag_end __attribute((visibility("hidden")));
#define gtk_buildable_custom_tag_end IA__gtk_buildable_custom_tag_end

extern __typeof (gtk_buildable_custom_finished) IA__gtk_buildable_custom_finished __attribute((visibility("hidden")));
#define gtk_buildable_custom_finished IA__gtk_buildable_custom_finished

extern __typeof (gtk_buildable_get_internal_child) IA__gtk_buildable_get_internal_child __attribute((visibility("hidden")));
#define gtk_buildable_get_internal_child IA__gtk_buildable_get_internal_child

extern __typeof (gtk_buildable_get_name) IA__gtk_buildable_get_name __attribute((visibility("hidden")));
#define gtk_buildable_get_name IA__gtk_buildable_get_name

extern __typeof (gtk_buildable_get_type) IA__gtk_buildable_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_buildable_get_type IA__gtk_buildable_get_type

extern __typeof (gtk_buildable_parser_finished) IA__gtk_buildable_parser_finished __attribute((visibility("hidden")));
#define gtk_buildable_parser_finished IA__gtk_buildable_parser_finished

extern __typeof (gtk_buildable_set_name) IA__gtk_buildable_set_name __attribute((visibility("hidden")));
#define gtk_buildable_set_name IA__gtk_buildable_set_name

extern __typeof (gtk_buildable_set_buildable_property) IA__gtk_buildable_set_buildable_property __attribute((visibility("hidden")));
#define gtk_buildable_set_buildable_property IA__gtk_buildable_set_buildable_property

#endif
#endif
#if IN_HEADER(__GTK_BUILDER_H__)
#if IN_FILE(__GTK_BUILDER_C__)
extern __typeof (gtk_builder_add_from_file) IA__gtk_builder_add_from_file __attribute((visibility("hidden")));
#define gtk_builder_add_from_file IA__gtk_builder_add_from_file

extern __typeof (gtk_builder_add_from_string) IA__gtk_builder_add_from_string __attribute((visibility("hidden")));
#define gtk_builder_add_from_string IA__gtk_builder_add_from_string

extern __typeof (gtk_builder_add_objects_from_file) IA__gtk_builder_add_objects_from_file __attribute((visibility("hidden")));
#define gtk_builder_add_objects_from_file IA__gtk_builder_add_objects_from_file

extern __typeof (gtk_builder_add_objects_from_string) IA__gtk_builder_add_objects_from_string __attribute((visibility("hidden")));
#define gtk_builder_add_objects_from_string IA__gtk_builder_add_objects_from_string

extern __typeof (gtk_builder_error_quark) IA__gtk_builder_error_quark __attribute((visibility("hidden")));
#define gtk_builder_error_quark IA__gtk_builder_error_quark

extern __typeof (gtk_builder_get_object) IA__gtk_builder_get_object __attribute((visibility("hidden")));
#define gtk_builder_get_object IA__gtk_builder_get_object

extern __typeof (gtk_builder_get_objects) IA__gtk_builder_get_objects __attribute((visibility("hidden")));
#define gtk_builder_get_objects IA__gtk_builder_get_objects

extern __typeof (gtk_builder_get_translation_domain) IA__gtk_builder_get_translation_domain __attribute((visibility("hidden")));
#define gtk_builder_get_translation_domain IA__gtk_builder_get_translation_domain

extern __typeof (gtk_builder_get_type) IA__gtk_builder_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_builder_get_type IA__gtk_builder_get_type

extern __typeof (gtk_builder_get_type_from_name) IA__gtk_builder_get_type_from_name __attribute((visibility("hidden")));
#define gtk_builder_get_type_from_name IA__gtk_builder_get_type_from_name

extern __typeof (gtk_builder_new) IA__gtk_builder_new __attribute((visibility("hidden")));
#define gtk_builder_new IA__gtk_builder_new

extern __typeof (gtk_builder_set_translation_domain) IA__gtk_builder_set_translation_domain __attribute((visibility("hidden")));
#define gtk_builder_set_translation_domain IA__gtk_builder_set_translation_domain

extern __typeof (gtk_builder_connect_signals) IA__gtk_builder_connect_signals __attribute((visibility("hidden")));
#define gtk_builder_connect_signals IA__gtk_builder_connect_signals

extern __typeof (gtk_builder_connect_signals_full) IA__gtk_builder_connect_signals_full __attribute((visibility("hidden")));
#define gtk_builder_connect_signals_full IA__gtk_builder_connect_signals_full

extern __typeof (gtk_builder_value_from_string) IA__gtk_builder_value_from_string __attribute((visibility("hidden")));
#define gtk_builder_value_from_string IA__gtk_builder_value_from_string

extern __typeof (gtk_builder_value_from_string_type) IA__gtk_builder_value_from_string_type __attribute((visibility("hidden")));
#define gtk_builder_value_from_string_type IA__gtk_builder_value_from_string_type

#endif
#endif
#if IN_HEADER(__GTK_BUTTON_BOX_H__)
#if IN_FILE(__GTK_BUTTON_BOX_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_button_box_set_child_size) IA__gtk_button_box_set_child_size __attribute((visibility("hidden")));
#define gtk_button_box_set_child_size IA__gtk_button_box_set_child_size

extern __typeof (gtk_button_box_set_child_ipadding) IA__gtk_button_box_set_child_ipadding __attribute((visibility("hidden")));
#define gtk_button_box_set_child_ipadding IA__gtk_button_box_set_child_ipadding

extern __typeof (gtk_button_box_get_child_ipadding) IA__gtk_button_box_get_child_ipadding __attribute((visibility("hidden")));
#define gtk_button_box_get_child_ipadding IA__gtk_button_box_get_child_ipadding

extern __typeof (gtk_button_box_get_child_size) IA__gtk_button_box_get_child_size __attribute((visibility("hidden")));
#define gtk_button_box_get_child_size IA__gtk_button_box_get_child_size

#endif
extern __typeof (gtk_button_box_get_child_secondary) IA__gtk_button_box_get_child_secondary __attribute((visibility("hidden")));
#define gtk_button_box_get_child_secondary IA__gtk_button_box_get_child_secondary

extern __typeof (gtk_button_box_get_layout) IA__gtk_button_box_get_layout __attribute((visibility("hidden")));
#define gtk_button_box_get_layout IA__gtk_button_box_get_layout

extern __typeof (gtk_button_box_get_type) IA__gtk_button_box_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_button_box_get_type IA__gtk_button_box_get_type

extern __typeof (gtk_button_box_set_child_secondary) IA__gtk_button_box_set_child_secondary __attribute((visibility("hidden")));
#define gtk_button_box_set_child_secondary IA__gtk_button_box_set_child_secondary

extern __typeof (gtk_button_box_set_layout) IA__gtk_button_box_set_layout __attribute((visibility("hidden")));
#define gtk_button_box_set_layout IA__gtk_button_box_set_layout

#endif
#endif
#if IN_HEADER(__GTK_BUTTON_H__)
#if IN_FILE(__GTK_BUTTON_C__)
extern __typeof (gtk_button_clicked) IA__gtk_button_clicked __attribute((visibility("hidden")));
#define gtk_button_clicked IA__gtk_button_clicked

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_button_enter) IA__gtk_button_enter __attribute((visibility("hidden")));
#define gtk_button_enter IA__gtk_button_enter

#endif
extern __typeof (gtk_button_get_alignment) IA__gtk_button_get_alignment __attribute((visibility("hidden")));
#define gtk_button_get_alignment IA__gtk_button_get_alignment

extern __typeof (gtk_button_get_event_window) IA__gtk_button_get_event_window __attribute((visibility("hidden")));
#define gtk_button_get_event_window IA__gtk_button_get_event_window

extern __typeof (gtk_button_get_focus_on_click) IA__gtk_button_get_focus_on_click __attribute((visibility("hidden")));
#define gtk_button_get_focus_on_click IA__gtk_button_get_focus_on_click

extern __typeof (gtk_button_get_image) IA__gtk_button_get_image __attribute((visibility("hidden")));
#define gtk_button_get_image IA__gtk_button_get_image

extern __typeof (gtk_button_get_image_position) IA__gtk_button_get_image_position __attribute((visibility("hidden")));
#define gtk_button_get_image_position IA__gtk_button_get_image_position

extern __typeof (gtk_button_get_label) IA__gtk_button_get_label __attribute((visibility("hidden")));
#define gtk_button_get_label IA__gtk_button_get_label

extern __typeof (gtk_button_get_relief) IA__gtk_button_get_relief __attribute((visibility("hidden")));
#define gtk_button_get_relief IA__gtk_button_get_relief

extern __typeof (gtk_button_get_type) IA__gtk_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_button_get_type IA__gtk_button_get_type

extern __typeof (gtk_button_get_use_stock) IA__gtk_button_get_use_stock __attribute((visibility("hidden")));
#define gtk_button_get_use_stock IA__gtk_button_get_use_stock

extern __typeof (gtk_button_get_use_underline) IA__gtk_button_get_use_underline __attribute((visibility("hidden")));
#define gtk_button_get_use_underline IA__gtk_button_get_use_underline

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_button_leave) IA__gtk_button_leave __attribute((visibility("hidden")));
#define gtk_button_leave IA__gtk_button_leave

#endif
extern __typeof (gtk_button_new) IA__gtk_button_new __attribute((visibility("hidden")));
#define gtk_button_new IA__gtk_button_new

extern __typeof (gtk_button_new_from_stock) IA__gtk_button_new_from_stock __attribute((visibility("hidden")));
#define gtk_button_new_from_stock IA__gtk_button_new_from_stock

extern __typeof (gtk_button_new_with_label) IA__gtk_button_new_with_label __attribute((visibility("hidden")));
#define gtk_button_new_with_label IA__gtk_button_new_with_label

extern __typeof (gtk_button_new_with_mnemonic) IA__gtk_button_new_with_mnemonic __attribute((visibility("hidden")));
#define gtk_button_new_with_mnemonic IA__gtk_button_new_with_mnemonic

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_button_pressed) IA__gtk_button_pressed __attribute((visibility("hidden")));
#define gtk_button_pressed IA__gtk_button_pressed

extern __typeof (gtk_button_released) IA__gtk_button_released __attribute((visibility("hidden")));
#define gtk_button_released IA__gtk_button_released

#endif
extern __typeof (gtk_button_set_alignment) IA__gtk_button_set_alignment __attribute((visibility("hidden")));
#define gtk_button_set_alignment IA__gtk_button_set_alignment

extern __typeof (gtk_button_set_focus_on_click) IA__gtk_button_set_focus_on_click __attribute((visibility("hidden")));
#define gtk_button_set_focus_on_click IA__gtk_button_set_focus_on_click

extern __typeof (gtk_button_set_image) IA__gtk_button_set_image __attribute((visibility("hidden")));
#define gtk_button_set_image IA__gtk_button_set_image

extern __typeof (gtk_button_set_image_position) IA__gtk_button_set_image_position __attribute((visibility("hidden")));
#define gtk_button_set_image_position IA__gtk_button_set_image_position

extern __typeof (gtk_button_set_label) IA__gtk_button_set_label __attribute((visibility("hidden")));
#define gtk_button_set_label IA__gtk_button_set_label

extern __typeof (gtk_button_set_relief) IA__gtk_button_set_relief __attribute((visibility("hidden")));
#define gtk_button_set_relief IA__gtk_button_set_relief

extern __typeof (gtk_button_set_use_stock) IA__gtk_button_set_use_stock __attribute((visibility("hidden")));
#define gtk_button_set_use_stock IA__gtk_button_set_use_stock

extern __typeof (gtk_button_set_use_underline) IA__gtk_button_set_use_underline __attribute((visibility("hidden")));
#define gtk_button_set_use_underline IA__gtk_button_set_use_underline

#endif
#endif
#if IN_HEADER(__GTK_CALENDAR_H__)
#if IN_FILE(__GTK_CALENDAR_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_calendar_display_options) IA__gtk_calendar_display_options __attribute((visibility("hidden")));
#define gtk_calendar_display_options IA__gtk_calendar_display_options

extern __typeof (gtk_calendar_freeze) IA__gtk_calendar_freeze __attribute((visibility("hidden")));
#define gtk_calendar_freeze IA__gtk_calendar_freeze

extern __typeof (gtk_calendar_thaw) IA__gtk_calendar_thaw __attribute((visibility("hidden")));
#define gtk_calendar_thaw IA__gtk_calendar_thaw

#endif
extern __typeof (gtk_calendar_clear_marks) IA__gtk_calendar_clear_marks __attribute((visibility("hidden")));
#define gtk_calendar_clear_marks IA__gtk_calendar_clear_marks

extern __typeof (gtk_calendar_get_date) IA__gtk_calendar_get_date __attribute((visibility("hidden")));
#define gtk_calendar_get_date IA__gtk_calendar_get_date

extern __typeof (gtk_calendar_get_detail_height_rows) IA__gtk_calendar_get_detail_height_rows __attribute((visibility("hidden")));
#define gtk_calendar_get_detail_height_rows IA__gtk_calendar_get_detail_height_rows

extern __typeof (gtk_calendar_get_detail_width_chars) IA__gtk_calendar_get_detail_width_chars __attribute((visibility("hidden")));
#define gtk_calendar_get_detail_width_chars IA__gtk_calendar_get_detail_width_chars

extern __typeof (gtk_calendar_get_display_options) IA__gtk_calendar_get_display_options __attribute((visibility("hidden")));
#define gtk_calendar_get_display_options IA__gtk_calendar_get_display_options

extern __typeof (gtk_calendar_get_type) IA__gtk_calendar_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_calendar_get_type IA__gtk_calendar_get_type

extern __typeof (gtk_calendar_mark_day) IA__gtk_calendar_mark_day __attribute((visibility("hidden")));
#define gtk_calendar_mark_day IA__gtk_calendar_mark_day

extern __typeof (gtk_calendar_new) IA__gtk_calendar_new __attribute((visibility("hidden")));
#define gtk_calendar_new IA__gtk_calendar_new

extern __typeof (gtk_calendar_select_day) IA__gtk_calendar_select_day __attribute((visibility("hidden")));
#define gtk_calendar_select_day IA__gtk_calendar_select_day

extern __typeof (gtk_calendar_select_month) IA__gtk_calendar_select_month __attribute((visibility("hidden")));
#define gtk_calendar_select_month IA__gtk_calendar_select_month

extern __typeof (gtk_calendar_set_detail_func) IA__gtk_calendar_set_detail_func __attribute((visibility("hidden")));
#define gtk_calendar_set_detail_func IA__gtk_calendar_set_detail_func

extern __typeof (gtk_calendar_set_detail_height_rows) IA__gtk_calendar_set_detail_height_rows __attribute((visibility("hidden")));
#define gtk_calendar_set_detail_height_rows IA__gtk_calendar_set_detail_height_rows

extern __typeof (gtk_calendar_set_detail_width_chars) IA__gtk_calendar_set_detail_width_chars __attribute((visibility("hidden")));
#define gtk_calendar_set_detail_width_chars IA__gtk_calendar_set_detail_width_chars

extern __typeof (gtk_calendar_set_display_options) IA__gtk_calendar_set_display_options __attribute((visibility("hidden")));
#define gtk_calendar_set_display_options IA__gtk_calendar_set_display_options

extern __typeof (gtk_calendar_unmark_day) IA__gtk_calendar_unmark_day __attribute((visibility("hidden")));
#define gtk_calendar_unmark_day IA__gtk_calendar_unmark_day

#endif
#endif
#if IN_HEADER(__GTK_CELL_EDITABLE_H__)
#if IN_FILE(__GTK_CELL_EDITABLE_C__)
extern __typeof (gtk_cell_editable_editing_done) IA__gtk_cell_editable_editing_done __attribute((visibility("hidden")));
#define gtk_cell_editable_editing_done IA__gtk_cell_editable_editing_done

extern __typeof (gtk_cell_editable_get_type) IA__gtk_cell_editable_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_editable_get_type IA__gtk_cell_editable_get_type

extern __typeof (gtk_cell_editable_remove_widget) IA__gtk_cell_editable_remove_widget __attribute((visibility("hidden")));
#define gtk_cell_editable_remove_widget IA__gtk_cell_editable_remove_widget

extern __typeof (gtk_cell_editable_start_editing) IA__gtk_cell_editable_start_editing __attribute((visibility("hidden")));
#define gtk_cell_editable_start_editing IA__gtk_cell_editable_start_editing

#endif
#endif
#if IN_HEADER(__GTK_CELL_LAYOUT_H__)
#if IN_FILE(__GTK_CELL_LAYOUT_C__)
extern __typeof (gtk_cell_layout_add_attribute) IA__gtk_cell_layout_add_attribute __attribute((visibility("hidden")));
#define gtk_cell_layout_add_attribute IA__gtk_cell_layout_add_attribute

extern __typeof (gtk_cell_layout_clear) IA__gtk_cell_layout_clear __attribute((visibility("hidden")));
#define gtk_cell_layout_clear IA__gtk_cell_layout_clear

extern __typeof (gtk_cell_layout_clear_attributes) IA__gtk_cell_layout_clear_attributes __attribute((visibility("hidden")));
#define gtk_cell_layout_clear_attributes IA__gtk_cell_layout_clear_attributes

extern __typeof (gtk_cell_layout_get_type) IA__gtk_cell_layout_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_layout_get_type IA__gtk_cell_layout_get_type

extern __typeof (gtk_cell_layout_pack_end) IA__gtk_cell_layout_pack_end __attribute((visibility("hidden")));
#define gtk_cell_layout_pack_end IA__gtk_cell_layout_pack_end

extern __typeof (gtk_cell_layout_pack_start) IA__gtk_cell_layout_pack_start __attribute((visibility("hidden")));
#define gtk_cell_layout_pack_start IA__gtk_cell_layout_pack_start

extern __typeof (gtk_cell_layout_get_cells) IA__gtk_cell_layout_get_cells __attribute((visibility("hidden")));
#define gtk_cell_layout_get_cells IA__gtk_cell_layout_get_cells

extern __typeof (gtk_cell_layout_reorder) IA__gtk_cell_layout_reorder __attribute((visibility("hidden")));
#define gtk_cell_layout_reorder IA__gtk_cell_layout_reorder

extern __typeof (gtk_cell_layout_set_attributes) IA__gtk_cell_layout_set_attributes __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_cell_layout_set_attributes IA__gtk_cell_layout_set_attributes

extern __typeof (gtk_cell_layout_set_cell_data_func) IA__gtk_cell_layout_set_cell_data_func __attribute((visibility("hidden")));
#define gtk_cell_layout_set_cell_data_func IA__gtk_cell_layout_set_cell_data_func

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_H__)
#if IN_FILE(__GTK_CELL_RENDERER_C__)
extern __typeof (gtk_cell_renderer_activate) IA__gtk_cell_renderer_activate __attribute((visibility("hidden")));
#define gtk_cell_renderer_activate IA__gtk_cell_renderer_activate

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_cell_renderer_editing_canceled) IA__gtk_cell_renderer_editing_canceled __attribute((visibility("hidden")));
#define gtk_cell_renderer_editing_canceled IA__gtk_cell_renderer_editing_canceled

#endif
extern __typeof (gtk_cell_renderer_get_alignment) IA__gtk_cell_renderer_get_alignment __attribute((visibility("hidden")));
#define gtk_cell_renderer_get_alignment IA__gtk_cell_renderer_get_alignment

extern __typeof (gtk_cell_renderer_get_fixed_size) IA__gtk_cell_renderer_get_fixed_size __attribute((visibility("hidden")));
#define gtk_cell_renderer_get_fixed_size IA__gtk_cell_renderer_get_fixed_size

extern __typeof (gtk_cell_renderer_get_padding) IA__gtk_cell_renderer_get_padding __attribute((visibility("hidden")));
#define gtk_cell_renderer_get_padding IA__gtk_cell_renderer_get_padding

extern __typeof (gtk_cell_renderer_get_sensitive) IA__gtk_cell_renderer_get_sensitive __attribute((visibility("hidden")));
#define gtk_cell_renderer_get_sensitive IA__gtk_cell_renderer_get_sensitive

extern __typeof (gtk_cell_renderer_get_size) IA__gtk_cell_renderer_get_size __attribute((visibility("hidden")));
#define gtk_cell_renderer_get_size IA__gtk_cell_renderer_get_size

extern __typeof (gtk_cell_renderer_get_type) IA__gtk_cell_renderer_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_renderer_get_type IA__gtk_cell_renderer_get_type

extern __typeof (gtk_cell_renderer_get_visible) IA__gtk_cell_renderer_get_visible __attribute((visibility("hidden")));
#define gtk_cell_renderer_get_visible IA__gtk_cell_renderer_get_visible

extern __typeof (gtk_cell_renderer_render) IA__gtk_cell_renderer_render __attribute((visibility("hidden")));
#define gtk_cell_renderer_render IA__gtk_cell_renderer_render

extern __typeof (gtk_cell_renderer_set_alignment) IA__gtk_cell_renderer_set_alignment __attribute((visibility("hidden")));
#define gtk_cell_renderer_set_alignment IA__gtk_cell_renderer_set_alignment

extern __typeof (gtk_cell_renderer_set_fixed_size) IA__gtk_cell_renderer_set_fixed_size __attribute((visibility("hidden")));
#define gtk_cell_renderer_set_fixed_size IA__gtk_cell_renderer_set_fixed_size

extern __typeof (gtk_cell_renderer_set_padding) IA__gtk_cell_renderer_set_padding __attribute((visibility("hidden")));
#define gtk_cell_renderer_set_padding IA__gtk_cell_renderer_set_padding

extern __typeof (gtk_cell_renderer_set_sensitive) IA__gtk_cell_renderer_set_sensitive __attribute((visibility("hidden")));
#define gtk_cell_renderer_set_sensitive IA__gtk_cell_renderer_set_sensitive

extern __typeof (gtk_cell_renderer_set_visible) IA__gtk_cell_renderer_set_visible __attribute((visibility("hidden")));
#define gtk_cell_renderer_set_visible IA__gtk_cell_renderer_set_visible

extern __typeof (gtk_cell_renderer_start_editing) IA__gtk_cell_renderer_start_editing __attribute((visibility("hidden")));
#define gtk_cell_renderer_start_editing IA__gtk_cell_renderer_start_editing

extern __typeof (gtk_cell_renderer_stop_editing) IA__gtk_cell_renderer_stop_editing __attribute((visibility("hidden")));
#define gtk_cell_renderer_stop_editing IA__gtk_cell_renderer_stop_editing

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_ACCEL_H__)
#if IN_FILE(__GTK_CELL_RENDERER_ACCEL_C__)
extern __typeof (gtk_cell_renderer_accel_get_type) IA__gtk_cell_renderer_accel_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_renderer_accel_get_type IA__gtk_cell_renderer_accel_get_type

extern __typeof (gtk_cell_renderer_accel_new) IA__gtk_cell_renderer_accel_new __attribute((visibility("hidden")));
#define gtk_cell_renderer_accel_new IA__gtk_cell_renderer_accel_new

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_COMBO_H__)
#if IN_FILE(__GTK_CELL_RENDERER_COMBO_C__)
extern __typeof (gtk_cell_renderer_combo_get_type) IA__gtk_cell_renderer_combo_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_renderer_combo_get_type IA__gtk_cell_renderer_combo_get_type

extern __typeof (gtk_cell_renderer_combo_new) IA__gtk_cell_renderer_combo_new __attribute((visibility("hidden")));
#define gtk_cell_renderer_combo_new IA__gtk_cell_renderer_combo_new

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_PIXBUF_H__)
#if IN_FILE(__GTK_CELL_RENDERER_PIXBUF_C__)
extern __typeof (gtk_cell_renderer_pixbuf_get_type) IA__gtk_cell_renderer_pixbuf_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_renderer_pixbuf_get_type IA__gtk_cell_renderer_pixbuf_get_type

extern __typeof (gtk_cell_renderer_pixbuf_new) IA__gtk_cell_renderer_pixbuf_new __attribute((visibility("hidden")));
#define gtk_cell_renderer_pixbuf_new IA__gtk_cell_renderer_pixbuf_new

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_SPIN_H__)
#if IN_FILE(__GTK_CELL_RENDERER_SPIN_C__)
extern __typeof (gtk_cell_renderer_spin_get_type) IA__gtk_cell_renderer_spin_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_renderer_spin_get_type IA__gtk_cell_renderer_spin_get_type

extern __typeof (gtk_cell_renderer_spin_new) IA__gtk_cell_renderer_spin_new __attribute((visibility("hidden")));
#define gtk_cell_renderer_spin_new IA__gtk_cell_renderer_spin_new

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_SPINNER_H__)
#if IN_FILE(__GTK_CELL_RENDERER_SPINNER_C__)
extern __typeof (gtk_cell_renderer_spinner_get_type) IA__gtk_cell_renderer_spinner_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_renderer_spinner_get_type IA__gtk_cell_renderer_spinner_get_type

extern __typeof (gtk_cell_renderer_spinner_new) IA__gtk_cell_renderer_spinner_new __attribute((visibility("hidden")));
#define gtk_cell_renderer_spinner_new IA__gtk_cell_renderer_spinner_new

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_PROGRESS_H__)
#if IN_FILE(__GTK_CELL_RENDERER_PROGRESS_C__)
extern __typeof (gtk_cell_renderer_progress_get_type) IA__gtk_cell_renderer_progress_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_renderer_progress_get_type IA__gtk_cell_renderer_progress_get_type

extern __typeof (gtk_cell_renderer_progress_new) IA__gtk_cell_renderer_progress_new __attribute((visibility("hidden")));
#define gtk_cell_renderer_progress_new IA__gtk_cell_renderer_progress_new

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_TEXT_H__)
#if IN_FILE(__GTK_CELL_RENDERER_TEXT_C__)
extern __typeof (gtk_cell_renderer_text_get_type) IA__gtk_cell_renderer_text_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_renderer_text_get_type IA__gtk_cell_renderer_text_get_type

extern __typeof (gtk_cell_renderer_text_new) IA__gtk_cell_renderer_text_new __attribute((visibility("hidden")));
#define gtk_cell_renderer_text_new IA__gtk_cell_renderer_text_new

extern __typeof (gtk_cell_renderer_text_set_fixed_height_from_font) IA__gtk_cell_renderer_text_set_fixed_height_from_font __attribute((visibility("hidden")));
#define gtk_cell_renderer_text_set_fixed_height_from_font IA__gtk_cell_renderer_text_set_fixed_height_from_font

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_TOGGLE_H__)
#if IN_FILE(__GTK_CELL_RENDERER_TOGGLE_C__)
extern __typeof (gtk_cell_renderer_toggle_get_activatable) IA__gtk_cell_renderer_toggle_get_activatable __attribute((visibility("hidden")));
#define gtk_cell_renderer_toggle_get_activatable IA__gtk_cell_renderer_toggle_get_activatable

extern __typeof (gtk_cell_renderer_toggle_get_active) IA__gtk_cell_renderer_toggle_get_active __attribute((visibility("hidden")));
#define gtk_cell_renderer_toggle_get_active IA__gtk_cell_renderer_toggle_get_active

extern __typeof (gtk_cell_renderer_toggle_get_radio) IA__gtk_cell_renderer_toggle_get_radio __attribute((visibility("hidden")));
#define gtk_cell_renderer_toggle_get_radio IA__gtk_cell_renderer_toggle_get_radio

extern __typeof (gtk_cell_renderer_toggle_get_type) IA__gtk_cell_renderer_toggle_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_renderer_toggle_get_type IA__gtk_cell_renderer_toggle_get_type

extern __typeof (gtk_cell_renderer_toggle_new) IA__gtk_cell_renderer_toggle_new __attribute((visibility("hidden")));
#define gtk_cell_renderer_toggle_new IA__gtk_cell_renderer_toggle_new

extern __typeof (gtk_cell_renderer_toggle_set_activatable) IA__gtk_cell_renderer_toggle_set_activatable __attribute((visibility("hidden")));
#define gtk_cell_renderer_toggle_set_activatable IA__gtk_cell_renderer_toggle_set_activatable

extern __typeof (gtk_cell_renderer_toggle_set_active) IA__gtk_cell_renderer_toggle_set_active __attribute((visibility("hidden")));
#define gtk_cell_renderer_toggle_set_active IA__gtk_cell_renderer_toggle_set_active

extern __typeof (gtk_cell_renderer_toggle_set_radio) IA__gtk_cell_renderer_toggle_set_radio __attribute((visibility("hidden")));
#define gtk_cell_renderer_toggle_set_radio IA__gtk_cell_renderer_toggle_set_radio

#endif
#endif
#if IN_HEADER(__GTK_CELL_VIEW_H__)
#if IN_FILE(__GTK_CELL_VIEW_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_cell_view_get_cell_renderers) IA__gtk_cell_view_get_cell_renderers __attribute((visibility("hidden")));
#define gtk_cell_view_get_cell_renderers IA__gtk_cell_view_get_cell_renderers

#endif
extern __typeof (gtk_cell_view_get_displayed_row) IA__gtk_cell_view_get_displayed_row __attribute((visibility("hidden")));
#define gtk_cell_view_get_displayed_row IA__gtk_cell_view_get_displayed_row

extern __typeof (gtk_cell_view_get_model) IA__gtk_cell_view_get_model __attribute((visibility("hidden")));
#define gtk_cell_view_get_model IA__gtk_cell_view_get_model

extern __typeof (gtk_cell_view_get_size_of_row) IA__gtk_cell_view_get_size_of_row __attribute((visibility("hidden")));
#define gtk_cell_view_get_size_of_row IA__gtk_cell_view_get_size_of_row

extern __typeof (gtk_cell_view_get_type) IA__gtk_cell_view_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_cell_view_get_type IA__gtk_cell_view_get_type

extern __typeof (gtk_cell_view_new) IA__gtk_cell_view_new __attribute((visibility("hidden")));
#define gtk_cell_view_new IA__gtk_cell_view_new

extern __typeof (gtk_cell_view_new_with_markup) IA__gtk_cell_view_new_with_markup __attribute((visibility("hidden")));
#define gtk_cell_view_new_with_markup IA__gtk_cell_view_new_with_markup

extern __typeof (gtk_cell_view_new_with_pixbuf) IA__gtk_cell_view_new_with_pixbuf __attribute((visibility("hidden")));
#define gtk_cell_view_new_with_pixbuf IA__gtk_cell_view_new_with_pixbuf

extern __typeof (gtk_cell_view_new_with_text) IA__gtk_cell_view_new_with_text __attribute((visibility("hidden")));
#define gtk_cell_view_new_with_text IA__gtk_cell_view_new_with_text

extern __typeof (gtk_cell_view_set_background_color) IA__gtk_cell_view_set_background_color __attribute((visibility("hidden")));
#define gtk_cell_view_set_background_color IA__gtk_cell_view_set_background_color

extern __typeof (gtk_cell_view_set_displayed_row) IA__gtk_cell_view_set_displayed_row __attribute((visibility("hidden")));
#define gtk_cell_view_set_displayed_row IA__gtk_cell_view_set_displayed_row

extern __typeof (gtk_cell_view_set_model) IA__gtk_cell_view_set_model __attribute((visibility("hidden")));
#define gtk_cell_view_set_model IA__gtk_cell_view_set_model

#endif
#endif
#if IN_HEADER(__GTK_CHECK_BUTTON_H__)
#if IN_FILE(__GTK_CHECK_BUTTON_C__)
extern __typeof (gtk_check_button_get_type) IA__gtk_check_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_check_button_get_type IA__gtk_check_button_get_type

extern __typeof (gtk_check_button_new) IA__gtk_check_button_new __attribute((visibility("hidden")));
#define gtk_check_button_new IA__gtk_check_button_new

extern __typeof (gtk_check_button_new_with_label) IA__gtk_check_button_new_with_label __attribute((visibility("hidden")));
#define gtk_check_button_new_with_label IA__gtk_check_button_new_with_label

extern __typeof (gtk_check_button_new_with_mnemonic) IA__gtk_check_button_new_with_mnemonic __attribute((visibility("hidden")));
#define gtk_check_button_new_with_mnemonic IA__gtk_check_button_new_with_mnemonic

#endif
#endif
#if IN_HEADER(__GTK_CHECK_MENU_ITEM_H__)
#if IN_FILE(__GTK_CHECK_MENU_ITEM_C__)
extern __typeof (gtk_check_menu_item_get_active) IA__gtk_check_menu_item_get_active __attribute((visibility("hidden")));
#define gtk_check_menu_item_get_active IA__gtk_check_menu_item_get_active

extern __typeof (gtk_check_menu_item_get_draw_as_radio) IA__gtk_check_menu_item_get_draw_as_radio __attribute((visibility("hidden")));
#define gtk_check_menu_item_get_draw_as_radio IA__gtk_check_menu_item_get_draw_as_radio

extern __typeof (gtk_check_menu_item_get_inconsistent) IA__gtk_check_menu_item_get_inconsistent __attribute((visibility("hidden")));
#define gtk_check_menu_item_get_inconsistent IA__gtk_check_menu_item_get_inconsistent

extern __typeof (gtk_check_menu_item_get_type) IA__gtk_check_menu_item_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_check_menu_item_get_type IA__gtk_check_menu_item_get_type

extern __typeof (gtk_check_menu_item_new) IA__gtk_check_menu_item_new __attribute((visibility("hidden")));
#define gtk_check_menu_item_new IA__gtk_check_menu_item_new

extern __typeof (gtk_check_menu_item_new_with_label) IA__gtk_check_menu_item_new_with_label __attribute((visibility("hidden")));
#define gtk_check_menu_item_new_with_label IA__gtk_check_menu_item_new_with_label

extern __typeof (gtk_check_menu_item_new_with_mnemonic) IA__gtk_check_menu_item_new_with_mnemonic __attribute((visibility("hidden")));
#define gtk_check_menu_item_new_with_mnemonic IA__gtk_check_menu_item_new_with_mnemonic

extern __typeof (gtk_check_menu_item_set_active) IA__gtk_check_menu_item_set_active __attribute((visibility("hidden")));
#define gtk_check_menu_item_set_active IA__gtk_check_menu_item_set_active

extern __typeof (gtk_check_menu_item_set_draw_as_radio) IA__gtk_check_menu_item_set_draw_as_radio __attribute((visibility("hidden")));
#define gtk_check_menu_item_set_draw_as_radio IA__gtk_check_menu_item_set_draw_as_radio

extern __typeof (gtk_check_menu_item_set_inconsistent) IA__gtk_check_menu_item_set_inconsistent __attribute((visibility("hidden")));
#define gtk_check_menu_item_set_inconsistent IA__gtk_check_menu_item_set_inconsistent

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_check_menu_item_set_show_toggle) IA__gtk_check_menu_item_set_show_toggle __attribute((visibility("hidden")));
#define gtk_check_menu_item_set_show_toggle IA__gtk_check_menu_item_set_show_toggle

#endif
extern __typeof (gtk_check_menu_item_toggled) IA__gtk_check_menu_item_toggled __attribute((visibility("hidden")));
#define gtk_check_menu_item_toggled IA__gtk_check_menu_item_toggled

#endif
#endif
#if IN_HEADER(__GTK_CLIPBOARD_H__)
#if IN_FILE(__GTK_CLIPBOARD_C__)
extern __typeof (gtk_clipboard_clear) IA__gtk_clipboard_clear __attribute((visibility("hidden")));
#define gtk_clipboard_clear IA__gtk_clipboard_clear

extern __typeof (gtk_clipboard_get) IA__gtk_clipboard_get __attribute((visibility("hidden")));
#define gtk_clipboard_get IA__gtk_clipboard_get

extern __typeof (gtk_clipboard_get_display) IA__gtk_clipboard_get_display __attribute((visibility("hidden")));
#define gtk_clipboard_get_display IA__gtk_clipboard_get_display

extern __typeof (gtk_clipboard_get_for_display) IA__gtk_clipboard_get_for_display __attribute((visibility("hidden")));
#define gtk_clipboard_get_for_display IA__gtk_clipboard_get_for_display

extern __typeof (gtk_clipboard_get_owner) IA__gtk_clipboard_get_owner __attribute((visibility("hidden")));
#define gtk_clipboard_get_owner IA__gtk_clipboard_get_owner

extern __typeof (gtk_clipboard_get_type) IA__gtk_clipboard_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_clipboard_get_type IA__gtk_clipboard_get_type

extern __typeof (gtk_clipboard_request_contents) IA__gtk_clipboard_request_contents __attribute((visibility("hidden")));
#define gtk_clipboard_request_contents IA__gtk_clipboard_request_contents

extern __typeof (gtk_clipboard_request_image) IA__gtk_clipboard_request_image __attribute((visibility("hidden")));
#define gtk_clipboard_request_image IA__gtk_clipboard_request_image

extern __typeof (gtk_clipboard_request_rich_text) IA__gtk_clipboard_request_rich_text __attribute((visibility("hidden")));
#define gtk_clipboard_request_rich_text IA__gtk_clipboard_request_rich_text

extern __typeof (gtk_clipboard_request_targets) IA__gtk_clipboard_request_targets __attribute((visibility("hidden")));
#define gtk_clipboard_request_targets IA__gtk_clipboard_request_targets

extern __typeof (gtk_clipboard_request_text) IA__gtk_clipboard_request_text __attribute((visibility("hidden")));
#define gtk_clipboard_request_text IA__gtk_clipboard_request_text

extern __typeof (gtk_clipboard_request_uris) IA__gtk_clipboard_request_uris __attribute((visibility("hidden")));
#define gtk_clipboard_request_uris IA__gtk_clipboard_request_uris

extern __typeof (gtk_clipboard_set_can_store) IA__gtk_clipboard_set_can_store __attribute((visibility("hidden")));
#define gtk_clipboard_set_can_store IA__gtk_clipboard_set_can_store

extern __typeof (gtk_clipboard_set_image) IA__gtk_clipboard_set_image __attribute((visibility("hidden")));
#define gtk_clipboard_set_image IA__gtk_clipboard_set_image

extern __typeof (gtk_clipboard_set_text) IA__gtk_clipboard_set_text __attribute((visibility("hidden")));
#define gtk_clipboard_set_text IA__gtk_clipboard_set_text

extern __typeof (gtk_clipboard_set_with_data) IA__gtk_clipboard_set_with_data __attribute((visibility("hidden")));
#define gtk_clipboard_set_with_data IA__gtk_clipboard_set_with_data

extern __typeof (gtk_clipboard_set_with_owner) IA__gtk_clipboard_set_with_owner __attribute((visibility("hidden")));
#define gtk_clipboard_set_with_owner IA__gtk_clipboard_set_with_owner

extern __typeof (gtk_clipboard_store) IA__gtk_clipboard_store __attribute((visibility("hidden")));
#define gtk_clipboard_store IA__gtk_clipboard_store

extern __typeof (gtk_clipboard_wait_for_contents) IA__gtk_clipboard_wait_for_contents __attribute((visibility("hidden")));
#define gtk_clipboard_wait_for_contents IA__gtk_clipboard_wait_for_contents

extern __typeof (gtk_clipboard_wait_for_image) IA__gtk_clipboard_wait_for_image __attribute((visibility("hidden")));
#define gtk_clipboard_wait_for_image IA__gtk_clipboard_wait_for_image

extern __typeof (gtk_clipboard_wait_for_rich_text) IA__gtk_clipboard_wait_for_rich_text __attribute((visibility("hidden")));
#define gtk_clipboard_wait_for_rich_text IA__gtk_clipboard_wait_for_rich_text

extern __typeof (gtk_clipboard_wait_for_uris) IA__gtk_clipboard_wait_for_uris __attribute((visibility("hidden")));
#define gtk_clipboard_wait_for_uris IA__gtk_clipboard_wait_for_uris

extern __typeof (gtk_clipboard_wait_for_targets) IA__gtk_clipboard_wait_for_targets __attribute((visibility("hidden")));
#define gtk_clipboard_wait_for_targets IA__gtk_clipboard_wait_for_targets

extern __typeof (gtk_clipboard_wait_for_text) IA__gtk_clipboard_wait_for_text __attribute((visibility("hidden")));
#define gtk_clipboard_wait_for_text IA__gtk_clipboard_wait_for_text

extern __typeof (gtk_clipboard_wait_is_image_available) IA__gtk_clipboard_wait_is_image_available __attribute((visibility("hidden")));
#define gtk_clipboard_wait_is_image_available IA__gtk_clipboard_wait_is_image_available

extern __typeof (gtk_clipboard_wait_is_rich_text_available) IA__gtk_clipboard_wait_is_rich_text_available __attribute((visibility("hidden")));
#define gtk_clipboard_wait_is_rich_text_available IA__gtk_clipboard_wait_is_rich_text_available

extern __typeof (gtk_clipboard_wait_is_text_available) IA__gtk_clipboard_wait_is_text_available __attribute((visibility("hidden")));
#define gtk_clipboard_wait_is_text_available IA__gtk_clipboard_wait_is_text_available

extern __typeof (gtk_clipboard_wait_is_uris_available) IA__gtk_clipboard_wait_is_uris_available __attribute((visibility("hidden")));
#define gtk_clipboard_wait_is_uris_available IA__gtk_clipboard_wait_is_uris_available

extern __typeof (gtk_clipboard_wait_is_target_available) IA__gtk_clipboard_wait_is_target_available __attribute((visibility("hidden")));
#define gtk_clipboard_wait_is_target_available IA__gtk_clipboard_wait_is_target_available

#endif
#endif
#if IN_HEADER(__GTK_CLIST_H__)
#if IN_FILE(__GTK_CLIST_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_clist_append) IA__gtk_clist_append __attribute((visibility("hidden")));
#define gtk_clist_append IA__gtk_clist_append

extern __typeof (gtk_clist_clear) IA__gtk_clist_clear __attribute((visibility("hidden")));
#define gtk_clist_clear IA__gtk_clist_clear

extern __typeof (gtk_clist_columns_autosize) IA__gtk_clist_columns_autosize __attribute((visibility("hidden")));
#define gtk_clist_columns_autosize IA__gtk_clist_columns_autosize

extern __typeof (gtk_clist_column_title_active) IA__gtk_clist_column_title_active __attribute((visibility("hidden")));
#define gtk_clist_column_title_active IA__gtk_clist_column_title_active

extern __typeof (gtk_clist_column_title_passive) IA__gtk_clist_column_title_passive __attribute((visibility("hidden")));
#define gtk_clist_column_title_passive IA__gtk_clist_column_title_passive

extern __typeof (gtk_clist_column_titles_active) IA__gtk_clist_column_titles_active __attribute((visibility("hidden")));
#define gtk_clist_column_titles_active IA__gtk_clist_column_titles_active

extern __typeof (gtk_clist_column_titles_hide) IA__gtk_clist_column_titles_hide __attribute((visibility("hidden")));
#define gtk_clist_column_titles_hide IA__gtk_clist_column_titles_hide

extern __typeof (gtk_clist_column_titles_passive) IA__gtk_clist_column_titles_passive __attribute((visibility("hidden")));
#define gtk_clist_column_titles_passive IA__gtk_clist_column_titles_passive

extern __typeof (gtk_clist_column_titles_show) IA__gtk_clist_column_titles_show __attribute((visibility("hidden")));
#define gtk_clist_column_titles_show IA__gtk_clist_column_titles_show

extern __typeof (gtk_clist_find_row_from_data) IA__gtk_clist_find_row_from_data __attribute((visibility("hidden")));
#define gtk_clist_find_row_from_data IA__gtk_clist_find_row_from_data

extern __typeof (gtk_clist_freeze) IA__gtk_clist_freeze __attribute((visibility("hidden")));
#define gtk_clist_freeze IA__gtk_clist_freeze

extern __typeof (gtk_clist_get_cell_style) IA__gtk_clist_get_cell_style __attribute((visibility("hidden")));
#define gtk_clist_get_cell_style IA__gtk_clist_get_cell_style

extern __typeof (gtk_clist_get_cell_type) IA__gtk_clist_get_cell_type __attribute((visibility("hidden")));
#define gtk_clist_get_cell_type IA__gtk_clist_get_cell_type

extern __typeof (gtk_clist_get_column_title) IA__gtk_clist_get_column_title __attribute((visibility("hidden")));
#define gtk_clist_get_column_title IA__gtk_clist_get_column_title

extern __typeof (gtk_clist_get_column_widget) IA__gtk_clist_get_column_widget __attribute((visibility("hidden")));
#define gtk_clist_get_column_widget IA__gtk_clist_get_column_widget

extern __typeof (gtk_clist_get_hadjustment) IA__gtk_clist_get_hadjustment __attribute((visibility("hidden")));
#define gtk_clist_get_hadjustment IA__gtk_clist_get_hadjustment

extern __typeof (gtk_clist_get_pixmap) IA__gtk_clist_get_pixmap __attribute((visibility("hidden")));
#define gtk_clist_get_pixmap IA__gtk_clist_get_pixmap

extern __typeof (gtk_clist_get_pixtext) IA__gtk_clist_get_pixtext __attribute((visibility("hidden")));
#define gtk_clist_get_pixtext IA__gtk_clist_get_pixtext

extern __typeof (gtk_clist_get_row_data) IA__gtk_clist_get_row_data __attribute((visibility("hidden")));
#define gtk_clist_get_row_data IA__gtk_clist_get_row_data

extern __typeof (gtk_clist_get_row_style) IA__gtk_clist_get_row_style __attribute((visibility("hidden")));
#define gtk_clist_get_row_style IA__gtk_clist_get_row_style

extern __typeof (gtk_clist_get_selectable) IA__gtk_clist_get_selectable __attribute((visibility("hidden")));
#define gtk_clist_get_selectable IA__gtk_clist_get_selectable

extern __typeof (gtk_clist_get_selection_info) IA__gtk_clist_get_selection_info __attribute((visibility("hidden")));
#define gtk_clist_get_selection_info IA__gtk_clist_get_selection_info

extern __typeof (gtk_clist_get_text) IA__gtk_clist_get_text __attribute((visibility("hidden")));
#define gtk_clist_get_text IA__gtk_clist_get_text

extern __typeof (gtk_clist_get_type) IA__gtk_clist_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_clist_get_type IA__gtk_clist_get_type

extern __typeof (gtk_clist_get_vadjustment) IA__gtk_clist_get_vadjustment __attribute((visibility("hidden")));
#define gtk_clist_get_vadjustment IA__gtk_clist_get_vadjustment

extern __typeof (gtk_clist_insert) IA__gtk_clist_insert __attribute((visibility("hidden")));
#define gtk_clist_insert IA__gtk_clist_insert

extern __typeof (gtk_clist_moveto) IA__gtk_clist_moveto __attribute((visibility("hidden")));
#define gtk_clist_moveto IA__gtk_clist_moveto

extern __typeof (gtk_clist_new) IA__gtk_clist_new __attribute((visibility("hidden")));
#define gtk_clist_new IA__gtk_clist_new

extern __typeof (gtk_clist_new_with_titles) IA__gtk_clist_new_with_titles __attribute((visibility("hidden")));
#define gtk_clist_new_with_titles IA__gtk_clist_new_with_titles

extern __typeof (gtk_clist_optimal_column_width) IA__gtk_clist_optimal_column_width __attribute((visibility("hidden")));
#define gtk_clist_optimal_column_width IA__gtk_clist_optimal_column_width

extern __typeof (gtk_clist_prepend) IA__gtk_clist_prepend __attribute((visibility("hidden")));
#define gtk_clist_prepend IA__gtk_clist_prepend

extern __typeof (gtk_clist_remove) IA__gtk_clist_remove __attribute((visibility("hidden")));
#define gtk_clist_remove IA__gtk_clist_remove

extern __typeof (gtk_clist_row_is_visible) IA__gtk_clist_row_is_visible __attribute((visibility("hidden")));
#define gtk_clist_row_is_visible IA__gtk_clist_row_is_visible

extern __typeof (gtk_clist_row_move) IA__gtk_clist_row_move __attribute((visibility("hidden")));
#define gtk_clist_row_move IA__gtk_clist_row_move

extern __typeof (gtk_clist_select_all) IA__gtk_clist_select_all __attribute((visibility("hidden")));
#define gtk_clist_select_all IA__gtk_clist_select_all

extern __typeof (gtk_clist_select_row) IA__gtk_clist_select_row __attribute((visibility("hidden")));
#define gtk_clist_select_row IA__gtk_clist_select_row

extern __typeof (gtk_clist_set_auto_sort) IA__gtk_clist_set_auto_sort __attribute((visibility("hidden")));
#define gtk_clist_set_auto_sort IA__gtk_clist_set_auto_sort

extern __typeof (gtk_clist_set_background) IA__gtk_clist_set_background __attribute((visibility("hidden")));
#define gtk_clist_set_background IA__gtk_clist_set_background

extern __typeof (gtk_clist_set_button_actions) IA__gtk_clist_set_button_actions __attribute((visibility("hidden")));
#define gtk_clist_set_button_actions IA__gtk_clist_set_button_actions

extern __typeof (gtk_clist_set_cell_style) IA__gtk_clist_set_cell_style __attribute((visibility("hidden")));
#define gtk_clist_set_cell_style IA__gtk_clist_set_cell_style

extern __typeof (gtk_clist_set_column_auto_resize) IA__gtk_clist_set_column_auto_resize __attribute((visibility("hidden")));
#define gtk_clist_set_column_auto_resize IA__gtk_clist_set_column_auto_resize

extern __typeof (gtk_clist_set_column_justification) IA__gtk_clist_set_column_justification __attribute((visibility("hidden")));
#define gtk_clist_set_column_justification IA__gtk_clist_set_column_justification

extern __typeof (gtk_clist_set_column_max_width) IA__gtk_clist_set_column_max_width __attribute((visibility("hidden")));
#define gtk_clist_set_column_max_width IA__gtk_clist_set_column_max_width

extern __typeof (gtk_clist_set_column_min_width) IA__gtk_clist_set_column_min_width __attribute((visibility("hidden")));
#define gtk_clist_set_column_min_width IA__gtk_clist_set_column_min_width

extern __typeof (gtk_clist_set_column_resizeable) IA__gtk_clist_set_column_resizeable __attribute((visibility("hidden")));
#define gtk_clist_set_column_resizeable IA__gtk_clist_set_column_resizeable

extern __typeof (gtk_clist_set_column_title) IA__gtk_clist_set_column_title __attribute((visibility("hidden")));
#define gtk_clist_set_column_title IA__gtk_clist_set_column_title

extern __typeof (gtk_clist_set_column_visibility) IA__gtk_clist_set_column_visibility __attribute((visibility("hidden")));
#define gtk_clist_set_column_visibility IA__gtk_clist_set_column_visibility

extern __typeof (gtk_clist_set_column_widget) IA__gtk_clist_set_column_widget __attribute((visibility("hidden")));
#define gtk_clist_set_column_widget IA__gtk_clist_set_column_widget

extern __typeof (gtk_clist_set_column_width) IA__gtk_clist_set_column_width __attribute((visibility("hidden")));
#define gtk_clist_set_column_width IA__gtk_clist_set_column_width

extern __typeof (gtk_clist_set_compare_func) IA__gtk_clist_set_compare_func __attribute((visibility("hidden")));
#define gtk_clist_set_compare_func IA__gtk_clist_set_compare_func

extern __typeof (gtk_clist_set_foreground) IA__gtk_clist_set_foreground __attribute((visibility("hidden")));
#define gtk_clist_set_foreground IA__gtk_clist_set_foreground

extern __typeof (gtk_clist_set_hadjustment) IA__gtk_clist_set_hadjustment __attribute((visibility("hidden")));
#define gtk_clist_set_hadjustment IA__gtk_clist_set_hadjustment

extern __typeof (gtk_clist_set_pixmap) IA__gtk_clist_set_pixmap __attribute((visibility("hidden")));
#define gtk_clist_set_pixmap IA__gtk_clist_set_pixmap

extern __typeof (gtk_clist_set_pixtext) IA__gtk_clist_set_pixtext __attribute((visibility("hidden")));
#define gtk_clist_set_pixtext IA__gtk_clist_set_pixtext

extern __typeof (gtk_clist_set_reorderable) IA__gtk_clist_set_reorderable __attribute((visibility("hidden")));
#define gtk_clist_set_reorderable IA__gtk_clist_set_reorderable

extern __typeof (gtk_clist_set_row_data) IA__gtk_clist_set_row_data __attribute((visibility("hidden")));
#define gtk_clist_set_row_data IA__gtk_clist_set_row_data

extern __typeof (gtk_clist_set_row_data_full) IA__gtk_clist_set_row_data_full __attribute((visibility("hidden")));
#define gtk_clist_set_row_data_full IA__gtk_clist_set_row_data_full

extern __typeof (gtk_clist_set_row_height) IA__gtk_clist_set_row_height __attribute((visibility("hidden")));
#define gtk_clist_set_row_height IA__gtk_clist_set_row_height

extern __typeof (gtk_clist_set_row_style) IA__gtk_clist_set_row_style __attribute((visibility("hidden")));
#define gtk_clist_set_row_style IA__gtk_clist_set_row_style

extern __typeof (gtk_clist_set_selectable) IA__gtk_clist_set_selectable __attribute((visibility("hidden")));
#define gtk_clist_set_selectable IA__gtk_clist_set_selectable

extern __typeof (gtk_clist_set_selection_mode) IA__gtk_clist_set_selection_mode __attribute((visibility("hidden")));
#define gtk_clist_set_selection_mode IA__gtk_clist_set_selection_mode

extern __typeof (gtk_clist_set_shadow_type) IA__gtk_clist_set_shadow_type __attribute((visibility("hidden")));
#define gtk_clist_set_shadow_type IA__gtk_clist_set_shadow_type

extern __typeof (gtk_clist_set_shift) IA__gtk_clist_set_shift __attribute((visibility("hidden")));
#define gtk_clist_set_shift IA__gtk_clist_set_shift

extern __typeof (gtk_clist_set_sort_column) IA__gtk_clist_set_sort_column __attribute((visibility("hidden")));
#define gtk_clist_set_sort_column IA__gtk_clist_set_sort_column

extern __typeof (gtk_clist_set_sort_type) IA__gtk_clist_set_sort_type __attribute((visibility("hidden")));
#define gtk_clist_set_sort_type IA__gtk_clist_set_sort_type

extern __typeof (gtk_clist_set_text) IA__gtk_clist_set_text __attribute((visibility("hidden")));
#define gtk_clist_set_text IA__gtk_clist_set_text

extern __typeof (gtk_clist_set_use_drag_icons) IA__gtk_clist_set_use_drag_icons __attribute((visibility("hidden")));
#define gtk_clist_set_use_drag_icons IA__gtk_clist_set_use_drag_icons

extern __typeof (gtk_clist_set_vadjustment) IA__gtk_clist_set_vadjustment __attribute((visibility("hidden")));
#define gtk_clist_set_vadjustment IA__gtk_clist_set_vadjustment

extern __typeof (gtk_clist_sort) IA__gtk_clist_sort __attribute((visibility("hidden")));
#define gtk_clist_sort IA__gtk_clist_sort

extern __typeof (gtk_clist_swap_rows) IA__gtk_clist_swap_rows __attribute((visibility("hidden")));
#define gtk_clist_swap_rows IA__gtk_clist_swap_rows

extern __typeof (gtk_clist_thaw) IA__gtk_clist_thaw __attribute((visibility("hidden")));
#define gtk_clist_thaw IA__gtk_clist_thaw

extern __typeof (gtk_clist_undo_selection) IA__gtk_clist_undo_selection __attribute((visibility("hidden")));
#define gtk_clist_undo_selection IA__gtk_clist_undo_selection

extern __typeof (gtk_clist_unselect_all) IA__gtk_clist_unselect_all __attribute((visibility("hidden")));
#define gtk_clist_unselect_all IA__gtk_clist_unselect_all

extern __typeof (gtk_clist_unselect_row) IA__gtk_clist_unselect_row __attribute((visibility("hidden")));
#define gtk_clist_unselect_row IA__gtk_clist_unselect_row

#endif
#endif
#endif
#if IN_HEADER(__GTK_COLOR_BUTTON_H__)
#if IN_FILE(__GTK_COLOR_BUTTON_C__)
extern __typeof (gtk_color_button_get_alpha) IA__gtk_color_button_get_alpha __attribute((visibility("hidden")));
#define gtk_color_button_get_alpha IA__gtk_color_button_get_alpha

extern __typeof (gtk_color_button_get_color) IA__gtk_color_button_get_color __attribute((visibility("hidden")));
#define gtk_color_button_get_color IA__gtk_color_button_get_color

extern __typeof (gtk_color_button_get_title) IA__gtk_color_button_get_title __attribute((visibility("hidden")));
#define gtk_color_button_get_title IA__gtk_color_button_get_title

extern __typeof (gtk_color_button_get_type) IA__gtk_color_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_color_button_get_type IA__gtk_color_button_get_type

extern __typeof (gtk_color_button_get_use_alpha) IA__gtk_color_button_get_use_alpha __attribute((visibility("hidden")));
#define gtk_color_button_get_use_alpha IA__gtk_color_button_get_use_alpha

extern __typeof (gtk_color_button_new) IA__gtk_color_button_new __attribute((visibility("hidden")));
#define gtk_color_button_new IA__gtk_color_button_new

extern __typeof (gtk_color_button_new_with_color) IA__gtk_color_button_new_with_color __attribute((visibility("hidden")));
#define gtk_color_button_new_with_color IA__gtk_color_button_new_with_color

extern __typeof (gtk_color_button_set_alpha) IA__gtk_color_button_set_alpha __attribute((visibility("hidden")));
#define gtk_color_button_set_alpha IA__gtk_color_button_set_alpha

extern __typeof (gtk_color_button_set_color) IA__gtk_color_button_set_color __attribute((visibility("hidden")));
#define gtk_color_button_set_color IA__gtk_color_button_set_color

extern __typeof (gtk_color_button_set_title) IA__gtk_color_button_set_title __attribute((visibility("hidden")));
#define gtk_color_button_set_title IA__gtk_color_button_set_title

extern __typeof (gtk_color_button_set_use_alpha) IA__gtk_color_button_set_use_alpha __attribute((visibility("hidden")));
#define gtk_color_button_set_use_alpha IA__gtk_color_button_set_use_alpha

#endif
#endif
#if IN_HEADER(__GTK_COLOR_SELECTION_H__)
#if IN_FILE(__GTK_COLOR_SELECTION_C__)
extern __typeof (gtk_color_selection_get_current_alpha) IA__gtk_color_selection_get_current_alpha __attribute((visibility("hidden")));
#define gtk_color_selection_get_current_alpha IA__gtk_color_selection_get_current_alpha

extern __typeof (gtk_color_selection_get_current_color) IA__gtk_color_selection_get_current_color __attribute((visibility("hidden")));
#define gtk_color_selection_get_current_color IA__gtk_color_selection_get_current_color

extern __typeof (gtk_color_selection_get_has_opacity_control) IA__gtk_color_selection_get_has_opacity_control __attribute((visibility("hidden")));
#define gtk_color_selection_get_has_opacity_control IA__gtk_color_selection_get_has_opacity_control

extern __typeof (gtk_color_selection_get_has_palette) IA__gtk_color_selection_get_has_palette __attribute((visibility("hidden")));
#define gtk_color_selection_get_has_palette IA__gtk_color_selection_get_has_palette

extern __typeof (gtk_color_selection_get_previous_alpha) IA__gtk_color_selection_get_previous_alpha __attribute((visibility("hidden")));
#define gtk_color_selection_get_previous_alpha IA__gtk_color_selection_get_previous_alpha

extern __typeof (gtk_color_selection_get_previous_color) IA__gtk_color_selection_get_previous_color __attribute((visibility("hidden")));
#define gtk_color_selection_get_previous_color IA__gtk_color_selection_get_previous_color

extern __typeof (gtk_color_selection_get_type) IA__gtk_color_selection_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_color_selection_get_type IA__gtk_color_selection_get_type

extern __typeof (gtk_color_selection_is_adjusting) IA__gtk_color_selection_is_adjusting __attribute((visibility("hidden")));
#define gtk_color_selection_is_adjusting IA__gtk_color_selection_is_adjusting

extern __typeof (gtk_color_selection_new) IA__gtk_color_selection_new __attribute((visibility("hidden")));
#define gtk_color_selection_new IA__gtk_color_selection_new

extern __typeof (gtk_color_selection_palette_from_string) IA__gtk_color_selection_palette_from_string __attribute((visibility("hidden")));
#define gtk_color_selection_palette_from_string IA__gtk_color_selection_palette_from_string

extern __typeof (gtk_color_selection_palette_to_string) IA__gtk_color_selection_palette_to_string __attribute((visibility("hidden")));
#define gtk_color_selection_palette_to_string IA__gtk_color_selection_palette_to_string

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_color_selection_set_change_palette_hook) IA__gtk_color_selection_set_change_palette_hook __attribute((visibility("hidden")));
#define gtk_color_selection_set_change_palette_hook IA__gtk_color_selection_set_change_palette_hook

extern __typeof (gtk_color_selection_set_color) IA__gtk_color_selection_set_color __attribute((visibility("hidden")));
#define gtk_color_selection_set_color IA__gtk_color_selection_set_color

extern __typeof (gtk_color_selection_get_color) IA__gtk_color_selection_get_color __attribute((visibility("hidden")));
#define gtk_color_selection_get_color IA__gtk_color_selection_get_color

extern __typeof (gtk_color_selection_set_update_policy) IA__gtk_color_selection_set_update_policy __attribute((visibility("hidden")));
#define gtk_color_selection_set_update_policy IA__gtk_color_selection_set_update_policy

#endif
extern __typeof (gtk_color_selection_set_change_palette_with_screen_hook) IA__gtk_color_selection_set_change_palette_with_screen_hook __attribute((visibility("hidden")));
#define gtk_color_selection_set_change_palette_with_screen_hook IA__gtk_color_selection_set_change_palette_with_screen_hook

extern __typeof (gtk_color_selection_set_current_alpha) IA__gtk_color_selection_set_current_alpha __attribute((visibility("hidden")));
#define gtk_color_selection_set_current_alpha IA__gtk_color_selection_set_current_alpha

extern __typeof (gtk_color_selection_set_current_color) IA__gtk_color_selection_set_current_color __attribute((visibility("hidden")));
#define gtk_color_selection_set_current_color IA__gtk_color_selection_set_current_color

extern __typeof (gtk_color_selection_set_has_opacity_control) IA__gtk_color_selection_set_has_opacity_control __attribute((visibility("hidden")));
#define gtk_color_selection_set_has_opacity_control IA__gtk_color_selection_set_has_opacity_control

extern __typeof (gtk_color_selection_set_has_palette) IA__gtk_color_selection_set_has_palette __attribute((visibility("hidden")));
#define gtk_color_selection_set_has_palette IA__gtk_color_selection_set_has_palette

extern __typeof (gtk_color_selection_set_previous_alpha) IA__gtk_color_selection_set_previous_alpha __attribute((visibility("hidden")));
#define gtk_color_selection_set_previous_alpha IA__gtk_color_selection_set_previous_alpha

extern __typeof (gtk_color_selection_set_previous_color) IA__gtk_color_selection_set_previous_color __attribute((visibility("hidden")));
#define gtk_color_selection_set_previous_color IA__gtk_color_selection_set_previous_color

#endif
#endif
#if IN_HEADER(__GTK_COLOR_SELECTION_DIALOG_H__)
#if IN_FILE(__GTK_COLOR_SELECTION_DIALOG_C__)
extern __typeof (gtk_color_selection_dialog_get_type) IA__gtk_color_selection_dialog_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_color_selection_dialog_get_type IA__gtk_color_selection_dialog_get_type

extern __typeof (gtk_color_selection_dialog_new) IA__gtk_color_selection_dialog_new __attribute((visibility("hidden")));
#define gtk_color_selection_dialog_new IA__gtk_color_selection_dialog_new

extern __typeof (gtk_color_selection_dialog_get_color_selection) IA__gtk_color_selection_dialog_get_color_selection __attribute((visibility("hidden")));
#define gtk_color_selection_dialog_get_color_selection IA__gtk_color_selection_dialog_get_color_selection

#endif
#endif
#if IN_HEADER(__GTK_COMBO_BOX_H__)
#if IN_FILE(__GTK_COMBO_BOX_C__)
extern __typeof (gtk_combo_box_append_text) IA__gtk_combo_box_append_text __attribute((visibility("hidden")));
#define gtk_combo_box_append_text IA__gtk_combo_box_append_text

extern __typeof (gtk_combo_box_get_active) IA__gtk_combo_box_get_active __attribute((visibility("hidden")));
#define gtk_combo_box_get_active IA__gtk_combo_box_get_active

extern __typeof (gtk_combo_box_get_active_iter) IA__gtk_combo_box_get_active_iter __attribute((visibility("hidden")));
#define gtk_combo_box_get_active_iter IA__gtk_combo_box_get_active_iter

extern __typeof (gtk_combo_box_get_active_text) IA__gtk_combo_box_get_active_text __attribute((visibility("hidden")));
#define gtk_combo_box_get_active_text IA__gtk_combo_box_get_active_text

extern __typeof (gtk_combo_box_get_add_tearoffs) IA__gtk_combo_box_get_add_tearoffs __attribute((visibility("hidden")));
#define gtk_combo_box_get_add_tearoffs IA__gtk_combo_box_get_add_tearoffs

extern __typeof (gtk_combo_box_get_column_span_column) IA__gtk_combo_box_get_column_span_column __attribute((visibility("hidden")));
#define gtk_combo_box_get_column_span_column IA__gtk_combo_box_get_column_span_column

extern __typeof (gtk_combo_box_get_entry_text_column) IA__gtk_combo_box_get_entry_text_column __attribute((visibility("hidden")));
#define gtk_combo_box_get_entry_text_column IA__gtk_combo_box_get_entry_text_column

extern __typeof (gtk_combo_box_get_focus_on_click) IA__gtk_combo_box_get_focus_on_click __attribute((visibility("hidden")));
#define gtk_combo_box_get_focus_on_click IA__gtk_combo_box_get_focus_on_click

extern __typeof (gtk_combo_box_get_has_entry) IA__gtk_combo_box_get_has_entry __attribute((visibility("hidden")));
#define gtk_combo_box_get_has_entry IA__gtk_combo_box_get_has_entry

extern __typeof (gtk_combo_box_get_model) IA__gtk_combo_box_get_model __attribute((visibility("hidden")));
#define gtk_combo_box_get_model IA__gtk_combo_box_get_model

extern __typeof (gtk_combo_box_get_popup_accessible) IA__gtk_combo_box_get_popup_accessible __attribute((visibility("hidden")));
#define gtk_combo_box_get_popup_accessible IA__gtk_combo_box_get_popup_accessible

extern __typeof (gtk_combo_box_get_row_separator_func) IA__gtk_combo_box_get_row_separator_func __attribute((visibility("hidden")));
#define gtk_combo_box_get_row_separator_func IA__gtk_combo_box_get_row_separator_func

extern __typeof (gtk_combo_box_get_row_span_column) IA__gtk_combo_box_get_row_span_column __attribute((visibility("hidden")));
#define gtk_combo_box_get_row_span_column IA__gtk_combo_box_get_row_span_column

extern __typeof (gtk_combo_box_get_button_sensitivity) IA__gtk_combo_box_get_button_sensitivity __attribute((visibility("hidden")));
#define gtk_combo_box_get_button_sensitivity IA__gtk_combo_box_get_button_sensitivity

extern __typeof (gtk_combo_box_get_title) IA__gtk_combo_box_get_title __attribute((visibility("hidden")));
#define gtk_combo_box_get_title IA__gtk_combo_box_get_title

extern __typeof (gtk_combo_box_get_type) IA__gtk_combo_box_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_combo_box_get_type IA__gtk_combo_box_get_type

extern __typeof (gtk_combo_box_get_wrap_width) IA__gtk_combo_box_get_wrap_width __attribute((visibility("hidden")));
#define gtk_combo_box_get_wrap_width IA__gtk_combo_box_get_wrap_width

extern __typeof (gtk_combo_box_insert_text) IA__gtk_combo_box_insert_text __attribute((visibility("hidden")));
#define gtk_combo_box_insert_text IA__gtk_combo_box_insert_text

extern __typeof (gtk_combo_box_new) IA__gtk_combo_box_new __attribute((visibility("hidden")));
#define gtk_combo_box_new IA__gtk_combo_box_new

extern __typeof (gtk_combo_box_new_text) IA__gtk_combo_box_new_text __attribute((visibility("hidden")));
#define gtk_combo_box_new_text IA__gtk_combo_box_new_text

extern __typeof (gtk_combo_box_new_with_entry) IA__gtk_combo_box_new_with_entry __attribute((visibility("hidden")));
#define gtk_combo_box_new_with_entry IA__gtk_combo_box_new_with_entry

extern __typeof (gtk_combo_box_new_with_model) IA__gtk_combo_box_new_with_model __attribute((visibility("hidden")));
#define gtk_combo_box_new_with_model IA__gtk_combo_box_new_with_model

extern __typeof (gtk_combo_box_new_with_model_and_entry) IA__gtk_combo_box_new_with_model_and_entry __attribute((visibility("hidden")));
#define gtk_combo_box_new_with_model_and_entry IA__gtk_combo_box_new_with_model_and_entry

extern __typeof (gtk_combo_box_popdown) IA__gtk_combo_box_popdown __attribute((visibility("hidden")));
#define gtk_combo_box_popdown IA__gtk_combo_box_popdown

extern __typeof (gtk_combo_box_popup) IA__gtk_combo_box_popup __attribute((visibility("hidden")));
#define gtk_combo_box_popup IA__gtk_combo_box_popup

extern __typeof (gtk_combo_box_prepend_text) IA__gtk_combo_box_prepend_text __attribute((visibility("hidden")));
#define gtk_combo_box_prepend_text IA__gtk_combo_box_prepend_text

extern __typeof (gtk_combo_box_remove_text) IA__gtk_combo_box_remove_text __attribute((visibility("hidden")));
#define gtk_combo_box_remove_text IA__gtk_combo_box_remove_text

extern __typeof (gtk_combo_box_set_active) IA__gtk_combo_box_set_active __attribute((visibility("hidden")));
#define gtk_combo_box_set_active IA__gtk_combo_box_set_active

extern __typeof (gtk_combo_box_set_active_iter) IA__gtk_combo_box_set_active_iter __attribute((visibility("hidden")));
#define gtk_combo_box_set_active_iter IA__gtk_combo_box_set_active_iter

extern __typeof (gtk_combo_box_set_add_tearoffs) IA__gtk_combo_box_set_add_tearoffs __attribute((visibility("hidden")));
#define gtk_combo_box_set_add_tearoffs IA__gtk_combo_box_set_add_tearoffs

extern __typeof (gtk_combo_box_set_column_span_column) IA__gtk_combo_box_set_column_span_column __attribute((visibility("hidden")));
#define gtk_combo_box_set_column_span_column IA__gtk_combo_box_set_column_span_column

extern __typeof (gtk_combo_box_set_entry_text_column) IA__gtk_combo_box_set_entry_text_column __attribute((visibility("hidden")));
#define gtk_combo_box_set_entry_text_column IA__gtk_combo_box_set_entry_text_column

extern __typeof (gtk_combo_box_set_focus_on_click) IA__gtk_combo_box_set_focus_on_click __attribute((visibility("hidden")));
#define gtk_combo_box_set_focus_on_click IA__gtk_combo_box_set_focus_on_click

extern __typeof (gtk_combo_box_set_model) IA__gtk_combo_box_set_model __attribute((visibility("hidden")));
#define gtk_combo_box_set_model IA__gtk_combo_box_set_model

extern __typeof (gtk_combo_box_set_row_separator_func) IA__gtk_combo_box_set_row_separator_func __attribute((visibility("hidden")));
#define gtk_combo_box_set_row_separator_func IA__gtk_combo_box_set_row_separator_func

extern __typeof (gtk_combo_box_set_row_span_column) IA__gtk_combo_box_set_row_span_column __attribute((visibility("hidden")));
#define gtk_combo_box_set_row_span_column IA__gtk_combo_box_set_row_span_column

extern __typeof (gtk_combo_box_set_button_sensitivity) IA__gtk_combo_box_set_button_sensitivity __attribute((visibility("hidden")));
#define gtk_combo_box_set_button_sensitivity IA__gtk_combo_box_set_button_sensitivity

extern __typeof (gtk_combo_box_set_title) IA__gtk_combo_box_set_title __attribute((visibility("hidden")));
#define gtk_combo_box_set_title IA__gtk_combo_box_set_title

extern __typeof (gtk_combo_box_set_wrap_width) IA__gtk_combo_box_set_wrap_width __attribute((visibility("hidden")));
#define gtk_combo_box_set_wrap_width IA__gtk_combo_box_set_wrap_width

#endif
#endif
#if IN_HEADER(__GTK_COMBO_BOX_TEXT_H__)
#if IN_FILE(__GTK_COMBO_BOX_TEXT_C__)
extern __typeof (gtk_combo_box_text_append_text) IA__gtk_combo_box_text_append_text __attribute((visibility("hidden")));
#define gtk_combo_box_text_append_text IA__gtk_combo_box_text_append_text

extern __typeof (gtk_combo_box_text_get_active_text) IA__gtk_combo_box_text_get_active_text __attribute((visibility("hidden")));
#define gtk_combo_box_text_get_active_text IA__gtk_combo_box_text_get_active_text

extern __typeof (gtk_combo_box_text_get_type) IA__gtk_combo_box_text_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_combo_box_text_get_type IA__gtk_combo_box_text_get_type

extern __typeof (gtk_combo_box_text_insert_text) IA__gtk_combo_box_text_insert_text __attribute((visibility("hidden")));
#define gtk_combo_box_text_insert_text IA__gtk_combo_box_text_insert_text

extern __typeof (gtk_combo_box_text_new) IA__gtk_combo_box_text_new __attribute((visibility("hidden")));
#define gtk_combo_box_text_new IA__gtk_combo_box_text_new

extern __typeof (gtk_combo_box_text_new_with_entry) IA__gtk_combo_box_text_new_with_entry __attribute((visibility("hidden")));
#define gtk_combo_box_text_new_with_entry IA__gtk_combo_box_text_new_with_entry

extern __typeof (gtk_combo_box_text_prepend_text) IA__gtk_combo_box_text_prepend_text __attribute((visibility("hidden")));
#define gtk_combo_box_text_prepend_text IA__gtk_combo_box_text_prepend_text

extern __typeof (gtk_combo_box_text_remove) IA__gtk_combo_box_text_remove __attribute((visibility("hidden")));
#define gtk_combo_box_text_remove IA__gtk_combo_box_text_remove

#endif
#endif
#if IN_HEADER(__GTK_COMBO_BOX_ENTRY_H__)
#if IN_FILE(__GTK_COMBO_BOX_ENTRY_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_combo_box_entry_get_text_column) IA__gtk_combo_box_entry_get_text_column __attribute((visibility("hidden")));
#define gtk_combo_box_entry_get_text_column IA__gtk_combo_box_entry_get_text_column

extern __typeof (gtk_combo_box_entry_get_type) IA__gtk_combo_box_entry_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_combo_box_entry_get_type IA__gtk_combo_box_entry_get_type

extern __typeof (gtk_combo_box_entry_new) IA__gtk_combo_box_entry_new __attribute((visibility("hidden")));
#define gtk_combo_box_entry_new IA__gtk_combo_box_entry_new

extern __typeof (gtk_combo_box_entry_new_text) IA__gtk_combo_box_entry_new_text __attribute((visibility("hidden")));
#define gtk_combo_box_entry_new_text IA__gtk_combo_box_entry_new_text

extern __typeof (gtk_combo_box_entry_new_with_model) IA__gtk_combo_box_entry_new_with_model __attribute((visibility("hidden")));
#define gtk_combo_box_entry_new_with_model IA__gtk_combo_box_entry_new_with_model

extern __typeof (gtk_combo_box_entry_set_text_column) IA__gtk_combo_box_entry_set_text_column __attribute((visibility("hidden")));
#define gtk_combo_box_entry_set_text_column IA__gtk_combo_box_entry_set_text_column

#endif
#endif
#endif
#if IN_HEADER(__GTK_SMART_COMBO_H__)
#if IN_FILE(__GTK_SMART_COMBO_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_combo_disable_activate) IA__gtk_combo_disable_activate __attribute((visibility("hidden")));
#define gtk_combo_disable_activate IA__gtk_combo_disable_activate

extern __typeof (gtk_combo_get_type) IA__gtk_combo_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_combo_get_type IA__gtk_combo_get_type

extern __typeof (gtk_combo_new) IA__gtk_combo_new __attribute((visibility("hidden")));
#define gtk_combo_new IA__gtk_combo_new

extern __typeof (gtk_combo_set_case_sensitive) IA__gtk_combo_set_case_sensitive __attribute((visibility("hidden")));
#define gtk_combo_set_case_sensitive IA__gtk_combo_set_case_sensitive

extern __typeof (gtk_combo_set_item_string) IA__gtk_combo_set_item_string __attribute((visibility("hidden")));
#define gtk_combo_set_item_string IA__gtk_combo_set_item_string

extern __typeof (gtk_combo_set_popdown_strings) IA__gtk_combo_set_popdown_strings __attribute((visibility("hidden")));
#define gtk_combo_set_popdown_strings IA__gtk_combo_set_popdown_strings

extern __typeof (gtk_combo_set_use_arrows) IA__gtk_combo_set_use_arrows __attribute((visibility("hidden")));
#define gtk_combo_set_use_arrows IA__gtk_combo_set_use_arrows

extern __typeof (gtk_combo_set_use_arrows_always) IA__gtk_combo_set_use_arrows_always __attribute((visibility("hidden")));
#define gtk_combo_set_use_arrows_always IA__gtk_combo_set_use_arrows_always

extern __typeof (gtk_combo_set_value_in_list) IA__gtk_combo_set_value_in_list __attribute((visibility("hidden")));
#define gtk_combo_set_value_in_list IA__gtk_combo_set_value_in_list

#endif
#endif
#endif
#if IN_HEADER(__GTK_CONTAINER_H__)
#if IN_FILE(__GTK_CONTAINER_C__)
extern __typeof (gtk_container_add) IA__gtk_container_add __attribute((visibility("hidden")));
#define gtk_container_add IA__gtk_container_add

extern __typeof (gtk_container_add_with_properties) IA__gtk_container_add_with_properties __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_container_add_with_properties IA__gtk_container_add_with_properties

extern __typeof (gtk_container_check_resize) IA__gtk_container_check_resize __attribute((visibility("hidden")));
#define gtk_container_check_resize IA__gtk_container_check_resize

extern __typeof (gtk_container_child_get) IA__gtk_container_child_get __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_container_child_get IA__gtk_container_child_get

extern __typeof (gtk_container_child_get_property) IA__gtk_container_child_get_property __attribute((visibility("hidden")));
#define gtk_container_child_get_property IA__gtk_container_child_get_property

extern __typeof (gtk_container_child_get_valist) IA__gtk_container_child_get_valist __attribute((visibility("hidden")));
#define gtk_container_child_get_valist IA__gtk_container_child_get_valist

extern __typeof (gtk_container_child_set) IA__gtk_container_child_set __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_container_child_set IA__gtk_container_child_set

extern __typeof (gtk_container_child_set_property) IA__gtk_container_child_set_property __attribute((visibility("hidden")));
#define gtk_container_child_set_property IA__gtk_container_child_set_property

extern __typeof (gtk_container_child_set_valist) IA__gtk_container_child_set_valist __attribute((visibility("hidden")));
#define gtk_container_child_set_valist IA__gtk_container_child_set_valist

extern __typeof (gtk_container_child_type) IA__gtk_container_child_type __attribute((visibility("hidden")));
#define gtk_container_child_type IA__gtk_container_child_type

extern __typeof (gtk_container_class_find_child_property) IA__gtk_container_class_find_child_property __attribute((visibility("hidden")));
#define gtk_container_class_find_child_property IA__gtk_container_class_find_child_property

extern __typeof (gtk_container_class_install_child_property) IA__gtk_container_class_install_child_property __attribute((visibility("hidden")));
#define gtk_container_class_install_child_property IA__gtk_container_class_install_child_property

extern __typeof (gtk_container_class_list_child_properties) IA__gtk_container_class_list_child_properties __attribute((visibility("hidden")));
#define gtk_container_class_list_child_properties IA__gtk_container_class_list_child_properties

extern __typeof (gtk_container_forall) IA__gtk_container_forall __attribute((visibility("hidden")));
#define gtk_container_forall IA__gtk_container_forall

extern __typeof (gtk_container_foreach) IA__gtk_container_foreach __attribute((visibility("hidden")));
#define gtk_container_foreach IA__gtk_container_foreach

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_container_foreach_full) IA__gtk_container_foreach_full __attribute((visibility("hidden")));
#define gtk_container_foreach_full IA__gtk_container_foreach_full

#endif
extern __typeof (gtk_container_get_border_width) IA__gtk_container_get_border_width __attribute((visibility("hidden")));
#define gtk_container_get_border_width IA__gtk_container_get_border_width

extern __typeof (gtk_container_get_children) IA__gtk_container_get_children __attribute((visibility("hidden")));
#define gtk_container_get_children IA__gtk_container_get_children

extern __typeof (gtk_container_get_focus_chain) IA__gtk_container_get_focus_chain __attribute((visibility("hidden")));
#define gtk_container_get_focus_chain IA__gtk_container_get_focus_chain

extern __typeof (gtk_container_get_focus_child) IA__gtk_container_get_focus_child __attribute((visibility("hidden")));
#define gtk_container_get_focus_child IA__gtk_container_get_focus_child

extern __typeof (gtk_container_get_focus_hadjustment) IA__gtk_container_get_focus_hadjustment __attribute((visibility("hidden")));
#define gtk_container_get_focus_hadjustment IA__gtk_container_get_focus_hadjustment

extern __typeof (gtk_container_get_focus_vadjustment) IA__gtk_container_get_focus_vadjustment __attribute((visibility("hidden")));
#define gtk_container_get_focus_vadjustment IA__gtk_container_get_focus_vadjustment

extern __typeof (gtk_container_get_resize_mode) IA__gtk_container_get_resize_mode __attribute((visibility("hidden")));
#define gtk_container_get_resize_mode IA__gtk_container_get_resize_mode

extern __typeof (gtk_container_get_type) IA__gtk_container_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_container_get_type IA__gtk_container_get_type

extern __typeof (gtk_container_propagate_expose) IA__gtk_container_propagate_expose __attribute((visibility("hidden")));
#define gtk_container_propagate_expose IA__gtk_container_propagate_expose

extern __typeof (gtk_container_remove) IA__gtk_container_remove __attribute((visibility("hidden")));
#define gtk_container_remove IA__gtk_container_remove

extern __typeof (gtk_container_resize_children) IA__gtk_container_resize_children __attribute((visibility("hidden")));
#define gtk_container_resize_children IA__gtk_container_resize_children

extern __typeof (gtk_container_set_border_width) IA__gtk_container_set_border_width __attribute((visibility("hidden")));
#define gtk_container_set_border_width IA__gtk_container_set_border_width

extern __typeof (gtk_container_set_focus_chain) IA__gtk_container_set_focus_chain __attribute((visibility("hidden")));
#define gtk_container_set_focus_chain IA__gtk_container_set_focus_chain

extern __typeof (gtk_container_set_focus_child) IA__gtk_container_set_focus_child __attribute((visibility("hidden")));
#define gtk_container_set_focus_child IA__gtk_container_set_focus_child

extern __typeof (gtk_container_set_focus_hadjustment) IA__gtk_container_set_focus_hadjustment __attribute((visibility("hidden")));
#define gtk_container_set_focus_hadjustment IA__gtk_container_set_focus_hadjustment

extern __typeof (gtk_container_set_focus_vadjustment) IA__gtk_container_set_focus_vadjustment __attribute((visibility("hidden")));
#define gtk_container_set_focus_vadjustment IA__gtk_container_set_focus_vadjustment

extern __typeof (gtk_container_set_reallocate_redraws) IA__gtk_container_set_reallocate_redraws __attribute((visibility("hidden")));
#define gtk_container_set_reallocate_redraws IA__gtk_container_set_reallocate_redraws

extern __typeof (gtk_container_set_resize_mode) IA__gtk_container_set_resize_mode __attribute((visibility("hidden")));
#define gtk_container_set_resize_mode IA__gtk_container_set_resize_mode

extern __typeof (gtk_container_unset_focus_chain) IA__gtk_container_unset_focus_chain __attribute((visibility("hidden")));
#define gtk_container_unset_focus_chain IA__gtk_container_unset_focus_chain

#endif
#endif
#if IN_HEADER(__GTK_CTREE_H__)
#if IN_FILE(__GTK_CTREE_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_ctree_collapse) IA__gtk_ctree_collapse __attribute((visibility("hidden")));
#define gtk_ctree_collapse IA__gtk_ctree_collapse

extern __typeof (gtk_ctree_collapse_recursive) IA__gtk_ctree_collapse_recursive __attribute((visibility("hidden")));
#define gtk_ctree_collapse_recursive IA__gtk_ctree_collapse_recursive

extern __typeof (gtk_ctree_collapse_to_depth) IA__gtk_ctree_collapse_to_depth __attribute((visibility("hidden")));
#define gtk_ctree_collapse_to_depth IA__gtk_ctree_collapse_to_depth

extern __typeof (gtk_ctree_expand) IA__gtk_ctree_expand __attribute((visibility("hidden")));
#define gtk_ctree_expand IA__gtk_ctree_expand

extern __typeof (gtk_ctree_expand_recursive) IA__gtk_ctree_expand_recursive __attribute((visibility("hidden")));
#define gtk_ctree_expand_recursive IA__gtk_ctree_expand_recursive

extern __typeof (gtk_ctree_expand_to_depth) IA__gtk_ctree_expand_to_depth __attribute((visibility("hidden")));
#define gtk_ctree_expand_to_depth IA__gtk_ctree_expand_to_depth

extern __typeof (gtk_ctree_export_to_gnode) IA__gtk_ctree_export_to_gnode __attribute((visibility("hidden")));
#define gtk_ctree_export_to_gnode IA__gtk_ctree_export_to_gnode

extern __typeof (gtk_ctree_find) IA__gtk_ctree_find __attribute((visibility("hidden")));
#define gtk_ctree_find IA__gtk_ctree_find

extern __typeof (gtk_ctree_find_all_by_row_data) IA__gtk_ctree_find_all_by_row_data __attribute((visibility("hidden")));
#define gtk_ctree_find_all_by_row_data IA__gtk_ctree_find_all_by_row_data

extern __typeof (gtk_ctree_find_all_by_row_data_custom) IA__gtk_ctree_find_all_by_row_data_custom __attribute((visibility("hidden")));
#define gtk_ctree_find_all_by_row_data_custom IA__gtk_ctree_find_all_by_row_data_custom

extern __typeof (gtk_ctree_find_by_row_data) IA__gtk_ctree_find_by_row_data __attribute((visibility("hidden")));
#define gtk_ctree_find_by_row_data IA__gtk_ctree_find_by_row_data

extern __typeof (gtk_ctree_find_by_row_data_custom) IA__gtk_ctree_find_by_row_data_custom __attribute((visibility("hidden")));
#define gtk_ctree_find_by_row_data_custom IA__gtk_ctree_find_by_row_data_custom

extern __typeof (gtk_ctree_find_node_ptr) IA__gtk_ctree_find_node_ptr __attribute((visibility("hidden")));
#define gtk_ctree_find_node_ptr IA__gtk_ctree_find_node_ptr

extern __typeof (gtk_ctree_get_node_info) IA__gtk_ctree_get_node_info __attribute((visibility("hidden")));
#define gtk_ctree_get_node_info IA__gtk_ctree_get_node_info

extern __typeof (gtk_ctree_get_type) IA__gtk_ctree_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_ctree_get_type IA__gtk_ctree_get_type

extern __typeof (gtk_ctree_insert_gnode) IA__gtk_ctree_insert_gnode __attribute((visibility("hidden")));
#define gtk_ctree_insert_gnode IA__gtk_ctree_insert_gnode

extern __typeof (gtk_ctree_insert_node) IA__gtk_ctree_insert_node __attribute((visibility("hidden")));
#define gtk_ctree_insert_node IA__gtk_ctree_insert_node

extern __typeof (gtk_ctree_is_ancestor) IA__gtk_ctree_is_ancestor __attribute((visibility("hidden")));
#define gtk_ctree_is_ancestor IA__gtk_ctree_is_ancestor

extern __typeof (gtk_ctree_is_hot_spot) IA__gtk_ctree_is_hot_spot __attribute((visibility("hidden")));
#define gtk_ctree_is_hot_spot IA__gtk_ctree_is_hot_spot

extern __typeof (gtk_ctree_is_viewable) IA__gtk_ctree_is_viewable __attribute((visibility("hidden")));
#define gtk_ctree_is_viewable IA__gtk_ctree_is_viewable

extern __typeof (gtk_ctree_last) IA__gtk_ctree_last __attribute((visibility("hidden")));
#define gtk_ctree_last IA__gtk_ctree_last

extern __typeof (gtk_ctree_move) IA__gtk_ctree_move __attribute((visibility("hidden")));
#define gtk_ctree_move IA__gtk_ctree_move

extern __typeof (gtk_ctree_new) IA__gtk_ctree_new __attribute((visibility("hidden")));
#define gtk_ctree_new IA__gtk_ctree_new

extern __typeof (gtk_ctree_new_with_titles) IA__gtk_ctree_new_with_titles __attribute((visibility("hidden")));
#define gtk_ctree_new_with_titles IA__gtk_ctree_new_with_titles

extern __typeof (gtk_ctree_node_get_cell_style) IA__gtk_ctree_node_get_cell_style __attribute((visibility("hidden")));
#define gtk_ctree_node_get_cell_style IA__gtk_ctree_node_get_cell_style

extern __typeof (gtk_ctree_node_get_cell_type) IA__gtk_ctree_node_get_cell_type __attribute((visibility("hidden")));
#define gtk_ctree_node_get_cell_type IA__gtk_ctree_node_get_cell_type

extern __typeof (gtk_ctree_node_get_pixmap) IA__gtk_ctree_node_get_pixmap __attribute((visibility("hidden")));
#define gtk_ctree_node_get_pixmap IA__gtk_ctree_node_get_pixmap

extern __typeof (gtk_ctree_node_get_pixtext) IA__gtk_ctree_node_get_pixtext __attribute((visibility("hidden")));
#define gtk_ctree_node_get_pixtext IA__gtk_ctree_node_get_pixtext

extern __typeof (gtk_ctree_node_get_row_data) IA__gtk_ctree_node_get_row_data __attribute((visibility("hidden")));
#define gtk_ctree_node_get_row_data IA__gtk_ctree_node_get_row_data

extern __typeof (gtk_ctree_node_get_row_style) IA__gtk_ctree_node_get_row_style __attribute((visibility("hidden")));
#define gtk_ctree_node_get_row_style IA__gtk_ctree_node_get_row_style

extern __typeof (gtk_ctree_node_get_selectable) IA__gtk_ctree_node_get_selectable __attribute((visibility("hidden")));
#define gtk_ctree_node_get_selectable IA__gtk_ctree_node_get_selectable

extern __typeof (gtk_ctree_node_get_text) IA__gtk_ctree_node_get_text __attribute((visibility("hidden")));
#define gtk_ctree_node_get_text IA__gtk_ctree_node_get_text

extern __typeof (gtk_ctree_node_get_type) IA__gtk_ctree_node_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_ctree_node_get_type IA__gtk_ctree_node_get_type

extern __typeof (gtk_ctree_node_is_visible) IA__gtk_ctree_node_is_visible __attribute((visibility("hidden")));
#define gtk_ctree_node_is_visible IA__gtk_ctree_node_is_visible

extern __typeof (gtk_ctree_node_moveto) IA__gtk_ctree_node_moveto __attribute((visibility("hidden")));
#define gtk_ctree_node_moveto IA__gtk_ctree_node_moveto

extern __typeof (gtk_ctree_node_nth) IA__gtk_ctree_node_nth __attribute((visibility("hidden")));
#define gtk_ctree_node_nth IA__gtk_ctree_node_nth

extern __typeof (gtk_ctree_node_set_background) IA__gtk_ctree_node_set_background __attribute((visibility("hidden")));
#define gtk_ctree_node_set_background IA__gtk_ctree_node_set_background

extern __typeof (gtk_ctree_node_set_cell_style) IA__gtk_ctree_node_set_cell_style __attribute((visibility("hidden")));
#define gtk_ctree_node_set_cell_style IA__gtk_ctree_node_set_cell_style

extern __typeof (gtk_ctree_node_set_foreground) IA__gtk_ctree_node_set_foreground __attribute((visibility("hidden")));
#define gtk_ctree_node_set_foreground IA__gtk_ctree_node_set_foreground

extern __typeof (gtk_ctree_node_set_pixmap) IA__gtk_ctree_node_set_pixmap __attribute((visibility("hidden")));
#define gtk_ctree_node_set_pixmap IA__gtk_ctree_node_set_pixmap

extern __typeof (gtk_ctree_node_set_pixtext) IA__gtk_ctree_node_set_pixtext __attribute((visibility("hidden")));
#define gtk_ctree_node_set_pixtext IA__gtk_ctree_node_set_pixtext

extern __typeof (gtk_ctree_node_set_row_data) IA__gtk_ctree_node_set_row_data __attribute((visibility("hidden")));
#define gtk_ctree_node_set_row_data IA__gtk_ctree_node_set_row_data

extern __typeof (gtk_ctree_node_set_row_data_full) IA__gtk_ctree_node_set_row_data_full __attribute((visibility("hidden")));
#define gtk_ctree_node_set_row_data_full IA__gtk_ctree_node_set_row_data_full

extern __typeof (gtk_ctree_node_set_row_style) IA__gtk_ctree_node_set_row_style __attribute((visibility("hidden")));
#define gtk_ctree_node_set_row_style IA__gtk_ctree_node_set_row_style

extern __typeof (gtk_ctree_node_set_selectable) IA__gtk_ctree_node_set_selectable __attribute((visibility("hidden")));
#define gtk_ctree_node_set_selectable IA__gtk_ctree_node_set_selectable

extern __typeof (gtk_ctree_node_set_shift) IA__gtk_ctree_node_set_shift __attribute((visibility("hidden")));
#define gtk_ctree_node_set_shift IA__gtk_ctree_node_set_shift

extern __typeof (gtk_ctree_node_set_text) IA__gtk_ctree_node_set_text __attribute((visibility("hidden")));
#define gtk_ctree_node_set_text IA__gtk_ctree_node_set_text

extern __typeof (gtk_ctree_post_recursive) IA__gtk_ctree_post_recursive __attribute((visibility("hidden")));
#define gtk_ctree_post_recursive IA__gtk_ctree_post_recursive

extern __typeof (gtk_ctree_post_recursive_to_depth) IA__gtk_ctree_post_recursive_to_depth __attribute((visibility("hidden")));
#define gtk_ctree_post_recursive_to_depth IA__gtk_ctree_post_recursive_to_depth

extern __typeof (gtk_ctree_pre_recursive) IA__gtk_ctree_pre_recursive __attribute((visibility("hidden")));
#define gtk_ctree_pre_recursive IA__gtk_ctree_pre_recursive

extern __typeof (gtk_ctree_pre_recursive_to_depth) IA__gtk_ctree_pre_recursive_to_depth __attribute((visibility("hidden")));
#define gtk_ctree_pre_recursive_to_depth IA__gtk_ctree_pre_recursive_to_depth

extern __typeof (gtk_ctree_real_select_recursive) IA__gtk_ctree_real_select_recursive __attribute((visibility("hidden")));
#define gtk_ctree_real_select_recursive IA__gtk_ctree_real_select_recursive

extern __typeof (gtk_ctree_remove_node) IA__gtk_ctree_remove_node __attribute((visibility("hidden")));
#define gtk_ctree_remove_node IA__gtk_ctree_remove_node

extern __typeof (gtk_ctree_select) IA__gtk_ctree_select __attribute((visibility("hidden")));
#define gtk_ctree_select IA__gtk_ctree_select

extern __typeof (gtk_ctree_select_recursive) IA__gtk_ctree_select_recursive __attribute((visibility("hidden")));
#define gtk_ctree_select_recursive IA__gtk_ctree_select_recursive

extern __typeof (gtk_ctree_set_drag_compare_func) IA__gtk_ctree_set_drag_compare_func __attribute((visibility("hidden")));
#define gtk_ctree_set_drag_compare_func IA__gtk_ctree_set_drag_compare_func

extern __typeof (gtk_ctree_set_expander_style) IA__gtk_ctree_set_expander_style __attribute((visibility("hidden")));
#define gtk_ctree_set_expander_style IA__gtk_ctree_set_expander_style

extern __typeof (gtk_ctree_set_indent) IA__gtk_ctree_set_indent __attribute((visibility("hidden")));
#define gtk_ctree_set_indent IA__gtk_ctree_set_indent

extern __typeof (gtk_ctree_set_line_style) IA__gtk_ctree_set_line_style __attribute((visibility("hidden")));
#define gtk_ctree_set_line_style IA__gtk_ctree_set_line_style

extern __typeof (gtk_ctree_set_node_info) IA__gtk_ctree_set_node_info __attribute((visibility("hidden")));
#define gtk_ctree_set_node_info IA__gtk_ctree_set_node_info

extern __typeof (gtk_ctree_set_show_stub) IA__gtk_ctree_set_show_stub __attribute((visibility("hidden")));
#define gtk_ctree_set_show_stub IA__gtk_ctree_set_show_stub

extern __typeof (gtk_ctree_set_spacing) IA__gtk_ctree_set_spacing __attribute((visibility("hidden")));
#define gtk_ctree_set_spacing IA__gtk_ctree_set_spacing

extern __typeof (gtk_ctree_sort_node) IA__gtk_ctree_sort_node __attribute((visibility("hidden")));
#define gtk_ctree_sort_node IA__gtk_ctree_sort_node

extern __typeof (gtk_ctree_sort_recursive) IA__gtk_ctree_sort_recursive __attribute((visibility("hidden")));
#define gtk_ctree_sort_recursive IA__gtk_ctree_sort_recursive

extern __typeof (gtk_ctree_toggle_expansion) IA__gtk_ctree_toggle_expansion __attribute((visibility("hidden")));
#define gtk_ctree_toggle_expansion IA__gtk_ctree_toggle_expansion

extern __typeof (gtk_ctree_toggle_expansion_recursive) IA__gtk_ctree_toggle_expansion_recursive __attribute((visibility("hidden")));
#define gtk_ctree_toggle_expansion_recursive IA__gtk_ctree_toggle_expansion_recursive

extern __typeof (gtk_ctree_unselect) IA__gtk_ctree_unselect __attribute((visibility("hidden")));
#define gtk_ctree_unselect IA__gtk_ctree_unselect

extern __typeof (gtk_ctree_unselect_recursive) IA__gtk_ctree_unselect_recursive __attribute((visibility("hidden")));
#define gtk_ctree_unselect_recursive IA__gtk_ctree_unselect_recursive

#endif
#endif
#endif
#if IN_HEADER(__GTK_CURVE_H__)
#if IN_FILE(__GTK_CURVE_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_curve_get_type) IA__gtk_curve_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_curve_get_type IA__gtk_curve_get_type

extern __typeof (gtk_curve_get_vector) IA__gtk_curve_get_vector __attribute((visibility("hidden")));
#define gtk_curve_get_vector IA__gtk_curve_get_vector

extern __typeof (gtk_curve_new) IA__gtk_curve_new __attribute((visibility("hidden")));
#define gtk_curve_new IA__gtk_curve_new

extern __typeof (gtk_curve_reset) IA__gtk_curve_reset __attribute((visibility("hidden")));
#define gtk_curve_reset IA__gtk_curve_reset

extern __typeof (gtk_curve_set_curve_type) IA__gtk_curve_set_curve_type __attribute((visibility("hidden")));
#define gtk_curve_set_curve_type IA__gtk_curve_set_curve_type

extern __typeof (gtk_curve_set_gamma) IA__gtk_curve_set_gamma __attribute((visibility("hidden")));
#define gtk_curve_set_gamma IA__gtk_curve_set_gamma

extern __typeof (gtk_curve_set_range) IA__gtk_curve_set_range __attribute((visibility("hidden")));
#define gtk_curve_set_range IA__gtk_curve_set_range

extern __typeof (gtk_curve_set_vector) IA__gtk_curve_set_vector __attribute((visibility("hidden")));
#define gtk_curve_set_vector IA__gtk_curve_set_vector

#endif
#endif
#endif
#if IN_HEADER(__GTK_WINDOW_DECORATE_H__)
#if IN_FILE(__GTK_WINDOW_DECORATE_C__)
extern __typeof (gtk_decorated_window_calculate_frame_size) IA__gtk_decorated_window_calculate_frame_size __attribute((visibility("hidden")));
#define gtk_decorated_window_calculate_frame_size IA__gtk_decorated_window_calculate_frame_size

extern __typeof (gtk_decorated_window_init) IA__gtk_decorated_window_init __attribute((visibility("hidden")));
#define gtk_decorated_window_init IA__gtk_decorated_window_init

extern __typeof (gtk_decorated_window_move_resize_window) IA__gtk_decorated_window_move_resize_window __attribute((visibility("hidden")));
#define gtk_decorated_window_move_resize_window IA__gtk_decorated_window_move_resize_window

extern __typeof (gtk_decorated_window_set_title) IA__gtk_decorated_window_set_title __attribute((visibility("hidden")));
#define gtk_decorated_window_set_title IA__gtk_decorated_window_set_title

#endif
#endif
#if IN_HEADER(__GTK_DIALOG_H__)
#if IN_FILE(__GTK_DIALOG_C__)
extern __typeof (gtk_dialog_add_action_widget) IA__gtk_dialog_add_action_widget __attribute((visibility("hidden")));
#define gtk_dialog_add_action_widget IA__gtk_dialog_add_action_widget

extern __typeof (gtk_dialog_add_button) IA__gtk_dialog_add_button __attribute((visibility("hidden")));
#define gtk_dialog_add_button IA__gtk_dialog_add_button

extern __typeof (gtk_dialog_add_buttons) IA__gtk_dialog_add_buttons __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_dialog_add_buttons IA__gtk_dialog_add_buttons

extern __typeof (gtk_dialog_get_action_area) IA__gtk_dialog_get_action_area __attribute((visibility("hidden")));
#define gtk_dialog_get_action_area IA__gtk_dialog_get_action_area

extern __typeof (gtk_dialog_get_content_area) IA__gtk_dialog_get_content_area __attribute((visibility("hidden")));
#define gtk_dialog_get_content_area IA__gtk_dialog_get_content_area

extern __typeof (gtk_dialog_get_widget_for_response) IA__gtk_dialog_get_widget_for_response __attribute((visibility("hidden")));
#define gtk_dialog_get_widget_for_response IA__gtk_dialog_get_widget_for_response

extern __typeof (gtk_dialog_get_response_for_widget) IA__gtk_dialog_get_response_for_widget __attribute((visibility("hidden")));
#define gtk_dialog_get_response_for_widget IA__gtk_dialog_get_response_for_widget

extern __typeof (gtk_dialog_get_type) IA__gtk_dialog_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_dialog_get_type IA__gtk_dialog_get_type

extern __typeof (gtk_dialog_new) IA__gtk_dialog_new __attribute((visibility("hidden")));
#define gtk_dialog_new IA__gtk_dialog_new

extern __typeof (gtk_dialog_new_with_buttons) IA__gtk_dialog_new_with_buttons __attribute((visibility("hidden")));
#define gtk_dialog_new_with_buttons IA__gtk_dialog_new_with_buttons

extern __typeof (gtk_dialog_response) IA__gtk_dialog_response __attribute((visibility("hidden")));
#define gtk_dialog_response IA__gtk_dialog_response

extern __typeof (gtk_dialog_run) IA__gtk_dialog_run __attribute((visibility("hidden")));
#define gtk_dialog_run IA__gtk_dialog_run

extern __typeof (gtk_alternative_dialog_button_order) IA__gtk_alternative_dialog_button_order __attribute((visibility("hidden")));
#define gtk_alternative_dialog_button_order IA__gtk_alternative_dialog_button_order

extern __typeof (gtk_dialog_set_alternative_button_order) IA__gtk_dialog_set_alternative_button_order __attribute((visibility("hidden")));
#define gtk_dialog_set_alternative_button_order IA__gtk_dialog_set_alternative_button_order

extern __typeof (gtk_dialog_set_alternative_button_order_from_array) IA__gtk_dialog_set_alternative_button_order_from_array __attribute((visibility("hidden")));
#define gtk_dialog_set_alternative_button_order_from_array IA__gtk_dialog_set_alternative_button_order_from_array

extern __typeof (gtk_dialog_set_default_response) IA__gtk_dialog_set_default_response __attribute((visibility("hidden")));
#define gtk_dialog_set_default_response IA__gtk_dialog_set_default_response

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_dialog_get_has_separator) IA__gtk_dialog_get_has_separator __attribute((visibility("hidden")));
#define gtk_dialog_get_has_separator IA__gtk_dialog_get_has_separator

extern __typeof (gtk_dialog_set_has_separator) IA__gtk_dialog_set_has_separator __attribute((visibility("hidden")));
#define gtk_dialog_set_has_separator IA__gtk_dialog_set_has_separator

#endif
extern __typeof (gtk_dialog_set_response_sensitive) IA__gtk_dialog_set_response_sensitive __attribute((visibility("hidden")));
#define gtk_dialog_set_response_sensitive IA__gtk_dialog_set_response_sensitive

#endif
#endif
#if IN_HEADER(__GTK_DND_H__)
#if IN_FILE(__GTK_DND_C__)
extern __typeof (gtk_drag_begin) IA__gtk_drag_begin __attribute((visibility("hidden")));
#define gtk_drag_begin IA__gtk_drag_begin

extern __typeof (gtk_drag_check_threshold) IA__gtk_drag_check_threshold __attribute((visibility("hidden")));
#define gtk_drag_check_threshold IA__gtk_drag_check_threshold

extern __typeof (gtk_drag_dest_add_image_targets) IA__gtk_drag_dest_add_image_targets __attribute((visibility("hidden")));
#define gtk_drag_dest_add_image_targets IA__gtk_drag_dest_add_image_targets

extern __typeof (gtk_drag_dest_add_text_targets) IA__gtk_drag_dest_add_text_targets __attribute((visibility("hidden")));
#define gtk_drag_dest_add_text_targets IA__gtk_drag_dest_add_text_targets

extern __typeof (gtk_drag_dest_add_uri_targets) IA__gtk_drag_dest_add_uri_targets __attribute((visibility("hidden")));
#define gtk_drag_dest_add_uri_targets IA__gtk_drag_dest_add_uri_targets

extern __typeof (gtk_drag_dest_find_target) IA__gtk_drag_dest_find_target __attribute((visibility("hidden")));
#define gtk_drag_dest_find_target IA__gtk_drag_dest_find_target

extern __typeof (gtk_drag_dest_get_target_list) IA__gtk_drag_dest_get_target_list __attribute((visibility("hidden")));
#define gtk_drag_dest_get_target_list IA__gtk_drag_dest_get_target_list

extern __typeof (gtk_drag_dest_set) IA__gtk_drag_dest_set __attribute((visibility("hidden")));
#define gtk_drag_dest_set IA__gtk_drag_dest_set

extern __typeof (gtk_drag_dest_set_proxy) IA__gtk_drag_dest_set_proxy __attribute((visibility("hidden")));
#define gtk_drag_dest_set_proxy IA__gtk_drag_dest_set_proxy

extern __typeof (gtk_drag_dest_set_target_list) IA__gtk_drag_dest_set_target_list __attribute((visibility("hidden")));
#define gtk_drag_dest_set_target_list IA__gtk_drag_dest_set_target_list

extern __typeof (gtk_drag_dest_set_track_motion) IA__gtk_drag_dest_set_track_motion __attribute((visibility("hidden")));
#define gtk_drag_dest_set_track_motion IA__gtk_drag_dest_set_track_motion

extern __typeof (gtk_drag_dest_get_track_motion) IA__gtk_drag_dest_get_track_motion __attribute((visibility("hidden")));
#define gtk_drag_dest_get_track_motion IA__gtk_drag_dest_get_track_motion

extern __typeof (gtk_drag_dest_unset) IA__gtk_drag_dest_unset __attribute((visibility("hidden")));
#define gtk_drag_dest_unset IA__gtk_drag_dest_unset

extern __typeof (gtk_drag_finish) IA__gtk_drag_finish __attribute((visibility("hidden")));
#define gtk_drag_finish IA__gtk_drag_finish

extern __typeof (gtk_drag_get_data) IA__gtk_drag_get_data __attribute((visibility("hidden")));
#define gtk_drag_get_data IA__gtk_drag_get_data

extern __typeof (gtk_drag_get_source_widget) IA__gtk_drag_get_source_widget __attribute((visibility("hidden")));
#define gtk_drag_get_source_widget IA__gtk_drag_get_source_widget

extern __typeof (gtk_drag_highlight) IA__gtk_drag_highlight __attribute((visibility("hidden")));
#define gtk_drag_highlight IA__gtk_drag_highlight

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_drag_set_default_icon) IA__gtk_drag_set_default_icon __attribute((visibility("hidden")));
#define gtk_drag_set_default_icon IA__gtk_drag_set_default_icon

#endif
extern __typeof (gtk_drag_set_icon_default) IA__gtk_drag_set_icon_default __attribute((visibility("hidden")));
#define gtk_drag_set_icon_default IA__gtk_drag_set_icon_default

extern __typeof (gtk_drag_set_icon_name) IA__gtk_drag_set_icon_name __attribute((visibility("hidden")));
#define gtk_drag_set_icon_name IA__gtk_drag_set_icon_name

extern __typeof (gtk_drag_set_icon_pixbuf) IA__gtk_drag_set_icon_pixbuf __attribute((visibility("hidden")));
#define gtk_drag_set_icon_pixbuf IA__gtk_drag_set_icon_pixbuf

extern __typeof (gtk_drag_set_icon_pixmap) IA__gtk_drag_set_icon_pixmap __attribute((visibility("hidden")));
#define gtk_drag_set_icon_pixmap IA__gtk_drag_set_icon_pixmap

extern __typeof (gtk_drag_set_icon_stock) IA__gtk_drag_set_icon_stock __attribute((visibility("hidden")));
#define gtk_drag_set_icon_stock IA__gtk_drag_set_icon_stock

extern __typeof (gtk_drag_set_icon_widget) IA__gtk_drag_set_icon_widget __attribute((visibility("hidden")));
#define gtk_drag_set_icon_widget IA__gtk_drag_set_icon_widget

extern __typeof (gtk_drag_source_add_image_targets) IA__gtk_drag_source_add_image_targets __attribute((visibility("hidden")));
#define gtk_drag_source_add_image_targets IA__gtk_drag_source_add_image_targets

extern __typeof (gtk_drag_source_add_text_targets) IA__gtk_drag_source_add_text_targets __attribute((visibility("hidden")));
#define gtk_drag_source_add_text_targets IA__gtk_drag_source_add_text_targets

extern __typeof (gtk_drag_source_add_uri_targets) IA__gtk_drag_source_add_uri_targets __attribute((visibility("hidden")));
#define gtk_drag_source_add_uri_targets IA__gtk_drag_source_add_uri_targets

extern __typeof (gtk_drag_source_get_target_list) IA__gtk_drag_source_get_target_list __attribute((visibility("hidden")));
#define gtk_drag_source_get_target_list IA__gtk_drag_source_get_target_list

extern __typeof (gtk_drag_source_set) IA__gtk_drag_source_set __attribute((visibility("hidden")));
#define gtk_drag_source_set IA__gtk_drag_source_set

extern __typeof (gtk_drag_source_set_icon) IA__gtk_drag_source_set_icon __attribute((visibility("hidden")));
#define gtk_drag_source_set_icon IA__gtk_drag_source_set_icon

extern __typeof (gtk_drag_source_set_icon_name) IA__gtk_drag_source_set_icon_name __attribute((visibility("hidden")));
#define gtk_drag_source_set_icon_name IA__gtk_drag_source_set_icon_name

extern __typeof (gtk_drag_source_set_icon_pixbuf) IA__gtk_drag_source_set_icon_pixbuf __attribute((visibility("hidden")));
#define gtk_drag_source_set_icon_pixbuf IA__gtk_drag_source_set_icon_pixbuf

extern __typeof (gtk_drag_source_set_icon_stock) IA__gtk_drag_source_set_icon_stock __attribute((visibility("hidden")));
#define gtk_drag_source_set_icon_stock IA__gtk_drag_source_set_icon_stock

extern __typeof (gtk_drag_source_set_target_list) IA__gtk_drag_source_set_target_list __attribute((visibility("hidden")));
#define gtk_drag_source_set_target_list IA__gtk_drag_source_set_target_list

extern __typeof (gtk_drag_source_unset) IA__gtk_drag_source_unset __attribute((visibility("hidden")));
#define gtk_drag_source_unset IA__gtk_drag_source_unset

extern __typeof (gtk_drag_unhighlight) IA__gtk_drag_unhighlight __attribute((visibility("hidden")));
#define gtk_drag_unhighlight IA__gtk_drag_unhighlight

#endif
#endif
#if IN_HEADER(__GTK_STATUS_ICON_H__)
#if IN_FILE(__GTK_STATUS_ICON_C__)
extern __typeof (gtk_status_icon_get_type) IA__gtk_status_icon_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_status_icon_get_type IA__gtk_status_icon_get_type

extern __typeof (gtk_status_icon_new) IA__gtk_status_icon_new __attribute((visibility("hidden")));
#define gtk_status_icon_new IA__gtk_status_icon_new

extern __typeof (gtk_status_icon_new_from_pixbuf) IA__gtk_status_icon_new_from_pixbuf __attribute((visibility("hidden")));
#define gtk_status_icon_new_from_pixbuf IA__gtk_status_icon_new_from_pixbuf

extern __typeof (gtk_status_icon_new_from_file) IA__gtk_status_icon_new_from_file __attribute((visibility("hidden")));
#define gtk_status_icon_new_from_file IA__gtk_status_icon_new_from_file

extern __typeof (gtk_status_icon_new_from_stock) IA__gtk_status_icon_new_from_stock __attribute((visibility("hidden")));
#define gtk_status_icon_new_from_stock IA__gtk_status_icon_new_from_stock

extern __typeof (gtk_status_icon_new_from_icon_name) IA__gtk_status_icon_new_from_icon_name __attribute((visibility("hidden")));
#define gtk_status_icon_new_from_icon_name IA__gtk_status_icon_new_from_icon_name

extern __typeof (gtk_status_icon_new_from_gicon) IA__gtk_status_icon_new_from_gicon __attribute((visibility("hidden")));
#define gtk_status_icon_new_from_gicon IA__gtk_status_icon_new_from_gicon

extern __typeof (gtk_status_icon_set_from_pixbuf) IA__gtk_status_icon_set_from_pixbuf __attribute((visibility("hidden")));
#define gtk_status_icon_set_from_pixbuf IA__gtk_status_icon_set_from_pixbuf

extern __typeof (gtk_status_icon_set_from_file) IA__gtk_status_icon_set_from_file __attribute((visibility("hidden")));
#define gtk_status_icon_set_from_file IA__gtk_status_icon_set_from_file

extern __typeof (gtk_status_icon_set_from_stock) IA__gtk_status_icon_set_from_stock __attribute((visibility("hidden")));
#define gtk_status_icon_set_from_stock IA__gtk_status_icon_set_from_stock

extern __typeof (gtk_status_icon_set_from_icon_name) IA__gtk_status_icon_set_from_icon_name __attribute((visibility("hidden")));
#define gtk_status_icon_set_from_icon_name IA__gtk_status_icon_set_from_icon_name

extern __typeof (gtk_status_icon_set_from_gicon) IA__gtk_status_icon_set_from_gicon __attribute((visibility("hidden")));
#define gtk_status_icon_set_from_gicon IA__gtk_status_icon_set_from_gicon

extern __typeof (gtk_status_icon_get_storage_type) IA__gtk_status_icon_get_storage_type __attribute((visibility("hidden")));
#define gtk_status_icon_get_storage_type IA__gtk_status_icon_get_storage_type

extern __typeof (gtk_status_icon_get_pixbuf) IA__gtk_status_icon_get_pixbuf __attribute((visibility("hidden")));
#define gtk_status_icon_get_pixbuf IA__gtk_status_icon_get_pixbuf

extern __typeof (gtk_status_icon_get_screen) IA__gtk_status_icon_get_screen __attribute((visibility("hidden")));
#define gtk_status_icon_get_screen IA__gtk_status_icon_get_screen

extern __typeof (gtk_status_icon_get_stock) IA__gtk_status_icon_get_stock __attribute((visibility("hidden")));
#define gtk_status_icon_get_stock IA__gtk_status_icon_get_stock

extern __typeof (gtk_status_icon_get_icon_name) IA__gtk_status_icon_get_icon_name __attribute((visibility("hidden")));
#define gtk_status_icon_get_icon_name IA__gtk_status_icon_get_icon_name

extern __typeof (gtk_status_icon_get_gicon) IA__gtk_status_icon_get_gicon __attribute((visibility("hidden")));
#define gtk_status_icon_get_gicon IA__gtk_status_icon_get_gicon

extern __typeof (gtk_status_icon_get_size) IA__gtk_status_icon_get_size __attribute((visibility("hidden")));
#define gtk_status_icon_get_size IA__gtk_status_icon_get_size

extern __typeof (gtk_status_icon_set_screen) IA__gtk_status_icon_set_screen __attribute((visibility("hidden")));
#define gtk_status_icon_set_screen IA__gtk_status_icon_set_screen

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_status_icon_set_tooltip) IA__gtk_status_icon_set_tooltip __attribute((visibility("hidden")));
#define gtk_status_icon_set_tooltip IA__gtk_status_icon_set_tooltip

#endif
extern __typeof (gtk_status_icon_set_tooltip_text) IA__gtk_status_icon_set_tooltip_text __attribute((visibility("hidden")));
#define gtk_status_icon_set_tooltip_text IA__gtk_status_icon_set_tooltip_text

extern __typeof (gtk_status_icon_get_tooltip_text) IA__gtk_status_icon_get_tooltip_text __attribute((visibility("hidden")));
#define gtk_status_icon_get_tooltip_text IA__gtk_status_icon_get_tooltip_text

extern __typeof (gtk_status_icon_set_tooltip_markup) IA__gtk_status_icon_set_tooltip_markup __attribute((visibility("hidden")));
#define gtk_status_icon_set_tooltip_markup IA__gtk_status_icon_set_tooltip_markup

extern __typeof (gtk_status_icon_get_tooltip_markup) IA__gtk_status_icon_get_tooltip_markup __attribute((visibility("hidden")));
#define gtk_status_icon_get_tooltip_markup IA__gtk_status_icon_get_tooltip_markup

extern __typeof (gtk_status_icon_set_has_tooltip) IA__gtk_status_icon_set_has_tooltip __attribute((visibility("hidden")));
#define gtk_status_icon_set_has_tooltip IA__gtk_status_icon_set_has_tooltip

extern __typeof (gtk_status_icon_get_has_tooltip) IA__gtk_status_icon_get_has_tooltip __attribute((visibility("hidden")));
#define gtk_status_icon_get_has_tooltip IA__gtk_status_icon_get_has_tooltip

extern __typeof (gtk_status_icon_set_visible) IA__gtk_status_icon_set_visible __attribute((visibility("hidden")));
#define gtk_status_icon_set_visible IA__gtk_status_icon_set_visible

extern __typeof (gtk_status_icon_get_visible) IA__gtk_status_icon_get_visible __attribute((visibility("hidden")));
#define gtk_status_icon_get_visible IA__gtk_status_icon_get_visible

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_status_icon_set_blinking) IA__gtk_status_icon_set_blinking __attribute((visibility("hidden")));
#define gtk_status_icon_set_blinking IA__gtk_status_icon_set_blinking

extern __typeof (gtk_status_icon_get_blinking) IA__gtk_status_icon_get_blinking __attribute((visibility("hidden")));
#define gtk_status_icon_get_blinking IA__gtk_status_icon_get_blinking

#endif
extern __typeof (gtk_status_icon_is_embedded) IA__gtk_status_icon_is_embedded __attribute((visibility("hidden")));
#define gtk_status_icon_is_embedded IA__gtk_status_icon_is_embedded

extern __typeof (gtk_status_icon_position_menu) IA__gtk_status_icon_position_menu __attribute((visibility("hidden")));
#define gtk_status_icon_position_menu IA__gtk_status_icon_position_menu

extern __typeof (gtk_status_icon_get_geometry) IA__gtk_status_icon_get_geometry __attribute((visibility("hidden")));
#define gtk_status_icon_get_geometry IA__gtk_status_icon_get_geometry

extern __typeof (gtk_status_icon_get_x11_window_id) IA__gtk_status_icon_get_x11_window_id __attribute((visibility("hidden")));
#define gtk_status_icon_get_x11_window_id IA__gtk_status_icon_get_x11_window_id

extern __typeof (gtk_status_icon_get_title) IA__gtk_status_icon_get_title __attribute((visibility("hidden")));
#define gtk_status_icon_get_title IA__gtk_status_icon_get_title

extern __typeof (gtk_status_icon_set_title) IA__gtk_status_icon_set_title __attribute((visibility("hidden")));
#define gtk_status_icon_set_title IA__gtk_status_icon_set_title

extern __typeof (gtk_status_icon_set_name) IA__gtk_status_icon_set_name __attribute((visibility("hidden")));
#define gtk_status_icon_set_name IA__gtk_status_icon_set_name

#endif
#endif
#if IN_HEADER(__GTK_STYLE_H__)
#if IN_FILE(__GTK_STYLE_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_draw_arrow) IA__gtk_draw_arrow __attribute((visibility("hidden")));
#define gtk_draw_arrow IA__gtk_draw_arrow

extern __typeof (gtk_draw_box) IA__gtk_draw_box __attribute((visibility("hidden")));
#define gtk_draw_box IA__gtk_draw_box

extern __typeof (gtk_draw_box_gap) IA__gtk_draw_box_gap __attribute((visibility("hidden")));
#define gtk_draw_box_gap IA__gtk_draw_box_gap

extern __typeof (gtk_draw_check) IA__gtk_draw_check __attribute((visibility("hidden")));
#define gtk_draw_check IA__gtk_draw_check

extern __typeof (gtk_draw_diamond) IA__gtk_draw_diamond __attribute((visibility("hidden")));
#define gtk_draw_diamond IA__gtk_draw_diamond

extern __typeof (gtk_draw_expander) IA__gtk_draw_expander __attribute((visibility("hidden")));
#define gtk_draw_expander IA__gtk_draw_expander

extern __typeof (gtk_draw_extension) IA__gtk_draw_extension __attribute((visibility("hidden")));
#define gtk_draw_extension IA__gtk_draw_extension

extern __typeof (gtk_draw_flat_box) IA__gtk_draw_flat_box __attribute((visibility("hidden")));
#define gtk_draw_flat_box IA__gtk_draw_flat_box

extern __typeof (gtk_draw_focus) IA__gtk_draw_focus __attribute((visibility("hidden")));
#define gtk_draw_focus IA__gtk_draw_focus

extern __typeof (gtk_draw_handle) IA__gtk_draw_handle __attribute((visibility("hidden")));
#define gtk_draw_handle IA__gtk_draw_handle

extern __typeof (gtk_draw_hline) IA__gtk_draw_hline __attribute((visibility("hidden")));
#define gtk_draw_hline IA__gtk_draw_hline

extern __typeof (gtk_draw_layout) IA__gtk_draw_layout __attribute((visibility("hidden")));
#define gtk_draw_layout IA__gtk_draw_layout

extern __typeof (gtk_draw_option) IA__gtk_draw_option __attribute((visibility("hidden")));
#define gtk_draw_option IA__gtk_draw_option

extern __typeof (gtk_draw_polygon) IA__gtk_draw_polygon __attribute((visibility("hidden")));
#define gtk_draw_polygon IA__gtk_draw_polygon

extern __typeof (gtk_draw_resize_grip) IA__gtk_draw_resize_grip __attribute((visibility("hidden")));
#define gtk_draw_resize_grip IA__gtk_draw_resize_grip

extern __typeof (gtk_draw_shadow) IA__gtk_draw_shadow __attribute((visibility("hidden")));
#define gtk_draw_shadow IA__gtk_draw_shadow

extern __typeof (gtk_draw_shadow_gap) IA__gtk_draw_shadow_gap __attribute((visibility("hidden")));
#define gtk_draw_shadow_gap IA__gtk_draw_shadow_gap

extern __typeof (gtk_draw_slider) IA__gtk_draw_slider __attribute((visibility("hidden")));
#define gtk_draw_slider IA__gtk_draw_slider

extern __typeof (gtk_draw_string) IA__gtk_draw_string __attribute((visibility("hidden")));
#define gtk_draw_string IA__gtk_draw_string

extern __typeof (gtk_draw_tab) IA__gtk_draw_tab __attribute((visibility("hidden")));
#define gtk_draw_tab IA__gtk_draw_tab

extern __typeof (gtk_draw_vline) IA__gtk_draw_vline __attribute((visibility("hidden")));
#define gtk_draw_vline IA__gtk_draw_vline

extern __typeof (gtk_style_set_font) IA__gtk_style_set_font __attribute((visibility("hidden")));
#define gtk_style_set_font IA__gtk_style_set_font

extern __typeof (gtk_style_get_font) IA__gtk_style_get_font __attribute((visibility("hidden")));
#define gtk_style_get_font IA__gtk_style_get_font

extern __typeof (gtk_style_ref) IA__gtk_style_ref __attribute((visibility("hidden")));
#define gtk_style_ref IA__gtk_style_ref

extern __typeof (gtk_style_unref) IA__gtk_style_unref __attribute((visibility("hidden")));
#define gtk_style_unref IA__gtk_style_unref

extern __typeof (gtk_paint_string) IA__gtk_paint_string __attribute((visibility("hidden")));
#define gtk_paint_string IA__gtk_paint_string

#endif
extern __typeof (gtk_paint_arrow) IA__gtk_paint_arrow __attribute((visibility("hidden")));
#define gtk_paint_arrow IA__gtk_paint_arrow

extern __typeof (gtk_paint_box) IA__gtk_paint_box __attribute((visibility("hidden")));
#define gtk_paint_box IA__gtk_paint_box

extern __typeof (gtk_paint_box_gap) IA__gtk_paint_box_gap __attribute((visibility("hidden")));
#define gtk_paint_box_gap IA__gtk_paint_box_gap

extern __typeof (gtk_paint_check) IA__gtk_paint_check __attribute((visibility("hidden")));
#define gtk_paint_check IA__gtk_paint_check

extern __typeof (gtk_paint_diamond) IA__gtk_paint_diamond __attribute((visibility("hidden")));
#define gtk_paint_diamond IA__gtk_paint_diamond

extern __typeof (gtk_paint_expander) IA__gtk_paint_expander __attribute((visibility("hidden")));
#define gtk_paint_expander IA__gtk_paint_expander

extern __typeof (gtk_paint_extension) IA__gtk_paint_extension __attribute((visibility("hidden")));
#define gtk_paint_extension IA__gtk_paint_extension

extern __typeof (gtk_paint_flat_box) IA__gtk_paint_flat_box __attribute((visibility("hidden")));
#define gtk_paint_flat_box IA__gtk_paint_flat_box

extern __typeof (gtk_paint_focus) IA__gtk_paint_focus __attribute((visibility("hidden")));
#define gtk_paint_focus IA__gtk_paint_focus

extern __typeof (gtk_paint_handle) IA__gtk_paint_handle __attribute((visibility("hidden")));
#define gtk_paint_handle IA__gtk_paint_handle

extern __typeof (gtk_paint_hline) IA__gtk_paint_hline __attribute((visibility("hidden")));
#define gtk_paint_hline IA__gtk_paint_hline

extern __typeof (gtk_paint_layout) IA__gtk_paint_layout __attribute((visibility("hidden")));
#define gtk_paint_layout IA__gtk_paint_layout

extern __typeof (gtk_paint_option) IA__gtk_paint_option __attribute((visibility("hidden")));
#define gtk_paint_option IA__gtk_paint_option

extern __typeof (gtk_paint_polygon) IA__gtk_paint_polygon __attribute((visibility("hidden")));
#define gtk_paint_polygon IA__gtk_paint_polygon

extern __typeof (gtk_paint_resize_grip) IA__gtk_paint_resize_grip __attribute((visibility("hidden")));
#define gtk_paint_resize_grip IA__gtk_paint_resize_grip

extern __typeof (gtk_paint_shadow) IA__gtk_paint_shadow __attribute((visibility("hidden")));
#define gtk_paint_shadow IA__gtk_paint_shadow

extern __typeof (gtk_paint_shadow_gap) IA__gtk_paint_shadow_gap __attribute((visibility("hidden")));
#define gtk_paint_shadow_gap IA__gtk_paint_shadow_gap

extern __typeof (gtk_paint_slider) IA__gtk_paint_slider __attribute((visibility("hidden")));
#define gtk_paint_slider IA__gtk_paint_slider

extern __typeof (gtk_paint_spinner) IA__gtk_paint_spinner __attribute((visibility("hidden")));
#define gtk_paint_spinner IA__gtk_paint_spinner

extern __typeof (gtk_paint_tab) IA__gtk_paint_tab __attribute((visibility("hidden")));
#define gtk_paint_tab IA__gtk_paint_tab

extern __typeof (gtk_paint_vline) IA__gtk_paint_vline __attribute((visibility("hidden")));
#define gtk_paint_vline IA__gtk_paint_vline

extern __typeof (gtk_border_new) IA__gtk_border_new __attribute((visibility("hidden"))) G_GNUC_MALLOC;
#define gtk_border_new IA__gtk_border_new

extern __typeof (gtk_border_copy) IA__gtk_border_copy __attribute((visibility("hidden")));
#define gtk_border_copy IA__gtk_border_copy

extern __typeof (gtk_border_free) IA__gtk_border_free __attribute((visibility("hidden")));
#define gtk_border_free IA__gtk_border_free

extern __typeof (gtk_border_get_type) IA__gtk_border_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_border_get_type IA__gtk_border_get_type

extern __typeof (gtk_style_apply_default_background) IA__gtk_style_apply_default_background __attribute((visibility("hidden")));
#define gtk_style_apply_default_background IA__gtk_style_apply_default_background

extern __typeof (gtk_style_attach) IA__gtk_style_attach __attribute((visibility("hidden")));
#define gtk_style_attach IA__gtk_style_attach

extern __typeof (gtk_style_copy) IA__gtk_style_copy __attribute((visibility("hidden")));
#define gtk_style_copy IA__gtk_style_copy

extern __typeof (gtk_style_detach) IA__gtk_style_detach __attribute((visibility("hidden")));
#define gtk_style_detach IA__gtk_style_detach

extern __typeof (gtk_style_get_type) IA__gtk_style_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_style_get_type IA__gtk_style_get_type

extern __typeof (gtk_style_lookup_icon_set) IA__gtk_style_lookup_icon_set __attribute((visibility("hidden")));
#define gtk_style_lookup_icon_set IA__gtk_style_lookup_icon_set

extern __typeof (gtk_style_lookup_color) IA__gtk_style_lookup_color __attribute((visibility("hidden")));
#define gtk_style_lookup_color IA__gtk_style_lookup_color

extern __typeof (gtk_style_new) IA__gtk_style_new __attribute((visibility("hidden")));
#define gtk_style_new IA__gtk_style_new

extern __typeof (gtk_style_render_icon) IA__gtk_style_render_icon __attribute((visibility("hidden")));
#define gtk_style_render_icon IA__gtk_style_render_icon

extern __typeof (gtk_style_set_background) IA__gtk_style_set_background __attribute((visibility("hidden")));
#define gtk_style_set_background IA__gtk_style_set_background

extern __typeof (gtk_draw_insertion_cursor) IA__gtk_draw_insertion_cursor __attribute((visibility("hidden")));
#define gtk_draw_insertion_cursor IA__gtk_draw_insertion_cursor

extern __typeof (gtk_style_get_style_property) IA__gtk_style_get_style_property __attribute((visibility("hidden")));
#define gtk_style_get_style_property IA__gtk_style_get_style_property

extern __typeof (gtk_style_get_valist) IA__gtk_style_get_valist __attribute((visibility("hidden")));
#define gtk_style_get_valist IA__gtk_style_get_valist

extern __typeof (gtk_style_get) IA__gtk_style_get __attribute((visibility("hidden")));
#define gtk_style_get IA__gtk_style_get

#endif
#endif
#if IN_HEADER(__GTK_DRAWING_AREA_H__)
#if IN_FILE(__GTK_DRAWING_AREA_C__)
extern __typeof (gtk_drawing_area_get_type) IA__gtk_drawing_area_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_drawing_area_get_type IA__gtk_drawing_area_get_type

extern __typeof (gtk_drawing_area_new) IA__gtk_drawing_area_new __attribute((visibility("hidden")));
#define gtk_drawing_area_new IA__gtk_drawing_area_new

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_drawing_area_size) IA__gtk_drawing_area_size __attribute((visibility("hidden")));
#define gtk_drawing_area_size IA__gtk_drawing_area_size

#endif
#endif
#endif
#if IN_HEADER(__GTK_EDITABLE_H__)
#if IN_FILE(__GTK_EDITABLE_C__)
extern __typeof (gtk_editable_copy_clipboard) IA__gtk_editable_copy_clipboard __attribute((visibility("hidden")));
#define gtk_editable_copy_clipboard IA__gtk_editable_copy_clipboard

extern __typeof (gtk_editable_cut_clipboard) IA__gtk_editable_cut_clipboard __attribute((visibility("hidden")));
#define gtk_editable_cut_clipboard IA__gtk_editable_cut_clipboard

extern __typeof (gtk_editable_delete_selection) IA__gtk_editable_delete_selection __attribute((visibility("hidden")));
#define gtk_editable_delete_selection IA__gtk_editable_delete_selection

extern __typeof (gtk_editable_delete_text) IA__gtk_editable_delete_text __attribute((visibility("hidden")));
#define gtk_editable_delete_text IA__gtk_editable_delete_text

extern __typeof (gtk_editable_get_chars) IA__gtk_editable_get_chars __attribute((visibility("hidden")));
#define gtk_editable_get_chars IA__gtk_editable_get_chars

extern __typeof (gtk_editable_get_editable) IA__gtk_editable_get_editable __attribute((visibility("hidden")));
#define gtk_editable_get_editable IA__gtk_editable_get_editable

extern __typeof (gtk_editable_get_position) IA__gtk_editable_get_position __attribute((visibility("hidden")));
#define gtk_editable_get_position IA__gtk_editable_get_position

extern __typeof (gtk_editable_get_selection_bounds) IA__gtk_editable_get_selection_bounds __attribute((visibility("hidden")));
#define gtk_editable_get_selection_bounds IA__gtk_editable_get_selection_bounds

extern __typeof (gtk_editable_get_type) IA__gtk_editable_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_editable_get_type IA__gtk_editable_get_type

extern __typeof (gtk_editable_insert_text) IA__gtk_editable_insert_text __attribute((visibility("hidden")));
#define gtk_editable_insert_text IA__gtk_editable_insert_text

extern __typeof (gtk_editable_paste_clipboard) IA__gtk_editable_paste_clipboard __attribute((visibility("hidden")));
#define gtk_editable_paste_clipboard IA__gtk_editable_paste_clipboard

extern __typeof (gtk_editable_select_region) IA__gtk_editable_select_region __attribute((visibility("hidden")));
#define gtk_editable_select_region IA__gtk_editable_select_region

extern __typeof (gtk_editable_set_editable) IA__gtk_editable_set_editable __attribute((visibility("hidden")));
#define gtk_editable_set_editable IA__gtk_editable_set_editable

extern __typeof (gtk_editable_set_position) IA__gtk_editable_set_position __attribute((visibility("hidden")));
#define gtk_editable_set_position IA__gtk_editable_set_position

#endif
#endif
#if IN_HEADER(__GTK_ENTRY_H__)
#if IN_FILE(__GTK_ENTRY_C__)
extern __typeof (gtk_entry_get_activates_default) IA__gtk_entry_get_activates_default __attribute((visibility("hidden")));
#define gtk_entry_get_activates_default IA__gtk_entry_get_activates_default

extern __typeof (gtk_entry_get_alignment) IA__gtk_entry_get_alignment __attribute((visibility("hidden")));
#define gtk_entry_get_alignment IA__gtk_entry_get_alignment

extern __typeof (gtk_entry_get_buffer) IA__gtk_entry_get_buffer __attribute((visibility("hidden")));
#define gtk_entry_get_buffer IA__gtk_entry_get_buffer

extern __typeof (gtk_entry_get_completion) IA__gtk_entry_get_completion __attribute((visibility("hidden")));
#define gtk_entry_get_completion IA__gtk_entry_get_completion

extern __typeof (gtk_entry_get_current_icon_drag_source) IA__gtk_entry_get_current_icon_drag_source __attribute((visibility("hidden")));
#define gtk_entry_get_current_icon_drag_source IA__gtk_entry_get_current_icon_drag_source

extern __typeof (gtk_entry_get_cursor_hadjustment) IA__gtk_entry_get_cursor_hadjustment __attribute((visibility("hidden")));
#define gtk_entry_get_cursor_hadjustment IA__gtk_entry_get_cursor_hadjustment

extern __typeof (gtk_entry_get_has_frame) IA__gtk_entry_get_has_frame __attribute((visibility("hidden")));
#define gtk_entry_get_has_frame IA__gtk_entry_get_has_frame

extern __typeof (gtk_entry_get_icon_activatable) IA__gtk_entry_get_icon_activatable __attribute((visibility("hidden")));
#define gtk_entry_get_icon_activatable IA__gtk_entry_get_icon_activatable

extern __typeof (gtk_entry_get_icon_at_pos) IA__gtk_entry_get_icon_at_pos __attribute((visibility("hidden")));
#define gtk_entry_get_icon_at_pos IA__gtk_entry_get_icon_at_pos

extern __typeof (gtk_entry_get_icon_gicon) IA__gtk_entry_get_icon_gicon __attribute((visibility("hidden")));
#define gtk_entry_get_icon_gicon IA__gtk_entry_get_icon_gicon

extern __typeof (gtk_entry_get_icon_name) IA__gtk_entry_get_icon_name __attribute((visibility("hidden")));
#define gtk_entry_get_icon_name IA__gtk_entry_get_icon_name

extern __typeof (gtk_entry_get_icon_pixbuf) IA__gtk_entry_get_icon_pixbuf __attribute((visibility("hidden")));
#define gtk_entry_get_icon_pixbuf IA__gtk_entry_get_icon_pixbuf

extern __typeof (gtk_entry_get_icon_sensitive) IA__gtk_entry_get_icon_sensitive __attribute((visibility("hidden")));
#define gtk_entry_get_icon_sensitive IA__gtk_entry_get_icon_sensitive

extern __typeof (gtk_entry_get_icon_stock) IA__gtk_entry_get_icon_stock __attribute((visibility("hidden")));
#define gtk_entry_get_icon_stock IA__gtk_entry_get_icon_stock

extern __typeof (gtk_entry_get_icon_storage_type) IA__gtk_entry_get_icon_storage_type __attribute((visibility("hidden")));
#define gtk_entry_get_icon_storage_type IA__gtk_entry_get_icon_storage_type

extern __typeof (gtk_entry_get_icon_tooltip_markup) IA__gtk_entry_get_icon_tooltip_markup __attribute((visibility("hidden")));
#define gtk_entry_get_icon_tooltip_markup IA__gtk_entry_get_icon_tooltip_markup

extern __typeof (gtk_entry_get_icon_tooltip_text) IA__gtk_entry_get_icon_tooltip_text __attribute((visibility("hidden")));
#define gtk_entry_get_icon_tooltip_text IA__gtk_entry_get_icon_tooltip_text

extern __typeof (gtk_entry_get_icon_window) IA__gtk_entry_get_icon_window __attribute((visibility("hidden")));
#define gtk_entry_get_icon_window IA__gtk_entry_get_icon_window

extern __typeof (gtk_entry_get_inner_border) IA__gtk_entry_get_inner_border __attribute((visibility("hidden")));
#define gtk_entry_get_inner_border IA__gtk_entry_get_inner_border

extern __typeof (gtk_entry_get_invisible_char) IA__gtk_entry_get_invisible_char __attribute((visibility("hidden")));
#define gtk_entry_get_invisible_char IA__gtk_entry_get_invisible_char

extern __typeof (gtk_entry_get_layout) IA__gtk_entry_get_layout __attribute((visibility("hidden")));
#define gtk_entry_get_layout IA__gtk_entry_get_layout

extern __typeof (gtk_entry_get_layout_offsets) IA__gtk_entry_get_layout_offsets __attribute((visibility("hidden")));
#define gtk_entry_get_layout_offsets IA__gtk_entry_get_layout_offsets

extern __typeof (gtk_entry_get_max_length) IA__gtk_entry_get_max_length __attribute((visibility("hidden")));
#define gtk_entry_get_max_length IA__gtk_entry_get_max_length

extern __typeof (gtk_entry_get_overwrite_mode) IA__gtk_entry_get_overwrite_mode __attribute((visibility("hidden")));
#define gtk_entry_get_overwrite_mode IA__gtk_entry_get_overwrite_mode

extern __typeof (gtk_entry_get_progress_fraction) IA__gtk_entry_get_progress_fraction __attribute((visibility("hidden")));
#define gtk_entry_get_progress_fraction IA__gtk_entry_get_progress_fraction

extern __typeof (gtk_entry_get_progress_pulse_step) IA__gtk_entry_get_progress_pulse_step __attribute((visibility("hidden")));
#define gtk_entry_get_progress_pulse_step IA__gtk_entry_get_progress_pulse_step

extern __typeof (gtk_entry_get_text) IA__gtk_entry_get_text __attribute((visibility("hidden")));
#define gtk_entry_get_text IA__gtk_entry_get_text

extern __typeof (gtk_entry_get_text_length) IA__gtk_entry_get_text_length __attribute((visibility("hidden")));
#define gtk_entry_get_text_length IA__gtk_entry_get_text_length

extern __typeof (gtk_entry_get_type) IA__gtk_entry_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_entry_get_type IA__gtk_entry_get_type

extern __typeof (gtk_entry_get_visibility) IA__gtk_entry_get_visibility __attribute((visibility("hidden")));
#define gtk_entry_get_visibility IA__gtk_entry_get_visibility

extern __typeof (gtk_entry_get_width_chars) IA__gtk_entry_get_width_chars __attribute((visibility("hidden")));
#define gtk_entry_get_width_chars IA__gtk_entry_get_width_chars

extern __typeof (gtk_entry_get_text_window) IA__gtk_entry_get_text_window __attribute((visibility("hidden")));
#define gtk_entry_get_text_window IA__gtk_entry_get_text_window

extern __typeof (gtk_entry_im_context_filter_keypress) IA__gtk_entry_im_context_filter_keypress __attribute((visibility("hidden")));
#define gtk_entry_im_context_filter_keypress IA__gtk_entry_im_context_filter_keypress

extern __typeof (gtk_entry_layout_index_to_text_index) IA__gtk_entry_layout_index_to_text_index __attribute((visibility("hidden")));
#define gtk_entry_layout_index_to_text_index IA__gtk_entry_layout_index_to_text_index

extern __typeof (gtk_entry_new) IA__gtk_entry_new __attribute((visibility("hidden")));
#define gtk_entry_new IA__gtk_entry_new

extern __typeof (gtk_entry_new_with_buffer) IA__gtk_entry_new_with_buffer __attribute((visibility("hidden")));
#define gtk_entry_new_with_buffer IA__gtk_entry_new_with_buffer

extern __typeof (gtk_entry_progress_pulse) IA__gtk_entry_progress_pulse __attribute((visibility("hidden")));
#define gtk_entry_progress_pulse IA__gtk_entry_progress_pulse

extern __typeof (gtk_entry_reset_im_context) IA__gtk_entry_reset_im_context __attribute((visibility("hidden")));
#define gtk_entry_reset_im_context IA__gtk_entry_reset_im_context

extern __typeof (gtk_entry_set_activates_default) IA__gtk_entry_set_activates_default __attribute((visibility("hidden")));
#define gtk_entry_set_activates_default IA__gtk_entry_set_activates_default

extern __typeof (gtk_entry_set_alignment) IA__gtk_entry_set_alignment __attribute((visibility("hidden")));
#define gtk_entry_set_alignment IA__gtk_entry_set_alignment

extern __typeof (gtk_entry_set_buffer) IA__gtk_entry_set_buffer __attribute((visibility("hidden")));
#define gtk_entry_set_buffer IA__gtk_entry_set_buffer

extern __typeof (gtk_entry_set_completion) IA__gtk_entry_set_completion __attribute((visibility("hidden")));
#define gtk_entry_set_completion IA__gtk_entry_set_completion

extern __typeof (gtk_entry_set_cursor_hadjustment) IA__gtk_entry_set_cursor_hadjustment __attribute((visibility("hidden")));
#define gtk_entry_set_cursor_hadjustment IA__gtk_entry_set_cursor_hadjustment

extern __typeof (gtk_entry_set_has_frame) IA__gtk_entry_set_has_frame __attribute((visibility("hidden")));
#define gtk_entry_set_has_frame IA__gtk_entry_set_has_frame

extern __typeof (gtk_entry_set_icon_activatable) IA__gtk_entry_set_icon_activatable __attribute((visibility("hidden")));
#define gtk_entry_set_icon_activatable IA__gtk_entry_set_icon_activatable

extern __typeof (gtk_entry_set_icon_drag_source) IA__gtk_entry_set_icon_drag_source __attribute((visibility("hidden")));
#define gtk_entry_set_icon_drag_source IA__gtk_entry_set_icon_drag_source

extern __typeof (gtk_entry_set_icon_from_gicon) IA__gtk_entry_set_icon_from_gicon __attribute((visibility("hidden")));
#define gtk_entry_set_icon_from_gicon IA__gtk_entry_set_icon_from_gicon

extern __typeof (gtk_entry_set_icon_from_icon_name) IA__gtk_entry_set_icon_from_icon_name __attribute((visibility("hidden")));
#define gtk_entry_set_icon_from_icon_name IA__gtk_entry_set_icon_from_icon_name

extern __typeof (gtk_entry_set_icon_from_pixbuf) IA__gtk_entry_set_icon_from_pixbuf __attribute((visibility("hidden")));
#define gtk_entry_set_icon_from_pixbuf IA__gtk_entry_set_icon_from_pixbuf

extern __typeof (gtk_entry_set_icon_from_stock) IA__gtk_entry_set_icon_from_stock __attribute((visibility("hidden")));
#define gtk_entry_set_icon_from_stock IA__gtk_entry_set_icon_from_stock

extern __typeof (gtk_entry_set_icon_sensitive) IA__gtk_entry_set_icon_sensitive __attribute((visibility("hidden")));
#define gtk_entry_set_icon_sensitive IA__gtk_entry_set_icon_sensitive

extern __typeof (gtk_entry_set_icon_tooltip_markup) IA__gtk_entry_set_icon_tooltip_markup __attribute((visibility("hidden")));
#define gtk_entry_set_icon_tooltip_markup IA__gtk_entry_set_icon_tooltip_markup

extern __typeof (gtk_entry_set_icon_tooltip_text) IA__gtk_entry_set_icon_tooltip_text __attribute((visibility("hidden")));
#define gtk_entry_set_icon_tooltip_text IA__gtk_entry_set_icon_tooltip_text

extern __typeof (gtk_entry_set_inner_border) IA__gtk_entry_set_inner_border __attribute((visibility("hidden")));
#define gtk_entry_set_inner_border IA__gtk_entry_set_inner_border

extern __typeof (gtk_entry_set_invisible_char) IA__gtk_entry_set_invisible_char __attribute((visibility("hidden")));
#define gtk_entry_set_invisible_char IA__gtk_entry_set_invisible_char

extern __typeof (gtk_entry_set_max_length) IA__gtk_entry_set_max_length __attribute((visibility("hidden")));
#define gtk_entry_set_max_length IA__gtk_entry_set_max_length

extern __typeof (gtk_entry_set_overwrite_mode) IA__gtk_entry_set_overwrite_mode __attribute((visibility("hidden")));
#define gtk_entry_set_overwrite_mode IA__gtk_entry_set_overwrite_mode

extern __typeof (gtk_entry_set_progress_fraction) IA__gtk_entry_set_progress_fraction __attribute((visibility("hidden")));
#define gtk_entry_set_progress_fraction IA__gtk_entry_set_progress_fraction

extern __typeof (gtk_entry_set_progress_pulse_step) IA__gtk_entry_set_progress_pulse_step __attribute((visibility("hidden")));
#define gtk_entry_set_progress_pulse_step IA__gtk_entry_set_progress_pulse_step

extern __typeof (gtk_entry_set_text) IA__gtk_entry_set_text __attribute((visibility("hidden")));
#define gtk_entry_set_text IA__gtk_entry_set_text

extern __typeof (gtk_entry_set_visibility) IA__gtk_entry_set_visibility __attribute((visibility("hidden")));
#define gtk_entry_set_visibility IA__gtk_entry_set_visibility

extern __typeof (gtk_entry_set_width_chars) IA__gtk_entry_set_width_chars __attribute((visibility("hidden")));
#define gtk_entry_set_width_chars IA__gtk_entry_set_width_chars

extern __typeof (gtk_entry_text_index_to_layout_index) IA__gtk_entry_text_index_to_layout_index __attribute((visibility("hidden")));
#define gtk_entry_text_index_to_layout_index IA__gtk_entry_text_index_to_layout_index

extern __typeof (gtk_entry_unset_invisible_char) IA__gtk_entry_unset_invisible_char __attribute((visibility("hidden")));
#define gtk_entry_unset_invisible_char IA__gtk_entry_unset_invisible_char

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_entry_append_text) IA__gtk_entry_append_text __attribute((visibility("hidden")));
#define gtk_entry_append_text IA__gtk_entry_append_text

extern __typeof (gtk_entry_new_with_max_length) IA__gtk_entry_new_with_max_length __attribute((visibility("hidden")));
#define gtk_entry_new_with_max_length IA__gtk_entry_new_with_max_length

extern __typeof (gtk_entry_prepend_text) IA__gtk_entry_prepend_text __attribute((visibility("hidden")));
#define gtk_entry_prepend_text IA__gtk_entry_prepend_text

extern __typeof (gtk_entry_select_region) IA__gtk_entry_select_region __attribute((visibility("hidden")));
#define gtk_entry_select_region IA__gtk_entry_select_region

extern __typeof (gtk_entry_set_editable) IA__gtk_entry_set_editable __attribute((visibility("hidden")));
#define gtk_entry_set_editable IA__gtk_entry_set_editable

extern __typeof (gtk_entry_set_position) IA__gtk_entry_set_position __attribute((visibility("hidden")));
#define gtk_entry_set_position IA__gtk_entry_set_position

#endif
#endif
#endif
#if IN_HEADER(__GTK_ENTRY_COMPLETION_H__)
#if IN_FILE(__GTK_ENTRY_COMPLETION_C__)
extern __typeof (gtk_entry_completion_complete) IA__gtk_entry_completion_complete __attribute((visibility("hidden")));
#define gtk_entry_completion_complete IA__gtk_entry_completion_complete

extern __typeof (gtk_entry_completion_delete_action) IA__gtk_entry_completion_delete_action __attribute((visibility("hidden")));
#define gtk_entry_completion_delete_action IA__gtk_entry_completion_delete_action

extern __typeof (gtk_entry_completion_get_entry) IA__gtk_entry_completion_get_entry __attribute((visibility("hidden")));
#define gtk_entry_completion_get_entry IA__gtk_entry_completion_get_entry

extern __typeof (gtk_entry_completion_get_inline_completion) IA__gtk_entry_completion_get_inline_completion __attribute((visibility("hidden")));
#define gtk_entry_completion_get_inline_completion IA__gtk_entry_completion_get_inline_completion

extern __typeof (gtk_entry_completion_get_inline_selection) IA__gtk_entry_completion_get_inline_selection __attribute((visibility("hidden")));
#define gtk_entry_completion_get_inline_selection IA__gtk_entry_completion_get_inline_selection

extern __typeof (gtk_entry_completion_get_minimum_key_length) IA__gtk_entry_completion_get_minimum_key_length __attribute((visibility("hidden")));
#define gtk_entry_completion_get_minimum_key_length IA__gtk_entry_completion_get_minimum_key_length

extern __typeof (gtk_entry_completion_get_completion_prefix) IA__gtk_entry_completion_get_completion_prefix __attribute((visibility("hidden")));
#define gtk_entry_completion_get_completion_prefix IA__gtk_entry_completion_get_completion_prefix

extern __typeof (gtk_entry_completion_get_model) IA__gtk_entry_completion_get_model __attribute((visibility("hidden")));
#define gtk_entry_completion_get_model IA__gtk_entry_completion_get_model

extern __typeof (gtk_entry_completion_get_popup_completion) IA__gtk_entry_completion_get_popup_completion __attribute((visibility("hidden")));
#define gtk_entry_completion_get_popup_completion IA__gtk_entry_completion_get_popup_completion

extern __typeof (gtk_entry_completion_get_popup_set_width) IA__gtk_entry_completion_get_popup_set_width __attribute((visibility("hidden")));
#define gtk_entry_completion_get_popup_set_width IA__gtk_entry_completion_get_popup_set_width

extern __typeof (gtk_entry_completion_get_popup_single_match) IA__gtk_entry_completion_get_popup_single_match __attribute((visibility("hidden")));
#define gtk_entry_completion_get_popup_single_match IA__gtk_entry_completion_get_popup_single_match

extern __typeof (gtk_entry_completion_get_text_column) IA__gtk_entry_completion_get_text_column __attribute((visibility("hidden")));
#define gtk_entry_completion_get_text_column IA__gtk_entry_completion_get_text_column

extern __typeof (gtk_entry_completion_get_type) IA__gtk_entry_completion_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_entry_completion_get_type IA__gtk_entry_completion_get_type

extern __typeof (gtk_entry_completion_insert_action_markup) IA__gtk_entry_completion_insert_action_markup __attribute((visibility("hidden")));
#define gtk_entry_completion_insert_action_markup IA__gtk_entry_completion_insert_action_markup

extern __typeof (gtk_entry_completion_insert_action_text) IA__gtk_entry_completion_insert_action_text __attribute((visibility("hidden")));
#define gtk_entry_completion_insert_action_text IA__gtk_entry_completion_insert_action_text

extern __typeof (gtk_entry_completion_insert_prefix) IA__gtk_entry_completion_insert_prefix __attribute((visibility("hidden")));
#define gtk_entry_completion_insert_prefix IA__gtk_entry_completion_insert_prefix

extern __typeof (gtk_entry_completion_new) IA__gtk_entry_completion_new __attribute((visibility("hidden")));
#define gtk_entry_completion_new IA__gtk_entry_completion_new

extern __typeof (gtk_entry_completion_set_inline_completion) IA__gtk_entry_completion_set_inline_completion __attribute((visibility("hidden")));
#define gtk_entry_completion_set_inline_completion IA__gtk_entry_completion_set_inline_completion

extern __typeof (gtk_entry_completion_set_inline_selection) IA__gtk_entry_completion_set_inline_selection __attribute((visibility("hidden")));
#define gtk_entry_completion_set_inline_selection IA__gtk_entry_completion_set_inline_selection

extern __typeof (gtk_entry_completion_set_match_func) IA__gtk_entry_completion_set_match_func __attribute((visibility("hidden")));
#define gtk_entry_completion_set_match_func IA__gtk_entry_completion_set_match_func

extern __typeof (gtk_entry_completion_set_minimum_key_length) IA__gtk_entry_completion_set_minimum_key_length __attribute((visibility("hidden")));
#define gtk_entry_completion_set_minimum_key_length IA__gtk_entry_completion_set_minimum_key_length

extern __typeof (gtk_entry_completion_set_model) IA__gtk_entry_completion_set_model __attribute((visibility("hidden")));
#define gtk_entry_completion_set_model IA__gtk_entry_completion_set_model

extern __typeof (gtk_entry_completion_set_popup_completion) IA__gtk_entry_completion_set_popup_completion __attribute((visibility("hidden")));
#define gtk_entry_completion_set_popup_completion IA__gtk_entry_completion_set_popup_completion

extern __typeof (gtk_entry_completion_set_popup_set_width) IA__gtk_entry_completion_set_popup_set_width __attribute((visibility("hidden")));
#define gtk_entry_completion_set_popup_set_width IA__gtk_entry_completion_set_popup_set_width

extern __typeof (gtk_entry_completion_set_popup_single_match) IA__gtk_entry_completion_set_popup_single_match __attribute((visibility("hidden")));
#define gtk_entry_completion_set_popup_single_match IA__gtk_entry_completion_set_popup_single_match

extern __typeof (gtk_entry_completion_set_text_column) IA__gtk_entry_completion_set_text_column __attribute((visibility("hidden")));
#define gtk_entry_completion_set_text_column IA__gtk_entry_completion_set_text_column

#endif
#endif
#if IN_HEADER(__GTK_ENTRY_BUFFER_H__)
#if IN_FILE(__GTK_ENTRY_BUFFER_C__)
extern __typeof (gtk_entry_buffer_get_type) IA__gtk_entry_buffer_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_entry_buffer_get_type IA__gtk_entry_buffer_get_type

extern __typeof (gtk_entry_buffer_new) IA__gtk_entry_buffer_new __attribute((visibility("hidden")));
#define gtk_entry_buffer_new IA__gtk_entry_buffer_new

extern __typeof (gtk_entry_buffer_get_bytes) IA__gtk_entry_buffer_get_bytes __attribute((visibility("hidden")));
#define gtk_entry_buffer_get_bytes IA__gtk_entry_buffer_get_bytes

extern __typeof (gtk_entry_buffer_get_length) IA__gtk_entry_buffer_get_length __attribute((visibility("hidden")));
#define gtk_entry_buffer_get_length IA__gtk_entry_buffer_get_length

extern __typeof (gtk_entry_buffer_get_text) IA__gtk_entry_buffer_get_text __attribute((visibility("hidden")));
#define gtk_entry_buffer_get_text IA__gtk_entry_buffer_get_text

extern __typeof (gtk_entry_buffer_set_text) IA__gtk_entry_buffer_set_text __attribute((visibility("hidden")));
#define gtk_entry_buffer_set_text IA__gtk_entry_buffer_set_text

extern __typeof (gtk_entry_buffer_set_max_length) IA__gtk_entry_buffer_set_max_length __attribute((visibility("hidden")));
#define gtk_entry_buffer_set_max_length IA__gtk_entry_buffer_set_max_length

extern __typeof (gtk_entry_buffer_get_max_length) IA__gtk_entry_buffer_get_max_length __attribute((visibility("hidden")));
#define gtk_entry_buffer_get_max_length IA__gtk_entry_buffer_get_max_length

extern __typeof (gtk_entry_buffer_insert_text) IA__gtk_entry_buffer_insert_text __attribute((visibility("hidden")));
#define gtk_entry_buffer_insert_text IA__gtk_entry_buffer_insert_text

extern __typeof (gtk_entry_buffer_delete_text) IA__gtk_entry_buffer_delete_text __attribute((visibility("hidden")));
#define gtk_entry_buffer_delete_text IA__gtk_entry_buffer_delete_text

extern __typeof (gtk_entry_buffer_emit_inserted_text) IA__gtk_entry_buffer_emit_inserted_text __attribute((visibility("hidden")));
#define gtk_entry_buffer_emit_inserted_text IA__gtk_entry_buffer_emit_inserted_text

extern __typeof (gtk_entry_buffer_emit_deleted_text) IA__gtk_entry_buffer_emit_deleted_text __attribute((visibility("hidden")));
#define gtk_entry_buffer_emit_deleted_text IA__gtk_entry_buffer_emit_deleted_text

#endif
#endif
#if IN_HEADER(__GTK_EVENT_BOX_H__)
#if IN_FILE(__GTK_EVENT_BOX_C__)
extern __typeof (gtk_event_box_get_above_child) IA__gtk_event_box_get_above_child __attribute((visibility("hidden")));
#define gtk_event_box_get_above_child IA__gtk_event_box_get_above_child

extern __typeof (gtk_event_box_get_type) IA__gtk_event_box_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_event_box_get_type IA__gtk_event_box_get_type

extern __typeof (gtk_event_box_get_visible_window) IA__gtk_event_box_get_visible_window __attribute((visibility("hidden")));
#define gtk_event_box_get_visible_window IA__gtk_event_box_get_visible_window

extern __typeof (gtk_event_box_new) IA__gtk_event_box_new __attribute((visibility("hidden")));
#define gtk_event_box_new IA__gtk_event_box_new

extern __typeof (gtk_event_box_set_above_child) IA__gtk_event_box_set_above_child __attribute((visibility("hidden")));
#define gtk_event_box_set_above_child IA__gtk_event_box_set_above_child

extern __typeof (gtk_event_box_set_visible_window) IA__gtk_event_box_set_visible_window __attribute((visibility("hidden")));
#define gtk_event_box_set_visible_window IA__gtk_event_box_set_visible_window

#endif
#endif
#if IN_HEADER(__GTK_EXPANDER_H__)
#if IN_FILE(__GTK_EXPANDER_C__)
extern __typeof (gtk_expander_get_expanded) IA__gtk_expander_get_expanded __attribute((visibility("hidden")));
#define gtk_expander_get_expanded IA__gtk_expander_get_expanded

extern __typeof (gtk_expander_get_label) IA__gtk_expander_get_label __attribute((visibility("hidden")));
#define gtk_expander_get_label IA__gtk_expander_get_label

extern __typeof (gtk_expander_get_label_fill) IA__gtk_expander_get_label_fill __attribute((visibility("hidden")));
#define gtk_expander_get_label_fill IA__gtk_expander_get_label_fill

extern __typeof (gtk_expander_get_label_widget) IA__gtk_expander_get_label_widget __attribute((visibility("hidden")));
#define gtk_expander_get_label_widget IA__gtk_expander_get_label_widget

extern __typeof (gtk_expander_get_spacing) IA__gtk_expander_get_spacing __attribute((visibility("hidden")));
#define gtk_expander_get_spacing IA__gtk_expander_get_spacing

extern __typeof (gtk_expander_get_type) IA__gtk_expander_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_expander_get_type IA__gtk_expander_get_type

extern __typeof (gtk_expander_get_use_markup) IA__gtk_expander_get_use_markup __attribute((visibility("hidden")));
#define gtk_expander_get_use_markup IA__gtk_expander_get_use_markup

extern __typeof (gtk_expander_get_use_underline) IA__gtk_expander_get_use_underline __attribute((visibility("hidden")));
#define gtk_expander_get_use_underline IA__gtk_expander_get_use_underline

extern __typeof (gtk_expander_new) IA__gtk_expander_new __attribute((visibility("hidden")));
#define gtk_expander_new IA__gtk_expander_new

extern __typeof (gtk_expander_new_with_mnemonic) IA__gtk_expander_new_with_mnemonic __attribute((visibility("hidden")));
#define gtk_expander_new_with_mnemonic IA__gtk_expander_new_with_mnemonic

extern __typeof (gtk_expander_set_expanded) IA__gtk_expander_set_expanded __attribute((visibility("hidden")));
#define gtk_expander_set_expanded IA__gtk_expander_set_expanded

extern __typeof (gtk_expander_set_label) IA__gtk_expander_set_label __attribute((visibility("hidden")));
#define gtk_expander_set_label IA__gtk_expander_set_label

extern __typeof (gtk_expander_set_label_fill) IA__gtk_expander_set_label_fill __attribute((visibility("hidden")));
#define gtk_expander_set_label_fill IA__gtk_expander_set_label_fill

extern __typeof (gtk_expander_set_label_widget) IA__gtk_expander_set_label_widget __attribute((visibility("hidden")));
#define gtk_expander_set_label_widget IA__gtk_expander_set_label_widget

extern __typeof (gtk_expander_set_spacing) IA__gtk_expander_set_spacing __attribute((visibility("hidden")));
#define gtk_expander_set_spacing IA__gtk_expander_set_spacing

extern __typeof (gtk_expander_set_use_markup) IA__gtk_expander_set_use_markup __attribute((visibility("hidden")));
#define gtk_expander_set_use_markup IA__gtk_expander_set_use_markup

extern __typeof (gtk_expander_set_use_underline) IA__gtk_expander_set_use_underline __attribute((visibility("hidden")));
#define gtk_expander_set_use_underline IA__gtk_expander_set_use_underline

#endif
#endif
#if IN_HEADER(__GTK_FILE_CHOOSER_H__)
#if IN_FILE(__GTK_FILE_CHOOSER_C__)
extern __typeof (gtk_file_chooser_add_filter) IA__gtk_file_chooser_add_filter __attribute((visibility("hidden")));
#define gtk_file_chooser_add_filter IA__gtk_file_chooser_add_filter

#ifndef _WIN64
extern __typeof (gtk_file_chooser_add_shortcut_folder) IA__gtk_file_chooser_add_shortcut_folder __attribute((visibility("hidden")));
#define gtk_file_chooser_add_shortcut_folder IA__gtk_file_chooser_add_shortcut_folder

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_chooser_add_shortcut_folder_utf8) IA__gtk_file_chooser_add_shortcut_folder_utf8 __attribute((visibility("hidden")));
#define gtk_file_chooser_add_shortcut_folder_utf8 IA__gtk_file_chooser_add_shortcut_folder_utf8

#endif
extern __typeof (gtk_file_chooser_add_shortcut_folder_uri) IA__gtk_file_chooser_add_shortcut_folder_uri __attribute((visibility("hidden")));
#define gtk_file_chooser_add_shortcut_folder_uri IA__gtk_file_chooser_add_shortcut_folder_uri

extern __typeof (gtk_file_chooser_error_quark) IA__gtk_file_chooser_error_quark __attribute((visibility("hidden")));
#define gtk_file_chooser_error_quark IA__gtk_file_chooser_error_quark

extern __typeof (gtk_file_chooser_get_action) IA__gtk_file_chooser_get_action __attribute((visibility("hidden")));
#define gtk_file_chooser_get_action IA__gtk_file_chooser_get_action

#ifndef _WIN64
extern __typeof (gtk_file_chooser_get_current_folder) IA__gtk_file_chooser_get_current_folder __attribute((visibility("hidden")));
#define gtk_file_chooser_get_current_folder IA__gtk_file_chooser_get_current_folder

#endif
extern __typeof (gtk_file_chooser_get_current_folder_file) IA__gtk_file_chooser_get_current_folder_file __attribute((visibility("hidden")));
#define gtk_file_chooser_get_current_folder_file IA__gtk_file_chooser_get_current_folder_file

#ifdef G_OS_WIN32
extern __typeof (gtk_file_chooser_get_current_folder_utf8) IA__gtk_file_chooser_get_current_folder_utf8 __attribute((visibility("hidden")));
#define gtk_file_chooser_get_current_folder_utf8 IA__gtk_file_chooser_get_current_folder_utf8

#endif
extern __typeof (gtk_file_chooser_get_current_folder_uri) IA__gtk_file_chooser_get_current_folder_uri __attribute((visibility("hidden")));
#define gtk_file_chooser_get_current_folder_uri IA__gtk_file_chooser_get_current_folder_uri

extern __typeof (gtk_file_chooser_get_extra_widget) IA__gtk_file_chooser_get_extra_widget __attribute((visibility("hidden")));
#define gtk_file_chooser_get_extra_widget IA__gtk_file_chooser_get_extra_widget

extern __typeof (gtk_file_chooser_get_file) IA__gtk_file_chooser_get_file __attribute((visibility("hidden")));
#define gtk_file_chooser_get_file IA__gtk_file_chooser_get_file

#ifndef _WIN64
extern __typeof (gtk_file_chooser_get_filename) IA__gtk_file_chooser_get_filename __attribute((visibility("hidden")));
#define gtk_file_chooser_get_filename IA__gtk_file_chooser_get_filename

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_chooser_get_filename_utf8) IA__gtk_file_chooser_get_filename_utf8 __attribute((visibility("hidden")));
#define gtk_file_chooser_get_filename_utf8 IA__gtk_file_chooser_get_filename_utf8

#endif
#ifndef _WIN64
extern __typeof (gtk_file_chooser_get_filenames) IA__gtk_file_chooser_get_filenames __attribute((visibility("hidden")));
#define gtk_file_chooser_get_filenames IA__gtk_file_chooser_get_filenames

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_chooser_get_filenames_utf8) IA__gtk_file_chooser_get_filenames_utf8 __attribute((visibility("hidden")));
#define gtk_file_chooser_get_filenames_utf8 IA__gtk_file_chooser_get_filenames_utf8

#endif
extern __typeof (gtk_file_chooser_get_files) IA__gtk_file_chooser_get_files __attribute((visibility("hidden")));
#define gtk_file_chooser_get_files IA__gtk_file_chooser_get_files

extern __typeof (gtk_file_chooser_get_filter) IA__gtk_file_chooser_get_filter __attribute((visibility("hidden")));
#define gtk_file_chooser_get_filter IA__gtk_file_chooser_get_filter

extern __typeof (gtk_file_chooser_get_local_only) IA__gtk_file_chooser_get_local_only __attribute((visibility("hidden")));
#define gtk_file_chooser_get_local_only IA__gtk_file_chooser_get_local_only

extern __typeof (gtk_file_chooser_get_preview_file) IA__gtk_file_chooser_get_preview_file __attribute((visibility("hidden")));
#define gtk_file_chooser_get_preview_file IA__gtk_file_chooser_get_preview_file

#ifndef _WIN64
extern __typeof (gtk_file_chooser_get_preview_filename) IA__gtk_file_chooser_get_preview_filename __attribute((visibility("hidden")));
#define gtk_file_chooser_get_preview_filename IA__gtk_file_chooser_get_preview_filename

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_chooser_get_preview_filename_utf8) IA__gtk_file_chooser_get_preview_filename_utf8 __attribute((visibility("hidden")));
#define gtk_file_chooser_get_preview_filename_utf8 IA__gtk_file_chooser_get_preview_filename_utf8

#endif
extern __typeof (gtk_file_chooser_get_preview_uri) IA__gtk_file_chooser_get_preview_uri __attribute((visibility("hidden")));
#define gtk_file_chooser_get_preview_uri IA__gtk_file_chooser_get_preview_uri

extern __typeof (gtk_file_chooser_get_preview_widget) IA__gtk_file_chooser_get_preview_widget __attribute((visibility("hidden")));
#define gtk_file_chooser_get_preview_widget IA__gtk_file_chooser_get_preview_widget

extern __typeof (gtk_file_chooser_get_preview_widget_active) IA__gtk_file_chooser_get_preview_widget_active __attribute((visibility("hidden")));
#define gtk_file_chooser_get_preview_widget_active IA__gtk_file_chooser_get_preview_widget_active

extern __typeof (gtk_file_chooser_get_select_multiple) IA__gtk_file_chooser_get_select_multiple __attribute((visibility("hidden")));
#define gtk_file_chooser_get_select_multiple IA__gtk_file_chooser_get_select_multiple

extern __typeof (gtk_file_chooser_get_show_hidden) IA__gtk_file_chooser_get_show_hidden __attribute((visibility("hidden")));
#define gtk_file_chooser_get_show_hidden IA__gtk_file_chooser_get_show_hidden

extern __typeof (gtk_file_chooser_set_do_overwrite_confirmation) IA__gtk_file_chooser_set_do_overwrite_confirmation __attribute((visibility("hidden")));
#define gtk_file_chooser_set_do_overwrite_confirmation IA__gtk_file_chooser_set_do_overwrite_confirmation

extern __typeof (gtk_file_chooser_get_do_overwrite_confirmation) IA__gtk_file_chooser_get_do_overwrite_confirmation __attribute((visibility("hidden")));
#define gtk_file_chooser_get_do_overwrite_confirmation IA__gtk_file_chooser_get_do_overwrite_confirmation

extern __typeof (gtk_file_chooser_set_create_folders) IA__gtk_file_chooser_set_create_folders __attribute((visibility("hidden")));
#define gtk_file_chooser_set_create_folders IA__gtk_file_chooser_set_create_folders

extern __typeof (gtk_file_chooser_get_create_folders) IA__gtk_file_chooser_get_create_folders __attribute((visibility("hidden")));
#define gtk_file_chooser_get_create_folders IA__gtk_file_chooser_get_create_folders

extern __typeof (gtk_file_chooser_get_type) IA__gtk_file_chooser_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_file_chooser_get_type IA__gtk_file_chooser_get_type

extern __typeof (gtk_file_chooser_get_uri) IA__gtk_file_chooser_get_uri __attribute((visibility("hidden")));
#define gtk_file_chooser_get_uri IA__gtk_file_chooser_get_uri

extern __typeof (gtk_file_chooser_get_uris) IA__gtk_file_chooser_get_uris __attribute((visibility("hidden")));
#define gtk_file_chooser_get_uris IA__gtk_file_chooser_get_uris

extern __typeof (gtk_file_chooser_get_use_preview_label) IA__gtk_file_chooser_get_use_preview_label __attribute((visibility("hidden")));
#define gtk_file_chooser_get_use_preview_label IA__gtk_file_chooser_get_use_preview_label

extern __typeof (gtk_file_chooser_list_filters) IA__gtk_file_chooser_list_filters __attribute((visibility("hidden")));
#define gtk_file_chooser_list_filters IA__gtk_file_chooser_list_filters

#ifndef _WIN64
extern __typeof (gtk_file_chooser_list_shortcut_folders) IA__gtk_file_chooser_list_shortcut_folders __attribute((visibility("hidden")));
#define gtk_file_chooser_list_shortcut_folders IA__gtk_file_chooser_list_shortcut_folders

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_chooser_list_shortcut_folders_utf8) IA__gtk_file_chooser_list_shortcut_folders_utf8 __attribute((visibility("hidden")));
#define gtk_file_chooser_list_shortcut_folders_utf8 IA__gtk_file_chooser_list_shortcut_folders_utf8

#endif
extern __typeof (gtk_file_chooser_list_shortcut_folder_uris) IA__gtk_file_chooser_list_shortcut_folder_uris __attribute((visibility("hidden")));
#define gtk_file_chooser_list_shortcut_folder_uris IA__gtk_file_chooser_list_shortcut_folder_uris

extern __typeof (gtk_file_chooser_remove_filter) IA__gtk_file_chooser_remove_filter __attribute((visibility("hidden")));
#define gtk_file_chooser_remove_filter IA__gtk_file_chooser_remove_filter

#ifndef _WIN64
extern __typeof (gtk_file_chooser_remove_shortcut_folder) IA__gtk_file_chooser_remove_shortcut_folder __attribute((visibility("hidden")));
#define gtk_file_chooser_remove_shortcut_folder IA__gtk_file_chooser_remove_shortcut_folder

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_chooser_remove_shortcut_folder_utf8) IA__gtk_file_chooser_remove_shortcut_folder_utf8 __attribute((visibility("hidden")));
#define gtk_file_chooser_remove_shortcut_folder_utf8 IA__gtk_file_chooser_remove_shortcut_folder_utf8

#endif
extern __typeof (gtk_file_chooser_remove_shortcut_folder_uri) IA__gtk_file_chooser_remove_shortcut_folder_uri __attribute((visibility("hidden")));
#define gtk_file_chooser_remove_shortcut_folder_uri IA__gtk_file_chooser_remove_shortcut_folder_uri

extern __typeof (gtk_file_chooser_select_all) IA__gtk_file_chooser_select_all __attribute((visibility("hidden")));
#define gtk_file_chooser_select_all IA__gtk_file_chooser_select_all

extern __typeof (gtk_file_chooser_select_file) IA__gtk_file_chooser_select_file __attribute((visibility("hidden")));
#define gtk_file_chooser_select_file IA__gtk_file_chooser_select_file

#ifndef _WIN64
extern __typeof (gtk_file_chooser_select_filename) IA__gtk_file_chooser_select_filename __attribute((visibility("hidden")));
#define gtk_file_chooser_select_filename IA__gtk_file_chooser_select_filename

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_chooser_select_filename_utf8) IA__gtk_file_chooser_select_filename_utf8 __attribute((visibility("hidden")));
#define gtk_file_chooser_select_filename_utf8 IA__gtk_file_chooser_select_filename_utf8

#endif
extern __typeof (gtk_file_chooser_select_uri) IA__gtk_file_chooser_select_uri __attribute((visibility("hidden")));
#define gtk_file_chooser_select_uri IA__gtk_file_chooser_select_uri

extern __typeof (gtk_file_chooser_set_action) IA__gtk_file_chooser_set_action __attribute((visibility("hidden")));
#define gtk_file_chooser_set_action IA__gtk_file_chooser_set_action

#ifndef _WIN64
extern __typeof (gtk_file_chooser_set_current_folder) IA__gtk_file_chooser_set_current_folder __attribute((visibility("hidden")));
#define gtk_file_chooser_set_current_folder IA__gtk_file_chooser_set_current_folder

#endif
extern __typeof (gtk_file_chooser_set_current_folder_file) IA__gtk_file_chooser_set_current_folder_file __attribute((visibility("hidden")));
#define gtk_file_chooser_set_current_folder_file IA__gtk_file_chooser_set_current_folder_file

#ifdef G_OS_WIN32
extern __typeof (gtk_file_chooser_set_current_folder_utf8) IA__gtk_file_chooser_set_current_folder_utf8 __attribute((visibility("hidden")));
#define gtk_file_chooser_set_current_folder_utf8 IA__gtk_file_chooser_set_current_folder_utf8

#endif
extern __typeof (gtk_file_chooser_set_current_folder_uri) IA__gtk_file_chooser_set_current_folder_uri __attribute((visibility("hidden")));
#define gtk_file_chooser_set_current_folder_uri IA__gtk_file_chooser_set_current_folder_uri

extern __typeof (gtk_file_chooser_set_current_name) IA__gtk_file_chooser_set_current_name __attribute((visibility("hidden")));
#define gtk_file_chooser_set_current_name IA__gtk_file_chooser_set_current_name

extern __typeof (gtk_file_chooser_set_extra_widget) IA__gtk_file_chooser_set_extra_widget __attribute((visibility("hidden")));
#define gtk_file_chooser_set_extra_widget IA__gtk_file_chooser_set_extra_widget

extern __typeof (gtk_file_chooser_set_file) IA__gtk_file_chooser_set_file __attribute((visibility("hidden")));
#define gtk_file_chooser_set_file IA__gtk_file_chooser_set_file

#ifndef _WIN64
extern __typeof (gtk_file_chooser_set_filename) IA__gtk_file_chooser_set_filename __attribute((visibility("hidden")));
#define gtk_file_chooser_set_filename IA__gtk_file_chooser_set_filename

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_chooser_set_filename_utf8) IA__gtk_file_chooser_set_filename_utf8 __attribute((visibility("hidden")));
#define gtk_file_chooser_set_filename_utf8 IA__gtk_file_chooser_set_filename_utf8

#endif
extern __typeof (gtk_file_chooser_set_filter) IA__gtk_file_chooser_set_filter __attribute((visibility("hidden")));
#define gtk_file_chooser_set_filter IA__gtk_file_chooser_set_filter

extern __typeof (gtk_file_chooser_set_local_only) IA__gtk_file_chooser_set_local_only __attribute((visibility("hidden")));
#define gtk_file_chooser_set_local_only IA__gtk_file_chooser_set_local_only

extern __typeof (gtk_file_chooser_set_preview_widget) IA__gtk_file_chooser_set_preview_widget __attribute((visibility("hidden")));
#define gtk_file_chooser_set_preview_widget IA__gtk_file_chooser_set_preview_widget

extern __typeof (gtk_file_chooser_set_preview_widget_active) IA__gtk_file_chooser_set_preview_widget_active __attribute((visibility("hidden")));
#define gtk_file_chooser_set_preview_widget_active IA__gtk_file_chooser_set_preview_widget_active

extern __typeof (gtk_file_chooser_set_select_multiple) IA__gtk_file_chooser_set_select_multiple __attribute((visibility("hidden")));
#define gtk_file_chooser_set_select_multiple IA__gtk_file_chooser_set_select_multiple

extern __typeof (gtk_file_chooser_set_show_hidden) IA__gtk_file_chooser_set_show_hidden __attribute((visibility("hidden")));
#define gtk_file_chooser_set_show_hidden IA__gtk_file_chooser_set_show_hidden

extern __typeof (gtk_file_chooser_set_uri) IA__gtk_file_chooser_set_uri __attribute((visibility("hidden")));
#define gtk_file_chooser_set_uri IA__gtk_file_chooser_set_uri

extern __typeof (gtk_file_chooser_set_use_preview_label) IA__gtk_file_chooser_set_use_preview_label __attribute((visibility("hidden")));
#define gtk_file_chooser_set_use_preview_label IA__gtk_file_chooser_set_use_preview_label

extern __typeof (gtk_file_chooser_unselect_all) IA__gtk_file_chooser_unselect_all __attribute((visibility("hidden")));
#define gtk_file_chooser_unselect_all IA__gtk_file_chooser_unselect_all

extern __typeof (gtk_file_chooser_unselect_file) IA__gtk_file_chooser_unselect_file __attribute((visibility("hidden")));
#define gtk_file_chooser_unselect_file IA__gtk_file_chooser_unselect_file

#ifndef _WIN64
extern __typeof (gtk_file_chooser_unselect_filename) IA__gtk_file_chooser_unselect_filename __attribute((visibility("hidden")));
#define gtk_file_chooser_unselect_filename IA__gtk_file_chooser_unselect_filename

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_chooser_unselect_filename_utf8) IA__gtk_file_chooser_unselect_filename_utf8 __attribute((visibility("hidden")));
#define gtk_file_chooser_unselect_filename_utf8 IA__gtk_file_chooser_unselect_filename_utf8

#endif
extern __typeof (gtk_file_chooser_unselect_uri) IA__gtk_file_chooser_unselect_uri __attribute((visibility("hidden")));
#define gtk_file_chooser_unselect_uri IA__gtk_file_chooser_unselect_uri

#endif
#endif
#if IN_HEADER(__GTK_FILE_CHOOSER_DIALOG_H__)
#if IN_FILE(__GTK_FILE_CHOOSER_DIALOG_C__)
extern __typeof (gtk_file_chooser_dialog_get_type) IA__gtk_file_chooser_dialog_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_file_chooser_dialog_get_type IA__gtk_file_chooser_dialog_get_type

extern __typeof (gtk_file_chooser_dialog_new) IA__gtk_file_chooser_dialog_new __attribute((visibility("hidden")));
#define gtk_file_chooser_dialog_new IA__gtk_file_chooser_dialog_new

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_file_chooser_dialog_new_with_backend) IA__gtk_file_chooser_dialog_new_with_backend __attribute((visibility("hidden")));
#define gtk_file_chooser_dialog_new_with_backend IA__gtk_file_chooser_dialog_new_with_backend

#endif
#endif
#endif
#if IN_HEADER(__GTK_FILE_CHOOSER_WIDGET_H__)
#if IN_FILE(__GTK_FILE_CHOOSER_WIDGET_C__)
extern __typeof (gtk_file_chooser_widget_get_type) IA__gtk_file_chooser_widget_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_file_chooser_widget_get_type IA__gtk_file_chooser_widget_get_type

extern __typeof (gtk_file_chooser_widget_new) IA__gtk_file_chooser_widget_new __attribute((visibility("hidden")));
#define gtk_file_chooser_widget_new IA__gtk_file_chooser_widget_new

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_file_chooser_widget_new_with_backend) IA__gtk_file_chooser_widget_new_with_backend __attribute((visibility("hidden")));
#define gtk_file_chooser_widget_new_with_backend IA__gtk_file_chooser_widget_new_with_backend

#endif
#endif
#endif
#if IN_HEADER(__GTK_FILE_CHOOSER_BUTTON_H__)
#if IN_FILE(__GTK_FILE_CHOOSER_BUTTON_C__)
extern __typeof (gtk_file_chooser_button_get_title) IA__gtk_file_chooser_button_get_title __attribute((visibility("hidden")));
#define gtk_file_chooser_button_get_title IA__gtk_file_chooser_button_get_title

extern __typeof (gtk_file_chooser_button_get_type) IA__gtk_file_chooser_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_file_chooser_button_get_type IA__gtk_file_chooser_button_get_type

extern __typeof (gtk_file_chooser_button_get_width_chars) IA__gtk_file_chooser_button_get_width_chars __attribute((visibility("hidden")));
#define gtk_file_chooser_button_get_width_chars IA__gtk_file_chooser_button_get_width_chars

extern __typeof (gtk_file_chooser_button_new) IA__gtk_file_chooser_button_new __attribute((visibility("hidden")));
#define gtk_file_chooser_button_new IA__gtk_file_chooser_button_new

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_file_chooser_button_new_with_backend) IA__gtk_file_chooser_button_new_with_backend __attribute((visibility("hidden")));
#define gtk_file_chooser_button_new_with_backend IA__gtk_file_chooser_button_new_with_backend

#endif
extern __typeof (gtk_file_chooser_button_new_with_dialog) IA__gtk_file_chooser_button_new_with_dialog __attribute((visibility("hidden")));
#define gtk_file_chooser_button_new_with_dialog IA__gtk_file_chooser_button_new_with_dialog

extern __typeof (gtk_file_chooser_button_set_title) IA__gtk_file_chooser_button_set_title __attribute((visibility("hidden")));
#define gtk_file_chooser_button_set_title IA__gtk_file_chooser_button_set_title

extern __typeof (gtk_file_chooser_button_set_width_chars) IA__gtk_file_chooser_button_set_width_chars __attribute((visibility("hidden")));
#define gtk_file_chooser_button_set_width_chars IA__gtk_file_chooser_button_set_width_chars

extern __typeof (gtk_file_chooser_button_get_focus_on_click) IA__gtk_file_chooser_button_get_focus_on_click __attribute((visibility("hidden")));
#define gtk_file_chooser_button_get_focus_on_click IA__gtk_file_chooser_button_get_focus_on_click

extern __typeof (gtk_file_chooser_button_set_focus_on_click) IA__gtk_file_chooser_button_set_focus_on_click __attribute((visibility("hidden")));
#define gtk_file_chooser_button_set_focus_on_click IA__gtk_file_chooser_button_set_focus_on_click

#endif
#endif
#if IN_HEADER(__GTK_FILE_FILTER_H__)
#if IN_FILE(__GTK_FILE_FILTER_C__)
extern __typeof (gtk_file_filter_add_custom) IA__gtk_file_filter_add_custom __attribute((visibility("hidden")));
#define gtk_file_filter_add_custom IA__gtk_file_filter_add_custom

extern __typeof (gtk_file_filter_add_mime_type) IA__gtk_file_filter_add_mime_type __attribute((visibility("hidden")));
#define gtk_file_filter_add_mime_type IA__gtk_file_filter_add_mime_type

extern __typeof (gtk_file_filter_add_pattern) IA__gtk_file_filter_add_pattern __attribute((visibility("hidden")));
#define gtk_file_filter_add_pattern IA__gtk_file_filter_add_pattern

extern __typeof (gtk_file_filter_add_pixbuf_formats) IA__gtk_file_filter_add_pixbuf_formats __attribute((visibility("hidden")));
#define gtk_file_filter_add_pixbuf_formats IA__gtk_file_filter_add_pixbuf_formats

extern __typeof (gtk_file_filter_filter) IA__gtk_file_filter_filter __attribute((visibility("hidden")));
#define gtk_file_filter_filter IA__gtk_file_filter_filter

extern __typeof (gtk_file_filter_get_name) IA__gtk_file_filter_get_name __attribute((visibility("hidden")));
#define gtk_file_filter_get_name IA__gtk_file_filter_get_name

extern __typeof (gtk_file_filter_get_needed) IA__gtk_file_filter_get_needed __attribute((visibility("hidden")));
#define gtk_file_filter_get_needed IA__gtk_file_filter_get_needed

extern __typeof (gtk_file_filter_get_type) IA__gtk_file_filter_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_file_filter_get_type IA__gtk_file_filter_get_type

extern __typeof (gtk_file_filter_new) IA__gtk_file_filter_new __attribute((visibility("hidden")));
#define gtk_file_filter_new IA__gtk_file_filter_new

extern __typeof (gtk_file_filter_set_name) IA__gtk_file_filter_set_name __attribute((visibility("hidden")));
#define gtk_file_filter_set_name IA__gtk_file_filter_set_name

#endif
#endif
#if IN_HEADER(__GTK_FILESEL_H__)
#if IN_FILE(__GTK_FILESEL_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_file_selection_complete) IA__gtk_file_selection_complete __attribute((visibility("hidden")));
#define gtk_file_selection_complete IA__gtk_file_selection_complete

#ifndef _WIN64
extern __typeof (gtk_file_selection_get_filename) IA__gtk_file_selection_get_filename __attribute((visibility("hidden")));
#define gtk_file_selection_get_filename IA__gtk_file_selection_get_filename

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_selection_get_filename_utf8) IA__gtk_file_selection_get_filename_utf8 __attribute((visibility("hidden")));
#define gtk_file_selection_get_filename_utf8 IA__gtk_file_selection_get_filename_utf8

#endif
#ifndef _WIN64
extern __typeof (gtk_file_selection_get_selections) IA__gtk_file_selection_get_selections __attribute((visibility("hidden")));
#define gtk_file_selection_get_selections IA__gtk_file_selection_get_selections

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_selection_get_selections_utf8) IA__gtk_file_selection_get_selections_utf8 __attribute((visibility("hidden")));
#define gtk_file_selection_get_selections_utf8 IA__gtk_file_selection_get_selections_utf8

#endif
extern __typeof (gtk_file_selection_get_select_multiple) IA__gtk_file_selection_get_select_multiple __attribute((visibility("hidden")));
#define gtk_file_selection_get_select_multiple IA__gtk_file_selection_get_select_multiple

extern __typeof (gtk_file_selection_get_type) IA__gtk_file_selection_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_file_selection_get_type IA__gtk_file_selection_get_type

extern __typeof (gtk_file_selection_hide_fileop_buttons) IA__gtk_file_selection_hide_fileop_buttons __attribute((visibility("hidden")));
#define gtk_file_selection_hide_fileop_buttons IA__gtk_file_selection_hide_fileop_buttons

extern __typeof (gtk_file_selection_new) IA__gtk_file_selection_new __attribute((visibility("hidden")));
#define gtk_file_selection_new IA__gtk_file_selection_new

#ifndef _WIN64
extern __typeof (gtk_file_selection_set_filename) IA__gtk_file_selection_set_filename __attribute((visibility("hidden")));
#define gtk_file_selection_set_filename IA__gtk_file_selection_set_filename

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_file_selection_set_filename_utf8) IA__gtk_file_selection_set_filename_utf8 __attribute((visibility("hidden")));
#define gtk_file_selection_set_filename_utf8 IA__gtk_file_selection_set_filename_utf8

#endif
extern __typeof (gtk_file_selection_set_select_multiple) IA__gtk_file_selection_set_select_multiple __attribute((visibility("hidden")));
#define gtk_file_selection_set_select_multiple IA__gtk_file_selection_set_select_multiple

extern __typeof (gtk_file_selection_show_fileop_buttons) IA__gtk_file_selection_show_fileop_buttons __attribute((visibility("hidden")));
#define gtk_file_selection_show_fileop_buttons IA__gtk_file_selection_show_fileop_buttons

#endif
#endif
#endif
#if IN_HEADER(__GTK_FIXED_H__)
#if IN_FILE(__GTK_FIXED_C__)
extern __typeof (gtk_fixed_get_type) IA__gtk_fixed_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_fixed_get_type IA__gtk_fixed_get_type

extern __typeof (gtk_fixed_move) IA__gtk_fixed_move __attribute((visibility("hidden")));
#define gtk_fixed_move IA__gtk_fixed_move

extern __typeof (gtk_fixed_new) IA__gtk_fixed_new __attribute((visibility("hidden")));
#define gtk_fixed_new IA__gtk_fixed_new

extern __typeof (gtk_fixed_put) IA__gtk_fixed_put __attribute((visibility("hidden")));
#define gtk_fixed_put IA__gtk_fixed_put

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_fixed_set_has_window) IA__gtk_fixed_set_has_window __attribute((visibility("hidden")));
#define gtk_fixed_set_has_window IA__gtk_fixed_set_has_window

extern __typeof (gtk_fixed_get_has_window) IA__gtk_fixed_get_has_window __attribute((visibility("hidden")));
#define gtk_fixed_get_has_window IA__gtk_fixed_get_has_window

#endif
#endif
#endif
#if IN_HEADER(__GTK_FONT_BUTTON_H__)
#if IN_FILE(__GTK_FONT_BUTTON_C__)
extern __typeof (gtk_font_button_get_font_name) IA__gtk_font_button_get_font_name __attribute((visibility("hidden")));
#define gtk_font_button_get_font_name IA__gtk_font_button_get_font_name

extern __typeof (gtk_font_button_get_show_size) IA__gtk_font_button_get_show_size __attribute((visibility("hidden")));
#define gtk_font_button_get_show_size IA__gtk_font_button_get_show_size

extern __typeof (gtk_font_button_get_show_style) IA__gtk_font_button_get_show_style __attribute((visibility("hidden")));
#define gtk_font_button_get_show_style IA__gtk_font_button_get_show_style

extern __typeof (gtk_font_button_get_title) IA__gtk_font_button_get_title __attribute((visibility("hidden")));
#define gtk_font_button_get_title IA__gtk_font_button_get_title

extern __typeof (gtk_font_button_get_type) IA__gtk_font_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_font_button_get_type IA__gtk_font_button_get_type

extern __typeof (gtk_font_button_get_use_font) IA__gtk_font_button_get_use_font __attribute((visibility("hidden")));
#define gtk_font_button_get_use_font IA__gtk_font_button_get_use_font

extern __typeof (gtk_font_button_get_use_size) IA__gtk_font_button_get_use_size __attribute((visibility("hidden")));
#define gtk_font_button_get_use_size IA__gtk_font_button_get_use_size

extern __typeof (gtk_font_button_new) IA__gtk_font_button_new __attribute((visibility("hidden")));
#define gtk_font_button_new IA__gtk_font_button_new

extern __typeof (gtk_font_button_new_with_font) IA__gtk_font_button_new_with_font __attribute((visibility("hidden")));
#define gtk_font_button_new_with_font IA__gtk_font_button_new_with_font

extern __typeof (gtk_font_button_set_font_name) IA__gtk_font_button_set_font_name __attribute((visibility("hidden")));
#define gtk_font_button_set_font_name IA__gtk_font_button_set_font_name

extern __typeof (gtk_font_button_set_show_size) IA__gtk_font_button_set_show_size __attribute((visibility("hidden")));
#define gtk_font_button_set_show_size IA__gtk_font_button_set_show_size

extern __typeof (gtk_font_button_set_show_style) IA__gtk_font_button_set_show_style __attribute((visibility("hidden")));
#define gtk_font_button_set_show_style IA__gtk_font_button_set_show_style

extern __typeof (gtk_font_button_set_title) IA__gtk_font_button_set_title __attribute((visibility("hidden")));
#define gtk_font_button_set_title IA__gtk_font_button_set_title

extern __typeof (gtk_font_button_set_use_font) IA__gtk_font_button_set_use_font __attribute((visibility("hidden")));
#define gtk_font_button_set_use_font IA__gtk_font_button_set_use_font

extern __typeof (gtk_font_button_set_use_size) IA__gtk_font_button_set_use_size __attribute((visibility("hidden")));
#define gtk_font_button_set_use_size IA__gtk_font_button_set_use_size

#endif
#endif
#if IN_HEADER(__GTK_FONTSEL_H__)
#if IN_FILE(__GTK_FONTSEL_C__)
extern __typeof (gtk_font_selection_dialog_get_cancel_button) IA__gtk_font_selection_dialog_get_cancel_button __attribute((visibility("hidden")));
#define gtk_font_selection_dialog_get_cancel_button IA__gtk_font_selection_dialog_get_cancel_button

extern __typeof (gtk_font_selection_dialog_get_font_name) IA__gtk_font_selection_dialog_get_font_name __attribute((visibility("hidden")));
#define gtk_font_selection_dialog_get_font_name IA__gtk_font_selection_dialog_get_font_name

extern __typeof (gtk_font_selection_dialog_get_font_selection) IA__gtk_font_selection_dialog_get_font_selection __attribute((visibility("hidden")));
#define gtk_font_selection_dialog_get_font_selection IA__gtk_font_selection_dialog_get_font_selection

extern __typeof (gtk_font_selection_dialog_get_ok_button) IA__gtk_font_selection_dialog_get_ok_button __attribute((visibility("hidden")));
#define gtk_font_selection_dialog_get_ok_button IA__gtk_font_selection_dialog_get_ok_button

extern __typeof (gtk_font_selection_dialog_get_preview_text) IA__gtk_font_selection_dialog_get_preview_text __attribute((visibility("hidden")));
#define gtk_font_selection_dialog_get_preview_text IA__gtk_font_selection_dialog_get_preview_text

extern __typeof (gtk_font_selection_dialog_get_type) IA__gtk_font_selection_dialog_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_font_selection_dialog_get_type IA__gtk_font_selection_dialog_get_type

extern __typeof (gtk_font_selection_dialog_new) IA__gtk_font_selection_dialog_new __attribute((visibility("hidden")));
#define gtk_font_selection_dialog_new IA__gtk_font_selection_dialog_new

extern __typeof (gtk_font_selection_dialog_set_font_name) IA__gtk_font_selection_dialog_set_font_name __attribute((visibility("hidden")));
#define gtk_font_selection_dialog_set_font_name IA__gtk_font_selection_dialog_set_font_name

extern __typeof (gtk_font_selection_dialog_set_preview_text) IA__gtk_font_selection_dialog_set_preview_text __attribute((visibility("hidden")));
#define gtk_font_selection_dialog_set_preview_text IA__gtk_font_selection_dialog_set_preview_text

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_font_selection_dialog_get_apply_button) IA__gtk_font_selection_dialog_get_apply_button __attribute((visibility("hidden")));
#define gtk_font_selection_dialog_get_apply_button IA__gtk_font_selection_dialog_get_apply_button

extern __typeof (gtk_font_selection_get_font) IA__gtk_font_selection_get_font __attribute((visibility("hidden")));
#define gtk_font_selection_get_font IA__gtk_font_selection_get_font

extern __typeof (gtk_font_selection_dialog_get_font) IA__gtk_font_selection_dialog_get_font __attribute((visibility("hidden")));
#define gtk_font_selection_dialog_get_font IA__gtk_font_selection_dialog_get_font

#endif
extern __typeof (gtk_font_selection_get_face) IA__gtk_font_selection_get_face __attribute((visibility("hidden")));
#define gtk_font_selection_get_face IA__gtk_font_selection_get_face

extern __typeof (gtk_font_selection_get_face_list) IA__gtk_font_selection_get_face_list __attribute((visibility("hidden")));
#define gtk_font_selection_get_face_list IA__gtk_font_selection_get_face_list

extern __typeof (gtk_font_selection_get_family) IA__gtk_font_selection_get_family __attribute((visibility("hidden")));
#define gtk_font_selection_get_family IA__gtk_font_selection_get_family

extern __typeof (gtk_font_selection_get_family_list) IA__gtk_font_selection_get_family_list __attribute((visibility("hidden")));
#define gtk_font_selection_get_family_list IA__gtk_font_selection_get_family_list

extern __typeof (gtk_font_selection_get_font_name) IA__gtk_font_selection_get_font_name __attribute((visibility("hidden")));
#define gtk_font_selection_get_font_name IA__gtk_font_selection_get_font_name

extern __typeof (gtk_font_selection_get_preview_entry) IA__gtk_font_selection_get_preview_entry __attribute((visibility("hidden")));
#define gtk_font_selection_get_preview_entry IA__gtk_font_selection_get_preview_entry

extern __typeof (gtk_font_selection_get_preview_text) IA__gtk_font_selection_get_preview_text __attribute((visibility("hidden")));
#define gtk_font_selection_get_preview_text IA__gtk_font_selection_get_preview_text

extern __typeof (gtk_font_selection_get_size) IA__gtk_font_selection_get_size __attribute((visibility("hidden")));
#define gtk_font_selection_get_size IA__gtk_font_selection_get_size

extern __typeof (gtk_font_selection_get_size_entry) IA__gtk_font_selection_get_size_entry __attribute((visibility("hidden")));
#define gtk_font_selection_get_size_entry IA__gtk_font_selection_get_size_entry

extern __typeof (gtk_font_selection_get_size_list) IA__gtk_font_selection_get_size_list __attribute((visibility("hidden")));
#define gtk_font_selection_get_size_list IA__gtk_font_selection_get_size_list

extern __typeof (gtk_font_selection_get_type) IA__gtk_font_selection_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_font_selection_get_type IA__gtk_font_selection_get_type

extern __typeof (gtk_font_selection_new) IA__gtk_font_selection_new __attribute((visibility("hidden")));
#define gtk_font_selection_new IA__gtk_font_selection_new

extern __typeof (gtk_font_selection_set_font_name) IA__gtk_font_selection_set_font_name __attribute((visibility("hidden")));
#define gtk_font_selection_set_font_name IA__gtk_font_selection_set_font_name

extern __typeof (gtk_font_selection_set_preview_text) IA__gtk_font_selection_set_preview_text __attribute((visibility("hidden")));
#define gtk_font_selection_set_preview_text IA__gtk_font_selection_set_preview_text

#endif
#endif
#if IN_HEADER(__GTK_FRAME_H__)
#if IN_FILE(__GTK_FRAME_C__)
extern __typeof (gtk_frame_get_label) IA__gtk_frame_get_label __attribute((visibility("hidden")));
#define gtk_frame_get_label IA__gtk_frame_get_label

extern __typeof (gtk_frame_get_label_align) IA__gtk_frame_get_label_align __attribute((visibility("hidden")));
#define gtk_frame_get_label_align IA__gtk_frame_get_label_align

extern __typeof (gtk_frame_get_label_widget) IA__gtk_frame_get_label_widget __attribute((visibility("hidden")));
#define gtk_frame_get_label_widget IA__gtk_frame_get_label_widget

extern __typeof (gtk_frame_get_shadow_type) IA__gtk_frame_get_shadow_type __attribute((visibility("hidden")));
#define gtk_frame_get_shadow_type IA__gtk_frame_get_shadow_type

extern __typeof (gtk_frame_get_type) IA__gtk_frame_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_frame_get_type IA__gtk_frame_get_type

extern __typeof (gtk_frame_new) IA__gtk_frame_new __attribute((visibility("hidden")));
#define gtk_frame_new IA__gtk_frame_new

extern __typeof (gtk_frame_set_label) IA__gtk_frame_set_label __attribute((visibility("hidden")));
#define gtk_frame_set_label IA__gtk_frame_set_label

extern __typeof (gtk_frame_set_label_align) IA__gtk_frame_set_label_align __attribute((visibility("hidden")));
#define gtk_frame_set_label_align IA__gtk_frame_set_label_align

extern __typeof (gtk_frame_set_label_widget) IA__gtk_frame_set_label_widget __attribute((visibility("hidden")));
#define gtk_frame_set_label_widget IA__gtk_frame_set_label_widget

extern __typeof (gtk_frame_set_shadow_type) IA__gtk_frame_set_shadow_type __attribute((visibility("hidden")));
#define gtk_frame_set_shadow_type IA__gtk_frame_set_shadow_type

#endif
#endif
#if IN_HEADER(__GTK_GAMMA_CURVE_H__)
#if IN_FILE(__GTK_GAMMA_CURVE_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_gamma_curve_get_type) IA__gtk_gamma_curve_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_gamma_curve_get_type IA__gtk_gamma_curve_get_type

extern __typeof (gtk_gamma_curve_new) IA__gtk_gamma_curve_new __attribute((visibility("hidden")));
#define gtk_gamma_curve_new IA__gtk_gamma_curve_new

#endif
#endif
#endif
#if IN_HEADER(__GTK_GC_H__)
#if IN_FILE(__GTK_GC_C__)
extern __typeof (gtk_gc_get) IA__gtk_gc_get __attribute((visibility("hidden")));
#define gtk_gc_get IA__gtk_gc_get

extern __typeof (gtk_gc_release) IA__gtk_gc_release __attribute((visibility("hidden")));
#define gtk_gc_release IA__gtk_gc_release

#endif
#endif
#if IN_HEADER(__GTK_HANDLE_BOX_H__)
#if IN_FILE(__GTK_HANDLE_BOX_C__)
extern __typeof (gtk_handle_box_get_handle_position) IA__gtk_handle_box_get_handle_position __attribute((visibility("hidden")));
#define gtk_handle_box_get_handle_position IA__gtk_handle_box_get_handle_position

extern __typeof (gtk_handle_box_get_shadow_type) IA__gtk_handle_box_get_shadow_type __attribute((visibility("hidden")));
#define gtk_handle_box_get_shadow_type IA__gtk_handle_box_get_shadow_type

extern __typeof (gtk_handle_box_get_snap_edge) IA__gtk_handle_box_get_snap_edge __attribute((visibility("hidden")));
#define gtk_handle_box_get_snap_edge IA__gtk_handle_box_get_snap_edge

extern __typeof (gtk_handle_box_get_child_detached) IA__gtk_handle_box_get_child_detached __attribute((visibility("hidden")));
#define gtk_handle_box_get_child_detached IA__gtk_handle_box_get_child_detached

extern __typeof (gtk_handle_box_get_type) IA__gtk_handle_box_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_handle_box_get_type IA__gtk_handle_box_get_type

extern __typeof (gtk_handle_box_new) IA__gtk_handle_box_new __attribute((visibility("hidden")));
#define gtk_handle_box_new IA__gtk_handle_box_new

extern __typeof (gtk_handle_box_set_handle_position) IA__gtk_handle_box_set_handle_position __attribute((visibility("hidden")));
#define gtk_handle_box_set_handle_position IA__gtk_handle_box_set_handle_position

extern __typeof (gtk_handle_box_set_shadow_type) IA__gtk_handle_box_set_shadow_type __attribute((visibility("hidden")));
#define gtk_handle_box_set_shadow_type IA__gtk_handle_box_set_shadow_type

extern __typeof (gtk_handle_box_set_snap_edge) IA__gtk_handle_box_set_snap_edge __attribute((visibility("hidden")));
#define gtk_handle_box_set_snap_edge IA__gtk_handle_box_set_snap_edge

#endif
#endif
#if IN_HEADER(__GTK_HBOX_H__)
#if IN_FILE(__GTK_HBOX_C__)
extern __typeof (gtk_hbox_get_type) IA__gtk_hbox_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_hbox_get_type IA__gtk_hbox_get_type

extern __typeof (gtk_hbox_new) IA__gtk_hbox_new __attribute((visibility("hidden")));
#define gtk_hbox_new IA__gtk_hbox_new

#endif
#endif
#if IN_HEADER(__GTK_HBUTTON_BOX_H__)
#if IN_FILE(__GTK_HBUTTON_BOX_C__)
extern __typeof (gtk_hbutton_box_new) IA__gtk_hbutton_box_new __attribute((visibility("hidden")));
#define gtk_hbutton_box_new IA__gtk_hbutton_box_new

extern __typeof (gtk_hbutton_box_get_type) IA__gtk_hbutton_box_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_hbutton_box_get_type IA__gtk_hbutton_box_get_type

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_hbutton_box_get_layout_default) IA__gtk_hbutton_box_get_layout_default __attribute((visibility("hidden")));
#define gtk_hbutton_box_get_layout_default IA__gtk_hbutton_box_get_layout_default

extern __typeof (gtk_hbutton_box_get_spacing_default) IA__gtk_hbutton_box_get_spacing_default __attribute((visibility("hidden")));
#define gtk_hbutton_box_get_spacing_default IA__gtk_hbutton_box_get_spacing_default

extern __typeof (gtk_hbutton_box_set_layout_default) IA__gtk_hbutton_box_set_layout_default __attribute((visibility("hidden")));
#define gtk_hbutton_box_set_layout_default IA__gtk_hbutton_box_set_layout_default

extern __typeof (gtk_hbutton_box_set_spacing_default) IA__gtk_hbutton_box_set_spacing_default __attribute((visibility("hidden")));
#define gtk_hbutton_box_set_spacing_default IA__gtk_hbutton_box_set_spacing_default

#endif
#endif
#endif
#if IN_HEADER(__GTK_HPANED_H__)
#if IN_FILE(__GTK_HPANED_C__)
extern __typeof (gtk_hpaned_get_type) IA__gtk_hpaned_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_hpaned_get_type IA__gtk_hpaned_get_type

extern __typeof (gtk_hpaned_new) IA__gtk_hpaned_new __attribute((visibility("hidden")));
#define gtk_hpaned_new IA__gtk_hpaned_new

#endif
#endif
#if IN_HEADER(__GTK_HRULER_H__)
#if IN_FILE(__GTK_HRULER_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_hruler_get_type) IA__gtk_hruler_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_hruler_get_type IA__gtk_hruler_get_type

extern __typeof (gtk_hruler_new) IA__gtk_hruler_new __attribute((visibility("hidden")));
#define gtk_hruler_new IA__gtk_hruler_new

#endif
#endif
#endif
#if IN_HEADER(__GTK_HSCALE_H__)
#if IN_FILE(__GTK_HSCALE_C__)
extern __typeof (gtk_hscale_get_type) IA__gtk_hscale_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_hscale_get_type IA__gtk_hscale_get_type

extern __typeof (gtk_hscale_new) IA__gtk_hscale_new __attribute((visibility("hidden")));
#define gtk_hscale_new IA__gtk_hscale_new

extern __typeof (gtk_hscale_new_with_range) IA__gtk_hscale_new_with_range __attribute((visibility("hidden")));
#define gtk_hscale_new_with_range IA__gtk_hscale_new_with_range

#endif
#endif
#if IN_HEADER(__GTK_HSCROLLBAR_H__)
#if IN_FILE(__GTK_HSCROLLBAR_C__)
extern __typeof (gtk_hscrollbar_get_type) IA__gtk_hscrollbar_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_hscrollbar_get_type IA__gtk_hscrollbar_get_type

extern __typeof (gtk_hscrollbar_new) IA__gtk_hscrollbar_new __attribute((visibility("hidden")));
#define gtk_hscrollbar_new IA__gtk_hscrollbar_new

#endif
#endif
#if IN_HEADER(__GTK_HSEPARATOR_H__)
#if IN_FILE(__GTK_HSEPARATOR_C__)
extern __typeof (gtk_hseparator_get_type) IA__gtk_hseparator_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_hseparator_get_type IA__gtk_hseparator_get_type

extern __typeof (gtk_hseparator_new) IA__gtk_hseparator_new __attribute((visibility("hidden")));
#define gtk_hseparator_new IA__gtk_hseparator_new

#endif
#endif
#if IN_HEADER(__GTK_HSV_H__)
#if IN_FILE(__GTK_HSV_C__)
extern __typeof (gtk_hsv_get_color) IA__gtk_hsv_get_color __attribute((visibility("hidden")));
#define gtk_hsv_get_color IA__gtk_hsv_get_color

extern __typeof (gtk_hsv_get_metrics) IA__gtk_hsv_get_metrics __attribute((visibility("hidden")));
#define gtk_hsv_get_metrics IA__gtk_hsv_get_metrics

extern __typeof (gtk_hsv_get_type) IA__gtk_hsv_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_hsv_get_type IA__gtk_hsv_get_type

extern __typeof (gtk_hsv_is_adjusting) IA__gtk_hsv_is_adjusting __attribute((visibility("hidden")));
#define gtk_hsv_is_adjusting IA__gtk_hsv_is_adjusting

extern __typeof (gtk_hsv_new) IA__gtk_hsv_new __attribute((visibility("hidden")));
#define gtk_hsv_new IA__gtk_hsv_new

extern __typeof (gtk_hsv_set_color) IA__gtk_hsv_set_color __attribute((visibility("hidden")));
#define gtk_hsv_set_color IA__gtk_hsv_set_color

extern __typeof (gtk_hsv_set_metrics) IA__gtk_hsv_set_metrics __attribute((visibility("hidden")));
#define gtk_hsv_set_metrics IA__gtk_hsv_set_metrics

extern __typeof (gtk_hsv_to_rgb) IA__gtk_hsv_to_rgb __attribute((visibility("hidden")));
#define gtk_hsv_to_rgb IA__gtk_hsv_to_rgb

extern __typeof (gtk_rgb_to_hsv) IA__gtk_rgb_to_hsv __attribute((visibility("hidden")));
#define gtk_rgb_to_hsv IA__gtk_rgb_to_hsv

#endif
#endif
#if IN_HEADER(__GTK_ICON_FACTORY_H__)
#if IN_FILE(__GTK_ICON_FACTORY_C__)
extern __typeof (gtk_icon_factory_add) IA__gtk_icon_factory_add __attribute((visibility("hidden")));
#define gtk_icon_factory_add IA__gtk_icon_factory_add

extern __typeof (gtk_icon_factory_add_default) IA__gtk_icon_factory_add_default __attribute((visibility("hidden")));
#define gtk_icon_factory_add_default IA__gtk_icon_factory_add_default

extern __typeof (gtk_icon_factory_get_type) IA__gtk_icon_factory_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_icon_factory_get_type IA__gtk_icon_factory_get_type

extern __typeof (gtk_icon_factory_lookup) IA__gtk_icon_factory_lookup __attribute((visibility("hidden")));
#define gtk_icon_factory_lookup IA__gtk_icon_factory_lookup

extern __typeof (gtk_icon_factory_lookup_default) IA__gtk_icon_factory_lookup_default __attribute((visibility("hidden")));
#define gtk_icon_factory_lookup_default IA__gtk_icon_factory_lookup_default

extern __typeof (gtk_icon_factory_new) IA__gtk_icon_factory_new __attribute((visibility("hidden")));
#define gtk_icon_factory_new IA__gtk_icon_factory_new

extern __typeof (gtk_icon_factory_remove_default) IA__gtk_icon_factory_remove_default __attribute((visibility("hidden")));
#define gtk_icon_factory_remove_default IA__gtk_icon_factory_remove_default

extern __typeof (gtk_icon_size_from_name) IA__gtk_icon_size_from_name __attribute((visibility("hidden")));
#define gtk_icon_size_from_name IA__gtk_icon_size_from_name

extern __typeof (gtk_icon_size_get_name) IA__gtk_icon_size_get_name __attribute((visibility("hidden")));
#define gtk_icon_size_get_name IA__gtk_icon_size_get_name

extern __typeof (gtk_icon_size_lookup) IA__gtk_icon_size_lookup __attribute((visibility("hidden")));
#define gtk_icon_size_lookup IA__gtk_icon_size_lookup

extern __typeof (gtk_icon_size_lookup_for_settings) IA__gtk_icon_size_lookup_for_settings __attribute((visibility("hidden")));
#define gtk_icon_size_lookup_for_settings IA__gtk_icon_size_lookup_for_settings

extern __typeof (gtk_icon_size_register) IA__gtk_icon_size_register __attribute((visibility("hidden")));
#define gtk_icon_size_register IA__gtk_icon_size_register

extern __typeof (gtk_icon_size_register_alias) IA__gtk_icon_size_register_alias __attribute((visibility("hidden")));
#define gtk_icon_size_register_alias IA__gtk_icon_size_register_alias

extern __typeof (gtk_icon_set_add_source) IA__gtk_icon_set_add_source __attribute((visibility("hidden")));
#define gtk_icon_set_add_source IA__gtk_icon_set_add_source

extern __typeof (gtk_icon_set_copy) IA__gtk_icon_set_copy __attribute((visibility("hidden")));
#define gtk_icon_set_copy IA__gtk_icon_set_copy

extern __typeof (gtk_icon_set_get_sizes) IA__gtk_icon_set_get_sizes __attribute((visibility("hidden")));
#define gtk_icon_set_get_sizes IA__gtk_icon_set_get_sizes

extern __typeof (gtk_icon_set_get_type) IA__gtk_icon_set_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_icon_set_get_type IA__gtk_icon_set_get_type

extern __typeof (gtk_icon_set_new) IA__gtk_icon_set_new __attribute((visibility("hidden")));
#define gtk_icon_set_new IA__gtk_icon_set_new

extern __typeof (gtk_icon_set_new_from_pixbuf) IA__gtk_icon_set_new_from_pixbuf __attribute((visibility("hidden")));
#define gtk_icon_set_new_from_pixbuf IA__gtk_icon_set_new_from_pixbuf

extern __typeof (gtk_icon_set_ref) IA__gtk_icon_set_ref __attribute((visibility("hidden")));
#define gtk_icon_set_ref IA__gtk_icon_set_ref

extern __typeof (gtk_icon_set_render_icon) IA__gtk_icon_set_render_icon __attribute((visibility("hidden")));
#define gtk_icon_set_render_icon IA__gtk_icon_set_render_icon

extern __typeof (gtk_icon_set_unref) IA__gtk_icon_set_unref __attribute((visibility("hidden")));
#define gtk_icon_set_unref IA__gtk_icon_set_unref

extern __typeof (gtk_icon_source_copy) IA__gtk_icon_source_copy __attribute((visibility("hidden")));
#define gtk_icon_source_copy IA__gtk_icon_source_copy

extern __typeof (gtk_icon_source_free) IA__gtk_icon_source_free __attribute((visibility("hidden")));
#define gtk_icon_source_free IA__gtk_icon_source_free

extern __typeof (gtk_icon_source_get_direction) IA__gtk_icon_source_get_direction __attribute((visibility("hidden")));
#define gtk_icon_source_get_direction IA__gtk_icon_source_get_direction

extern __typeof (gtk_icon_source_get_direction_wildcarded) IA__gtk_icon_source_get_direction_wildcarded __attribute((visibility("hidden")));
#define gtk_icon_source_get_direction_wildcarded IA__gtk_icon_source_get_direction_wildcarded

#ifndef _WIN64
extern __typeof (gtk_icon_source_get_filename) IA__gtk_icon_source_get_filename __attribute((visibility("hidden")));
#define gtk_icon_source_get_filename IA__gtk_icon_source_get_filename

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_icon_source_get_filename_utf8) IA__gtk_icon_source_get_filename_utf8 __attribute((visibility("hidden")));
#define gtk_icon_source_get_filename_utf8 IA__gtk_icon_source_get_filename_utf8

#endif
extern __typeof (gtk_icon_source_get_icon_name) IA__gtk_icon_source_get_icon_name __attribute((visibility("hidden")));
#define gtk_icon_source_get_icon_name IA__gtk_icon_source_get_icon_name

extern __typeof (gtk_icon_source_get_pixbuf) IA__gtk_icon_source_get_pixbuf __attribute((visibility("hidden")));
#define gtk_icon_source_get_pixbuf IA__gtk_icon_source_get_pixbuf

extern __typeof (gtk_icon_source_get_size) IA__gtk_icon_source_get_size __attribute((visibility("hidden")));
#define gtk_icon_source_get_size IA__gtk_icon_source_get_size

extern __typeof (gtk_icon_source_get_size_wildcarded) IA__gtk_icon_source_get_size_wildcarded __attribute((visibility("hidden")));
#define gtk_icon_source_get_size_wildcarded IA__gtk_icon_source_get_size_wildcarded

extern __typeof (gtk_icon_source_get_state) IA__gtk_icon_source_get_state __attribute((visibility("hidden")));
#define gtk_icon_source_get_state IA__gtk_icon_source_get_state

extern __typeof (gtk_icon_source_get_state_wildcarded) IA__gtk_icon_source_get_state_wildcarded __attribute((visibility("hidden")));
#define gtk_icon_source_get_state_wildcarded IA__gtk_icon_source_get_state_wildcarded

extern __typeof (gtk_icon_source_get_type) IA__gtk_icon_source_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_icon_source_get_type IA__gtk_icon_source_get_type

extern __typeof (gtk_icon_source_new) IA__gtk_icon_source_new __attribute((visibility("hidden")));
#define gtk_icon_source_new IA__gtk_icon_source_new

extern __typeof (gtk_icon_source_set_direction) IA__gtk_icon_source_set_direction __attribute((visibility("hidden")));
#define gtk_icon_source_set_direction IA__gtk_icon_source_set_direction

extern __typeof (gtk_icon_source_set_direction_wildcarded) IA__gtk_icon_source_set_direction_wildcarded __attribute((visibility("hidden")));
#define gtk_icon_source_set_direction_wildcarded IA__gtk_icon_source_set_direction_wildcarded

#ifndef _WIN64
extern __typeof (gtk_icon_source_set_filename) IA__gtk_icon_source_set_filename __attribute((visibility("hidden")));
#define gtk_icon_source_set_filename IA__gtk_icon_source_set_filename

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_icon_source_set_filename_utf8) IA__gtk_icon_source_set_filename_utf8 __attribute((visibility("hidden")));
#define gtk_icon_source_set_filename_utf8 IA__gtk_icon_source_set_filename_utf8

#endif
extern __typeof (gtk_icon_source_set_icon_name) IA__gtk_icon_source_set_icon_name __attribute((visibility("hidden")));
#define gtk_icon_source_set_icon_name IA__gtk_icon_source_set_icon_name

extern __typeof (gtk_icon_source_set_pixbuf) IA__gtk_icon_source_set_pixbuf __attribute((visibility("hidden")));
#define gtk_icon_source_set_pixbuf IA__gtk_icon_source_set_pixbuf

extern __typeof (gtk_icon_source_set_size) IA__gtk_icon_source_set_size __attribute((visibility("hidden")));
#define gtk_icon_source_set_size IA__gtk_icon_source_set_size

extern __typeof (gtk_icon_source_set_size_wildcarded) IA__gtk_icon_source_set_size_wildcarded __attribute((visibility("hidden")));
#define gtk_icon_source_set_size_wildcarded IA__gtk_icon_source_set_size_wildcarded

extern __typeof (gtk_icon_source_set_state) IA__gtk_icon_source_set_state __attribute((visibility("hidden")));
#define gtk_icon_source_set_state IA__gtk_icon_source_set_state

extern __typeof (gtk_icon_source_set_state_wildcarded) IA__gtk_icon_source_set_state_wildcarded __attribute((visibility("hidden")));
#define gtk_icon_source_set_state_wildcarded IA__gtk_icon_source_set_state_wildcarded

#endif
#endif
#if IN_HEADER(__GTK_ICON_THEME_H__)
#if IN_FILE(__GTK_ICON_THEME_C__)
extern __typeof (gtk_icon_info_copy) IA__gtk_icon_info_copy __attribute((visibility("hidden")));
#define gtk_icon_info_copy IA__gtk_icon_info_copy

extern __typeof (gtk_icon_info_free) IA__gtk_icon_info_free __attribute((visibility("hidden")));
#define gtk_icon_info_free IA__gtk_icon_info_free

extern __typeof (gtk_icon_info_new_for_pixbuf) IA__gtk_icon_info_new_for_pixbuf __attribute((visibility("hidden")));
#define gtk_icon_info_new_for_pixbuf IA__gtk_icon_info_new_for_pixbuf

extern __typeof (gtk_icon_info_get_attach_points) IA__gtk_icon_info_get_attach_points __attribute((visibility("hidden")));
#define gtk_icon_info_get_attach_points IA__gtk_icon_info_get_attach_points

extern __typeof (gtk_icon_info_get_base_size) IA__gtk_icon_info_get_base_size __attribute((visibility("hidden")));
#define gtk_icon_info_get_base_size IA__gtk_icon_info_get_base_size

extern __typeof (gtk_icon_info_get_builtin_pixbuf) IA__gtk_icon_info_get_builtin_pixbuf __attribute((visibility("hidden")));
#define gtk_icon_info_get_builtin_pixbuf IA__gtk_icon_info_get_builtin_pixbuf

extern __typeof (gtk_icon_info_get_display_name) IA__gtk_icon_info_get_display_name __attribute((visibility("hidden")));
#define gtk_icon_info_get_display_name IA__gtk_icon_info_get_display_name

extern __typeof (gtk_icon_info_get_embedded_rect) IA__gtk_icon_info_get_embedded_rect __attribute((visibility("hidden")));
#define gtk_icon_info_get_embedded_rect IA__gtk_icon_info_get_embedded_rect

#ifndef _WIN64
extern __typeof (gtk_icon_info_get_filename) IA__gtk_icon_info_get_filename __attribute((visibility("hidden")));
#define gtk_icon_info_get_filename IA__gtk_icon_info_get_filename

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_icon_info_get_filename_utf8) IA__gtk_icon_info_get_filename_utf8 __attribute((visibility("hidden")));
#define gtk_icon_info_get_filename_utf8 IA__gtk_icon_info_get_filename_utf8

#endif
extern __typeof (gtk_icon_info_get_type) IA__gtk_icon_info_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_icon_info_get_type IA__gtk_icon_info_get_type

extern __typeof (gtk_icon_info_load_icon) IA__gtk_icon_info_load_icon __attribute((visibility("hidden")));
#define gtk_icon_info_load_icon IA__gtk_icon_info_load_icon

extern __typeof (gtk_icon_info_set_raw_coordinates) IA__gtk_icon_info_set_raw_coordinates __attribute((visibility("hidden")));
#define gtk_icon_info_set_raw_coordinates IA__gtk_icon_info_set_raw_coordinates

extern __typeof (gtk_icon_theme_add_builtin_icon) IA__gtk_icon_theme_add_builtin_icon __attribute((visibility("hidden")));
#define gtk_icon_theme_add_builtin_icon IA__gtk_icon_theme_add_builtin_icon

#ifndef _WIN64
extern __typeof (gtk_icon_theme_append_search_path) IA__gtk_icon_theme_append_search_path __attribute((visibility("hidden")));
#define gtk_icon_theme_append_search_path IA__gtk_icon_theme_append_search_path

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_icon_theme_append_search_path_utf8) IA__gtk_icon_theme_append_search_path_utf8 __attribute((visibility("hidden")));
#define gtk_icon_theme_append_search_path_utf8 IA__gtk_icon_theme_append_search_path_utf8

#endif
extern __typeof (gtk_icon_theme_error_quark) IA__gtk_icon_theme_error_quark __attribute((visibility("hidden")));
#define gtk_icon_theme_error_quark IA__gtk_icon_theme_error_quark

extern __typeof (gtk_icon_theme_get_default) IA__gtk_icon_theme_get_default __attribute((visibility("hidden")));
#define gtk_icon_theme_get_default IA__gtk_icon_theme_get_default

extern __typeof (gtk_icon_theme_get_example_icon_name) IA__gtk_icon_theme_get_example_icon_name __attribute((visibility("hidden")));
#define gtk_icon_theme_get_example_icon_name IA__gtk_icon_theme_get_example_icon_name

extern __typeof (gtk_icon_theme_get_for_screen) IA__gtk_icon_theme_get_for_screen __attribute((visibility("hidden")));
#define gtk_icon_theme_get_for_screen IA__gtk_icon_theme_get_for_screen

extern __typeof (gtk_icon_theme_get_icon_sizes) IA__gtk_icon_theme_get_icon_sizes __attribute((visibility("hidden")));
#define gtk_icon_theme_get_icon_sizes IA__gtk_icon_theme_get_icon_sizes

#ifndef _WIN64
extern __typeof (gtk_icon_theme_get_search_path) IA__gtk_icon_theme_get_search_path __attribute((visibility("hidden")));
#define gtk_icon_theme_get_search_path IA__gtk_icon_theme_get_search_path

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_icon_theme_get_search_path_utf8) IA__gtk_icon_theme_get_search_path_utf8 __attribute((visibility("hidden")));
#define gtk_icon_theme_get_search_path_utf8 IA__gtk_icon_theme_get_search_path_utf8

#endif
extern __typeof (gtk_icon_theme_get_type) IA__gtk_icon_theme_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_icon_theme_get_type IA__gtk_icon_theme_get_type

extern __typeof (gtk_icon_theme_has_icon) IA__gtk_icon_theme_has_icon __attribute((visibility("hidden")));
#define gtk_icon_theme_has_icon IA__gtk_icon_theme_has_icon

extern __typeof (gtk_icon_theme_list_contexts) IA__gtk_icon_theme_list_contexts __attribute((visibility("hidden")));
#define gtk_icon_theme_list_contexts IA__gtk_icon_theme_list_contexts

extern __typeof (gtk_icon_theme_list_icons) IA__gtk_icon_theme_list_icons __attribute((visibility("hidden")));
#define gtk_icon_theme_list_icons IA__gtk_icon_theme_list_icons

extern __typeof (gtk_icon_theme_load_icon) IA__gtk_icon_theme_load_icon __attribute((visibility("hidden")));
#define gtk_icon_theme_load_icon IA__gtk_icon_theme_load_icon

extern __typeof (gtk_icon_theme_lookup_icon) IA__gtk_icon_theme_lookup_icon __attribute((visibility("hidden")));
#define gtk_icon_theme_lookup_icon IA__gtk_icon_theme_lookup_icon

extern __typeof (gtk_icon_theme_lookup_by_gicon) IA__gtk_icon_theme_lookup_by_gicon __attribute((visibility("hidden")));
#define gtk_icon_theme_lookup_by_gicon IA__gtk_icon_theme_lookup_by_gicon

extern __typeof (gtk_icon_theme_choose_icon) IA__gtk_icon_theme_choose_icon __attribute((visibility("hidden")));
#define gtk_icon_theme_choose_icon IA__gtk_icon_theme_choose_icon

extern __typeof (gtk_icon_theme_new) IA__gtk_icon_theme_new __attribute((visibility("hidden")));
#define gtk_icon_theme_new IA__gtk_icon_theme_new

#ifndef _WIN64
extern __typeof (gtk_icon_theme_prepend_search_path) IA__gtk_icon_theme_prepend_search_path __attribute((visibility("hidden")));
#define gtk_icon_theme_prepend_search_path IA__gtk_icon_theme_prepend_search_path

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_icon_theme_prepend_search_path_utf8) IA__gtk_icon_theme_prepend_search_path_utf8 __attribute((visibility("hidden")));
#define gtk_icon_theme_prepend_search_path_utf8 IA__gtk_icon_theme_prepend_search_path_utf8

#endif
extern __typeof (gtk_icon_theme_rescan_if_needed) IA__gtk_icon_theme_rescan_if_needed __attribute((visibility("hidden")));
#define gtk_icon_theme_rescan_if_needed IA__gtk_icon_theme_rescan_if_needed

extern __typeof (gtk_icon_theme_set_custom_theme) IA__gtk_icon_theme_set_custom_theme __attribute((visibility("hidden")));
#define gtk_icon_theme_set_custom_theme IA__gtk_icon_theme_set_custom_theme

extern __typeof (gtk_icon_theme_set_screen) IA__gtk_icon_theme_set_screen __attribute((visibility("hidden")));
#define gtk_icon_theme_set_screen IA__gtk_icon_theme_set_screen

#ifndef _WIN64
extern __typeof (gtk_icon_theme_set_search_path) IA__gtk_icon_theme_set_search_path __attribute((visibility("hidden")));
#define gtk_icon_theme_set_search_path IA__gtk_icon_theme_set_search_path

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_icon_theme_set_search_path_utf8) IA__gtk_icon_theme_set_search_path_utf8 __attribute((visibility("hidden")));
#define gtk_icon_theme_set_search_path_utf8 IA__gtk_icon_theme_set_search_path_utf8

#endif
#endif
#endif
#if IN_HEADER(__GTK_ICON_VIEW_H__)
#if IN_FILE(__GTK_ICON_VIEW_C__)
extern __typeof (gtk_icon_view_get_column_spacing) IA__gtk_icon_view_get_column_spacing __attribute((visibility("hidden")));
#define gtk_icon_view_get_column_spacing IA__gtk_icon_view_get_column_spacing

extern __typeof (gtk_icon_view_get_columns) IA__gtk_icon_view_get_columns __attribute((visibility("hidden")));
#define gtk_icon_view_get_columns IA__gtk_icon_view_get_columns

extern __typeof (gtk_icon_view_get_cursor) IA__gtk_icon_view_get_cursor __attribute((visibility("hidden")));
#define gtk_icon_view_get_cursor IA__gtk_icon_view_get_cursor

extern __typeof (gtk_icon_view_get_item_padding) IA__gtk_icon_view_get_item_padding __attribute((visibility("hidden")));
#define gtk_icon_view_get_item_padding IA__gtk_icon_view_get_item_padding

extern __typeof (gtk_icon_view_get_item_width) IA__gtk_icon_view_get_item_width __attribute((visibility("hidden")));
#define gtk_icon_view_get_item_width IA__gtk_icon_view_get_item_width

extern __typeof (gtk_icon_view_get_margin) IA__gtk_icon_view_get_margin __attribute((visibility("hidden")));
#define gtk_icon_view_get_margin IA__gtk_icon_view_get_margin

extern __typeof (gtk_icon_view_get_markup_column) IA__gtk_icon_view_get_markup_column __attribute((visibility("hidden")));
#define gtk_icon_view_get_markup_column IA__gtk_icon_view_get_markup_column

extern __typeof (gtk_icon_view_get_model) IA__gtk_icon_view_get_model __attribute((visibility("hidden")));
#define gtk_icon_view_get_model IA__gtk_icon_view_get_model

extern __typeof (gtk_icon_view_get_orientation) IA__gtk_icon_view_get_orientation __attribute((visibility("hidden")));
#define gtk_icon_view_get_orientation IA__gtk_icon_view_get_orientation

extern __typeof (gtk_icon_view_get_item_orientation) IA__gtk_icon_view_get_item_orientation __attribute((visibility("hidden")));
#define gtk_icon_view_get_item_orientation IA__gtk_icon_view_get_item_orientation

extern __typeof (gtk_icon_view_get_path_at_pos) IA__gtk_icon_view_get_path_at_pos __attribute((visibility("hidden")));
#define gtk_icon_view_get_path_at_pos IA__gtk_icon_view_get_path_at_pos

extern __typeof (gtk_icon_view_get_item_at_pos) IA__gtk_icon_view_get_item_at_pos __attribute((visibility("hidden")));
#define gtk_icon_view_get_item_at_pos IA__gtk_icon_view_get_item_at_pos

extern __typeof (gtk_icon_view_convert_widget_to_bin_window_coords) IA__gtk_icon_view_convert_widget_to_bin_window_coords __attribute((visibility("hidden")));
#define gtk_icon_view_convert_widget_to_bin_window_coords IA__gtk_icon_view_convert_widget_to_bin_window_coords

extern __typeof (gtk_icon_view_get_pixbuf_column) IA__gtk_icon_view_get_pixbuf_column __attribute((visibility("hidden")));
#define gtk_icon_view_get_pixbuf_column IA__gtk_icon_view_get_pixbuf_column

extern __typeof (gtk_icon_view_get_row_spacing) IA__gtk_icon_view_get_row_spacing __attribute((visibility("hidden")));
#define gtk_icon_view_get_row_spacing IA__gtk_icon_view_get_row_spacing

extern __typeof (gtk_icon_view_get_selected_items) IA__gtk_icon_view_get_selected_items __attribute((visibility("hidden")));
#define gtk_icon_view_get_selected_items IA__gtk_icon_view_get_selected_items

extern __typeof (gtk_icon_view_get_selection_mode) IA__gtk_icon_view_get_selection_mode __attribute((visibility("hidden")));
#define gtk_icon_view_get_selection_mode IA__gtk_icon_view_get_selection_mode

extern __typeof (gtk_icon_view_get_spacing) IA__gtk_icon_view_get_spacing __attribute((visibility("hidden")));
#define gtk_icon_view_get_spacing IA__gtk_icon_view_get_spacing

extern __typeof (gtk_icon_view_get_text_column) IA__gtk_icon_view_get_text_column __attribute((visibility("hidden")));
#define gtk_icon_view_get_text_column IA__gtk_icon_view_get_text_column

extern __typeof (gtk_icon_view_get_type) IA__gtk_icon_view_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_icon_view_get_type IA__gtk_icon_view_get_type

extern __typeof (gtk_icon_view_get_visible_range) IA__gtk_icon_view_get_visible_range __attribute((visibility("hidden")));
#define gtk_icon_view_get_visible_range IA__gtk_icon_view_get_visible_range

extern __typeof (gtk_icon_view_item_activated) IA__gtk_icon_view_item_activated __attribute((visibility("hidden")));
#define gtk_icon_view_item_activated IA__gtk_icon_view_item_activated

extern __typeof (gtk_icon_view_new) IA__gtk_icon_view_new __attribute((visibility("hidden")));
#define gtk_icon_view_new IA__gtk_icon_view_new

extern __typeof (gtk_icon_view_new_with_model) IA__gtk_icon_view_new_with_model __attribute((visibility("hidden")));
#define gtk_icon_view_new_with_model IA__gtk_icon_view_new_with_model

extern __typeof (gtk_icon_view_path_is_selected) IA__gtk_icon_view_path_is_selected __attribute((visibility("hidden")));
#define gtk_icon_view_path_is_selected IA__gtk_icon_view_path_is_selected

extern __typeof (gtk_icon_view_scroll_to_path) IA__gtk_icon_view_scroll_to_path __attribute((visibility("hidden")));
#define gtk_icon_view_scroll_to_path IA__gtk_icon_view_scroll_to_path

extern __typeof (gtk_icon_view_select_all) IA__gtk_icon_view_select_all __attribute((visibility("hidden")));
#define gtk_icon_view_select_all IA__gtk_icon_view_select_all

extern __typeof (gtk_icon_view_selected_foreach) IA__gtk_icon_view_selected_foreach __attribute((visibility("hidden")));
#define gtk_icon_view_selected_foreach IA__gtk_icon_view_selected_foreach

extern __typeof (gtk_icon_view_select_path) IA__gtk_icon_view_select_path __attribute((visibility("hidden")));
#define gtk_icon_view_select_path IA__gtk_icon_view_select_path

extern __typeof (gtk_icon_view_set_column_spacing) IA__gtk_icon_view_set_column_spacing __attribute((visibility("hidden")));
#define gtk_icon_view_set_column_spacing IA__gtk_icon_view_set_column_spacing

extern __typeof (gtk_icon_view_set_columns) IA__gtk_icon_view_set_columns __attribute((visibility("hidden")));
#define gtk_icon_view_set_columns IA__gtk_icon_view_set_columns

extern __typeof (gtk_icon_view_set_cursor) IA__gtk_icon_view_set_cursor __attribute((visibility("hidden")));
#define gtk_icon_view_set_cursor IA__gtk_icon_view_set_cursor

extern __typeof (gtk_icon_view_set_item_padding) IA__gtk_icon_view_set_item_padding __attribute((visibility("hidden")));
#define gtk_icon_view_set_item_padding IA__gtk_icon_view_set_item_padding

extern __typeof (gtk_icon_view_set_item_width) IA__gtk_icon_view_set_item_width __attribute((visibility("hidden")));
#define gtk_icon_view_set_item_width IA__gtk_icon_view_set_item_width

extern __typeof (gtk_icon_view_set_margin) IA__gtk_icon_view_set_margin __attribute((visibility("hidden")));
#define gtk_icon_view_set_margin IA__gtk_icon_view_set_margin

extern __typeof (gtk_icon_view_set_markup_column) IA__gtk_icon_view_set_markup_column __attribute((visibility("hidden")));
#define gtk_icon_view_set_markup_column IA__gtk_icon_view_set_markup_column

extern __typeof (gtk_icon_view_set_model) IA__gtk_icon_view_set_model __attribute((visibility("hidden")));
#define gtk_icon_view_set_model IA__gtk_icon_view_set_model

extern __typeof (gtk_icon_view_set_orientation) IA__gtk_icon_view_set_orientation __attribute((visibility("hidden")));
#define gtk_icon_view_set_orientation IA__gtk_icon_view_set_orientation

extern __typeof (gtk_icon_view_set_item_orientation) IA__gtk_icon_view_set_item_orientation __attribute((visibility("hidden")));
#define gtk_icon_view_set_item_orientation IA__gtk_icon_view_set_item_orientation

extern __typeof (gtk_icon_view_set_pixbuf_column) IA__gtk_icon_view_set_pixbuf_column __attribute((visibility("hidden")));
#define gtk_icon_view_set_pixbuf_column IA__gtk_icon_view_set_pixbuf_column

extern __typeof (gtk_icon_view_set_row_spacing) IA__gtk_icon_view_set_row_spacing __attribute((visibility("hidden")));
#define gtk_icon_view_set_row_spacing IA__gtk_icon_view_set_row_spacing

extern __typeof (gtk_icon_view_set_selection_mode) IA__gtk_icon_view_set_selection_mode __attribute((visibility("hidden")));
#define gtk_icon_view_set_selection_mode IA__gtk_icon_view_set_selection_mode

extern __typeof (gtk_icon_view_set_spacing) IA__gtk_icon_view_set_spacing __attribute((visibility("hidden")));
#define gtk_icon_view_set_spacing IA__gtk_icon_view_set_spacing

extern __typeof (gtk_icon_view_set_text_column) IA__gtk_icon_view_set_text_column __attribute((visibility("hidden")));
#define gtk_icon_view_set_text_column IA__gtk_icon_view_set_text_column

extern __typeof (gtk_icon_view_unselect_all) IA__gtk_icon_view_unselect_all __attribute((visibility("hidden")));
#define gtk_icon_view_unselect_all IA__gtk_icon_view_unselect_all

extern __typeof (gtk_icon_view_unselect_path) IA__gtk_icon_view_unselect_path __attribute((visibility("hidden")));
#define gtk_icon_view_unselect_path IA__gtk_icon_view_unselect_path

extern __typeof (gtk_icon_view_enable_model_drag_source) IA__gtk_icon_view_enable_model_drag_source __attribute((visibility("hidden")));
#define gtk_icon_view_enable_model_drag_source IA__gtk_icon_view_enable_model_drag_source

extern __typeof (gtk_icon_view_enable_model_drag_dest) IA__gtk_icon_view_enable_model_drag_dest __attribute((visibility("hidden")));
#define gtk_icon_view_enable_model_drag_dest IA__gtk_icon_view_enable_model_drag_dest

extern __typeof (gtk_icon_view_unset_model_drag_source) IA__gtk_icon_view_unset_model_drag_source __attribute((visibility("hidden")));
#define gtk_icon_view_unset_model_drag_source IA__gtk_icon_view_unset_model_drag_source

extern __typeof (gtk_icon_view_unset_model_drag_dest) IA__gtk_icon_view_unset_model_drag_dest __attribute((visibility("hidden")));
#define gtk_icon_view_unset_model_drag_dest IA__gtk_icon_view_unset_model_drag_dest

extern __typeof (gtk_icon_view_set_reorderable) IA__gtk_icon_view_set_reorderable __attribute((visibility("hidden")));
#define gtk_icon_view_set_reorderable IA__gtk_icon_view_set_reorderable

extern __typeof (gtk_icon_view_get_reorderable) IA__gtk_icon_view_get_reorderable __attribute((visibility("hidden")));
#define gtk_icon_view_get_reorderable IA__gtk_icon_view_get_reorderable

extern __typeof (gtk_icon_view_set_drag_dest_item) IA__gtk_icon_view_set_drag_dest_item __attribute((visibility("hidden")));
#define gtk_icon_view_set_drag_dest_item IA__gtk_icon_view_set_drag_dest_item

extern __typeof (gtk_icon_view_get_drag_dest_item) IA__gtk_icon_view_get_drag_dest_item __attribute((visibility("hidden")));
#define gtk_icon_view_get_drag_dest_item IA__gtk_icon_view_get_drag_dest_item

extern __typeof (gtk_icon_view_get_dest_item_at_pos) IA__gtk_icon_view_get_dest_item_at_pos __attribute((visibility("hidden")));
#define gtk_icon_view_get_dest_item_at_pos IA__gtk_icon_view_get_dest_item_at_pos

extern __typeof (gtk_icon_view_create_drag_icon) IA__gtk_icon_view_create_drag_icon __attribute((visibility("hidden")));
#define gtk_icon_view_create_drag_icon IA__gtk_icon_view_create_drag_icon

extern __typeof (gtk_icon_view_set_tooltip_item) IA__gtk_icon_view_set_tooltip_item __attribute((visibility("hidden")));
#define gtk_icon_view_set_tooltip_item IA__gtk_icon_view_set_tooltip_item

extern __typeof (gtk_icon_view_set_tooltip_cell) IA__gtk_icon_view_set_tooltip_cell __attribute((visibility("hidden")));
#define gtk_icon_view_set_tooltip_cell IA__gtk_icon_view_set_tooltip_cell

extern __typeof (gtk_icon_view_get_tooltip_context) IA__gtk_icon_view_get_tooltip_context __attribute((visibility("hidden")));
#define gtk_icon_view_get_tooltip_context IA__gtk_icon_view_get_tooltip_context

extern __typeof (gtk_icon_view_set_tooltip_column) IA__gtk_icon_view_set_tooltip_column __attribute((visibility("hidden")));
#define gtk_icon_view_set_tooltip_column IA__gtk_icon_view_set_tooltip_column

extern __typeof (gtk_icon_view_get_tooltip_column) IA__gtk_icon_view_get_tooltip_column __attribute((visibility("hidden")));
#define gtk_icon_view_get_tooltip_column IA__gtk_icon_view_get_tooltip_column

extern __typeof (gtk_icon_view_get_item_row) IA__gtk_icon_view_get_item_row __attribute((visibility("hidden")));
#define gtk_icon_view_get_item_row IA__gtk_icon_view_get_item_row

extern __typeof (gtk_icon_view_get_item_column) IA__gtk_icon_view_get_item_column __attribute((visibility("hidden")));
#define gtk_icon_view_get_item_column IA__gtk_icon_view_get_item_column

#endif
#endif
#if IN_HEADER(__GTK_IMAGE_H__)
#if IN_FILE(__GTK_IMAGE_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_image_get) IA__gtk_image_get __attribute((visibility("hidden")));
#define gtk_image_get IA__gtk_image_get

extern __typeof (gtk_image_set) IA__gtk_image_set __attribute((visibility("hidden")));
#define gtk_image_set IA__gtk_image_set

#endif
extern __typeof (gtk_image_clear) IA__gtk_image_clear __attribute((visibility("hidden")));
#define gtk_image_clear IA__gtk_image_clear

extern __typeof (gtk_image_get_animation) IA__gtk_image_get_animation __attribute((visibility("hidden")));
#define gtk_image_get_animation IA__gtk_image_get_animation

extern __typeof (gtk_image_get_icon_name) IA__gtk_image_get_icon_name __attribute((visibility("hidden")));
#define gtk_image_get_icon_name IA__gtk_image_get_icon_name

extern __typeof (gtk_image_get_icon_set) IA__gtk_image_get_icon_set __attribute((visibility("hidden")));
#define gtk_image_get_icon_set IA__gtk_image_get_icon_set

extern __typeof (gtk_image_get_image) IA__gtk_image_get_image __attribute((visibility("hidden")));
#define gtk_image_get_image IA__gtk_image_get_image

extern __typeof (gtk_image_get_pixbuf) IA__gtk_image_get_pixbuf __attribute((visibility("hidden")));
#define gtk_image_get_pixbuf IA__gtk_image_get_pixbuf

extern __typeof (gtk_image_get_pixel_size) IA__gtk_image_get_pixel_size __attribute((visibility("hidden")));
#define gtk_image_get_pixel_size IA__gtk_image_get_pixel_size

extern __typeof (gtk_image_get_pixmap) IA__gtk_image_get_pixmap __attribute((visibility("hidden")));
#define gtk_image_get_pixmap IA__gtk_image_get_pixmap

extern __typeof (gtk_image_get_stock) IA__gtk_image_get_stock __attribute((visibility("hidden")));
#define gtk_image_get_stock IA__gtk_image_get_stock

extern __typeof (gtk_image_get_gicon) IA__gtk_image_get_gicon __attribute((visibility("hidden")));
#define gtk_image_get_gicon IA__gtk_image_get_gicon

extern __typeof (gtk_image_get_storage_type) IA__gtk_image_get_storage_type __attribute((visibility("hidden")));
#define gtk_image_get_storage_type IA__gtk_image_get_storage_type

extern __typeof (gtk_image_get_type) IA__gtk_image_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_image_get_type IA__gtk_image_get_type

extern __typeof (gtk_image_new) IA__gtk_image_new __attribute((visibility("hidden")));
#define gtk_image_new IA__gtk_image_new

extern __typeof (gtk_image_new_from_animation) IA__gtk_image_new_from_animation __attribute((visibility("hidden")));
#define gtk_image_new_from_animation IA__gtk_image_new_from_animation

#ifndef _WIN64
extern __typeof (gtk_image_new_from_file) IA__gtk_image_new_from_file __attribute((visibility("hidden")));
#define gtk_image_new_from_file IA__gtk_image_new_from_file

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_image_new_from_file_utf8) IA__gtk_image_new_from_file_utf8 __attribute((visibility("hidden")));
#define gtk_image_new_from_file_utf8 IA__gtk_image_new_from_file_utf8

#endif
extern __typeof (gtk_image_new_from_icon_name) IA__gtk_image_new_from_icon_name __attribute((visibility("hidden")));
#define gtk_image_new_from_icon_name IA__gtk_image_new_from_icon_name

extern __typeof (gtk_image_new_from_icon_set) IA__gtk_image_new_from_icon_set __attribute((visibility("hidden")));
#define gtk_image_new_from_icon_set IA__gtk_image_new_from_icon_set

extern __typeof (gtk_image_new_from_image) IA__gtk_image_new_from_image __attribute((visibility("hidden")));
#define gtk_image_new_from_image IA__gtk_image_new_from_image

extern __typeof (gtk_image_new_from_pixbuf) IA__gtk_image_new_from_pixbuf __attribute((visibility("hidden")));
#define gtk_image_new_from_pixbuf IA__gtk_image_new_from_pixbuf

extern __typeof (gtk_image_new_from_pixmap) IA__gtk_image_new_from_pixmap __attribute((visibility("hidden")));
#define gtk_image_new_from_pixmap IA__gtk_image_new_from_pixmap

extern __typeof (gtk_image_new_from_stock) IA__gtk_image_new_from_stock __attribute((visibility("hidden")));
#define gtk_image_new_from_stock IA__gtk_image_new_from_stock

extern __typeof (gtk_image_new_from_gicon) IA__gtk_image_new_from_gicon __attribute((visibility("hidden")));
#define gtk_image_new_from_gicon IA__gtk_image_new_from_gicon

extern __typeof (gtk_image_set_from_animation) IA__gtk_image_set_from_animation __attribute((visibility("hidden")));
#define gtk_image_set_from_animation IA__gtk_image_set_from_animation

#ifndef _WIN64
extern __typeof (gtk_image_set_from_file) IA__gtk_image_set_from_file __attribute((visibility("hidden")));
#define gtk_image_set_from_file IA__gtk_image_set_from_file

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_image_set_from_file_utf8) IA__gtk_image_set_from_file_utf8 __attribute((visibility("hidden")));
#define gtk_image_set_from_file_utf8 IA__gtk_image_set_from_file_utf8

#endif
extern __typeof (gtk_image_set_from_icon_name) IA__gtk_image_set_from_icon_name __attribute((visibility("hidden")));
#define gtk_image_set_from_icon_name IA__gtk_image_set_from_icon_name

extern __typeof (gtk_image_set_from_icon_set) IA__gtk_image_set_from_icon_set __attribute((visibility("hidden")));
#define gtk_image_set_from_icon_set IA__gtk_image_set_from_icon_set

extern __typeof (gtk_image_set_from_image) IA__gtk_image_set_from_image __attribute((visibility("hidden")));
#define gtk_image_set_from_image IA__gtk_image_set_from_image

extern __typeof (gtk_image_set_from_pixbuf) IA__gtk_image_set_from_pixbuf __attribute((visibility("hidden")));
#define gtk_image_set_from_pixbuf IA__gtk_image_set_from_pixbuf

extern __typeof (gtk_image_set_from_pixmap) IA__gtk_image_set_from_pixmap __attribute((visibility("hidden")));
#define gtk_image_set_from_pixmap IA__gtk_image_set_from_pixmap

extern __typeof (gtk_image_set_from_stock) IA__gtk_image_set_from_stock __attribute((visibility("hidden")));
#define gtk_image_set_from_stock IA__gtk_image_set_from_stock

extern __typeof (gtk_image_set_from_gicon) IA__gtk_image_set_from_gicon __attribute((visibility("hidden")));
#define gtk_image_set_from_gicon IA__gtk_image_set_from_gicon

extern __typeof (gtk_image_set_pixel_size) IA__gtk_image_set_pixel_size __attribute((visibility("hidden")));
#define gtk_image_set_pixel_size IA__gtk_image_set_pixel_size

#endif
#endif
#if IN_HEADER(__GTK_IMAGE_MENU_ITEM_H__)
#if IN_FILE(__GTK_IMAGE_MENU_ITEM_C__)
extern __typeof (gtk_image_menu_item_get_always_show_image) IA__gtk_image_menu_item_get_always_show_image __attribute((visibility("hidden")));
#define gtk_image_menu_item_get_always_show_image IA__gtk_image_menu_item_get_always_show_image

extern __typeof (gtk_image_menu_item_get_image) IA__gtk_image_menu_item_get_image __attribute((visibility("hidden")));
#define gtk_image_menu_item_get_image IA__gtk_image_menu_item_get_image

extern __typeof (gtk_image_menu_item_get_type) IA__gtk_image_menu_item_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_image_menu_item_get_type IA__gtk_image_menu_item_get_type

extern __typeof (gtk_image_menu_item_get_use_stock) IA__gtk_image_menu_item_get_use_stock __attribute((visibility("hidden")));
#define gtk_image_menu_item_get_use_stock IA__gtk_image_menu_item_get_use_stock

extern __typeof (gtk_image_menu_item_new) IA__gtk_image_menu_item_new __attribute((visibility("hidden")));
#define gtk_image_menu_item_new IA__gtk_image_menu_item_new

extern __typeof (gtk_image_menu_item_new_from_stock) IA__gtk_image_menu_item_new_from_stock __attribute((visibility("hidden")));
#define gtk_image_menu_item_new_from_stock IA__gtk_image_menu_item_new_from_stock

extern __typeof (gtk_image_menu_item_new_with_label) IA__gtk_image_menu_item_new_with_label __attribute((visibility("hidden")));
#define gtk_image_menu_item_new_with_label IA__gtk_image_menu_item_new_with_label

extern __typeof (gtk_image_menu_item_new_with_mnemonic) IA__gtk_image_menu_item_new_with_mnemonic __attribute((visibility("hidden")));
#define gtk_image_menu_item_new_with_mnemonic IA__gtk_image_menu_item_new_with_mnemonic

extern __typeof (gtk_image_menu_item_set_accel_group) IA__gtk_image_menu_item_set_accel_group __attribute((visibility("hidden")));
#define gtk_image_menu_item_set_accel_group IA__gtk_image_menu_item_set_accel_group

extern __typeof (gtk_image_menu_item_set_always_show_image) IA__gtk_image_menu_item_set_always_show_image __attribute((visibility("hidden")));
#define gtk_image_menu_item_set_always_show_image IA__gtk_image_menu_item_set_always_show_image

extern __typeof (gtk_image_menu_item_set_image) IA__gtk_image_menu_item_set_image __attribute((visibility("hidden")));
#define gtk_image_menu_item_set_image IA__gtk_image_menu_item_set_image

extern __typeof (gtk_image_menu_item_set_use_stock) IA__gtk_image_menu_item_set_use_stock __attribute((visibility("hidden")));
#define gtk_image_menu_item_set_use_stock IA__gtk_image_menu_item_set_use_stock

#endif
#endif
#if IN_HEADER(__GTK_IM_CONTEXT_H__)
#if IN_FILE(__GTK_IM_CONTEXT_C__)
extern __typeof (gtk_im_context_delete_surrounding) IA__gtk_im_context_delete_surrounding __attribute((visibility("hidden")));
#define gtk_im_context_delete_surrounding IA__gtk_im_context_delete_surrounding

extern __typeof (gtk_im_context_filter_keypress) IA__gtk_im_context_filter_keypress __attribute((visibility("hidden")));
#define gtk_im_context_filter_keypress IA__gtk_im_context_filter_keypress

extern __typeof (gtk_im_context_focus_in) IA__gtk_im_context_focus_in __attribute((visibility("hidden")));
#define gtk_im_context_focus_in IA__gtk_im_context_focus_in

extern __typeof (gtk_im_context_focus_out) IA__gtk_im_context_focus_out __attribute((visibility("hidden")));
#define gtk_im_context_focus_out IA__gtk_im_context_focus_out

extern __typeof (gtk_im_context_get_preedit_string) IA__gtk_im_context_get_preedit_string __attribute((visibility("hidden")));
#define gtk_im_context_get_preedit_string IA__gtk_im_context_get_preedit_string

extern __typeof (gtk_im_context_get_surrounding) IA__gtk_im_context_get_surrounding __attribute((visibility("hidden")));
#define gtk_im_context_get_surrounding IA__gtk_im_context_get_surrounding

extern __typeof (gtk_im_context_get_type) IA__gtk_im_context_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_im_context_get_type IA__gtk_im_context_get_type

extern __typeof (gtk_im_context_reset) IA__gtk_im_context_reset __attribute((visibility("hidden")));
#define gtk_im_context_reset IA__gtk_im_context_reset

extern __typeof (gtk_im_context_set_client_window) IA__gtk_im_context_set_client_window __attribute((visibility("hidden")));
#define gtk_im_context_set_client_window IA__gtk_im_context_set_client_window

extern __typeof (gtk_im_context_set_cursor_location) IA__gtk_im_context_set_cursor_location __attribute((visibility("hidden")));
#define gtk_im_context_set_cursor_location IA__gtk_im_context_set_cursor_location

extern __typeof (gtk_im_context_set_surrounding) IA__gtk_im_context_set_surrounding __attribute((visibility("hidden")));
#define gtk_im_context_set_surrounding IA__gtk_im_context_set_surrounding

extern __typeof (gtk_im_context_set_use_preedit) IA__gtk_im_context_set_use_preedit __attribute((visibility("hidden")));
#define gtk_im_context_set_use_preedit IA__gtk_im_context_set_use_preedit

#endif
#endif
#if IN_HEADER(__GTK_IM_CONTEXT_SIMPLE_H__)
#if IN_FILE(__GTK_IM_CONTEXT_SIMPLE_C__)
extern __typeof (gtk_im_context_simple_add_table) IA__gtk_im_context_simple_add_table __attribute((visibility("hidden")));
#define gtk_im_context_simple_add_table IA__gtk_im_context_simple_add_table

extern __typeof (gtk_im_context_simple_get_type) IA__gtk_im_context_simple_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_im_context_simple_get_type IA__gtk_im_context_simple_get_type

extern __typeof (gtk_im_context_simple_new) IA__gtk_im_context_simple_new __attribute((visibility("hidden")));
#define gtk_im_context_simple_new IA__gtk_im_context_simple_new

#endif
#endif
#if IN_HEADER(__GTK_IM_MULTICONTEXT_H__)
#if IN_FILE(__GTK_IM_MULTICONTEXT_C__)
extern __typeof (gtk_im_multicontext_append_menuitems) IA__gtk_im_multicontext_append_menuitems __attribute((visibility("hidden")));
#define gtk_im_multicontext_append_menuitems IA__gtk_im_multicontext_append_menuitems

extern __typeof (gtk_im_multicontext_get_context_id) IA__gtk_im_multicontext_get_context_id __attribute((visibility("hidden")));
#define gtk_im_multicontext_get_context_id IA__gtk_im_multicontext_get_context_id

extern __typeof (gtk_im_multicontext_set_context_id) IA__gtk_im_multicontext_set_context_id __attribute((visibility("hidden")));
#define gtk_im_multicontext_set_context_id IA__gtk_im_multicontext_set_context_id

extern __typeof (gtk_im_multicontext_get_type) IA__gtk_im_multicontext_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_im_multicontext_get_type IA__gtk_im_multicontext_get_type

extern __typeof (gtk_im_multicontext_new) IA__gtk_im_multicontext_new __attribute((visibility("hidden")));
#define gtk_im_multicontext_new IA__gtk_im_multicontext_new

#endif
#endif
#if IN_HEADER(__GTK_INVISIBLE_H__)
#if IN_FILE(__GTK_INVISIBLE_C__)
extern __typeof (gtk_invisible_get_screen) IA__gtk_invisible_get_screen __attribute((visibility("hidden")));
#define gtk_invisible_get_screen IA__gtk_invisible_get_screen

extern __typeof (gtk_invisible_get_type) IA__gtk_invisible_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_invisible_get_type IA__gtk_invisible_get_type

extern __typeof (gtk_invisible_new) IA__gtk_invisible_new __attribute((visibility("hidden")));
#define gtk_invisible_new IA__gtk_invisible_new

extern __typeof (gtk_invisible_new_for_screen) IA__gtk_invisible_new_for_screen __attribute((visibility("hidden")));
#define gtk_invisible_new_for_screen IA__gtk_invisible_new_for_screen

extern __typeof (gtk_invisible_set_screen) IA__gtk_invisible_set_screen __attribute((visibility("hidden")));
#define gtk_invisible_set_screen IA__gtk_invisible_set_screen

#endif
#endif
#if IN_HEADER(__GTK_ITEM_H__)
#if IN_FILE(__GTK_ITEM_C__)
extern __typeof (gtk_item_get_type) IA__gtk_item_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_item_get_type IA__gtk_item_get_type

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_item_deselect) IA__gtk_item_deselect __attribute((visibility("hidden")));
#define gtk_item_deselect IA__gtk_item_deselect

extern __typeof (gtk_item_select) IA__gtk_item_select __attribute((visibility("hidden")));
#define gtk_item_select IA__gtk_item_select

extern __typeof (gtk_item_toggle) IA__gtk_item_toggle __attribute((visibility("hidden")));
#define gtk_item_toggle IA__gtk_item_toggle

#endif
#endif
#endif
#if IN_HEADER(__GTK_ITEM_FACTORY_H__)
#if IN_FILE(__GTK_ITEM_FACTORY_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_item_factories_path_delete) IA__gtk_item_factories_path_delete __attribute((visibility("hidden")));
#define gtk_item_factories_path_delete IA__gtk_item_factories_path_delete

extern __typeof (gtk_item_factory_add_foreign) IA__gtk_item_factory_add_foreign __attribute((visibility("hidden")));
#define gtk_item_factory_add_foreign IA__gtk_item_factory_add_foreign

extern __typeof (gtk_item_factory_construct) IA__gtk_item_factory_construct __attribute((visibility("hidden")));
#define gtk_item_factory_construct IA__gtk_item_factory_construct

extern __typeof (gtk_item_factory_create_item) IA__gtk_item_factory_create_item __attribute((visibility("hidden")));
#define gtk_item_factory_create_item IA__gtk_item_factory_create_item

extern __typeof (gtk_item_factory_create_items) IA__gtk_item_factory_create_items __attribute((visibility("hidden")));
#define gtk_item_factory_create_items IA__gtk_item_factory_create_items

extern __typeof (gtk_item_factory_create_items_ac) IA__gtk_item_factory_create_items_ac __attribute((visibility("hidden")));
#define gtk_item_factory_create_items_ac IA__gtk_item_factory_create_items_ac

extern __typeof (gtk_item_factory_create_menu_entries) IA__gtk_item_factory_create_menu_entries __attribute((visibility("hidden")));
#define gtk_item_factory_create_menu_entries IA__gtk_item_factory_create_menu_entries

extern __typeof (gtk_item_factory_delete_entries) IA__gtk_item_factory_delete_entries __attribute((visibility("hidden")));
#define gtk_item_factory_delete_entries IA__gtk_item_factory_delete_entries

extern __typeof (gtk_item_factory_delete_entry) IA__gtk_item_factory_delete_entry __attribute((visibility("hidden")));
#define gtk_item_factory_delete_entry IA__gtk_item_factory_delete_entry

extern __typeof (gtk_item_factory_delete_item) IA__gtk_item_factory_delete_item __attribute((visibility("hidden")));
#define gtk_item_factory_delete_item IA__gtk_item_factory_delete_item

extern __typeof (gtk_item_factory_from_path) IA__gtk_item_factory_from_path __attribute((visibility("hidden")));
#define gtk_item_factory_from_path IA__gtk_item_factory_from_path

extern __typeof (gtk_item_factory_from_widget) IA__gtk_item_factory_from_widget __attribute((visibility("hidden")));
#define gtk_item_factory_from_widget IA__gtk_item_factory_from_widget

extern __typeof (gtk_item_factory_get_item) IA__gtk_item_factory_get_item __attribute((visibility("hidden")));
#define gtk_item_factory_get_item IA__gtk_item_factory_get_item

extern __typeof (gtk_item_factory_get_item_by_action) IA__gtk_item_factory_get_item_by_action __attribute((visibility("hidden")));
#define gtk_item_factory_get_item_by_action IA__gtk_item_factory_get_item_by_action

extern __typeof (gtk_item_factory_get_type) IA__gtk_item_factory_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_item_factory_get_type IA__gtk_item_factory_get_type

extern __typeof (gtk_item_factory_get_widget) IA__gtk_item_factory_get_widget __attribute((visibility("hidden")));
#define gtk_item_factory_get_widget IA__gtk_item_factory_get_widget

extern __typeof (gtk_item_factory_get_widget_by_action) IA__gtk_item_factory_get_widget_by_action __attribute((visibility("hidden")));
#define gtk_item_factory_get_widget_by_action IA__gtk_item_factory_get_widget_by_action

extern __typeof (gtk_item_factory_new) IA__gtk_item_factory_new __attribute((visibility("hidden")));
#define gtk_item_factory_new IA__gtk_item_factory_new

extern __typeof (gtk_item_factory_path_from_widget) IA__gtk_item_factory_path_from_widget __attribute((visibility("hidden")));
#define gtk_item_factory_path_from_widget IA__gtk_item_factory_path_from_widget

extern __typeof (gtk_item_factory_popup) IA__gtk_item_factory_popup __attribute((visibility("hidden")));
#define gtk_item_factory_popup IA__gtk_item_factory_popup

extern __typeof (gtk_item_factory_popup_data) IA__gtk_item_factory_popup_data __attribute((visibility("hidden")));
#define gtk_item_factory_popup_data IA__gtk_item_factory_popup_data

extern __typeof (gtk_item_factory_popup_data_from_widget) IA__gtk_item_factory_popup_data_from_widget __attribute((visibility("hidden")));
#define gtk_item_factory_popup_data_from_widget IA__gtk_item_factory_popup_data_from_widget

extern __typeof (gtk_item_factory_popup_with_data) IA__gtk_item_factory_popup_with_data __attribute((visibility("hidden")));
#define gtk_item_factory_popup_with_data IA__gtk_item_factory_popup_with_data

extern __typeof (gtk_item_factory_set_translate_func) IA__gtk_item_factory_set_translate_func __attribute((visibility("hidden")));
#define gtk_item_factory_set_translate_func IA__gtk_item_factory_set_translate_func

#endif
#endif
#endif
#if IN_HEADER(__GTK_LABEL_H__)
#if IN_FILE(__GTK_LABEL_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_label_get) IA__gtk_label_get __attribute((visibility("hidden")));
#define gtk_label_get IA__gtk_label_get

extern __typeof (gtk_label_parse_uline) IA__gtk_label_parse_uline __attribute((visibility("hidden")));
#define gtk_label_parse_uline IA__gtk_label_parse_uline

#endif
extern __typeof (gtk_label_get_angle) IA__gtk_label_get_angle __attribute((visibility("hidden")));
#define gtk_label_get_angle IA__gtk_label_get_angle

extern __typeof (gtk_label_get_attributes) IA__gtk_label_get_attributes __attribute((visibility("hidden")));
#define gtk_label_get_attributes IA__gtk_label_get_attributes

extern __typeof (gtk_label_get_ellipsize) IA__gtk_label_get_ellipsize __attribute((visibility("hidden")));
#define gtk_label_get_ellipsize IA__gtk_label_get_ellipsize

extern __typeof (gtk_label_get_justify) IA__gtk_label_get_justify __attribute((visibility("hidden")));
#define gtk_label_get_justify IA__gtk_label_get_justify

extern __typeof (gtk_label_get_label) IA__gtk_label_get_label __attribute((visibility("hidden")));
#define gtk_label_get_label IA__gtk_label_get_label

extern __typeof (gtk_label_get_layout) IA__gtk_label_get_layout __attribute((visibility("hidden")));
#define gtk_label_get_layout IA__gtk_label_get_layout

extern __typeof (gtk_label_get_layout_offsets) IA__gtk_label_get_layout_offsets __attribute((visibility("hidden")));
#define gtk_label_get_layout_offsets IA__gtk_label_get_layout_offsets

extern __typeof (gtk_label_get_line_wrap) IA__gtk_label_get_line_wrap __attribute((visibility("hidden")));
#define gtk_label_get_line_wrap IA__gtk_label_get_line_wrap

extern __typeof (gtk_label_get_line_wrap_mode) IA__gtk_label_get_line_wrap_mode __attribute((visibility("hidden")));
#define gtk_label_get_line_wrap_mode IA__gtk_label_get_line_wrap_mode

extern __typeof (gtk_label_get_max_width_chars) IA__gtk_label_get_max_width_chars __attribute((visibility("hidden")));
#define gtk_label_get_max_width_chars IA__gtk_label_get_max_width_chars

extern __typeof (gtk_label_get_mnemonic_keyval) IA__gtk_label_get_mnemonic_keyval __attribute((visibility("hidden")));
#define gtk_label_get_mnemonic_keyval IA__gtk_label_get_mnemonic_keyval

extern __typeof (gtk_label_get_mnemonic_widget) IA__gtk_label_get_mnemonic_widget __attribute((visibility("hidden")));
#define gtk_label_get_mnemonic_widget IA__gtk_label_get_mnemonic_widget

extern __typeof (gtk_label_get_selectable) IA__gtk_label_get_selectable __attribute((visibility("hidden")));
#define gtk_label_get_selectable IA__gtk_label_get_selectable

extern __typeof (gtk_label_get_selection_bounds) IA__gtk_label_get_selection_bounds __attribute((visibility("hidden")));
#define gtk_label_get_selection_bounds IA__gtk_label_get_selection_bounds

extern __typeof (gtk_label_get_single_line_mode) IA__gtk_label_get_single_line_mode __attribute((visibility("hidden")));
#define gtk_label_get_single_line_mode IA__gtk_label_get_single_line_mode

extern __typeof (gtk_label_get_text) IA__gtk_label_get_text __attribute((visibility("hidden")));
#define gtk_label_get_text IA__gtk_label_get_text

extern __typeof (gtk_label_get_type) IA__gtk_label_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_label_get_type IA__gtk_label_get_type

extern __typeof (gtk_label_get_use_markup) IA__gtk_label_get_use_markup __attribute((visibility("hidden")));
#define gtk_label_get_use_markup IA__gtk_label_get_use_markup

extern __typeof (gtk_label_get_use_underline) IA__gtk_label_get_use_underline __attribute((visibility("hidden")));
#define gtk_label_get_use_underline IA__gtk_label_get_use_underline

extern __typeof (gtk_label_get_width_chars) IA__gtk_label_get_width_chars __attribute((visibility("hidden")));
#define gtk_label_get_width_chars IA__gtk_label_get_width_chars

extern __typeof (gtk_label_new) IA__gtk_label_new __attribute((visibility("hidden")));
#define gtk_label_new IA__gtk_label_new

extern __typeof (gtk_label_new_with_mnemonic) IA__gtk_label_new_with_mnemonic __attribute((visibility("hidden")));
#define gtk_label_new_with_mnemonic IA__gtk_label_new_with_mnemonic

extern __typeof (gtk_label_select_region) IA__gtk_label_select_region __attribute((visibility("hidden")));
#define gtk_label_select_region IA__gtk_label_select_region

extern __typeof (gtk_label_set_angle) IA__gtk_label_set_angle __attribute((visibility("hidden")));
#define gtk_label_set_angle IA__gtk_label_set_angle

extern __typeof (gtk_label_set_attributes) IA__gtk_label_set_attributes __attribute((visibility("hidden")));
#define gtk_label_set_attributes IA__gtk_label_set_attributes

extern __typeof (gtk_label_set_ellipsize) IA__gtk_label_set_ellipsize __attribute((visibility("hidden")));
#define gtk_label_set_ellipsize IA__gtk_label_set_ellipsize

extern __typeof (gtk_label_set_justify) IA__gtk_label_set_justify __attribute((visibility("hidden")));
#define gtk_label_set_justify IA__gtk_label_set_justify

extern __typeof (gtk_label_set_label) IA__gtk_label_set_label __attribute((visibility("hidden")));
#define gtk_label_set_label IA__gtk_label_set_label

extern __typeof (gtk_label_set_line_wrap) IA__gtk_label_set_line_wrap __attribute((visibility("hidden")));
#define gtk_label_set_line_wrap IA__gtk_label_set_line_wrap

extern __typeof (gtk_label_set_line_wrap_mode) IA__gtk_label_set_line_wrap_mode __attribute((visibility("hidden")));
#define gtk_label_set_line_wrap_mode IA__gtk_label_set_line_wrap_mode

extern __typeof (gtk_label_set_markup) IA__gtk_label_set_markup __attribute((visibility("hidden")));
#define gtk_label_set_markup IA__gtk_label_set_markup

extern __typeof (gtk_label_set_markup_with_mnemonic) IA__gtk_label_set_markup_with_mnemonic __attribute((visibility("hidden")));
#define gtk_label_set_markup_with_mnemonic IA__gtk_label_set_markup_with_mnemonic

extern __typeof (gtk_label_set_max_width_chars) IA__gtk_label_set_max_width_chars __attribute((visibility("hidden")));
#define gtk_label_set_max_width_chars IA__gtk_label_set_max_width_chars

extern __typeof (gtk_label_set_mnemonic_widget) IA__gtk_label_set_mnemonic_widget __attribute((visibility("hidden")));
#define gtk_label_set_mnemonic_widget IA__gtk_label_set_mnemonic_widget

extern __typeof (gtk_label_set_pattern) IA__gtk_label_set_pattern __attribute((visibility("hidden")));
#define gtk_label_set_pattern IA__gtk_label_set_pattern

extern __typeof (gtk_label_set_selectable) IA__gtk_label_set_selectable __attribute((visibility("hidden")));
#define gtk_label_set_selectable IA__gtk_label_set_selectable

extern __typeof (gtk_label_set_single_line_mode) IA__gtk_label_set_single_line_mode __attribute((visibility("hidden")));
#define gtk_label_set_single_line_mode IA__gtk_label_set_single_line_mode

extern __typeof (gtk_label_set_text) IA__gtk_label_set_text __attribute((visibility("hidden")));
#define gtk_label_set_text IA__gtk_label_set_text

extern __typeof (gtk_label_set_text_with_mnemonic) IA__gtk_label_set_text_with_mnemonic __attribute((visibility("hidden")));
#define gtk_label_set_text_with_mnemonic IA__gtk_label_set_text_with_mnemonic

extern __typeof (gtk_label_set_use_markup) IA__gtk_label_set_use_markup __attribute((visibility("hidden")));
#define gtk_label_set_use_markup IA__gtk_label_set_use_markup

extern __typeof (gtk_label_set_use_underline) IA__gtk_label_set_use_underline __attribute((visibility("hidden")));
#define gtk_label_set_use_underline IA__gtk_label_set_use_underline

extern __typeof (gtk_label_set_width_chars) IA__gtk_label_set_width_chars __attribute((visibility("hidden")));
#define gtk_label_set_width_chars IA__gtk_label_set_width_chars

extern __typeof (gtk_label_get_current_uri) IA__gtk_label_get_current_uri __attribute((visibility("hidden")));
#define gtk_label_get_current_uri IA__gtk_label_get_current_uri

extern __typeof (gtk_label_set_track_visited_links) IA__gtk_label_set_track_visited_links __attribute((visibility("hidden")));
#define gtk_label_set_track_visited_links IA__gtk_label_set_track_visited_links

extern __typeof (gtk_label_get_track_visited_links) IA__gtk_label_get_track_visited_links __attribute((visibility("hidden")));
#define gtk_label_get_track_visited_links IA__gtk_label_get_track_visited_links

#endif
#endif
#if IN_HEADER(__GTK_LAYOUT_H__)
#if IN_FILE(__GTK_LAYOUT_C__)
extern __typeof (gtk_layout_get_bin_window) IA__gtk_layout_get_bin_window __attribute((visibility("hidden")));
#define gtk_layout_get_bin_window IA__gtk_layout_get_bin_window

extern __typeof (gtk_layout_get_hadjustment) IA__gtk_layout_get_hadjustment __attribute((visibility("hidden")));
#define gtk_layout_get_hadjustment IA__gtk_layout_get_hadjustment

extern __typeof (gtk_layout_get_size) IA__gtk_layout_get_size __attribute((visibility("hidden")));
#define gtk_layout_get_size IA__gtk_layout_get_size

extern __typeof (gtk_layout_get_type) IA__gtk_layout_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_layout_get_type IA__gtk_layout_get_type

extern __typeof (gtk_layout_get_vadjustment) IA__gtk_layout_get_vadjustment __attribute((visibility("hidden")));
#define gtk_layout_get_vadjustment IA__gtk_layout_get_vadjustment

extern __typeof (gtk_layout_move) IA__gtk_layout_move __attribute((visibility("hidden")));
#define gtk_layout_move IA__gtk_layout_move

extern __typeof (gtk_layout_new) IA__gtk_layout_new __attribute((visibility("hidden")));
#define gtk_layout_new IA__gtk_layout_new

extern __typeof (gtk_layout_put) IA__gtk_layout_put __attribute((visibility("hidden")));
#define gtk_layout_put IA__gtk_layout_put

extern __typeof (gtk_layout_set_hadjustment) IA__gtk_layout_set_hadjustment __attribute((visibility("hidden")));
#define gtk_layout_set_hadjustment IA__gtk_layout_set_hadjustment

extern __typeof (gtk_layout_set_size) IA__gtk_layout_set_size __attribute((visibility("hidden")));
#define gtk_layout_set_size IA__gtk_layout_set_size

extern __typeof (gtk_layout_set_vadjustment) IA__gtk_layout_set_vadjustment __attribute((visibility("hidden")));
#define gtk_layout_set_vadjustment IA__gtk_layout_set_vadjustment

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_layout_freeze) IA__gtk_layout_freeze __attribute((visibility("hidden")));
#define gtk_layout_freeze IA__gtk_layout_freeze

extern __typeof (gtk_layout_thaw) IA__gtk_layout_thaw __attribute((visibility("hidden")));
#define gtk_layout_thaw IA__gtk_layout_thaw

#endif
#endif
#endif
#if IN_HEADER(__GTK_LINK_BUTTON_H__)
#if IN_FILE(__GTK_LINK_BUTTON_C__)
extern __typeof (gtk_link_button_get_type) IA__gtk_link_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_link_button_get_type IA__gtk_link_button_get_type

extern __typeof (gtk_link_button_new) IA__gtk_link_button_new __attribute((visibility("hidden")));
#define gtk_link_button_new IA__gtk_link_button_new

extern __typeof (gtk_link_button_new_with_label) IA__gtk_link_button_new_with_label __attribute((visibility("hidden")));
#define gtk_link_button_new_with_label IA__gtk_link_button_new_with_label

extern __typeof (gtk_link_button_get_uri) IA__gtk_link_button_get_uri __attribute((visibility("hidden")));
#define gtk_link_button_get_uri IA__gtk_link_button_get_uri

extern __typeof (gtk_link_button_set_uri) IA__gtk_link_button_set_uri __attribute((visibility("hidden")));
#define gtk_link_button_set_uri IA__gtk_link_button_set_uri

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_link_button_set_uri_hook) IA__gtk_link_button_set_uri_hook __attribute((visibility("hidden")));
#define gtk_link_button_set_uri_hook IA__gtk_link_button_set_uri_hook

#endif
extern __typeof (gtk_link_button_get_visited) IA__gtk_link_button_get_visited __attribute((visibility("hidden")));
#define gtk_link_button_get_visited IA__gtk_link_button_get_visited

extern __typeof (gtk_link_button_set_visited) IA__gtk_link_button_set_visited __attribute((visibility("hidden")));
#define gtk_link_button_set_visited IA__gtk_link_button_set_visited

#endif
#endif
#if IN_HEADER(__GTK_LIST_H__)
#if IN_FILE(__GTK_LIST_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_list_append_items) IA__gtk_list_append_items __attribute((visibility("hidden")));
#define gtk_list_append_items IA__gtk_list_append_items

extern __typeof (gtk_list_child_position) IA__gtk_list_child_position __attribute((visibility("hidden")));
#define gtk_list_child_position IA__gtk_list_child_position

extern __typeof (gtk_list_clear_items) IA__gtk_list_clear_items __attribute((visibility("hidden")));
#define gtk_list_clear_items IA__gtk_list_clear_items

extern __typeof (gtk_list_end_drag_selection) IA__gtk_list_end_drag_selection __attribute((visibility("hidden")));
#define gtk_list_end_drag_selection IA__gtk_list_end_drag_selection

extern __typeof (gtk_list_end_selection) IA__gtk_list_end_selection __attribute((visibility("hidden")));
#define gtk_list_end_selection IA__gtk_list_end_selection

extern __typeof (gtk_list_extend_selection) IA__gtk_list_extend_selection __attribute((visibility("hidden")));
#define gtk_list_extend_selection IA__gtk_list_extend_selection

extern __typeof (gtk_list_get_type) IA__gtk_list_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_list_get_type IA__gtk_list_get_type

extern __typeof (gtk_list_insert_items) IA__gtk_list_insert_items __attribute((visibility("hidden")));
#define gtk_list_insert_items IA__gtk_list_insert_items

extern __typeof (gtk_list_new) IA__gtk_list_new __attribute((visibility("hidden")));
#define gtk_list_new IA__gtk_list_new

extern __typeof (gtk_list_prepend_items) IA__gtk_list_prepend_items __attribute((visibility("hidden")));
#define gtk_list_prepend_items IA__gtk_list_prepend_items

extern __typeof (gtk_list_remove_items) IA__gtk_list_remove_items __attribute((visibility("hidden")));
#define gtk_list_remove_items IA__gtk_list_remove_items

extern __typeof (gtk_list_remove_items_no_unref) IA__gtk_list_remove_items_no_unref __attribute((visibility("hidden")));
#define gtk_list_remove_items_no_unref IA__gtk_list_remove_items_no_unref

extern __typeof (gtk_list_scroll_horizontal) IA__gtk_list_scroll_horizontal __attribute((visibility("hidden")));
#define gtk_list_scroll_horizontal IA__gtk_list_scroll_horizontal

extern __typeof (gtk_list_scroll_vertical) IA__gtk_list_scroll_vertical __attribute((visibility("hidden")));
#define gtk_list_scroll_vertical IA__gtk_list_scroll_vertical

extern __typeof (gtk_list_select_all) IA__gtk_list_select_all __attribute((visibility("hidden")));
#define gtk_list_select_all IA__gtk_list_select_all

extern __typeof (gtk_list_select_child) IA__gtk_list_select_child __attribute((visibility("hidden")));
#define gtk_list_select_child IA__gtk_list_select_child

extern __typeof (gtk_list_select_item) IA__gtk_list_select_item __attribute((visibility("hidden")));
#define gtk_list_select_item IA__gtk_list_select_item

extern __typeof (gtk_list_set_selection_mode) IA__gtk_list_set_selection_mode __attribute((visibility("hidden")));
#define gtk_list_set_selection_mode IA__gtk_list_set_selection_mode

extern __typeof (gtk_list_start_selection) IA__gtk_list_start_selection __attribute((visibility("hidden")));
#define gtk_list_start_selection IA__gtk_list_start_selection

extern __typeof (gtk_list_toggle_add_mode) IA__gtk_list_toggle_add_mode __attribute((visibility("hidden")));
#define gtk_list_toggle_add_mode IA__gtk_list_toggle_add_mode

extern __typeof (gtk_list_toggle_focus_row) IA__gtk_list_toggle_focus_row __attribute((visibility("hidden")));
#define gtk_list_toggle_focus_row IA__gtk_list_toggle_focus_row

extern __typeof (gtk_list_toggle_row) IA__gtk_list_toggle_row __attribute((visibility("hidden")));
#define gtk_list_toggle_row IA__gtk_list_toggle_row

extern __typeof (gtk_list_undo_selection) IA__gtk_list_undo_selection __attribute((visibility("hidden")));
#define gtk_list_undo_selection IA__gtk_list_undo_selection

extern __typeof (gtk_list_unselect_all) IA__gtk_list_unselect_all __attribute((visibility("hidden")));
#define gtk_list_unselect_all IA__gtk_list_unselect_all

extern __typeof (gtk_list_unselect_child) IA__gtk_list_unselect_child __attribute((visibility("hidden")));
#define gtk_list_unselect_child IA__gtk_list_unselect_child

extern __typeof (gtk_list_unselect_item) IA__gtk_list_unselect_item __attribute((visibility("hidden")));
#define gtk_list_unselect_item IA__gtk_list_unselect_item

#endif
#endif
#endif
#if IN_HEADER(__GTK_LIST_ITEM_H__)
#if IN_FILE(__GTK_LIST_ITEM_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_list_item_deselect) IA__gtk_list_item_deselect __attribute((visibility("hidden")));
#define gtk_list_item_deselect IA__gtk_list_item_deselect

extern __typeof (gtk_list_item_get_type) IA__gtk_list_item_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_list_item_get_type IA__gtk_list_item_get_type

extern __typeof (gtk_list_item_new) IA__gtk_list_item_new __attribute((visibility("hidden")));
#define gtk_list_item_new IA__gtk_list_item_new

extern __typeof (gtk_list_item_new_with_label) IA__gtk_list_item_new_with_label __attribute((visibility("hidden")));
#define gtk_list_item_new_with_label IA__gtk_list_item_new_with_label

extern __typeof (gtk_list_item_select) IA__gtk_list_item_select __attribute((visibility("hidden")));
#define gtk_list_item_select IA__gtk_list_item_select

#endif
#endif
#endif
#if IN_HEADER(__GTK_LIST_STORE_H__)
#if IN_FILE(__GTK_LIST_STORE_C__)
extern __typeof (gtk_list_store_append) IA__gtk_list_store_append __attribute((visibility("hidden")));
#define gtk_list_store_append IA__gtk_list_store_append

extern __typeof (gtk_list_store_clear) IA__gtk_list_store_clear __attribute((visibility("hidden")));
#define gtk_list_store_clear IA__gtk_list_store_clear

extern __typeof (gtk_list_store_get_type) IA__gtk_list_store_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_list_store_get_type IA__gtk_list_store_get_type

extern __typeof (gtk_list_store_insert) IA__gtk_list_store_insert __attribute((visibility("hidden")));
#define gtk_list_store_insert IA__gtk_list_store_insert

extern __typeof (gtk_list_store_insert_after) IA__gtk_list_store_insert_after __attribute((visibility("hidden")));
#define gtk_list_store_insert_after IA__gtk_list_store_insert_after

extern __typeof (gtk_list_store_insert_before) IA__gtk_list_store_insert_before __attribute((visibility("hidden")));
#define gtk_list_store_insert_before IA__gtk_list_store_insert_before

extern __typeof (gtk_list_store_insert_with_values) IA__gtk_list_store_insert_with_values __attribute((visibility("hidden")));
#define gtk_list_store_insert_with_values IA__gtk_list_store_insert_with_values

extern __typeof (gtk_list_store_insert_with_valuesv) IA__gtk_list_store_insert_with_valuesv __attribute((visibility("hidden")));
#define gtk_list_store_insert_with_valuesv IA__gtk_list_store_insert_with_valuesv

extern __typeof (gtk_list_store_iter_is_valid) IA__gtk_list_store_iter_is_valid __attribute((visibility("hidden")));
#define gtk_list_store_iter_is_valid IA__gtk_list_store_iter_is_valid

extern __typeof (gtk_list_store_move_after) IA__gtk_list_store_move_after __attribute((visibility("hidden")));
#define gtk_list_store_move_after IA__gtk_list_store_move_after

extern __typeof (gtk_list_store_move_before) IA__gtk_list_store_move_before __attribute((visibility("hidden")));
#define gtk_list_store_move_before IA__gtk_list_store_move_before

extern __typeof (gtk_list_store_new) IA__gtk_list_store_new __attribute((visibility("hidden")));
#define gtk_list_store_new IA__gtk_list_store_new

extern __typeof (gtk_list_store_newv) IA__gtk_list_store_newv __attribute((visibility("hidden")));
#define gtk_list_store_newv IA__gtk_list_store_newv

extern __typeof (gtk_list_store_prepend) IA__gtk_list_store_prepend __attribute((visibility("hidden")));
#define gtk_list_store_prepend IA__gtk_list_store_prepend

extern __typeof (gtk_list_store_remove) IA__gtk_list_store_remove __attribute((visibility("hidden")));
#define gtk_list_store_remove IA__gtk_list_store_remove

extern __typeof (gtk_list_store_reorder) IA__gtk_list_store_reorder __attribute((visibility("hidden")));
#define gtk_list_store_reorder IA__gtk_list_store_reorder

extern __typeof (gtk_list_store_set) IA__gtk_list_store_set __attribute((visibility("hidden")));
#define gtk_list_store_set IA__gtk_list_store_set

extern __typeof (gtk_list_store_set_column_types) IA__gtk_list_store_set_column_types __attribute((visibility("hidden")));
#define gtk_list_store_set_column_types IA__gtk_list_store_set_column_types

extern __typeof (gtk_list_store_set_valist) IA__gtk_list_store_set_valist __attribute((visibility("hidden")));
#define gtk_list_store_set_valist IA__gtk_list_store_set_valist

extern __typeof (gtk_list_store_set_value) IA__gtk_list_store_set_value __attribute((visibility("hidden")));
#define gtk_list_store_set_value IA__gtk_list_store_set_value

extern __typeof (gtk_list_store_set_valuesv) IA__gtk_list_store_set_valuesv __attribute((visibility("hidden")));
#define gtk_list_store_set_valuesv IA__gtk_list_store_set_valuesv

extern __typeof (gtk_list_store_swap) IA__gtk_list_store_swap __attribute((visibility("hidden")));
#define gtk_list_store_swap IA__gtk_list_store_swap

#endif
#endif
#if IN_HEADER(__GTK_MAIN_H__)
#if IN_FILE(__GTK_MAIN_C__)
extern __typeof (gtk_get_option_group) IA__gtk_get_option_group __attribute((visibility("hidden")));
#define gtk_get_option_group IA__gtk_get_option_group

extern __typeof (gtk_get_current_event) IA__gtk_get_current_event __attribute((visibility("hidden")));
#define gtk_get_current_event IA__gtk_get_current_event

extern __typeof (gtk_get_current_event_state) IA__gtk_get_current_event_state __attribute((visibility("hidden")));
#define gtk_get_current_event_state IA__gtk_get_current_event_state

extern __typeof (gtk_get_current_event_time) IA__gtk_get_current_event_time __attribute((visibility("hidden")));
#define gtk_get_current_event_time IA__gtk_get_current_event_time

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_exit) IA__gtk_exit __attribute((visibility("hidden")));
#define gtk_exit IA__gtk_exit

extern __typeof (gtk_set_locale) IA__gtk_set_locale __attribute((visibility("hidden")));
#define gtk_set_locale IA__gtk_set_locale

extern __typeof (gtk_idle_add) IA__gtk_idle_add __attribute((visibility("hidden")));
#define gtk_idle_add IA__gtk_idle_add

extern __typeof (gtk_idle_add_full) IA__gtk_idle_add_full __attribute((visibility("hidden")));
#define gtk_idle_add_full IA__gtk_idle_add_full

extern __typeof (gtk_idle_add_priority) IA__gtk_idle_add_priority __attribute((visibility("hidden")));
#define gtk_idle_add_priority IA__gtk_idle_add_priority

extern __typeof (gtk_idle_remove) IA__gtk_idle_remove __attribute((visibility("hidden")));
#define gtk_idle_remove IA__gtk_idle_remove

extern __typeof (gtk_idle_remove_by_data) IA__gtk_idle_remove_by_data __attribute((visibility("hidden")));
#define gtk_idle_remove_by_data IA__gtk_idle_remove_by_data

extern __typeof (gtk_timeout_add) IA__gtk_timeout_add __attribute((visibility("hidden")));
#define gtk_timeout_add IA__gtk_timeout_add

extern __typeof (gtk_timeout_add_full) IA__gtk_timeout_add_full __attribute((visibility("hidden")));
#define gtk_timeout_add_full IA__gtk_timeout_add_full

extern __typeof (gtk_timeout_remove) IA__gtk_timeout_remove __attribute((visibility("hidden")));
#define gtk_timeout_remove IA__gtk_timeout_remove

extern __typeof (gtk_input_add_full) IA__gtk_input_add_full __attribute((visibility("hidden")));
#define gtk_input_add_full IA__gtk_input_add_full

extern __typeof (gtk_input_remove) IA__gtk_input_remove __attribute((visibility("hidden")));
#define gtk_input_remove IA__gtk_input_remove

#endif
extern __typeof (gtk_false) IA__gtk_false __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_false IA__gtk_false

extern __typeof (gtk_true) IA__gtk_true __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_true IA__gtk_true

extern __typeof (gtk_events_pending) IA__gtk_events_pending __attribute((visibility("hidden")));
#define gtk_events_pending IA__gtk_events_pending

extern __typeof (gtk_disable_setlocale) IA__gtk_disable_setlocale __attribute((visibility("hidden")));
#define gtk_disable_setlocale IA__gtk_disable_setlocale

extern __typeof (gtk_check_version) IA__gtk_check_version __attribute((visibility("hidden")));
#define gtk_check_version IA__gtk_check_version

extern __typeof (gtk_get_default_language) IA__gtk_get_default_language __attribute((visibility("hidden")));
#define gtk_get_default_language IA__gtk_get_default_language

extern __typeof (gtk_get_event_widget) IA__gtk_get_event_widget __attribute((visibility("hidden")));
#define gtk_get_event_widget IA__gtk_get_event_widget

extern __typeof (gtk_grab_add) IA__gtk_grab_add __attribute((visibility("hidden")));
#define gtk_grab_add IA__gtk_grab_add

extern __typeof (gtk_grab_get_current) IA__gtk_grab_get_current __attribute((visibility("hidden")));
#define gtk_grab_get_current IA__gtk_grab_get_current

extern __typeof (gtk_grab_remove) IA__gtk_grab_remove __attribute((visibility("hidden")));
#define gtk_grab_remove IA__gtk_grab_remove

extern __typeof (gtk_propagate_event) IA__gtk_propagate_event __attribute((visibility("hidden")));
#define gtk_propagate_event IA__gtk_propagate_event

extern __typeof (gtk_quit_add) IA__gtk_quit_add __attribute((visibility("hidden")));
#define gtk_quit_add IA__gtk_quit_add

extern __typeof (gtk_quit_add_destroy) IA__gtk_quit_add_destroy __attribute((visibility("hidden")));
#define gtk_quit_add_destroy IA__gtk_quit_add_destroy

extern __typeof (gtk_quit_add_full) IA__gtk_quit_add_full __attribute((visibility("hidden")));
#define gtk_quit_add_full IA__gtk_quit_add_full

extern __typeof (gtk_quit_remove) IA__gtk_quit_remove __attribute((visibility("hidden")));
#define gtk_quit_remove IA__gtk_quit_remove

extern __typeof (gtk_quit_remove_by_data) IA__gtk_quit_remove_by_data __attribute((visibility("hidden")));
#define gtk_quit_remove_by_data IA__gtk_quit_remove_by_data

extern __typeof (gtk_key_snooper_install) IA__gtk_key_snooper_install __attribute((visibility("hidden")));
#define gtk_key_snooper_install IA__gtk_key_snooper_install

extern __typeof (gtk_key_snooper_remove) IA__gtk_key_snooper_remove __attribute((visibility("hidden")));
#define gtk_key_snooper_remove IA__gtk_key_snooper_remove

extern __typeof (gtk_init) IA__gtk_init __attribute((visibility("hidden")));
#define gtk_init IA__gtk_init

#ifdef G_OS_WIN32
extern __typeof (gtk_init_abi_check) IA__gtk_init_abi_check __attribute((visibility("hidden")));
#define gtk_init_abi_check IA__gtk_init_abi_check

#endif
extern __typeof (gtk_init_add) IA__gtk_init_add __attribute((visibility("hidden")));
#define gtk_init_add IA__gtk_init_add

extern __typeof (gtk_init_check) IA__gtk_init_check __attribute((visibility("hidden")));
#define gtk_init_check IA__gtk_init_check

#ifdef G_OS_WIN32
extern __typeof (gtk_init_check_abi_check) IA__gtk_init_check_abi_check __attribute((visibility("hidden")));
#define gtk_init_check_abi_check IA__gtk_init_check_abi_check

#endif
extern __typeof (gtk_init_with_args) IA__gtk_init_with_args __attribute((visibility("hidden")));
#define gtk_init_with_args IA__gtk_init_with_args

extern __typeof (gtk_parse_args) IA__gtk_parse_args __attribute((visibility("hidden")));
#define gtk_parse_args IA__gtk_parse_args

extern __typeof (gtk_main) IA__gtk_main __attribute((visibility("hidden")));
#define gtk_main IA__gtk_main

extern __typeof (gtk_main_do_event) IA__gtk_main_do_event __attribute((visibility("hidden")));
#define gtk_main_do_event IA__gtk_main_do_event

extern __typeof (gtk_main_iteration) IA__gtk_main_iteration __attribute((visibility("hidden")));
#define gtk_main_iteration IA__gtk_main_iteration

extern __typeof (gtk_main_iteration_do) IA__gtk_main_iteration_do __attribute((visibility("hidden")));
#define gtk_main_iteration_do IA__gtk_main_iteration_do

extern __typeof (gtk_main_level) IA__gtk_main_level __attribute((visibility("hidden")));
#define gtk_main_level IA__gtk_main_level

extern __typeof (gtk_main_quit) IA__gtk_main_quit __attribute((visibility("hidden")));
#define gtk_main_quit IA__gtk_main_quit

#endif
#endif
#if IN_HEADER(__GTK_INPUTDIALOG_H__)
#if IN_FILE(__GTK_INPUTDIALOG_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_input_dialog_get_type) IA__gtk_input_dialog_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_input_dialog_get_type IA__gtk_input_dialog_get_type

extern __typeof (gtk_input_dialog_new) IA__gtk_input_dialog_new __attribute((visibility("hidden")));
#define gtk_input_dialog_new IA__gtk_input_dialog_new

#endif
#endif
#endif
#if IN_HEADER(__gtk_marshal_MARSHAL_H__) || IN_HEADER(__GTK_MARSHAL_MARSHAL_H__)
#if IN_FILE(__gtk_marshal_MARSHAL_C__) || IN_FILE(__GTK_MARSHAL_MARSHAL_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_marshal_BOOLEAN__POINTER) IA__gtk_marshal_BOOLEAN__POINTER __attribute((visibility("hidden")));
#define gtk_marshal_BOOLEAN__POINTER IA__gtk_marshal_BOOLEAN__POINTER

extern __typeof (gtk_marshal_BOOLEAN__POINTER_INT_INT) IA__gtk_marshal_BOOLEAN__POINTER_INT_INT __attribute((visibility("hidden")));
#define gtk_marshal_BOOLEAN__POINTER_INT_INT IA__gtk_marshal_BOOLEAN__POINTER_INT_INT

extern __typeof (gtk_marshal_BOOLEAN__POINTER_INT_INT_UINT) IA__gtk_marshal_BOOLEAN__POINTER_INT_INT_UINT __attribute((visibility("hidden")));
#define gtk_marshal_BOOLEAN__POINTER_INT_INT_UINT IA__gtk_marshal_BOOLEAN__POINTER_INT_INT_UINT

extern __typeof (gtk_marshal_BOOLEAN__POINTER_POINTER_INT_INT) IA__gtk_marshal_BOOLEAN__POINTER_POINTER_INT_INT __attribute((visibility("hidden")));
#define gtk_marshal_BOOLEAN__POINTER_POINTER_INT_INT IA__gtk_marshal_BOOLEAN__POINTER_POINTER_INT_INT

extern __typeof (gtk_marshal_BOOLEAN__POINTER_STRING_STRING_POINTER) IA__gtk_marshal_BOOLEAN__POINTER_STRING_STRING_POINTER __attribute((visibility("hidden")));
#define gtk_marshal_BOOLEAN__POINTER_STRING_STRING_POINTER IA__gtk_marshal_BOOLEAN__POINTER_STRING_STRING_POINTER

extern __typeof (gtk_marshal_BOOLEAN__VOID) IA__gtk_marshal_BOOLEAN__VOID __attribute((visibility("hidden")));
#define gtk_marshal_BOOLEAN__VOID IA__gtk_marshal_BOOLEAN__VOID

extern __typeof (gtk_marshal_ENUM__ENUM) IA__gtk_marshal_ENUM__ENUM __attribute((visibility("hidden")));
#define gtk_marshal_ENUM__ENUM IA__gtk_marshal_ENUM__ENUM

extern __typeof (gtk_marshal_INT__POINTER) IA__gtk_marshal_INT__POINTER __attribute((visibility("hidden")));
#define gtk_marshal_INT__POINTER IA__gtk_marshal_INT__POINTER

extern __typeof (gtk_marshal_INT__POINTER_CHAR_CHAR) IA__gtk_marshal_INT__POINTER_CHAR_CHAR __attribute((visibility("hidden")));
#define gtk_marshal_INT__POINTER_CHAR_CHAR IA__gtk_marshal_INT__POINTER_CHAR_CHAR

extern __typeof (gtk_marshal_VOID__ENUM_FLOAT) IA__gtk_marshal_VOID__ENUM_FLOAT __attribute((visibility("hidden")));
#define gtk_marshal_VOID__ENUM_FLOAT IA__gtk_marshal_VOID__ENUM_FLOAT

extern __typeof (gtk_marshal_VOID__ENUM_FLOAT_BOOLEAN) IA__gtk_marshal_VOID__ENUM_FLOAT_BOOLEAN __attribute((visibility("hidden")));
#define gtk_marshal_VOID__ENUM_FLOAT_BOOLEAN IA__gtk_marshal_VOID__ENUM_FLOAT_BOOLEAN

extern __typeof (gtk_marshal_VOID__INT_INT) IA__gtk_marshal_VOID__INT_INT __attribute((visibility("hidden")));
#define gtk_marshal_VOID__INT_INT IA__gtk_marshal_VOID__INT_INT

extern __typeof (gtk_marshal_VOID__INT_INT_POINTER) IA__gtk_marshal_VOID__INT_INT_POINTER __attribute((visibility("hidden")));
#define gtk_marshal_VOID__INT_INT_POINTER IA__gtk_marshal_VOID__INT_INT_POINTER

extern __typeof (gtk_marshal_VOID__POINTER_INT) IA__gtk_marshal_VOID__POINTER_INT __attribute((visibility("hidden")));
#define gtk_marshal_VOID__POINTER_INT IA__gtk_marshal_VOID__POINTER_INT

extern __typeof (gtk_marshal_VOID__POINTER_INT_INT_POINTER_UINT_UINT) IA__gtk_marshal_VOID__POINTER_INT_INT_POINTER_UINT_UINT __attribute((visibility("hidden")));
#define gtk_marshal_VOID__POINTER_INT_INT_POINTER_UINT_UINT IA__gtk_marshal_VOID__POINTER_INT_INT_POINTER_UINT_UINT

extern __typeof (gtk_marshal_VOID__POINTER_POINTER) IA__gtk_marshal_VOID__POINTER_POINTER __attribute((visibility("hidden")));
#define gtk_marshal_VOID__POINTER_POINTER IA__gtk_marshal_VOID__POINTER_POINTER

extern __typeof (gtk_marshal_VOID__POINTER_POINTER_POINTER) IA__gtk_marshal_VOID__POINTER_POINTER_POINTER __attribute((visibility("hidden")));
#define gtk_marshal_VOID__POINTER_POINTER_POINTER IA__gtk_marshal_VOID__POINTER_POINTER_POINTER

extern __typeof (gtk_marshal_VOID__POINTER_POINTER_UINT_UINT) IA__gtk_marshal_VOID__POINTER_POINTER_UINT_UINT __attribute((visibility("hidden")));
#define gtk_marshal_VOID__POINTER_POINTER_UINT_UINT IA__gtk_marshal_VOID__POINTER_POINTER_UINT_UINT

extern __typeof (gtk_marshal_VOID__POINTER_STRING_STRING) IA__gtk_marshal_VOID__POINTER_STRING_STRING __attribute((visibility("hidden")));
#define gtk_marshal_VOID__POINTER_STRING_STRING IA__gtk_marshal_VOID__POINTER_STRING_STRING

extern __typeof (gtk_marshal_VOID__POINTER_UINT) IA__gtk_marshal_VOID__POINTER_UINT __attribute((visibility("hidden")));
#define gtk_marshal_VOID__POINTER_UINT IA__gtk_marshal_VOID__POINTER_UINT

extern __typeof (gtk_marshal_VOID__POINTER_UINT_ENUM) IA__gtk_marshal_VOID__POINTER_UINT_ENUM __attribute((visibility("hidden")));
#define gtk_marshal_VOID__POINTER_UINT_ENUM IA__gtk_marshal_VOID__POINTER_UINT_ENUM

extern __typeof (gtk_marshal_VOID__POINTER_UINT_UINT) IA__gtk_marshal_VOID__POINTER_UINT_UINT __attribute((visibility("hidden")));
#define gtk_marshal_VOID__POINTER_UINT_UINT IA__gtk_marshal_VOID__POINTER_UINT_UINT

extern __typeof (gtk_marshal_VOID__STRING_INT_POINTER) IA__gtk_marshal_VOID__STRING_INT_POINTER __attribute((visibility("hidden")));
#define gtk_marshal_VOID__STRING_INT_POINTER IA__gtk_marshal_VOID__STRING_INT_POINTER

extern __typeof (gtk_marshal_VOID__UINT_POINTER_UINT_ENUM_ENUM_POINTER) IA__gtk_marshal_VOID__UINT_POINTER_UINT_ENUM_ENUM_POINTER __attribute((visibility("hidden")));
#define gtk_marshal_VOID__UINT_POINTER_UINT_ENUM_ENUM_POINTER IA__gtk_marshal_VOID__UINT_POINTER_UINT_ENUM_ENUM_POINTER

extern __typeof (gtk_marshal_VOID__UINT_POINTER_UINT_UINT_ENUM) IA__gtk_marshal_VOID__UINT_POINTER_UINT_UINT_ENUM __attribute((visibility("hidden")));
#define gtk_marshal_VOID__UINT_POINTER_UINT_UINT_ENUM IA__gtk_marshal_VOID__UINT_POINTER_UINT_UINT_ENUM

extern __typeof (gtk_marshal_VOID__UINT_STRING) IA__gtk_marshal_VOID__UINT_STRING __attribute((visibility("hidden")));
#define gtk_marshal_VOID__UINT_STRING IA__gtk_marshal_VOID__UINT_STRING

#endif
#endif
#endif
#if IN_HEADER(__GTK_MENU_BAR_H__)
#if IN_FILE(__GTK_MENU_BAR_C__)
extern __typeof (gtk_menu_bar_get_type) IA__gtk_menu_bar_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_menu_bar_get_type IA__gtk_menu_bar_get_type

extern __typeof (gtk_menu_bar_new) IA__gtk_menu_bar_new __attribute((visibility("hidden")));
#define gtk_menu_bar_new IA__gtk_menu_bar_new

extern __typeof (gtk_menu_bar_get_child_pack_direction) IA__gtk_menu_bar_get_child_pack_direction __attribute((visibility("hidden")));
#define gtk_menu_bar_get_child_pack_direction IA__gtk_menu_bar_get_child_pack_direction

extern __typeof (gtk_menu_bar_set_child_pack_direction) IA__gtk_menu_bar_set_child_pack_direction __attribute((visibility("hidden")));
#define gtk_menu_bar_set_child_pack_direction IA__gtk_menu_bar_set_child_pack_direction

extern __typeof (gtk_menu_bar_get_pack_direction) IA__gtk_menu_bar_get_pack_direction __attribute((visibility("hidden")));
#define gtk_menu_bar_get_pack_direction IA__gtk_menu_bar_get_pack_direction

extern __typeof (gtk_menu_bar_set_pack_direction) IA__gtk_menu_bar_set_pack_direction __attribute((visibility("hidden")));
#define gtk_menu_bar_set_pack_direction IA__gtk_menu_bar_set_pack_direction

#endif
#endif
#if IN_HEADER(__GTK_MENU_H__)
#if IN_FILE(__GTK_MENU_C__)
extern __typeof (gtk_menu_attach) IA__gtk_menu_attach __attribute((visibility("hidden")));
#define gtk_menu_attach IA__gtk_menu_attach

extern __typeof (gtk_menu_attach_to_widget) IA__gtk_menu_attach_to_widget __attribute((visibility("hidden")));
#define gtk_menu_attach_to_widget IA__gtk_menu_attach_to_widget

extern __typeof (gtk_menu_detach) IA__gtk_menu_detach __attribute((visibility("hidden")));
#define gtk_menu_detach IA__gtk_menu_detach

extern __typeof (gtk_menu_get_accel_group) IA__gtk_menu_get_accel_group __attribute((visibility("hidden")));
#define gtk_menu_get_accel_group IA__gtk_menu_get_accel_group

extern __typeof (gtk_menu_get_accel_path) IA__gtk_menu_get_accel_path __attribute((visibility("hidden")));
#define gtk_menu_get_accel_path IA__gtk_menu_get_accel_path

extern __typeof (gtk_menu_get_active) IA__gtk_menu_get_active __attribute((visibility("hidden")));
#define gtk_menu_get_active IA__gtk_menu_get_active

extern __typeof (gtk_menu_get_monitor) IA__gtk_menu_get_monitor __attribute((visibility("hidden")));
#define gtk_menu_get_monitor IA__gtk_menu_get_monitor

extern __typeof (gtk_menu_get_attach_widget) IA__gtk_menu_get_attach_widget __attribute((visibility("hidden")));
#define gtk_menu_get_attach_widget IA__gtk_menu_get_attach_widget

extern __typeof (gtk_menu_get_for_attach_widget) IA__gtk_menu_get_for_attach_widget __attribute((visibility("hidden")));
#define gtk_menu_get_for_attach_widget IA__gtk_menu_get_for_attach_widget

extern __typeof (gtk_menu_get_reserve_toggle_size) IA__gtk_menu_get_reserve_toggle_size __attribute((visibility("hidden")));
#define gtk_menu_get_reserve_toggle_size IA__gtk_menu_get_reserve_toggle_size

extern __typeof (gtk_menu_get_tearoff_state) IA__gtk_menu_get_tearoff_state __attribute((visibility("hidden")));
#define gtk_menu_get_tearoff_state IA__gtk_menu_get_tearoff_state

extern __typeof (gtk_menu_get_title) IA__gtk_menu_get_title __attribute((visibility("hidden")));
#define gtk_menu_get_title IA__gtk_menu_get_title

extern __typeof (gtk_menu_get_type) IA__gtk_menu_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_menu_get_type IA__gtk_menu_get_type

extern __typeof (gtk_menu_new) IA__gtk_menu_new __attribute((visibility("hidden")));
#define gtk_menu_new IA__gtk_menu_new

extern __typeof (gtk_menu_popdown) IA__gtk_menu_popdown __attribute((visibility("hidden")));
#define gtk_menu_popdown IA__gtk_menu_popdown

extern __typeof (gtk_menu_popup) IA__gtk_menu_popup __attribute((visibility("hidden")));
#define gtk_menu_popup IA__gtk_menu_popup

extern __typeof (gtk_menu_reorder_child) IA__gtk_menu_reorder_child __attribute((visibility("hidden")));
#define gtk_menu_reorder_child IA__gtk_menu_reorder_child

extern __typeof (gtk_menu_reposition) IA__gtk_menu_reposition __attribute((visibility("hidden")));
#define gtk_menu_reposition IA__gtk_menu_reposition

extern __typeof (gtk_menu_set_accel_group) IA__gtk_menu_set_accel_group __attribute((visibility("hidden")));
#define gtk_menu_set_accel_group IA__gtk_menu_set_accel_group

extern __typeof (gtk_menu_set_accel_path) IA__gtk_menu_set_accel_path __attribute((visibility("hidden")));
#define gtk_menu_set_accel_path IA__gtk_menu_set_accel_path

extern __typeof (gtk_menu_set_active) IA__gtk_menu_set_active __attribute((visibility("hidden")));
#define gtk_menu_set_active IA__gtk_menu_set_active

extern __typeof (gtk_menu_set_monitor) IA__gtk_menu_set_monitor __attribute((visibility("hidden")));
#define gtk_menu_set_monitor IA__gtk_menu_set_monitor

extern __typeof (gtk_menu_set_reserve_toggle_size) IA__gtk_menu_set_reserve_toggle_size __attribute((visibility("hidden")));
#define gtk_menu_set_reserve_toggle_size IA__gtk_menu_set_reserve_toggle_size

extern __typeof (gtk_menu_set_screen) IA__gtk_menu_set_screen __attribute((visibility("hidden")));
#define gtk_menu_set_screen IA__gtk_menu_set_screen

extern __typeof (gtk_menu_set_tearoff_state) IA__gtk_menu_set_tearoff_state __attribute((visibility("hidden")));
#define gtk_menu_set_tearoff_state IA__gtk_menu_set_tearoff_state

extern __typeof (gtk_menu_set_title) IA__gtk_menu_set_title __attribute((visibility("hidden")));
#define gtk_menu_set_title IA__gtk_menu_set_title

#endif
#endif
#if IN_HEADER(__GTK_MENU_ITEM_H__)
#if IN_FILE(__GTK_MENU_ITEM_C__)
extern __typeof (gtk_menu_item_activate) IA__gtk_menu_item_activate __attribute((visibility("hidden")));
#define gtk_menu_item_activate IA__gtk_menu_item_activate

extern __typeof (gtk_menu_item_deselect) IA__gtk_menu_item_deselect __attribute((visibility("hidden")));
#define gtk_menu_item_deselect IA__gtk_menu_item_deselect

extern __typeof (gtk_menu_item_get_accel_path) IA__gtk_menu_item_get_accel_path __attribute((visibility("hidden")));
#define gtk_menu_item_get_accel_path IA__gtk_menu_item_get_accel_path

extern __typeof (gtk_menu_item_get_label) IA__gtk_menu_item_get_label __attribute((visibility("hidden")));
#define gtk_menu_item_get_label IA__gtk_menu_item_get_label

extern __typeof (gtk_menu_item_get_right_justified) IA__gtk_menu_item_get_right_justified __attribute((visibility("hidden")));
#define gtk_menu_item_get_right_justified IA__gtk_menu_item_get_right_justified

extern __typeof (gtk_menu_item_get_submenu) IA__gtk_menu_item_get_submenu __attribute((visibility("hidden")));
#define gtk_menu_item_get_submenu IA__gtk_menu_item_get_submenu

extern __typeof (gtk_menu_item_get_type) IA__gtk_menu_item_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_menu_item_get_type IA__gtk_menu_item_get_type

extern __typeof (gtk_menu_item_get_use_underline) IA__gtk_menu_item_get_use_underline __attribute((visibility("hidden")));
#define gtk_menu_item_get_use_underline IA__gtk_menu_item_get_use_underline

extern __typeof (gtk_menu_item_new) IA__gtk_menu_item_new __attribute((visibility("hidden")));
#define gtk_menu_item_new IA__gtk_menu_item_new

extern __typeof (gtk_menu_item_new_with_label) IA__gtk_menu_item_new_with_label __attribute((visibility("hidden")));
#define gtk_menu_item_new_with_label IA__gtk_menu_item_new_with_label

extern __typeof (gtk_menu_item_new_with_mnemonic) IA__gtk_menu_item_new_with_mnemonic __attribute((visibility("hidden")));
#define gtk_menu_item_new_with_mnemonic IA__gtk_menu_item_new_with_mnemonic

extern __typeof (gtk_menu_item_select) IA__gtk_menu_item_select __attribute((visibility("hidden")));
#define gtk_menu_item_select IA__gtk_menu_item_select

extern __typeof (gtk_menu_item_set_accel_path) IA__gtk_menu_item_set_accel_path __attribute((visibility("hidden")));
#define gtk_menu_item_set_accel_path IA__gtk_menu_item_set_accel_path

extern __typeof (gtk_menu_item_set_label) IA__gtk_menu_item_set_label __attribute((visibility("hidden")));
#define gtk_menu_item_set_label IA__gtk_menu_item_set_label

extern __typeof (gtk_menu_item_set_right_justified) IA__gtk_menu_item_set_right_justified __attribute((visibility("hidden")));
#define gtk_menu_item_set_right_justified IA__gtk_menu_item_set_right_justified

extern __typeof (gtk_menu_item_set_submenu) IA__gtk_menu_item_set_submenu __attribute((visibility("hidden")));
#define gtk_menu_item_set_submenu IA__gtk_menu_item_set_submenu

extern __typeof (gtk_menu_item_set_use_underline) IA__gtk_menu_item_set_use_underline __attribute((visibility("hidden")));
#define gtk_menu_item_set_use_underline IA__gtk_menu_item_set_use_underline

extern __typeof (gtk_menu_item_toggle_size_allocate) IA__gtk_menu_item_toggle_size_allocate __attribute((visibility("hidden")));
#define gtk_menu_item_toggle_size_allocate IA__gtk_menu_item_toggle_size_allocate

extern __typeof (gtk_menu_item_toggle_size_request) IA__gtk_menu_item_toggle_size_request __attribute((visibility("hidden")));
#define gtk_menu_item_toggle_size_request IA__gtk_menu_item_toggle_size_request

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_menu_item_remove_submenu) IA__gtk_menu_item_remove_submenu __attribute((visibility("hidden")));
#define gtk_menu_item_remove_submenu IA__gtk_menu_item_remove_submenu

#endif
#endif
#endif
#if IN_HEADER(__GTK_MENU_SHELL_H__)
#if IN_FILE(__GTK_MENU_SHELL_C__)
extern __typeof (gtk_menu_shell_activate_item) IA__gtk_menu_shell_activate_item __attribute((visibility("hidden")));
#define gtk_menu_shell_activate_item IA__gtk_menu_shell_activate_item

extern __typeof (gtk_menu_shell_append) IA__gtk_menu_shell_append __attribute((visibility("hidden")));
#define gtk_menu_shell_append IA__gtk_menu_shell_append

extern __typeof (gtk_menu_shell_cancel) IA__gtk_menu_shell_cancel __attribute((visibility("hidden")));
#define gtk_menu_shell_cancel IA__gtk_menu_shell_cancel

extern __typeof (gtk_menu_shell_deactivate) IA__gtk_menu_shell_deactivate __attribute((visibility("hidden")));
#define gtk_menu_shell_deactivate IA__gtk_menu_shell_deactivate

extern __typeof (gtk_menu_shell_deselect) IA__gtk_menu_shell_deselect __attribute((visibility("hidden")));
#define gtk_menu_shell_deselect IA__gtk_menu_shell_deselect

extern __typeof (gtk_menu_shell_get_type) IA__gtk_menu_shell_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_menu_shell_get_type IA__gtk_menu_shell_get_type

extern __typeof (gtk_menu_shell_insert) IA__gtk_menu_shell_insert __attribute((visibility("hidden")));
#define gtk_menu_shell_insert IA__gtk_menu_shell_insert

extern __typeof (gtk_menu_shell_prepend) IA__gtk_menu_shell_prepend __attribute((visibility("hidden")));
#define gtk_menu_shell_prepend IA__gtk_menu_shell_prepend

extern __typeof (gtk_menu_shell_select_first) IA__gtk_menu_shell_select_first __attribute((visibility("hidden")));
#define gtk_menu_shell_select_first IA__gtk_menu_shell_select_first

extern __typeof (gtk_menu_shell_select_item) IA__gtk_menu_shell_select_item __attribute((visibility("hidden")));
#define gtk_menu_shell_select_item IA__gtk_menu_shell_select_item

extern __typeof (gtk_menu_shell_set_take_focus) IA__gtk_menu_shell_set_take_focus __attribute((visibility("hidden")));
#define gtk_menu_shell_set_take_focus IA__gtk_menu_shell_set_take_focus

extern __typeof (gtk_menu_shell_get_take_focus) IA__gtk_menu_shell_get_take_focus __attribute((visibility("hidden")));
#define gtk_menu_shell_get_take_focus IA__gtk_menu_shell_get_take_focus

#endif
#endif
#if IN_HEADER(__GTK_MENU_TOOL_BUTTON_H__)
#if IN_FILE(__GTK_MENU_TOOL_BUTTON_C__)
extern __typeof (gtk_menu_tool_button_get_menu) IA__gtk_menu_tool_button_get_menu __attribute((visibility("hidden")));
#define gtk_menu_tool_button_get_menu IA__gtk_menu_tool_button_get_menu

extern __typeof (gtk_menu_tool_button_get_type) IA__gtk_menu_tool_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_menu_tool_button_get_type IA__gtk_menu_tool_button_get_type

extern __typeof (gtk_menu_tool_button_new) IA__gtk_menu_tool_button_new __attribute((visibility("hidden")));
#define gtk_menu_tool_button_new IA__gtk_menu_tool_button_new

extern __typeof (gtk_menu_tool_button_new_from_stock) IA__gtk_menu_tool_button_new_from_stock __attribute((visibility("hidden")));
#define gtk_menu_tool_button_new_from_stock IA__gtk_menu_tool_button_new_from_stock

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_menu_tool_button_set_arrow_tooltip) IA__gtk_menu_tool_button_set_arrow_tooltip __attribute((visibility("hidden")));
#define gtk_menu_tool_button_set_arrow_tooltip IA__gtk_menu_tool_button_set_arrow_tooltip

#endif
extern __typeof (gtk_menu_tool_button_set_arrow_tooltip_markup) IA__gtk_menu_tool_button_set_arrow_tooltip_markup __attribute((visibility("hidden")));
#define gtk_menu_tool_button_set_arrow_tooltip_markup IA__gtk_menu_tool_button_set_arrow_tooltip_markup

extern __typeof (gtk_menu_tool_button_set_arrow_tooltip_text) IA__gtk_menu_tool_button_set_arrow_tooltip_text __attribute((visibility("hidden")));
#define gtk_menu_tool_button_set_arrow_tooltip_text IA__gtk_menu_tool_button_set_arrow_tooltip_text

extern __typeof (gtk_menu_tool_button_set_menu) IA__gtk_menu_tool_button_set_menu __attribute((visibility("hidden")));
#define gtk_menu_tool_button_set_menu IA__gtk_menu_tool_button_set_menu

#endif
#endif
#if IN_HEADER(__GTK_MESSAGE_DIALOG_H__)
#if IN_FILE(__GTK_MESSAGE_DIALOG_C__)
extern __typeof (gtk_message_dialog_format_secondary_markup) IA__gtk_message_dialog_format_secondary_markup __attribute((visibility("hidden"))) G_GNUC_PRINTF(2,3);
#define gtk_message_dialog_format_secondary_markup IA__gtk_message_dialog_format_secondary_markup

extern __typeof (gtk_message_dialog_format_secondary_text) IA__gtk_message_dialog_format_secondary_text __attribute((visibility("hidden"))) G_GNUC_PRINTF(2,3);
#define gtk_message_dialog_format_secondary_text IA__gtk_message_dialog_format_secondary_text

extern __typeof (gtk_message_dialog_get_message_area) IA__gtk_message_dialog_get_message_area __attribute((visibility("hidden")));
#define gtk_message_dialog_get_message_area IA__gtk_message_dialog_get_message_area

extern __typeof (gtk_message_dialog_get_type) IA__gtk_message_dialog_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_message_dialog_get_type IA__gtk_message_dialog_get_type

extern __typeof (gtk_message_dialog_new) IA__gtk_message_dialog_new __attribute((visibility("hidden"))) G_GNUC_PRINTF(5,6);
#define gtk_message_dialog_new IA__gtk_message_dialog_new

extern __typeof (gtk_message_dialog_new_with_markup) IA__gtk_message_dialog_new_with_markup __attribute((visibility("hidden"))) G_GNUC_PRINTF(5,6);
#define gtk_message_dialog_new_with_markup IA__gtk_message_dialog_new_with_markup

extern __typeof (gtk_message_dialog_set_markup) IA__gtk_message_dialog_set_markup __attribute((visibility("hidden")));
#define gtk_message_dialog_set_markup IA__gtk_message_dialog_set_markup

extern __typeof (gtk_message_dialog_set_image) IA__gtk_message_dialog_set_image __attribute((visibility("hidden")));
#define gtk_message_dialog_set_image IA__gtk_message_dialog_set_image

extern __typeof (gtk_message_dialog_get_image) IA__gtk_message_dialog_get_image __attribute((visibility("hidden")));
#define gtk_message_dialog_get_image IA__gtk_message_dialog_get_image

#endif
#endif
#if IN_HEADER(__GTK_MISC_H__)
#if IN_FILE(__GTK_MISC_C__)
extern __typeof (gtk_misc_get_alignment) IA__gtk_misc_get_alignment __attribute((visibility("hidden")));
#define gtk_misc_get_alignment IA__gtk_misc_get_alignment

extern __typeof (gtk_misc_get_padding) IA__gtk_misc_get_padding __attribute((visibility("hidden")));
#define gtk_misc_get_padding IA__gtk_misc_get_padding

extern __typeof (gtk_misc_get_type) IA__gtk_misc_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_misc_get_type IA__gtk_misc_get_type

extern __typeof (gtk_misc_set_alignment) IA__gtk_misc_set_alignment __attribute((visibility("hidden")));
#define gtk_misc_set_alignment IA__gtk_misc_set_alignment

extern __typeof (gtk_misc_set_padding) IA__gtk_misc_set_padding __attribute((visibility("hidden")));
#define gtk_misc_set_padding IA__gtk_misc_set_padding

#endif
#endif
#if IN_HEADER(__GTK_MOUNT_OPERATION_H__)
#if IN_FILE(__GTK_MOUNT_OPERATION_C__)
extern __typeof (gtk_mount_operation_get_type) IA__gtk_mount_operation_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_mount_operation_get_type IA__gtk_mount_operation_get_type

extern __typeof (gtk_mount_operation_new) IA__gtk_mount_operation_new __attribute((visibility("hidden")));
#define gtk_mount_operation_new IA__gtk_mount_operation_new

extern __typeof (gtk_mount_operation_is_showing) IA__gtk_mount_operation_is_showing __attribute((visibility("hidden")));
#define gtk_mount_operation_is_showing IA__gtk_mount_operation_is_showing

extern __typeof (gtk_mount_operation_set_parent) IA__gtk_mount_operation_set_parent __attribute((visibility("hidden")));
#define gtk_mount_operation_set_parent IA__gtk_mount_operation_set_parent

extern __typeof (gtk_mount_operation_get_parent) IA__gtk_mount_operation_get_parent __attribute((visibility("hidden")));
#define gtk_mount_operation_get_parent IA__gtk_mount_operation_get_parent

extern __typeof (gtk_mount_operation_set_screen) IA__gtk_mount_operation_set_screen __attribute((visibility("hidden")));
#define gtk_mount_operation_set_screen IA__gtk_mount_operation_set_screen

extern __typeof (gtk_mount_operation_get_screen) IA__gtk_mount_operation_get_screen __attribute((visibility("hidden")));
#define gtk_mount_operation_get_screen IA__gtk_mount_operation_get_screen

#endif
#endif
#if IN_HEADER(__GTK_NOTEBOOK_H__)
#if IN_FILE(__GTK_NOTEBOOK_C__)
extern __typeof (gtk_notebook_append_page) IA__gtk_notebook_append_page __attribute((visibility("hidden")));
#define gtk_notebook_append_page IA__gtk_notebook_append_page

extern __typeof (gtk_notebook_append_page_menu) IA__gtk_notebook_append_page_menu __attribute((visibility("hidden")));
#define gtk_notebook_append_page_menu IA__gtk_notebook_append_page_menu

extern __typeof (gtk_notebook_get_current_page) IA__gtk_notebook_get_current_page __attribute((visibility("hidden")));
#define gtk_notebook_get_current_page IA__gtk_notebook_get_current_page

extern __typeof (gtk_notebook_get_menu_label) IA__gtk_notebook_get_menu_label __attribute((visibility("hidden")));
#define gtk_notebook_get_menu_label IA__gtk_notebook_get_menu_label

extern __typeof (gtk_notebook_get_menu_label_text) IA__gtk_notebook_get_menu_label_text __attribute((visibility("hidden")));
#define gtk_notebook_get_menu_label_text IA__gtk_notebook_get_menu_label_text

extern __typeof (gtk_notebook_get_n_pages) IA__gtk_notebook_get_n_pages __attribute((visibility("hidden")));
#define gtk_notebook_get_n_pages IA__gtk_notebook_get_n_pages

extern __typeof (gtk_notebook_get_nth_page) IA__gtk_notebook_get_nth_page __attribute((visibility("hidden")));
#define gtk_notebook_get_nth_page IA__gtk_notebook_get_nth_page

extern __typeof (gtk_notebook_get_scrollable) IA__gtk_notebook_get_scrollable __attribute((visibility("hidden")));
#define gtk_notebook_get_scrollable IA__gtk_notebook_get_scrollable

extern __typeof (gtk_notebook_get_show_border) IA__gtk_notebook_get_show_border __attribute((visibility("hidden")));
#define gtk_notebook_get_show_border IA__gtk_notebook_get_show_border

extern __typeof (gtk_notebook_get_show_tabs) IA__gtk_notebook_get_show_tabs __attribute((visibility("hidden")));
#define gtk_notebook_get_show_tabs IA__gtk_notebook_get_show_tabs

extern __typeof (gtk_notebook_get_tab_label) IA__gtk_notebook_get_tab_label __attribute((visibility("hidden")));
#define gtk_notebook_get_tab_label IA__gtk_notebook_get_tab_label

extern __typeof (gtk_notebook_get_tab_label_text) IA__gtk_notebook_get_tab_label_text __attribute((visibility("hidden")));
#define gtk_notebook_get_tab_label_text IA__gtk_notebook_get_tab_label_text

extern __typeof (gtk_notebook_get_tab_pos) IA__gtk_notebook_get_tab_pos __attribute((visibility("hidden")));
#define gtk_notebook_get_tab_pos IA__gtk_notebook_get_tab_pos

extern __typeof (gtk_notebook_get_type) IA__gtk_notebook_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_notebook_get_type IA__gtk_notebook_get_type

extern __typeof (gtk_notebook_insert_page) IA__gtk_notebook_insert_page __attribute((visibility("hidden")));
#define gtk_notebook_insert_page IA__gtk_notebook_insert_page

extern __typeof (gtk_notebook_insert_page_menu) IA__gtk_notebook_insert_page_menu __attribute((visibility("hidden")));
#define gtk_notebook_insert_page_menu IA__gtk_notebook_insert_page_menu

extern __typeof (gtk_notebook_new) IA__gtk_notebook_new __attribute((visibility("hidden")));
#define gtk_notebook_new IA__gtk_notebook_new

extern __typeof (gtk_notebook_next_page) IA__gtk_notebook_next_page __attribute((visibility("hidden")));
#define gtk_notebook_next_page IA__gtk_notebook_next_page

extern __typeof (gtk_notebook_page_num) IA__gtk_notebook_page_num __attribute((visibility("hidden")));
#define gtk_notebook_page_num IA__gtk_notebook_page_num

extern __typeof (gtk_notebook_popup_disable) IA__gtk_notebook_popup_disable __attribute((visibility("hidden")));
#define gtk_notebook_popup_disable IA__gtk_notebook_popup_disable

extern __typeof (gtk_notebook_popup_enable) IA__gtk_notebook_popup_enable __attribute((visibility("hidden")));
#define gtk_notebook_popup_enable IA__gtk_notebook_popup_enable

extern __typeof (gtk_notebook_prepend_page) IA__gtk_notebook_prepend_page __attribute((visibility("hidden")));
#define gtk_notebook_prepend_page IA__gtk_notebook_prepend_page

extern __typeof (gtk_notebook_prepend_page_menu) IA__gtk_notebook_prepend_page_menu __attribute((visibility("hidden")));
#define gtk_notebook_prepend_page_menu IA__gtk_notebook_prepend_page_menu

extern __typeof (gtk_notebook_prev_page) IA__gtk_notebook_prev_page __attribute((visibility("hidden")));
#define gtk_notebook_prev_page IA__gtk_notebook_prev_page

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_notebook_query_tab_label_packing) IA__gtk_notebook_query_tab_label_packing __attribute((visibility("hidden")));
#define gtk_notebook_query_tab_label_packing IA__gtk_notebook_query_tab_label_packing

#endif
extern __typeof (gtk_notebook_remove_page) IA__gtk_notebook_remove_page __attribute((visibility("hidden")));
#define gtk_notebook_remove_page IA__gtk_notebook_remove_page

extern __typeof (gtk_notebook_reorder_child) IA__gtk_notebook_reorder_child __attribute((visibility("hidden")));
#define gtk_notebook_reorder_child IA__gtk_notebook_reorder_child

extern __typeof (gtk_notebook_set_current_page) IA__gtk_notebook_set_current_page __attribute((visibility("hidden")));
#define gtk_notebook_set_current_page IA__gtk_notebook_set_current_page

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_notebook_set_homogeneous_tabs) IA__gtk_notebook_set_homogeneous_tabs __attribute((visibility("hidden")));
#define gtk_notebook_set_homogeneous_tabs IA__gtk_notebook_set_homogeneous_tabs

extern __typeof (gtk_notebook_set_tab_border) IA__gtk_notebook_set_tab_border __attribute((visibility("hidden")));
#define gtk_notebook_set_tab_border IA__gtk_notebook_set_tab_border

extern __typeof (gtk_notebook_set_tab_hborder) IA__gtk_notebook_set_tab_hborder __attribute((visibility("hidden")));
#define gtk_notebook_set_tab_hborder IA__gtk_notebook_set_tab_hborder

extern __typeof (gtk_notebook_set_tab_vborder) IA__gtk_notebook_set_tab_vborder __attribute((visibility("hidden")));
#define gtk_notebook_set_tab_vborder IA__gtk_notebook_set_tab_vborder

#endif
extern __typeof (gtk_notebook_set_menu_label) IA__gtk_notebook_set_menu_label __attribute((visibility("hidden")));
#define gtk_notebook_set_menu_label IA__gtk_notebook_set_menu_label

extern __typeof (gtk_notebook_set_menu_label_text) IA__gtk_notebook_set_menu_label_text __attribute((visibility("hidden")));
#define gtk_notebook_set_menu_label_text IA__gtk_notebook_set_menu_label_text

extern __typeof (gtk_notebook_set_scrollable) IA__gtk_notebook_set_scrollable __attribute((visibility("hidden")));
#define gtk_notebook_set_scrollable IA__gtk_notebook_set_scrollable

extern __typeof (gtk_notebook_set_show_border) IA__gtk_notebook_set_show_border __attribute((visibility("hidden")));
#define gtk_notebook_set_show_border IA__gtk_notebook_set_show_border

extern __typeof (gtk_notebook_set_show_tabs) IA__gtk_notebook_set_show_tabs __attribute((visibility("hidden")));
#define gtk_notebook_set_show_tabs IA__gtk_notebook_set_show_tabs

extern __typeof (gtk_notebook_set_tab_label) IA__gtk_notebook_set_tab_label __attribute((visibility("hidden")));
#define gtk_notebook_set_tab_label IA__gtk_notebook_set_tab_label

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_notebook_set_tab_label_packing) IA__gtk_notebook_set_tab_label_packing __attribute((visibility("hidden")));
#define gtk_notebook_set_tab_label_packing IA__gtk_notebook_set_tab_label_packing

#endif
extern __typeof (gtk_notebook_set_tab_label_text) IA__gtk_notebook_set_tab_label_text __attribute((visibility("hidden")));
#define gtk_notebook_set_tab_label_text IA__gtk_notebook_set_tab_label_text

extern __typeof (gtk_notebook_set_tab_pos) IA__gtk_notebook_set_tab_pos __attribute((visibility("hidden")));
#define gtk_notebook_set_tab_pos IA__gtk_notebook_set_tab_pos

extern __typeof (gtk_notebook_get_tab_hborder) IA__gtk_notebook_get_tab_hborder __attribute((visibility("hidden")));
#define gtk_notebook_get_tab_hborder IA__gtk_notebook_get_tab_hborder

extern __typeof (gtk_notebook_get_tab_vborder) IA__gtk_notebook_get_tab_vborder __attribute((visibility("hidden")));
#define gtk_notebook_get_tab_vborder IA__gtk_notebook_get_tab_vborder

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_notebook_set_window_creation_hook) IA__gtk_notebook_set_window_creation_hook __attribute((visibility("hidden")));
#define gtk_notebook_set_window_creation_hook IA__gtk_notebook_set_window_creation_hook

extern __typeof (gtk_notebook_get_group_id) IA__gtk_notebook_get_group_id __attribute((visibility("hidden")));
#define gtk_notebook_get_group_id IA__gtk_notebook_get_group_id

extern __typeof (gtk_notebook_set_group_id) IA__gtk_notebook_set_group_id __attribute((visibility("hidden")));
#define gtk_notebook_set_group_id IA__gtk_notebook_set_group_id

extern __typeof (gtk_notebook_get_group) IA__gtk_notebook_get_group __attribute((visibility("hidden")));
#define gtk_notebook_get_group IA__gtk_notebook_get_group

extern __typeof (gtk_notebook_set_group) IA__gtk_notebook_set_group __attribute((visibility("hidden")));
#define gtk_notebook_set_group IA__gtk_notebook_set_group

#endif
extern __typeof (gtk_notebook_get_group_name) IA__gtk_notebook_get_group_name __attribute((visibility("hidden")));
#define gtk_notebook_get_group_name IA__gtk_notebook_get_group_name

extern __typeof (gtk_notebook_set_group_name) IA__gtk_notebook_set_group_name __attribute((visibility("hidden")));
#define gtk_notebook_set_group_name IA__gtk_notebook_set_group_name

extern __typeof (gtk_notebook_get_tab_reorderable) IA__gtk_notebook_get_tab_reorderable __attribute((visibility("hidden")));
#define gtk_notebook_get_tab_reorderable IA__gtk_notebook_get_tab_reorderable

extern __typeof (gtk_notebook_set_tab_reorderable) IA__gtk_notebook_set_tab_reorderable __attribute((visibility("hidden")));
#define gtk_notebook_set_tab_reorderable IA__gtk_notebook_set_tab_reorderable

extern __typeof (gtk_notebook_get_tab_detachable) IA__gtk_notebook_get_tab_detachable __attribute((visibility("hidden")));
#define gtk_notebook_get_tab_detachable IA__gtk_notebook_get_tab_detachable

extern __typeof (gtk_notebook_set_tab_detachable) IA__gtk_notebook_set_tab_detachable __attribute((visibility("hidden")));
#define gtk_notebook_set_tab_detachable IA__gtk_notebook_set_tab_detachable

extern __typeof (gtk_notebook_get_action_widget) IA__gtk_notebook_get_action_widget __attribute((visibility("hidden")));
#define gtk_notebook_get_action_widget IA__gtk_notebook_get_action_widget

extern __typeof (gtk_notebook_set_action_widget) IA__gtk_notebook_set_action_widget __attribute((visibility("hidden")));
#define gtk_notebook_set_action_widget IA__gtk_notebook_set_action_widget

#endif
#endif
#if IN_HEADER(__GTK_OBJECT_H__)
#if IN_FILE(__GTK_OBJECT_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_object_add_arg_type) IA__gtk_object_add_arg_type __attribute((visibility("hidden")));
#define gtk_object_add_arg_type IA__gtk_object_add_arg_type

extern __typeof (gtk_object_get) IA__gtk_object_get __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_object_get IA__gtk_object_get

extern __typeof (gtk_object_set) IA__gtk_object_set __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_object_set IA__gtk_object_set

extern __typeof (gtk_object_get_data) IA__gtk_object_get_data __attribute((visibility("hidden")));
#define gtk_object_get_data IA__gtk_object_get_data

extern __typeof (gtk_object_get_data_by_id) IA__gtk_object_get_data_by_id __attribute((visibility("hidden")));
#define gtk_object_get_data_by_id IA__gtk_object_get_data_by_id

extern __typeof (gtk_object_get_user_data) IA__gtk_object_get_user_data __attribute((visibility("hidden")));
#define gtk_object_get_user_data IA__gtk_object_get_user_data

extern __typeof (gtk_object_new) IA__gtk_object_new __attribute((visibility("hidden")));
#define gtk_object_new IA__gtk_object_new

extern __typeof (gtk_object_remove_data) IA__gtk_object_remove_data __attribute((visibility("hidden")));
#define gtk_object_remove_data IA__gtk_object_remove_data

extern __typeof (gtk_object_remove_data_by_id) IA__gtk_object_remove_data_by_id __attribute((visibility("hidden")));
#define gtk_object_remove_data_by_id IA__gtk_object_remove_data_by_id

extern __typeof (gtk_object_remove_no_notify) IA__gtk_object_remove_no_notify __attribute((visibility("hidden")));
#define gtk_object_remove_no_notify IA__gtk_object_remove_no_notify

extern __typeof (gtk_object_remove_no_notify_by_id) IA__gtk_object_remove_no_notify_by_id __attribute((visibility("hidden")));
#define gtk_object_remove_no_notify_by_id IA__gtk_object_remove_no_notify_by_id

extern __typeof (gtk_object_ref) IA__gtk_object_ref __attribute((visibility("hidden")));
#define gtk_object_ref IA__gtk_object_ref

extern __typeof (gtk_object_unref) IA__gtk_object_unref __attribute((visibility("hidden")));
#define gtk_object_unref IA__gtk_object_unref

extern __typeof (gtk_object_weakref) IA__gtk_object_weakref __attribute((visibility("hidden")));
#define gtk_object_weakref IA__gtk_object_weakref

extern __typeof (gtk_object_weakunref) IA__gtk_object_weakunref __attribute((visibility("hidden")));
#define gtk_object_weakunref IA__gtk_object_weakunref

extern __typeof (gtk_object_set_data) IA__gtk_object_set_data __attribute((visibility("hidden")));
#define gtk_object_set_data IA__gtk_object_set_data

extern __typeof (gtk_object_set_data_by_id) IA__gtk_object_set_data_by_id __attribute((visibility("hidden")));
#define gtk_object_set_data_by_id IA__gtk_object_set_data_by_id

extern __typeof (gtk_object_set_data_by_id_full) IA__gtk_object_set_data_by_id_full __attribute((visibility("hidden")));
#define gtk_object_set_data_by_id_full IA__gtk_object_set_data_by_id_full

extern __typeof (gtk_object_set_data_full) IA__gtk_object_set_data_full __attribute((visibility("hidden")));
#define gtk_object_set_data_full IA__gtk_object_set_data_full

extern __typeof (gtk_object_set_user_data) IA__gtk_object_set_user_data __attribute((visibility("hidden")));
#define gtk_object_set_user_data IA__gtk_object_set_user_data

extern __typeof (gtk_object_sink) IA__gtk_object_sink __attribute((visibility("hidden")));
#define gtk_object_sink IA__gtk_object_sink

#endif
extern __typeof (gtk_object_destroy) IA__gtk_object_destroy __attribute((visibility("hidden")));
#define gtk_object_destroy IA__gtk_object_destroy

extern __typeof (gtk_object_get_type) IA__gtk_object_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_object_get_type IA__gtk_object_get_type

#endif
#endif
#if IN_HEADER(__GTK_OFFSCREEN_WINDOW_H__)
#if IN_FILE(__GTK_OFFSCREEN_WINDOW_C__)
extern __typeof (gtk_offscreen_window_get_type) IA__gtk_offscreen_window_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_offscreen_window_get_type IA__gtk_offscreen_window_get_type

extern __typeof (gtk_offscreen_window_new) IA__gtk_offscreen_window_new __attribute((visibility("hidden")));
#define gtk_offscreen_window_new IA__gtk_offscreen_window_new

extern __typeof (gtk_offscreen_window_get_pixmap) IA__gtk_offscreen_window_get_pixmap __attribute((visibility("hidden")));
#define gtk_offscreen_window_get_pixmap IA__gtk_offscreen_window_get_pixmap

extern __typeof (gtk_offscreen_window_get_pixbuf) IA__gtk_offscreen_window_get_pixbuf __attribute((visibility("hidden")));
#define gtk_offscreen_window_get_pixbuf IA__gtk_offscreen_window_get_pixbuf

#endif
#endif
#if IN_HEADER(__GTK_ORIENTABLE_H__)
#if IN_FILE(__GTK_ORIENTABLE_C__)
extern __typeof (gtk_orientable_get_orientation) IA__gtk_orientable_get_orientation __attribute((visibility("hidden")));
#define gtk_orientable_get_orientation IA__gtk_orientable_get_orientation

extern __typeof (gtk_orientable_get_type) IA__gtk_orientable_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_orientable_get_type IA__gtk_orientable_get_type

extern __typeof (gtk_orientable_set_orientation) IA__gtk_orientable_set_orientation __attribute((visibility("hidden")));
#define gtk_orientable_set_orientation IA__gtk_orientable_set_orientation

#endif
#endif
#if IN_HEADER(__GTK_OLD_EDITABLE_H__)
#if IN_FILE(__GTK_OLD_EDITABLE_C__)
extern __typeof (gtk_old_editable_changed) IA__gtk_old_editable_changed __attribute((visibility("hidden")));
#define gtk_old_editable_changed IA__gtk_old_editable_changed

extern __typeof (gtk_old_editable_claim_selection) IA__gtk_old_editable_claim_selection __attribute((visibility("hidden")));
#define gtk_old_editable_claim_selection IA__gtk_old_editable_claim_selection

extern __typeof (gtk_old_editable_get_type) IA__gtk_old_editable_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_old_editable_get_type IA__gtk_old_editable_get_type

#endif
#endif
#if IN_HEADER(__GTK_OPTION_MENU_H__)
#if IN_FILE(__GTK_OPTION_MENU_C__)
extern __typeof (gtk_option_menu_get_history) IA__gtk_option_menu_get_history __attribute((visibility("hidden")));
#define gtk_option_menu_get_history IA__gtk_option_menu_get_history

extern __typeof (gtk_option_menu_get_menu) IA__gtk_option_menu_get_menu __attribute((visibility("hidden")));
#define gtk_option_menu_get_menu IA__gtk_option_menu_get_menu

extern __typeof (gtk_option_menu_get_type) IA__gtk_option_menu_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_option_menu_get_type IA__gtk_option_menu_get_type

extern __typeof (gtk_option_menu_new) IA__gtk_option_menu_new __attribute((visibility("hidden")));
#define gtk_option_menu_new IA__gtk_option_menu_new

extern __typeof (gtk_option_menu_remove_menu) IA__gtk_option_menu_remove_menu __attribute((visibility("hidden")));
#define gtk_option_menu_remove_menu IA__gtk_option_menu_remove_menu

extern __typeof (gtk_option_menu_set_history) IA__gtk_option_menu_set_history __attribute((visibility("hidden")));
#define gtk_option_menu_set_history IA__gtk_option_menu_set_history

extern __typeof (gtk_option_menu_set_menu) IA__gtk_option_menu_set_menu __attribute((visibility("hidden")));
#define gtk_option_menu_set_menu IA__gtk_option_menu_set_menu

#endif
#endif
#if IN_HEADER(__GTK_PANED_H__)
#if IN_FILE(__GTK_PANED_C__)
extern __typeof (gtk_paned_add1) IA__gtk_paned_add1 __attribute((visibility("hidden")));
#define gtk_paned_add1 IA__gtk_paned_add1

extern __typeof (gtk_paned_add2) IA__gtk_paned_add2 __attribute((visibility("hidden")));
#define gtk_paned_add2 IA__gtk_paned_add2

extern __typeof (gtk_paned_get_child1) IA__gtk_paned_get_child1 __attribute((visibility("hidden")));
#define gtk_paned_get_child1 IA__gtk_paned_get_child1

extern __typeof (gtk_paned_get_child2) IA__gtk_paned_get_child2 __attribute((visibility("hidden")));
#define gtk_paned_get_child2 IA__gtk_paned_get_child2

extern __typeof (gtk_paned_get_handle_window) IA__gtk_paned_get_handle_window __attribute((visibility("hidden")));
#define gtk_paned_get_handle_window IA__gtk_paned_get_handle_window

extern __typeof (gtk_paned_get_position) IA__gtk_paned_get_position __attribute((visibility("hidden")));
#define gtk_paned_get_position IA__gtk_paned_get_position

extern __typeof (gtk_paned_get_type) IA__gtk_paned_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_paned_get_type IA__gtk_paned_get_type

extern __typeof (gtk_paned_pack1) IA__gtk_paned_pack1 __attribute((visibility("hidden")));
#define gtk_paned_pack1 IA__gtk_paned_pack1

extern __typeof (gtk_paned_pack2) IA__gtk_paned_pack2 __attribute((visibility("hidden")));
#define gtk_paned_pack2 IA__gtk_paned_pack2

extern __typeof (gtk_paned_set_position) IA__gtk_paned_set_position __attribute((visibility("hidden")));
#define gtk_paned_set_position IA__gtk_paned_set_position

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_paned_compute_position) IA__gtk_paned_compute_position __attribute((visibility("hidden")));
#define gtk_paned_compute_position IA__gtk_paned_compute_position

#endif
#endif
#endif
#if IN_HEADER(__GTK_PATH_BAR_H__)
#if IN_FILE(__GTK_PATH_BAR_C__)
extern __typeof (gtk_path_bar_get_type) IA__gtk_path_bar_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_path_bar_get_type IA__gtk_path_bar_get_type

#endif
#endif
#if IN_HEADER(__GTK_PIXMAP_H__)
#if IN_FILE(__GTK_PIXMAP_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_pixmap_get) IA__gtk_pixmap_get __attribute((visibility("hidden")));
#define gtk_pixmap_get IA__gtk_pixmap_get

extern __typeof (gtk_pixmap_get_type) IA__gtk_pixmap_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_pixmap_get_type IA__gtk_pixmap_get_type

extern __typeof (gtk_pixmap_new) IA__gtk_pixmap_new __attribute((visibility("hidden")));
#define gtk_pixmap_new IA__gtk_pixmap_new

extern __typeof (gtk_pixmap_set) IA__gtk_pixmap_set __attribute((visibility("hidden")));
#define gtk_pixmap_set IA__gtk_pixmap_set

extern __typeof (gtk_pixmap_set_build_insensitive) IA__gtk_pixmap_set_build_insensitive __attribute((visibility("hidden")));
#define gtk_pixmap_set_build_insensitive IA__gtk_pixmap_set_build_insensitive

#endif
#endif
#endif
#if IN_HEADER(__GTK_PLUG_H__)
#if IN_FILE(__GTK_PLUG_C__)
extern __typeof (gtk_plug_construct) IA__gtk_plug_construct __attribute((visibility("hidden")));
#define gtk_plug_construct IA__gtk_plug_construct

extern __typeof (gtk_plug_construct_for_display) IA__gtk_plug_construct_for_display __attribute((visibility("hidden")));
#define gtk_plug_construct_for_display IA__gtk_plug_construct_for_display

extern __typeof (gtk_plug_get_id) IA__gtk_plug_get_id __attribute((visibility("hidden")));
#define gtk_plug_get_id IA__gtk_plug_get_id

extern __typeof (gtk_plug_get_embedded) IA__gtk_plug_get_embedded __attribute((visibility("hidden")));
#define gtk_plug_get_embedded IA__gtk_plug_get_embedded

extern __typeof (gtk_plug_get_socket_window) IA__gtk_plug_get_socket_window __attribute((visibility("hidden")));
#define gtk_plug_get_socket_window IA__gtk_plug_get_socket_window

extern __typeof (gtk_plug_get_type) IA__gtk_plug_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_plug_get_type IA__gtk_plug_get_type

extern __typeof (gtk_plug_new) IA__gtk_plug_new __attribute((visibility("hidden")));
#define gtk_plug_new IA__gtk_plug_new

extern __typeof (gtk_plug_new_for_display) IA__gtk_plug_new_for_display __attribute((visibility("hidden")));
#define gtk_plug_new_for_display IA__gtk_plug_new_for_display

#endif
#endif
#if IN_HEADER(__GTK_PREVIEW_H__)
#if IN_FILE(__GTK_PREVIEW_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_preview_draw_row) IA__gtk_preview_draw_row __attribute((visibility("hidden")));
#define gtk_preview_draw_row IA__gtk_preview_draw_row

extern __typeof (gtk_preview_get_cmap) IA__gtk_preview_get_cmap __attribute((visibility("hidden")));
#define gtk_preview_get_cmap IA__gtk_preview_get_cmap

extern __typeof (gtk_preview_get_info) IA__gtk_preview_get_info __attribute((visibility("hidden")));
#define gtk_preview_get_info IA__gtk_preview_get_info

extern __typeof (gtk_preview_get_type) IA__gtk_preview_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_preview_get_type IA__gtk_preview_get_type

extern __typeof (gtk_preview_get_visual) IA__gtk_preview_get_visual __attribute((visibility("hidden")));
#define gtk_preview_get_visual IA__gtk_preview_get_visual

extern __typeof (gtk_preview_new) IA__gtk_preview_new __attribute((visibility("hidden")));
#define gtk_preview_new IA__gtk_preview_new

extern __typeof (gtk_preview_put) IA__gtk_preview_put __attribute((visibility("hidden")));
#define gtk_preview_put IA__gtk_preview_put

extern __typeof (gtk_preview_reset) IA__gtk_preview_reset __attribute((visibility("hidden")));
#define gtk_preview_reset IA__gtk_preview_reset

extern __typeof (gtk_preview_set_color_cube) IA__gtk_preview_set_color_cube __attribute((visibility("hidden")));
#define gtk_preview_set_color_cube IA__gtk_preview_set_color_cube

extern __typeof (gtk_preview_set_dither) IA__gtk_preview_set_dither __attribute((visibility("hidden")));
#define gtk_preview_set_dither IA__gtk_preview_set_dither

extern __typeof (gtk_preview_set_expand) IA__gtk_preview_set_expand __attribute((visibility("hidden")));
#define gtk_preview_set_expand IA__gtk_preview_set_expand

extern __typeof (gtk_preview_set_gamma) IA__gtk_preview_set_gamma __attribute((visibility("hidden")));
#define gtk_preview_set_gamma IA__gtk_preview_set_gamma

extern __typeof (gtk_preview_set_install_cmap) IA__gtk_preview_set_install_cmap __attribute((visibility("hidden")));
#define gtk_preview_set_install_cmap IA__gtk_preview_set_install_cmap

extern __typeof (gtk_preview_set_reserved) IA__gtk_preview_set_reserved __attribute((visibility("hidden")));
#define gtk_preview_set_reserved IA__gtk_preview_set_reserved

extern __typeof (gtk_preview_size) IA__gtk_preview_size __attribute((visibility("hidden")));
#define gtk_preview_size IA__gtk_preview_size

extern __typeof (gtk_preview_uninit) IA__gtk_preview_uninit __attribute((visibility("hidden")));
#define gtk_preview_uninit IA__gtk_preview_uninit

#endif
#endif
#endif
#if IN_HEADER(__GTK_PAGE_SETUP_H__)
#if IN_FILE(__GTK_PAGE_SETUP_C__)
extern __typeof (gtk_page_setup_get_type) IA__gtk_page_setup_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_page_setup_get_type IA__gtk_page_setup_get_type

extern __typeof (gtk_page_setup_new) IA__gtk_page_setup_new __attribute((visibility("hidden")));
#define gtk_page_setup_new IA__gtk_page_setup_new

extern __typeof (gtk_page_setup_new_from_file) IA__gtk_page_setup_new_from_file __attribute((visibility("hidden")));
#define gtk_page_setup_new_from_file IA__gtk_page_setup_new_from_file

extern __typeof (gtk_page_setup_new_from_key_file) IA__gtk_page_setup_new_from_key_file __attribute((visibility("hidden")));
#define gtk_page_setup_new_from_key_file IA__gtk_page_setup_new_from_key_file

extern __typeof (gtk_page_setup_load_file) IA__gtk_page_setup_load_file __attribute((visibility("hidden")));
#define gtk_page_setup_load_file IA__gtk_page_setup_load_file

extern __typeof (gtk_page_setup_load_key_file) IA__gtk_page_setup_load_key_file __attribute((visibility("hidden")));
#define gtk_page_setup_load_key_file IA__gtk_page_setup_load_key_file

extern __typeof (gtk_page_setup_to_file) IA__gtk_page_setup_to_file __attribute((visibility("hidden")));
#define gtk_page_setup_to_file IA__gtk_page_setup_to_file

extern __typeof (gtk_page_setup_to_key_file) IA__gtk_page_setup_to_key_file __attribute((visibility("hidden")));
#define gtk_page_setup_to_key_file IA__gtk_page_setup_to_key_file

extern __typeof (gtk_page_setup_copy) IA__gtk_page_setup_copy __attribute((visibility("hidden")));
#define gtk_page_setup_copy IA__gtk_page_setup_copy

extern __typeof (gtk_page_setup_get_orientation) IA__gtk_page_setup_get_orientation __attribute((visibility("hidden")));
#define gtk_page_setup_get_orientation IA__gtk_page_setup_get_orientation

extern __typeof (gtk_page_setup_set_orientation) IA__gtk_page_setup_set_orientation __attribute((visibility("hidden")));
#define gtk_page_setup_set_orientation IA__gtk_page_setup_set_orientation

extern __typeof (gtk_page_setup_get_paper_size) IA__gtk_page_setup_get_paper_size __attribute((visibility("hidden")));
#define gtk_page_setup_get_paper_size IA__gtk_page_setup_get_paper_size

extern __typeof (gtk_page_setup_set_paper_size) IA__gtk_page_setup_set_paper_size __attribute((visibility("hidden")));
#define gtk_page_setup_set_paper_size IA__gtk_page_setup_set_paper_size

extern __typeof (gtk_page_setup_get_top_margin) IA__gtk_page_setup_get_top_margin __attribute((visibility("hidden")));
#define gtk_page_setup_get_top_margin IA__gtk_page_setup_get_top_margin

extern __typeof (gtk_page_setup_set_top_margin) IA__gtk_page_setup_set_top_margin __attribute((visibility("hidden")));
#define gtk_page_setup_set_top_margin IA__gtk_page_setup_set_top_margin

extern __typeof (gtk_page_setup_get_bottom_margin) IA__gtk_page_setup_get_bottom_margin __attribute((visibility("hidden")));
#define gtk_page_setup_get_bottom_margin IA__gtk_page_setup_get_bottom_margin

extern __typeof (gtk_page_setup_set_bottom_margin) IA__gtk_page_setup_set_bottom_margin __attribute((visibility("hidden")));
#define gtk_page_setup_set_bottom_margin IA__gtk_page_setup_set_bottom_margin

extern __typeof (gtk_page_setup_get_left_margin) IA__gtk_page_setup_get_left_margin __attribute((visibility("hidden")));
#define gtk_page_setup_get_left_margin IA__gtk_page_setup_get_left_margin

extern __typeof (gtk_page_setup_set_left_margin) IA__gtk_page_setup_set_left_margin __attribute((visibility("hidden")));
#define gtk_page_setup_set_left_margin IA__gtk_page_setup_set_left_margin

extern __typeof (gtk_page_setup_get_right_margin) IA__gtk_page_setup_get_right_margin __attribute((visibility("hidden")));
#define gtk_page_setup_get_right_margin IA__gtk_page_setup_get_right_margin

extern __typeof (gtk_page_setup_set_right_margin) IA__gtk_page_setup_set_right_margin __attribute((visibility("hidden")));
#define gtk_page_setup_set_right_margin IA__gtk_page_setup_set_right_margin

extern __typeof (gtk_page_setup_set_paper_size_and_default_margins) IA__gtk_page_setup_set_paper_size_and_default_margins __attribute((visibility("hidden")));
#define gtk_page_setup_set_paper_size_and_default_margins IA__gtk_page_setup_set_paper_size_and_default_margins

extern __typeof (gtk_page_setup_get_paper_width) IA__gtk_page_setup_get_paper_width __attribute((visibility("hidden")));
#define gtk_page_setup_get_paper_width IA__gtk_page_setup_get_paper_width

extern __typeof (gtk_page_setup_get_paper_height) IA__gtk_page_setup_get_paper_height __attribute((visibility("hidden")));
#define gtk_page_setup_get_paper_height IA__gtk_page_setup_get_paper_height

extern __typeof (gtk_page_setup_get_page_width) IA__gtk_page_setup_get_page_width __attribute((visibility("hidden")));
#define gtk_page_setup_get_page_width IA__gtk_page_setup_get_page_width

extern __typeof (gtk_page_setup_get_page_height) IA__gtk_page_setup_get_page_height __attribute((visibility("hidden")));
#define gtk_page_setup_get_page_height IA__gtk_page_setup_get_page_height

#endif
#endif
#if IN_HEADER(__GTK_PAGE_SETUP_UNIX_DIALOG_H__)
#if IN_FILE(__GTK_PAGE_SETUP_UNIX_DIALOG_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_page_setup_unix_dialog_get_type) IA__gtk_page_setup_unix_dialog_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_page_setup_unix_dialog_get_type IA__gtk_page_setup_unix_dialog_get_type

extern __typeof (gtk_page_setup_unix_dialog_new) IA__gtk_page_setup_unix_dialog_new __attribute((visibility("hidden")));
#define gtk_page_setup_unix_dialog_new IA__gtk_page_setup_unix_dialog_new

extern __typeof (gtk_page_setup_unix_dialog_set_page_setup) IA__gtk_page_setup_unix_dialog_set_page_setup __attribute((visibility("hidden")));
#define gtk_page_setup_unix_dialog_set_page_setup IA__gtk_page_setup_unix_dialog_set_page_setup

extern __typeof (gtk_page_setup_unix_dialog_get_page_setup) IA__gtk_page_setup_unix_dialog_get_page_setup __attribute((visibility("hidden")));
#define gtk_page_setup_unix_dialog_get_page_setup IA__gtk_page_setup_unix_dialog_get_page_setup

extern __typeof (gtk_page_setup_unix_dialog_set_print_settings) IA__gtk_page_setup_unix_dialog_set_print_settings __attribute((visibility("hidden")));
#define gtk_page_setup_unix_dialog_set_print_settings IA__gtk_page_setup_unix_dialog_set_print_settings

extern __typeof (gtk_page_setup_unix_dialog_get_print_settings) IA__gtk_page_setup_unix_dialog_get_print_settings __attribute((visibility("hidden")));
#define gtk_page_setup_unix_dialog_get_print_settings IA__gtk_page_setup_unix_dialog_get_print_settings

#endif
#endif
#endif
#if IN_HEADER(__GTK_PAPER_SIZE_H__)
#if IN_FILE(__GTK_PAPER_SIZE_C__)
extern __typeof (gtk_paper_size_get_type) IA__gtk_paper_size_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_paper_size_get_type IA__gtk_paper_size_get_type

extern __typeof (gtk_paper_size_new) IA__gtk_paper_size_new __attribute((visibility("hidden")));
#define gtk_paper_size_new IA__gtk_paper_size_new

extern __typeof (gtk_paper_size_new_from_ppd) IA__gtk_paper_size_new_from_ppd __attribute((visibility("hidden")));
#define gtk_paper_size_new_from_ppd IA__gtk_paper_size_new_from_ppd

extern __typeof (gtk_paper_size_new_custom) IA__gtk_paper_size_new_custom __attribute((visibility("hidden")));
#define gtk_paper_size_new_custom IA__gtk_paper_size_new_custom

extern __typeof (gtk_paper_size_new_from_key_file) IA__gtk_paper_size_new_from_key_file __attribute((visibility("hidden")));
#define gtk_paper_size_new_from_key_file IA__gtk_paper_size_new_from_key_file

extern __typeof (gtk_paper_size_to_key_file) IA__gtk_paper_size_to_key_file __attribute((visibility("hidden")));
#define gtk_paper_size_to_key_file IA__gtk_paper_size_to_key_file

extern __typeof (gtk_paper_size_copy) IA__gtk_paper_size_copy __attribute((visibility("hidden")));
#define gtk_paper_size_copy IA__gtk_paper_size_copy

extern __typeof (gtk_paper_size_free) IA__gtk_paper_size_free __attribute((visibility("hidden")));
#define gtk_paper_size_free IA__gtk_paper_size_free

extern __typeof (gtk_paper_size_is_equal) IA__gtk_paper_size_is_equal __attribute((visibility("hidden")));
#define gtk_paper_size_is_equal IA__gtk_paper_size_is_equal

extern __typeof (gtk_paper_size_get_paper_sizes) IA__gtk_paper_size_get_paper_sizes __attribute((visibility("hidden")));
#define gtk_paper_size_get_paper_sizes IA__gtk_paper_size_get_paper_sizes

extern __typeof (gtk_paper_size_get_name) IA__gtk_paper_size_get_name __attribute((visibility("hidden")));
#define gtk_paper_size_get_name IA__gtk_paper_size_get_name

extern __typeof (gtk_paper_size_get_display_name) IA__gtk_paper_size_get_display_name __attribute((visibility("hidden")));
#define gtk_paper_size_get_display_name IA__gtk_paper_size_get_display_name

extern __typeof (gtk_paper_size_get_ppd_name) IA__gtk_paper_size_get_ppd_name __attribute((visibility("hidden")));
#define gtk_paper_size_get_ppd_name IA__gtk_paper_size_get_ppd_name

extern __typeof (gtk_paper_size_get_width) IA__gtk_paper_size_get_width __attribute((visibility("hidden")));
#define gtk_paper_size_get_width IA__gtk_paper_size_get_width

extern __typeof (gtk_paper_size_get_height) IA__gtk_paper_size_get_height __attribute((visibility("hidden")));
#define gtk_paper_size_get_height IA__gtk_paper_size_get_height

extern __typeof (gtk_paper_size_is_custom) IA__gtk_paper_size_is_custom __attribute((visibility("hidden")));
#define gtk_paper_size_is_custom IA__gtk_paper_size_is_custom

extern __typeof (gtk_paper_size_set_size) IA__gtk_paper_size_set_size __attribute((visibility("hidden")));
#define gtk_paper_size_set_size IA__gtk_paper_size_set_size

extern __typeof (gtk_paper_size_get_default_top_margin) IA__gtk_paper_size_get_default_top_margin __attribute((visibility("hidden")));
#define gtk_paper_size_get_default_top_margin IA__gtk_paper_size_get_default_top_margin

extern __typeof (gtk_paper_size_get_default_bottom_margin) IA__gtk_paper_size_get_default_bottom_margin __attribute((visibility("hidden")));
#define gtk_paper_size_get_default_bottom_margin IA__gtk_paper_size_get_default_bottom_margin

extern __typeof (gtk_paper_size_get_default_left_margin) IA__gtk_paper_size_get_default_left_margin __attribute((visibility("hidden")));
#define gtk_paper_size_get_default_left_margin IA__gtk_paper_size_get_default_left_margin

extern __typeof (gtk_paper_size_get_default_right_margin) IA__gtk_paper_size_get_default_right_margin __attribute((visibility("hidden")));
#define gtk_paper_size_get_default_right_margin IA__gtk_paper_size_get_default_right_margin

extern __typeof (gtk_paper_size_get_default) IA__gtk_paper_size_get_default __attribute((visibility("hidden")));
#define gtk_paper_size_get_default IA__gtk_paper_size_get_default

#endif
#endif
#if IN_HEADER(__GTK_CUSTOM_PAPER_UNIX_DIALOG_H__)
#if IN_FILE(__GTK_CUSTOM_PAPER_UNIX_DIALOG_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_custom_paper_unix_dialog_get_type) IA__gtk_custom_paper_unix_dialog_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_custom_paper_unix_dialog_get_type IA__gtk_custom_paper_unix_dialog_get_type

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_BACKEND_H__)
#if IN_FILE(__GTK_PRINT_BACKEND_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_print_backend_error_quark) IA__gtk_print_backend_error_quark __attribute((visibility("hidden")));
#define gtk_print_backend_error_quark IA__gtk_print_backend_error_quark

extern __typeof (gtk_print_backend_get_type) IA__gtk_print_backend_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_print_backend_get_type IA__gtk_print_backend_get_type

extern __typeof (gtk_print_backend_add_printer) IA__gtk_print_backend_add_printer __attribute((visibility("hidden")));
#define gtk_print_backend_add_printer IA__gtk_print_backend_add_printer

extern __typeof (gtk_print_backend_remove_printer) IA__gtk_print_backend_remove_printer __attribute((visibility("hidden")));
#define gtk_print_backend_remove_printer IA__gtk_print_backend_remove_printer

extern __typeof (gtk_print_backend_set_list_done) IA__gtk_print_backend_set_list_done __attribute((visibility("hidden")));
#define gtk_print_backend_set_list_done IA__gtk_print_backend_set_list_done

extern __typeof (gtk_print_backend_destroy) IA__gtk_print_backend_destroy __attribute((visibility("hidden")));
#define gtk_print_backend_destroy IA__gtk_print_backend_destroy

extern __typeof (gtk_print_backend_get_printer_list) IA__gtk_print_backend_get_printer_list __attribute((visibility("hidden")));
#define gtk_print_backend_get_printer_list IA__gtk_print_backend_get_printer_list

extern __typeof (gtk_print_backend_printer_list_is_done) IA__gtk_print_backend_printer_list_is_done __attribute((visibility("hidden")));
#define gtk_print_backend_printer_list_is_done IA__gtk_print_backend_printer_list_is_done

extern __typeof (gtk_print_backend_find_printer) IA__gtk_print_backend_find_printer __attribute((visibility("hidden")));
#define gtk_print_backend_find_printer IA__gtk_print_backend_find_printer

extern __typeof (gtk_print_backend_print_stream) IA__gtk_print_backend_print_stream __attribute((visibility("hidden")));
#define gtk_print_backend_print_stream IA__gtk_print_backend_print_stream

extern __typeof (gtk_print_backend_load_modules) IA__gtk_print_backend_load_modules __attribute((visibility("hidden")));
#define gtk_print_backend_load_modules IA__gtk_print_backend_load_modules

extern __typeof (gtk_print_backend_set_password) IA__gtk_print_backend_set_password __attribute((visibility("hidden")));
#define gtk_print_backend_set_password IA__gtk_print_backend_set_password

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_CONTEXT_H__)
#if IN_FILE(__GTK_PRINT_CONTEXT_C__)
extern __typeof (gtk_print_context_get_type) IA__gtk_print_context_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_print_context_get_type IA__gtk_print_context_get_type

extern __typeof (gtk_print_context_get_cairo_context) IA__gtk_print_context_get_cairo_context __attribute((visibility("hidden")));
#define gtk_print_context_get_cairo_context IA__gtk_print_context_get_cairo_context

extern __typeof (gtk_print_context_get_page_setup) IA__gtk_print_context_get_page_setup __attribute((visibility("hidden")));
#define gtk_print_context_get_page_setup IA__gtk_print_context_get_page_setup

extern __typeof (gtk_print_context_get_width) IA__gtk_print_context_get_width __attribute((visibility("hidden")));
#define gtk_print_context_get_width IA__gtk_print_context_get_width

extern __typeof (gtk_print_context_get_height) IA__gtk_print_context_get_height __attribute((visibility("hidden")));
#define gtk_print_context_get_height IA__gtk_print_context_get_height

extern __typeof (gtk_print_context_get_dpi_x) IA__gtk_print_context_get_dpi_x __attribute((visibility("hidden")));
#define gtk_print_context_get_dpi_x IA__gtk_print_context_get_dpi_x

extern __typeof (gtk_print_context_get_dpi_y) IA__gtk_print_context_get_dpi_y __attribute((visibility("hidden")));
#define gtk_print_context_get_dpi_y IA__gtk_print_context_get_dpi_y

extern __typeof (gtk_print_context_get_hard_margins) IA__gtk_print_context_get_hard_margins __attribute((visibility("hidden")));
#define gtk_print_context_get_hard_margins IA__gtk_print_context_get_hard_margins

extern __typeof (gtk_print_context_get_pango_fontmap) IA__gtk_print_context_get_pango_fontmap __attribute((visibility("hidden")));
#define gtk_print_context_get_pango_fontmap IA__gtk_print_context_get_pango_fontmap

extern __typeof (gtk_print_context_create_pango_context) IA__gtk_print_context_create_pango_context __attribute((visibility("hidden")));
#define gtk_print_context_create_pango_context IA__gtk_print_context_create_pango_context

extern __typeof (gtk_print_context_create_pango_layout) IA__gtk_print_context_create_pango_layout __attribute((visibility("hidden")));
#define gtk_print_context_create_pango_layout IA__gtk_print_context_create_pango_layout

extern __typeof (gtk_print_context_set_cairo_context) IA__gtk_print_context_set_cairo_context __attribute((visibility("hidden")));
#define gtk_print_context_set_cairo_context IA__gtk_print_context_set_cairo_context

#endif
#endif
#if IN_HEADER(__GTK_PRINTER_H__)
#if IN_FILE(__GTK_PRINTER_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_printer_get_type) IA__gtk_printer_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_printer_get_type IA__gtk_printer_get_type

extern __typeof (gtk_printer_get_backend) IA__gtk_printer_get_backend __attribute((visibility("hidden")));
#define gtk_printer_get_backend IA__gtk_printer_get_backend

extern __typeof (gtk_printer_get_name) IA__gtk_printer_get_name __attribute((visibility("hidden")));
#define gtk_printer_get_name IA__gtk_printer_get_name

extern __typeof (gtk_printer_get_description) IA__gtk_printer_get_description __attribute((visibility("hidden")));
#define gtk_printer_get_description IA__gtk_printer_get_description

extern __typeof (gtk_printer_get_default_page_size) IA__gtk_printer_get_default_page_size __attribute((visibility("hidden")));
#define gtk_printer_get_default_page_size IA__gtk_printer_get_default_page_size

extern __typeof (gtk_printer_get_state_message) IA__gtk_printer_get_state_message __attribute((visibility("hidden")));
#define gtk_printer_get_state_message IA__gtk_printer_get_state_message

extern __typeof (gtk_printer_get_location) IA__gtk_printer_get_location __attribute((visibility("hidden")));
#define gtk_printer_get_location IA__gtk_printer_get_location

extern __typeof (gtk_printer_get_icon_name) IA__gtk_printer_get_icon_name __attribute((visibility("hidden")));
#define gtk_printer_get_icon_name IA__gtk_printer_get_icon_name

extern __typeof (gtk_printer_get_job_count) IA__gtk_printer_get_job_count __attribute((visibility("hidden")));
#define gtk_printer_get_job_count IA__gtk_printer_get_job_count

extern __typeof (gtk_printer_is_virtual) IA__gtk_printer_is_virtual __attribute((visibility("hidden")));
#define gtk_printer_is_virtual IA__gtk_printer_is_virtual

extern __typeof (gtk_printer_list_papers) IA__gtk_printer_list_papers __attribute((visibility("hidden")));
#define gtk_printer_list_papers IA__gtk_printer_list_papers

extern __typeof (gtk_printer_accepts_pdf) IA__gtk_printer_accepts_pdf __attribute((visibility("hidden")));
#define gtk_printer_accepts_pdf IA__gtk_printer_accepts_pdf

extern __typeof (gtk_printer_accepts_ps) IA__gtk_printer_accepts_ps __attribute((visibility("hidden")));
#define gtk_printer_accepts_ps IA__gtk_printer_accepts_ps

extern __typeof (gtk_printer_compare) IA__gtk_printer_compare __attribute((visibility("hidden")));
#define gtk_printer_compare IA__gtk_printer_compare

extern __typeof (gtk_printer_has_details) IA__gtk_printer_has_details __attribute((visibility("hidden")));
#define gtk_printer_has_details IA__gtk_printer_has_details

extern __typeof (gtk_printer_request_details) IA__gtk_printer_request_details __attribute((visibility("hidden")));
#define gtk_printer_request_details IA__gtk_printer_request_details

extern __typeof (gtk_printer_get_capabilities) IA__gtk_printer_get_capabilities __attribute((visibility("hidden")));
#define gtk_printer_get_capabilities IA__gtk_printer_get_capabilities

extern __typeof (gtk_printer_get_hard_margins) IA__gtk_printer_get_hard_margins __attribute((visibility("hidden")));
#define gtk_printer_get_hard_margins IA__gtk_printer_get_hard_margins

extern __typeof (gtk_enumerate_printers) IA__gtk_enumerate_printers __attribute((visibility("hidden")));
#define gtk_enumerate_printers IA__gtk_enumerate_printers

extern __typeof (gtk_print_capabilities_get_type) IA__gtk_print_capabilities_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_print_capabilities_get_type IA__gtk_print_capabilities_get_type

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_BACKEND_H__)
#if IN_FILE(__GTK_PRINTER_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_printer_new) IA__gtk_printer_new __attribute((visibility("hidden")));
#define gtk_printer_new IA__gtk_printer_new

extern __typeof (gtk_printer_set_accepts_pdf) IA__gtk_printer_set_accepts_pdf __attribute((visibility("hidden")));
#define gtk_printer_set_accepts_pdf IA__gtk_printer_set_accepts_pdf

extern __typeof (gtk_printer_set_accepts_ps) IA__gtk_printer_set_accepts_ps __attribute((visibility("hidden")));
#define gtk_printer_set_accepts_ps IA__gtk_printer_set_accepts_ps

extern __typeof (gtk_printer_set_description) IA__gtk_printer_set_description __attribute((visibility("hidden")));
#define gtk_printer_set_description IA__gtk_printer_set_description

extern __typeof (gtk_printer_set_has_details) IA__gtk_printer_set_has_details __attribute((visibility("hidden")));
#define gtk_printer_set_has_details IA__gtk_printer_set_has_details

extern __typeof (gtk_printer_set_icon_name) IA__gtk_printer_set_icon_name __attribute((visibility("hidden")));
#define gtk_printer_set_icon_name IA__gtk_printer_set_icon_name

extern __typeof (gtk_printer_set_is_active) IA__gtk_printer_set_is_active __attribute((visibility("hidden")));
#define gtk_printer_set_is_active IA__gtk_printer_set_is_active

extern __typeof (gtk_printer_set_is_paused) IA__gtk_printer_set_is_paused __attribute((visibility("hidden")));
#define gtk_printer_set_is_paused IA__gtk_printer_set_is_paused

extern __typeof (gtk_printer_set_is_accepting_jobs) IA__gtk_printer_set_is_accepting_jobs __attribute((visibility("hidden")));
#define gtk_printer_set_is_accepting_jobs IA__gtk_printer_set_is_accepting_jobs

extern __typeof (gtk_printer_set_is_default) IA__gtk_printer_set_is_default __attribute((visibility("hidden")));
#define gtk_printer_set_is_default IA__gtk_printer_set_is_default

extern __typeof (gtk_printer_set_is_new) IA__gtk_printer_set_is_new __attribute((visibility("hidden")));
#define gtk_printer_set_is_new IA__gtk_printer_set_is_new

extern __typeof (gtk_printer_set_job_count) IA__gtk_printer_set_job_count __attribute((visibility("hidden")));
#define gtk_printer_set_job_count IA__gtk_printer_set_job_count

extern __typeof (gtk_printer_set_location) IA__gtk_printer_set_location __attribute((visibility("hidden")));
#define gtk_printer_set_location IA__gtk_printer_set_location

extern __typeof (gtk_printer_set_state_message) IA__gtk_printer_set_state_message __attribute((visibility("hidden")));
#define gtk_printer_set_state_message IA__gtk_printer_set_state_message

extern __typeof (gtk_printer_is_active) IA__gtk_printer_is_active __attribute((visibility("hidden")));
#define gtk_printer_is_active IA__gtk_printer_is_active

extern __typeof (gtk_printer_is_paused) IA__gtk_printer_is_paused __attribute((visibility("hidden")));
#define gtk_printer_is_paused IA__gtk_printer_is_paused

extern __typeof (gtk_printer_is_accepting_jobs) IA__gtk_printer_is_accepting_jobs __attribute((visibility("hidden")));
#define gtk_printer_is_accepting_jobs IA__gtk_printer_is_accepting_jobs

extern __typeof (gtk_printer_is_new) IA__gtk_printer_is_new __attribute((visibility("hidden")));
#define gtk_printer_is_new IA__gtk_printer_is_new

extern __typeof (gtk_printer_is_default) IA__gtk_printer_is_default __attribute((visibility("hidden")));
#define gtk_printer_is_default IA__gtk_printer_is_default

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINTER_OPTION_H__)
#if IN_FILE(__GTK_PRINTER_OPTION_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_printer_option_get_type) IA__gtk_printer_option_get_type __attribute((visibility("hidden")));
#define gtk_printer_option_get_type IA__gtk_printer_option_get_type

extern __typeof (gtk_printer_option_has_choice) IA__gtk_printer_option_has_choice __attribute((visibility("hidden")));
#define gtk_printer_option_has_choice IA__gtk_printer_option_has_choice

extern __typeof (gtk_printer_option_new) IA__gtk_printer_option_new __attribute((visibility("hidden")));
#define gtk_printer_option_new IA__gtk_printer_option_new

extern __typeof (gtk_printer_option_set) IA__gtk_printer_option_set __attribute((visibility("hidden")));
#define gtk_printer_option_set IA__gtk_printer_option_set

extern __typeof (gtk_printer_option_set_has_conflict) IA__gtk_printer_option_set_has_conflict __attribute((visibility("hidden")));
#define gtk_printer_option_set_has_conflict IA__gtk_printer_option_set_has_conflict

extern __typeof (gtk_printer_option_clear_has_conflict) IA__gtk_printer_option_clear_has_conflict __attribute((visibility("hidden")));
#define gtk_printer_option_clear_has_conflict IA__gtk_printer_option_clear_has_conflict

extern __typeof (gtk_printer_option_set_boolean) IA__gtk_printer_option_set_boolean __attribute((visibility("hidden")));
#define gtk_printer_option_set_boolean IA__gtk_printer_option_set_boolean

extern __typeof (gtk_printer_option_allocate_choices) IA__gtk_printer_option_allocate_choices __attribute((visibility("hidden")));
#define gtk_printer_option_allocate_choices IA__gtk_printer_option_allocate_choices

extern __typeof (gtk_printer_option_choices_from_array) IA__gtk_printer_option_choices_from_array __attribute((visibility("hidden")));
#define gtk_printer_option_choices_from_array IA__gtk_printer_option_choices_from_array

extern __typeof (gtk_printer_option_set_activates_default) IA__gtk_printer_option_set_activates_default __attribute((visibility("hidden")));
#define gtk_printer_option_set_activates_default IA__gtk_printer_option_set_activates_default

extern __typeof (gtk_printer_option_get_activates_default) IA__gtk_printer_option_get_activates_default __attribute((visibility("hidden")));
#define gtk_printer_option_get_activates_default IA__gtk_printer_option_get_activates_default

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINTER_OPTION_SET_H__)
#if IN_FILE(__GTK_PRINTER_OPTION_SET_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_printer_option_set_get_type) IA__gtk_printer_option_set_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_printer_option_set_get_type IA__gtk_printer_option_set_get_type

extern __typeof (gtk_printer_option_set_new) IA__gtk_printer_option_set_new __attribute((visibility("hidden")));
#define gtk_printer_option_set_new IA__gtk_printer_option_set_new

extern __typeof (gtk_printer_option_set_add) IA__gtk_printer_option_set_add __attribute((visibility("hidden")));
#define gtk_printer_option_set_add IA__gtk_printer_option_set_add

extern __typeof (gtk_printer_option_set_lookup) IA__gtk_printer_option_set_lookup __attribute((visibility("hidden")));
#define gtk_printer_option_set_lookup IA__gtk_printer_option_set_lookup

extern __typeof (gtk_printer_option_set_remove) IA__gtk_printer_option_set_remove __attribute((visibility("hidden")));
#define gtk_printer_option_set_remove IA__gtk_printer_option_set_remove

extern __typeof (gtk_printer_option_set_foreach) IA__gtk_printer_option_set_foreach __attribute((visibility("hidden")));
#define gtk_printer_option_set_foreach IA__gtk_printer_option_set_foreach

extern __typeof (gtk_printer_option_set_clear_conflicts) IA__gtk_printer_option_set_clear_conflicts __attribute((visibility("hidden")));
#define gtk_printer_option_set_clear_conflicts IA__gtk_printer_option_set_clear_conflicts

extern __typeof (gtk_printer_option_set_get_groups) IA__gtk_printer_option_set_get_groups __attribute((visibility("hidden")));
#define gtk_printer_option_set_get_groups IA__gtk_printer_option_set_get_groups

extern __typeof (gtk_printer_option_set_foreach_in_group) IA__gtk_printer_option_set_foreach_in_group __attribute((visibility("hidden")));
#define gtk_printer_option_set_foreach_in_group IA__gtk_printer_option_set_foreach_in_group

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINTER_OPTION_WIDGET_H__)
#if IN_FILE(__GTK_PRINTER_OPTION_WIDGET_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_printer_option_widget_get_type) IA__gtk_printer_option_widget_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_printer_option_widget_get_type IA__gtk_printer_option_widget_get_type

extern __typeof (gtk_printer_option_widget_new) IA__gtk_printer_option_widget_new __attribute((visibility("hidden")));
#define gtk_printer_option_widget_new IA__gtk_printer_option_widget_new

extern __typeof (gtk_printer_option_widget_set_source) IA__gtk_printer_option_widget_set_source __attribute((visibility("hidden")));
#define gtk_printer_option_widget_set_source IA__gtk_printer_option_widget_set_source

extern __typeof (gtk_printer_option_widget_has_external_label) IA__gtk_printer_option_widget_has_external_label __attribute((visibility("hidden")));
#define gtk_printer_option_widget_has_external_label IA__gtk_printer_option_widget_has_external_label

extern __typeof (gtk_printer_option_widget_get_external_label) IA__gtk_printer_option_widget_get_external_label __attribute((visibility("hidden")));
#define gtk_printer_option_widget_get_external_label IA__gtk_printer_option_widget_get_external_label

extern __typeof (gtk_printer_option_widget_get_value) IA__gtk_printer_option_widget_get_value __attribute((visibility("hidden")));
#define gtk_printer_option_widget_get_value IA__gtk_printer_option_widget_get_value

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_JOB_H__)
#if IN_FILE(__GTK_PRINT_JOB_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_print_job_get_type) IA__gtk_print_job_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_print_job_get_type IA__gtk_print_job_get_type

extern __typeof (gtk_print_job_new) IA__gtk_print_job_new __attribute((visibility("hidden")));
#define gtk_print_job_new IA__gtk_print_job_new

extern __typeof (gtk_print_job_get_settings) IA__gtk_print_job_get_settings __attribute((visibility("hidden")));
#define gtk_print_job_get_settings IA__gtk_print_job_get_settings

extern __typeof (gtk_print_job_get_printer) IA__gtk_print_job_get_printer __attribute((visibility("hidden")));
#define gtk_print_job_get_printer IA__gtk_print_job_get_printer

extern __typeof (gtk_print_job_get_title) IA__gtk_print_job_get_title __attribute((visibility("hidden")));
#define gtk_print_job_get_title IA__gtk_print_job_get_title

extern __typeof (gtk_print_job_get_status) IA__gtk_print_job_get_status __attribute((visibility("hidden")));
#define gtk_print_job_get_status IA__gtk_print_job_get_status

extern __typeof (gtk_print_job_set_source_file) IA__gtk_print_job_set_source_file __attribute((visibility("hidden")));
#define gtk_print_job_set_source_file IA__gtk_print_job_set_source_file

extern __typeof (gtk_print_job_get_surface) IA__gtk_print_job_get_surface __attribute((visibility("hidden")));
#define gtk_print_job_get_surface IA__gtk_print_job_get_surface

extern __typeof (gtk_print_job_send) IA__gtk_print_job_send __attribute((visibility("hidden")));
#define gtk_print_job_send IA__gtk_print_job_send

extern __typeof (gtk_print_job_set_track_print_status) IA__gtk_print_job_set_track_print_status __attribute((visibility("hidden")));
#define gtk_print_job_set_track_print_status IA__gtk_print_job_set_track_print_status

extern __typeof (gtk_print_job_get_track_print_status) IA__gtk_print_job_get_track_print_status __attribute((visibility("hidden")));
#define gtk_print_job_get_track_print_status IA__gtk_print_job_get_track_print_status

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINTER_PRIVATE_H__)
#if IN_FILE(__GTK_PRINT_JOB_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_print_job_set_status) IA__gtk_print_job_set_status __attribute((visibility("hidden")));
#define gtk_print_job_set_status IA__gtk_print_job_set_status

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_OPERATION_H__)
#if IN_FILE(__GTK_PRINT_OPERATION_C__)
extern __typeof (gtk_print_error_quark) IA__gtk_print_error_quark __attribute((visibility("hidden")));
#define gtk_print_error_quark IA__gtk_print_error_quark

extern __typeof (gtk_print_operation_get_type) IA__gtk_print_operation_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_print_operation_get_type IA__gtk_print_operation_get_type

extern __typeof (gtk_print_operation_new) IA__gtk_print_operation_new __attribute((visibility("hidden")));
#define gtk_print_operation_new IA__gtk_print_operation_new

extern __typeof (gtk_print_operation_set_allow_async) IA__gtk_print_operation_set_allow_async __attribute((visibility("hidden")));
#define gtk_print_operation_set_allow_async IA__gtk_print_operation_set_allow_async

extern __typeof (gtk_print_operation_set_default_page_setup) IA__gtk_print_operation_set_default_page_setup __attribute((visibility("hidden")));
#define gtk_print_operation_set_default_page_setup IA__gtk_print_operation_set_default_page_setup

extern __typeof (gtk_print_operation_get_default_page_setup) IA__gtk_print_operation_get_default_page_setup __attribute((visibility("hidden")));
#define gtk_print_operation_get_default_page_setup IA__gtk_print_operation_get_default_page_setup

extern __typeof (gtk_print_operation_set_print_settings) IA__gtk_print_operation_set_print_settings __attribute((visibility("hidden")));
#define gtk_print_operation_set_print_settings IA__gtk_print_operation_set_print_settings

extern __typeof (gtk_print_operation_get_n_pages_to_print) IA__gtk_print_operation_get_n_pages_to_print __attribute((visibility("hidden")));
#define gtk_print_operation_get_n_pages_to_print IA__gtk_print_operation_get_n_pages_to_print

extern __typeof (gtk_print_operation_get_print_settings) IA__gtk_print_operation_get_print_settings __attribute((visibility("hidden")));
#define gtk_print_operation_get_print_settings IA__gtk_print_operation_get_print_settings

extern __typeof (gtk_print_operation_set_job_name) IA__gtk_print_operation_set_job_name __attribute((visibility("hidden")));
#define gtk_print_operation_set_job_name IA__gtk_print_operation_set_job_name

extern __typeof (gtk_print_operation_set_n_pages) IA__gtk_print_operation_set_n_pages __attribute((visibility("hidden")));
#define gtk_print_operation_set_n_pages IA__gtk_print_operation_set_n_pages

extern __typeof (gtk_print_operation_set_current_page) IA__gtk_print_operation_set_current_page __attribute((visibility("hidden")));
#define gtk_print_operation_set_current_page IA__gtk_print_operation_set_current_page

extern __typeof (gtk_print_operation_set_use_full_page) IA__gtk_print_operation_set_use_full_page __attribute((visibility("hidden")));
#define gtk_print_operation_set_use_full_page IA__gtk_print_operation_set_use_full_page

extern __typeof (gtk_print_operation_set_unit) IA__gtk_print_operation_set_unit __attribute((visibility("hidden")));
#define gtk_print_operation_set_unit IA__gtk_print_operation_set_unit

extern __typeof (gtk_print_operation_set_export_filename) IA__gtk_print_operation_set_export_filename __attribute((visibility("hidden")));
#define gtk_print_operation_set_export_filename IA__gtk_print_operation_set_export_filename

extern __typeof (gtk_print_operation_set_track_print_status) IA__gtk_print_operation_set_track_print_status __attribute((visibility("hidden")));
#define gtk_print_operation_set_track_print_status IA__gtk_print_operation_set_track_print_status

extern __typeof (gtk_print_operation_set_show_progress) IA__gtk_print_operation_set_show_progress __attribute((visibility("hidden")));
#define gtk_print_operation_set_show_progress IA__gtk_print_operation_set_show_progress

extern __typeof (gtk_print_operation_set_custom_tab_label) IA__gtk_print_operation_set_custom_tab_label __attribute((visibility("hidden")));
#define gtk_print_operation_set_custom_tab_label IA__gtk_print_operation_set_custom_tab_label

extern __typeof (gtk_print_operation_get_error) IA__gtk_print_operation_get_error __attribute((visibility("hidden")));
#define gtk_print_operation_get_error IA__gtk_print_operation_get_error

extern __typeof (gtk_print_operation_run) IA__gtk_print_operation_run __attribute((visibility("hidden")));
#define gtk_print_operation_run IA__gtk_print_operation_run

extern __typeof (gtk_print_operation_get_status) IA__gtk_print_operation_get_status __attribute((visibility("hidden")));
#define gtk_print_operation_get_status IA__gtk_print_operation_get_status

extern __typeof (gtk_print_operation_get_status_string) IA__gtk_print_operation_get_status_string __attribute((visibility("hidden")));
#define gtk_print_operation_get_status_string IA__gtk_print_operation_get_status_string

extern __typeof (gtk_print_operation_is_finished) IA__gtk_print_operation_is_finished __attribute((visibility("hidden")));
#define gtk_print_operation_is_finished IA__gtk_print_operation_is_finished

extern __typeof (gtk_print_operation_cancel) IA__gtk_print_operation_cancel __attribute((visibility("hidden")));
#define gtk_print_operation_cancel IA__gtk_print_operation_cancel

extern __typeof (gtk_print_operation_draw_page_finish) IA__gtk_print_operation_draw_page_finish __attribute((visibility("hidden")));
#define gtk_print_operation_draw_page_finish IA__gtk_print_operation_draw_page_finish

extern __typeof (gtk_print_operation_set_defer_drawing) IA__gtk_print_operation_set_defer_drawing __attribute((visibility("hidden")));
#define gtk_print_operation_set_defer_drawing IA__gtk_print_operation_set_defer_drawing

extern __typeof (gtk_print_operation_set_support_selection) IA__gtk_print_operation_set_support_selection __attribute((visibility("hidden")));
#define gtk_print_operation_set_support_selection IA__gtk_print_operation_set_support_selection

extern __typeof (gtk_print_operation_get_support_selection) IA__gtk_print_operation_get_support_selection __attribute((visibility("hidden")));
#define gtk_print_operation_get_support_selection IA__gtk_print_operation_get_support_selection

extern __typeof (gtk_print_operation_set_has_selection) IA__gtk_print_operation_set_has_selection __attribute((visibility("hidden")));
#define gtk_print_operation_set_has_selection IA__gtk_print_operation_set_has_selection

extern __typeof (gtk_print_operation_get_has_selection) IA__gtk_print_operation_get_has_selection __attribute((visibility("hidden")));
#define gtk_print_operation_get_has_selection IA__gtk_print_operation_get_has_selection

extern __typeof (gtk_print_operation_set_embed_page_setup) IA__gtk_print_operation_set_embed_page_setup __attribute((visibility("hidden")));
#define gtk_print_operation_set_embed_page_setup IA__gtk_print_operation_set_embed_page_setup

extern __typeof (gtk_print_operation_get_embed_page_setup) IA__gtk_print_operation_get_embed_page_setup __attribute((visibility("hidden")));
#define gtk_print_operation_get_embed_page_setup IA__gtk_print_operation_get_embed_page_setup

#endif
#endif
#if IN_HEADER(__GTK_PRINT_OPERATION_PREVIEW_H__)
#if IN_FILE(__GTK_PRINT_OPERATION_PREVIEW_C__)
extern __typeof (gtk_print_operation_preview_end_preview) IA__gtk_print_operation_preview_end_preview __attribute((visibility("hidden")));
#define gtk_print_operation_preview_end_preview IA__gtk_print_operation_preview_end_preview

extern __typeof (gtk_print_operation_preview_get_type) IA__gtk_print_operation_preview_get_type __attribute((visibility("hidden")));
#define gtk_print_operation_preview_get_type IA__gtk_print_operation_preview_get_type

extern __typeof (gtk_print_operation_preview_is_selected) IA__gtk_print_operation_preview_is_selected __attribute((visibility("hidden")));
#define gtk_print_operation_preview_is_selected IA__gtk_print_operation_preview_is_selected

extern __typeof (gtk_print_operation_preview_render_page) IA__gtk_print_operation_preview_render_page __attribute((visibility("hidden")));
#define gtk_print_operation_preview_render_page IA__gtk_print_operation_preview_render_page

#endif
#endif
#if IN_HEADER(__GTK_PRINT_OPERATION_H__)
#if IN_FILE(__GTK_PRINT_OPERATION_UNIX_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_print_run_page_setup_dialog) IA__gtk_print_run_page_setup_dialog __attribute((visibility("hidden")));
#define gtk_print_run_page_setup_dialog IA__gtk_print_run_page_setup_dialog

extern __typeof (gtk_print_run_page_setup_dialog_async) IA__gtk_print_run_page_setup_dialog_async __attribute((visibility("hidden")));
#define gtk_print_run_page_setup_dialog_async IA__gtk_print_run_page_setup_dialog_async

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_OPERATION_H__)
#if IN_FILE(__GTK_PRINT_OPERATION_WIN32_C__)
#ifdef G_OS_WIN32
extern __typeof (gtk_print_run_page_setup_dialog) IA__gtk_print_run_page_setup_dialog __attribute((visibility("hidden")));
#define gtk_print_run_page_setup_dialog IA__gtk_print_run_page_setup_dialog

extern __typeof (gtk_print_run_page_setup_dialog_async) IA__gtk_print_run_page_setup_dialog_async __attribute((visibility("hidden")));
#define gtk_print_run_page_setup_dialog_async IA__gtk_print_run_page_setup_dialog_async

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_SETTINGS_H__)
#if IN_FILE(__GTK_PRINT_SETTINGS_C__)
extern __typeof (gtk_print_settings_get_type) IA__gtk_print_settings_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_print_settings_get_type IA__gtk_print_settings_get_type

extern __typeof (gtk_print_settings_new) IA__gtk_print_settings_new __attribute((visibility("hidden")));
#define gtk_print_settings_new IA__gtk_print_settings_new

extern __typeof (gtk_print_settings_new_from_file) IA__gtk_print_settings_new_from_file __attribute((visibility("hidden")));
#define gtk_print_settings_new_from_file IA__gtk_print_settings_new_from_file

extern __typeof (gtk_print_settings_new_from_key_file) IA__gtk_print_settings_new_from_key_file __attribute((visibility("hidden")));
#define gtk_print_settings_new_from_key_file IA__gtk_print_settings_new_from_key_file

extern __typeof (gtk_print_settings_load_file) IA__gtk_print_settings_load_file __attribute((visibility("hidden")));
#define gtk_print_settings_load_file IA__gtk_print_settings_load_file

extern __typeof (gtk_print_settings_load_key_file) IA__gtk_print_settings_load_key_file __attribute((visibility("hidden")));
#define gtk_print_settings_load_key_file IA__gtk_print_settings_load_key_file

extern __typeof (gtk_print_settings_to_file) IA__gtk_print_settings_to_file __attribute((visibility("hidden")));
#define gtk_print_settings_to_file IA__gtk_print_settings_to_file

extern __typeof (gtk_print_settings_to_key_file) IA__gtk_print_settings_to_key_file __attribute((visibility("hidden")));
#define gtk_print_settings_to_key_file IA__gtk_print_settings_to_key_file

extern __typeof (gtk_print_settings_copy) IA__gtk_print_settings_copy __attribute((visibility("hidden")));
#define gtk_print_settings_copy IA__gtk_print_settings_copy

extern __typeof (gtk_print_settings_has_key) IA__gtk_print_settings_has_key __attribute((visibility("hidden")));
#define gtk_print_settings_has_key IA__gtk_print_settings_has_key

extern __typeof (gtk_print_settings_get) IA__gtk_print_settings_get __attribute((visibility("hidden")));
#define gtk_print_settings_get IA__gtk_print_settings_get

extern __typeof (gtk_print_settings_set) IA__gtk_print_settings_set __attribute((visibility("hidden")));
#define gtk_print_settings_set IA__gtk_print_settings_set

extern __typeof (gtk_print_settings_unset) IA__gtk_print_settings_unset __attribute((visibility("hidden")));
#define gtk_print_settings_unset IA__gtk_print_settings_unset

extern __typeof (gtk_print_settings_foreach) IA__gtk_print_settings_foreach __attribute((visibility("hidden")));
#define gtk_print_settings_foreach IA__gtk_print_settings_foreach

extern __typeof (gtk_print_settings_get_bool) IA__gtk_print_settings_get_bool __attribute((visibility("hidden")));
#define gtk_print_settings_get_bool IA__gtk_print_settings_get_bool

extern __typeof (gtk_print_settings_set_bool) IA__gtk_print_settings_set_bool __attribute((visibility("hidden")));
#define gtk_print_settings_set_bool IA__gtk_print_settings_set_bool

extern __typeof (gtk_print_settings_get_double) IA__gtk_print_settings_get_double __attribute((visibility("hidden")));
#define gtk_print_settings_get_double IA__gtk_print_settings_get_double

extern __typeof (gtk_print_settings_get_double_with_default) IA__gtk_print_settings_get_double_with_default __attribute((visibility("hidden")));
#define gtk_print_settings_get_double_with_default IA__gtk_print_settings_get_double_with_default

extern __typeof (gtk_print_settings_set_double) IA__gtk_print_settings_set_double __attribute((visibility("hidden")));
#define gtk_print_settings_set_double IA__gtk_print_settings_set_double

extern __typeof (gtk_print_settings_get_length) IA__gtk_print_settings_get_length __attribute((visibility("hidden")));
#define gtk_print_settings_get_length IA__gtk_print_settings_get_length

extern __typeof (gtk_print_settings_set_length) IA__gtk_print_settings_set_length __attribute((visibility("hidden")));
#define gtk_print_settings_set_length IA__gtk_print_settings_set_length

extern __typeof (gtk_print_settings_get_int) IA__gtk_print_settings_get_int __attribute((visibility("hidden")));
#define gtk_print_settings_get_int IA__gtk_print_settings_get_int

extern __typeof (gtk_print_settings_get_int_with_default) IA__gtk_print_settings_get_int_with_default __attribute((visibility("hidden")));
#define gtk_print_settings_get_int_with_default IA__gtk_print_settings_get_int_with_default

extern __typeof (gtk_print_settings_set_int) IA__gtk_print_settings_set_int __attribute((visibility("hidden")));
#define gtk_print_settings_set_int IA__gtk_print_settings_set_int

extern __typeof (gtk_print_settings_get_printer) IA__gtk_print_settings_get_printer __attribute((visibility("hidden")));
#define gtk_print_settings_get_printer IA__gtk_print_settings_get_printer

extern __typeof (gtk_print_settings_set_printer) IA__gtk_print_settings_set_printer __attribute((visibility("hidden")));
#define gtk_print_settings_set_printer IA__gtk_print_settings_set_printer

extern __typeof (gtk_print_settings_get_orientation) IA__gtk_print_settings_get_orientation __attribute((visibility("hidden")));
#define gtk_print_settings_get_orientation IA__gtk_print_settings_get_orientation

extern __typeof (gtk_print_settings_set_orientation) IA__gtk_print_settings_set_orientation __attribute((visibility("hidden")));
#define gtk_print_settings_set_orientation IA__gtk_print_settings_set_orientation

extern __typeof (gtk_print_settings_get_paper_size) IA__gtk_print_settings_get_paper_size __attribute((visibility("hidden")));
#define gtk_print_settings_get_paper_size IA__gtk_print_settings_get_paper_size

extern __typeof (gtk_print_settings_set_paper_size) IA__gtk_print_settings_set_paper_size __attribute((visibility("hidden")));
#define gtk_print_settings_set_paper_size IA__gtk_print_settings_set_paper_size

extern __typeof (gtk_print_settings_get_paper_width) IA__gtk_print_settings_get_paper_width __attribute((visibility("hidden")));
#define gtk_print_settings_get_paper_width IA__gtk_print_settings_get_paper_width

extern __typeof (gtk_print_settings_set_paper_width) IA__gtk_print_settings_set_paper_width __attribute((visibility("hidden")));
#define gtk_print_settings_set_paper_width IA__gtk_print_settings_set_paper_width

extern __typeof (gtk_print_settings_get_paper_height) IA__gtk_print_settings_get_paper_height __attribute((visibility("hidden")));
#define gtk_print_settings_get_paper_height IA__gtk_print_settings_get_paper_height

extern __typeof (gtk_print_settings_set_paper_height) IA__gtk_print_settings_set_paper_height __attribute((visibility("hidden")));
#define gtk_print_settings_set_paper_height IA__gtk_print_settings_set_paper_height

extern __typeof (gtk_print_settings_get_use_color) IA__gtk_print_settings_get_use_color __attribute((visibility("hidden")));
#define gtk_print_settings_get_use_color IA__gtk_print_settings_get_use_color

extern __typeof (gtk_print_settings_set_use_color) IA__gtk_print_settings_set_use_color __attribute((visibility("hidden")));
#define gtk_print_settings_set_use_color IA__gtk_print_settings_set_use_color

extern __typeof (gtk_print_settings_get_collate) IA__gtk_print_settings_get_collate __attribute((visibility("hidden")));
#define gtk_print_settings_get_collate IA__gtk_print_settings_get_collate

extern __typeof (gtk_print_settings_set_collate) IA__gtk_print_settings_set_collate __attribute((visibility("hidden")));
#define gtk_print_settings_set_collate IA__gtk_print_settings_set_collate

extern __typeof (gtk_print_settings_get_reverse) IA__gtk_print_settings_get_reverse __attribute((visibility("hidden")));
#define gtk_print_settings_get_reverse IA__gtk_print_settings_get_reverse

extern __typeof (gtk_print_settings_set_reverse) IA__gtk_print_settings_set_reverse __attribute((visibility("hidden")));
#define gtk_print_settings_set_reverse IA__gtk_print_settings_set_reverse

extern __typeof (gtk_print_settings_get_duplex) IA__gtk_print_settings_get_duplex __attribute((visibility("hidden")));
#define gtk_print_settings_get_duplex IA__gtk_print_settings_get_duplex

extern __typeof (gtk_print_settings_set_duplex) IA__gtk_print_settings_set_duplex __attribute((visibility("hidden")));
#define gtk_print_settings_set_duplex IA__gtk_print_settings_set_duplex

extern __typeof (gtk_print_settings_get_quality) IA__gtk_print_settings_get_quality __attribute((visibility("hidden")));
#define gtk_print_settings_get_quality IA__gtk_print_settings_get_quality

extern __typeof (gtk_print_settings_set_quality) IA__gtk_print_settings_set_quality __attribute((visibility("hidden")));
#define gtk_print_settings_set_quality IA__gtk_print_settings_set_quality

extern __typeof (gtk_print_settings_get_n_copies) IA__gtk_print_settings_get_n_copies __attribute((visibility("hidden")));
#define gtk_print_settings_get_n_copies IA__gtk_print_settings_get_n_copies

extern __typeof (gtk_print_settings_set_n_copies) IA__gtk_print_settings_set_n_copies __attribute((visibility("hidden")));
#define gtk_print_settings_set_n_copies IA__gtk_print_settings_set_n_copies

extern __typeof (gtk_print_settings_get_number_up) IA__gtk_print_settings_get_number_up __attribute((visibility("hidden")));
#define gtk_print_settings_get_number_up IA__gtk_print_settings_get_number_up

extern __typeof (gtk_print_settings_set_number_up) IA__gtk_print_settings_set_number_up __attribute((visibility("hidden")));
#define gtk_print_settings_set_number_up IA__gtk_print_settings_set_number_up

extern __typeof (gtk_print_settings_get_number_up_layout) IA__gtk_print_settings_get_number_up_layout __attribute((visibility("hidden")));
#define gtk_print_settings_get_number_up_layout IA__gtk_print_settings_get_number_up_layout

extern __typeof (gtk_print_settings_set_number_up_layout) IA__gtk_print_settings_set_number_up_layout __attribute((visibility("hidden")));
#define gtk_print_settings_set_number_up_layout IA__gtk_print_settings_set_number_up_layout

extern __typeof (gtk_print_settings_get_resolution) IA__gtk_print_settings_get_resolution __attribute((visibility("hidden")));
#define gtk_print_settings_get_resolution IA__gtk_print_settings_get_resolution

extern __typeof (gtk_print_settings_set_resolution) IA__gtk_print_settings_set_resolution __attribute((visibility("hidden")));
#define gtk_print_settings_set_resolution IA__gtk_print_settings_set_resolution

extern __typeof (gtk_print_settings_get_resolution_x) IA__gtk_print_settings_get_resolution_x __attribute((visibility("hidden")));
#define gtk_print_settings_get_resolution_x IA__gtk_print_settings_get_resolution_x

extern __typeof (gtk_print_settings_get_resolution_y) IA__gtk_print_settings_get_resolution_y __attribute((visibility("hidden")));
#define gtk_print_settings_get_resolution_y IA__gtk_print_settings_get_resolution_y

extern __typeof (gtk_print_settings_set_resolution_xy) IA__gtk_print_settings_set_resolution_xy __attribute((visibility("hidden")));
#define gtk_print_settings_set_resolution_xy IA__gtk_print_settings_set_resolution_xy

extern __typeof (gtk_print_settings_get_printer_lpi) IA__gtk_print_settings_get_printer_lpi __attribute((visibility("hidden")));
#define gtk_print_settings_get_printer_lpi IA__gtk_print_settings_get_printer_lpi

extern __typeof (gtk_print_settings_set_printer_lpi) IA__gtk_print_settings_set_printer_lpi __attribute((visibility("hidden")));
#define gtk_print_settings_set_printer_lpi IA__gtk_print_settings_set_printer_lpi

extern __typeof (gtk_print_settings_get_scale) IA__gtk_print_settings_get_scale __attribute((visibility("hidden")));
#define gtk_print_settings_get_scale IA__gtk_print_settings_get_scale

extern __typeof (gtk_print_settings_set_scale) IA__gtk_print_settings_set_scale __attribute((visibility("hidden")));
#define gtk_print_settings_set_scale IA__gtk_print_settings_set_scale

extern __typeof (gtk_print_settings_get_print_pages) IA__gtk_print_settings_get_print_pages __attribute((visibility("hidden")));
#define gtk_print_settings_get_print_pages IA__gtk_print_settings_get_print_pages

extern __typeof (gtk_print_settings_set_print_pages) IA__gtk_print_settings_set_print_pages __attribute((visibility("hidden")));
#define gtk_print_settings_set_print_pages IA__gtk_print_settings_set_print_pages

extern __typeof (gtk_print_settings_get_page_ranges) IA__gtk_print_settings_get_page_ranges __attribute((visibility("hidden")));
#define gtk_print_settings_get_page_ranges IA__gtk_print_settings_get_page_ranges

extern __typeof (gtk_print_settings_set_page_ranges) IA__gtk_print_settings_set_page_ranges __attribute((visibility("hidden")));
#define gtk_print_settings_set_page_ranges IA__gtk_print_settings_set_page_ranges

extern __typeof (gtk_print_settings_get_page_set) IA__gtk_print_settings_get_page_set __attribute((visibility("hidden")));
#define gtk_print_settings_get_page_set IA__gtk_print_settings_get_page_set

extern __typeof (gtk_print_settings_set_page_set) IA__gtk_print_settings_set_page_set __attribute((visibility("hidden")));
#define gtk_print_settings_set_page_set IA__gtk_print_settings_set_page_set

extern __typeof (gtk_print_settings_get_default_source) IA__gtk_print_settings_get_default_source __attribute((visibility("hidden")));
#define gtk_print_settings_get_default_source IA__gtk_print_settings_get_default_source

extern __typeof (gtk_print_settings_set_default_source) IA__gtk_print_settings_set_default_source __attribute((visibility("hidden")));
#define gtk_print_settings_set_default_source IA__gtk_print_settings_set_default_source

extern __typeof (gtk_print_settings_get_media_type) IA__gtk_print_settings_get_media_type __attribute((visibility("hidden")));
#define gtk_print_settings_get_media_type IA__gtk_print_settings_get_media_type

extern __typeof (gtk_print_settings_set_media_type) IA__gtk_print_settings_set_media_type __attribute((visibility("hidden")));
#define gtk_print_settings_set_media_type IA__gtk_print_settings_set_media_type

extern __typeof (gtk_print_settings_get_dither) IA__gtk_print_settings_get_dither __attribute((visibility("hidden")));
#define gtk_print_settings_get_dither IA__gtk_print_settings_get_dither

extern __typeof (gtk_print_settings_set_dither) IA__gtk_print_settings_set_dither __attribute((visibility("hidden")));
#define gtk_print_settings_set_dither IA__gtk_print_settings_set_dither

extern __typeof (gtk_print_settings_get_finishings) IA__gtk_print_settings_get_finishings __attribute((visibility("hidden")));
#define gtk_print_settings_get_finishings IA__gtk_print_settings_get_finishings

extern __typeof (gtk_print_settings_set_finishings) IA__gtk_print_settings_set_finishings __attribute((visibility("hidden")));
#define gtk_print_settings_set_finishings IA__gtk_print_settings_set_finishings

extern __typeof (gtk_print_settings_get_output_bin) IA__gtk_print_settings_get_output_bin __attribute((visibility("hidden")));
#define gtk_print_settings_get_output_bin IA__gtk_print_settings_get_output_bin

extern __typeof (gtk_print_settings_set_output_bin) IA__gtk_print_settings_set_output_bin __attribute((visibility("hidden")));
#define gtk_print_settings_set_output_bin IA__gtk_print_settings_set_output_bin

#endif
#endif
#if IN_HEADER(__GTK_PRINT_UNIX_DIALOG_H__)
#if IN_FILE(__GTK_PRINT_UNIX_DIALOG_C__)
#ifdef G_OS_UNIX
extern __typeof (gtk_print_unix_dialog_get_type) IA__gtk_print_unix_dialog_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_print_unix_dialog_get_type IA__gtk_print_unix_dialog_get_type

extern __typeof (gtk_print_unix_dialog_new) IA__gtk_print_unix_dialog_new __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_new IA__gtk_print_unix_dialog_new

extern __typeof (gtk_print_unix_dialog_set_page_setup) IA__gtk_print_unix_dialog_set_page_setup __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_set_page_setup IA__gtk_print_unix_dialog_set_page_setup

extern __typeof (gtk_print_unix_dialog_get_page_setup) IA__gtk_print_unix_dialog_get_page_setup __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_get_page_setup IA__gtk_print_unix_dialog_get_page_setup

extern __typeof (gtk_print_unix_dialog_set_current_page) IA__gtk_print_unix_dialog_set_current_page __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_set_current_page IA__gtk_print_unix_dialog_set_current_page

extern __typeof (gtk_print_unix_dialog_get_current_page) IA__gtk_print_unix_dialog_get_current_page __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_get_current_page IA__gtk_print_unix_dialog_get_current_page

extern __typeof (gtk_print_unix_dialog_set_settings) IA__gtk_print_unix_dialog_set_settings __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_set_settings IA__gtk_print_unix_dialog_set_settings

extern __typeof (gtk_print_unix_dialog_get_settings) IA__gtk_print_unix_dialog_get_settings __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_get_settings IA__gtk_print_unix_dialog_get_settings

extern __typeof (gtk_print_unix_dialog_get_selected_printer) IA__gtk_print_unix_dialog_get_selected_printer __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_get_selected_printer IA__gtk_print_unix_dialog_get_selected_printer

extern __typeof (gtk_print_unix_dialog_add_custom_tab) IA__gtk_print_unix_dialog_add_custom_tab __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_add_custom_tab IA__gtk_print_unix_dialog_add_custom_tab

extern __typeof (gtk_print_unix_dialog_set_manual_capabilities) IA__gtk_print_unix_dialog_set_manual_capabilities __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_set_manual_capabilities IA__gtk_print_unix_dialog_set_manual_capabilities

extern __typeof (gtk_print_unix_dialog_get_manual_capabilities) IA__gtk_print_unix_dialog_get_manual_capabilities __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_get_manual_capabilities IA__gtk_print_unix_dialog_get_manual_capabilities

extern __typeof (gtk_print_unix_dialog_set_support_selection) IA__gtk_print_unix_dialog_set_support_selection __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_set_support_selection IA__gtk_print_unix_dialog_set_support_selection

extern __typeof (gtk_print_unix_dialog_get_support_selection) IA__gtk_print_unix_dialog_get_support_selection __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_get_support_selection IA__gtk_print_unix_dialog_get_support_selection

extern __typeof (gtk_print_unix_dialog_set_has_selection) IA__gtk_print_unix_dialog_set_has_selection __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_set_has_selection IA__gtk_print_unix_dialog_set_has_selection

extern __typeof (gtk_print_unix_dialog_get_has_selection) IA__gtk_print_unix_dialog_get_has_selection __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_get_has_selection IA__gtk_print_unix_dialog_get_has_selection

extern __typeof (gtk_print_unix_dialog_set_embed_page_setup) IA__gtk_print_unix_dialog_set_embed_page_setup __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_set_embed_page_setup IA__gtk_print_unix_dialog_set_embed_page_setup

extern __typeof (gtk_print_unix_dialog_get_embed_page_setup) IA__gtk_print_unix_dialog_get_embed_page_setup __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_get_embed_page_setup IA__gtk_print_unix_dialog_get_embed_page_setup

extern __typeof (gtk_print_unix_dialog_get_page_setup_set) IA__gtk_print_unix_dialog_get_page_setup_set __attribute((visibility("hidden")));
#define gtk_print_unix_dialog_get_page_setup_set IA__gtk_print_unix_dialog_get_page_setup_set

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_WIN32_H__)
#if IN_FILE(__GTK_PRINT_WIN32_C__)
#ifdef G_OS_WIN32
extern __typeof (gtk_print_win32_devnames_free) IA__gtk_print_win32_devnames_free __attribute((visibility("hidden")));
#define gtk_print_win32_devnames_free IA__gtk_print_win32_devnames_free

extern __typeof (gtk_print_win32_devnames_from_win32) IA__gtk_print_win32_devnames_from_win32 __attribute((visibility("hidden")));
#define gtk_print_win32_devnames_from_win32 IA__gtk_print_win32_devnames_from_win32

extern __typeof (gtk_print_win32_devnames_to_win32) IA__gtk_print_win32_devnames_to_win32 __attribute((visibility("hidden")));
#define gtk_print_win32_devnames_to_win32 IA__gtk_print_win32_devnames_to_win32

extern __typeof (gtk_print_win32_devnames_from_printer_name) IA__gtk_print_win32_devnames_from_printer_name __attribute((visibility("hidden")));
#define gtk_print_win32_devnames_from_printer_name IA__gtk_print_win32_devnames_from_printer_name

#endif
#endif
#endif
#if IN_HEADER(__GTK_PROGRESS_BAR_H__)
#if IN_FILE(__GTK_PROGRESS_BAR_C__)
extern __typeof (gtk_progress_bar_get_fraction) IA__gtk_progress_bar_get_fraction __attribute((visibility("hidden")));
#define gtk_progress_bar_get_fraction IA__gtk_progress_bar_get_fraction

extern __typeof (gtk_progress_bar_get_orientation) IA__gtk_progress_bar_get_orientation __attribute((visibility("hidden")));
#define gtk_progress_bar_get_orientation IA__gtk_progress_bar_get_orientation

extern __typeof (gtk_progress_bar_get_pulse_step) IA__gtk_progress_bar_get_pulse_step __attribute((visibility("hidden")));
#define gtk_progress_bar_get_pulse_step IA__gtk_progress_bar_get_pulse_step

extern __typeof (gtk_progress_bar_get_text) IA__gtk_progress_bar_get_text __attribute((visibility("hidden")));
#define gtk_progress_bar_get_text IA__gtk_progress_bar_get_text

extern __typeof (gtk_progress_bar_get_ellipsize) IA__gtk_progress_bar_get_ellipsize __attribute((visibility("hidden")));
#define gtk_progress_bar_get_ellipsize IA__gtk_progress_bar_get_ellipsize

extern __typeof (gtk_progress_bar_get_type) IA__gtk_progress_bar_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_progress_bar_get_type IA__gtk_progress_bar_get_type

extern __typeof (gtk_progress_bar_new) IA__gtk_progress_bar_new __attribute((visibility("hidden")));
#define gtk_progress_bar_new IA__gtk_progress_bar_new

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_progress_bar_new_with_adjustment) IA__gtk_progress_bar_new_with_adjustment __attribute((visibility("hidden")));
#define gtk_progress_bar_new_with_adjustment IA__gtk_progress_bar_new_with_adjustment

extern __typeof (gtk_progress_bar_set_bar_style) IA__gtk_progress_bar_set_bar_style __attribute((visibility("hidden")));
#define gtk_progress_bar_set_bar_style IA__gtk_progress_bar_set_bar_style

extern __typeof (gtk_progress_bar_set_discrete_blocks) IA__gtk_progress_bar_set_discrete_blocks __attribute((visibility("hidden")));
#define gtk_progress_bar_set_discrete_blocks IA__gtk_progress_bar_set_discrete_blocks

extern __typeof (gtk_progress_bar_set_activity_blocks) IA__gtk_progress_bar_set_activity_blocks __attribute((visibility("hidden")));
#define gtk_progress_bar_set_activity_blocks IA__gtk_progress_bar_set_activity_blocks

extern __typeof (gtk_progress_bar_set_activity_step) IA__gtk_progress_bar_set_activity_step __attribute((visibility("hidden")));
#define gtk_progress_bar_set_activity_step IA__gtk_progress_bar_set_activity_step

extern __typeof (gtk_progress_bar_update) IA__gtk_progress_bar_update __attribute((visibility("hidden")));
#define gtk_progress_bar_update IA__gtk_progress_bar_update

#endif
extern __typeof (gtk_progress_bar_pulse) IA__gtk_progress_bar_pulse __attribute((visibility("hidden")));
#define gtk_progress_bar_pulse IA__gtk_progress_bar_pulse

extern __typeof (gtk_progress_bar_set_fraction) IA__gtk_progress_bar_set_fraction __attribute((visibility("hidden")));
#define gtk_progress_bar_set_fraction IA__gtk_progress_bar_set_fraction

extern __typeof (gtk_progress_bar_set_orientation) IA__gtk_progress_bar_set_orientation __attribute((visibility("hidden")));
#define gtk_progress_bar_set_orientation IA__gtk_progress_bar_set_orientation

extern __typeof (gtk_progress_bar_set_pulse_step) IA__gtk_progress_bar_set_pulse_step __attribute((visibility("hidden")));
#define gtk_progress_bar_set_pulse_step IA__gtk_progress_bar_set_pulse_step

extern __typeof (gtk_progress_bar_set_text) IA__gtk_progress_bar_set_text __attribute((visibility("hidden")));
#define gtk_progress_bar_set_text IA__gtk_progress_bar_set_text

extern __typeof (gtk_progress_bar_set_ellipsize) IA__gtk_progress_bar_set_ellipsize __attribute((visibility("hidden")));
#define gtk_progress_bar_set_ellipsize IA__gtk_progress_bar_set_ellipsize

#endif
#endif
#if IN_HEADER(__GTK_PROGRESS_H__)
#if IN_FILE(__GTK_PROGRESS_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_progress_configure) IA__gtk_progress_configure __attribute((visibility("hidden")));
#define gtk_progress_configure IA__gtk_progress_configure

extern __typeof (gtk_progress_get_current_percentage) IA__gtk_progress_get_current_percentage __attribute((visibility("hidden")));
#define gtk_progress_get_current_percentage IA__gtk_progress_get_current_percentage

extern __typeof (gtk_progress_get_current_text) IA__gtk_progress_get_current_text __attribute((visibility("hidden")));
#define gtk_progress_get_current_text IA__gtk_progress_get_current_text

extern __typeof (gtk_progress_get_percentage_from_value) IA__gtk_progress_get_percentage_from_value __attribute((visibility("hidden")));
#define gtk_progress_get_percentage_from_value IA__gtk_progress_get_percentage_from_value

extern __typeof (gtk_progress_get_text_from_value) IA__gtk_progress_get_text_from_value __attribute((visibility("hidden")));
#define gtk_progress_get_text_from_value IA__gtk_progress_get_text_from_value

extern __typeof (gtk_progress_get_type) IA__gtk_progress_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_progress_get_type IA__gtk_progress_get_type

extern __typeof (gtk_progress_get_value) IA__gtk_progress_get_value __attribute((visibility("hidden")));
#define gtk_progress_get_value IA__gtk_progress_get_value

extern __typeof (gtk_progress_set_activity_mode) IA__gtk_progress_set_activity_mode __attribute((visibility("hidden")));
#define gtk_progress_set_activity_mode IA__gtk_progress_set_activity_mode

extern __typeof (gtk_progress_set_adjustment) IA__gtk_progress_set_adjustment __attribute((visibility("hidden")));
#define gtk_progress_set_adjustment IA__gtk_progress_set_adjustment

extern __typeof (gtk_progress_set_format_string) IA__gtk_progress_set_format_string __attribute((visibility("hidden")));
#define gtk_progress_set_format_string IA__gtk_progress_set_format_string

extern __typeof (gtk_progress_set_percentage) IA__gtk_progress_set_percentage __attribute((visibility("hidden")));
#define gtk_progress_set_percentage IA__gtk_progress_set_percentage

extern __typeof (gtk_progress_set_show_text) IA__gtk_progress_set_show_text __attribute((visibility("hidden")));
#define gtk_progress_set_show_text IA__gtk_progress_set_show_text

extern __typeof (gtk_progress_set_text_alignment) IA__gtk_progress_set_text_alignment __attribute((visibility("hidden")));
#define gtk_progress_set_text_alignment IA__gtk_progress_set_text_alignment

extern __typeof (gtk_progress_set_value) IA__gtk_progress_set_value __attribute((visibility("hidden")));
#define gtk_progress_set_value IA__gtk_progress_set_value

#endif
#endif
#endif
#if IN_HEADER(__GTK_RADIO_ACTION_H__)
#if IN_FILE(__GTK_RADIO_ACTION_C__)
extern __typeof (gtk_radio_action_get_current_value) IA__gtk_radio_action_get_current_value __attribute((visibility("hidden")));
#define gtk_radio_action_get_current_value IA__gtk_radio_action_get_current_value

extern __typeof (gtk_radio_action_get_group) IA__gtk_radio_action_get_group __attribute((visibility("hidden")));
#define gtk_radio_action_get_group IA__gtk_radio_action_get_group

extern __typeof (gtk_radio_action_get_type) IA__gtk_radio_action_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_radio_action_get_type IA__gtk_radio_action_get_type

extern __typeof (gtk_radio_action_new) IA__gtk_radio_action_new __attribute((visibility("hidden")));
#define gtk_radio_action_new IA__gtk_radio_action_new

extern __typeof (gtk_radio_action_set_current_value) IA__gtk_radio_action_set_current_value __attribute((visibility("hidden")));
#define gtk_radio_action_set_current_value IA__gtk_radio_action_set_current_value

extern __typeof (gtk_radio_action_set_group) IA__gtk_radio_action_set_group __attribute((visibility("hidden")));
#define gtk_radio_action_set_group IA__gtk_radio_action_set_group

#endif
#endif
#if IN_HEADER(__GTK_RADIO_BUTTON_H__)
#if IN_FILE(__GTK_RADIO_BUTTON_C__)
extern __typeof (gtk_radio_button_get_group) IA__gtk_radio_button_get_group __attribute((visibility("hidden")));
#define gtk_radio_button_get_group IA__gtk_radio_button_get_group

extern __typeof (gtk_radio_button_get_type) IA__gtk_radio_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_radio_button_get_type IA__gtk_radio_button_get_type

extern __typeof (gtk_radio_button_new) IA__gtk_radio_button_new __attribute((visibility("hidden")));
#define gtk_radio_button_new IA__gtk_radio_button_new

extern __typeof (gtk_radio_button_new_from_widget) IA__gtk_radio_button_new_from_widget __attribute((visibility("hidden")));
#define gtk_radio_button_new_from_widget IA__gtk_radio_button_new_from_widget

extern __typeof (gtk_radio_button_new_with_label) IA__gtk_radio_button_new_with_label __attribute((visibility("hidden")));
#define gtk_radio_button_new_with_label IA__gtk_radio_button_new_with_label

extern __typeof (gtk_radio_button_new_with_label_from_widget) IA__gtk_radio_button_new_with_label_from_widget __attribute((visibility("hidden")));
#define gtk_radio_button_new_with_label_from_widget IA__gtk_radio_button_new_with_label_from_widget

extern __typeof (gtk_radio_button_new_with_mnemonic) IA__gtk_radio_button_new_with_mnemonic __attribute((visibility("hidden")));
#define gtk_radio_button_new_with_mnemonic IA__gtk_radio_button_new_with_mnemonic

extern __typeof (gtk_radio_button_new_with_mnemonic_from_widget) IA__gtk_radio_button_new_with_mnemonic_from_widget __attribute((visibility("hidden")));
#define gtk_radio_button_new_with_mnemonic_from_widget IA__gtk_radio_button_new_with_mnemonic_from_widget

extern __typeof (gtk_radio_button_set_group) IA__gtk_radio_button_set_group __attribute((visibility("hidden")));
#define gtk_radio_button_set_group IA__gtk_radio_button_set_group

#endif
#endif
#if IN_HEADER(__GTK_RADIO_MENU_ITEM_H__)
#if IN_FILE(__GTK_RADIO_MENU_ITEM_C__)
extern __typeof (gtk_radio_menu_item_get_group) IA__gtk_radio_menu_item_get_group __attribute((visibility("hidden")));
#define gtk_radio_menu_item_get_group IA__gtk_radio_menu_item_get_group

extern __typeof (gtk_radio_menu_item_get_type) IA__gtk_radio_menu_item_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_radio_menu_item_get_type IA__gtk_radio_menu_item_get_type

extern __typeof (gtk_radio_menu_item_new) IA__gtk_radio_menu_item_new __attribute((visibility("hidden")));
#define gtk_radio_menu_item_new IA__gtk_radio_menu_item_new

extern __typeof (gtk_radio_menu_item_new_from_widget) IA__gtk_radio_menu_item_new_from_widget __attribute((visibility("hidden")));
#define gtk_radio_menu_item_new_from_widget IA__gtk_radio_menu_item_new_from_widget

extern __typeof (gtk_radio_menu_item_new_with_label) IA__gtk_radio_menu_item_new_with_label __attribute((visibility("hidden")));
#define gtk_radio_menu_item_new_with_label IA__gtk_radio_menu_item_new_with_label

extern __typeof (gtk_radio_menu_item_new_with_label_from_widget) IA__gtk_radio_menu_item_new_with_label_from_widget __attribute((visibility("hidden")));
#define gtk_radio_menu_item_new_with_label_from_widget IA__gtk_radio_menu_item_new_with_label_from_widget

extern __typeof (gtk_radio_menu_item_new_with_mnemonic) IA__gtk_radio_menu_item_new_with_mnemonic __attribute((visibility("hidden")));
#define gtk_radio_menu_item_new_with_mnemonic IA__gtk_radio_menu_item_new_with_mnemonic

extern __typeof (gtk_radio_menu_item_new_with_mnemonic_from_widget) IA__gtk_radio_menu_item_new_with_mnemonic_from_widget __attribute((visibility("hidden")));
#define gtk_radio_menu_item_new_with_mnemonic_from_widget IA__gtk_radio_menu_item_new_with_mnemonic_from_widget

extern __typeof (gtk_radio_menu_item_set_group) IA__gtk_radio_menu_item_set_group __attribute((visibility("hidden")));
#define gtk_radio_menu_item_set_group IA__gtk_radio_menu_item_set_group

#endif
#endif
#if IN_HEADER(__GTK_RADIO_TOOL_BUTTON_H__)
#if IN_FILE(__GTK_RADIO_TOOL_BUTTON_C__)
extern __typeof (gtk_radio_tool_button_get_group) IA__gtk_radio_tool_button_get_group __attribute((visibility("hidden")));
#define gtk_radio_tool_button_get_group IA__gtk_radio_tool_button_get_group

extern __typeof (gtk_radio_tool_button_get_type) IA__gtk_radio_tool_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_radio_tool_button_get_type IA__gtk_radio_tool_button_get_type

extern __typeof (gtk_radio_tool_button_new) IA__gtk_radio_tool_button_new __attribute((visibility("hidden")));
#define gtk_radio_tool_button_new IA__gtk_radio_tool_button_new

extern __typeof (gtk_radio_tool_button_new_from_stock) IA__gtk_radio_tool_button_new_from_stock __attribute((visibility("hidden")));
#define gtk_radio_tool_button_new_from_stock IA__gtk_radio_tool_button_new_from_stock

extern __typeof (gtk_radio_tool_button_new_from_widget) IA__gtk_radio_tool_button_new_from_widget __attribute((visibility("hidden")));
#define gtk_radio_tool_button_new_from_widget IA__gtk_radio_tool_button_new_from_widget

extern __typeof (gtk_radio_tool_button_new_with_stock_from_widget) IA__gtk_radio_tool_button_new_with_stock_from_widget __attribute((visibility("hidden")));
#define gtk_radio_tool_button_new_with_stock_from_widget IA__gtk_radio_tool_button_new_with_stock_from_widget

extern __typeof (gtk_radio_tool_button_set_group) IA__gtk_radio_tool_button_set_group __attribute((visibility("hidden")));
#define gtk_radio_tool_button_set_group IA__gtk_radio_tool_button_set_group

#endif
#endif
#if IN_HEADER(__GTK_RANGE_H__)
#if IN_FILE(__GTK_RANGE_C__)
extern __typeof (gtk_range_get_adjustment) IA__gtk_range_get_adjustment __attribute((visibility("hidden")));
#define gtk_range_get_adjustment IA__gtk_range_get_adjustment

extern __typeof (gtk_range_get_fill_level) IA__gtk_range_get_fill_level __attribute((visibility("hidden")));
#define gtk_range_get_fill_level IA__gtk_range_get_fill_level

extern __typeof (gtk_range_get_flippable) IA__gtk_range_get_flippable __attribute((visibility("hidden")));
#define gtk_range_get_flippable IA__gtk_range_get_flippable

extern __typeof (gtk_range_get_inverted) IA__gtk_range_get_inverted __attribute((visibility("hidden")));
#define gtk_range_get_inverted IA__gtk_range_get_inverted

extern __typeof (gtk_range_get_lower_stepper_sensitivity) IA__gtk_range_get_lower_stepper_sensitivity __attribute((visibility("hidden")));
#define gtk_range_get_lower_stepper_sensitivity IA__gtk_range_get_lower_stepper_sensitivity

extern __typeof (gtk_range_get_min_slider_size) IA__gtk_range_get_min_slider_size __attribute((visibility("hidden")));
#define gtk_range_get_min_slider_size IA__gtk_range_get_min_slider_size

extern __typeof (gtk_range_get_range_rect) IA__gtk_range_get_range_rect __attribute((visibility("hidden")));
#define gtk_range_get_range_rect IA__gtk_range_get_range_rect

extern __typeof (gtk_range_get_restrict_to_fill_level) IA__gtk_range_get_restrict_to_fill_level __attribute((visibility("hidden")));
#define gtk_range_get_restrict_to_fill_level IA__gtk_range_get_restrict_to_fill_level

extern __typeof (gtk_range_get_show_fill_level) IA__gtk_range_get_show_fill_level __attribute((visibility("hidden")));
#define gtk_range_get_show_fill_level IA__gtk_range_get_show_fill_level

extern __typeof (gtk_range_get_slider_range) IA__gtk_range_get_slider_range __attribute((visibility("hidden")));
#define gtk_range_get_slider_range IA__gtk_range_get_slider_range

extern __typeof (gtk_range_get_slider_size_fixed) IA__gtk_range_get_slider_size_fixed __attribute((visibility("hidden")));
#define gtk_range_get_slider_size_fixed IA__gtk_range_get_slider_size_fixed

extern __typeof (gtk_range_get_type) IA__gtk_range_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_range_get_type IA__gtk_range_get_type

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_range_get_update_policy) IA__gtk_range_get_update_policy __attribute((visibility("hidden")));
#define gtk_range_get_update_policy IA__gtk_range_get_update_policy

#endif
extern __typeof (gtk_range_get_upper_stepper_sensitivity) IA__gtk_range_get_upper_stepper_sensitivity __attribute((visibility("hidden")));
#define gtk_range_get_upper_stepper_sensitivity IA__gtk_range_get_upper_stepper_sensitivity

extern __typeof (gtk_range_get_value) IA__gtk_range_get_value __attribute((visibility("hidden")));
#define gtk_range_get_value IA__gtk_range_get_value

extern __typeof (gtk_range_set_adjustment) IA__gtk_range_set_adjustment __attribute((visibility("hidden")));
#define gtk_range_set_adjustment IA__gtk_range_set_adjustment

extern __typeof (gtk_range_set_fill_level) IA__gtk_range_set_fill_level __attribute((visibility("hidden")));
#define gtk_range_set_fill_level IA__gtk_range_set_fill_level

extern __typeof (gtk_range_set_flippable) IA__gtk_range_set_flippable __attribute((visibility("hidden")));
#define gtk_range_set_flippable IA__gtk_range_set_flippable

extern __typeof (gtk_range_set_increments) IA__gtk_range_set_increments __attribute((visibility("hidden")));
#define gtk_range_set_increments IA__gtk_range_set_increments

extern __typeof (gtk_range_set_inverted) IA__gtk_range_set_inverted __attribute((visibility("hidden")));
#define gtk_range_set_inverted IA__gtk_range_set_inverted

extern __typeof (gtk_range_set_lower_stepper_sensitivity) IA__gtk_range_set_lower_stepper_sensitivity __attribute((visibility("hidden")));
#define gtk_range_set_lower_stepper_sensitivity IA__gtk_range_set_lower_stepper_sensitivity

extern __typeof (gtk_range_set_min_slider_size) IA__gtk_range_set_min_slider_size __attribute((visibility("hidden")));
#define gtk_range_set_min_slider_size IA__gtk_range_set_min_slider_size

extern __typeof (gtk_range_set_range) IA__gtk_range_set_range __attribute((visibility("hidden")));
#define gtk_range_set_range IA__gtk_range_set_range

extern __typeof (gtk_range_set_restrict_to_fill_level) IA__gtk_range_set_restrict_to_fill_level __attribute((visibility("hidden")));
#define gtk_range_set_restrict_to_fill_level IA__gtk_range_set_restrict_to_fill_level

extern __typeof (gtk_range_set_show_fill_level) IA__gtk_range_set_show_fill_level __attribute((visibility("hidden")));
#define gtk_range_set_show_fill_level IA__gtk_range_set_show_fill_level

extern __typeof (gtk_range_set_slider_size_fixed) IA__gtk_range_set_slider_size_fixed __attribute((visibility("hidden")));
#define gtk_range_set_slider_size_fixed IA__gtk_range_set_slider_size_fixed

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_range_set_update_policy) IA__gtk_range_set_update_policy __attribute((visibility("hidden")));
#define gtk_range_set_update_policy IA__gtk_range_set_update_policy

#endif
extern __typeof (gtk_range_set_upper_stepper_sensitivity) IA__gtk_range_set_upper_stepper_sensitivity __attribute((visibility("hidden")));
#define gtk_range_set_upper_stepper_sensitivity IA__gtk_range_set_upper_stepper_sensitivity

extern __typeof (gtk_range_set_value) IA__gtk_range_set_value __attribute((visibility("hidden")));
#define gtk_range_set_value IA__gtk_range_set_value

extern __typeof (gtk_range_set_round_digits) IA__gtk_range_set_round_digits __attribute((visibility("hidden")));
#define gtk_range_set_round_digits IA__gtk_range_set_round_digits

extern __typeof (gtk_range_get_round_digits) IA__gtk_range_get_round_digits __attribute((visibility("hidden")));
#define gtk_range_get_round_digits IA__gtk_range_get_round_digits

#endif
#endif
#if IN_HEADER(__GTK_RC_H__)
#if IN_FILE(__GTK_RC_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_rc_add_class_style) IA__gtk_rc_add_class_style __attribute((visibility("hidden")));
#define gtk_rc_add_class_style IA__gtk_rc_add_class_style

extern __typeof (gtk_rc_add_widget_class_style) IA__gtk_rc_add_widget_class_style __attribute((visibility("hidden")));
#define gtk_rc_add_widget_class_style IA__gtk_rc_add_widget_class_style

extern __typeof (gtk_rc_add_widget_name_style) IA__gtk_rc_add_widget_name_style __attribute((visibility("hidden")));
#define gtk_rc_add_widget_name_style IA__gtk_rc_add_widget_name_style

#endif
extern __typeof (gtk_rc_find_module_in_path) IA__gtk_rc_find_module_in_path __attribute((visibility("hidden")));
#define gtk_rc_find_module_in_path IA__gtk_rc_find_module_in_path

extern __typeof (gtk_rc_find_pixmap_in_path) IA__gtk_rc_find_pixmap_in_path __attribute((visibility("hidden")));
#define gtk_rc_find_pixmap_in_path IA__gtk_rc_find_pixmap_in_path

extern __typeof (gtk_rc_get_im_module_file) IA__gtk_rc_get_im_module_file __attribute((visibility("hidden")));
#define gtk_rc_get_im_module_file IA__gtk_rc_get_im_module_file

extern __typeof (gtk_rc_get_im_module_path) IA__gtk_rc_get_im_module_path __attribute((visibility("hidden")));
#define gtk_rc_get_im_module_path IA__gtk_rc_get_im_module_path

extern __typeof (gtk_rc_get_module_dir) IA__gtk_rc_get_module_dir __attribute((visibility("hidden")));
#define gtk_rc_get_module_dir IA__gtk_rc_get_module_dir

extern __typeof (gtk_rc_get_style) IA__gtk_rc_get_style __attribute((visibility("hidden")));
#define gtk_rc_get_style IA__gtk_rc_get_style

extern __typeof (gtk_rc_get_style_by_paths) IA__gtk_rc_get_style_by_paths __attribute((visibility("hidden")));
#define gtk_rc_get_style_by_paths IA__gtk_rc_get_style_by_paths

#ifndef _WIN64
extern __typeof (gtk_rc_parse) IA__gtk_rc_parse __attribute((visibility("hidden")));
#define gtk_rc_parse IA__gtk_rc_parse

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_rc_parse_utf8) IA__gtk_rc_parse_utf8 __attribute((visibility("hidden")));
#define gtk_rc_parse_utf8 IA__gtk_rc_parse_utf8

#endif
extern __typeof (gtk_rc_parse_color) IA__gtk_rc_parse_color __attribute((visibility("hidden")));
#define gtk_rc_parse_color IA__gtk_rc_parse_color

extern __typeof (gtk_rc_parse_color_full) IA__gtk_rc_parse_color_full __attribute((visibility("hidden")));
#define gtk_rc_parse_color_full IA__gtk_rc_parse_color_full

extern __typeof (gtk_rc_parse_priority) IA__gtk_rc_parse_priority __attribute((visibility("hidden")));
#define gtk_rc_parse_priority IA__gtk_rc_parse_priority

extern __typeof (gtk_rc_parse_state) IA__gtk_rc_parse_state __attribute((visibility("hidden")));
#define gtk_rc_parse_state IA__gtk_rc_parse_state

extern __typeof (gtk_rc_parse_string) IA__gtk_rc_parse_string __attribute((visibility("hidden")));
#define gtk_rc_parse_string IA__gtk_rc_parse_string

extern __typeof (gtk_rc_reparse_all) IA__gtk_rc_reparse_all __attribute((visibility("hidden")));
#define gtk_rc_reparse_all IA__gtk_rc_reparse_all

extern __typeof (gtk_rc_reparse_all_for_settings) IA__gtk_rc_reparse_all_for_settings __attribute((visibility("hidden")));
#define gtk_rc_reparse_all_for_settings IA__gtk_rc_reparse_all_for_settings

extern __typeof (gtk_rc_reset_styles) IA__gtk_rc_reset_styles __attribute((visibility("hidden")));
#define gtk_rc_reset_styles IA__gtk_rc_reset_styles

extern __typeof (gtk_rc_scanner_new) IA__gtk_rc_scanner_new __attribute((visibility("hidden")));
#define gtk_rc_scanner_new IA__gtk_rc_scanner_new

extern __typeof (gtk_rc_style_copy) IA__gtk_rc_style_copy __attribute((visibility("hidden")));
#define gtk_rc_style_copy IA__gtk_rc_style_copy

extern __typeof (gtk_rc_style_get_type) IA__gtk_rc_style_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_rc_style_get_type IA__gtk_rc_style_get_type

extern __typeof (gtk_rc_style_new) IA__gtk_rc_style_new __attribute((visibility("hidden")));
#define gtk_rc_style_new IA__gtk_rc_style_new

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_rc_style_ref) IA__gtk_rc_style_ref __attribute((visibility("hidden")));
#define gtk_rc_style_ref IA__gtk_rc_style_ref

extern __typeof (gtk_rc_style_unref) IA__gtk_rc_style_unref __attribute((visibility("hidden")));
#define gtk_rc_style_unref IA__gtk_rc_style_unref

#endif
#endif
#endif
#if IN_HEADER(__GTK_RECENT_ACTION_H__)
#if IN_FILE(__GTK_RECENT_ACTION_C__)
extern __typeof (gtk_recent_action_get_type) IA__gtk_recent_action_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_action_get_type IA__gtk_recent_action_get_type

extern __typeof (gtk_recent_action_new) IA__gtk_recent_action_new __attribute((visibility("hidden")));
#define gtk_recent_action_new IA__gtk_recent_action_new

extern __typeof (gtk_recent_action_new_for_manager) IA__gtk_recent_action_new_for_manager __attribute((visibility("hidden")));
#define gtk_recent_action_new_for_manager IA__gtk_recent_action_new_for_manager

extern __typeof (gtk_recent_action_set_show_numbers) IA__gtk_recent_action_set_show_numbers __attribute((visibility("hidden")));
#define gtk_recent_action_set_show_numbers IA__gtk_recent_action_set_show_numbers

extern __typeof (gtk_recent_action_get_show_numbers) IA__gtk_recent_action_get_show_numbers __attribute((visibility("hidden")));
#define gtk_recent_action_get_show_numbers IA__gtk_recent_action_get_show_numbers

#endif
#endif
#if IN_HEADER(__GTK_RECENT_CHOOSER_H__)
#if IN_FILE(__GTK_RECENT_CHOOSER_C__)
extern __typeof (gtk_recent_chooser_set_show_private) IA__gtk_recent_chooser_set_show_private __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_show_private IA__gtk_recent_chooser_set_show_private

extern __typeof (gtk_recent_chooser_get_show_private) IA__gtk_recent_chooser_get_show_private __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_show_private IA__gtk_recent_chooser_get_show_private

extern __typeof (gtk_recent_chooser_set_show_not_found) IA__gtk_recent_chooser_set_show_not_found __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_show_not_found IA__gtk_recent_chooser_set_show_not_found

extern __typeof (gtk_recent_chooser_get_show_not_found) IA__gtk_recent_chooser_get_show_not_found __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_show_not_found IA__gtk_recent_chooser_get_show_not_found

extern __typeof (gtk_recent_chooser_set_show_icons) IA__gtk_recent_chooser_set_show_icons __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_show_icons IA__gtk_recent_chooser_set_show_icons

extern __typeof (gtk_recent_chooser_get_show_icons) IA__gtk_recent_chooser_get_show_icons __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_show_icons IA__gtk_recent_chooser_get_show_icons

extern __typeof (gtk_recent_chooser_set_select_multiple) IA__gtk_recent_chooser_set_select_multiple __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_select_multiple IA__gtk_recent_chooser_set_select_multiple

extern __typeof (gtk_recent_chooser_get_select_multiple) IA__gtk_recent_chooser_get_select_multiple __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_select_multiple IA__gtk_recent_chooser_get_select_multiple

extern __typeof (gtk_recent_chooser_set_local_only) IA__gtk_recent_chooser_set_local_only __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_local_only IA__gtk_recent_chooser_set_local_only

extern __typeof (gtk_recent_chooser_get_local_only) IA__gtk_recent_chooser_get_local_only __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_local_only IA__gtk_recent_chooser_get_local_only

extern __typeof (gtk_recent_chooser_set_limit) IA__gtk_recent_chooser_set_limit __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_limit IA__gtk_recent_chooser_set_limit

extern __typeof (gtk_recent_chooser_get_limit) IA__gtk_recent_chooser_get_limit __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_limit IA__gtk_recent_chooser_get_limit

extern __typeof (gtk_recent_chooser_set_show_tips) IA__gtk_recent_chooser_set_show_tips __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_show_tips IA__gtk_recent_chooser_set_show_tips

extern __typeof (gtk_recent_chooser_get_show_tips) IA__gtk_recent_chooser_get_show_tips __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_show_tips IA__gtk_recent_chooser_get_show_tips

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_recent_chooser_set_show_numbers) IA__gtk_recent_chooser_set_show_numbers __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_show_numbers IA__gtk_recent_chooser_set_show_numbers

extern __typeof (gtk_recent_chooser_get_show_numbers) IA__gtk_recent_chooser_get_show_numbers __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_show_numbers IA__gtk_recent_chooser_get_show_numbers

#endif
extern __typeof (gtk_recent_chooser_set_sort_type) IA__gtk_recent_chooser_set_sort_type __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_sort_type IA__gtk_recent_chooser_set_sort_type

extern __typeof (gtk_recent_chooser_get_sort_type) IA__gtk_recent_chooser_get_sort_type __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_sort_type IA__gtk_recent_chooser_get_sort_type

extern __typeof (gtk_recent_chooser_set_sort_func) IA__gtk_recent_chooser_set_sort_func __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_sort_func IA__gtk_recent_chooser_set_sort_func

extern __typeof (gtk_recent_chooser_set_current_uri) IA__gtk_recent_chooser_set_current_uri __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_current_uri IA__gtk_recent_chooser_set_current_uri

extern __typeof (gtk_recent_chooser_get_current_uri) IA__gtk_recent_chooser_get_current_uri __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_current_uri IA__gtk_recent_chooser_get_current_uri

extern __typeof (gtk_recent_chooser_get_current_item) IA__gtk_recent_chooser_get_current_item __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_current_item IA__gtk_recent_chooser_get_current_item

extern __typeof (gtk_recent_chooser_select_uri) IA__gtk_recent_chooser_select_uri __attribute((visibility("hidden")));
#define gtk_recent_chooser_select_uri IA__gtk_recent_chooser_select_uri

extern __typeof (gtk_recent_chooser_unselect_uri) IA__gtk_recent_chooser_unselect_uri __attribute((visibility("hidden")));
#define gtk_recent_chooser_unselect_uri IA__gtk_recent_chooser_unselect_uri

extern __typeof (gtk_recent_chooser_select_all) IA__gtk_recent_chooser_select_all __attribute((visibility("hidden")));
#define gtk_recent_chooser_select_all IA__gtk_recent_chooser_select_all

extern __typeof (gtk_recent_chooser_unselect_all) IA__gtk_recent_chooser_unselect_all __attribute((visibility("hidden")));
#define gtk_recent_chooser_unselect_all IA__gtk_recent_chooser_unselect_all

extern __typeof (gtk_recent_chooser_get_items) IA__gtk_recent_chooser_get_items __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_items IA__gtk_recent_chooser_get_items

extern __typeof (gtk_recent_chooser_get_uris) IA__gtk_recent_chooser_get_uris __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_uris IA__gtk_recent_chooser_get_uris

extern __typeof (gtk_recent_chooser_add_filter) IA__gtk_recent_chooser_add_filter __attribute((visibility("hidden")));
#define gtk_recent_chooser_add_filter IA__gtk_recent_chooser_add_filter

extern __typeof (gtk_recent_chooser_remove_filter) IA__gtk_recent_chooser_remove_filter __attribute((visibility("hidden")));
#define gtk_recent_chooser_remove_filter IA__gtk_recent_chooser_remove_filter

extern __typeof (gtk_recent_chooser_list_filters) IA__gtk_recent_chooser_list_filters __attribute((visibility("hidden")));
#define gtk_recent_chooser_list_filters IA__gtk_recent_chooser_list_filters

extern __typeof (gtk_recent_chooser_set_filter) IA__gtk_recent_chooser_set_filter __attribute((visibility("hidden")));
#define gtk_recent_chooser_set_filter IA__gtk_recent_chooser_set_filter

extern __typeof (gtk_recent_chooser_get_filter) IA__gtk_recent_chooser_get_filter __attribute((visibility("hidden")));
#define gtk_recent_chooser_get_filter IA__gtk_recent_chooser_get_filter

extern __typeof (gtk_recent_chooser_get_type) IA__gtk_recent_chooser_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_chooser_get_type IA__gtk_recent_chooser_get_type

extern __typeof (gtk_recent_chooser_error_quark) IA__gtk_recent_chooser_error_quark __attribute((visibility("hidden")));
#define gtk_recent_chooser_error_quark IA__gtk_recent_chooser_error_quark

#endif
#endif
#if IN_HEADER(__GTK_RECENT_CHOOSER_DIALOG_H__)
#if IN_FILE(__GTK_RECENT_CHOOSER_DIALOG_C__)
extern __typeof (gtk_recent_chooser_dialog_get_type) IA__gtk_recent_chooser_dialog_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_chooser_dialog_get_type IA__gtk_recent_chooser_dialog_get_type

extern __typeof (gtk_recent_chooser_dialog_new) IA__gtk_recent_chooser_dialog_new __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_recent_chooser_dialog_new IA__gtk_recent_chooser_dialog_new

extern __typeof (gtk_recent_chooser_dialog_new_for_manager) IA__gtk_recent_chooser_dialog_new_for_manager __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_recent_chooser_dialog_new_for_manager IA__gtk_recent_chooser_dialog_new_for_manager

#endif
#endif
#if IN_HEADER(__GTK_RECENT_CHOOSER_MENU_H__)
#if IN_FILE(__GTK_RECENT_CHOOSER_MENU_C__)
extern __typeof (gtk_recent_chooser_menu_get_type) IA__gtk_recent_chooser_menu_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_chooser_menu_get_type IA__gtk_recent_chooser_menu_get_type

extern __typeof (gtk_recent_chooser_menu_new) IA__gtk_recent_chooser_menu_new __attribute((visibility("hidden")));
#define gtk_recent_chooser_menu_new IA__gtk_recent_chooser_menu_new

extern __typeof (gtk_recent_chooser_menu_new_for_manager) IA__gtk_recent_chooser_menu_new_for_manager __attribute((visibility("hidden")));
#define gtk_recent_chooser_menu_new_for_manager IA__gtk_recent_chooser_menu_new_for_manager

extern __typeof (gtk_recent_chooser_menu_get_show_numbers) IA__gtk_recent_chooser_menu_get_show_numbers __attribute((visibility("hidden")));
#define gtk_recent_chooser_menu_get_show_numbers IA__gtk_recent_chooser_menu_get_show_numbers

extern __typeof (gtk_recent_chooser_menu_set_show_numbers) IA__gtk_recent_chooser_menu_set_show_numbers __attribute((visibility("hidden")));
#define gtk_recent_chooser_menu_set_show_numbers IA__gtk_recent_chooser_menu_set_show_numbers

#endif
#endif
#if IN_HEADER(__GTK_RECENT_CHOOSER_WIDGET_H__)
#if IN_FILE(__GTK_RECENT_CHOOSER_WIDGET_C__)
extern __typeof (gtk_recent_chooser_widget_get_type) IA__gtk_recent_chooser_widget_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_chooser_widget_get_type IA__gtk_recent_chooser_widget_get_type

extern __typeof (gtk_recent_chooser_widget_new) IA__gtk_recent_chooser_widget_new __attribute((visibility("hidden")));
#define gtk_recent_chooser_widget_new IA__gtk_recent_chooser_widget_new

extern __typeof (gtk_recent_chooser_widget_new_for_manager) IA__gtk_recent_chooser_widget_new_for_manager __attribute((visibility("hidden")));
#define gtk_recent_chooser_widget_new_for_manager IA__gtk_recent_chooser_widget_new_for_manager

#endif
#endif
#if IN_HEADER(__GTK_RECENT_FILTER_H__)
#if IN_FILE(__GTK_RECENT_FILTER_C__)
extern __typeof (gtk_recent_filter_get_type) IA__gtk_recent_filter_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_filter_get_type IA__gtk_recent_filter_get_type

extern __typeof (gtk_recent_filter_new) IA__gtk_recent_filter_new __attribute((visibility("hidden")));
#define gtk_recent_filter_new IA__gtk_recent_filter_new

extern __typeof (gtk_recent_filter_set_name) IA__gtk_recent_filter_set_name __attribute((visibility("hidden")));
#define gtk_recent_filter_set_name IA__gtk_recent_filter_set_name

extern __typeof (gtk_recent_filter_get_name) IA__gtk_recent_filter_get_name __attribute((visibility("hidden")));
#define gtk_recent_filter_get_name IA__gtk_recent_filter_get_name

extern __typeof (gtk_recent_filter_add_mime_type) IA__gtk_recent_filter_add_mime_type __attribute((visibility("hidden")));
#define gtk_recent_filter_add_mime_type IA__gtk_recent_filter_add_mime_type

extern __typeof (gtk_recent_filter_add_pattern) IA__gtk_recent_filter_add_pattern __attribute((visibility("hidden")));
#define gtk_recent_filter_add_pattern IA__gtk_recent_filter_add_pattern

extern __typeof (gtk_recent_filter_add_pixbuf_formats) IA__gtk_recent_filter_add_pixbuf_formats __attribute((visibility("hidden")));
#define gtk_recent_filter_add_pixbuf_formats IA__gtk_recent_filter_add_pixbuf_formats

extern __typeof (gtk_recent_filter_add_application) IA__gtk_recent_filter_add_application __attribute((visibility("hidden")));
#define gtk_recent_filter_add_application IA__gtk_recent_filter_add_application

extern __typeof (gtk_recent_filter_add_group) IA__gtk_recent_filter_add_group __attribute((visibility("hidden")));
#define gtk_recent_filter_add_group IA__gtk_recent_filter_add_group

extern __typeof (gtk_recent_filter_add_age) IA__gtk_recent_filter_add_age __attribute((visibility("hidden")));
#define gtk_recent_filter_add_age IA__gtk_recent_filter_add_age

extern __typeof (gtk_recent_filter_add_custom) IA__gtk_recent_filter_add_custom __attribute((visibility("hidden")));
#define gtk_recent_filter_add_custom IA__gtk_recent_filter_add_custom

extern __typeof (gtk_recent_filter_get_needed) IA__gtk_recent_filter_get_needed __attribute((visibility("hidden")));
#define gtk_recent_filter_get_needed IA__gtk_recent_filter_get_needed

extern __typeof (gtk_recent_filter_filter) IA__gtk_recent_filter_filter __attribute((visibility("hidden")));
#define gtk_recent_filter_filter IA__gtk_recent_filter_filter

#endif
#endif
#if IN_HEADER(__GTK_RECENT_MANAGER_H__)
#if IN_FILE(__GTK_RECENT_MANAGER_C__)
extern __typeof (gtk_recent_manager_error_quark) IA__gtk_recent_manager_error_quark __attribute((visibility("hidden")));
#define gtk_recent_manager_error_quark IA__gtk_recent_manager_error_quark

extern __typeof (gtk_recent_manager_get_type) IA__gtk_recent_manager_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_manager_get_type IA__gtk_recent_manager_get_type

extern __typeof (gtk_recent_manager_new) IA__gtk_recent_manager_new __attribute((visibility("hidden")));
#define gtk_recent_manager_new IA__gtk_recent_manager_new

extern __typeof (gtk_recent_manager_get_items) IA__gtk_recent_manager_get_items __attribute((visibility("hidden")));
#define gtk_recent_manager_get_items IA__gtk_recent_manager_get_items

extern __typeof (gtk_recent_manager_get_default) IA__gtk_recent_manager_get_default __attribute((visibility("hidden")));
#define gtk_recent_manager_get_default IA__gtk_recent_manager_get_default

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_recent_manager_get_for_screen) IA__gtk_recent_manager_get_for_screen __attribute((visibility("hidden")));
#define gtk_recent_manager_get_for_screen IA__gtk_recent_manager_get_for_screen

extern __typeof (gtk_recent_manager_set_screen) IA__gtk_recent_manager_set_screen __attribute((visibility("hidden")));
#define gtk_recent_manager_set_screen IA__gtk_recent_manager_set_screen

#endif
extern __typeof (gtk_recent_manager_add_item) IA__gtk_recent_manager_add_item __attribute((visibility("hidden")));
#define gtk_recent_manager_add_item IA__gtk_recent_manager_add_item

extern __typeof (gtk_recent_manager_add_full) IA__gtk_recent_manager_add_full __attribute((visibility("hidden")));
#define gtk_recent_manager_add_full IA__gtk_recent_manager_add_full

extern __typeof (gtk_recent_manager_remove_item) IA__gtk_recent_manager_remove_item __attribute((visibility("hidden")));
#define gtk_recent_manager_remove_item IA__gtk_recent_manager_remove_item

extern __typeof (gtk_recent_manager_lookup_item) IA__gtk_recent_manager_lookup_item __attribute((visibility("hidden")));
#define gtk_recent_manager_lookup_item IA__gtk_recent_manager_lookup_item

extern __typeof (gtk_recent_manager_has_item) IA__gtk_recent_manager_has_item __attribute((visibility("hidden")));
#define gtk_recent_manager_has_item IA__gtk_recent_manager_has_item

extern __typeof (gtk_recent_manager_move_item) IA__gtk_recent_manager_move_item __attribute((visibility("hidden")));
#define gtk_recent_manager_move_item IA__gtk_recent_manager_move_item

extern __typeof (gtk_recent_manager_set_limit) IA__gtk_recent_manager_set_limit __attribute((visibility("hidden")));
#define gtk_recent_manager_set_limit IA__gtk_recent_manager_set_limit

extern __typeof (gtk_recent_manager_get_limit) IA__gtk_recent_manager_get_limit __attribute((visibility("hidden")));
#define gtk_recent_manager_get_limit IA__gtk_recent_manager_get_limit

extern __typeof (gtk_recent_manager_purge_items) IA__gtk_recent_manager_purge_items __attribute((visibility("hidden")));
#define gtk_recent_manager_purge_items IA__gtk_recent_manager_purge_items

extern __typeof (gtk_recent_info_get_type) IA__gtk_recent_info_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_recent_info_get_type IA__gtk_recent_info_get_type

extern __typeof (gtk_recent_info_ref) IA__gtk_recent_info_ref __attribute((visibility("hidden")));
#define gtk_recent_info_ref IA__gtk_recent_info_ref

extern __typeof (gtk_recent_info_unref) IA__gtk_recent_info_unref __attribute((visibility("hidden")));
#define gtk_recent_info_unref IA__gtk_recent_info_unref

extern __typeof (gtk_recent_info_get_uri) IA__gtk_recent_info_get_uri __attribute((visibility("hidden")));
#define gtk_recent_info_get_uri IA__gtk_recent_info_get_uri

extern __typeof (gtk_recent_info_get_display_name) IA__gtk_recent_info_get_display_name __attribute((visibility("hidden")));
#define gtk_recent_info_get_display_name IA__gtk_recent_info_get_display_name

extern __typeof (gtk_recent_info_get_description) IA__gtk_recent_info_get_description __attribute((visibility("hidden")));
#define gtk_recent_info_get_description IA__gtk_recent_info_get_description

extern __typeof (gtk_recent_info_get_mime_type) IA__gtk_recent_info_get_mime_type __attribute((visibility("hidden")));
#define gtk_recent_info_get_mime_type IA__gtk_recent_info_get_mime_type

extern __typeof (gtk_recent_info_get_added) IA__gtk_recent_info_get_added __attribute((visibility("hidden")));
#define gtk_recent_info_get_added IA__gtk_recent_info_get_added

extern __typeof (gtk_recent_info_get_modified) IA__gtk_recent_info_get_modified __attribute((visibility("hidden")));
#define gtk_recent_info_get_modified IA__gtk_recent_info_get_modified

extern __typeof (gtk_recent_info_get_visited) IA__gtk_recent_info_get_visited __attribute((visibility("hidden")));
#define gtk_recent_info_get_visited IA__gtk_recent_info_get_visited

extern __typeof (gtk_recent_info_get_private_hint) IA__gtk_recent_info_get_private_hint __attribute((visibility("hidden")));
#define gtk_recent_info_get_private_hint IA__gtk_recent_info_get_private_hint

extern __typeof (gtk_recent_info_get_application_info) IA__gtk_recent_info_get_application_info __attribute((visibility("hidden")));
#define gtk_recent_info_get_application_info IA__gtk_recent_info_get_application_info

extern __typeof (gtk_recent_info_get_applications) IA__gtk_recent_info_get_applications __attribute((visibility("hidden"))) G_GNUC_MALLOC;
#define gtk_recent_info_get_applications IA__gtk_recent_info_get_applications

extern __typeof (gtk_recent_info_last_application) IA__gtk_recent_info_last_application __attribute((visibility("hidden"))) G_GNUC_MALLOC;
#define gtk_recent_info_last_application IA__gtk_recent_info_last_application

extern __typeof (gtk_recent_info_has_application) IA__gtk_recent_info_has_application __attribute((visibility("hidden")));
#define gtk_recent_info_has_application IA__gtk_recent_info_has_application

extern __typeof (gtk_recent_info_get_groups) IA__gtk_recent_info_get_groups __attribute((visibility("hidden"))) G_GNUC_MALLOC;
#define gtk_recent_info_get_groups IA__gtk_recent_info_get_groups

extern __typeof (gtk_recent_info_has_group) IA__gtk_recent_info_has_group __attribute((visibility("hidden")));
#define gtk_recent_info_has_group IA__gtk_recent_info_has_group

extern __typeof (gtk_recent_info_get_icon) IA__gtk_recent_info_get_icon __attribute((visibility("hidden")));
#define gtk_recent_info_get_icon IA__gtk_recent_info_get_icon

extern __typeof (gtk_recent_info_get_short_name) IA__gtk_recent_info_get_short_name __attribute((visibility("hidden"))) G_GNUC_MALLOC;
#define gtk_recent_info_get_short_name IA__gtk_recent_info_get_short_name

extern __typeof (gtk_recent_info_get_uri_display) IA__gtk_recent_info_get_uri_display __attribute((visibility("hidden"))) G_GNUC_MALLOC;
#define gtk_recent_info_get_uri_display IA__gtk_recent_info_get_uri_display

extern __typeof (gtk_recent_info_get_age) IA__gtk_recent_info_get_age __attribute((visibility("hidden")));
#define gtk_recent_info_get_age IA__gtk_recent_info_get_age

extern __typeof (gtk_recent_info_is_local) IA__gtk_recent_info_is_local __attribute((visibility("hidden")));
#define gtk_recent_info_is_local IA__gtk_recent_info_is_local

extern __typeof (gtk_recent_info_exists) IA__gtk_recent_info_exists __attribute((visibility("hidden")));
#define gtk_recent_info_exists IA__gtk_recent_info_exists

extern __typeof (gtk_recent_info_match) IA__gtk_recent_info_match __attribute((visibility("hidden")));
#define gtk_recent_info_match IA__gtk_recent_info_match

#endif
#endif
#if IN_HEADER(__GTK_TEST_UTILS_H__)
#if IN_FILE(__GTK_TEST_UTILS_C__)
extern __typeof (gtk_test_create_simple_window) IA__gtk_test_create_simple_window __attribute((visibility("hidden")));
#define gtk_test_create_simple_window IA__gtk_test_create_simple_window

extern __typeof (gtk_test_create_widget) IA__gtk_test_create_widget __attribute((visibility("hidden")));
#define gtk_test_create_widget IA__gtk_test_create_widget

extern __typeof (gtk_test_display_button_window) IA__gtk_test_display_button_window __attribute((visibility("hidden")));
#define gtk_test_display_button_window IA__gtk_test_display_button_window

extern __typeof (gtk_test_find_label) IA__gtk_test_find_label __attribute((visibility("hidden")));
#define gtk_test_find_label IA__gtk_test_find_label

extern __typeof (gtk_test_find_sibling) IA__gtk_test_find_sibling __attribute((visibility("hidden")));
#define gtk_test_find_sibling IA__gtk_test_find_sibling

extern __typeof (gtk_test_find_widget) IA__gtk_test_find_widget __attribute((visibility("hidden")));
#define gtk_test_find_widget IA__gtk_test_find_widget

extern __typeof (gtk_test_init) IA__gtk_test_init __attribute((visibility("hidden")));
#define gtk_test_init IA__gtk_test_init

extern __typeof (gtk_test_register_all_types) IA__gtk_test_register_all_types __attribute((visibility("hidden")));
#define gtk_test_register_all_types IA__gtk_test_register_all_types

extern __typeof (gtk_test_list_all_types) IA__gtk_test_list_all_types __attribute((visibility("hidden")));
#define gtk_test_list_all_types IA__gtk_test_list_all_types

extern __typeof (gtk_test_slider_get_value) IA__gtk_test_slider_get_value __attribute((visibility("hidden")));
#define gtk_test_slider_get_value IA__gtk_test_slider_get_value

extern __typeof (gtk_test_slider_set_perc) IA__gtk_test_slider_set_perc __attribute((visibility("hidden")));
#define gtk_test_slider_set_perc IA__gtk_test_slider_set_perc

extern __typeof (gtk_test_spin_button_click) IA__gtk_test_spin_button_click __attribute((visibility("hidden")));
#define gtk_test_spin_button_click IA__gtk_test_spin_button_click

extern __typeof (gtk_test_text_get) IA__gtk_test_text_get __attribute((visibility("hidden")));
#define gtk_test_text_get IA__gtk_test_text_get

extern __typeof (gtk_test_text_set) IA__gtk_test_text_set __attribute((visibility("hidden")));
#define gtk_test_text_set IA__gtk_test_text_set

extern __typeof (gtk_test_widget_click) IA__gtk_test_widget_click __attribute((visibility("hidden")));
#define gtk_test_widget_click IA__gtk_test_widget_click

extern __typeof (gtk_test_widget_send_key) IA__gtk_test_widget_send_key __attribute((visibility("hidden")));
#define gtk_test_widget_send_key IA__gtk_test_widget_send_key

#endif
#endif
#if IN_HEADER(__GTK_TEXT_BUFFER_RICH_TEXT_H__)
#if IN_FILE(__GTK_TEXT_BUFFER_RICH_TEXT_C__)
extern __typeof (gtk_text_buffer_deserialize) IA__gtk_text_buffer_deserialize __attribute((visibility("hidden")));
#define gtk_text_buffer_deserialize IA__gtk_text_buffer_deserialize

extern __typeof (gtk_text_buffer_deserialize_get_can_create_tags) IA__gtk_text_buffer_deserialize_get_can_create_tags __attribute((visibility("hidden")));
#define gtk_text_buffer_deserialize_get_can_create_tags IA__gtk_text_buffer_deserialize_get_can_create_tags

extern __typeof (gtk_text_buffer_deserialize_set_can_create_tags) IA__gtk_text_buffer_deserialize_set_can_create_tags __attribute((visibility("hidden")));
#define gtk_text_buffer_deserialize_set_can_create_tags IA__gtk_text_buffer_deserialize_set_can_create_tags

extern __typeof (gtk_text_buffer_get_deserialize_formats) IA__gtk_text_buffer_get_deserialize_formats __attribute((visibility("hidden")));
#define gtk_text_buffer_get_deserialize_formats IA__gtk_text_buffer_get_deserialize_formats

extern __typeof (gtk_text_buffer_get_serialize_formats) IA__gtk_text_buffer_get_serialize_formats __attribute((visibility("hidden")));
#define gtk_text_buffer_get_serialize_formats IA__gtk_text_buffer_get_serialize_formats

extern __typeof (gtk_text_buffer_register_deserialize_format) IA__gtk_text_buffer_register_deserialize_format __attribute((visibility("hidden")));
#define gtk_text_buffer_register_deserialize_format IA__gtk_text_buffer_register_deserialize_format

extern __typeof (gtk_text_buffer_register_deserialize_tagset) IA__gtk_text_buffer_register_deserialize_tagset __attribute((visibility("hidden")));
#define gtk_text_buffer_register_deserialize_tagset IA__gtk_text_buffer_register_deserialize_tagset

extern __typeof (gtk_text_buffer_register_serialize_format) IA__gtk_text_buffer_register_serialize_format __attribute((visibility("hidden")));
#define gtk_text_buffer_register_serialize_format IA__gtk_text_buffer_register_serialize_format

extern __typeof (gtk_text_buffer_register_serialize_tagset) IA__gtk_text_buffer_register_serialize_tagset __attribute((visibility("hidden")));
#define gtk_text_buffer_register_serialize_tagset IA__gtk_text_buffer_register_serialize_tagset

extern __typeof (gtk_text_buffer_serialize) IA__gtk_text_buffer_serialize __attribute((visibility("hidden")));
#define gtk_text_buffer_serialize IA__gtk_text_buffer_serialize

extern __typeof (gtk_text_buffer_unregister_deserialize_format) IA__gtk_text_buffer_unregister_deserialize_format __attribute((visibility("hidden")));
#define gtk_text_buffer_unregister_deserialize_format IA__gtk_text_buffer_unregister_deserialize_format

extern __typeof (gtk_text_buffer_unregister_serialize_format) IA__gtk_text_buffer_unregister_serialize_format __attribute((visibility("hidden")));
#define gtk_text_buffer_unregister_serialize_format IA__gtk_text_buffer_unregister_serialize_format

#endif
#endif
#if IN_HEADER(__GTK_RULER_H__)
#if IN_FILE(__GTK_RULER_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_ruler_draw_pos) IA__gtk_ruler_draw_pos __attribute((visibility("hidden")));
#define gtk_ruler_draw_pos IA__gtk_ruler_draw_pos

extern __typeof (gtk_ruler_draw_ticks) IA__gtk_ruler_draw_ticks __attribute((visibility("hidden")));
#define gtk_ruler_draw_ticks IA__gtk_ruler_draw_ticks

extern __typeof (gtk_ruler_get_metric) IA__gtk_ruler_get_metric __attribute((visibility("hidden")));
#define gtk_ruler_get_metric IA__gtk_ruler_get_metric

extern __typeof (gtk_ruler_get_range) IA__gtk_ruler_get_range __attribute((visibility("hidden")));
#define gtk_ruler_get_range IA__gtk_ruler_get_range

extern __typeof (gtk_ruler_get_type) IA__gtk_ruler_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_ruler_get_type IA__gtk_ruler_get_type

extern __typeof (gtk_ruler_set_metric) IA__gtk_ruler_set_metric __attribute((visibility("hidden")));
#define gtk_ruler_set_metric IA__gtk_ruler_set_metric

extern __typeof (gtk_ruler_set_range) IA__gtk_ruler_set_range __attribute((visibility("hidden")));
#define gtk_ruler_set_range IA__gtk_ruler_set_range

#endif
#endif
#endif
#if IN_HEADER(__GTK_SCALE_H__)
#if IN_FILE(__GTK_SCALE_C__)
extern __typeof (gtk_scale_get_digits) IA__gtk_scale_get_digits __attribute((visibility("hidden")));
#define gtk_scale_get_digits IA__gtk_scale_get_digits

extern __typeof (gtk_scale_get_draw_value) IA__gtk_scale_get_draw_value __attribute((visibility("hidden")));
#define gtk_scale_get_draw_value IA__gtk_scale_get_draw_value

extern __typeof (gtk_scale_get_layout) IA__gtk_scale_get_layout __attribute((visibility("hidden")));
#define gtk_scale_get_layout IA__gtk_scale_get_layout

extern __typeof (gtk_scale_get_layout_offsets) IA__gtk_scale_get_layout_offsets __attribute((visibility("hidden")));
#define gtk_scale_get_layout_offsets IA__gtk_scale_get_layout_offsets

extern __typeof (gtk_scale_get_type) IA__gtk_scale_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_scale_get_type IA__gtk_scale_get_type

extern __typeof (gtk_scale_get_value_pos) IA__gtk_scale_get_value_pos __attribute((visibility("hidden")));
#define gtk_scale_get_value_pos IA__gtk_scale_get_value_pos

extern __typeof (gtk_scale_set_digits) IA__gtk_scale_set_digits __attribute((visibility("hidden")));
#define gtk_scale_set_digits IA__gtk_scale_set_digits

extern __typeof (gtk_scale_set_draw_value) IA__gtk_scale_set_draw_value __attribute((visibility("hidden")));
#define gtk_scale_set_draw_value IA__gtk_scale_set_draw_value

extern __typeof (gtk_scale_set_value_pos) IA__gtk_scale_set_value_pos __attribute((visibility("hidden")));
#define gtk_scale_set_value_pos IA__gtk_scale_set_value_pos

extern __typeof (gtk_scale_add_mark) IA__gtk_scale_add_mark __attribute((visibility("hidden")));
#define gtk_scale_add_mark IA__gtk_scale_add_mark

extern __typeof (gtk_scale_clear_marks) IA__gtk_scale_clear_marks __attribute((visibility("hidden")));
#define gtk_scale_clear_marks IA__gtk_scale_clear_marks

#endif
#endif
#if IN_HEADER(__GTK_SCALE_BUTTON_H__)
#if IN_FILE(__GTK_SCALE_BUTTON_C__)
extern __typeof (gtk_scale_button_get_type) IA__gtk_scale_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_scale_button_get_type IA__gtk_scale_button_get_type

extern __typeof (gtk_scale_button_get_value) IA__gtk_scale_button_get_value __attribute((visibility("hidden")));
#define gtk_scale_button_get_value IA__gtk_scale_button_get_value

extern __typeof (gtk_scale_button_new) IA__gtk_scale_button_new __attribute((visibility("hidden")));
#define gtk_scale_button_new IA__gtk_scale_button_new

extern __typeof (gtk_scale_button_set_icons) IA__gtk_scale_button_set_icons __attribute((visibility("hidden")));
#define gtk_scale_button_set_icons IA__gtk_scale_button_set_icons

extern __typeof (gtk_scale_button_set_value) IA__gtk_scale_button_set_value __attribute((visibility("hidden")));
#define gtk_scale_button_set_value IA__gtk_scale_button_set_value

extern __typeof (gtk_scale_button_get_adjustment) IA__gtk_scale_button_get_adjustment __attribute((visibility("hidden")));
#define gtk_scale_button_get_adjustment IA__gtk_scale_button_get_adjustment

extern __typeof (gtk_scale_button_set_adjustment) IA__gtk_scale_button_set_adjustment __attribute((visibility("hidden")));
#define gtk_scale_button_set_adjustment IA__gtk_scale_button_set_adjustment

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_scale_button_get_orientation) IA__gtk_scale_button_get_orientation __attribute((visibility("hidden")));
#define gtk_scale_button_get_orientation IA__gtk_scale_button_get_orientation

extern __typeof (gtk_scale_button_set_orientation) IA__gtk_scale_button_set_orientation __attribute((visibility("hidden")));
#define gtk_scale_button_set_orientation IA__gtk_scale_button_set_orientation

#endif
extern __typeof (gtk_scale_button_get_plus_button) IA__gtk_scale_button_get_plus_button __attribute((visibility("hidden")));
#define gtk_scale_button_get_plus_button IA__gtk_scale_button_get_plus_button

extern __typeof (gtk_scale_button_get_minus_button) IA__gtk_scale_button_get_minus_button __attribute((visibility("hidden")));
#define gtk_scale_button_get_minus_button IA__gtk_scale_button_get_minus_button

extern __typeof (gtk_scale_button_get_popup) IA__gtk_scale_button_get_popup __attribute((visibility("hidden")));
#define gtk_scale_button_get_popup IA__gtk_scale_button_get_popup

#endif
#endif
#if IN_HEADER(__GTK_SCROLLBAR_H__)
#if IN_FILE(__GTK_SCROLLBAR_C__)
extern __typeof (gtk_scrollbar_get_type) IA__gtk_scrollbar_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_scrollbar_get_type IA__gtk_scrollbar_get_type

#endif
#endif
#if IN_HEADER(__GTK_SCROLLED_WINDOW_H__)
#if IN_FILE(__GTK_SCROLLED_WINDOW_C__)
extern __typeof (gtk_scrolled_window_add_with_viewport) IA__gtk_scrolled_window_add_with_viewport __attribute((visibility("hidden")));
#define gtk_scrolled_window_add_with_viewport IA__gtk_scrolled_window_add_with_viewport

extern __typeof (gtk_scrolled_window_get_hadjustment) IA__gtk_scrolled_window_get_hadjustment __attribute((visibility("hidden")));
#define gtk_scrolled_window_get_hadjustment IA__gtk_scrolled_window_get_hadjustment

extern __typeof (gtk_scrolled_window_get_hscrollbar) IA__gtk_scrolled_window_get_hscrollbar __attribute((visibility("hidden")));
#define gtk_scrolled_window_get_hscrollbar IA__gtk_scrolled_window_get_hscrollbar

extern __typeof (gtk_scrolled_window_get_placement) IA__gtk_scrolled_window_get_placement __attribute((visibility("hidden")));
#define gtk_scrolled_window_get_placement IA__gtk_scrolled_window_get_placement

extern __typeof (gtk_scrolled_window_get_policy) IA__gtk_scrolled_window_get_policy __attribute((visibility("hidden")));
#define gtk_scrolled_window_get_policy IA__gtk_scrolled_window_get_policy

extern __typeof (gtk_scrolled_window_get_shadow_type) IA__gtk_scrolled_window_get_shadow_type __attribute((visibility("hidden")));
#define gtk_scrolled_window_get_shadow_type IA__gtk_scrolled_window_get_shadow_type

extern __typeof (gtk_scrolled_window_get_type) IA__gtk_scrolled_window_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_scrolled_window_get_type IA__gtk_scrolled_window_get_type

extern __typeof (gtk_scrolled_window_get_vadjustment) IA__gtk_scrolled_window_get_vadjustment __attribute((visibility("hidden")));
#define gtk_scrolled_window_get_vadjustment IA__gtk_scrolled_window_get_vadjustment

extern __typeof (gtk_scrolled_window_get_vscrollbar) IA__gtk_scrolled_window_get_vscrollbar __attribute((visibility("hidden")));
#define gtk_scrolled_window_get_vscrollbar IA__gtk_scrolled_window_get_vscrollbar

extern __typeof (gtk_scrolled_window_new) IA__gtk_scrolled_window_new __attribute((visibility("hidden")));
#define gtk_scrolled_window_new IA__gtk_scrolled_window_new

extern __typeof (gtk_scrolled_window_set_hadjustment) IA__gtk_scrolled_window_set_hadjustment __attribute((visibility("hidden")));
#define gtk_scrolled_window_set_hadjustment IA__gtk_scrolled_window_set_hadjustment

extern __typeof (gtk_scrolled_window_set_placement) IA__gtk_scrolled_window_set_placement __attribute((visibility("hidden")));
#define gtk_scrolled_window_set_placement IA__gtk_scrolled_window_set_placement

extern __typeof (gtk_scrolled_window_set_policy) IA__gtk_scrolled_window_set_policy __attribute((visibility("hidden")));
#define gtk_scrolled_window_set_policy IA__gtk_scrolled_window_set_policy

extern __typeof (gtk_scrolled_window_set_shadow_type) IA__gtk_scrolled_window_set_shadow_type __attribute((visibility("hidden")));
#define gtk_scrolled_window_set_shadow_type IA__gtk_scrolled_window_set_shadow_type

extern __typeof (gtk_scrolled_window_set_vadjustment) IA__gtk_scrolled_window_set_vadjustment __attribute((visibility("hidden")));
#define gtk_scrolled_window_set_vadjustment IA__gtk_scrolled_window_set_vadjustment

extern __typeof (gtk_scrolled_window_unset_placement) IA__gtk_scrolled_window_unset_placement __attribute((visibility("hidden")));
#define gtk_scrolled_window_unset_placement IA__gtk_scrolled_window_unset_placement

#endif
#endif
#if IN_HEADER(__GTK_SELECTION_H__)
#if IN_FILE(__GTK_SELECTION_C__)
extern __typeof (gtk_selection_add_target) IA__gtk_selection_add_target __attribute((visibility("hidden")));
#define gtk_selection_add_target IA__gtk_selection_add_target

extern __typeof (gtk_selection_add_targets) IA__gtk_selection_add_targets __attribute((visibility("hidden")));
#define gtk_selection_add_targets IA__gtk_selection_add_targets

extern __typeof (gtk_selection_clear) IA__gtk_selection_clear __attribute((visibility("hidden")));
#define gtk_selection_clear IA__gtk_selection_clear

extern __typeof (gtk_selection_clear_targets) IA__gtk_selection_clear_targets __attribute((visibility("hidden")));
#define gtk_selection_clear_targets IA__gtk_selection_clear_targets

extern __typeof (gtk_selection_convert) IA__gtk_selection_convert __attribute((visibility("hidden")));
#define gtk_selection_convert IA__gtk_selection_convert

extern __typeof (gtk_selection_data_copy) IA__gtk_selection_data_copy __attribute((visibility("hidden")));
#define gtk_selection_data_copy IA__gtk_selection_data_copy

extern __typeof (gtk_selection_data_free) IA__gtk_selection_data_free __attribute((visibility("hidden")));
#define gtk_selection_data_free IA__gtk_selection_data_free

extern __typeof (gtk_selection_data_get_selection) IA__gtk_selection_data_get_selection __attribute((visibility("hidden")));
#define gtk_selection_data_get_selection IA__gtk_selection_data_get_selection

extern __typeof (gtk_selection_data_get_target) IA__gtk_selection_data_get_target __attribute((visibility("hidden")));
#define gtk_selection_data_get_target IA__gtk_selection_data_get_target

extern __typeof (gtk_selection_data_get_data_type) IA__gtk_selection_data_get_data_type __attribute((visibility("hidden")));
#define gtk_selection_data_get_data_type IA__gtk_selection_data_get_data_type

extern __typeof (gtk_selection_data_get_display) IA__gtk_selection_data_get_display __attribute((visibility("hidden")));
#define gtk_selection_data_get_display IA__gtk_selection_data_get_display

extern __typeof (gtk_selection_data_get_format) IA__gtk_selection_data_get_format __attribute((visibility("hidden")));
#define gtk_selection_data_get_format IA__gtk_selection_data_get_format

extern __typeof (gtk_selection_data_get_data) IA__gtk_selection_data_get_data __attribute((visibility("hidden")));
#define gtk_selection_data_get_data IA__gtk_selection_data_get_data

extern __typeof (gtk_selection_data_get_length) IA__gtk_selection_data_get_length __attribute((visibility("hidden")));
#define gtk_selection_data_get_length IA__gtk_selection_data_get_length

extern __typeof (gtk_selection_data_get_pixbuf) IA__gtk_selection_data_get_pixbuf __attribute((visibility("hidden")));
#define gtk_selection_data_get_pixbuf IA__gtk_selection_data_get_pixbuf

extern __typeof (gtk_selection_data_get_targets) IA__gtk_selection_data_get_targets __attribute((visibility("hidden")));
#define gtk_selection_data_get_targets IA__gtk_selection_data_get_targets

extern __typeof (gtk_selection_data_get_text) IA__gtk_selection_data_get_text __attribute((visibility("hidden")));
#define gtk_selection_data_get_text IA__gtk_selection_data_get_text

extern __typeof (gtk_selection_data_get_type) IA__gtk_selection_data_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_selection_data_get_type IA__gtk_selection_data_get_type

extern __typeof (gtk_selection_data_get_uris) IA__gtk_selection_data_get_uris __attribute((visibility("hidden")));
#define gtk_selection_data_get_uris IA__gtk_selection_data_get_uris

extern __typeof (gtk_selection_data_set) IA__gtk_selection_data_set __attribute((visibility("hidden")));
#define gtk_selection_data_set IA__gtk_selection_data_set

extern __typeof (gtk_selection_data_set_pixbuf) IA__gtk_selection_data_set_pixbuf __attribute((visibility("hidden")));
#define gtk_selection_data_set_pixbuf IA__gtk_selection_data_set_pixbuf

extern __typeof (gtk_selection_data_set_text) IA__gtk_selection_data_set_text __attribute((visibility("hidden")));
#define gtk_selection_data_set_text IA__gtk_selection_data_set_text

extern __typeof (gtk_selection_data_set_uris) IA__gtk_selection_data_set_uris __attribute((visibility("hidden")));
#define gtk_selection_data_set_uris IA__gtk_selection_data_set_uris

extern __typeof (gtk_selection_data_targets_include_image) IA__gtk_selection_data_targets_include_image __attribute((visibility("hidden")));
#define gtk_selection_data_targets_include_image IA__gtk_selection_data_targets_include_image

extern __typeof (gtk_selection_data_targets_include_rich_text) IA__gtk_selection_data_targets_include_rich_text __attribute((visibility("hidden")));
#define gtk_selection_data_targets_include_rich_text IA__gtk_selection_data_targets_include_rich_text

extern __typeof (gtk_selection_data_targets_include_text) IA__gtk_selection_data_targets_include_text __attribute((visibility("hidden")));
#define gtk_selection_data_targets_include_text IA__gtk_selection_data_targets_include_text

extern __typeof (gtk_selection_data_targets_include_uri) IA__gtk_selection_data_targets_include_uri __attribute((visibility("hidden")));
#define gtk_selection_data_targets_include_uri IA__gtk_selection_data_targets_include_uri

extern __typeof (gtk_targets_include_image) IA__gtk_targets_include_image __attribute((visibility("hidden")));
#define gtk_targets_include_image IA__gtk_targets_include_image

extern __typeof (gtk_targets_include_rich_text) IA__gtk_targets_include_rich_text __attribute((visibility("hidden")));
#define gtk_targets_include_rich_text IA__gtk_targets_include_rich_text

extern __typeof (gtk_targets_include_text) IA__gtk_targets_include_text __attribute((visibility("hidden")));
#define gtk_targets_include_text IA__gtk_targets_include_text

extern __typeof (gtk_targets_include_uri) IA__gtk_targets_include_uri __attribute((visibility("hidden")));
#define gtk_targets_include_uri IA__gtk_targets_include_uri

extern __typeof (gtk_selection_owner_set) IA__gtk_selection_owner_set __attribute((visibility("hidden")));
#define gtk_selection_owner_set IA__gtk_selection_owner_set

extern __typeof (gtk_selection_owner_set_for_display) IA__gtk_selection_owner_set_for_display __attribute((visibility("hidden")));
#define gtk_selection_owner_set_for_display IA__gtk_selection_owner_set_for_display

extern __typeof (gtk_selection_remove_all) IA__gtk_selection_remove_all __attribute((visibility("hidden")));
#define gtk_selection_remove_all IA__gtk_selection_remove_all

extern __typeof (gtk_target_list_add) IA__gtk_target_list_add __attribute((visibility("hidden")));
#define gtk_target_list_add IA__gtk_target_list_add

extern __typeof (gtk_target_list_add_image_targets) IA__gtk_target_list_add_image_targets __attribute((visibility("hidden")));
#define gtk_target_list_add_image_targets IA__gtk_target_list_add_image_targets

extern __typeof (gtk_target_list_add_rich_text_targets) IA__gtk_target_list_add_rich_text_targets __attribute((visibility("hidden")));
#define gtk_target_list_add_rich_text_targets IA__gtk_target_list_add_rich_text_targets

extern __typeof (gtk_target_list_add_table) IA__gtk_target_list_add_table __attribute((visibility("hidden")));
#define gtk_target_list_add_table IA__gtk_target_list_add_table

extern __typeof (gtk_target_list_add_text_targets) IA__gtk_target_list_add_text_targets __attribute((visibility("hidden")));
#define gtk_target_list_add_text_targets IA__gtk_target_list_add_text_targets

extern __typeof (gtk_target_list_add_uri_targets) IA__gtk_target_list_add_uri_targets __attribute((visibility("hidden")));
#define gtk_target_list_add_uri_targets IA__gtk_target_list_add_uri_targets

extern __typeof (gtk_target_list_find) IA__gtk_target_list_find __attribute((visibility("hidden")));
#define gtk_target_list_find IA__gtk_target_list_find

extern __typeof (gtk_target_list_get_type) IA__gtk_target_list_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_target_list_get_type IA__gtk_target_list_get_type

extern __typeof (gtk_target_list_new) IA__gtk_target_list_new __attribute((visibility("hidden")));
#define gtk_target_list_new IA__gtk_target_list_new

extern __typeof (gtk_target_list_ref) IA__gtk_target_list_ref __attribute((visibility("hidden")));
#define gtk_target_list_ref IA__gtk_target_list_ref

extern __typeof (gtk_target_list_remove) IA__gtk_target_list_remove __attribute((visibility("hidden")));
#define gtk_target_list_remove IA__gtk_target_list_remove

extern __typeof (gtk_target_list_unref) IA__gtk_target_list_unref __attribute((visibility("hidden")));
#define gtk_target_list_unref IA__gtk_target_list_unref

extern __typeof (gtk_target_table_new_from_list) IA__gtk_target_table_new_from_list __attribute((visibility("hidden")));
#define gtk_target_table_new_from_list IA__gtk_target_table_new_from_list

extern __typeof (gtk_target_table_free) IA__gtk_target_table_free __attribute((visibility("hidden")));
#define gtk_target_table_free IA__gtk_target_table_free

#endif
#endif
#if IN_HEADER(__GTK_SEPARATOR_H__)
#if IN_FILE(__GTK_SEPARATOR_C__)
extern __typeof (gtk_separator_get_type) IA__gtk_separator_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_separator_get_type IA__gtk_separator_get_type

#endif
#endif
#if IN_HEADER(__GTK_SEPARATOR_MENU_ITEM_H__)
#if IN_FILE(__GTK_SEPARATOR_MENU_ITEM_C__)
extern __typeof (gtk_separator_menu_item_get_type) IA__gtk_separator_menu_item_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_separator_menu_item_get_type IA__gtk_separator_menu_item_get_type

extern __typeof (gtk_separator_menu_item_new) IA__gtk_separator_menu_item_new __attribute((visibility("hidden")));
#define gtk_separator_menu_item_new IA__gtk_separator_menu_item_new

#endif
#endif
#if IN_HEADER(__GTK_SEPARATOR_TOOL_ITEM_H__)
#if IN_FILE(__GTK_SEPARATOR_TOOL_ITEM_C__)
extern __typeof (gtk_separator_tool_item_get_draw) IA__gtk_separator_tool_item_get_draw __attribute((visibility("hidden")));
#define gtk_separator_tool_item_get_draw IA__gtk_separator_tool_item_get_draw

extern __typeof (gtk_separator_tool_item_get_type) IA__gtk_separator_tool_item_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_separator_tool_item_get_type IA__gtk_separator_tool_item_get_type

extern __typeof (gtk_separator_tool_item_new) IA__gtk_separator_tool_item_new __attribute((visibility("hidden")));
#define gtk_separator_tool_item_new IA__gtk_separator_tool_item_new

extern __typeof (gtk_separator_tool_item_set_draw) IA__gtk_separator_tool_item_set_draw __attribute((visibility("hidden")));
#define gtk_separator_tool_item_set_draw IA__gtk_separator_tool_item_set_draw

#endif
#endif
#if IN_HEADER(__GTK_SETTINGS_H__)
#if IN_FILE(__GTK_SETTINGS_C__)
extern __typeof (gtk_settings_get_default) IA__gtk_settings_get_default __attribute((visibility("hidden")));
#define gtk_settings_get_default IA__gtk_settings_get_default

extern __typeof (gtk_settings_get_for_screen) IA__gtk_settings_get_for_screen __attribute((visibility("hidden")));
#define gtk_settings_get_for_screen IA__gtk_settings_get_for_screen

extern __typeof (gtk_settings_get_type) IA__gtk_settings_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_settings_get_type IA__gtk_settings_get_type

extern __typeof (gtk_settings_install_property) IA__gtk_settings_install_property __attribute((visibility("hidden")));
#define gtk_settings_install_property IA__gtk_settings_install_property

extern __typeof (gtk_settings_install_property_parser) IA__gtk_settings_install_property_parser __attribute((visibility("hidden")));
#define gtk_settings_install_property_parser IA__gtk_settings_install_property_parser

extern __typeof (gtk_settings_set_double_property) IA__gtk_settings_set_double_property __attribute((visibility("hidden")));
#define gtk_settings_set_double_property IA__gtk_settings_set_double_property

extern __typeof (gtk_settings_set_long_property) IA__gtk_settings_set_long_property __attribute((visibility("hidden")));
#define gtk_settings_set_long_property IA__gtk_settings_set_long_property

extern __typeof (gtk_settings_set_property_value) IA__gtk_settings_set_property_value __attribute((visibility("hidden")));
#define gtk_settings_set_property_value IA__gtk_settings_set_property_value

extern __typeof (gtk_settings_set_string_property) IA__gtk_settings_set_string_property __attribute((visibility("hidden")));
#define gtk_settings_set_string_property IA__gtk_settings_set_string_property

extern __typeof (gtk_rc_property_parse_border) IA__gtk_rc_property_parse_border __attribute((visibility("hidden")));
#define gtk_rc_property_parse_border IA__gtk_rc_property_parse_border

extern __typeof (gtk_rc_property_parse_color) IA__gtk_rc_property_parse_color __attribute((visibility("hidden")));
#define gtk_rc_property_parse_color IA__gtk_rc_property_parse_color

extern __typeof (gtk_rc_property_parse_enum) IA__gtk_rc_property_parse_enum __attribute((visibility("hidden")));
#define gtk_rc_property_parse_enum IA__gtk_rc_property_parse_enum

extern __typeof (gtk_rc_property_parse_flags) IA__gtk_rc_property_parse_flags __attribute((visibility("hidden")));
#define gtk_rc_property_parse_flags IA__gtk_rc_property_parse_flags

extern __typeof (gtk_rc_property_parse_requisition) IA__gtk_rc_property_parse_requisition __attribute((visibility("hidden")));
#define gtk_rc_property_parse_requisition IA__gtk_rc_property_parse_requisition

#endif
#endif
#if IN_HEADER(__GTK_SIGNAL_H__)
#if IN_FILE(__GTK_SIGNAL_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_signal_compat_matched) IA__gtk_signal_compat_matched __attribute((visibility("hidden")));
#define gtk_signal_compat_matched IA__gtk_signal_compat_matched

extern __typeof (gtk_signal_connect_full) IA__gtk_signal_connect_full __attribute((visibility("hidden")));
#define gtk_signal_connect_full IA__gtk_signal_connect_full

extern __typeof (gtk_signal_connect_object_while_alive) IA__gtk_signal_connect_object_while_alive __attribute((visibility("hidden")));
#define gtk_signal_connect_object_while_alive IA__gtk_signal_connect_object_while_alive

extern __typeof (gtk_signal_connect_while_alive) IA__gtk_signal_connect_while_alive __attribute((visibility("hidden")));
#define gtk_signal_connect_while_alive IA__gtk_signal_connect_while_alive

extern __typeof (gtk_signal_emit) IA__gtk_signal_emit __attribute((visibility("hidden")));
#define gtk_signal_emit IA__gtk_signal_emit

extern __typeof (gtk_signal_emit_by_name) IA__gtk_signal_emit_by_name __attribute((visibility("hidden")));
#define gtk_signal_emit_by_name IA__gtk_signal_emit_by_name

extern __typeof (gtk_signal_emit_stop_by_name) IA__gtk_signal_emit_stop_by_name __attribute((visibility("hidden")));
#define gtk_signal_emit_stop_by_name IA__gtk_signal_emit_stop_by_name

extern __typeof (gtk_signal_emitv) IA__gtk_signal_emitv __attribute((visibility("hidden")));
#define gtk_signal_emitv IA__gtk_signal_emitv

extern __typeof (gtk_signal_emitv_by_name) IA__gtk_signal_emitv_by_name __attribute((visibility("hidden")));
#define gtk_signal_emitv_by_name IA__gtk_signal_emitv_by_name

extern __typeof (gtk_signal_new) IA__gtk_signal_new __attribute((visibility("hidden")));
#define gtk_signal_new IA__gtk_signal_new

extern __typeof (gtk_signal_newv) IA__gtk_signal_newv __attribute((visibility("hidden")));
#define gtk_signal_newv IA__gtk_signal_newv

#endif
#endif
#endif
#if IN_HEADER(__GTK_SIZE_GROUP_H__)
#if IN_FILE(__GTK_SIZE_GROUP_C__)
extern __typeof (gtk_size_group_add_widget) IA__gtk_size_group_add_widget __attribute((visibility("hidden")));
#define gtk_size_group_add_widget IA__gtk_size_group_add_widget

extern __typeof (gtk_size_group_get_ignore_hidden) IA__gtk_size_group_get_ignore_hidden __attribute((visibility("hidden")));
#define gtk_size_group_get_ignore_hidden IA__gtk_size_group_get_ignore_hidden

extern __typeof (gtk_size_group_get_mode) IA__gtk_size_group_get_mode __attribute((visibility("hidden")));
#define gtk_size_group_get_mode IA__gtk_size_group_get_mode

extern __typeof (gtk_size_group_get_type) IA__gtk_size_group_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_size_group_get_type IA__gtk_size_group_get_type

extern __typeof (gtk_size_group_get_widgets) IA__gtk_size_group_get_widgets __attribute((visibility("hidden")));
#define gtk_size_group_get_widgets IA__gtk_size_group_get_widgets

extern __typeof (gtk_size_group_new) IA__gtk_size_group_new __attribute((visibility("hidden")));
#define gtk_size_group_new IA__gtk_size_group_new

extern __typeof (gtk_size_group_remove_widget) IA__gtk_size_group_remove_widget __attribute((visibility("hidden")));
#define gtk_size_group_remove_widget IA__gtk_size_group_remove_widget

extern __typeof (gtk_size_group_set_ignore_hidden) IA__gtk_size_group_set_ignore_hidden __attribute((visibility("hidden")));
#define gtk_size_group_set_ignore_hidden IA__gtk_size_group_set_ignore_hidden

extern __typeof (gtk_size_group_set_mode) IA__gtk_size_group_set_mode __attribute((visibility("hidden")));
#define gtk_size_group_set_mode IA__gtk_size_group_set_mode

#endif
#endif
#if IN_HEADER(__GTK_SHOW_H__)
#if IN_FILE(__GTK_SHOW_C__)
extern __typeof (gtk_show_uri) IA__gtk_show_uri __attribute((visibility("hidden")));
#define gtk_show_uri IA__gtk_show_uri

#endif
#endif
#if IN_HEADER(__GTK_SOCKET_H__)
#if IN_FILE(__GTK_SOCKET_C__)
extern __typeof (gtk_socket_add_id) IA__gtk_socket_add_id __attribute((visibility("hidden")));
#define gtk_socket_add_id IA__gtk_socket_add_id

extern __typeof (gtk_socket_get_id) IA__gtk_socket_get_id __attribute((visibility("hidden")));
#define gtk_socket_get_id IA__gtk_socket_get_id

extern __typeof (gtk_socket_get_plug_window) IA__gtk_socket_get_plug_window __attribute((visibility("hidden")));
#define gtk_socket_get_plug_window IA__gtk_socket_get_plug_window

extern __typeof (gtk_socket_get_type) IA__gtk_socket_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_socket_get_type IA__gtk_socket_get_type

extern __typeof (gtk_socket_new) IA__gtk_socket_new __attribute((visibility("hidden")));
#define gtk_socket_new IA__gtk_socket_new

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_socket_steal) IA__gtk_socket_steal __attribute((visibility("hidden")));
#define gtk_socket_steal IA__gtk_socket_steal

#endif
#endif
#endif
#if IN_HEADER(__GTK_SPIN_BUTTON_H__)
#if IN_FILE(__GTK_SPIN_BUTTON_C__)
extern __typeof (gtk_spin_button_configure) IA__gtk_spin_button_configure __attribute((visibility("hidden")));
#define gtk_spin_button_configure IA__gtk_spin_button_configure

extern __typeof (gtk_spin_button_get_adjustment) IA__gtk_spin_button_get_adjustment __attribute((visibility("hidden")));
#define gtk_spin_button_get_adjustment IA__gtk_spin_button_get_adjustment

extern __typeof (gtk_spin_button_get_digits) IA__gtk_spin_button_get_digits __attribute((visibility("hidden")));
#define gtk_spin_button_get_digits IA__gtk_spin_button_get_digits

extern __typeof (gtk_spin_button_get_increments) IA__gtk_spin_button_get_increments __attribute((visibility("hidden")));
#define gtk_spin_button_get_increments IA__gtk_spin_button_get_increments

extern __typeof (gtk_spin_button_get_numeric) IA__gtk_spin_button_get_numeric __attribute((visibility("hidden")));
#define gtk_spin_button_get_numeric IA__gtk_spin_button_get_numeric

extern __typeof (gtk_spin_button_get_range) IA__gtk_spin_button_get_range __attribute((visibility("hidden")));
#define gtk_spin_button_get_range IA__gtk_spin_button_get_range

extern __typeof (gtk_spin_button_get_snap_to_ticks) IA__gtk_spin_button_get_snap_to_ticks __attribute((visibility("hidden")));
#define gtk_spin_button_get_snap_to_ticks IA__gtk_spin_button_get_snap_to_ticks

extern __typeof (gtk_spin_button_get_type) IA__gtk_spin_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_spin_button_get_type IA__gtk_spin_button_get_type

extern __typeof (gtk_spin_button_get_update_policy) IA__gtk_spin_button_get_update_policy __attribute((visibility("hidden")));
#define gtk_spin_button_get_update_policy IA__gtk_spin_button_get_update_policy

extern __typeof (gtk_spin_button_get_value) IA__gtk_spin_button_get_value __attribute((visibility("hidden")));
#define gtk_spin_button_get_value IA__gtk_spin_button_get_value

extern __typeof (gtk_spin_button_get_value_as_int) IA__gtk_spin_button_get_value_as_int __attribute((visibility("hidden")));
#define gtk_spin_button_get_value_as_int IA__gtk_spin_button_get_value_as_int

extern __typeof (gtk_spin_button_get_wrap) IA__gtk_spin_button_get_wrap __attribute((visibility("hidden")));
#define gtk_spin_button_get_wrap IA__gtk_spin_button_get_wrap

extern __typeof (gtk_spin_button_new) IA__gtk_spin_button_new __attribute((visibility("hidden")));
#define gtk_spin_button_new IA__gtk_spin_button_new

extern __typeof (gtk_spin_button_new_with_range) IA__gtk_spin_button_new_with_range __attribute((visibility("hidden")));
#define gtk_spin_button_new_with_range IA__gtk_spin_button_new_with_range

extern __typeof (gtk_spin_button_set_adjustment) IA__gtk_spin_button_set_adjustment __attribute((visibility("hidden")));
#define gtk_spin_button_set_adjustment IA__gtk_spin_button_set_adjustment

extern __typeof (gtk_spin_button_set_digits) IA__gtk_spin_button_set_digits __attribute((visibility("hidden")));
#define gtk_spin_button_set_digits IA__gtk_spin_button_set_digits

extern __typeof (gtk_spin_button_set_increments) IA__gtk_spin_button_set_increments __attribute((visibility("hidden")));
#define gtk_spin_button_set_increments IA__gtk_spin_button_set_increments

extern __typeof (gtk_spin_button_set_numeric) IA__gtk_spin_button_set_numeric __attribute((visibility("hidden")));
#define gtk_spin_button_set_numeric IA__gtk_spin_button_set_numeric

extern __typeof (gtk_spin_button_set_range) IA__gtk_spin_button_set_range __attribute((visibility("hidden")));
#define gtk_spin_button_set_range IA__gtk_spin_button_set_range

extern __typeof (gtk_spin_button_set_snap_to_ticks) IA__gtk_spin_button_set_snap_to_ticks __attribute((visibility("hidden")));
#define gtk_spin_button_set_snap_to_ticks IA__gtk_spin_button_set_snap_to_ticks

extern __typeof (gtk_spin_button_set_update_policy) IA__gtk_spin_button_set_update_policy __attribute((visibility("hidden")));
#define gtk_spin_button_set_update_policy IA__gtk_spin_button_set_update_policy

extern __typeof (gtk_spin_button_set_value) IA__gtk_spin_button_set_value __attribute((visibility("hidden")));
#define gtk_spin_button_set_value IA__gtk_spin_button_set_value

extern __typeof (gtk_spin_button_set_wrap) IA__gtk_spin_button_set_wrap __attribute((visibility("hidden")));
#define gtk_spin_button_set_wrap IA__gtk_spin_button_set_wrap

extern __typeof (gtk_spin_button_spin) IA__gtk_spin_button_spin __attribute((visibility("hidden")));
#define gtk_spin_button_spin IA__gtk_spin_button_spin

extern __typeof (gtk_spin_button_update) IA__gtk_spin_button_update __attribute((visibility("hidden")));
#define gtk_spin_button_update IA__gtk_spin_button_update

#endif
#endif
#if IN_HEADER(__GTK_SPINNER_H__)
#if IN_FILE(__GTK_SPINNER_C__)
extern __typeof (gtk_spinner_get_type) IA__gtk_spinner_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_spinner_get_type IA__gtk_spinner_get_type

extern __typeof (gtk_spinner_new) IA__gtk_spinner_new __attribute((visibility("hidden")));
#define gtk_spinner_new IA__gtk_spinner_new

extern __typeof (gtk_spinner_start) IA__gtk_spinner_start __attribute((visibility("hidden")));
#define gtk_spinner_start IA__gtk_spinner_start

extern __typeof (gtk_spinner_stop) IA__gtk_spinner_stop __attribute((visibility("hidden")));
#define gtk_spinner_stop IA__gtk_spinner_stop

#endif
#endif
#if IN_HEADER(__GTK_STATUSBAR_H__)
#if IN_FILE(__GTK_STATUSBAR_C__)
extern __typeof (gtk_statusbar_get_context_id) IA__gtk_statusbar_get_context_id __attribute((visibility("hidden")));
#define gtk_statusbar_get_context_id IA__gtk_statusbar_get_context_id

extern __typeof (gtk_statusbar_get_has_resize_grip) IA__gtk_statusbar_get_has_resize_grip __attribute((visibility("hidden")));
#define gtk_statusbar_get_has_resize_grip IA__gtk_statusbar_get_has_resize_grip

extern __typeof (gtk_statusbar_get_message_area) IA__gtk_statusbar_get_message_area __attribute((visibility("hidden")));
#define gtk_statusbar_get_message_area IA__gtk_statusbar_get_message_area

extern __typeof (gtk_statusbar_get_type) IA__gtk_statusbar_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_statusbar_get_type IA__gtk_statusbar_get_type

extern __typeof (gtk_statusbar_new) IA__gtk_statusbar_new __attribute((visibility("hidden")));
#define gtk_statusbar_new IA__gtk_statusbar_new

extern __typeof (gtk_statusbar_pop) IA__gtk_statusbar_pop __attribute((visibility("hidden")));
#define gtk_statusbar_pop IA__gtk_statusbar_pop

extern __typeof (gtk_statusbar_push) IA__gtk_statusbar_push __attribute((visibility("hidden")));
#define gtk_statusbar_push IA__gtk_statusbar_push

extern __typeof (gtk_statusbar_remove) IA__gtk_statusbar_remove __attribute((visibility("hidden")));
#define gtk_statusbar_remove IA__gtk_statusbar_remove

extern __typeof (gtk_statusbar_remove_all) IA__gtk_statusbar_remove_all __attribute((visibility("hidden")));
#define gtk_statusbar_remove_all IA__gtk_statusbar_remove_all

extern __typeof (gtk_statusbar_set_has_resize_grip) IA__gtk_statusbar_set_has_resize_grip __attribute((visibility("hidden")));
#define gtk_statusbar_set_has_resize_grip IA__gtk_statusbar_set_has_resize_grip

#endif
#endif
#if IN_HEADER(__GTK_STOCK_H__)
#if IN_FILE(__GTK_STOCK_C__)
extern __typeof (gtk_stock_add) IA__gtk_stock_add __attribute((visibility("hidden")));
#define gtk_stock_add IA__gtk_stock_add

extern __typeof (gtk_stock_add_static) IA__gtk_stock_add_static __attribute((visibility("hidden")));
#define gtk_stock_add_static IA__gtk_stock_add_static

extern __typeof (gtk_stock_item_copy) IA__gtk_stock_item_copy __attribute((visibility("hidden")));
#define gtk_stock_item_copy IA__gtk_stock_item_copy

extern __typeof (gtk_stock_item_free) IA__gtk_stock_item_free __attribute((visibility("hidden")));
#define gtk_stock_item_free IA__gtk_stock_item_free

extern __typeof (gtk_stock_list_ids) IA__gtk_stock_list_ids __attribute((visibility("hidden")));
#define gtk_stock_list_ids IA__gtk_stock_list_ids

extern __typeof (gtk_stock_lookup) IA__gtk_stock_lookup __attribute((visibility("hidden")));
#define gtk_stock_lookup IA__gtk_stock_lookup

extern __typeof (gtk_stock_set_translate_func) IA__gtk_stock_set_translate_func __attribute((visibility("hidden")));
#define gtk_stock_set_translate_func IA__gtk_stock_set_translate_func

#endif
#endif
#if IN_HEADER(__GTK_TABLE_H__)
#if IN_FILE(__GTK_TABLE_C__)
extern __typeof (gtk_table_attach) IA__gtk_table_attach __attribute((visibility("hidden")));
#define gtk_table_attach IA__gtk_table_attach

extern __typeof (gtk_table_attach_defaults) IA__gtk_table_attach_defaults __attribute((visibility("hidden")));
#define gtk_table_attach_defaults IA__gtk_table_attach_defaults

extern __typeof (gtk_table_get_col_spacing) IA__gtk_table_get_col_spacing __attribute((visibility("hidden")));
#define gtk_table_get_col_spacing IA__gtk_table_get_col_spacing

extern __typeof (gtk_table_get_default_col_spacing) IA__gtk_table_get_default_col_spacing __attribute((visibility("hidden")));
#define gtk_table_get_default_col_spacing IA__gtk_table_get_default_col_spacing

extern __typeof (gtk_table_get_default_row_spacing) IA__gtk_table_get_default_row_spacing __attribute((visibility("hidden")));
#define gtk_table_get_default_row_spacing IA__gtk_table_get_default_row_spacing

extern __typeof (gtk_table_get_homogeneous) IA__gtk_table_get_homogeneous __attribute((visibility("hidden")));
#define gtk_table_get_homogeneous IA__gtk_table_get_homogeneous

extern __typeof (gtk_table_get_row_spacing) IA__gtk_table_get_row_spacing __attribute((visibility("hidden")));
#define gtk_table_get_row_spacing IA__gtk_table_get_row_spacing

extern __typeof (gtk_table_get_size) IA__gtk_table_get_size __attribute((visibility("hidden")));
#define gtk_table_get_size IA__gtk_table_get_size

extern __typeof (gtk_table_get_type) IA__gtk_table_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_table_get_type IA__gtk_table_get_type

extern __typeof (gtk_table_new) IA__gtk_table_new __attribute((visibility("hidden")));
#define gtk_table_new IA__gtk_table_new

extern __typeof (gtk_table_resize) IA__gtk_table_resize __attribute((visibility("hidden")));
#define gtk_table_resize IA__gtk_table_resize

extern __typeof (gtk_table_set_col_spacing) IA__gtk_table_set_col_spacing __attribute((visibility("hidden")));
#define gtk_table_set_col_spacing IA__gtk_table_set_col_spacing

extern __typeof (gtk_table_set_col_spacings) IA__gtk_table_set_col_spacings __attribute((visibility("hidden")));
#define gtk_table_set_col_spacings IA__gtk_table_set_col_spacings

extern __typeof (gtk_table_set_homogeneous) IA__gtk_table_set_homogeneous __attribute((visibility("hidden")));
#define gtk_table_set_homogeneous IA__gtk_table_set_homogeneous

extern __typeof (gtk_table_set_row_spacing) IA__gtk_table_set_row_spacing __attribute((visibility("hidden")));
#define gtk_table_set_row_spacing IA__gtk_table_set_row_spacing

extern __typeof (gtk_table_set_row_spacings) IA__gtk_table_set_row_spacings __attribute((visibility("hidden")));
#define gtk_table_set_row_spacings IA__gtk_table_set_row_spacings

#endif
#endif
#if IN_HEADER(__GTK_TEAROFF_MENU_ITEM_H__)
#if IN_FILE(__GTK_TEAROFF_MENU_ITEM_C__)
extern __typeof (gtk_tearoff_menu_item_get_type) IA__gtk_tearoff_menu_item_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tearoff_menu_item_get_type IA__gtk_tearoff_menu_item_get_type

extern __typeof (gtk_tearoff_menu_item_new) IA__gtk_tearoff_menu_item_new __attribute((visibility("hidden")));
#define gtk_tearoff_menu_item_new IA__gtk_tearoff_menu_item_new

#endif
#endif
#if IN_HEADER(__GTK_TEXT_TAG_H__)
#if IN_FILE(__GTK_TEXT_TAG_C__)
extern __typeof (gtk_text_attributes_copy) IA__gtk_text_attributes_copy __attribute((visibility("hidden")));
#define gtk_text_attributes_copy IA__gtk_text_attributes_copy

extern __typeof (gtk_text_attributes_copy_values) IA__gtk_text_attributes_copy_values __attribute((visibility("hidden")));
#define gtk_text_attributes_copy_values IA__gtk_text_attributes_copy_values

extern __typeof (gtk_text_attributes_get_type) IA__gtk_text_attributes_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_attributes_get_type IA__gtk_text_attributes_get_type

extern __typeof (gtk_text_attributes_new) IA__gtk_text_attributes_new __attribute((visibility("hidden")));
#define gtk_text_attributes_new IA__gtk_text_attributes_new

extern __typeof (gtk_text_attributes_ref) IA__gtk_text_attributes_ref __attribute((visibility("hidden")));
#define gtk_text_attributes_ref IA__gtk_text_attributes_ref

extern __typeof (gtk_text_attributes_unref) IA__gtk_text_attributes_unref __attribute((visibility("hidden")));
#define gtk_text_attributes_unref IA__gtk_text_attributes_unref

extern __typeof (gtk_text_tag_event) IA__gtk_text_tag_event __attribute((visibility("hidden")));
#define gtk_text_tag_event IA__gtk_text_tag_event

extern __typeof (gtk_text_tag_get_priority) IA__gtk_text_tag_get_priority __attribute((visibility("hidden")));
#define gtk_text_tag_get_priority IA__gtk_text_tag_get_priority

extern __typeof (gtk_text_tag_get_type) IA__gtk_text_tag_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_tag_get_type IA__gtk_text_tag_get_type

extern __typeof (gtk_text_tag_new) IA__gtk_text_tag_new __attribute((visibility("hidden")));
#define gtk_text_tag_new IA__gtk_text_tag_new

extern __typeof (gtk_text_tag_set_priority) IA__gtk_text_tag_set_priority __attribute((visibility("hidden")));
#define gtk_text_tag_set_priority IA__gtk_text_tag_set_priority

#endif
#endif
#if IN_HEADER(__GTK_TEXT_TAG_TABLE_H__)
#if IN_FILE(__GTK_TEXT_TAG_TABLE_C__)
extern __typeof (gtk_text_tag_table_add) IA__gtk_text_tag_table_add __attribute((visibility("hidden")));
#define gtk_text_tag_table_add IA__gtk_text_tag_table_add

extern __typeof (gtk_text_tag_table_foreach) IA__gtk_text_tag_table_foreach __attribute((visibility("hidden")));
#define gtk_text_tag_table_foreach IA__gtk_text_tag_table_foreach

extern __typeof (gtk_text_tag_table_get_size) IA__gtk_text_tag_table_get_size __attribute((visibility("hidden")));
#define gtk_text_tag_table_get_size IA__gtk_text_tag_table_get_size

extern __typeof (gtk_text_tag_table_get_type) IA__gtk_text_tag_table_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_tag_table_get_type IA__gtk_text_tag_table_get_type

extern __typeof (gtk_text_tag_table_lookup) IA__gtk_text_tag_table_lookup __attribute((visibility("hidden")));
#define gtk_text_tag_table_lookup IA__gtk_text_tag_table_lookup

extern __typeof (gtk_text_tag_table_new) IA__gtk_text_tag_table_new __attribute((visibility("hidden")));
#define gtk_text_tag_table_new IA__gtk_text_tag_table_new

extern __typeof (gtk_text_tag_table_remove) IA__gtk_text_tag_table_remove __attribute((visibility("hidden")));
#define gtk_text_tag_table_remove IA__gtk_text_tag_table_remove

#endif
#endif
#if IN_HEADER(__GTK_TEXT_H__)
#if IN_FILE(__GTK_TEXT_C__)
extern __typeof (gtk_text_backward_delete) IA__gtk_text_backward_delete __attribute((visibility("hidden")));
#define gtk_text_backward_delete IA__gtk_text_backward_delete

extern __typeof (gtk_text_forward_delete) IA__gtk_text_forward_delete __attribute((visibility("hidden")));
#define gtk_text_forward_delete IA__gtk_text_forward_delete

extern __typeof (gtk_text_freeze) IA__gtk_text_freeze __attribute((visibility("hidden")));
#define gtk_text_freeze IA__gtk_text_freeze

extern __typeof (gtk_text_get_length) IA__gtk_text_get_length __attribute((visibility("hidden")));
#define gtk_text_get_length IA__gtk_text_get_length

extern __typeof (gtk_text_get_point) IA__gtk_text_get_point __attribute((visibility("hidden")));
#define gtk_text_get_point IA__gtk_text_get_point

extern __typeof (gtk_text_get_type) IA__gtk_text_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_get_type IA__gtk_text_get_type

extern __typeof (gtk_text_insert) IA__gtk_text_insert __attribute((visibility("hidden")));
#define gtk_text_insert IA__gtk_text_insert

extern __typeof (gtk_text_new) IA__gtk_text_new __attribute((visibility("hidden")));
#define gtk_text_new IA__gtk_text_new

extern __typeof (gtk_text_set_adjustments) IA__gtk_text_set_adjustments __attribute((visibility("hidden")));
#define gtk_text_set_adjustments IA__gtk_text_set_adjustments

extern __typeof (gtk_text_set_editable) IA__gtk_text_set_editable __attribute((visibility("hidden")));
#define gtk_text_set_editable IA__gtk_text_set_editable

extern __typeof (gtk_text_set_line_wrap) IA__gtk_text_set_line_wrap __attribute((visibility("hidden")));
#define gtk_text_set_line_wrap IA__gtk_text_set_line_wrap

extern __typeof (gtk_text_set_point) IA__gtk_text_set_point __attribute((visibility("hidden")));
#define gtk_text_set_point IA__gtk_text_set_point

extern __typeof (gtk_text_set_word_wrap) IA__gtk_text_set_word_wrap __attribute((visibility("hidden")));
#define gtk_text_set_word_wrap IA__gtk_text_set_word_wrap

extern __typeof (gtk_text_thaw) IA__gtk_text_thaw __attribute((visibility("hidden")));
#define gtk_text_thaw IA__gtk_text_thaw

#endif
#endif
#if IN_HEADER(__GTK_TEXT_BUFFER_H__)
#if IN_FILE(__GTK_TEXT_BUFFER_C__)
extern __typeof (gtk_text_buffer_add_selection_clipboard) IA__gtk_text_buffer_add_selection_clipboard __attribute((visibility("hidden")));
#define gtk_text_buffer_add_selection_clipboard IA__gtk_text_buffer_add_selection_clipboard

extern __typeof (gtk_text_buffer_apply_tag) IA__gtk_text_buffer_apply_tag __attribute((visibility("hidden")));
#define gtk_text_buffer_apply_tag IA__gtk_text_buffer_apply_tag

extern __typeof (gtk_text_buffer_apply_tag_by_name) IA__gtk_text_buffer_apply_tag_by_name __attribute((visibility("hidden")));
#define gtk_text_buffer_apply_tag_by_name IA__gtk_text_buffer_apply_tag_by_name

extern __typeof (gtk_text_buffer_backspace) IA__gtk_text_buffer_backspace __attribute((visibility("hidden")));
#define gtk_text_buffer_backspace IA__gtk_text_buffer_backspace

extern __typeof (gtk_text_buffer_begin_user_action) IA__gtk_text_buffer_begin_user_action __attribute((visibility("hidden")));
#define gtk_text_buffer_begin_user_action IA__gtk_text_buffer_begin_user_action

extern __typeof (gtk_text_buffer_copy_clipboard) IA__gtk_text_buffer_copy_clipboard __attribute((visibility("hidden")));
#define gtk_text_buffer_copy_clipboard IA__gtk_text_buffer_copy_clipboard

extern __typeof (gtk_text_buffer_create_child_anchor) IA__gtk_text_buffer_create_child_anchor __attribute((visibility("hidden")));
#define gtk_text_buffer_create_child_anchor IA__gtk_text_buffer_create_child_anchor

extern __typeof (gtk_text_buffer_add_mark) IA__gtk_text_buffer_add_mark __attribute((visibility("hidden")));
#define gtk_text_buffer_add_mark IA__gtk_text_buffer_add_mark

extern __typeof (gtk_text_buffer_create_mark) IA__gtk_text_buffer_create_mark __attribute((visibility("hidden")));
#define gtk_text_buffer_create_mark IA__gtk_text_buffer_create_mark

extern __typeof (gtk_text_buffer_create_tag) IA__gtk_text_buffer_create_tag __attribute((visibility("hidden")));
#define gtk_text_buffer_create_tag IA__gtk_text_buffer_create_tag

extern __typeof (gtk_text_buffer_cut_clipboard) IA__gtk_text_buffer_cut_clipboard __attribute((visibility("hidden")));
#define gtk_text_buffer_cut_clipboard IA__gtk_text_buffer_cut_clipboard

extern __typeof (gtk_text_buffer_delete) IA__gtk_text_buffer_delete __attribute((visibility("hidden")));
#define gtk_text_buffer_delete IA__gtk_text_buffer_delete

extern __typeof (gtk_text_buffer_delete_interactive) IA__gtk_text_buffer_delete_interactive __attribute((visibility("hidden")));
#define gtk_text_buffer_delete_interactive IA__gtk_text_buffer_delete_interactive

extern __typeof (gtk_text_buffer_delete_mark) IA__gtk_text_buffer_delete_mark __attribute((visibility("hidden")));
#define gtk_text_buffer_delete_mark IA__gtk_text_buffer_delete_mark

extern __typeof (gtk_text_buffer_delete_mark_by_name) IA__gtk_text_buffer_delete_mark_by_name __attribute((visibility("hidden")));
#define gtk_text_buffer_delete_mark_by_name IA__gtk_text_buffer_delete_mark_by_name

extern __typeof (gtk_text_buffer_delete_selection) IA__gtk_text_buffer_delete_selection __attribute((visibility("hidden")));
#define gtk_text_buffer_delete_selection IA__gtk_text_buffer_delete_selection

extern __typeof (gtk_text_buffer_end_user_action) IA__gtk_text_buffer_end_user_action __attribute((visibility("hidden")));
#define gtk_text_buffer_end_user_action IA__gtk_text_buffer_end_user_action

extern __typeof (gtk_text_buffer_get_bounds) IA__gtk_text_buffer_get_bounds __attribute((visibility("hidden")));
#define gtk_text_buffer_get_bounds IA__gtk_text_buffer_get_bounds

extern __typeof (gtk_text_buffer_get_char_count) IA__gtk_text_buffer_get_char_count __attribute((visibility("hidden")));
#define gtk_text_buffer_get_char_count IA__gtk_text_buffer_get_char_count

extern __typeof (gtk_text_buffer_get_copy_target_list) IA__gtk_text_buffer_get_copy_target_list __attribute((visibility("hidden")));
#define gtk_text_buffer_get_copy_target_list IA__gtk_text_buffer_get_copy_target_list

extern __typeof (gtk_text_buffer_get_end_iter) IA__gtk_text_buffer_get_end_iter __attribute((visibility("hidden")));
#define gtk_text_buffer_get_end_iter IA__gtk_text_buffer_get_end_iter

extern __typeof (gtk_text_buffer_get_has_selection) IA__gtk_text_buffer_get_has_selection __attribute((visibility("hidden")));
#define gtk_text_buffer_get_has_selection IA__gtk_text_buffer_get_has_selection

extern __typeof (gtk_text_buffer_get_insert) IA__gtk_text_buffer_get_insert __attribute((visibility("hidden")));
#define gtk_text_buffer_get_insert IA__gtk_text_buffer_get_insert

extern __typeof (gtk_text_buffer_get_iter_at_child_anchor) IA__gtk_text_buffer_get_iter_at_child_anchor __attribute((visibility("hidden")));
#define gtk_text_buffer_get_iter_at_child_anchor IA__gtk_text_buffer_get_iter_at_child_anchor

extern __typeof (gtk_text_buffer_get_iter_at_line) IA__gtk_text_buffer_get_iter_at_line __attribute((visibility("hidden")));
#define gtk_text_buffer_get_iter_at_line IA__gtk_text_buffer_get_iter_at_line

extern __typeof (gtk_text_buffer_get_iter_at_line_index) IA__gtk_text_buffer_get_iter_at_line_index __attribute((visibility("hidden")));
#define gtk_text_buffer_get_iter_at_line_index IA__gtk_text_buffer_get_iter_at_line_index

extern __typeof (gtk_text_buffer_get_iter_at_line_offset) IA__gtk_text_buffer_get_iter_at_line_offset __attribute((visibility("hidden")));
#define gtk_text_buffer_get_iter_at_line_offset IA__gtk_text_buffer_get_iter_at_line_offset

extern __typeof (gtk_text_buffer_get_iter_at_mark) IA__gtk_text_buffer_get_iter_at_mark __attribute((visibility("hidden")));
#define gtk_text_buffer_get_iter_at_mark IA__gtk_text_buffer_get_iter_at_mark

extern __typeof (gtk_text_buffer_get_iter_at_offset) IA__gtk_text_buffer_get_iter_at_offset __attribute((visibility("hidden")));
#define gtk_text_buffer_get_iter_at_offset IA__gtk_text_buffer_get_iter_at_offset

extern __typeof (gtk_text_buffer_get_line_count) IA__gtk_text_buffer_get_line_count __attribute((visibility("hidden")));
#define gtk_text_buffer_get_line_count IA__gtk_text_buffer_get_line_count

extern __typeof (gtk_text_buffer_get_mark) IA__gtk_text_buffer_get_mark __attribute((visibility("hidden")));
#define gtk_text_buffer_get_mark IA__gtk_text_buffer_get_mark

extern __typeof (gtk_text_buffer_get_modified) IA__gtk_text_buffer_get_modified __attribute((visibility("hidden")));
#define gtk_text_buffer_get_modified IA__gtk_text_buffer_get_modified

extern __typeof (gtk_text_buffer_get_paste_target_list) IA__gtk_text_buffer_get_paste_target_list __attribute((visibility("hidden")));
#define gtk_text_buffer_get_paste_target_list IA__gtk_text_buffer_get_paste_target_list

extern __typeof (gtk_text_buffer_get_selection_bound) IA__gtk_text_buffer_get_selection_bound __attribute((visibility("hidden")));
#define gtk_text_buffer_get_selection_bound IA__gtk_text_buffer_get_selection_bound

extern __typeof (gtk_text_buffer_get_selection_bounds) IA__gtk_text_buffer_get_selection_bounds __attribute((visibility("hidden")));
#define gtk_text_buffer_get_selection_bounds IA__gtk_text_buffer_get_selection_bounds

extern __typeof (gtk_text_buffer_get_slice) IA__gtk_text_buffer_get_slice __attribute((visibility("hidden")));
#define gtk_text_buffer_get_slice IA__gtk_text_buffer_get_slice

extern __typeof (gtk_text_buffer_get_start_iter) IA__gtk_text_buffer_get_start_iter __attribute((visibility("hidden")));
#define gtk_text_buffer_get_start_iter IA__gtk_text_buffer_get_start_iter

extern __typeof (gtk_text_buffer_get_tag_table) IA__gtk_text_buffer_get_tag_table __attribute((visibility("hidden")));
#define gtk_text_buffer_get_tag_table IA__gtk_text_buffer_get_tag_table

extern __typeof (gtk_text_buffer_get_text) IA__gtk_text_buffer_get_text __attribute((visibility("hidden")));
#define gtk_text_buffer_get_text IA__gtk_text_buffer_get_text

extern __typeof (gtk_text_buffer_get_type) IA__gtk_text_buffer_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_buffer_get_type IA__gtk_text_buffer_get_type

extern __typeof (gtk_text_buffer_insert) IA__gtk_text_buffer_insert __attribute((visibility("hidden")));
#define gtk_text_buffer_insert IA__gtk_text_buffer_insert

extern __typeof (gtk_text_buffer_insert_at_cursor) IA__gtk_text_buffer_insert_at_cursor __attribute((visibility("hidden")));
#define gtk_text_buffer_insert_at_cursor IA__gtk_text_buffer_insert_at_cursor

extern __typeof (gtk_text_buffer_insert_child_anchor) IA__gtk_text_buffer_insert_child_anchor __attribute((visibility("hidden")));
#define gtk_text_buffer_insert_child_anchor IA__gtk_text_buffer_insert_child_anchor

extern __typeof (gtk_text_buffer_insert_interactive) IA__gtk_text_buffer_insert_interactive __attribute((visibility("hidden")));
#define gtk_text_buffer_insert_interactive IA__gtk_text_buffer_insert_interactive

extern __typeof (gtk_text_buffer_insert_interactive_at_cursor) IA__gtk_text_buffer_insert_interactive_at_cursor __attribute((visibility("hidden")));
#define gtk_text_buffer_insert_interactive_at_cursor IA__gtk_text_buffer_insert_interactive_at_cursor

extern __typeof (gtk_text_buffer_insert_pixbuf) IA__gtk_text_buffer_insert_pixbuf __attribute((visibility("hidden")));
#define gtk_text_buffer_insert_pixbuf IA__gtk_text_buffer_insert_pixbuf

extern __typeof (gtk_text_buffer_insert_range) IA__gtk_text_buffer_insert_range __attribute((visibility("hidden")));
#define gtk_text_buffer_insert_range IA__gtk_text_buffer_insert_range

extern __typeof (gtk_text_buffer_insert_range_interactive) IA__gtk_text_buffer_insert_range_interactive __attribute((visibility("hidden")));
#define gtk_text_buffer_insert_range_interactive IA__gtk_text_buffer_insert_range_interactive

extern __typeof (gtk_text_buffer_insert_with_tags) IA__gtk_text_buffer_insert_with_tags __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_text_buffer_insert_with_tags IA__gtk_text_buffer_insert_with_tags

extern __typeof (gtk_text_buffer_insert_with_tags_by_name) IA__gtk_text_buffer_insert_with_tags_by_name __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_text_buffer_insert_with_tags_by_name IA__gtk_text_buffer_insert_with_tags_by_name

extern __typeof (gtk_text_buffer_move_mark) IA__gtk_text_buffer_move_mark __attribute((visibility("hidden")));
#define gtk_text_buffer_move_mark IA__gtk_text_buffer_move_mark

extern __typeof (gtk_text_buffer_move_mark_by_name) IA__gtk_text_buffer_move_mark_by_name __attribute((visibility("hidden")));
#define gtk_text_buffer_move_mark_by_name IA__gtk_text_buffer_move_mark_by_name

extern __typeof (gtk_text_buffer_new) IA__gtk_text_buffer_new __attribute((visibility("hidden")));
#define gtk_text_buffer_new IA__gtk_text_buffer_new

extern __typeof (gtk_text_buffer_paste_clipboard) IA__gtk_text_buffer_paste_clipboard __attribute((visibility("hidden")));
#define gtk_text_buffer_paste_clipboard IA__gtk_text_buffer_paste_clipboard

extern __typeof (gtk_text_buffer_place_cursor) IA__gtk_text_buffer_place_cursor __attribute((visibility("hidden")));
#define gtk_text_buffer_place_cursor IA__gtk_text_buffer_place_cursor

extern __typeof (gtk_text_buffer_remove_all_tags) IA__gtk_text_buffer_remove_all_tags __attribute((visibility("hidden")));
#define gtk_text_buffer_remove_all_tags IA__gtk_text_buffer_remove_all_tags

extern __typeof (gtk_text_buffer_remove_selection_clipboard) IA__gtk_text_buffer_remove_selection_clipboard __attribute((visibility("hidden")));
#define gtk_text_buffer_remove_selection_clipboard IA__gtk_text_buffer_remove_selection_clipboard

extern __typeof (gtk_text_buffer_remove_tag) IA__gtk_text_buffer_remove_tag __attribute((visibility("hidden")));
#define gtk_text_buffer_remove_tag IA__gtk_text_buffer_remove_tag

extern __typeof (gtk_text_buffer_remove_tag_by_name) IA__gtk_text_buffer_remove_tag_by_name __attribute((visibility("hidden")));
#define gtk_text_buffer_remove_tag_by_name IA__gtk_text_buffer_remove_tag_by_name

extern __typeof (gtk_text_buffer_select_range) IA__gtk_text_buffer_select_range __attribute((visibility("hidden")));
#define gtk_text_buffer_select_range IA__gtk_text_buffer_select_range

extern __typeof (gtk_text_buffer_set_modified) IA__gtk_text_buffer_set_modified __attribute((visibility("hidden")));
#define gtk_text_buffer_set_modified IA__gtk_text_buffer_set_modified

extern __typeof (gtk_text_buffer_set_text) IA__gtk_text_buffer_set_text __attribute((visibility("hidden")));
#define gtk_text_buffer_set_text IA__gtk_text_buffer_set_text

#endif
#endif
#if IN_HEADER(__GTK_TEXT_TYPES_H__)
#if IN_FILE(__GTK_TEXT_TYPES_C__)
extern __typeof (gtk_text_byte_begins_utf8_char) IA__gtk_text_byte_begins_utf8_char __attribute((visibility("hidden")));
#define gtk_text_byte_begins_utf8_char IA__gtk_text_byte_begins_utf8_char

#endif
#endif
#if IN_HEADER(__GTK_TEXT_ITER_H__)
#if IN_FILE(__GTK_TEXT_ITER_C__)
extern __typeof (gtk_text_iter_backward_char) IA__gtk_text_iter_backward_char __attribute((visibility("hidden")));
#define gtk_text_iter_backward_char IA__gtk_text_iter_backward_char

extern __typeof (gtk_text_iter_backward_chars) IA__gtk_text_iter_backward_chars __attribute((visibility("hidden")));
#define gtk_text_iter_backward_chars IA__gtk_text_iter_backward_chars

extern __typeof (gtk_text_iter_backward_cursor_position) IA__gtk_text_iter_backward_cursor_position __attribute((visibility("hidden")));
#define gtk_text_iter_backward_cursor_position IA__gtk_text_iter_backward_cursor_position

extern __typeof (gtk_text_iter_backward_cursor_positions) IA__gtk_text_iter_backward_cursor_positions __attribute((visibility("hidden")));
#define gtk_text_iter_backward_cursor_positions IA__gtk_text_iter_backward_cursor_positions

extern __typeof (gtk_text_iter_backward_find_char) IA__gtk_text_iter_backward_find_char __attribute((visibility("hidden")));
#define gtk_text_iter_backward_find_char IA__gtk_text_iter_backward_find_char

extern __typeof (gtk_text_iter_backward_line) IA__gtk_text_iter_backward_line __attribute((visibility("hidden")));
#define gtk_text_iter_backward_line IA__gtk_text_iter_backward_line

extern __typeof (gtk_text_iter_backward_lines) IA__gtk_text_iter_backward_lines __attribute((visibility("hidden")));
#define gtk_text_iter_backward_lines IA__gtk_text_iter_backward_lines

extern __typeof (gtk_text_iter_backward_search) IA__gtk_text_iter_backward_search __attribute((visibility("hidden")));
#define gtk_text_iter_backward_search IA__gtk_text_iter_backward_search

extern __typeof (gtk_text_iter_backward_sentence_start) IA__gtk_text_iter_backward_sentence_start __attribute((visibility("hidden")));
#define gtk_text_iter_backward_sentence_start IA__gtk_text_iter_backward_sentence_start

extern __typeof (gtk_text_iter_backward_sentence_starts) IA__gtk_text_iter_backward_sentence_starts __attribute((visibility("hidden")));
#define gtk_text_iter_backward_sentence_starts IA__gtk_text_iter_backward_sentence_starts

extern __typeof (gtk_text_iter_backward_to_tag_toggle) IA__gtk_text_iter_backward_to_tag_toggle __attribute((visibility("hidden")));
#define gtk_text_iter_backward_to_tag_toggle IA__gtk_text_iter_backward_to_tag_toggle

extern __typeof (gtk_text_iter_backward_visible_cursor_position) IA__gtk_text_iter_backward_visible_cursor_position __attribute((visibility("hidden")));
#define gtk_text_iter_backward_visible_cursor_position IA__gtk_text_iter_backward_visible_cursor_position

extern __typeof (gtk_text_iter_backward_visible_cursor_positions) IA__gtk_text_iter_backward_visible_cursor_positions __attribute((visibility("hidden")));
#define gtk_text_iter_backward_visible_cursor_positions IA__gtk_text_iter_backward_visible_cursor_positions

extern __typeof (gtk_text_iter_backward_visible_line) IA__gtk_text_iter_backward_visible_line __attribute((visibility("hidden")));
#define gtk_text_iter_backward_visible_line IA__gtk_text_iter_backward_visible_line

extern __typeof (gtk_text_iter_backward_visible_lines) IA__gtk_text_iter_backward_visible_lines __attribute((visibility("hidden")));
#define gtk_text_iter_backward_visible_lines IA__gtk_text_iter_backward_visible_lines

extern __typeof (gtk_text_iter_backward_visible_word_start) IA__gtk_text_iter_backward_visible_word_start __attribute((visibility("hidden")));
#define gtk_text_iter_backward_visible_word_start IA__gtk_text_iter_backward_visible_word_start

extern __typeof (gtk_text_iter_backward_visible_word_starts) IA__gtk_text_iter_backward_visible_word_starts __attribute((visibility("hidden")));
#define gtk_text_iter_backward_visible_word_starts IA__gtk_text_iter_backward_visible_word_starts

extern __typeof (gtk_text_iter_backward_word_start) IA__gtk_text_iter_backward_word_start __attribute((visibility("hidden")));
#define gtk_text_iter_backward_word_start IA__gtk_text_iter_backward_word_start

extern __typeof (gtk_text_iter_backward_word_starts) IA__gtk_text_iter_backward_word_starts __attribute((visibility("hidden")));
#define gtk_text_iter_backward_word_starts IA__gtk_text_iter_backward_word_starts

extern __typeof (gtk_text_iter_begins_tag) IA__gtk_text_iter_begins_tag __attribute((visibility("hidden")));
#define gtk_text_iter_begins_tag IA__gtk_text_iter_begins_tag

extern __typeof (gtk_text_iter_can_insert) IA__gtk_text_iter_can_insert __attribute((visibility("hidden")));
#define gtk_text_iter_can_insert IA__gtk_text_iter_can_insert

extern __typeof (gtk_text_iter_compare) IA__gtk_text_iter_compare __attribute((visibility("hidden")));
#define gtk_text_iter_compare IA__gtk_text_iter_compare

extern __typeof (gtk_text_iter_copy) IA__gtk_text_iter_copy __attribute((visibility("hidden")));
#define gtk_text_iter_copy IA__gtk_text_iter_copy

extern __typeof (gtk_text_iter_editable) IA__gtk_text_iter_editable __attribute((visibility("hidden")));
#define gtk_text_iter_editable IA__gtk_text_iter_editable

extern __typeof (gtk_text_iter_ends_line) IA__gtk_text_iter_ends_line __attribute((visibility("hidden")));
#define gtk_text_iter_ends_line IA__gtk_text_iter_ends_line

extern __typeof (gtk_text_iter_ends_sentence) IA__gtk_text_iter_ends_sentence __attribute((visibility("hidden")));
#define gtk_text_iter_ends_sentence IA__gtk_text_iter_ends_sentence

extern __typeof (gtk_text_iter_ends_tag) IA__gtk_text_iter_ends_tag __attribute((visibility("hidden")));
#define gtk_text_iter_ends_tag IA__gtk_text_iter_ends_tag

extern __typeof (gtk_text_iter_ends_word) IA__gtk_text_iter_ends_word __attribute((visibility("hidden")));
#define gtk_text_iter_ends_word IA__gtk_text_iter_ends_word

extern __typeof (gtk_text_iter_equal) IA__gtk_text_iter_equal __attribute((visibility("hidden")));
#define gtk_text_iter_equal IA__gtk_text_iter_equal

extern __typeof (gtk_text_iter_forward_char) IA__gtk_text_iter_forward_char __attribute((visibility("hidden")));
#define gtk_text_iter_forward_char IA__gtk_text_iter_forward_char

extern __typeof (gtk_text_iter_forward_chars) IA__gtk_text_iter_forward_chars __attribute((visibility("hidden")));
#define gtk_text_iter_forward_chars IA__gtk_text_iter_forward_chars

extern __typeof (gtk_text_iter_forward_cursor_position) IA__gtk_text_iter_forward_cursor_position __attribute((visibility("hidden")));
#define gtk_text_iter_forward_cursor_position IA__gtk_text_iter_forward_cursor_position

extern __typeof (gtk_text_iter_forward_cursor_positions) IA__gtk_text_iter_forward_cursor_positions __attribute((visibility("hidden")));
#define gtk_text_iter_forward_cursor_positions IA__gtk_text_iter_forward_cursor_positions

extern __typeof (gtk_text_iter_forward_find_char) IA__gtk_text_iter_forward_find_char __attribute((visibility("hidden")));
#define gtk_text_iter_forward_find_char IA__gtk_text_iter_forward_find_char

extern __typeof (gtk_text_iter_forward_line) IA__gtk_text_iter_forward_line __attribute((visibility("hidden")));
#define gtk_text_iter_forward_line IA__gtk_text_iter_forward_line

extern __typeof (gtk_text_iter_forward_lines) IA__gtk_text_iter_forward_lines __attribute((visibility("hidden")));
#define gtk_text_iter_forward_lines IA__gtk_text_iter_forward_lines

extern __typeof (gtk_text_iter_forward_search) IA__gtk_text_iter_forward_search __attribute((visibility("hidden")));
#define gtk_text_iter_forward_search IA__gtk_text_iter_forward_search

extern __typeof (gtk_text_iter_forward_sentence_end) IA__gtk_text_iter_forward_sentence_end __attribute((visibility("hidden")));
#define gtk_text_iter_forward_sentence_end IA__gtk_text_iter_forward_sentence_end

extern __typeof (gtk_text_iter_forward_sentence_ends) IA__gtk_text_iter_forward_sentence_ends __attribute((visibility("hidden")));
#define gtk_text_iter_forward_sentence_ends IA__gtk_text_iter_forward_sentence_ends

extern __typeof (gtk_text_iter_forward_to_end) IA__gtk_text_iter_forward_to_end __attribute((visibility("hidden")));
#define gtk_text_iter_forward_to_end IA__gtk_text_iter_forward_to_end

extern __typeof (gtk_text_iter_forward_to_line_end) IA__gtk_text_iter_forward_to_line_end __attribute((visibility("hidden")));
#define gtk_text_iter_forward_to_line_end IA__gtk_text_iter_forward_to_line_end

extern __typeof (gtk_text_iter_forward_to_tag_toggle) IA__gtk_text_iter_forward_to_tag_toggle __attribute((visibility("hidden")));
#define gtk_text_iter_forward_to_tag_toggle IA__gtk_text_iter_forward_to_tag_toggle

extern __typeof (gtk_text_iter_forward_visible_cursor_position) IA__gtk_text_iter_forward_visible_cursor_position __attribute((visibility("hidden")));
#define gtk_text_iter_forward_visible_cursor_position IA__gtk_text_iter_forward_visible_cursor_position

extern __typeof (gtk_text_iter_forward_visible_cursor_positions) IA__gtk_text_iter_forward_visible_cursor_positions __attribute((visibility("hidden")));
#define gtk_text_iter_forward_visible_cursor_positions IA__gtk_text_iter_forward_visible_cursor_positions

extern __typeof (gtk_text_iter_forward_visible_line) IA__gtk_text_iter_forward_visible_line __attribute((visibility("hidden")));
#define gtk_text_iter_forward_visible_line IA__gtk_text_iter_forward_visible_line

extern __typeof (gtk_text_iter_forward_visible_lines) IA__gtk_text_iter_forward_visible_lines __attribute((visibility("hidden")));
#define gtk_text_iter_forward_visible_lines IA__gtk_text_iter_forward_visible_lines

extern __typeof (gtk_text_iter_forward_visible_word_end) IA__gtk_text_iter_forward_visible_word_end __attribute((visibility("hidden")));
#define gtk_text_iter_forward_visible_word_end IA__gtk_text_iter_forward_visible_word_end

extern __typeof (gtk_text_iter_forward_visible_word_ends) IA__gtk_text_iter_forward_visible_word_ends __attribute((visibility("hidden")));
#define gtk_text_iter_forward_visible_word_ends IA__gtk_text_iter_forward_visible_word_ends

extern __typeof (gtk_text_iter_forward_word_end) IA__gtk_text_iter_forward_word_end __attribute((visibility("hidden")));
#define gtk_text_iter_forward_word_end IA__gtk_text_iter_forward_word_end

extern __typeof (gtk_text_iter_forward_word_ends) IA__gtk_text_iter_forward_word_ends __attribute((visibility("hidden")));
#define gtk_text_iter_forward_word_ends IA__gtk_text_iter_forward_word_ends

extern __typeof (gtk_text_iter_free) IA__gtk_text_iter_free __attribute((visibility("hidden")));
#define gtk_text_iter_free IA__gtk_text_iter_free

extern __typeof (gtk_text_iter_get_attributes) IA__gtk_text_iter_get_attributes __attribute((visibility("hidden")));
#define gtk_text_iter_get_attributes IA__gtk_text_iter_get_attributes

extern __typeof (gtk_text_iter_get_buffer) IA__gtk_text_iter_get_buffer __attribute((visibility("hidden")));
#define gtk_text_iter_get_buffer IA__gtk_text_iter_get_buffer

extern __typeof (gtk_text_iter_get_bytes_in_line) IA__gtk_text_iter_get_bytes_in_line __attribute((visibility("hidden")));
#define gtk_text_iter_get_bytes_in_line IA__gtk_text_iter_get_bytes_in_line

extern __typeof (gtk_text_iter_get_char) IA__gtk_text_iter_get_char __attribute((visibility("hidden")));
#define gtk_text_iter_get_char IA__gtk_text_iter_get_char

extern __typeof (gtk_text_iter_get_chars_in_line) IA__gtk_text_iter_get_chars_in_line __attribute((visibility("hidden")));
#define gtk_text_iter_get_chars_in_line IA__gtk_text_iter_get_chars_in_line

extern __typeof (gtk_text_iter_get_child_anchor) IA__gtk_text_iter_get_child_anchor __attribute((visibility("hidden")));
#define gtk_text_iter_get_child_anchor IA__gtk_text_iter_get_child_anchor

extern __typeof (gtk_text_iter_get_language) IA__gtk_text_iter_get_language __attribute((visibility("hidden")));
#define gtk_text_iter_get_language IA__gtk_text_iter_get_language

extern __typeof (gtk_text_iter_get_line) IA__gtk_text_iter_get_line __attribute((visibility("hidden")));
#define gtk_text_iter_get_line IA__gtk_text_iter_get_line

extern __typeof (gtk_text_iter_get_line_index) IA__gtk_text_iter_get_line_index __attribute((visibility("hidden")));
#define gtk_text_iter_get_line_index IA__gtk_text_iter_get_line_index

extern __typeof (gtk_text_iter_get_line_offset) IA__gtk_text_iter_get_line_offset __attribute((visibility("hidden")));
#define gtk_text_iter_get_line_offset IA__gtk_text_iter_get_line_offset

extern __typeof (gtk_text_iter_get_marks) IA__gtk_text_iter_get_marks __attribute((visibility("hidden")));
#define gtk_text_iter_get_marks IA__gtk_text_iter_get_marks

extern __typeof (gtk_text_iter_get_offset) IA__gtk_text_iter_get_offset __attribute((visibility("hidden")));
#define gtk_text_iter_get_offset IA__gtk_text_iter_get_offset

extern __typeof (gtk_text_iter_get_pixbuf) IA__gtk_text_iter_get_pixbuf __attribute((visibility("hidden")));
#define gtk_text_iter_get_pixbuf IA__gtk_text_iter_get_pixbuf

extern __typeof (gtk_text_iter_get_slice) IA__gtk_text_iter_get_slice __attribute((visibility("hidden")));
#define gtk_text_iter_get_slice IA__gtk_text_iter_get_slice

extern __typeof (gtk_text_iter_get_tags) IA__gtk_text_iter_get_tags __attribute((visibility("hidden")));
#define gtk_text_iter_get_tags IA__gtk_text_iter_get_tags

extern __typeof (gtk_text_iter_get_text) IA__gtk_text_iter_get_text __attribute((visibility("hidden")));
#define gtk_text_iter_get_text IA__gtk_text_iter_get_text

extern __typeof (gtk_text_iter_get_toggled_tags) IA__gtk_text_iter_get_toggled_tags __attribute((visibility("hidden")));
#define gtk_text_iter_get_toggled_tags IA__gtk_text_iter_get_toggled_tags

extern __typeof (gtk_text_iter_get_type) IA__gtk_text_iter_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_iter_get_type IA__gtk_text_iter_get_type

extern __typeof (gtk_text_iter_get_visible_line_index) IA__gtk_text_iter_get_visible_line_index __attribute((visibility("hidden")));
#define gtk_text_iter_get_visible_line_index IA__gtk_text_iter_get_visible_line_index

extern __typeof (gtk_text_iter_get_visible_line_offset) IA__gtk_text_iter_get_visible_line_offset __attribute((visibility("hidden")));
#define gtk_text_iter_get_visible_line_offset IA__gtk_text_iter_get_visible_line_offset

extern __typeof (gtk_text_iter_get_visible_slice) IA__gtk_text_iter_get_visible_slice __attribute((visibility("hidden")));
#define gtk_text_iter_get_visible_slice IA__gtk_text_iter_get_visible_slice

extern __typeof (gtk_text_iter_get_visible_text) IA__gtk_text_iter_get_visible_text __attribute((visibility("hidden")));
#define gtk_text_iter_get_visible_text IA__gtk_text_iter_get_visible_text

extern __typeof (gtk_text_iter_has_tag) IA__gtk_text_iter_has_tag __attribute((visibility("hidden")));
#define gtk_text_iter_has_tag IA__gtk_text_iter_has_tag

extern __typeof (gtk_text_iter_in_range) IA__gtk_text_iter_in_range __attribute((visibility("hidden")));
#define gtk_text_iter_in_range IA__gtk_text_iter_in_range

extern __typeof (gtk_text_iter_inside_sentence) IA__gtk_text_iter_inside_sentence __attribute((visibility("hidden")));
#define gtk_text_iter_inside_sentence IA__gtk_text_iter_inside_sentence

extern __typeof (gtk_text_iter_inside_word) IA__gtk_text_iter_inside_word __attribute((visibility("hidden")));
#define gtk_text_iter_inside_word IA__gtk_text_iter_inside_word

extern __typeof (gtk_text_iter_is_cursor_position) IA__gtk_text_iter_is_cursor_position __attribute((visibility("hidden")));
#define gtk_text_iter_is_cursor_position IA__gtk_text_iter_is_cursor_position

extern __typeof (gtk_text_iter_is_end) IA__gtk_text_iter_is_end __attribute((visibility("hidden")));
#define gtk_text_iter_is_end IA__gtk_text_iter_is_end

extern __typeof (gtk_text_iter_is_start) IA__gtk_text_iter_is_start __attribute((visibility("hidden")));
#define gtk_text_iter_is_start IA__gtk_text_iter_is_start

extern __typeof (gtk_text_iter_order) IA__gtk_text_iter_order __attribute((visibility("hidden")));
#define gtk_text_iter_order IA__gtk_text_iter_order

extern __typeof (gtk_text_iter_set_line) IA__gtk_text_iter_set_line __attribute((visibility("hidden")));
#define gtk_text_iter_set_line IA__gtk_text_iter_set_line

extern __typeof (gtk_text_iter_set_line_index) IA__gtk_text_iter_set_line_index __attribute((visibility("hidden")));
#define gtk_text_iter_set_line_index IA__gtk_text_iter_set_line_index

extern __typeof (gtk_text_iter_set_line_offset) IA__gtk_text_iter_set_line_offset __attribute((visibility("hidden")));
#define gtk_text_iter_set_line_offset IA__gtk_text_iter_set_line_offset

extern __typeof (gtk_text_iter_set_offset) IA__gtk_text_iter_set_offset __attribute((visibility("hidden")));
#define gtk_text_iter_set_offset IA__gtk_text_iter_set_offset

extern __typeof (gtk_text_iter_set_visible_line_index) IA__gtk_text_iter_set_visible_line_index __attribute((visibility("hidden")));
#define gtk_text_iter_set_visible_line_index IA__gtk_text_iter_set_visible_line_index

extern __typeof (gtk_text_iter_set_visible_line_offset) IA__gtk_text_iter_set_visible_line_offset __attribute((visibility("hidden")));
#define gtk_text_iter_set_visible_line_offset IA__gtk_text_iter_set_visible_line_offset

extern __typeof (gtk_text_iter_starts_line) IA__gtk_text_iter_starts_line __attribute((visibility("hidden")));
#define gtk_text_iter_starts_line IA__gtk_text_iter_starts_line

extern __typeof (gtk_text_iter_starts_sentence) IA__gtk_text_iter_starts_sentence __attribute((visibility("hidden")));
#define gtk_text_iter_starts_sentence IA__gtk_text_iter_starts_sentence

extern __typeof (gtk_text_iter_starts_word) IA__gtk_text_iter_starts_word __attribute((visibility("hidden")));
#define gtk_text_iter_starts_word IA__gtk_text_iter_starts_word

extern __typeof (gtk_text_iter_toggles_tag) IA__gtk_text_iter_toggles_tag __attribute((visibility("hidden")));
#define gtk_text_iter_toggles_tag IA__gtk_text_iter_toggles_tag

#endif
#endif
#if IN_HEADER(__GTK_TEXT_LAYOUT_H__)
#if IN_FILE(__GTK_TEXT_LAYOUT_C__)
extern __typeof (gtk_text_layout_changed) IA__gtk_text_layout_changed __attribute((visibility("hidden")));
#define gtk_text_layout_changed IA__gtk_text_layout_changed

extern __typeof (gtk_text_layout_cursors_changed) IA__gtk_text_layout_cursors_changed __attribute((visibility("hidden")));
#define gtk_text_layout_cursors_changed IA__gtk_text_layout_cursors_changed

extern __typeof (gtk_text_layout_clamp_iter_to_vrange) IA__gtk_text_layout_clamp_iter_to_vrange __attribute((visibility("hidden")));
#define gtk_text_layout_clamp_iter_to_vrange IA__gtk_text_layout_clamp_iter_to_vrange

extern __typeof (gtk_text_layout_default_style_changed) IA__gtk_text_layout_default_style_changed __attribute((visibility("hidden")));
#define gtk_text_layout_default_style_changed IA__gtk_text_layout_default_style_changed

extern __typeof (gtk_text_layout_free_line_data) IA__gtk_text_layout_free_line_data __attribute((visibility("hidden")));
#define gtk_text_layout_free_line_data IA__gtk_text_layout_free_line_data

extern __typeof (gtk_text_layout_free_line_display) IA__gtk_text_layout_free_line_display __attribute((visibility("hidden")));
#define gtk_text_layout_free_line_display IA__gtk_text_layout_free_line_display

extern __typeof (gtk_text_layout_get_buffer) IA__gtk_text_layout_get_buffer __attribute((visibility("hidden")));
#define gtk_text_layout_get_buffer IA__gtk_text_layout_get_buffer

extern __typeof (gtk_text_layout_get_cursor_locations) IA__gtk_text_layout_get_cursor_locations __attribute((visibility("hidden")));
#define gtk_text_layout_get_cursor_locations IA__gtk_text_layout_get_cursor_locations

extern __typeof (gtk_text_layout_get_cursor_visible) IA__gtk_text_layout_get_cursor_visible __attribute((visibility("hidden")));
#define gtk_text_layout_get_cursor_visible IA__gtk_text_layout_get_cursor_visible

extern __typeof (gtk_text_layout_get_iter_at_line) IA__gtk_text_layout_get_iter_at_line __attribute((visibility("hidden")));
#define gtk_text_layout_get_iter_at_line IA__gtk_text_layout_get_iter_at_line

extern __typeof (gtk_text_layout_get_iter_at_pixel) IA__gtk_text_layout_get_iter_at_pixel __attribute((visibility("hidden")));
#define gtk_text_layout_get_iter_at_pixel IA__gtk_text_layout_get_iter_at_pixel

extern __typeof (gtk_text_layout_get_iter_at_position) IA__gtk_text_layout_get_iter_at_position __attribute((visibility("hidden")));
#define gtk_text_layout_get_iter_at_position IA__gtk_text_layout_get_iter_at_position

extern __typeof (gtk_text_layout_get_iter_location) IA__gtk_text_layout_get_iter_location __attribute((visibility("hidden")));
#define gtk_text_layout_get_iter_location IA__gtk_text_layout_get_iter_location

extern __typeof (gtk_text_layout_get_line_at_y) IA__gtk_text_layout_get_line_at_y __attribute((visibility("hidden")));
#define gtk_text_layout_get_line_at_y IA__gtk_text_layout_get_line_at_y

extern __typeof (gtk_text_layout_get_line_display) IA__gtk_text_layout_get_line_display __attribute((visibility("hidden")));
#define gtk_text_layout_get_line_display IA__gtk_text_layout_get_line_display

extern __typeof (gtk_text_layout_get_lines) IA__gtk_text_layout_get_lines __attribute((visibility("hidden")));
#define gtk_text_layout_get_lines IA__gtk_text_layout_get_lines

extern __typeof (gtk_text_layout_get_line_yrange) IA__gtk_text_layout_get_line_yrange __attribute((visibility("hidden")));
#define gtk_text_layout_get_line_yrange IA__gtk_text_layout_get_line_yrange

extern __typeof (gtk_text_layout_get_size) IA__gtk_text_layout_get_size __attribute((visibility("hidden")));
#define gtk_text_layout_get_size IA__gtk_text_layout_get_size

extern __typeof (gtk_text_layout_get_type) IA__gtk_text_layout_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_layout_get_type IA__gtk_text_layout_get_type

extern __typeof (gtk_text_layout_invalidate) IA__gtk_text_layout_invalidate __attribute((visibility("hidden")));
#define gtk_text_layout_invalidate IA__gtk_text_layout_invalidate

extern __typeof (gtk_text_layout_invalidate_cursors) IA__gtk_text_layout_invalidate_cursors __attribute((visibility("hidden")));
#define gtk_text_layout_invalidate_cursors IA__gtk_text_layout_invalidate_cursors

extern __typeof (gtk_text_layout_is_valid) IA__gtk_text_layout_is_valid __attribute((visibility("hidden")));
#define gtk_text_layout_is_valid IA__gtk_text_layout_is_valid

extern __typeof (gtk_text_layout_iter_starts_line) IA__gtk_text_layout_iter_starts_line __attribute((visibility("hidden")));
#define gtk_text_layout_iter_starts_line IA__gtk_text_layout_iter_starts_line

extern __typeof (gtk_text_layout_move_iter_to_line_end) IA__gtk_text_layout_move_iter_to_line_end __attribute((visibility("hidden")));
#define gtk_text_layout_move_iter_to_line_end IA__gtk_text_layout_move_iter_to_line_end

extern __typeof (gtk_text_layout_move_iter_to_next_line) IA__gtk_text_layout_move_iter_to_next_line __attribute((visibility("hidden")));
#define gtk_text_layout_move_iter_to_next_line IA__gtk_text_layout_move_iter_to_next_line

extern __typeof (gtk_text_layout_move_iter_to_previous_line) IA__gtk_text_layout_move_iter_to_previous_line __attribute((visibility("hidden")));
#define gtk_text_layout_move_iter_to_previous_line IA__gtk_text_layout_move_iter_to_previous_line

extern __typeof (gtk_text_layout_move_iter_to_x) IA__gtk_text_layout_move_iter_to_x __attribute((visibility("hidden")));
#define gtk_text_layout_move_iter_to_x IA__gtk_text_layout_move_iter_to_x

extern __typeof (gtk_text_layout_move_iter_visually) IA__gtk_text_layout_move_iter_visually __attribute((visibility("hidden")));
#define gtk_text_layout_move_iter_visually IA__gtk_text_layout_move_iter_visually

extern __typeof (gtk_text_layout_new) IA__gtk_text_layout_new __attribute((visibility("hidden")));
#define gtk_text_layout_new IA__gtk_text_layout_new

extern __typeof (gtk_text_layout_set_buffer) IA__gtk_text_layout_set_buffer __attribute((visibility("hidden")));
#define gtk_text_layout_set_buffer IA__gtk_text_layout_set_buffer

extern __typeof (gtk_text_layout_set_contexts) IA__gtk_text_layout_set_contexts __attribute((visibility("hidden")));
#define gtk_text_layout_set_contexts IA__gtk_text_layout_set_contexts

extern __typeof (gtk_text_layout_set_cursor_direction) IA__gtk_text_layout_set_cursor_direction __attribute((visibility("hidden")));
#define gtk_text_layout_set_cursor_direction IA__gtk_text_layout_set_cursor_direction

extern __typeof (gtk_text_layout_set_cursor_visible) IA__gtk_text_layout_set_cursor_visible __attribute((visibility("hidden")));
#define gtk_text_layout_set_cursor_visible IA__gtk_text_layout_set_cursor_visible

extern __typeof (gtk_text_layout_set_default_style) IA__gtk_text_layout_set_default_style __attribute((visibility("hidden")));
#define gtk_text_layout_set_default_style IA__gtk_text_layout_set_default_style

extern __typeof (gtk_text_layout_set_keyboard_direction) IA__gtk_text_layout_set_keyboard_direction __attribute((visibility("hidden")));
#define gtk_text_layout_set_keyboard_direction IA__gtk_text_layout_set_keyboard_direction

extern __typeof (gtk_text_layout_set_overwrite_mode) IA__gtk_text_layout_set_overwrite_mode __attribute((visibility("hidden")));
#define gtk_text_layout_set_overwrite_mode IA__gtk_text_layout_set_overwrite_mode

extern __typeof (gtk_text_layout_set_preedit_string) IA__gtk_text_layout_set_preedit_string __attribute((visibility("hidden")));
#define gtk_text_layout_set_preedit_string IA__gtk_text_layout_set_preedit_string

extern __typeof (gtk_text_layout_set_screen_width) IA__gtk_text_layout_set_screen_width __attribute((visibility("hidden")));
#define gtk_text_layout_set_screen_width IA__gtk_text_layout_set_screen_width

extern __typeof (gtk_text_layout_spew) IA__gtk_text_layout_spew __attribute((visibility("hidden")));
#define gtk_text_layout_spew IA__gtk_text_layout_spew

extern __typeof (gtk_text_layout_validate) IA__gtk_text_layout_validate __attribute((visibility("hidden")));
#define gtk_text_layout_validate IA__gtk_text_layout_validate

extern __typeof (gtk_text_layout_validate_yrange) IA__gtk_text_layout_validate_yrange __attribute((visibility("hidden")));
#define gtk_text_layout_validate_yrange IA__gtk_text_layout_validate_yrange

extern __typeof (gtk_text_layout_wrap) IA__gtk_text_layout_wrap __attribute((visibility("hidden")));
#define gtk_text_layout_wrap IA__gtk_text_layout_wrap

extern __typeof (gtk_text_layout_wrap_loop_end) IA__gtk_text_layout_wrap_loop_end __attribute((visibility("hidden")));
#define gtk_text_layout_wrap_loop_end IA__gtk_text_layout_wrap_loop_end

extern __typeof (gtk_text_layout_wrap_loop_start) IA__gtk_text_layout_wrap_loop_start __attribute((visibility("hidden")));
#define gtk_text_layout_wrap_loop_start IA__gtk_text_layout_wrap_loop_start

#endif
#endif
#if IN_HEADER(__GTK_TEXT_CHILD_H__)
#if IN_FILE(__GTK_TEXT_CHILD_C__)
extern __typeof (gtk_text_child_anchor_get_deleted) IA__gtk_text_child_anchor_get_deleted __attribute((visibility("hidden")));
#define gtk_text_child_anchor_get_deleted IA__gtk_text_child_anchor_get_deleted

extern __typeof (gtk_text_child_anchor_get_type) IA__gtk_text_child_anchor_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_child_anchor_get_type IA__gtk_text_child_anchor_get_type

extern __typeof (gtk_text_child_anchor_get_widgets) IA__gtk_text_child_anchor_get_widgets __attribute((visibility("hidden")));
#define gtk_text_child_anchor_get_widgets IA__gtk_text_child_anchor_get_widgets

extern __typeof (gtk_text_child_anchor_new) IA__gtk_text_child_anchor_new __attribute((visibility("hidden")));
#define gtk_text_child_anchor_new IA__gtk_text_child_anchor_new

#endif
#endif
#if IN_HEADER(__GTK_TEXT_LAYOUT_H__)
#if IN_FILE(__GTK_TEXT_CHILD_C__)
extern __typeof (gtk_text_child_anchor_queue_resize) IA__gtk_text_child_anchor_queue_resize __attribute((visibility("hidden")));
#define gtk_text_child_anchor_queue_resize IA__gtk_text_child_anchor_queue_resize

extern __typeof (gtk_text_child_anchor_register_child) IA__gtk_text_child_anchor_register_child __attribute((visibility("hidden")));
#define gtk_text_child_anchor_register_child IA__gtk_text_child_anchor_register_child

extern __typeof (gtk_text_child_anchor_unregister_child) IA__gtk_text_child_anchor_unregister_child __attribute((visibility("hidden")));
#define gtk_text_child_anchor_unregister_child IA__gtk_text_child_anchor_unregister_child

extern __typeof (gtk_text_anchored_child_set_layout) IA__gtk_text_anchored_child_set_layout __attribute((visibility("hidden")));
#define gtk_text_anchored_child_set_layout IA__gtk_text_anchored_child_set_layout

#endif
#endif
#if IN_HEADER(__GTK_TEXT_DISPLAY_H__)
#if IN_FILE(__GTK_TEXT_DISPLAY_C__)
extern __typeof (gtk_text_layout_draw) IA__gtk_text_layout_draw __attribute((visibility("hidden")));
#define gtk_text_layout_draw IA__gtk_text_layout_draw

#endif
#endif
#if IN_HEADER(__GTK_TEXT_SEGMENT_H__)
#if IN_FILE(__GTK_TEXT_SEGMENT_C__)
extern __typeof (gtk_text_line_segment_split) IA__gtk_text_line_segment_split __attribute((visibility("hidden")));
#define gtk_text_line_segment_split IA__gtk_text_line_segment_split

#endif
#endif
#if IN_HEADER(__GTK_TEXT_MARK_H__)
#if IN_FILE(__GTK_TEXT_MARK_C__)
extern __typeof (gtk_text_mark_new) IA__gtk_text_mark_new __attribute((visibility("hidden")));
#define gtk_text_mark_new IA__gtk_text_mark_new

extern __typeof (gtk_text_mark_get_buffer) IA__gtk_text_mark_get_buffer __attribute((visibility("hidden")));
#define gtk_text_mark_get_buffer IA__gtk_text_mark_get_buffer

extern __typeof (gtk_text_mark_get_deleted) IA__gtk_text_mark_get_deleted __attribute((visibility("hidden")));
#define gtk_text_mark_get_deleted IA__gtk_text_mark_get_deleted

extern __typeof (gtk_text_mark_get_left_gravity) IA__gtk_text_mark_get_left_gravity __attribute((visibility("hidden")));
#define gtk_text_mark_get_left_gravity IA__gtk_text_mark_get_left_gravity

extern __typeof (gtk_text_mark_get_name) IA__gtk_text_mark_get_name __attribute((visibility("hidden")));
#define gtk_text_mark_get_name IA__gtk_text_mark_get_name

extern __typeof (gtk_text_mark_get_type) IA__gtk_text_mark_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_mark_get_type IA__gtk_text_mark_get_type

extern __typeof (gtk_text_mark_get_visible) IA__gtk_text_mark_get_visible __attribute((visibility("hidden")));
#define gtk_text_mark_get_visible IA__gtk_text_mark_get_visible

#endif
#endif
#if IN_HEADER(__GTK_TEXT_MARK_H__)
#if IN_FILE(__GTK_TEXT_BTREE_C__)
extern __typeof (gtk_text_mark_set_visible) IA__gtk_text_mark_set_visible __attribute((visibility("hidden")));
#define gtk_text_mark_set_visible IA__gtk_text_mark_set_visible

#endif
#endif
#if IN_HEADER(__GTK_TEXT_VIEW_H__)
#if IN_FILE(__GTK_TEXT_VIEW_C__)
extern __typeof (gtk_text_view_add_child_at_anchor) IA__gtk_text_view_add_child_at_anchor __attribute((visibility("hidden")));
#define gtk_text_view_add_child_at_anchor IA__gtk_text_view_add_child_at_anchor

extern __typeof (gtk_text_view_add_child_in_window) IA__gtk_text_view_add_child_in_window __attribute((visibility("hidden")));
#define gtk_text_view_add_child_in_window IA__gtk_text_view_add_child_in_window

extern __typeof (gtk_text_view_backward_display_line) IA__gtk_text_view_backward_display_line __attribute((visibility("hidden")));
#define gtk_text_view_backward_display_line IA__gtk_text_view_backward_display_line

extern __typeof (gtk_text_view_backward_display_line_start) IA__gtk_text_view_backward_display_line_start __attribute((visibility("hidden")));
#define gtk_text_view_backward_display_line_start IA__gtk_text_view_backward_display_line_start

extern __typeof (gtk_text_view_buffer_to_window_coords) IA__gtk_text_view_buffer_to_window_coords __attribute((visibility("hidden")));
#define gtk_text_view_buffer_to_window_coords IA__gtk_text_view_buffer_to_window_coords

extern __typeof (gtk_text_view_forward_display_line) IA__gtk_text_view_forward_display_line __attribute((visibility("hidden")));
#define gtk_text_view_forward_display_line IA__gtk_text_view_forward_display_line

extern __typeof (gtk_text_view_forward_display_line_end) IA__gtk_text_view_forward_display_line_end __attribute((visibility("hidden")));
#define gtk_text_view_forward_display_line_end IA__gtk_text_view_forward_display_line_end

extern __typeof (gtk_text_view_get_accepts_tab) IA__gtk_text_view_get_accepts_tab __attribute((visibility("hidden")));
#define gtk_text_view_get_accepts_tab IA__gtk_text_view_get_accepts_tab

extern __typeof (gtk_text_view_get_border_window_size) IA__gtk_text_view_get_border_window_size __attribute((visibility("hidden")));
#define gtk_text_view_get_border_window_size IA__gtk_text_view_get_border_window_size

extern __typeof (gtk_text_view_get_buffer) IA__gtk_text_view_get_buffer __attribute((visibility("hidden")));
#define gtk_text_view_get_buffer IA__gtk_text_view_get_buffer

extern __typeof (gtk_text_view_get_cursor_visible) IA__gtk_text_view_get_cursor_visible __attribute((visibility("hidden")));
#define gtk_text_view_get_cursor_visible IA__gtk_text_view_get_cursor_visible

extern __typeof (gtk_text_view_get_default_attributes) IA__gtk_text_view_get_default_attributes __attribute((visibility("hidden")));
#define gtk_text_view_get_default_attributes IA__gtk_text_view_get_default_attributes

extern __typeof (gtk_text_view_get_editable) IA__gtk_text_view_get_editable __attribute((visibility("hidden")));
#define gtk_text_view_get_editable IA__gtk_text_view_get_editable

extern __typeof (gtk_text_view_get_hadjustment) IA__gtk_text_view_get_hadjustment __attribute((visibility("hidden")));
#define gtk_text_view_get_hadjustment IA__gtk_text_view_get_hadjustment

extern __typeof (gtk_text_view_get_indent) IA__gtk_text_view_get_indent __attribute((visibility("hidden")));
#define gtk_text_view_get_indent IA__gtk_text_view_get_indent

extern __typeof (gtk_text_view_get_iter_at_location) IA__gtk_text_view_get_iter_at_location __attribute((visibility("hidden")));
#define gtk_text_view_get_iter_at_location IA__gtk_text_view_get_iter_at_location

extern __typeof (gtk_text_view_get_iter_at_position) IA__gtk_text_view_get_iter_at_position __attribute((visibility("hidden")));
#define gtk_text_view_get_iter_at_position IA__gtk_text_view_get_iter_at_position

extern __typeof (gtk_text_view_get_iter_location) IA__gtk_text_view_get_iter_location __attribute((visibility("hidden")));
#define gtk_text_view_get_iter_location IA__gtk_text_view_get_iter_location

extern __typeof (gtk_text_view_get_justification) IA__gtk_text_view_get_justification __attribute((visibility("hidden")));
#define gtk_text_view_get_justification IA__gtk_text_view_get_justification

extern __typeof (gtk_text_view_get_left_margin) IA__gtk_text_view_get_left_margin __attribute((visibility("hidden")));
#define gtk_text_view_get_left_margin IA__gtk_text_view_get_left_margin

extern __typeof (gtk_text_view_get_line_at_y) IA__gtk_text_view_get_line_at_y __attribute((visibility("hidden")));
#define gtk_text_view_get_line_at_y IA__gtk_text_view_get_line_at_y

extern __typeof (gtk_text_view_get_line_yrange) IA__gtk_text_view_get_line_yrange __attribute((visibility("hidden")));
#define gtk_text_view_get_line_yrange IA__gtk_text_view_get_line_yrange

extern __typeof (gtk_text_view_get_overwrite) IA__gtk_text_view_get_overwrite __attribute((visibility("hidden")));
#define gtk_text_view_get_overwrite IA__gtk_text_view_get_overwrite

extern __typeof (gtk_text_view_get_pixels_above_lines) IA__gtk_text_view_get_pixels_above_lines __attribute((visibility("hidden")));
#define gtk_text_view_get_pixels_above_lines IA__gtk_text_view_get_pixels_above_lines

extern __typeof (gtk_text_view_get_pixels_below_lines) IA__gtk_text_view_get_pixels_below_lines __attribute((visibility("hidden")));
#define gtk_text_view_get_pixels_below_lines IA__gtk_text_view_get_pixels_below_lines

extern __typeof (gtk_text_view_get_pixels_inside_wrap) IA__gtk_text_view_get_pixels_inside_wrap __attribute((visibility("hidden")));
#define gtk_text_view_get_pixels_inside_wrap IA__gtk_text_view_get_pixels_inside_wrap

extern __typeof (gtk_text_view_get_right_margin) IA__gtk_text_view_get_right_margin __attribute((visibility("hidden")));
#define gtk_text_view_get_right_margin IA__gtk_text_view_get_right_margin

extern __typeof (gtk_text_view_get_tabs) IA__gtk_text_view_get_tabs __attribute((visibility("hidden")));
#define gtk_text_view_get_tabs IA__gtk_text_view_get_tabs

extern __typeof (gtk_text_view_get_type) IA__gtk_text_view_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_text_view_get_type IA__gtk_text_view_get_type

extern __typeof (gtk_text_view_get_vadjustment) IA__gtk_text_view_get_vadjustment __attribute((visibility("hidden")));
#define gtk_text_view_get_vadjustment IA__gtk_text_view_get_vadjustment

extern __typeof (gtk_text_view_get_visible_rect) IA__gtk_text_view_get_visible_rect __attribute((visibility("hidden")));
#define gtk_text_view_get_visible_rect IA__gtk_text_view_get_visible_rect

extern __typeof (gtk_text_view_get_window) IA__gtk_text_view_get_window __attribute((visibility("hidden")));
#define gtk_text_view_get_window IA__gtk_text_view_get_window

extern __typeof (gtk_text_view_get_window_type) IA__gtk_text_view_get_window_type __attribute((visibility("hidden")));
#define gtk_text_view_get_window_type IA__gtk_text_view_get_window_type

extern __typeof (gtk_text_view_get_wrap_mode) IA__gtk_text_view_get_wrap_mode __attribute((visibility("hidden")));
#define gtk_text_view_get_wrap_mode IA__gtk_text_view_get_wrap_mode

extern __typeof (gtk_text_view_im_context_filter_keypress) IA__gtk_text_view_im_context_filter_keypress __attribute((visibility("hidden")));
#define gtk_text_view_im_context_filter_keypress IA__gtk_text_view_im_context_filter_keypress

extern __typeof (gtk_text_view_move_child) IA__gtk_text_view_move_child __attribute((visibility("hidden")));
#define gtk_text_view_move_child IA__gtk_text_view_move_child

extern __typeof (gtk_text_view_move_mark_onscreen) IA__gtk_text_view_move_mark_onscreen __attribute((visibility("hidden")));
#define gtk_text_view_move_mark_onscreen IA__gtk_text_view_move_mark_onscreen

extern __typeof (gtk_text_view_move_visually) IA__gtk_text_view_move_visually __attribute((visibility("hidden")));
#define gtk_text_view_move_visually IA__gtk_text_view_move_visually

extern __typeof (gtk_text_view_new) IA__gtk_text_view_new __attribute((visibility("hidden")));
#define gtk_text_view_new IA__gtk_text_view_new

extern __typeof (gtk_text_view_new_with_buffer) IA__gtk_text_view_new_with_buffer __attribute((visibility("hidden")));
#define gtk_text_view_new_with_buffer IA__gtk_text_view_new_with_buffer

extern __typeof (gtk_text_view_place_cursor_onscreen) IA__gtk_text_view_place_cursor_onscreen __attribute((visibility("hidden")));
#define gtk_text_view_place_cursor_onscreen IA__gtk_text_view_place_cursor_onscreen

extern __typeof (gtk_text_view_reset_im_context) IA__gtk_text_view_reset_im_context __attribute((visibility("hidden")));
#define gtk_text_view_reset_im_context IA__gtk_text_view_reset_im_context

extern __typeof (gtk_text_view_scroll_mark_onscreen) IA__gtk_text_view_scroll_mark_onscreen __attribute((visibility("hidden")));
#define gtk_text_view_scroll_mark_onscreen IA__gtk_text_view_scroll_mark_onscreen

extern __typeof (gtk_text_view_scroll_to_iter) IA__gtk_text_view_scroll_to_iter __attribute((visibility("hidden")));
#define gtk_text_view_scroll_to_iter IA__gtk_text_view_scroll_to_iter

extern __typeof (gtk_text_view_scroll_to_mark) IA__gtk_text_view_scroll_to_mark __attribute((visibility("hidden")));
#define gtk_text_view_scroll_to_mark IA__gtk_text_view_scroll_to_mark

extern __typeof (gtk_text_view_set_accepts_tab) IA__gtk_text_view_set_accepts_tab __attribute((visibility("hidden")));
#define gtk_text_view_set_accepts_tab IA__gtk_text_view_set_accepts_tab

extern __typeof (gtk_text_view_set_border_window_size) IA__gtk_text_view_set_border_window_size __attribute((visibility("hidden")));
#define gtk_text_view_set_border_window_size IA__gtk_text_view_set_border_window_size

extern __typeof (gtk_text_view_set_buffer) IA__gtk_text_view_set_buffer __attribute((visibility("hidden")));
#define gtk_text_view_set_buffer IA__gtk_text_view_set_buffer

extern __typeof (gtk_text_view_set_cursor_visible) IA__gtk_text_view_set_cursor_visible __attribute((visibility("hidden")));
#define gtk_text_view_set_cursor_visible IA__gtk_text_view_set_cursor_visible

extern __typeof (gtk_text_view_set_editable) IA__gtk_text_view_set_editable __attribute((visibility("hidden")));
#define gtk_text_view_set_editable IA__gtk_text_view_set_editable

extern __typeof (gtk_text_view_set_indent) IA__gtk_text_view_set_indent __attribute((visibility("hidden")));
#define gtk_text_view_set_indent IA__gtk_text_view_set_indent

extern __typeof (gtk_text_view_set_justification) IA__gtk_text_view_set_justification __attribute((visibility("hidden")));
#define gtk_text_view_set_justification IA__gtk_text_view_set_justification

extern __typeof (gtk_text_view_set_left_margin) IA__gtk_text_view_set_left_margin __attribute((visibility("hidden")));
#define gtk_text_view_set_left_margin IA__gtk_text_view_set_left_margin

extern __typeof (gtk_text_view_set_overwrite) IA__gtk_text_view_set_overwrite __attribute((visibility("hidden")));
#define gtk_text_view_set_overwrite IA__gtk_text_view_set_overwrite

extern __typeof (gtk_text_view_set_pixels_above_lines) IA__gtk_text_view_set_pixels_above_lines __attribute((visibility("hidden")));
#define gtk_text_view_set_pixels_above_lines IA__gtk_text_view_set_pixels_above_lines

extern __typeof (gtk_text_view_set_pixels_below_lines) IA__gtk_text_view_set_pixels_below_lines __attribute((visibility("hidden")));
#define gtk_text_view_set_pixels_below_lines IA__gtk_text_view_set_pixels_below_lines

extern __typeof (gtk_text_view_set_pixels_inside_wrap) IA__gtk_text_view_set_pixels_inside_wrap __attribute((visibility("hidden")));
#define gtk_text_view_set_pixels_inside_wrap IA__gtk_text_view_set_pixels_inside_wrap

extern __typeof (gtk_text_view_set_right_margin) IA__gtk_text_view_set_right_margin __attribute((visibility("hidden")));
#define gtk_text_view_set_right_margin IA__gtk_text_view_set_right_margin

extern __typeof (gtk_text_view_set_tabs) IA__gtk_text_view_set_tabs __attribute((visibility("hidden")));
#define gtk_text_view_set_tabs IA__gtk_text_view_set_tabs

extern __typeof (gtk_text_view_set_wrap_mode) IA__gtk_text_view_set_wrap_mode __attribute((visibility("hidden")));
#define gtk_text_view_set_wrap_mode IA__gtk_text_view_set_wrap_mode

extern __typeof (gtk_text_view_starts_display_line) IA__gtk_text_view_starts_display_line __attribute((visibility("hidden")));
#define gtk_text_view_starts_display_line IA__gtk_text_view_starts_display_line

extern __typeof (gtk_text_view_window_to_buffer_coords) IA__gtk_text_view_window_to_buffer_coords __attribute((visibility("hidden")));
#define gtk_text_view_window_to_buffer_coords IA__gtk_text_view_window_to_buffer_coords

#endif
#endif
#if IN_HEADER(__GTK_THEMES_H__)
#if IN_FILE(__GTK_THEMES_C__)
extern __typeof (gtk_theme_engine_create_rc_style) IA__gtk_theme_engine_create_rc_style __attribute((visibility("hidden")));
#define gtk_theme_engine_create_rc_style IA__gtk_theme_engine_create_rc_style

extern __typeof (gtk_theme_engine_get) IA__gtk_theme_engine_get __attribute((visibility("hidden")));
#define gtk_theme_engine_get IA__gtk_theme_engine_get

extern __typeof (gtk_theme_engine_get_type) IA__gtk_theme_engine_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_theme_engine_get_type IA__gtk_theme_engine_get_type

#endif
#endif
#if IN_HEADER(__GTK_TIPS_QUERY_H__)
#if IN_FILE(__GTK_TIPS_QUERY_C__)
extern __typeof (gtk_tips_query_get_type) IA__gtk_tips_query_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tips_query_get_type IA__gtk_tips_query_get_type

extern __typeof (gtk_tips_query_new) IA__gtk_tips_query_new __attribute((visibility("hidden")));
#define gtk_tips_query_new IA__gtk_tips_query_new

extern __typeof (gtk_tips_query_set_caller) IA__gtk_tips_query_set_caller __attribute((visibility("hidden")));
#define gtk_tips_query_set_caller IA__gtk_tips_query_set_caller

extern __typeof (gtk_tips_query_set_labels) IA__gtk_tips_query_set_labels __attribute((visibility("hidden")));
#define gtk_tips_query_set_labels IA__gtk_tips_query_set_labels

extern __typeof (gtk_tips_query_start_query) IA__gtk_tips_query_start_query __attribute((visibility("hidden")));
#define gtk_tips_query_start_query IA__gtk_tips_query_start_query

extern __typeof (gtk_tips_query_stop_query) IA__gtk_tips_query_stop_query __attribute((visibility("hidden")));
#define gtk_tips_query_stop_query IA__gtk_tips_query_stop_query

#endif
#endif
#if IN_HEADER(__GTK_TOGGLE_ACTION_H__)
#if IN_FILE(__GTK_TOGGLE_ACTION_C__)
extern __typeof (gtk_toggle_action_get_active) IA__gtk_toggle_action_get_active __attribute((visibility("hidden")));
#define gtk_toggle_action_get_active IA__gtk_toggle_action_get_active

extern __typeof (gtk_toggle_action_get_draw_as_radio) IA__gtk_toggle_action_get_draw_as_radio __attribute((visibility("hidden")));
#define gtk_toggle_action_get_draw_as_radio IA__gtk_toggle_action_get_draw_as_radio

extern __typeof (gtk_toggle_action_get_type) IA__gtk_toggle_action_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_toggle_action_get_type IA__gtk_toggle_action_get_type

extern __typeof (gtk_toggle_action_new) IA__gtk_toggle_action_new __attribute((visibility("hidden")));
#define gtk_toggle_action_new IA__gtk_toggle_action_new

extern __typeof (gtk_toggle_action_set_active) IA__gtk_toggle_action_set_active __attribute((visibility("hidden")));
#define gtk_toggle_action_set_active IA__gtk_toggle_action_set_active

extern __typeof (gtk_toggle_action_set_draw_as_radio) IA__gtk_toggle_action_set_draw_as_radio __attribute((visibility("hidden")));
#define gtk_toggle_action_set_draw_as_radio IA__gtk_toggle_action_set_draw_as_radio

extern __typeof (gtk_toggle_action_toggled) IA__gtk_toggle_action_toggled __attribute((visibility("hidden")));
#define gtk_toggle_action_toggled IA__gtk_toggle_action_toggled

#endif
#endif
#if IN_HEADER(__GTK_TOGGLE_BUTTON_H__)
#if IN_FILE(__GTK_TOGGLE_BUTTON_C__)
extern __typeof (gtk_toggle_button_get_active) IA__gtk_toggle_button_get_active __attribute((visibility("hidden")));
#define gtk_toggle_button_get_active IA__gtk_toggle_button_get_active

extern __typeof (gtk_toggle_button_get_inconsistent) IA__gtk_toggle_button_get_inconsistent __attribute((visibility("hidden")));
#define gtk_toggle_button_get_inconsistent IA__gtk_toggle_button_get_inconsistent

extern __typeof (gtk_toggle_button_get_mode) IA__gtk_toggle_button_get_mode __attribute((visibility("hidden")));
#define gtk_toggle_button_get_mode IA__gtk_toggle_button_get_mode

extern __typeof (gtk_toggle_button_get_type) IA__gtk_toggle_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_toggle_button_get_type IA__gtk_toggle_button_get_type

extern __typeof (gtk_toggle_button_new) IA__gtk_toggle_button_new __attribute((visibility("hidden")));
#define gtk_toggle_button_new IA__gtk_toggle_button_new

extern __typeof (gtk_toggle_button_new_with_label) IA__gtk_toggle_button_new_with_label __attribute((visibility("hidden")));
#define gtk_toggle_button_new_with_label IA__gtk_toggle_button_new_with_label

extern __typeof (gtk_toggle_button_new_with_mnemonic) IA__gtk_toggle_button_new_with_mnemonic __attribute((visibility("hidden")));
#define gtk_toggle_button_new_with_mnemonic IA__gtk_toggle_button_new_with_mnemonic

extern __typeof (gtk_toggle_button_set_active) IA__gtk_toggle_button_set_active __attribute((visibility("hidden")));
#define gtk_toggle_button_set_active IA__gtk_toggle_button_set_active

extern __typeof (gtk_toggle_button_set_inconsistent) IA__gtk_toggle_button_set_inconsistent __attribute((visibility("hidden")));
#define gtk_toggle_button_set_inconsistent IA__gtk_toggle_button_set_inconsistent

extern __typeof (gtk_toggle_button_set_mode) IA__gtk_toggle_button_set_mode __attribute((visibility("hidden")));
#define gtk_toggle_button_set_mode IA__gtk_toggle_button_set_mode

extern __typeof (gtk_toggle_button_toggled) IA__gtk_toggle_button_toggled __attribute((visibility("hidden")));
#define gtk_toggle_button_toggled IA__gtk_toggle_button_toggled

#endif
#endif
#if IN_HEADER(__GTK_TOGGLE_TOOL_BUTTON_H__)
#if IN_FILE(__GTK_TOGGLE_TOOL_BUTTON_C__)
extern __typeof (gtk_toggle_tool_button_get_active) IA__gtk_toggle_tool_button_get_active __attribute((visibility("hidden")));
#define gtk_toggle_tool_button_get_active IA__gtk_toggle_tool_button_get_active

extern __typeof (gtk_toggle_tool_button_get_type) IA__gtk_toggle_tool_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_toggle_tool_button_get_type IA__gtk_toggle_tool_button_get_type

extern __typeof (gtk_toggle_tool_button_new) IA__gtk_toggle_tool_button_new __attribute((visibility("hidden")));
#define gtk_toggle_tool_button_new IA__gtk_toggle_tool_button_new

extern __typeof (gtk_toggle_tool_button_new_from_stock) IA__gtk_toggle_tool_button_new_from_stock __attribute((visibility("hidden")));
#define gtk_toggle_tool_button_new_from_stock IA__gtk_toggle_tool_button_new_from_stock

extern __typeof (gtk_toggle_tool_button_set_active) IA__gtk_toggle_tool_button_set_active __attribute((visibility("hidden")));
#define gtk_toggle_tool_button_set_active IA__gtk_toggle_tool_button_set_active

#endif
#endif
#if IN_HEADER(__GTK_TOOLBAR_H__)
#if IN_FILE(__GTK_TOOLBAR_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_toolbar_append_element) IA__gtk_toolbar_append_element __attribute((visibility("hidden")));
#define gtk_toolbar_append_element IA__gtk_toolbar_append_element

extern __typeof (gtk_toolbar_set_icon_size) IA__gtk_toolbar_set_icon_size __attribute((visibility("hidden")));
#define gtk_toolbar_set_icon_size IA__gtk_toolbar_set_icon_size

extern __typeof (gtk_toolbar_unset_icon_size) IA__gtk_toolbar_unset_icon_size __attribute((visibility("hidden")));
#define gtk_toolbar_unset_icon_size IA__gtk_toolbar_unset_icon_size

extern __typeof (gtk_toolbar_append_item) IA__gtk_toolbar_append_item __attribute((visibility("hidden")));
#define gtk_toolbar_append_item IA__gtk_toolbar_append_item

extern __typeof (gtk_toolbar_prepend_element) IA__gtk_toolbar_prepend_element __attribute((visibility("hidden")));
#define gtk_toolbar_prepend_element IA__gtk_toolbar_prepend_element

extern __typeof (gtk_toolbar_prepend_item) IA__gtk_toolbar_prepend_item __attribute((visibility("hidden")));
#define gtk_toolbar_prepend_item IA__gtk_toolbar_prepend_item

extern __typeof (gtk_toolbar_insert_element) IA__gtk_toolbar_insert_element __attribute((visibility("hidden")));
#define gtk_toolbar_insert_element IA__gtk_toolbar_insert_element

extern __typeof (gtk_toolbar_insert_item) IA__gtk_toolbar_insert_item __attribute((visibility("hidden")));
#define gtk_toolbar_insert_item IA__gtk_toolbar_insert_item

extern __typeof (gtk_toolbar_insert_space) IA__gtk_toolbar_insert_space __attribute((visibility("hidden")));
#define gtk_toolbar_insert_space IA__gtk_toolbar_insert_space

extern __typeof (gtk_toolbar_insert_stock) IA__gtk_toolbar_insert_stock __attribute((visibility("hidden")));
#define gtk_toolbar_insert_stock IA__gtk_toolbar_insert_stock

extern __typeof (gtk_toolbar_append_space) IA__gtk_toolbar_append_space __attribute((visibility("hidden")));
#define gtk_toolbar_append_space IA__gtk_toolbar_append_space

extern __typeof (gtk_toolbar_prepend_space) IA__gtk_toolbar_prepend_space __attribute((visibility("hidden")));
#define gtk_toolbar_prepend_space IA__gtk_toolbar_prepend_space

extern __typeof (gtk_toolbar_remove_space) IA__gtk_toolbar_remove_space __attribute((visibility("hidden")));
#define gtk_toolbar_remove_space IA__gtk_toolbar_remove_space

extern __typeof (gtk_toolbar_append_widget) IA__gtk_toolbar_append_widget __attribute((visibility("hidden")));
#define gtk_toolbar_append_widget IA__gtk_toolbar_append_widget

extern __typeof (gtk_toolbar_insert_widget) IA__gtk_toolbar_insert_widget __attribute((visibility("hidden")));
#define gtk_toolbar_insert_widget IA__gtk_toolbar_insert_widget

extern __typeof (gtk_toolbar_prepend_widget) IA__gtk_toolbar_prepend_widget __attribute((visibility("hidden")));
#define gtk_toolbar_prepend_widget IA__gtk_toolbar_prepend_widget

#endif
extern __typeof (gtk_toolbar_get_drop_index) IA__gtk_toolbar_get_drop_index __attribute((visibility("hidden")));
#define gtk_toolbar_get_drop_index IA__gtk_toolbar_get_drop_index

extern __typeof (gtk_toolbar_get_icon_size) IA__gtk_toolbar_get_icon_size __attribute((visibility("hidden")));
#define gtk_toolbar_get_icon_size IA__gtk_toolbar_get_icon_size

extern __typeof (gtk_toolbar_get_item_index) IA__gtk_toolbar_get_item_index __attribute((visibility("hidden")));
#define gtk_toolbar_get_item_index IA__gtk_toolbar_get_item_index

extern __typeof (gtk_toolbar_get_n_items) IA__gtk_toolbar_get_n_items __attribute((visibility("hidden")));
#define gtk_toolbar_get_n_items IA__gtk_toolbar_get_n_items

extern __typeof (gtk_toolbar_get_nth_item) IA__gtk_toolbar_get_nth_item __attribute((visibility("hidden")));
#define gtk_toolbar_get_nth_item IA__gtk_toolbar_get_nth_item

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_toolbar_get_orientation) IA__gtk_toolbar_get_orientation __attribute((visibility("hidden")));
#define gtk_toolbar_get_orientation IA__gtk_toolbar_get_orientation

#endif
extern __typeof (gtk_toolbar_get_relief_style) IA__gtk_toolbar_get_relief_style __attribute((visibility("hidden")));
#define gtk_toolbar_get_relief_style IA__gtk_toolbar_get_relief_style

extern __typeof (gtk_toolbar_get_show_arrow) IA__gtk_toolbar_get_show_arrow __attribute((visibility("hidden")));
#define gtk_toolbar_get_show_arrow IA__gtk_toolbar_get_show_arrow

extern __typeof (gtk_toolbar_get_style) IA__gtk_toolbar_get_style __attribute((visibility("hidden")));
#define gtk_toolbar_get_style IA__gtk_toolbar_get_style

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_toolbar_get_tooltips) IA__gtk_toolbar_get_tooltips __attribute((visibility("hidden")));
#define gtk_toolbar_get_tooltips IA__gtk_toolbar_get_tooltips

#endif
extern __typeof (gtk_toolbar_get_type) IA__gtk_toolbar_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_toolbar_get_type IA__gtk_toolbar_get_type

extern __typeof (gtk_toolbar_insert) IA__gtk_toolbar_insert __attribute((visibility("hidden")));
#define gtk_toolbar_insert IA__gtk_toolbar_insert

extern __typeof (gtk_toolbar_new) IA__gtk_toolbar_new __attribute((visibility("hidden")));
#define gtk_toolbar_new IA__gtk_toolbar_new

extern __typeof (gtk_toolbar_set_drop_highlight_item) IA__gtk_toolbar_set_drop_highlight_item __attribute((visibility("hidden")));
#define gtk_toolbar_set_drop_highlight_item IA__gtk_toolbar_set_drop_highlight_item

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_toolbar_set_orientation) IA__gtk_toolbar_set_orientation __attribute((visibility("hidden")));
#define gtk_toolbar_set_orientation IA__gtk_toolbar_set_orientation

#endif
extern __typeof (gtk_toolbar_set_show_arrow) IA__gtk_toolbar_set_show_arrow __attribute((visibility("hidden")));
#define gtk_toolbar_set_show_arrow IA__gtk_toolbar_set_show_arrow

extern __typeof (gtk_toolbar_set_style) IA__gtk_toolbar_set_style __attribute((visibility("hidden")));
#define gtk_toolbar_set_style IA__gtk_toolbar_set_style

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_toolbar_set_tooltips) IA__gtk_toolbar_set_tooltips __attribute((visibility("hidden")));
#define gtk_toolbar_set_tooltips IA__gtk_toolbar_set_tooltips

#endif
extern __typeof (gtk_toolbar_unset_style) IA__gtk_toolbar_unset_style __attribute((visibility("hidden")));
#define gtk_toolbar_unset_style IA__gtk_toolbar_unset_style

#endif
#endif
#if IN_HEADER(__GTK_TOOL_SHELL_H__)
#if IN_FILE(__GTK_TOOL_SHELL_C__)
extern __typeof (gtk_tool_shell_get_type) IA__gtk_tool_shell_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tool_shell_get_type IA__gtk_tool_shell_get_type

extern __typeof (gtk_tool_shell_get_ellipsize_mode) IA__gtk_tool_shell_get_ellipsize_mode __attribute((visibility("hidden")));
#define gtk_tool_shell_get_ellipsize_mode IA__gtk_tool_shell_get_ellipsize_mode

extern __typeof (gtk_tool_shell_get_icon_size) IA__gtk_tool_shell_get_icon_size __attribute((visibility("hidden")));
#define gtk_tool_shell_get_icon_size IA__gtk_tool_shell_get_icon_size

extern __typeof (gtk_tool_shell_get_orientation) IA__gtk_tool_shell_get_orientation __attribute((visibility("hidden")));
#define gtk_tool_shell_get_orientation IA__gtk_tool_shell_get_orientation

extern __typeof (gtk_tool_shell_get_style) IA__gtk_tool_shell_get_style __attribute((visibility("hidden")));
#define gtk_tool_shell_get_style IA__gtk_tool_shell_get_style

extern __typeof (gtk_tool_shell_get_relief_style) IA__gtk_tool_shell_get_relief_style __attribute((visibility("hidden")));
#define gtk_tool_shell_get_relief_style IA__gtk_tool_shell_get_relief_style

extern __typeof (gtk_tool_shell_get_text_alignment) IA__gtk_tool_shell_get_text_alignment __attribute((visibility("hidden")));
#define gtk_tool_shell_get_text_alignment IA__gtk_tool_shell_get_text_alignment

extern __typeof (gtk_tool_shell_get_text_orientation) IA__gtk_tool_shell_get_text_orientation __attribute((visibility("hidden")));
#define gtk_tool_shell_get_text_orientation IA__gtk_tool_shell_get_text_orientation

extern __typeof (gtk_tool_shell_get_text_size_group) IA__gtk_tool_shell_get_text_size_group __attribute((visibility("hidden")));
#define gtk_tool_shell_get_text_size_group IA__gtk_tool_shell_get_text_size_group

extern __typeof (gtk_tool_shell_rebuild_menu) IA__gtk_tool_shell_rebuild_menu __attribute((visibility("hidden")));
#define gtk_tool_shell_rebuild_menu IA__gtk_tool_shell_rebuild_menu

#endif
#endif
#if IN_HEADER(__GTK_TOOL_BUTTON_H__)
#if IN_FILE(__GTK_TOOL_BUTTON_C__)
extern __typeof (gtk_tool_button_get_icon_name) IA__gtk_tool_button_get_icon_name __attribute((visibility("hidden")));
#define gtk_tool_button_get_icon_name IA__gtk_tool_button_get_icon_name

extern __typeof (gtk_tool_button_get_icon_widget) IA__gtk_tool_button_get_icon_widget __attribute((visibility("hidden")));
#define gtk_tool_button_get_icon_widget IA__gtk_tool_button_get_icon_widget

extern __typeof (gtk_tool_button_get_label) IA__gtk_tool_button_get_label __attribute((visibility("hidden")));
#define gtk_tool_button_get_label IA__gtk_tool_button_get_label

extern __typeof (gtk_tool_button_get_label_widget) IA__gtk_tool_button_get_label_widget __attribute((visibility("hidden")));
#define gtk_tool_button_get_label_widget IA__gtk_tool_button_get_label_widget

extern __typeof (gtk_tool_button_get_stock_id) IA__gtk_tool_button_get_stock_id __attribute((visibility("hidden")));
#define gtk_tool_button_get_stock_id IA__gtk_tool_button_get_stock_id

extern __typeof (gtk_tool_button_get_type) IA__gtk_tool_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tool_button_get_type IA__gtk_tool_button_get_type

extern __typeof (gtk_tool_button_get_use_underline) IA__gtk_tool_button_get_use_underline __attribute((visibility("hidden")));
#define gtk_tool_button_get_use_underline IA__gtk_tool_button_get_use_underline

extern __typeof (gtk_tool_button_new) IA__gtk_tool_button_new __attribute((visibility("hidden")));
#define gtk_tool_button_new IA__gtk_tool_button_new

extern __typeof (gtk_tool_button_new_from_stock) IA__gtk_tool_button_new_from_stock __attribute((visibility("hidden")));
#define gtk_tool_button_new_from_stock IA__gtk_tool_button_new_from_stock

extern __typeof (gtk_tool_button_set_icon_name) IA__gtk_tool_button_set_icon_name __attribute((visibility("hidden")));
#define gtk_tool_button_set_icon_name IA__gtk_tool_button_set_icon_name

extern __typeof (gtk_tool_button_set_icon_widget) IA__gtk_tool_button_set_icon_widget __attribute((visibility("hidden")));
#define gtk_tool_button_set_icon_widget IA__gtk_tool_button_set_icon_widget

extern __typeof (gtk_tool_button_set_label) IA__gtk_tool_button_set_label __attribute((visibility("hidden")));
#define gtk_tool_button_set_label IA__gtk_tool_button_set_label

extern __typeof (gtk_tool_button_set_label_widget) IA__gtk_tool_button_set_label_widget __attribute((visibility("hidden")));
#define gtk_tool_button_set_label_widget IA__gtk_tool_button_set_label_widget

extern __typeof (gtk_tool_button_set_stock_id) IA__gtk_tool_button_set_stock_id __attribute((visibility("hidden")));
#define gtk_tool_button_set_stock_id IA__gtk_tool_button_set_stock_id

extern __typeof (gtk_tool_button_set_use_underline) IA__gtk_tool_button_set_use_underline __attribute((visibility("hidden")));
#define gtk_tool_button_set_use_underline IA__gtk_tool_button_set_use_underline

#endif
#endif
#if IN_HEADER(__GTK_TOOL_ITEM_H__)
#if IN_FILE(__GTK_TOOL_ITEM_C__)
extern __typeof (gtk_tool_item_get_ellipsize_mode) IA__gtk_tool_item_get_ellipsize_mode __attribute((visibility("hidden")));
#define gtk_tool_item_get_ellipsize_mode IA__gtk_tool_item_get_ellipsize_mode

extern __typeof (gtk_tool_item_get_expand) IA__gtk_tool_item_get_expand __attribute((visibility("hidden")));
#define gtk_tool_item_get_expand IA__gtk_tool_item_get_expand

extern __typeof (gtk_tool_item_get_homogeneous) IA__gtk_tool_item_get_homogeneous __attribute((visibility("hidden")));
#define gtk_tool_item_get_homogeneous IA__gtk_tool_item_get_homogeneous

extern __typeof (gtk_tool_item_get_icon_size) IA__gtk_tool_item_get_icon_size __attribute((visibility("hidden")));
#define gtk_tool_item_get_icon_size IA__gtk_tool_item_get_icon_size

extern __typeof (gtk_tool_item_get_is_important) IA__gtk_tool_item_get_is_important __attribute((visibility("hidden")));
#define gtk_tool_item_get_is_important IA__gtk_tool_item_get_is_important

extern __typeof (gtk_tool_item_get_orientation) IA__gtk_tool_item_get_orientation __attribute((visibility("hidden")));
#define gtk_tool_item_get_orientation IA__gtk_tool_item_get_orientation

extern __typeof (gtk_tool_item_get_proxy_menu_item) IA__gtk_tool_item_get_proxy_menu_item __attribute((visibility("hidden")));
#define gtk_tool_item_get_proxy_menu_item IA__gtk_tool_item_get_proxy_menu_item

extern __typeof (gtk_tool_item_get_relief_style) IA__gtk_tool_item_get_relief_style __attribute((visibility("hidden")));
#define gtk_tool_item_get_relief_style IA__gtk_tool_item_get_relief_style

extern __typeof (gtk_tool_item_get_text_alignment) IA__gtk_tool_item_get_text_alignment __attribute((visibility("hidden")));
#define gtk_tool_item_get_text_alignment IA__gtk_tool_item_get_text_alignment

extern __typeof (gtk_tool_item_get_text_orientation) IA__gtk_tool_item_get_text_orientation __attribute((visibility("hidden")));
#define gtk_tool_item_get_text_orientation IA__gtk_tool_item_get_text_orientation

extern __typeof (gtk_tool_item_get_text_size_group) IA__gtk_tool_item_get_text_size_group __attribute((visibility("hidden")));
#define gtk_tool_item_get_text_size_group IA__gtk_tool_item_get_text_size_group

extern __typeof (gtk_tool_item_get_toolbar_style) IA__gtk_tool_item_get_toolbar_style __attribute((visibility("hidden")));
#define gtk_tool_item_get_toolbar_style IA__gtk_tool_item_get_toolbar_style

extern __typeof (gtk_tool_item_get_type) IA__gtk_tool_item_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tool_item_get_type IA__gtk_tool_item_get_type

extern __typeof (gtk_tool_item_get_use_drag_window) IA__gtk_tool_item_get_use_drag_window __attribute((visibility("hidden")));
#define gtk_tool_item_get_use_drag_window IA__gtk_tool_item_get_use_drag_window

extern __typeof (gtk_tool_item_get_visible_horizontal) IA__gtk_tool_item_get_visible_horizontal __attribute((visibility("hidden")));
#define gtk_tool_item_get_visible_horizontal IA__gtk_tool_item_get_visible_horizontal

extern __typeof (gtk_tool_item_get_visible_vertical) IA__gtk_tool_item_get_visible_vertical __attribute((visibility("hidden")));
#define gtk_tool_item_get_visible_vertical IA__gtk_tool_item_get_visible_vertical

extern __typeof (gtk_tool_item_new) IA__gtk_tool_item_new __attribute((visibility("hidden")));
#define gtk_tool_item_new IA__gtk_tool_item_new

extern __typeof (gtk_tool_item_rebuild_menu) IA__gtk_tool_item_rebuild_menu __attribute((visibility("hidden")));
#define gtk_tool_item_rebuild_menu IA__gtk_tool_item_rebuild_menu

extern __typeof (gtk_tool_item_retrieve_proxy_menu_item) IA__gtk_tool_item_retrieve_proxy_menu_item __attribute((visibility("hidden")));
#define gtk_tool_item_retrieve_proxy_menu_item IA__gtk_tool_item_retrieve_proxy_menu_item

extern __typeof (gtk_tool_item_set_expand) IA__gtk_tool_item_set_expand __attribute((visibility("hidden")));
#define gtk_tool_item_set_expand IA__gtk_tool_item_set_expand

extern __typeof (gtk_tool_item_set_homogeneous) IA__gtk_tool_item_set_homogeneous __attribute((visibility("hidden")));
#define gtk_tool_item_set_homogeneous IA__gtk_tool_item_set_homogeneous

extern __typeof (gtk_tool_item_set_is_important) IA__gtk_tool_item_set_is_important __attribute((visibility("hidden")));
#define gtk_tool_item_set_is_important IA__gtk_tool_item_set_is_important

extern __typeof (gtk_tool_item_set_proxy_menu_item) IA__gtk_tool_item_set_proxy_menu_item __attribute((visibility("hidden")));
#define gtk_tool_item_set_proxy_menu_item IA__gtk_tool_item_set_proxy_menu_item

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_tool_item_set_tooltip) IA__gtk_tool_item_set_tooltip __attribute((visibility("hidden")));
#define gtk_tool_item_set_tooltip IA__gtk_tool_item_set_tooltip

#endif
extern __typeof (gtk_tool_item_set_tooltip_markup) IA__gtk_tool_item_set_tooltip_markup __attribute((visibility("hidden")));
#define gtk_tool_item_set_tooltip_markup IA__gtk_tool_item_set_tooltip_markup

extern __typeof (gtk_tool_item_set_tooltip_text) IA__gtk_tool_item_set_tooltip_text __attribute((visibility("hidden")));
#define gtk_tool_item_set_tooltip_text IA__gtk_tool_item_set_tooltip_text

extern __typeof (gtk_tool_item_set_use_drag_window) IA__gtk_tool_item_set_use_drag_window __attribute((visibility("hidden")));
#define gtk_tool_item_set_use_drag_window IA__gtk_tool_item_set_use_drag_window

extern __typeof (gtk_tool_item_set_visible_horizontal) IA__gtk_tool_item_set_visible_horizontal __attribute((visibility("hidden")));
#define gtk_tool_item_set_visible_horizontal IA__gtk_tool_item_set_visible_horizontal

extern __typeof (gtk_tool_item_set_visible_vertical) IA__gtk_tool_item_set_visible_vertical __attribute((visibility("hidden")));
#define gtk_tool_item_set_visible_vertical IA__gtk_tool_item_set_visible_vertical

extern __typeof (gtk_tool_item_toolbar_reconfigured) IA__gtk_tool_item_toolbar_reconfigured __attribute((visibility("hidden")));
#define gtk_tool_item_toolbar_reconfigured IA__gtk_tool_item_toolbar_reconfigured

#endif
#endif
#if IN_HEADER(__GTK_TOOL_ITEM_GROUP_H__)
#if IN_FILE(__GTK_TOOL_ITEM_GROUP_C__)
extern __typeof (gtk_tool_item_group_get_type) IA__gtk_tool_item_group_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tool_item_group_get_type IA__gtk_tool_item_group_get_type

extern __typeof (gtk_tool_item_group_new) IA__gtk_tool_item_group_new __attribute((visibility("hidden")));
#define gtk_tool_item_group_new IA__gtk_tool_item_group_new

extern __typeof (gtk_tool_item_group_set_label) IA__gtk_tool_item_group_set_label __attribute((visibility("hidden")));
#define gtk_tool_item_group_set_label IA__gtk_tool_item_group_set_label

extern __typeof (gtk_tool_item_group_set_label_widget) IA__gtk_tool_item_group_set_label_widget __attribute((visibility("hidden")));
#define gtk_tool_item_group_set_label_widget IA__gtk_tool_item_group_set_label_widget

extern __typeof (gtk_tool_item_group_set_collapsed) IA__gtk_tool_item_group_set_collapsed __attribute((visibility("hidden")));
#define gtk_tool_item_group_set_collapsed IA__gtk_tool_item_group_set_collapsed

extern __typeof (gtk_tool_item_group_set_ellipsize) IA__gtk_tool_item_group_set_ellipsize __attribute((visibility("hidden")));
#define gtk_tool_item_group_set_ellipsize IA__gtk_tool_item_group_set_ellipsize

extern __typeof (gtk_tool_item_group_set_header_relief) IA__gtk_tool_item_group_set_header_relief __attribute((visibility("hidden")));
#define gtk_tool_item_group_set_header_relief IA__gtk_tool_item_group_set_header_relief

extern __typeof (gtk_tool_item_group_get_label) IA__gtk_tool_item_group_get_label __attribute((visibility("hidden")));
#define gtk_tool_item_group_get_label IA__gtk_tool_item_group_get_label

extern __typeof (gtk_tool_item_group_get_label_widget) IA__gtk_tool_item_group_get_label_widget __attribute((visibility("hidden")));
#define gtk_tool_item_group_get_label_widget IA__gtk_tool_item_group_get_label_widget

extern __typeof (gtk_tool_item_group_get_ellipsize) IA__gtk_tool_item_group_get_ellipsize __attribute((visibility("hidden")));
#define gtk_tool_item_group_get_ellipsize IA__gtk_tool_item_group_get_ellipsize

extern __typeof (gtk_tool_item_group_get_header_relief) IA__gtk_tool_item_group_get_header_relief __attribute((visibility("hidden")));
#define gtk_tool_item_group_get_header_relief IA__gtk_tool_item_group_get_header_relief

extern __typeof (gtk_tool_item_group_insert) IA__gtk_tool_item_group_insert __attribute((visibility("hidden")));
#define gtk_tool_item_group_insert IA__gtk_tool_item_group_insert

extern __typeof (gtk_tool_item_group_set_item_position) IA__gtk_tool_item_group_set_item_position __attribute((visibility("hidden")));
#define gtk_tool_item_group_set_item_position IA__gtk_tool_item_group_set_item_position

extern __typeof (gtk_tool_item_group_get_item_position) IA__gtk_tool_item_group_get_item_position __attribute((visibility("hidden")));
#define gtk_tool_item_group_get_item_position IA__gtk_tool_item_group_get_item_position

extern __typeof (gtk_tool_item_group_get_n_items) IA__gtk_tool_item_group_get_n_items __attribute((visibility("hidden")));
#define gtk_tool_item_group_get_n_items IA__gtk_tool_item_group_get_n_items

extern __typeof (gtk_tool_item_group_get_nth_item) IA__gtk_tool_item_group_get_nth_item __attribute((visibility("hidden")));
#define gtk_tool_item_group_get_nth_item IA__gtk_tool_item_group_get_nth_item

extern __typeof (gtk_tool_item_group_get_drop_item) IA__gtk_tool_item_group_get_drop_item __attribute((visibility("hidden")));
#define gtk_tool_item_group_get_drop_item IA__gtk_tool_item_group_get_drop_item

extern __typeof (gtk_tool_item_group_get_collapsed) IA__gtk_tool_item_group_get_collapsed __attribute((visibility("hidden")));
#define gtk_tool_item_group_get_collapsed IA__gtk_tool_item_group_get_collapsed

#endif
#endif
#if IN_HEADER(__GTK_TOOL_PALETTE_H__)
#if IN_FILE(__GTK_TOOL_PALETTE_C__)
extern __typeof (gtk_tool_palette_get_type) IA__gtk_tool_palette_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tool_palette_get_type IA__gtk_tool_palette_get_type

extern __typeof (gtk_tool_palette_new) IA__gtk_tool_palette_new __attribute((visibility("hidden")));
#define gtk_tool_palette_new IA__gtk_tool_palette_new

extern __typeof (gtk_tool_palette_set_group_position) IA__gtk_tool_palette_set_group_position __attribute((visibility("hidden")));
#define gtk_tool_palette_set_group_position IA__gtk_tool_palette_set_group_position

extern __typeof (gtk_tool_palette_set_exclusive) IA__gtk_tool_palette_set_exclusive __attribute((visibility("hidden")));
#define gtk_tool_palette_set_exclusive IA__gtk_tool_palette_set_exclusive

extern __typeof (gtk_tool_palette_set_expand) IA__gtk_tool_palette_set_expand __attribute((visibility("hidden")));
#define gtk_tool_palette_set_expand IA__gtk_tool_palette_set_expand

extern __typeof (gtk_tool_palette_get_group_position) IA__gtk_tool_palette_get_group_position __attribute((visibility("hidden")));
#define gtk_tool_palette_get_group_position IA__gtk_tool_palette_get_group_position

extern __typeof (gtk_tool_palette_get_exclusive) IA__gtk_tool_palette_get_exclusive __attribute((visibility("hidden")));
#define gtk_tool_palette_get_exclusive IA__gtk_tool_palette_get_exclusive

extern __typeof (gtk_tool_palette_get_expand) IA__gtk_tool_palette_get_expand __attribute((visibility("hidden")));
#define gtk_tool_palette_get_expand IA__gtk_tool_palette_get_expand

extern __typeof (gtk_tool_palette_set_icon_size) IA__gtk_tool_palette_set_icon_size __attribute((visibility("hidden")));
#define gtk_tool_palette_set_icon_size IA__gtk_tool_palette_set_icon_size

extern __typeof (gtk_tool_palette_unset_icon_size) IA__gtk_tool_palette_unset_icon_size __attribute((visibility("hidden")));
#define gtk_tool_palette_unset_icon_size IA__gtk_tool_palette_unset_icon_size

extern __typeof (gtk_tool_palette_set_style) IA__gtk_tool_palette_set_style __attribute((visibility("hidden")));
#define gtk_tool_palette_set_style IA__gtk_tool_palette_set_style

extern __typeof (gtk_tool_palette_unset_style) IA__gtk_tool_palette_unset_style __attribute((visibility("hidden")));
#define gtk_tool_palette_unset_style IA__gtk_tool_palette_unset_style

extern __typeof (gtk_tool_palette_get_icon_size) IA__gtk_tool_palette_get_icon_size __attribute((visibility("hidden")));
#define gtk_tool_palette_get_icon_size IA__gtk_tool_palette_get_icon_size

extern __typeof (gtk_tool_palette_get_style) IA__gtk_tool_palette_get_style __attribute((visibility("hidden")));
#define gtk_tool_palette_get_style IA__gtk_tool_palette_get_style

extern __typeof (gtk_tool_palette_get_drop_item) IA__gtk_tool_palette_get_drop_item __attribute((visibility("hidden")));
#define gtk_tool_palette_get_drop_item IA__gtk_tool_palette_get_drop_item

extern __typeof (gtk_tool_palette_get_drop_group) IA__gtk_tool_palette_get_drop_group __attribute((visibility("hidden")));
#define gtk_tool_palette_get_drop_group IA__gtk_tool_palette_get_drop_group

extern __typeof (gtk_tool_palette_get_drag_item) IA__gtk_tool_palette_get_drag_item __attribute((visibility("hidden")));
#define gtk_tool_palette_get_drag_item IA__gtk_tool_palette_get_drag_item

extern __typeof (gtk_tool_palette_set_drag_source) IA__gtk_tool_palette_set_drag_source __attribute((visibility("hidden")));
#define gtk_tool_palette_set_drag_source IA__gtk_tool_palette_set_drag_source

extern __typeof (gtk_tool_palette_add_drag_dest) IA__gtk_tool_palette_add_drag_dest __attribute((visibility("hidden")));
#define gtk_tool_palette_add_drag_dest IA__gtk_tool_palette_add_drag_dest

extern __typeof (gtk_tool_palette_get_hadjustment) IA__gtk_tool_palette_get_hadjustment __attribute((visibility("hidden")));
#define gtk_tool_palette_get_hadjustment IA__gtk_tool_palette_get_hadjustment

extern __typeof (gtk_tool_palette_get_vadjustment) IA__gtk_tool_palette_get_vadjustment __attribute((visibility("hidden")));
#define gtk_tool_palette_get_vadjustment IA__gtk_tool_palette_get_vadjustment

extern __typeof (gtk_tool_palette_get_drag_target_item) IA__gtk_tool_palette_get_drag_target_item __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tool_palette_get_drag_target_item IA__gtk_tool_palette_get_drag_target_item

extern __typeof (gtk_tool_palette_get_drag_target_group) IA__gtk_tool_palette_get_drag_target_group __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tool_palette_get_drag_target_group IA__gtk_tool_palette_get_drag_target_group

#endif
#endif
#if IN_HEADER(__GTK_TOOLTIP_H__)
#if IN_FILE(__GTK_TOOLTIP_C__)
extern __typeof (gtk_tooltip_get_type) IA__gtk_tooltip_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tooltip_get_type IA__gtk_tooltip_get_type

extern __typeof (gtk_tooltip_set_custom) IA__gtk_tooltip_set_custom __attribute((visibility("hidden")));
#define gtk_tooltip_set_custom IA__gtk_tooltip_set_custom

extern __typeof (gtk_tooltip_set_icon) IA__gtk_tooltip_set_icon __attribute((visibility("hidden")));
#define gtk_tooltip_set_icon IA__gtk_tooltip_set_icon

extern __typeof (gtk_tooltip_set_icon_from_stock) IA__gtk_tooltip_set_icon_from_stock __attribute((visibility("hidden")));
#define gtk_tooltip_set_icon_from_stock IA__gtk_tooltip_set_icon_from_stock

extern __typeof (gtk_tooltip_set_icon_from_icon_name) IA__gtk_tooltip_set_icon_from_icon_name __attribute((visibility("hidden")));
#define gtk_tooltip_set_icon_from_icon_name IA__gtk_tooltip_set_icon_from_icon_name

extern __typeof (gtk_tooltip_set_icon_from_gicon) IA__gtk_tooltip_set_icon_from_gicon __attribute((visibility("hidden")));
#define gtk_tooltip_set_icon_from_gicon IA__gtk_tooltip_set_icon_from_gicon

extern __typeof (gtk_tooltip_set_markup) IA__gtk_tooltip_set_markup __attribute((visibility("hidden")));
#define gtk_tooltip_set_markup IA__gtk_tooltip_set_markup

extern __typeof (gtk_tooltip_set_text) IA__gtk_tooltip_set_text __attribute((visibility("hidden")));
#define gtk_tooltip_set_text IA__gtk_tooltip_set_text

extern __typeof (gtk_tooltip_set_tip_area) IA__gtk_tooltip_set_tip_area __attribute((visibility("hidden")));
#define gtk_tooltip_set_tip_area IA__gtk_tooltip_set_tip_area

extern __typeof (gtk_tooltip_trigger_tooltip_query) IA__gtk_tooltip_trigger_tooltip_query __attribute((visibility("hidden")));
#define gtk_tooltip_trigger_tooltip_query IA__gtk_tooltip_trigger_tooltip_query

#endif
#endif
#if IN_HEADER(__GTK_TOOLTIPS_H__)
#if IN_FILE(__GTK_TOOLTIPS_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_tooltips_data_get) IA__gtk_tooltips_data_get __attribute((visibility("hidden")));
#define gtk_tooltips_data_get IA__gtk_tooltips_data_get

extern __typeof (gtk_tooltips_disable) IA__gtk_tooltips_disable __attribute((visibility("hidden")));
#define gtk_tooltips_disable IA__gtk_tooltips_disable

extern __typeof (gtk_tooltips_enable) IA__gtk_tooltips_enable __attribute((visibility("hidden")));
#define gtk_tooltips_enable IA__gtk_tooltips_enable

extern __typeof (gtk_tooltips_force_window) IA__gtk_tooltips_force_window __attribute((visibility("hidden")));
#define gtk_tooltips_force_window IA__gtk_tooltips_force_window

extern __typeof (gtk_tooltips_get_info_from_tip_window) IA__gtk_tooltips_get_info_from_tip_window __attribute((visibility("hidden")));
#define gtk_tooltips_get_info_from_tip_window IA__gtk_tooltips_get_info_from_tip_window

extern __typeof (gtk_tooltips_get_type) IA__gtk_tooltips_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tooltips_get_type IA__gtk_tooltips_get_type

extern __typeof (gtk_tooltips_new) IA__gtk_tooltips_new __attribute((visibility("hidden")));
#define gtk_tooltips_new IA__gtk_tooltips_new

extern __typeof (gtk_tooltips_set_delay) IA__gtk_tooltips_set_delay __attribute((visibility("hidden")));
#define gtk_tooltips_set_delay IA__gtk_tooltips_set_delay

extern __typeof (gtk_tooltips_set_tip) IA__gtk_tooltips_set_tip __attribute((visibility("hidden")));
#define gtk_tooltips_set_tip IA__gtk_tooltips_set_tip

#endif
#endif
#endif
#if IN_HEADER(__GTK_TRAY_ICON_H__)
#if IN_FILE(__GTK_TRAY_ICON_X11_C__)
#ifndef G_OS_WIN32
extern __typeof (gtk_tray_icon_get_type) IA__gtk_tray_icon_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tray_icon_get_type IA__gtk_tray_icon_get_type

#endif
#endif
#endif
#if IN_HEADER(__GTK_TREE_H__)
#if IN_FILE(__GTK_TREE_C__)
extern __typeof (gtk_tree_append) IA__gtk_tree_append __attribute((visibility("hidden")));
#define gtk_tree_append IA__gtk_tree_append

extern __typeof (gtk_tree_child_position) IA__gtk_tree_child_position __attribute((visibility("hidden")));
#define gtk_tree_child_position IA__gtk_tree_child_position

extern __typeof (gtk_tree_clear_items) IA__gtk_tree_clear_items __attribute((visibility("hidden")));
#define gtk_tree_clear_items IA__gtk_tree_clear_items

extern __typeof (gtk_tree_get_type) IA__gtk_tree_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_get_type IA__gtk_tree_get_type

extern __typeof (gtk_tree_insert) IA__gtk_tree_insert __attribute((visibility("hidden")));
#define gtk_tree_insert IA__gtk_tree_insert

extern __typeof (gtk_tree_prepend) IA__gtk_tree_prepend __attribute((visibility("hidden")));
#define gtk_tree_prepend IA__gtk_tree_prepend

extern __typeof (gtk_tree_remove_item) IA__gtk_tree_remove_item __attribute((visibility("hidden")));
#define gtk_tree_remove_item IA__gtk_tree_remove_item

extern __typeof (gtk_tree_remove_items) IA__gtk_tree_remove_items __attribute((visibility("hidden")));
#define gtk_tree_remove_items IA__gtk_tree_remove_items

extern __typeof (gtk_tree_select_child) IA__gtk_tree_select_child __attribute((visibility("hidden")));
#define gtk_tree_select_child IA__gtk_tree_select_child

extern __typeof (gtk_tree_select_item) IA__gtk_tree_select_item __attribute((visibility("hidden")));
#define gtk_tree_select_item IA__gtk_tree_select_item

extern __typeof (gtk_tree_set_selection_mode) IA__gtk_tree_set_selection_mode __attribute((visibility("hidden")));
#define gtk_tree_set_selection_mode IA__gtk_tree_set_selection_mode

extern __typeof (gtk_tree_set_view_lines) IA__gtk_tree_set_view_lines __attribute((visibility("hidden")));
#define gtk_tree_set_view_lines IA__gtk_tree_set_view_lines

extern __typeof (gtk_tree_set_view_mode) IA__gtk_tree_set_view_mode __attribute((visibility("hidden")));
#define gtk_tree_set_view_mode IA__gtk_tree_set_view_mode

extern __typeof (gtk_tree_unselect_child) IA__gtk_tree_unselect_child __attribute((visibility("hidden")));
#define gtk_tree_unselect_child IA__gtk_tree_unselect_child

extern __typeof (gtk_tree_unselect_item) IA__gtk_tree_unselect_item __attribute((visibility("hidden")));
#define gtk_tree_unselect_item IA__gtk_tree_unselect_item

extern __typeof (gtk_tree_new) IA__gtk_tree_new __attribute((visibility("hidden")));
#define gtk_tree_new IA__gtk_tree_new

#endif
#endif
#if IN_HEADER(__GTK_TREE_DND_H__)
#if IN_FILE(__GTK_TREE_DND_C__)
extern __typeof (gtk_tree_drag_dest_drag_data_received) IA__gtk_tree_drag_dest_drag_data_received __attribute((visibility("hidden")));
#define gtk_tree_drag_dest_drag_data_received IA__gtk_tree_drag_dest_drag_data_received

extern __typeof (gtk_tree_drag_dest_get_type) IA__gtk_tree_drag_dest_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_drag_dest_get_type IA__gtk_tree_drag_dest_get_type

extern __typeof (gtk_tree_drag_dest_row_drop_possible) IA__gtk_tree_drag_dest_row_drop_possible __attribute((visibility("hidden")));
#define gtk_tree_drag_dest_row_drop_possible IA__gtk_tree_drag_dest_row_drop_possible

extern __typeof (gtk_tree_drag_source_drag_data_delete) IA__gtk_tree_drag_source_drag_data_delete __attribute((visibility("hidden")));
#define gtk_tree_drag_source_drag_data_delete IA__gtk_tree_drag_source_drag_data_delete

extern __typeof (gtk_tree_drag_source_drag_data_get) IA__gtk_tree_drag_source_drag_data_get __attribute((visibility("hidden")));
#define gtk_tree_drag_source_drag_data_get IA__gtk_tree_drag_source_drag_data_get

extern __typeof (gtk_tree_drag_source_get_type) IA__gtk_tree_drag_source_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_drag_source_get_type IA__gtk_tree_drag_source_get_type

extern __typeof (gtk_tree_drag_source_row_draggable) IA__gtk_tree_drag_source_row_draggable __attribute((visibility("hidden")));
#define gtk_tree_drag_source_row_draggable IA__gtk_tree_drag_source_row_draggable

extern __typeof (gtk_tree_get_row_drag_data) IA__gtk_tree_get_row_drag_data __attribute((visibility("hidden")));
#define gtk_tree_get_row_drag_data IA__gtk_tree_get_row_drag_data

extern __typeof (gtk_tree_set_row_drag_data) IA__gtk_tree_set_row_drag_data __attribute((visibility("hidden")));
#define gtk_tree_set_row_drag_data IA__gtk_tree_set_row_drag_data

#endif
#endif
#if IN_HEADER(__GTK_TREE_ITEM_H__)
#if IN_FILE(__GTK_TREE_ITEM_C__)
extern __typeof (gtk_tree_item_collapse) IA__gtk_tree_item_collapse __attribute((visibility("hidden")));
#define gtk_tree_item_collapse IA__gtk_tree_item_collapse

extern __typeof (gtk_tree_item_deselect) IA__gtk_tree_item_deselect __attribute((visibility("hidden")));
#define gtk_tree_item_deselect IA__gtk_tree_item_deselect

extern __typeof (gtk_tree_item_expand) IA__gtk_tree_item_expand __attribute((visibility("hidden")));
#define gtk_tree_item_expand IA__gtk_tree_item_expand

extern __typeof (gtk_tree_item_get_type) IA__gtk_tree_item_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_item_get_type IA__gtk_tree_item_get_type

extern __typeof (gtk_tree_item_new) IA__gtk_tree_item_new __attribute((visibility("hidden")));
#define gtk_tree_item_new IA__gtk_tree_item_new

extern __typeof (gtk_tree_item_new_with_label) IA__gtk_tree_item_new_with_label __attribute((visibility("hidden")));
#define gtk_tree_item_new_with_label IA__gtk_tree_item_new_with_label

extern __typeof (gtk_tree_item_remove_subtree) IA__gtk_tree_item_remove_subtree __attribute((visibility("hidden")));
#define gtk_tree_item_remove_subtree IA__gtk_tree_item_remove_subtree

extern __typeof (gtk_tree_item_select) IA__gtk_tree_item_select __attribute((visibility("hidden")));
#define gtk_tree_item_select IA__gtk_tree_item_select

extern __typeof (gtk_tree_item_set_subtree) IA__gtk_tree_item_set_subtree __attribute((visibility("hidden")));
#define gtk_tree_item_set_subtree IA__gtk_tree_item_set_subtree

#endif
#endif
#if IN_HEADER(__GTK_TREE_MODEL_H__)
#if IN_FILE(__GTK_TREE_MODEL_C__)
extern __typeof (gtk_tree_iter_copy) IA__gtk_tree_iter_copy __attribute((visibility("hidden")));
#define gtk_tree_iter_copy IA__gtk_tree_iter_copy

extern __typeof (gtk_tree_iter_free) IA__gtk_tree_iter_free __attribute((visibility("hidden")));
#define gtk_tree_iter_free IA__gtk_tree_iter_free

extern __typeof (gtk_tree_iter_get_type) IA__gtk_tree_iter_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_iter_get_type IA__gtk_tree_iter_get_type

extern __typeof (gtk_tree_model_foreach) IA__gtk_tree_model_foreach __attribute((visibility("hidden")));
#define gtk_tree_model_foreach IA__gtk_tree_model_foreach

extern __typeof (gtk_tree_model_get) IA__gtk_tree_model_get __attribute((visibility("hidden")));
#define gtk_tree_model_get IA__gtk_tree_model_get

extern __typeof (gtk_tree_model_get_column_type) IA__gtk_tree_model_get_column_type __attribute((visibility("hidden")));
#define gtk_tree_model_get_column_type IA__gtk_tree_model_get_column_type

extern __typeof (gtk_tree_model_get_flags) IA__gtk_tree_model_get_flags __attribute((visibility("hidden")));
#define gtk_tree_model_get_flags IA__gtk_tree_model_get_flags

extern __typeof (gtk_tree_model_get_iter) IA__gtk_tree_model_get_iter __attribute((visibility("hidden")));
#define gtk_tree_model_get_iter IA__gtk_tree_model_get_iter

extern __typeof (gtk_tree_model_get_iter_first) IA__gtk_tree_model_get_iter_first __attribute((visibility("hidden")));
#define gtk_tree_model_get_iter_first IA__gtk_tree_model_get_iter_first

extern __typeof (gtk_tree_model_get_iter_from_string) IA__gtk_tree_model_get_iter_from_string __attribute((visibility("hidden")));
#define gtk_tree_model_get_iter_from_string IA__gtk_tree_model_get_iter_from_string

extern __typeof (gtk_tree_model_get_n_columns) IA__gtk_tree_model_get_n_columns __attribute((visibility("hidden")));
#define gtk_tree_model_get_n_columns IA__gtk_tree_model_get_n_columns

extern __typeof (gtk_tree_model_get_path) IA__gtk_tree_model_get_path __attribute((visibility("hidden")));
#define gtk_tree_model_get_path IA__gtk_tree_model_get_path

extern __typeof (gtk_tree_model_get_string_from_iter) IA__gtk_tree_model_get_string_from_iter __attribute((visibility("hidden")));
#define gtk_tree_model_get_string_from_iter IA__gtk_tree_model_get_string_from_iter

extern __typeof (gtk_tree_model_get_type) IA__gtk_tree_model_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_model_get_type IA__gtk_tree_model_get_type

extern __typeof (gtk_tree_model_get_valist) IA__gtk_tree_model_get_valist __attribute((visibility("hidden")));
#define gtk_tree_model_get_valist IA__gtk_tree_model_get_valist

extern __typeof (gtk_tree_model_get_value) IA__gtk_tree_model_get_value __attribute((visibility("hidden")));
#define gtk_tree_model_get_value IA__gtk_tree_model_get_value

extern __typeof (gtk_tree_model_iter_children) IA__gtk_tree_model_iter_children __attribute((visibility("hidden")));
#define gtk_tree_model_iter_children IA__gtk_tree_model_iter_children

extern __typeof (gtk_tree_model_iter_has_child) IA__gtk_tree_model_iter_has_child __attribute((visibility("hidden")));
#define gtk_tree_model_iter_has_child IA__gtk_tree_model_iter_has_child

extern __typeof (gtk_tree_model_iter_n_children) IA__gtk_tree_model_iter_n_children __attribute((visibility("hidden")));
#define gtk_tree_model_iter_n_children IA__gtk_tree_model_iter_n_children

extern __typeof (gtk_tree_model_iter_next) IA__gtk_tree_model_iter_next __attribute((visibility("hidden")));
#define gtk_tree_model_iter_next IA__gtk_tree_model_iter_next

extern __typeof (gtk_tree_model_iter_nth_child) IA__gtk_tree_model_iter_nth_child __attribute((visibility("hidden")));
#define gtk_tree_model_iter_nth_child IA__gtk_tree_model_iter_nth_child

extern __typeof (gtk_tree_model_iter_parent) IA__gtk_tree_model_iter_parent __attribute((visibility("hidden")));
#define gtk_tree_model_iter_parent IA__gtk_tree_model_iter_parent

extern __typeof (gtk_tree_model_ref_node) IA__gtk_tree_model_ref_node __attribute((visibility("hidden")));
#define gtk_tree_model_ref_node IA__gtk_tree_model_ref_node

extern __typeof (gtk_tree_model_row_changed) IA__gtk_tree_model_row_changed __attribute((visibility("hidden")));
#define gtk_tree_model_row_changed IA__gtk_tree_model_row_changed

extern __typeof (gtk_tree_model_row_deleted) IA__gtk_tree_model_row_deleted __attribute((visibility("hidden")));
#define gtk_tree_model_row_deleted IA__gtk_tree_model_row_deleted

extern __typeof (gtk_tree_model_row_has_child_toggled) IA__gtk_tree_model_row_has_child_toggled __attribute((visibility("hidden")));
#define gtk_tree_model_row_has_child_toggled IA__gtk_tree_model_row_has_child_toggled

extern __typeof (gtk_tree_model_row_inserted) IA__gtk_tree_model_row_inserted __attribute((visibility("hidden")));
#define gtk_tree_model_row_inserted IA__gtk_tree_model_row_inserted

extern __typeof (gtk_tree_model_rows_reordered) IA__gtk_tree_model_rows_reordered __attribute((visibility("hidden")));
#define gtk_tree_model_rows_reordered IA__gtk_tree_model_rows_reordered

extern __typeof (gtk_tree_model_unref_node) IA__gtk_tree_model_unref_node __attribute((visibility("hidden")));
#define gtk_tree_model_unref_node IA__gtk_tree_model_unref_node

extern __typeof (gtk_tree_path_append_index) IA__gtk_tree_path_append_index __attribute((visibility("hidden")));
#define gtk_tree_path_append_index IA__gtk_tree_path_append_index

extern __typeof (gtk_tree_path_compare) IA__gtk_tree_path_compare __attribute((visibility("hidden")));
#define gtk_tree_path_compare IA__gtk_tree_path_compare

extern __typeof (gtk_tree_path_copy) IA__gtk_tree_path_copy __attribute((visibility("hidden")));
#define gtk_tree_path_copy IA__gtk_tree_path_copy

extern __typeof (gtk_tree_path_down) IA__gtk_tree_path_down __attribute((visibility("hidden")));
#define gtk_tree_path_down IA__gtk_tree_path_down

extern __typeof (gtk_tree_path_free) IA__gtk_tree_path_free __attribute((visibility("hidden")));
#define gtk_tree_path_free IA__gtk_tree_path_free

extern __typeof (gtk_tree_path_get_depth) IA__gtk_tree_path_get_depth __attribute((visibility("hidden")));
#define gtk_tree_path_get_depth IA__gtk_tree_path_get_depth

extern __typeof (gtk_tree_path_get_indices) IA__gtk_tree_path_get_indices __attribute((visibility("hidden")));
#define gtk_tree_path_get_indices IA__gtk_tree_path_get_indices

extern __typeof (gtk_tree_path_get_indices_with_depth) IA__gtk_tree_path_get_indices_with_depth __attribute((visibility("hidden")));
#define gtk_tree_path_get_indices_with_depth IA__gtk_tree_path_get_indices_with_depth

extern __typeof (gtk_tree_path_get_type) IA__gtk_tree_path_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_path_get_type IA__gtk_tree_path_get_type

extern __typeof (gtk_tree_path_is_ancestor) IA__gtk_tree_path_is_ancestor __attribute((visibility("hidden")));
#define gtk_tree_path_is_ancestor IA__gtk_tree_path_is_ancestor

extern __typeof (gtk_tree_path_is_descendant) IA__gtk_tree_path_is_descendant __attribute((visibility("hidden")));
#define gtk_tree_path_is_descendant IA__gtk_tree_path_is_descendant

extern __typeof (gtk_tree_path_new) IA__gtk_tree_path_new __attribute((visibility("hidden")));
#define gtk_tree_path_new IA__gtk_tree_path_new

extern __typeof (gtk_tree_path_new_first) IA__gtk_tree_path_new_first __attribute((visibility("hidden")));
#define gtk_tree_path_new_first IA__gtk_tree_path_new_first

extern __typeof (gtk_tree_path_new_from_indices) IA__gtk_tree_path_new_from_indices __attribute((visibility("hidden")));
#define gtk_tree_path_new_from_indices IA__gtk_tree_path_new_from_indices

extern __typeof (gtk_tree_path_new_from_string) IA__gtk_tree_path_new_from_string __attribute((visibility("hidden")));
#define gtk_tree_path_new_from_string IA__gtk_tree_path_new_from_string

extern __typeof (gtk_tree_path_next) IA__gtk_tree_path_next __attribute((visibility("hidden")));
#define gtk_tree_path_next IA__gtk_tree_path_next

extern __typeof (gtk_tree_path_prepend_index) IA__gtk_tree_path_prepend_index __attribute((visibility("hidden")));
#define gtk_tree_path_prepend_index IA__gtk_tree_path_prepend_index

extern __typeof (gtk_tree_path_prev) IA__gtk_tree_path_prev __attribute((visibility("hidden")));
#define gtk_tree_path_prev IA__gtk_tree_path_prev

extern __typeof (gtk_tree_path_to_string) IA__gtk_tree_path_to_string __attribute((visibility("hidden")));
#define gtk_tree_path_to_string IA__gtk_tree_path_to_string

extern __typeof (gtk_tree_path_up) IA__gtk_tree_path_up __attribute((visibility("hidden")));
#define gtk_tree_path_up IA__gtk_tree_path_up

extern __typeof (gtk_tree_row_reference_copy) IA__gtk_tree_row_reference_copy __attribute((visibility("hidden")));
#define gtk_tree_row_reference_copy IA__gtk_tree_row_reference_copy

extern __typeof (gtk_tree_row_reference_deleted) IA__gtk_tree_row_reference_deleted __attribute((visibility("hidden")));
#define gtk_tree_row_reference_deleted IA__gtk_tree_row_reference_deleted

extern __typeof (gtk_tree_row_reference_free) IA__gtk_tree_row_reference_free __attribute((visibility("hidden")));
#define gtk_tree_row_reference_free IA__gtk_tree_row_reference_free

extern __typeof (gtk_tree_row_reference_get_model) IA__gtk_tree_row_reference_get_model __attribute((visibility("hidden")));
#define gtk_tree_row_reference_get_model IA__gtk_tree_row_reference_get_model

extern __typeof (gtk_tree_row_reference_get_path) IA__gtk_tree_row_reference_get_path __attribute((visibility("hidden")));
#define gtk_tree_row_reference_get_path IA__gtk_tree_row_reference_get_path

extern __typeof (gtk_tree_row_reference_get_type) IA__gtk_tree_row_reference_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_row_reference_get_type IA__gtk_tree_row_reference_get_type

extern __typeof (gtk_tree_row_reference_inserted) IA__gtk_tree_row_reference_inserted __attribute((visibility("hidden")));
#define gtk_tree_row_reference_inserted IA__gtk_tree_row_reference_inserted

extern __typeof (gtk_tree_row_reference_new) IA__gtk_tree_row_reference_new __attribute((visibility("hidden")));
#define gtk_tree_row_reference_new IA__gtk_tree_row_reference_new

extern __typeof (gtk_tree_row_reference_new_proxy) IA__gtk_tree_row_reference_new_proxy __attribute((visibility("hidden")));
#define gtk_tree_row_reference_new_proxy IA__gtk_tree_row_reference_new_proxy

extern __typeof (gtk_tree_row_reference_reordered) IA__gtk_tree_row_reference_reordered __attribute((visibility("hidden")));
#define gtk_tree_row_reference_reordered IA__gtk_tree_row_reference_reordered

extern __typeof (gtk_tree_row_reference_valid) IA__gtk_tree_row_reference_valid __attribute((visibility("hidden")));
#define gtk_tree_row_reference_valid IA__gtk_tree_row_reference_valid

#endif
#endif
#if IN_HEADER(__GTK_TREE_MODEL_SORT_H__)
#if IN_FILE(__GTK_TREE_MODEL_SORT_C__)
extern __typeof (gtk_tree_model_sort_clear_cache) IA__gtk_tree_model_sort_clear_cache __attribute((visibility("hidden")));
#define gtk_tree_model_sort_clear_cache IA__gtk_tree_model_sort_clear_cache

extern __typeof (gtk_tree_model_sort_convert_child_iter_to_iter) IA__gtk_tree_model_sort_convert_child_iter_to_iter __attribute((visibility("hidden")));
#define gtk_tree_model_sort_convert_child_iter_to_iter IA__gtk_tree_model_sort_convert_child_iter_to_iter

extern __typeof (gtk_tree_model_sort_convert_child_path_to_path) IA__gtk_tree_model_sort_convert_child_path_to_path __attribute((visibility("hidden")));
#define gtk_tree_model_sort_convert_child_path_to_path IA__gtk_tree_model_sort_convert_child_path_to_path

extern __typeof (gtk_tree_model_sort_convert_iter_to_child_iter) IA__gtk_tree_model_sort_convert_iter_to_child_iter __attribute((visibility("hidden")));
#define gtk_tree_model_sort_convert_iter_to_child_iter IA__gtk_tree_model_sort_convert_iter_to_child_iter

extern __typeof (gtk_tree_model_sort_convert_path_to_child_path) IA__gtk_tree_model_sort_convert_path_to_child_path __attribute((visibility("hidden")));
#define gtk_tree_model_sort_convert_path_to_child_path IA__gtk_tree_model_sort_convert_path_to_child_path

extern __typeof (gtk_tree_model_sort_get_model) IA__gtk_tree_model_sort_get_model __attribute((visibility("hidden")));
#define gtk_tree_model_sort_get_model IA__gtk_tree_model_sort_get_model

extern __typeof (gtk_tree_model_sort_get_type) IA__gtk_tree_model_sort_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_model_sort_get_type IA__gtk_tree_model_sort_get_type

extern __typeof (gtk_tree_model_sort_iter_is_valid) IA__gtk_tree_model_sort_iter_is_valid __attribute((visibility("hidden")));
#define gtk_tree_model_sort_iter_is_valid IA__gtk_tree_model_sort_iter_is_valid

extern __typeof (gtk_tree_model_sort_new_with_model) IA__gtk_tree_model_sort_new_with_model __attribute((visibility("hidden")));
#define gtk_tree_model_sort_new_with_model IA__gtk_tree_model_sort_new_with_model

extern __typeof (gtk_tree_model_sort_reset_default_sort_func) IA__gtk_tree_model_sort_reset_default_sort_func __attribute((visibility("hidden")));
#define gtk_tree_model_sort_reset_default_sort_func IA__gtk_tree_model_sort_reset_default_sort_func

#endif
#endif
#if IN_HEADER(__GTK_TREE_MODEL_FILTER_H__)
#if IN_FILE(__GTK_TREE_MODEL_FILTER_C__)
extern __typeof (gtk_tree_model_filter_clear_cache) IA__gtk_tree_model_filter_clear_cache __attribute((visibility("hidden")));
#define gtk_tree_model_filter_clear_cache IA__gtk_tree_model_filter_clear_cache

extern __typeof (gtk_tree_model_filter_convert_child_iter_to_iter) IA__gtk_tree_model_filter_convert_child_iter_to_iter __attribute((visibility("hidden")));
#define gtk_tree_model_filter_convert_child_iter_to_iter IA__gtk_tree_model_filter_convert_child_iter_to_iter

extern __typeof (gtk_tree_model_filter_convert_child_path_to_path) IA__gtk_tree_model_filter_convert_child_path_to_path __attribute((visibility("hidden")));
#define gtk_tree_model_filter_convert_child_path_to_path IA__gtk_tree_model_filter_convert_child_path_to_path

extern __typeof (gtk_tree_model_filter_convert_iter_to_child_iter) IA__gtk_tree_model_filter_convert_iter_to_child_iter __attribute((visibility("hidden")));
#define gtk_tree_model_filter_convert_iter_to_child_iter IA__gtk_tree_model_filter_convert_iter_to_child_iter

extern __typeof (gtk_tree_model_filter_convert_path_to_child_path) IA__gtk_tree_model_filter_convert_path_to_child_path __attribute((visibility("hidden")));
#define gtk_tree_model_filter_convert_path_to_child_path IA__gtk_tree_model_filter_convert_path_to_child_path

extern __typeof (gtk_tree_model_filter_get_model) IA__gtk_tree_model_filter_get_model __attribute((visibility("hidden")));
#define gtk_tree_model_filter_get_model IA__gtk_tree_model_filter_get_model

extern __typeof (gtk_tree_model_filter_get_type) IA__gtk_tree_model_filter_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_model_filter_get_type IA__gtk_tree_model_filter_get_type

extern __typeof (gtk_tree_model_filter_new) IA__gtk_tree_model_filter_new __attribute((visibility("hidden")));
#define gtk_tree_model_filter_new IA__gtk_tree_model_filter_new

extern __typeof (gtk_tree_model_filter_refilter) IA__gtk_tree_model_filter_refilter __attribute((visibility("hidden")));
#define gtk_tree_model_filter_refilter IA__gtk_tree_model_filter_refilter

extern __typeof (gtk_tree_model_filter_set_modify_func) IA__gtk_tree_model_filter_set_modify_func __attribute((visibility("hidden")));
#define gtk_tree_model_filter_set_modify_func IA__gtk_tree_model_filter_set_modify_func

extern __typeof (gtk_tree_model_filter_set_visible_column) IA__gtk_tree_model_filter_set_visible_column __attribute((visibility("hidden")));
#define gtk_tree_model_filter_set_visible_column IA__gtk_tree_model_filter_set_visible_column

extern __typeof (gtk_tree_model_filter_set_visible_func) IA__gtk_tree_model_filter_set_visible_func __attribute((visibility("hidden")));
#define gtk_tree_model_filter_set_visible_func IA__gtk_tree_model_filter_set_visible_func

#endif
#endif
#if IN_HEADER(__GTK_TREE_SELECTION_H__)
#if IN_FILE(__GTK_TREE_SELECTION_C__)
extern __typeof (gtk_tree_selection_count_selected_rows) IA__gtk_tree_selection_count_selected_rows __attribute((visibility("hidden")));
#define gtk_tree_selection_count_selected_rows IA__gtk_tree_selection_count_selected_rows

extern __typeof (gtk_tree_selection_get_mode) IA__gtk_tree_selection_get_mode __attribute((visibility("hidden")));
#define gtk_tree_selection_get_mode IA__gtk_tree_selection_get_mode

extern __typeof (gtk_tree_selection_get_select_function) IA__gtk_tree_selection_get_select_function __attribute((visibility("hidden")));
#define gtk_tree_selection_get_select_function IA__gtk_tree_selection_get_select_function

extern __typeof (gtk_tree_selection_get_selected) IA__gtk_tree_selection_get_selected __attribute((visibility("hidden")));
#define gtk_tree_selection_get_selected IA__gtk_tree_selection_get_selected

extern __typeof (gtk_tree_selection_get_selected_rows) IA__gtk_tree_selection_get_selected_rows __attribute((visibility("hidden")));
#define gtk_tree_selection_get_selected_rows IA__gtk_tree_selection_get_selected_rows

extern __typeof (gtk_tree_selection_get_tree_view) IA__gtk_tree_selection_get_tree_view __attribute((visibility("hidden")));
#define gtk_tree_selection_get_tree_view IA__gtk_tree_selection_get_tree_view

extern __typeof (gtk_tree_selection_get_type) IA__gtk_tree_selection_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_selection_get_type IA__gtk_tree_selection_get_type

extern __typeof (gtk_tree_selection_get_user_data) IA__gtk_tree_selection_get_user_data __attribute((visibility("hidden")));
#define gtk_tree_selection_get_user_data IA__gtk_tree_selection_get_user_data

extern __typeof (gtk_tree_selection_iter_is_selected) IA__gtk_tree_selection_iter_is_selected __attribute((visibility("hidden")));
#define gtk_tree_selection_iter_is_selected IA__gtk_tree_selection_iter_is_selected

extern __typeof (gtk_tree_selection_path_is_selected) IA__gtk_tree_selection_path_is_selected __attribute((visibility("hidden")));
#define gtk_tree_selection_path_is_selected IA__gtk_tree_selection_path_is_selected

extern __typeof (gtk_tree_selection_select_all) IA__gtk_tree_selection_select_all __attribute((visibility("hidden")));
#define gtk_tree_selection_select_all IA__gtk_tree_selection_select_all

extern __typeof (gtk_tree_selection_selected_foreach) IA__gtk_tree_selection_selected_foreach __attribute((visibility("hidden")));
#define gtk_tree_selection_selected_foreach IA__gtk_tree_selection_selected_foreach

extern __typeof (gtk_tree_selection_select_iter) IA__gtk_tree_selection_select_iter __attribute((visibility("hidden")));
#define gtk_tree_selection_select_iter IA__gtk_tree_selection_select_iter

extern __typeof (gtk_tree_selection_select_path) IA__gtk_tree_selection_select_path __attribute((visibility("hidden")));
#define gtk_tree_selection_select_path IA__gtk_tree_selection_select_path

extern __typeof (gtk_tree_selection_select_range) IA__gtk_tree_selection_select_range __attribute((visibility("hidden")));
#define gtk_tree_selection_select_range IA__gtk_tree_selection_select_range

extern __typeof (gtk_tree_selection_set_mode) IA__gtk_tree_selection_set_mode __attribute((visibility("hidden")));
#define gtk_tree_selection_set_mode IA__gtk_tree_selection_set_mode

extern __typeof (gtk_tree_selection_set_select_function) IA__gtk_tree_selection_set_select_function __attribute((visibility("hidden")));
#define gtk_tree_selection_set_select_function IA__gtk_tree_selection_set_select_function

extern __typeof (gtk_tree_selection_unselect_all) IA__gtk_tree_selection_unselect_all __attribute((visibility("hidden")));
#define gtk_tree_selection_unselect_all IA__gtk_tree_selection_unselect_all

extern __typeof (gtk_tree_selection_unselect_iter) IA__gtk_tree_selection_unselect_iter __attribute((visibility("hidden")));
#define gtk_tree_selection_unselect_iter IA__gtk_tree_selection_unselect_iter

extern __typeof (gtk_tree_selection_unselect_path) IA__gtk_tree_selection_unselect_path __attribute((visibility("hidden")));
#define gtk_tree_selection_unselect_path IA__gtk_tree_selection_unselect_path

extern __typeof (gtk_tree_selection_unselect_range) IA__gtk_tree_selection_unselect_range __attribute((visibility("hidden")));
#define gtk_tree_selection_unselect_range IA__gtk_tree_selection_unselect_range

#endif
#endif
#if IN_HEADER(__GTK_TREE_SORTABLE_H__)
#if IN_FILE(__GTK_TREE_SORTABLE_C__)
extern __typeof (gtk_tree_sortable_get_sort_column_id) IA__gtk_tree_sortable_get_sort_column_id __attribute((visibility("hidden")));
#define gtk_tree_sortable_get_sort_column_id IA__gtk_tree_sortable_get_sort_column_id

extern __typeof (gtk_tree_sortable_get_type) IA__gtk_tree_sortable_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_sortable_get_type IA__gtk_tree_sortable_get_type

extern __typeof (gtk_tree_sortable_has_default_sort_func) IA__gtk_tree_sortable_has_default_sort_func __attribute((visibility("hidden")));
#define gtk_tree_sortable_has_default_sort_func IA__gtk_tree_sortable_has_default_sort_func

extern __typeof (gtk_tree_sortable_set_default_sort_func) IA__gtk_tree_sortable_set_default_sort_func __attribute((visibility("hidden")));
#define gtk_tree_sortable_set_default_sort_func IA__gtk_tree_sortable_set_default_sort_func

extern __typeof (gtk_tree_sortable_set_sort_column_id) IA__gtk_tree_sortable_set_sort_column_id __attribute((visibility("hidden")));
#define gtk_tree_sortable_set_sort_column_id IA__gtk_tree_sortable_set_sort_column_id

extern __typeof (gtk_tree_sortable_set_sort_func) IA__gtk_tree_sortable_set_sort_func __attribute((visibility("hidden")));
#define gtk_tree_sortable_set_sort_func IA__gtk_tree_sortable_set_sort_func

extern __typeof (gtk_tree_sortable_sort_column_changed) IA__gtk_tree_sortable_sort_column_changed __attribute((visibility("hidden")));
#define gtk_tree_sortable_sort_column_changed IA__gtk_tree_sortable_sort_column_changed

#endif
#endif
#if IN_HEADER(__GTK_TREE_STORE_H__)
#if IN_FILE(__GTK_TREE_STORE_C__)
extern __typeof (gtk_tree_store_append) IA__gtk_tree_store_append __attribute((visibility("hidden")));
#define gtk_tree_store_append IA__gtk_tree_store_append

extern __typeof (gtk_tree_store_clear) IA__gtk_tree_store_clear __attribute((visibility("hidden")));
#define gtk_tree_store_clear IA__gtk_tree_store_clear

extern __typeof (gtk_tree_store_get_type) IA__gtk_tree_store_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_store_get_type IA__gtk_tree_store_get_type

extern __typeof (gtk_tree_store_insert) IA__gtk_tree_store_insert __attribute((visibility("hidden")));
#define gtk_tree_store_insert IA__gtk_tree_store_insert

extern __typeof (gtk_tree_store_insert_after) IA__gtk_tree_store_insert_after __attribute((visibility("hidden")));
#define gtk_tree_store_insert_after IA__gtk_tree_store_insert_after

extern __typeof (gtk_tree_store_insert_before) IA__gtk_tree_store_insert_before __attribute((visibility("hidden")));
#define gtk_tree_store_insert_before IA__gtk_tree_store_insert_before

extern __typeof (gtk_tree_store_insert_with_values) IA__gtk_tree_store_insert_with_values __attribute((visibility("hidden")));
#define gtk_tree_store_insert_with_values IA__gtk_tree_store_insert_with_values

extern __typeof (gtk_tree_store_insert_with_valuesv) IA__gtk_tree_store_insert_with_valuesv __attribute((visibility("hidden")));
#define gtk_tree_store_insert_with_valuesv IA__gtk_tree_store_insert_with_valuesv

extern __typeof (gtk_tree_store_is_ancestor) IA__gtk_tree_store_is_ancestor __attribute((visibility("hidden")));
#define gtk_tree_store_is_ancestor IA__gtk_tree_store_is_ancestor

extern __typeof (gtk_tree_store_iter_depth) IA__gtk_tree_store_iter_depth __attribute((visibility("hidden")));
#define gtk_tree_store_iter_depth IA__gtk_tree_store_iter_depth

extern __typeof (gtk_tree_store_iter_is_valid) IA__gtk_tree_store_iter_is_valid __attribute((visibility("hidden")));
#define gtk_tree_store_iter_is_valid IA__gtk_tree_store_iter_is_valid

extern __typeof (gtk_tree_store_move_after) IA__gtk_tree_store_move_after __attribute((visibility("hidden")));
#define gtk_tree_store_move_after IA__gtk_tree_store_move_after

extern __typeof (gtk_tree_store_move_before) IA__gtk_tree_store_move_before __attribute((visibility("hidden")));
#define gtk_tree_store_move_before IA__gtk_tree_store_move_before

extern __typeof (gtk_tree_store_new) IA__gtk_tree_store_new __attribute((visibility("hidden")));
#define gtk_tree_store_new IA__gtk_tree_store_new

extern __typeof (gtk_tree_store_newv) IA__gtk_tree_store_newv __attribute((visibility("hidden")));
#define gtk_tree_store_newv IA__gtk_tree_store_newv

extern __typeof (gtk_tree_store_prepend) IA__gtk_tree_store_prepend __attribute((visibility("hidden")));
#define gtk_tree_store_prepend IA__gtk_tree_store_prepend

extern __typeof (gtk_tree_store_remove) IA__gtk_tree_store_remove __attribute((visibility("hidden")));
#define gtk_tree_store_remove IA__gtk_tree_store_remove

extern __typeof (gtk_tree_store_reorder) IA__gtk_tree_store_reorder __attribute((visibility("hidden")));
#define gtk_tree_store_reorder IA__gtk_tree_store_reorder

extern __typeof (gtk_tree_store_set) IA__gtk_tree_store_set __attribute((visibility("hidden")));
#define gtk_tree_store_set IA__gtk_tree_store_set

extern __typeof (gtk_tree_store_set_column_types) IA__gtk_tree_store_set_column_types __attribute((visibility("hidden")));
#define gtk_tree_store_set_column_types IA__gtk_tree_store_set_column_types

extern __typeof (gtk_tree_store_set_valist) IA__gtk_tree_store_set_valist __attribute((visibility("hidden")));
#define gtk_tree_store_set_valist IA__gtk_tree_store_set_valist

extern __typeof (gtk_tree_store_set_value) IA__gtk_tree_store_set_value __attribute((visibility("hidden")));
#define gtk_tree_store_set_value IA__gtk_tree_store_set_value

extern __typeof (gtk_tree_store_set_valuesv) IA__gtk_tree_store_set_valuesv __attribute((visibility("hidden")));
#define gtk_tree_store_set_valuesv IA__gtk_tree_store_set_valuesv

extern __typeof (gtk_tree_store_swap) IA__gtk_tree_store_swap __attribute((visibility("hidden")));
#define gtk_tree_store_swap IA__gtk_tree_store_swap

#endif
#endif
#if IN_HEADER(__GTK_TREE_VIEW_COLUMN_H__)
#if IN_FILE(__GTK_TREE_VIEW_COLUMN_C__)
extern __typeof (gtk_tree_view_column_add_attribute) IA__gtk_tree_view_column_add_attribute __attribute((visibility("hidden")));
#define gtk_tree_view_column_add_attribute IA__gtk_tree_view_column_add_attribute

extern __typeof (gtk_tree_view_column_cell_get_position) IA__gtk_tree_view_column_cell_get_position __attribute((visibility("hidden")));
#define gtk_tree_view_column_cell_get_position IA__gtk_tree_view_column_cell_get_position

extern __typeof (gtk_tree_view_column_cell_get_size) IA__gtk_tree_view_column_cell_get_size __attribute((visibility("hidden")));
#define gtk_tree_view_column_cell_get_size IA__gtk_tree_view_column_cell_get_size

extern __typeof (gtk_tree_view_column_cell_is_visible) IA__gtk_tree_view_column_cell_is_visible __attribute((visibility("hidden")));
#define gtk_tree_view_column_cell_is_visible IA__gtk_tree_view_column_cell_is_visible

extern __typeof (gtk_tree_view_column_cell_set_cell_data) IA__gtk_tree_view_column_cell_set_cell_data __attribute((visibility("hidden")));
#define gtk_tree_view_column_cell_set_cell_data IA__gtk_tree_view_column_cell_set_cell_data

extern __typeof (gtk_tree_view_column_clear) IA__gtk_tree_view_column_clear __attribute((visibility("hidden")));
#define gtk_tree_view_column_clear IA__gtk_tree_view_column_clear

extern __typeof (gtk_tree_view_column_clear_attributes) IA__gtk_tree_view_column_clear_attributes __attribute((visibility("hidden")));
#define gtk_tree_view_column_clear_attributes IA__gtk_tree_view_column_clear_attributes

extern __typeof (gtk_tree_view_column_clicked) IA__gtk_tree_view_column_clicked __attribute((visibility("hidden")));
#define gtk_tree_view_column_clicked IA__gtk_tree_view_column_clicked

extern __typeof (gtk_tree_view_column_focus_cell) IA__gtk_tree_view_column_focus_cell __attribute((visibility("hidden")));
#define gtk_tree_view_column_focus_cell IA__gtk_tree_view_column_focus_cell

extern __typeof (gtk_tree_view_column_get_alignment) IA__gtk_tree_view_column_get_alignment __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_alignment IA__gtk_tree_view_column_get_alignment

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_tree_view_column_get_cell_renderers) IA__gtk_tree_view_column_get_cell_renderers __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_cell_renderers IA__gtk_tree_view_column_get_cell_renderers

#endif
extern __typeof (gtk_tree_view_column_get_clickable) IA__gtk_tree_view_column_get_clickable __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_clickable IA__gtk_tree_view_column_get_clickable

extern __typeof (gtk_tree_view_column_get_expand) IA__gtk_tree_view_column_get_expand __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_expand IA__gtk_tree_view_column_get_expand

extern __typeof (gtk_tree_view_column_get_fixed_width) IA__gtk_tree_view_column_get_fixed_width __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_fixed_width IA__gtk_tree_view_column_get_fixed_width

extern __typeof (gtk_tree_view_column_get_max_width) IA__gtk_tree_view_column_get_max_width __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_max_width IA__gtk_tree_view_column_get_max_width

extern __typeof (gtk_tree_view_column_get_min_width) IA__gtk_tree_view_column_get_min_width __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_min_width IA__gtk_tree_view_column_get_min_width

extern __typeof (gtk_tree_view_column_get_reorderable) IA__gtk_tree_view_column_get_reorderable __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_reorderable IA__gtk_tree_view_column_get_reorderable

extern __typeof (gtk_tree_view_column_get_resizable) IA__gtk_tree_view_column_get_resizable __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_resizable IA__gtk_tree_view_column_get_resizable

extern __typeof (gtk_tree_view_column_get_sizing) IA__gtk_tree_view_column_get_sizing __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_sizing IA__gtk_tree_view_column_get_sizing

extern __typeof (gtk_tree_view_column_get_sort_column_id) IA__gtk_tree_view_column_get_sort_column_id __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_sort_column_id IA__gtk_tree_view_column_get_sort_column_id

extern __typeof (gtk_tree_view_column_get_sort_indicator) IA__gtk_tree_view_column_get_sort_indicator __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_sort_indicator IA__gtk_tree_view_column_get_sort_indicator

extern __typeof (gtk_tree_view_column_get_sort_order) IA__gtk_tree_view_column_get_sort_order __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_sort_order IA__gtk_tree_view_column_get_sort_order

extern __typeof (gtk_tree_view_column_get_spacing) IA__gtk_tree_view_column_get_spacing __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_spacing IA__gtk_tree_view_column_get_spacing

extern __typeof (gtk_tree_view_column_get_title) IA__gtk_tree_view_column_get_title __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_title IA__gtk_tree_view_column_get_title

extern __typeof (gtk_tree_view_column_get_tree_view) IA__gtk_tree_view_column_get_tree_view __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_tree_view IA__gtk_tree_view_column_get_tree_view

extern __typeof (gtk_tree_view_column_get_type) IA__gtk_tree_view_column_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_view_column_get_type IA__gtk_tree_view_column_get_type

extern __typeof (gtk_tree_view_column_get_visible) IA__gtk_tree_view_column_get_visible __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_visible IA__gtk_tree_view_column_get_visible

extern __typeof (gtk_tree_view_column_get_widget) IA__gtk_tree_view_column_get_widget __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_widget IA__gtk_tree_view_column_get_widget

extern __typeof (gtk_tree_view_column_get_width) IA__gtk_tree_view_column_get_width __attribute((visibility("hidden")));
#define gtk_tree_view_column_get_width IA__gtk_tree_view_column_get_width

extern __typeof (gtk_tree_view_column_new) IA__gtk_tree_view_column_new __attribute((visibility("hidden")));
#define gtk_tree_view_column_new IA__gtk_tree_view_column_new

extern __typeof (gtk_tree_view_column_new_with_attributes) IA__gtk_tree_view_column_new_with_attributes __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_tree_view_column_new_with_attributes IA__gtk_tree_view_column_new_with_attributes

extern __typeof (gtk_tree_view_column_pack_end) IA__gtk_tree_view_column_pack_end __attribute((visibility("hidden")));
#define gtk_tree_view_column_pack_end IA__gtk_tree_view_column_pack_end

extern __typeof (gtk_tree_view_column_queue_resize) IA__gtk_tree_view_column_queue_resize __attribute((visibility("hidden")));
#define gtk_tree_view_column_queue_resize IA__gtk_tree_view_column_queue_resize

extern __typeof (gtk_tree_view_column_pack_start) IA__gtk_tree_view_column_pack_start __attribute((visibility("hidden")));
#define gtk_tree_view_column_pack_start IA__gtk_tree_view_column_pack_start

extern __typeof (gtk_tree_view_column_set_alignment) IA__gtk_tree_view_column_set_alignment __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_alignment IA__gtk_tree_view_column_set_alignment

extern __typeof (gtk_tree_view_column_set_attributes) IA__gtk_tree_view_column_set_attributes __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_attributes IA__gtk_tree_view_column_set_attributes

extern __typeof (gtk_tree_view_column_set_cell_data_func) IA__gtk_tree_view_column_set_cell_data_func __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_cell_data_func IA__gtk_tree_view_column_set_cell_data_func

extern __typeof (gtk_tree_view_column_set_clickable) IA__gtk_tree_view_column_set_clickable __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_clickable IA__gtk_tree_view_column_set_clickable

extern __typeof (gtk_tree_view_column_set_expand) IA__gtk_tree_view_column_set_expand __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_expand IA__gtk_tree_view_column_set_expand

extern __typeof (gtk_tree_view_column_set_fixed_width) IA__gtk_tree_view_column_set_fixed_width __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_fixed_width IA__gtk_tree_view_column_set_fixed_width

extern __typeof (gtk_tree_view_column_set_max_width) IA__gtk_tree_view_column_set_max_width __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_max_width IA__gtk_tree_view_column_set_max_width

extern __typeof (gtk_tree_view_column_set_min_width) IA__gtk_tree_view_column_set_min_width __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_min_width IA__gtk_tree_view_column_set_min_width

extern __typeof (gtk_tree_view_column_set_reorderable) IA__gtk_tree_view_column_set_reorderable __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_reorderable IA__gtk_tree_view_column_set_reorderable

extern __typeof (gtk_tree_view_column_set_resizable) IA__gtk_tree_view_column_set_resizable __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_resizable IA__gtk_tree_view_column_set_resizable

extern __typeof (gtk_tree_view_column_set_sizing) IA__gtk_tree_view_column_set_sizing __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_sizing IA__gtk_tree_view_column_set_sizing

extern __typeof (gtk_tree_view_column_set_sort_column_id) IA__gtk_tree_view_column_set_sort_column_id __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_sort_column_id IA__gtk_tree_view_column_set_sort_column_id

extern __typeof (gtk_tree_view_column_set_sort_indicator) IA__gtk_tree_view_column_set_sort_indicator __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_sort_indicator IA__gtk_tree_view_column_set_sort_indicator

extern __typeof (gtk_tree_view_column_set_sort_order) IA__gtk_tree_view_column_set_sort_order __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_sort_order IA__gtk_tree_view_column_set_sort_order

extern __typeof (gtk_tree_view_column_set_spacing) IA__gtk_tree_view_column_set_spacing __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_spacing IA__gtk_tree_view_column_set_spacing

extern __typeof (gtk_tree_view_column_set_title) IA__gtk_tree_view_column_set_title __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_title IA__gtk_tree_view_column_set_title

extern __typeof (gtk_tree_view_column_set_visible) IA__gtk_tree_view_column_set_visible __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_visible IA__gtk_tree_view_column_set_visible

extern __typeof (gtk_tree_view_column_set_widget) IA__gtk_tree_view_column_set_widget __attribute((visibility("hidden")));
#define gtk_tree_view_column_set_widget IA__gtk_tree_view_column_set_widget

#endif
#endif
#if IN_HEADER(__GTK_TREE_VIEW_H__)
#if IN_FILE(__GTK_TREE_VIEW_C__)
extern __typeof (gtk_tree_view_append_column) IA__gtk_tree_view_append_column __attribute((visibility("hidden")));
#define gtk_tree_view_append_column IA__gtk_tree_view_append_column

extern __typeof (gtk_tree_view_collapse_all) IA__gtk_tree_view_collapse_all __attribute((visibility("hidden")));
#define gtk_tree_view_collapse_all IA__gtk_tree_view_collapse_all

extern __typeof (gtk_tree_view_collapse_row) IA__gtk_tree_view_collapse_row __attribute((visibility("hidden")));
#define gtk_tree_view_collapse_row IA__gtk_tree_view_collapse_row

extern __typeof (gtk_tree_view_convert_bin_window_to_tree_coords) IA__gtk_tree_view_convert_bin_window_to_tree_coords __attribute((visibility("hidden")));
#define gtk_tree_view_convert_bin_window_to_tree_coords IA__gtk_tree_view_convert_bin_window_to_tree_coords

extern __typeof (gtk_tree_view_convert_bin_window_to_widget_coords) IA__gtk_tree_view_convert_bin_window_to_widget_coords __attribute((visibility("hidden")));
#define gtk_tree_view_convert_bin_window_to_widget_coords IA__gtk_tree_view_convert_bin_window_to_widget_coords

extern __typeof (gtk_tree_view_convert_tree_to_bin_window_coords) IA__gtk_tree_view_convert_tree_to_bin_window_coords __attribute((visibility("hidden")));
#define gtk_tree_view_convert_tree_to_bin_window_coords IA__gtk_tree_view_convert_tree_to_bin_window_coords

extern __typeof (gtk_tree_view_convert_tree_to_widget_coords) IA__gtk_tree_view_convert_tree_to_widget_coords __attribute((visibility("hidden")));
#define gtk_tree_view_convert_tree_to_widget_coords IA__gtk_tree_view_convert_tree_to_widget_coords

extern __typeof (gtk_tree_view_convert_widget_to_bin_window_coords) IA__gtk_tree_view_convert_widget_to_bin_window_coords __attribute((visibility("hidden")));
#define gtk_tree_view_convert_widget_to_bin_window_coords IA__gtk_tree_view_convert_widget_to_bin_window_coords

extern __typeof (gtk_tree_view_convert_widget_to_tree_coords) IA__gtk_tree_view_convert_widget_to_tree_coords __attribute((visibility("hidden")));
#define gtk_tree_view_convert_widget_to_tree_coords IA__gtk_tree_view_convert_widget_to_tree_coords

extern __typeof (gtk_tree_view_create_row_drag_icon) IA__gtk_tree_view_create_row_drag_icon __attribute((visibility("hidden")));
#define gtk_tree_view_create_row_drag_icon IA__gtk_tree_view_create_row_drag_icon

extern __typeof (gtk_tree_view_enable_model_drag_dest) IA__gtk_tree_view_enable_model_drag_dest __attribute((visibility("hidden")));
#define gtk_tree_view_enable_model_drag_dest IA__gtk_tree_view_enable_model_drag_dest

extern __typeof (gtk_tree_view_enable_model_drag_source) IA__gtk_tree_view_enable_model_drag_source __attribute((visibility("hidden")));
#define gtk_tree_view_enable_model_drag_source IA__gtk_tree_view_enable_model_drag_source

extern __typeof (gtk_tree_view_expand_all) IA__gtk_tree_view_expand_all __attribute((visibility("hidden")));
#define gtk_tree_view_expand_all IA__gtk_tree_view_expand_all

extern __typeof (gtk_tree_view_expand_row) IA__gtk_tree_view_expand_row __attribute((visibility("hidden")));
#define gtk_tree_view_expand_row IA__gtk_tree_view_expand_row

extern __typeof (gtk_tree_view_expand_to_path) IA__gtk_tree_view_expand_to_path __attribute((visibility("hidden")));
#define gtk_tree_view_expand_to_path IA__gtk_tree_view_expand_to_path

extern __typeof (gtk_tree_view_get_background_area) IA__gtk_tree_view_get_background_area __attribute((visibility("hidden")));
#define gtk_tree_view_get_background_area IA__gtk_tree_view_get_background_area

extern __typeof (gtk_tree_view_get_bin_window) IA__gtk_tree_view_get_bin_window __attribute((visibility("hidden")));
#define gtk_tree_view_get_bin_window IA__gtk_tree_view_get_bin_window

extern __typeof (gtk_tree_view_get_cell_area) IA__gtk_tree_view_get_cell_area __attribute((visibility("hidden")));
#define gtk_tree_view_get_cell_area IA__gtk_tree_view_get_cell_area

extern __typeof (gtk_tree_view_get_column) IA__gtk_tree_view_get_column __attribute((visibility("hidden")));
#define gtk_tree_view_get_column IA__gtk_tree_view_get_column

extern __typeof (gtk_tree_view_get_columns) IA__gtk_tree_view_get_columns __attribute((visibility("hidden")));
#define gtk_tree_view_get_columns IA__gtk_tree_view_get_columns

extern __typeof (gtk_tree_view_get_cursor) IA__gtk_tree_view_get_cursor __attribute((visibility("hidden")));
#define gtk_tree_view_get_cursor IA__gtk_tree_view_get_cursor

extern __typeof (gtk_tree_view_get_dest_row_at_pos) IA__gtk_tree_view_get_dest_row_at_pos __attribute((visibility("hidden")));
#define gtk_tree_view_get_dest_row_at_pos IA__gtk_tree_view_get_dest_row_at_pos

extern __typeof (gtk_tree_view_get_drag_dest_row) IA__gtk_tree_view_get_drag_dest_row __attribute((visibility("hidden")));
#define gtk_tree_view_get_drag_dest_row IA__gtk_tree_view_get_drag_dest_row

extern __typeof (gtk_tree_view_get_enable_search) IA__gtk_tree_view_get_enable_search __attribute((visibility("hidden")));
#define gtk_tree_view_get_enable_search IA__gtk_tree_view_get_enable_search

extern __typeof (gtk_tree_view_get_enable_tree_lines) IA__gtk_tree_view_get_enable_tree_lines __attribute((visibility("hidden")));
#define gtk_tree_view_get_enable_tree_lines IA__gtk_tree_view_get_enable_tree_lines

extern __typeof (gtk_tree_view_get_expander_column) IA__gtk_tree_view_get_expander_column __attribute((visibility("hidden")));
#define gtk_tree_view_get_expander_column IA__gtk_tree_view_get_expander_column

extern __typeof (gtk_tree_view_get_fixed_height_mode) IA__gtk_tree_view_get_fixed_height_mode __attribute((visibility("hidden")));
#define gtk_tree_view_get_fixed_height_mode IA__gtk_tree_view_get_fixed_height_mode

extern __typeof (gtk_tree_view_get_grid_lines) IA__gtk_tree_view_get_grid_lines __attribute((visibility("hidden")));
#define gtk_tree_view_get_grid_lines IA__gtk_tree_view_get_grid_lines

extern __typeof (gtk_tree_view_get_hadjustment) IA__gtk_tree_view_get_hadjustment __attribute((visibility("hidden")));
#define gtk_tree_view_get_hadjustment IA__gtk_tree_view_get_hadjustment

extern __typeof (gtk_tree_view_get_headers_clickable) IA__gtk_tree_view_get_headers_clickable __attribute((visibility("hidden")));
#define gtk_tree_view_get_headers_clickable IA__gtk_tree_view_get_headers_clickable

extern __typeof (gtk_tree_view_get_headers_visible) IA__gtk_tree_view_get_headers_visible __attribute((visibility("hidden")));
#define gtk_tree_view_get_headers_visible IA__gtk_tree_view_get_headers_visible

extern __typeof (gtk_tree_view_get_hover_expand) IA__gtk_tree_view_get_hover_expand __attribute((visibility("hidden")));
#define gtk_tree_view_get_hover_expand IA__gtk_tree_view_get_hover_expand

extern __typeof (gtk_tree_view_get_hover_selection) IA__gtk_tree_view_get_hover_selection __attribute((visibility("hidden")));
#define gtk_tree_view_get_hover_selection IA__gtk_tree_view_get_hover_selection

extern __typeof (gtk_tree_view_get_level_indentation) IA__gtk_tree_view_get_level_indentation __attribute((visibility("hidden")));
#define gtk_tree_view_get_level_indentation IA__gtk_tree_view_get_level_indentation

extern __typeof (gtk_tree_view_get_model) IA__gtk_tree_view_get_model __attribute((visibility("hidden")));
#define gtk_tree_view_get_model IA__gtk_tree_view_get_model

extern __typeof (gtk_tree_view_get_path_at_pos) IA__gtk_tree_view_get_path_at_pos __attribute((visibility("hidden")));
#define gtk_tree_view_get_path_at_pos IA__gtk_tree_view_get_path_at_pos

extern __typeof (gtk_tree_view_get_reorderable) IA__gtk_tree_view_get_reorderable __attribute((visibility("hidden")));
#define gtk_tree_view_get_reorderable IA__gtk_tree_view_get_reorderable

extern __typeof (gtk_tree_view_get_row_separator_func) IA__gtk_tree_view_get_row_separator_func __attribute((visibility("hidden")));
#define gtk_tree_view_get_row_separator_func IA__gtk_tree_view_get_row_separator_func

extern __typeof (gtk_tree_view_get_rubber_banding) IA__gtk_tree_view_get_rubber_banding __attribute((visibility("hidden")));
#define gtk_tree_view_get_rubber_banding IA__gtk_tree_view_get_rubber_banding

extern __typeof (gtk_tree_view_get_rules_hint) IA__gtk_tree_view_get_rules_hint __attribute((visibility("hidden")));
#define gtk_tree_view_get_rules_hint IA__gtk_tree_view_get_rules_hint

extern __typeof (gtk_tree_view_get_search_column) IA__gtk_tree_view_get_search_column __attribute((visibility("hidden")));
#define gtk_tree_view_get_search_column IA__gtk_tree_view_get_search_column

extern __typeof (gtk_tree_view_get_search_entry) IA__gtk_tree_view_get_search_entry __attribute((visibility("hidden")));
#define gtk_tree_view_get_search_entry IA__gtk_tree_view_get_search_entry

extern __typeof (gtk_tree_view_get_search_equal_func) IA__gtk_tree_view_get_search_equal_func __attribute((visibility("hidden")));
#define gtk_tree_view_get_search_equal_func IA__gtk_tree_view_get_search_equal_func

extern __typeof (gtk_tree_view_get_search_position_func) IA__gtk_tree_view_get_search_position_func __attribute((visibility("hidden")));
#define gtk_tree_view_get_search_position_func IA__gtk_tree_view_get_search_position_func

extern __typeof (gtk_tree_view_get_selection) IA__gtk_tree_view_get_selection __attribute((visibility("hidden")));
#define gtk_tree_view_get_selection IA__gtk_tree_view_get_selection

extern __typeof (gtk_tree_view_get_show_expanders) IA__gtk_tree_view_get_show_expanders __attribute((visibility("hidden")));
#define gtk_tree_view_get_show_expanders IA__gtk_tree_view_get_show_expanders

extern __typeof (gtk_tree_view_get_tooltip_column) IA__gtk_tree_view_get_tooltip_column __attribute((visibility("hidden")));
#define gtk_tree_view_get_tooltip_column IA__gtk_tree_view_get_tooltip_column

extern __typeof (gtk_tree_view_get_tooltip_context) IA__gtk_tree_view_get_tooltip_context __attribute((visibility("hidden")));
#define gtk_tree_view_get_tooltip_context IA__gtk_tree_view_get_tooltip_context

extern __typeof (gtk_tree_view_get_type) IA__gtk_tree_view_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_tree_view_get_type IA__gtk_tree_view_get_type

extern __typeof (gtk_tree_view_get_vadjustment) IA__gtk_tree_view_get_vadjustment __attribute((visibility("hidden")));
#define gtk_tree_view_get_vadjustment IA__gtk_tree_view_get_vadjustment

extern __typeof (gtk_tree_view_get_visible_range) IA__gtk_tree_view_get_visible_range __attribute((visibility("hidden")));
#define gtk_tree_view_get_visible_range IA__gtk_tree_view_get_visible_range

extern __typeof (gtk_tree_view_get_visible_rect) IA__gtk_tree_view_get_visible_rect __attribute((visibility("hidden")));
#define gtk_tree_view_get_visible_rect IA__gtk_tree_view_get_visible_rect

extern __typeof (gtk_tree_view_insert_column) IA__gtk_tree_view_insert_column __attribute((visibility("hidden")));
#define gtk_tree_view_insert_column IA__gtk_tree_view_insert_column

extern __typeof (gtk_tree_view_insert_column_with_attributes) IA__gtk_tree_view_insert_column_with_attributes __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_tree_view_insert_column_with_attributes IA__gtk_tree_view_insert_column_with_attributes

extern __typeof (gtk_tree_view_insert_column_with_data_func) IA__gtk_tree_view_insert_column_with_data_func __attribute((visibility("hidden")));
#define gtk_tree_view_insert_column_with_data_func IA__gtk_tree_view_insert_column_with_data_func

extern __typeof (gtk_tree_view_is_rubber_banding_active) IA__gtk_tree_view_is_rubber_banding_active __attribute((visibility("hidden")));
#define gtk_tree_view_is_rubber_banding_active IA__gtk_tree_view_is_rubber_banding_active

extern __typeof (gtk_tree_view_map_expanded_rows) IA__gtk_tree_view_map_expanded_rows __attribute((visibility("hidden")));
#define gtk_tree_view_map_expanded_rows IA__gtk_tree_view_map_expanded_rows

extern __typeof (gtk_tree_view_move_column_after) IA__gtk_tree_view_move_column_after __attribute((visibility("hidden")));
#define gtk_tree_view_move_column_after IA__gtk_tree_view_move_column_after

extern __typeof (gtk_tree_view_new) IA__gtk_tree_view_new __attribute((visibility("hidden")));
#define gtk_tree_view_new IA__gtk_tree_view_new

extern __typeof (gtk_tree_view_new_with_model) IA__gtk_tree_view_new_with_model __attribute((visibility("hidden")));
#define gtk_tree_view_new_with_model IA__gtk_tree_view_new_with_model

extern __typeof (gtk_tree_view_remove_column) IA__gtk_tree_view_remove_column __attribute((visibility("hidden")));
#define gtk_tree_view_remove_column IA__gtk_tree_view_remove_column

extern __typeof (gtk_tree_view_row_activated) IA__gtk_tree_view_row_activated __attribute((visibility("hidden")));
#define gtk_tree_view_row_activated IA__gtk_tree_view_row_activated

extern __typeof (gtk_tree_view_row_expanded) IA__gtk_tree_view_row_expanded __attribute((visibility("hidden")));
#define gtk_tree_view_row_expanded IA__gtk_tree_view_row_expanded

extern __typeof (gtk_tree_view_scroll_to_cell) IA__gtk_tree_view_scroll_to_cell __attribute((visibility("hidden")));
#define gtk_tree_view_scroll_to_cell IA__gtk_tree_view_scroll_to_cell

extern __typeof (gtk_tree_view_scroll_to_point) IA__gtk_tree_view_scroll_to_point __attribute((visibility("hidden")));
#define gtk_tree_view_scroll_to_point IA__gtk_tree_view_scroll_to_point

extern __typeof (gtk_tree_view_set_column_drag_function) IA__gtk_tree_view_set_column_drag_function __attribute((visibility("hidden")));
#define gtk_tree_view_set_column_drag_function IA__gtk_tree_view_set_column_drag_function

extern __typeof (gtk_tree_view_set_cursor) IA__gtk_tree_view_set_cursor __attribute((visibility("hidden")));
#define gtk_tree_view_set_cursor IA__gtk_tree_view_set_cursor

extern __typeof (gtk_tree_view_set_cursor_on_cell) IA__gtk_tree_view_set_cursor_on_cell __attribute((visibility("hidden")));
#define gtk_tree_view_set_cursor_on_cell IA__gtk_tree_view_set_cursor_on_cell

extern __typeof (gtk_tree_view_set_destroy_count_func) IA__gtk_tree_view_set_destroy_count_func __attribute((visibility("hidden")));
#define gtk_tree_view_set_destroy_count_func IA__gtk_tree_view_set_destroy_count_func

extern __typeof (gtk_tree_view_set_drag_dest_row) IA__gtk_tree_view_set_drag_dest_row __attribute((visibility("hidden")));
#define gtk_tree_view_set_drag_dest_row IA__gtk_tree_view_set_drag_dest_row

extern __typeof (gtk_tree_view_set_enable_search) IA__gtk_tree_view_set_enable_search __attribute((visibility("hidden")));
#define gtk_tree_view_set_enable_search IA__gtk_tree_view_set_enable_search

extern __typeof (gtk_tree_view_set_enable_tree_lines) IA__gtk_tree_view_set_enable_tree_lines __attribute((visibility("hidden")));
#define gtk_tree_view_set_enable_tree_lines IA__gtk_tree_view_set_enable_tree_lines

extern __typeof (gtk_tree_view_set_expander_column) IA__gtk_tree_view_set_expander_column __attribute((visibility("hidden")));
#define gtk_tree_view_set_expander_column IA__gtk_tree_view_set_expander_column

extern __typeof (gtk_tree_view_set_fixed_height_mode) IA__gtk_tree_view_set_fixed_height_mode __attribute((visibility("hidden")));
#define gtk_tree_view_set_fixed_height_mode IA__gtk_tree_view_set_fixed_height_mode

extern __typeof (gtk_tree_view_set_grid_lines) IA__gtk_tree_view_set_grid_lines __attribute((visibility("hidden")));
#define gtk_tree_view_set_grid_lines IA__gtk_tree_view_set_grid_lines

extern __typeof (gtk_tree_view_set_hadjustment) IA__gtk_tree_view_set_hadjustment __attribute((visibility("hidden")));
#define gtk_tree_view_set_hadjustment IA__gtk_tree_view_set_hadjustment

extern __typeof (gtk_tree_view_set_headers_clickable) IA__gtk_tree_view_set_headers_clickable __attribute((visibility("hidden")));
#define gtk_tree_view_set_headers_clickable IA__gtk_tree_view_set_headers_clickable

extern __typeof (gtk_tree_view_set_headers_visible) IA__gtk_tree_view_set_headers_visible __attribute((visibility("hidden")));
#define gtk_tree_view_set_headers_visible IA__gtk_tree_view_set_headers_visible

extern __typeof (gtk_tree_view_set_hover_expand) IA__gtk_tree_view_set_hover_expand __attribute((visibility("hidden")));
#define gtk_tree_view_set_hover_expand IA__gtk_tree_view_set_hover_expand

extern __typeof (gtk_tree_view_set_hover_selection) IA__gtk_tree_view_set_hover_selection __attribute((visibility("hidden")));
#define gtk_tree_view_set_hover_selection IA__gtk_tree_view_set_hover_selection

extern __typeof (gtk_tree_view_set_level_indentation) IA__gtk_tree_view_set_level_indentation __attribute((visibility("hidden")));
#define gtk_tree_view_set_level_indentation IA__gtk_tree_view_set_level_indentation

extern __typeof (gtk_tree_view_set_model) IA__gtk_tree_view_set_model __attribute((visibility("hidden")));
#define gtk_tree_view_set_model IA__gtk_tree_view_set_model

extern __typeof (gtk_tree_view_set_reorderable) IA__gtk_tree_view_set_reorderable __attribute((visibility("hidden")));
#define gtk_tree_view_set_reorderable IA__gtk_tree_view_set_reorderable

extern __typeof (gtk_tree_view_set_row_separator_func) IA__gtk_tree_view_set_row_separator_func __attribute((visibility("hidden")));
#define gtk_tree_view_set_row_separator_func IA__gtk_tree_view_set_row_separator_func

extern __typeof (gtk_tree_view_set_rubber_banding) IA__gtk_tree_view_set_rubber_banding __attribute((visibility("hidden")));
#define gtk_tree_view_set_rubber_banding IA__gtk_tree_view_set_rubber_banding

extern __typeof (gtk_tree_view_set_rules_hint) IA__gtk_tree_view_set_rules_hint __attribute((visibility("hidden")));
#define gtk_tree_view_set_rules_hint IA__gtk_tree_view_set_rules_hint

extern __typeof (gtk_tree_view_set_search_column) IA__gtk_tree_view_set_search_column __attribute((visibility("hidden")));
#define gtk_tree_view_set_search_column IA__gtk_tree_view_set_search_column

extern __typeof (gtk_tree_view_set_search_entry) IA__gtk_tree_view_set_search_entry __attribute((visibility("hidden")));
#define gtk_tree_view_set_search_entry IA__gtk_tree_view_set_search_entry

extern __typeof (gtk_tree_view_set_search_equal_func) IA__gtk_tree_view_set_search_equal_func __attribute((visibility("hidden")));
#define gtk_tree_view_set_search_equal_func IA__gtk_tree_view_set_search_equal_func

extern __typeof (gtk_tree_view_set_search_position_func) IA__gtk_tree_view_set_search_position_func __attribute((visibility("hidden")));
#define gtk_tree_view_set_search_position_func IA__gtk_tree_view_set_search_position_func

extern __typeof (gtk_tree_view_set_show_expanders) IA__gtk_tree_view_set_show_expanders __attribute((visibility("hidden")));
#define gtk_tree_view_set_show_expanders IA__gtk_tree_view_set_show_expanders

extern __typeof (gtk_tree_view_set_tooltip_row) IA__gtk_tree_view_set_tooltip_row __attribute((visibility("hidden")));
#define gtk_tree_view_set_tooltip_row IA__gtk_tree_view_set_tooltip_row

extern __typeof (gtk_tree_view_set_tooltip_cell) IA__gtk_tree_view_set_tooltip_cell __attribute((visibility("hidden")));
#define gtk_tree_view_set_tooltip_cell IA__gtk_tree_view_set_tooltip_cell

extern __typeof (gtk_tree_view_set_tooltip_column) IA__gtk_tree_view_set_tooltip_column __attribute((visibility("hidden")));
#define gtk_tree_view_set_tooltip_column IA__gtk_tree_view_set_tooltip_column

extern __typeof (gtk_tree_view_set_vadjustment) IA__gtk_tree_view_set_vadjustment __attribute((visibility("hidden")));
#define gtk_tree_view_set_vadjustment IA__gtk_tree_view_set_vadjustment

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_tree_view_tree_to_widget_coords) IA__gtk_tree_view_tree_to_widget_coords __attribute((visibility("hidden")));
#define gtk_tree_view_tree_to_widget_coords IA__gtk_tree_view_tree_to_widget_coords

#endif
extern __typeof (gtk_tree_view_unset_rows_drag_dest) IA__gtk_tree_view_unset_rows_drag_dest __attribute((visibility("hidden")));
#define gtk_tree_view_unset_rows_drag_dest IA__gtk_tree_view_unset_rows_drag_dest

extern __typeof (gtk_tree_view_unset_rows_drag_source) IA__gtk_tree_view_unset_rows_drag_source __attribute((visibility("hidden")));
#define gtk_tree_view_unset_rows_drag_source IA__gtk_tree_view_unset_rows_drag_source

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_tree_view_widget_to_tree_coords) IA__gtk_tree_view_widget_to_tree_coords __attribute((visibility("hidden")));
#define gtk_tree_view_widget_to_tree_coords IA__gtk_tree_view_widget_to_tree_coords

#endif
extern __typeof (gtk_tree_view_columns_autosize) IA__gtk_tree_view_columns_autosize __attribute((visibility("hidden")));
#define gtk_tree_view_columns_autosize IA__gtk_tree_view_columns_autosize

#endif
#endif
#if IN_HEADER(__GTK_TYPE_UTILS_H__)
#if IN_FILE(__GTK_TYPE_UTILS_C__)
extern __typeof (gtk_identifier_get_type) IA__gtk_identifier_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_identifier_get_type IA__gtk_identifier_get_type

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_type_enum_find_value) IA__gtk_type_enum_find_value __attribute((visibility("hidden")));
#define gtk_type_enum_find_value IA__gtk_type_enum_find_value

extern __typeof (gtk_type_enum_get_values) IA__gtk_type_enum_get_values __attribute((visibility("hidden")));
#define gtk_type_enum_get_values IA__gtk_type_enum_get_values

extern __typeof (gtk_type_flags_find_value) IA__gtk_type_flags_find_value __attribute((visibility("hidden")));
#define gtk_type_flags_find_value IA__gtk_type_flags_find_value

extern __typeof (gtk_type_flags_get_values) IA__gtk_type_flags_get_values __attribute((visibility("hidden")));
#define gtk_type_flags_get_values IA__gtk_type_flags_get_values

extern __typeof (gtk_type_init) IA__gtk_type_init __attribute((visibility("hidden")));
#define gtk_type_init IA__gtk_type_init

extern __typeof (gtk_type_new) IA__gtk_type_new __attribute((visibility("hidden")));
#define gtk_type_new IA__gtk_type_new

extern __typeof (gtk_type_unique) IA__gtk_type_unique __attribute((visibility("hidden")));
#define gtk_type_unique IA__gtk_type_unique

extern __typeof (gtk_type_class) IA__gtk_type_class __attribute((visibility("hidden")));
#define gtk_type_class IA__gtk_type_class

#endif
#endif
#endif
#if IN_HEADER(__GTK_UI_MANAGER_H__)
#if IN_FILE(__GTK_UI_MANAGER_C__)
extern __typeof (gtk_ui_manager_add_ui) IA__gtk_ui_manager_add_ui __attribute((visibility("hidden")));
#define gtk_ui_manager_add_ui IA__gtk_ui_manager_add_ui

#ifndef _WIN64
extern __typeof (gtk_ui_manager_add_ui_from_file) IA__gtk_ui_manager_add_ui_from_file __attribute((visibility("hidden")));
#define gtk_ui_manager_add_ui_from_file IA__gtk_ui_manager_add_ui_from_file

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_ui_manager_add_ui_from_file_utf8) IA__gtk_ui_manager_add_ui_from_file_utf8 __attribute((visibility("hidden")));
#define gtk_ui_manager_add_ui_from_file_utf8 IA__gtk_ui_manager_add_ui_from_file_utf8

#endif
extern __typeof (gtk_ui_manager_add_ui_from_string) IA__gtk_ui_manager_add_ui_from_string __attribute((visibility("hidden")));
#define gtk_ui_manager_add_ui_from_string IA__gtk_ui_manager_add_ui_from_string

extern __typeof (gtk_ui_manager_ensure_update) IA__gtk_ui_manager_ensure_update __attribute((visibility("hidden")));
#define gtk_ui_manager_ensure_update IA__gtk_ui_manager_ensure_update

extern __typeof (gtk_ui_manager_get_accel_group) IA__gtk_ui_manager_get_accel_group __attribute((visibility("hidden")));
#define gtk_ui_manager_get_accel_group IA__gtk_ui_manager_get_accel_group

extern __typeof (gtk_ui_manager_get_action) IA__gtk_ui_manager_get_action __attribute((visibility("hidden")));
#define gtk_ui_manager_get_action IA__gtk_ui_manager_get_action

extern __typeof (gtk_ui_manager_get_action_groups) IA__gtk_ui_manager_get_action_groups __attribute((visibility("hidden")));
#define gtk_ui_manager_get_action_groups IA__gtk_ui_manager_get_action_groups

extern __typeof (gtk_ui_manager_get_add_tearoffs) IA__gtk_ui_manager_get_add_tearoffs __attribute((visibility("hidden")));
#define gtk_ui_manager_get_add_tearoffs IA__gtk_ui_manager_get_add_tearoffs

extern __typeof (gtk_ui_manager_get_toplevels) IA__gtk_ui_manager_get_toplevels __attribute((visibility("hidden")));
#define gtk_ui_manager_get_toplevels IA__gtk_ui_manager_get_toplevels

extern __typeof (gtk_ui_manager_get_type) IA__gtk_ui_manager_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_ui_manager_get_type IA__gtk_ui_manager_get_type

extern __typeof (gtk_ui_manager_get_ui) IA__gtk_ui_manager_get_ui __attribute((visibility("hidden")));
#define gtk_ui_manager_get_ui IA__gtk_ui_manager_get_ui

extern __typeof (gtk_ui_manager_get_widget) IA__gtk_ui_manager_get_widget __attribute((visibility("hidden")));
#define gtk_ui_manager_get_widget IA__gtk_ui_manager_get_widget

extern __typeof (gtk_ui_manager_insert_action_group) IA__gtk_ui_manager_insert_action_group __attribute((visibility("hidden")));
#define gtk_ui_manager_insert_action_group IA__gtk_ui_manager_insert_action_group

extern __typeof (gtk_ui_manager_new) IA__gtk_ui_manager_new __attribute((visibility("hidden")));
#define gtk_ui_manager_new IA__gtk_ui_manager_new

extern __typeof (gtk_ui_manager_new_merge_id) IA__gtk_ui_manager_new_merge_id __attribute((visibility("hidden")));
#define gtk_ui_manager_new_merge_id IA__gtk_ui_manager_new_merge_id

extern __typeof (gtk_ui_manager_remove_action_group) IA__gtk_ui_manager_remove_action_group __attribute((visibility("hidden")));
#define gtk_ui_manager_remove_action_group IA__gtk_ui_manager_remove_action_group

extern __typeof (gtk_ui_manager_remove_ui) IA__gtk_ui_manager_remove_ui __attribute((visibility("hidden")));
#define gtk_ui_manager_remove_ui IA__gtk_ui_manager_remove_ui

extern __typeof (gtk_ui_manager_set_add_tearoffs) IA__gtk_ui_manager_set_add_tearoffs __attribute((visibility("hidden")));
#define gtk_ui_manager_set_add_tearoffs IA__gtk_ui_manager_set_add_tearoffs

#endif
#endif
#if IN_HEADER(__GTK_VBOX_H__)
#if IN_FILE(__GTK_VBOX_C__)
extern __typeof (gtk_vbox_get_type) IA__gtk_vbox_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_vbox_get_type IA__gtk_vbox_get_type

extern __typeof (gtk_vbox_new) IA__gtk_vbox_new __attribute((visibility("hidden")));
#define gtk_vbox_new IA__gtk_vbox_new

#endif
#endif
#if IN_HEADER(__GTK_VBBOX_H__)
#if IN_FILE(__GTK_VBBOX_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_vbutton_box_get_layout_default) IA__gtk_vbutton_box_get_layout_default __attribute((visibility("hidden")));
#define gtk_vbutton_box_get_layout_default IA__gtk_vbutton_box_get_layout_default

extern __typeof (gtk_vbutton_box_get_spacing_default) IA__gtk_vbutton_box_get_spacing_default __attribute((visibility("hidden")));
#define gtk_vbutton_box_get_spacing_default IA__gtk_vbutton_box_get_spacing_default

extern __typeof (gtk_vbutton_box_set_layout_default) IA__gtk_vbutton_box_set_layout_default __attribute((visibility("hidden")));
#define gtk_vbutton_box_set_layout_default IA__gtk_vbutton_box_set_layout_default

extern __typeof (gtk_vbutton_box_set_spacing_default) IA__gtk_vbutton_box_set_spacing_default __attribute((visibility("hidden")));
#define gtk_vbutton_box_set_spacing_default IA__gtk_vbutton_box_set_spacing_default

#endif
extern __typeof (gtk_vbutton_box_get_type) IA__gtk_vbutton_box_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_vbutton_box_get_type IA__gtk_vbutton_box_get_type

extern __typeof (gtk_vbutton_box_new) IA__gtk_vbutton_box_new __attribute((visibility("hidden")));
#define gtk_vbutton_box_new IA__gtk_vbutton_box_new

#endif
#endif
#if IN_HEADER(__GTK_VIEWPORT_H__)
#if IN_FILE(__GTK_VIEWPORT_C__)
extern __typeof (gtk_viewport_get_hadjustment) IA__gtk_viewport_get_hadjustment __attribute((visibility("hidden")));
#define gtk_viewport_get_hadjustment IA__gtk_viewport_get_hadjustment

extern __typeof (gtk_viewport_get_shadow_type) IA__gtk_viewport_get_shadow_type __attribute((visibility("hidden")));
#define gtk_viewport_get_shadow_type IA__gtk_viewport_get_shadow_type

extern __typeof (gtk_viewport_get_type) IA__gtk_viewport_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_viewport_get_type IA__gtk_viewport_get_type

extern __typeof (gtk_viewport_get_vadjustment) IA__gtk_viewport_get_vadjustment __attribute((visibility("hidden")));
#define gtk_viewport_get_vadjustment IA__gtk_viewport_get_vadjustment

extern __typeof (gtk_viewport_get_bin_window) IA__gtk_viewport_get_bin_window __attribute((visibility("hidden")));
#define gtk_viewport_get_bin_window IA__gtk_viewport_get_bin_window

extern __typeof (gtk_viewport_get_view_window) IA__gtk_viewport_get_view_window __attribute((visibility("hidden")));
#define gtk_viewport_get_view_window IA__gtk_viewport_get_view_window

extern __typeof (gtk_viewport_new) IA__gtk_viewport_new __attribute((visibility("hidden")));
#define gtk_viewport_new IA__gtk_viewport_new

extern __typeof (gtk_viewport_set_hadjustment) IA__gtk_viewport_set_hadjustment __attribute((visibility("hidden")));
#define gtk_viewport_set_hadjustment IA__gtk_viewport_set_hadjustment

extern __typeof (gtk_viewport_set_shadow_type) IA__gtk_viewport_set_shadow_type __attribute((visibility("hidden")));
#define gtk_viewport_set_shadow_type IA__gtk_viewport_set_shadow_type

extern __typeof (gtk_viewport_set_vadjustment) IA__gtk_viewport_set_vadjustment __attribute((visibility("hidden")));
#define gtk_viewport_set_vadjustment IA__gtk_viewport_set_vadjustment

#endif
#endif
#if IN_HEADER(__GTK_VOLUME_BUTTON_H__)
#if IN_FILE(__GTK_VOLUME_BUTTON_C__)
extern __typeof (gtk_volume_button_get_type) IA__gtk_volume_button_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_volume_button_get_type IA__gtk_volume_button_get_type

extern __typeof (gtk_volume_button_new) IA__gtk_volume_button_new __attribute((visibility("hidden")));
#define gtk_volume_button_new IA__gtk_volume_button_new

#endif
#endif
#if IN_HEADER(__GTK_VPANED_H__)
#if IN_FILE(__GTK_VPANED_C__)
extern __typeof (gtk_vpaned_get_type) IA__gtk_vpaned_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_vpaned_get_type IA__gtk_vpaned_get_type

extern __typeof (gtk_vpaned_new) IA__gtk_vpaned_new __attribute((visibility("hidden")));
#define gtk_vpaned_new IA__gtk_vpaned_new

#endif
#endif
#if IN_HEADER(__GTK_VRULER_H__)
#if IN_FILE(__GTK_VRULER_C__)
#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_vruler_get_type) IA__gtk_vruler_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_vruler_get_type IA__gtk_vruler_get_type

extern __typeof (gtk_vruler_new) IA__gtk_vruler_new __attribute((visibility("hidden")));
#define gtk_vruler_new IA__gtk_vruler_new

#endif
#endif
#endif
#if IN_HEADER(__GTK_VSCALE_H__)
#if IN_FILE(__GTK_VSCALE_C__)
extern __typeof (gtk_vscale_get_type) IA__gtk_vscale_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_vscale_get_type IA__gtk_vscale_get_type

extern __typeof (gtk_vscale_new) IA__gtk_vscale_new __attribute((visibility("hidden")));
#define gtk_vscale_new IA__gtk_vscale_new

extern __typeof (gtk_vscale_new_with_range) IA__gtk_vscale_new_with_range __attribute((visibility("hidden")));
#define gtk_vscale_new_with_range IA__gtk_vscale_new_with_range

#endif
#endif
#if IN_HEADER(__GTK_VSCROLLBAR_H__)
#if IN_FILE(__GTK_VSCROLLBAR_C__)
extern __typeof (gtk_vscrollbar_get_type) IA__gtk_vscrollbar_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_vscrollbar_get_type IA__gtk_vscrollbar_get_type

extern __typeof (gtk_vscrollbar_new) IA__gtk_vscrollbar_new __attribute((visibility("hidden")));
#define gtk_vscrollbar_new IA__gtk_vscrollbar_new

#endif
#endif
#if IN_HEADER(__GTK_VSEPARATOR_H__)
#if IN_FILE(__GTK_VSEPARATOR_C__)
extern __typeof (gtk_vseparator_get_type) IA__gtk_vseparator_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_vseparator_get_type IA__gtk_vseparator_get_type

extern __typeof (gtk_vseparator_new) IA__gtk_vseparator_new __attribute((visibility("hidden")));
#define gtk_vseparator_new IA__gtk_vseparator_new

#endif
#endif
#if IN_HEADER(__GTK_WIDGET_H__)
#if IN_FILE(__GTK_WIDGET_C__)
extern __typeof (gtk_requisition_copy) IA__gtk_requisition_copy __attribute((visibility("hidden")));
#define gtk_requisition_copy IA__gtk_requisition_copy

extern __typeof (gtk_requisition_free) IA__gtk_requisition_free __attribute((visibility("hidden")));
#define gtk_requisition_free IA__gtk_requisition_free

extern __typeof (gtk_requisition_get_type) IA__gtk_requisition_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_requisition_get_type IA__gtk_requisition_get_type

extern __typeof (gtk_widget_activate) IA__gtk_widget_activate __attribute((visibility("hidden")));
#define gtk_widget_activate IA__gtk_widget_activate

extern __typeof (gtk_widget_is_composited) IA__gtk_widget_is_composited __attribute((visibility("hidden")));
#define gtk_widget_is_composited IA__gtk_widget_is_composited

extern __typeof (gtk_widget_add_accelerator) IA__gtk_widget_add_accelerator __attribute((visibility("hidden")));
#define gtk_widget_add_accelerator IA__gtk_widget_add_accelerator

extern __typeof (gtk_widget_add_events) IA__gtk_widget_add_events __attribute((visibility("hidden")));
#define gtk_widget_add_events IA__gtk_widget_add_events

extern __typeof (gtk_widget_add_mnemonic_label) IA__gtk_widget_add_mnemonic_label __attribute((visibility("hidden")));
#define gtk_widget_add_mnemonic_label IA__gtk_widget_add_mnemonic_label

extern __typeof (gtk_widget_can_activate_accel) IA__gtk_widget_can_activate_accel __attribute((visibility("hidden")));
#define gtk_widget_can_activate_accel IA__gtk_widget_can_activate_accel

extern __typeof (gtk_widget_child_focus) IA__gtk_widget_child_focus __attribute((visibility("hidden")));
#define gtk_widget_child_focus IA__gtk_widget_child_focus

extern __typeof (gtk_widget_child_notify) IA__gtk_widget_child_notify __attribute((visibility("hidden")));
#define gtk_widget_child_notify IA__gtk_widget_child_notify

extern __typeof (gtk_widget_class_find_style_property) IA__gtk_widget_class_find_style_property __attribute((visibility("hidden")));
#define gtk_widget_class_find_style_property IA__gtk_widget_class_find_style_property

extern __typeof (gtk_widget_class_install_style_property) IA__gtk_widget_class_install_style_property __attribute((visibility("hidden")));
#define gtk_widget_class_install_style_property IA__gtk_widget_class_install_style_property

extern __typeof (gtk_widget_class_install_style_property_parser) IA__gtk_widget_class_install_style_property_parser __attribute((visibility("hidden")));
#define gtk_widget_class_install_style_property_parser IA__gtk_widget_class_install_style_property_parser

extern __typeof (gtk_widget_class_list_style_properties) IA__gtk_widget_class_list_style_properties __attribute((visibility("hidden")));
#define gtk_widget_class_list_style_properties IA__gtk_widget_class_list_style_properties

extern __typeof (gtk_widget_class_path) IA__gtk_widget_class_path __attribute((visibility("hidden")));
#define gtk_widget_class_path IA__gtk_widget_class_path

extern __typeof (gtk_widget_create_pango_context) IA__gtk_widget_create_pango_context __attribute((visibility("hidden")));
#define gtk_widget_create_pango_context IA__gtk_widget_create_pango_context

extern __typeof (gtk_widget_create_pango_layout) IA__gtk_widget_create_pango_layout __attribute((visibility("hidden")));
#define gtk_widget_create_pango_layout IA__gtk_widget_create_pango_layout

extern __typeof (gtk_widget_destroy) IA__gtk_widget_destroy __attribute((visibility("hidden")));
#define gtk_widget_destroy IA__gtk_widget_destroy

extern __typeof (gtk_widget_destroyed) IA__gtk_widget_destroyed __attribute((visibility("hidden")));
#define gtk_widget_destroyed IA__gtk_widget_destroyed

extern __typeof (gtk_widget_ensure_style) IA__gtk_widget_ensure_style __attribute((visibility("hidden")));
#define gtk_widget_ensure_style IA__gtk_widget_ensure_style

extern __typeof (gtk_widget_error_bell) IA__gtk_widget_error_bell __attribute((visibility("hidden")));
#define gtk_widget_error_bell IA__gtk_widget_error_bell

extern __typeof (gtk_widget_event) IA__gtk_widget_event __attribute((visibility("hidden")));
#define gtk_widget_event IA__gtk_widget_event

extern __typeof (gtk_widget_freeze_child_notify) IA__gtk_widget_freeze_child_notify __attribute((visibility("hidden")));
#define gtk_widget_freeze_child_notify IA__gtk_widget_freeze_child_notify

extern __typeof (gtk_widget_get_accessible) IA__gtk_widget_get_accessible __attribute((visibility("hidden")));
#define gtk_widget_get_accessible IA__gtk_widget_get_accessible

extern __typeof (gtk_widget_get_allocation) IA__gtk_widget_get_allocation __attribute((visibility("hidden")));
#define gtk_widget_get_allocation IA__gtk_widget_get_allocation

extern __typeof (gtk_widget_get_ancestor) IA__gtk_widget_get_ancestor __attribute((visibility("hidden")));
#define gtk_widget_get_ancestor IA__gtk_widget_get_ancestor

extern __typeof (gtk_widget_get_app_paintable) IA__gtk_widget_get_app_paintable __attribute((visibility("hidden")));
#define gtk_widget_get_app_paintable IA__gtk_widget_get_app_paintable

extern __typeof (gtk_widget_get_child_requisition) IA__gtk_widget_get_child_requisition __attribute((visibility("hidden")));
#define gtk_widget_get_child_requisition IA__gtk_widget_get_child_requisition

extern __typeof (gtk_widget_get_child_visible) IA__gtk_widget_get_child_visible __attribute((visibility("hidden")));
#define gtk_widget_get_child_visible IA__gtk_widget_get_child_visible

extern __typeof (gtk_widget_get_clipboard) IA__gtk_widget_get_clipboard __attribute((visibility("hidden")));
#define gtk_widget_get_clipboard IA__gtk_widget_get_clipboard

extern __typeof (gtk_widget_get_colormap) IA__gtk_widget_get_colormap __attribute((visibility("hidden")));
#define gtk_widget_get_colormap IA__gtk_widget_get_colormap

extern __typeof (gtk_widget_get_composite_name) IA__gtk_widget_get_composite_name __attribute((visibility("hidden")));
#define gtk_widget_get_composite_name IA__gtk_widget_get_composite_name

extern __typeof (gtk_widget_get_default_colormap) IA__gtk_widget_get_default_colormap __attribute((visibility("hidden")));
#define gtk_widget_get_default_colormap IA__gtk_widget_get_default_colormap

extern __typeof (gtk_widget_get_default_direction) IA__gtk_widget_get_default_direction __attribute((visibility("hidden")));
#define gtk_widget_get_default_direction IA__gtk_widget_get_default_direction

extern __typeof (gtk_widget_get_default_style) IA__gtk_widget_get_default_style __attribute((visibility("hidden")));
#define gtk_widget_get_default_style IA__gtk_widget_get_default_style

extern __typeof (gtk_widget_get_default_visual) IA__gtk_widget_get_default_visual __attribute((visibility("hidden")));
#define gtk_widget_get_default_visual IA__gtk_widget_get_default_visual

extern __typeof (gtk_widget_get_direction) IA__gtk_widget_get_direction __attribute((visibility("hidden")));
#define gtk_widget_get_direction IA__gtk_widget_get_direction

extern __typeof (gtk_widget_get_display) IA__gtk_widget_get_display __attribute((visibility("hidden")));
#define gtk_widget_get_display IA__gtk_widget_get_display

extern __typeof (gtk_widget_get_double_buffered) IA__gtk_widget_get_double_buffered __attribute((visibility("hidden")));
#define gtk_widget_get_double_buffered IA__gtk_widget_get_double_buffered

extern __typeof (gtk_widget_get_events) IA__gtk_widget_get_events __attribute((visibility("hidden")));
#define gtk_widget_get_events IA__gtk_widget_get_events

extern __typeof (gtk_widget_get_extension_events) IA__gtk_widget_get_extension_events __attribute((visibility("hidden")));
#define gtk_widget_get_extension_events IA__gtk_widget_get_extension_events

extern __typeof (gtk_widget_get_has_tooltip) IA__gtk_widget_get_has_tooltip __attribute((visibility("hidden")));
#define gtk_widget_get_has_tooltip IA__gtk_widget_get_has_tooltip

extern __typeof (gtk_widget_get_modifier_style) IA__gtk_widget_get_modifier_style __attribute((visibility("hidden")));
#define gtk_widget_get_modifier_style IA__gtk_widget_get_modifier_style

extern __typeof (gtk_widget_get_name) IA__gtk_widget_get_name __attribute((visibility("hidden")));
#define gtk_widget_get_name IA__gtk_widget_get_name

extern __typeof (gtk_widget_get_no_show_all) IA__gtk_widget_get_no_show_all __attribute((visibility("hidden")));
#define gtk_widget_get_no_show_all IA__gtk_widget_get_no_show_all

extern __typeof (gtk_widget_get_pango_context) IA__gtk_widget_get_pango_context __attribute((visibility("hidden")));
#define gtk_widget_get_pango_context IA__gtk_widget_get_pango_context

extern __typeof (gtk_widget_get_parent) IA__gtk_widget_get_parent __attribute((visibility("hidden")));
#define gtk_widget_get_parent IA__gtk_widget_get_parent

extern __typeof (gtk_widget_get_parent_window) IA__gtk_widget_get_parent_window __attribute((visibility("hidden")));
#define gtk_widget_get_parent_window IA__gtk_widget_get_parent_window

extern __typeof (gtk_widget_get_pointer) IA__gtk_widget_get_pointer __attribute((visibility("hidden")));
#define gtk_widget_get_pointer IA__gtk_widget_get_pointer

extern __typeof (gtk_widget_get_receives_default) IA__gtk_widget_get_receives_default __attribute((visibility("hidden")));
#define gtk_widget_get_receives_default IA__gtk_widget_get_receives_default

extern __typeof (gtk_widget_get_root_window) IA__gtk_widget_get_root_window __attribute((visibility("hidden")));
#define gtk_widget_get_root_window IA__gtk_widget_get_root_window

extern __typeof (gtk_widget_get_requisition) IA__gtk_widget_get_requisition __attribute((visibility("hidden")));
#define gtk_widget_get_requisition IA__gtk_widget_get_requisition

extern __typeof (gtk_widget_get_screen) IA__gtk_widget_get_screen __attribute((visibility("hidden")));
#define gtk_widget_get_screen IA__gtk_widget_get_screen

extern __typeof (gtk_widget_get_sensitive) IA__gtk_widget_get_sensitive __attribute((visibility("hidden")));
#define gtk_widget_get_sensitive IA__gtk_widget_get_sensitive

extern __typeof (gtk_widget_get_settings) IA__gtk_widget_get_settings __attribute((visibility("hidden")));
#define gtk_widget_get_settings IA__gtk_widget_get_settings

extern __typeof (gtk_widget_get_size_request) IA__gtk_widget_get_size_request __attribute((visibility("hidden")));
#define gtk_widget_get_size_request IA__gtk_widget_get_size_request

extern __typeof (gtk_widget_get_snapshot) IA__gtk_widget_get_snapshot __attribute((visibility("hidden")));
#define gtk_widget_get_snapshot IA__gtk_widget_get_snapshot

extern __typeof (gtk_widget_get_state) IA__gtk_widget_get_state __attribute((visibility("hidden")));
#define gtk_widget_get_state IA__gtk_widget_get_state

extern __typeof (gtk_widget_get_style) IA__gtk_widget_get_style __attribute((visibility("hidden")));
#define gtk_widget_get_style IA__gtk_widget_get_style

extern __typeof (gtk_widget_get_tooltip_markup) IA__gtk_widget_get_tooltip_markup __attribute((visibility("hidden")));
#define gtk_widget_get_tooltip_markup IA__gtk_widget_get_tooltip_markup

extern __typeof (gtk_widget_get_tooltip_text) IA__gtk_widget_get_tooltip_text __attribute((visibility("hidden")));
#define gtk_widget_get_tooltip_text IA__gtk_widget_get_tooltip_text

extern __typeof (gtk_widget_get_tooltip_window) IA__gtk_widget_get_tooltip_window __attribute((visibility("hidden")));
#define gtk_widget_get_tooltip_window IA__gtk_widget_get_tooltip_window

extern __typeof (gtk_widget_get_toplevel) IA__gtk_widget_get_toplevel __attribute((visibility("hidden")));
#define gtk_widget_get_toplevel IA__gtk_widget_get_toplevel

extern __typeof (gtk_widget_get_type) IA__gtk_widget_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_widget_get_type IA__gtk_widget_get_type

extern __typeof (gtk_widget_get_visible) IA__gtk_widget_get_visible __attribute((visibility("hidden")));
#define gtk_widget_get_visible IA__gtk_widget_get_visible

extern __typeof (gtk_widget_get_visual) IA__gtk_widget_get_visual __attribute((visibility("hidden")));
#define gtk_widget_get_visual IA__gtk_widget_get_visual

extern __typeof (gtk_widget_grab_default) IA__gtk_widget_grab_default __attribute((visibility("hidden")));
#define gtk_widget_grab_default IA__gtk_widget_grab_default

extern __typeof (gtk_widget_grab_focus) IA__gtk_widget_grab_focus __attribute((visibility("hidden")));
#define gtk_widget_grab_focus IA__gtk_widget_grab_focus

extern __typeof (gtk_widget_has_default) IA__gtk_widget_has_default __attribute((visibility("hidden")));
#define gtk_widget_has_default IA__gtk_widget_has_default

extern __typeof (gtk_widget_has_focus) IA__gtk_widget_has_focus __attribute((visibility("hidden")));
#define gtk_widget_has_focus IA__gtk_widget_has_focus

extern __typeof (gtk_widget_has_grab) IA__gtk_widget_has_grab __attribute((visibility("hidden")));
#define gtk_widget_has_grab IA__gtk_widget_has_grab

extern __typeof (gtk_widget_has_rc_style) IA__gtk_widget_has_rc_style __attribute((visibility("hidden")));
#define gtk_widget_has_rc_style IA__gtk_widget_has_rc_style

extern __typeof (gtk_widget_has_screen) IA__gtk_widget_has_screen __attribute((visibility("hidden")));
#define gtk_widget_has_screen IA__gtk_widget_has_screen

extern __typeof (gtk_widget_hide) IA__gtk_widget_hide __attribute((visibility("hidden")));
#define gtk_widget_hide IA__gtk_widget_hide

extern __typeof (gtk_widget_hide_all) IA__gtk_widget_hide_all __attribute((visibility("hidden")));
#define gtk_widget_hide_all IA__gtk_widget_hide_all

extern __typeof (gtk_widget_hide_on_delete) IA__gtk_widget_hide_on_delete __attribute((visibility("hidden")));
#define gtk_widget_hide_on_delete IA__gtk_widget_hide_on_delete

extern __typeof (gtk_widget_intersect) IA__gtk_widget_intersect __attribute((visibility("hidden")));
#define gtk_widget_intersect IA__gtk_widget_intersect

extern __typeof (gtk_widget_is_ancestor) IA__gtk_widget_is_ancestor __attribute((visibility("hidden")));
#define gtk_widget_is_ancestor IA__gtk_widget_is_ancestor

extern __typeof (gtk_widget_is_focus) IA__gtk_widget_is_focus __attribute((visibility("hidden")));
#define gtk_widget_is_focus IA__gtk_widget_is_focus

extern __typeof (gtk_widget_is_sensitive) IA__gtk_widget_is_sensitive __attribute((visibility("hidden")));
#define gtk_widget_is_sensitive IA__gtk_widget_is_sensitive

extern __typeof (gtk_widget_is_toplevel) IA__gtk_widget_is_toplevel __attribute((visibility("hidden")));
#define gtk_widget_is_toplevel IA__gtk_widget_is_toplevel

extern __typeof (gtk_widget_is_drawable) IA__gtk_widget_is_drawable __attribute((visibility("hidden")));
#define gtk_widget_is_drawable IA__gtk_widget_is_drawable

extern __typeof (gtk_widget_keynav_failed) IA__gtk_widget_keynav_failed __attribute((visibility("hidden")));
#define gtk_widget_keynav_failed IA__gtk_widget_keynav_failed

extern __typeof (gtk_widget_list_accel_closures) IA__gtk_widget_list_accel_closures __attribute((visibility("hidden")));
#define gtk_widget_list_accel_closures IA__gtk_widget_list_accel_closures

extern __typeof (gtk_widget_list_mnemonic_labels) IA__gtk_widget_list_mnemonic_labels __attribute((visibility("hidden")));
#define gtk_widget_list_mnemonic_labels IA__gtk_widget_list_mnemonic_labels

extern __typeof (gtk_widget_map) IA__gtk_widget_map __attribute((visibility("hidden")));
#define gtk_widget_map IA__gtk_widget_map

extern __typeof (gtk_widget_mnemonic_activate) IA__gtk_widget_mnemonic_activate __attribute((visibility("hidden")));
#define gtk_widget_mnemonic_activate IA__gtk_widget_mnemonic_activate

extern __typeof (gtk_widget_modify_base) IA__gtk_widget_modify_base __attribute((visibility("hidden")));
#define gtk_widget_modify_base IA__gtk_widget_modify_base

extern __typeof (gtk_widget_modify_bg) IA__gtk_widget_modify_bg __attribute((visibility("hidden")));
#define gtk_widget_modify_bg IA__gtk_widget_modify_bg

extern __typeof (gtk_widget_modify_fg) IA__gtk_widget_modify_fg __attribute((visibility("hidden")));
#define gtk_widget_modify_fg IA__gtk_widget_modify_fg

extern __typeof (gtk_widget_modify_cursor) IA__gtk_widget_modify_cursor __attribute((visibility("hidden")));
#define gtk_widget_modify_cursor IA__gtk_widget_modify_cursor

extern __typeof (gtk_widget_modify_font) IA__gtk_widget_modify_font __attribute((visibility("hidden")));
#define gtk_widget_modify_font IA__gtk_widget_modify_font

extern __typeof (gtk_widget_modify_style) IA__gtk_widget_modify_style __attribute((visibility("hidden")));
#define gtk_widget_modify_style IA__gtk_widget_modify_style

extern __typeof (gtk_widget_modify_text) IA__gtk_widget_modify_text __attribute((visibility("hidden")));
#define gtk_widget_modify_text IA__gtk_widget_modify_text

extern __typeof (gtk_widget_new) IA__gtk_widget_new __attribute((visibility("hidden")));
#define gtk_widget_new IA__gtk_widget_new

extern __typeof (gtk_widget_path) IA__gtk_widget_path __attribute((visibility("hidden")));
#define gtk_widget_path IA__gtk_widget_path

extern __typeof (gtk_widget_pop_colormap) IA__gtk_widget_pop_colormap __attribute((visibility("hidden")));
#define gtk_widget_pop_colormap IA__gtk_widget_pop_colormap

extern __typeof (gtk_widget_pop_composite_child) IA__gtk_widget_pop_composite_child __attribute((visibility("hidden")));
#define gtk_widget_pop_composite_child IA__gtk_widget_pop_composite_child

extern __typeof (gtk_widget_push_colormap) IA__gtk_widget_push_colormap __attribute((visibility("hidden")));
#define gtk_widget_push_colormap IA__gtk_widget_push_colormap

extern __typeof (gtk_widget_push_composite_child) IA__gtk_widget_push_composite_child __attribute((visibility("hidden")));
#define gtk_widget_push_composite_child IA__gtk_widget_push_composite_child

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_widget_draw) IA__gtk_widget_draw __attribute((visibility("hidden")));
#define gtk_widget_draw IA__gtk_widget_draw

extern __typeof (gtk_widget_queue_clear) IA__gtk_widget_queue_clear __attribute((visibility("hidden")));
#define gtk_widget_queue_clear IA__gtk_widget_queue_clear

extern __typeof (gtk_widget_queue_clear_area) IA__gtk_widget_queue_clear_area __attribute((visibility("hidden")));
#define gtk_widget_queue_clear_area IA__gtk_widget_queue_clear_area

extern __typeof (gtk_widget_set) IA__gtk_widget_set __attribute((visibility("hidden")));
#define gtk_widget_set IA__gtk_widget_set

extern __typeof (gtk_widget_set_uposition) IA__gtk_widget_set_uposition __attribute((visibility("hidden")));
#define gtk_widget_set_uposition IA__gtk_widget_set_uposition

extern __typeof (gtk_widget_set_usize) IA__gtk_widget_set_usize __attribute((visibility("hidden")));
#define gtk_widget_set_usize IA__gtk_widget_set_usize

extern __typeof (gtk_widget_ref) IA__gtk_widget_ref __attribute((visibility("hidden")));
#define gtk_widget_ref IA__gtk_widget_ref

extern __typeof (gtk_widget_unref) IA__gtk_widget_unref __attribute((visibility("hidden")));
#define gtk_widget_unref IA__gtk_widget_unref

#endif
extern __typeof (gtk_widget_queue_draw) IA__gtk_widget_queue_draw __attribute((visibility("hidden")));
#define gtk_widget_queue_draw IA__gtk_widget_queue_draw

extern __typeof (gtk_widget_queue_draw_area) IA__gtk_widget_queue_draw_area __attribute((visibility("hidden")));
#define gtk_widget_queue_draw_area IA__gtk_widget_queue_draw_area

extern __typeof (gtk_widget_queue_resize) IA__gtk_widget_queue_resize __attribute((visibility("hidden")));
#define gtk_widget_queue_resize IA__gtk_widget_queue_resize

extern __typeof (gtk_widget_queue_resize_no_redraw) IA__gtk_widget_queue_resize_no_redraw __attribute((visibility("hidden")));
#define gtk_widget_queue_resize_no_redraw IA__gtk_widget_queue_resize_no_redraw

extern __typeof (gtk_widget_realize) IA__gtk_widget_realize __attribute((visibility("hidden")));
#define gtk_widget_realize IA__gtk_widget_realize

extern __typeof (gtk_widget_region_intersect) IA__gtk_widget_region_intersect __attribute((visibility("hidden")));
#define gtk_widget_region_intersect IA__gtk_widget_region_intersect

extern __typeof (gtk_widget_remove_accelerator) IA__gtk_widget_remove_accelerator __attribute((visibility("hidden")));
#define gtk_widget_remove_accelerator IA__gtk_widget_remove_accelerator

extern __typeof (gtk_widget_remove_mnemonic_label) IA__gtk_widget_remove_mnemonic_label __attribute((visibility("hidden")));
#define gtk_widget_remove_mnemonic_label IA__gtk_widget_remove_mnemonic_label

extern __typeof (gtk_widget_render_icon) IA__gtk_widget_render_icon __attribute((visibility("hidden")));
#define gtk_widget_render_icon IA__gtk_widget_render_icon

extern __typeof (gtk_widget_reparent) IA__gtk_widget_reparent __attribute((visibility("hidden")));
#define gtk_widget_reparent IA__gtk_widget_reparent

extern __typeof (gtk_widget_reset_rc_styles) IA__gtk_widget_reset_rc_styles __attribute((visibility("hidden")));
#define gtk_widget_reset_rc_styles IA__gtk_widget_reset_rc_styles

extern __typeof (gtk_widget_reset_shapes) IA__gtk_widget_reset_shapes __attribute((visibility("hidden")));
#define gtk_widget_reset_shapes IA__gtk_widget_reset_shapes

extern __typeof (gtk_widget_send_expose) IA__gtk_widget_send_expose __attribute((visibility("hidden")));
#define gtk_widget_send_expose IA__gtk_widget_send_expose

extern __typeof (gtk_widget_send_focus_change) IA__gtk_widget_send_focus_change __attribute((visibility("hidden")));
#define gtk_widget_send_focus_change IA__gtk_widget_send_focus_change

extern __typeof (gtk_widget_set_accel_path) IA__gtk_widget_set_accel_path __attribute((visibility("hidden")));
#define gtk_widget_set_accel_path IA__gtk_widget_set_accel_path

extern __typeof (gtk_widget_set_allocation) IA__gtk_widget_set_allocation __attribute((visibility("hidden")));
#define gtk_widget_set_allocation IA__gtk_widget_set_allocation

extern __typeof (gtk_widget_set_app_paintable) IA__gtk_widget_set_app_paintable __attribute((visibility("hidden")));
#define gtk_widget_set_app_paintable IA__gtk_widget_set_app_paintable

extern __typeof (gtk_widget_set_child_visible) IA__gtk_widget_set_child_visible __attribute((visibility("hidden")));
#define gtk_widget_set_child_visible IA__gtk_widget_set_child_visible

extern __typeof (gtk_widget_set_colormap) IA__gtk_widget_set_colormap __attribute((visibility("hidden")));
#define gtk_widget_set_colormap IA__gtk_widget_set_colormap

extern __typeof (gtk_widget_set_composite_name) IA__gtk_widget_set_composite_name __attribute((visibility("hidden")));
#define gtk_widget_set_composite_name IA__gtk_widget_set_composite_name

extern __typeof (gtk_widget_set_default_colormap) IA__gtk_widget_set_default_colormap __attribute((visibility("hidden")));
#define gtk_widget_set_default_colormap IA__gtk_widget_set_default_colormap

extern __typeof (gtk_widget_set_default_direction) IA__gtk_widget_set_default_direction __attribute((visibility("hidden")));
#define gtk_widget_set_default_direction IA__gtk_widget_set_default_direction

extern __typeof (gtk_widget_set_direction) IA__gtk_widget_set_direction __attribute((visibility("hidden")));
#define gtk_widget_set_direction IA__gtk_widget_set_direction

extern __typeof (gtk_widget_set_double_buffered) IA__gtk_widget_set_double_buffered __attribute((visibility("hidden")));
#define gtk_widget_set_double_buffered IA__gtk_widget_set_double_buffered

extern __typeof (gtk_widget_set_events) IA__gtk_widget_set_events __attribute((visibility("hidden")));
#define gtk_widget_set_events IA__gtk_widget_set_events

extern __typeof (gtk_widget_set_extension_events) IA__gtk_widget_set_extension_events __attribute((visibility("hidden")));
#define gtk_widget_set_extension_events IA__gtk_widget_set_extension_events

extern __typeof (gtk_widget_set_has_tooltip) IA__gtk_widget_set_has_tooltip __attribute((visibility("hidden")));
#define gtk_widget_set_has_tooltip IA__gtk_widget_set_has_tooltip

extern __typeof (gtk_widget_set_name) IA__gtk_widget_set_name __attribute((visibility("hidden")));
#define gtk_widget_set_name IA__gtk_widget_set_name

extern __typeof (gtk_widget_set_no_show_all) IA__gtk_widget_set_no_show_all __attribute((visibility("hidden")));
#define gtk_widget_set_no_show_all IA__gtk_widget_set_no_show_all

extern __typeof (gtk_widget_set_parent) IA__gtk_widget_set_parent __attribute((visibility("hidden")));
#define gtk_widget_set_parent IA__gtk_widget_set_parent

extern __typeof (gtk_widget_set_parent_window) IA__gtk_widget_set_parent_window __attribute((visibility("hidden")));
#define gtk_widget_set_parent_window IA__gtk_widget_set_parent_window

extern __typeof (gtk_widget_set_receives_default) IA__gtk_widget_set_receives_default __attribute((visibility("hidden")));
#define gtk_widget_set_receives_default IA__gtk_widget_set_receives_default

extern __typeof (gtk_widget_set_redraw_on_allocate) IA__gtk_widget_set_redraw_on_allocate __attribute((visibility("hidden")));
#define gtk_widget_set_redraw_on_allocate IA__gtk_widget_set_redraw_on_allocate

extern __typeof (gtk_widget_set_scroll_adjustments) IA__gtk_widget_set_scroll_adjustments __attribute((visibility("hidden")));
#define gtk_widget_set_scroll_adjustments IA__gtk_widget_set_scroll_adjustments

extern __typeof (gtk_widget_set_sensitive) IA__gtk_widget_set_sensitive __attribute((visibility("hidden")));
#define gtk_widget_set_sensitive IA__gtk_widget_set_sensitive

extern __typeof (gtk_widget_set_size_request) IA__gtk_widget_set_size_request __attribute((visibility("hidden")));
#define gtk_widget_set_size_request IA__gtk_widget_set_size_request

extern __typeof (gtk_widget_set_state) IA__gtk_widget_set_state __attribute((visibility("hidden")));
#define gtk_widget_set_state IA__gtk_widget_set_state

extern __typeof (gtk_widget_set_style) IA__gtk_widget_set_style __attribute((visibility("hidden")));
#define gtk_widget_set_style IA__gtk_widget_set_style

extern __typeof (gtk_widget_set_tooltip_markup) IA__gtk_widget_set_tooltip_markup __attribute((visibility("hidden")));
#define gtk_widget_set_tooltip_markup IA__gtk_widget_set_tooltip_markup

extern __typeof (gtk_widget_set_tooltip_text) IA__gtk_widget_set_tooltip_text __attribute((visibility("hidden")));
#define gtk_widget_set_tooltip_text IA__gtk_widget_set_tooltip_text

extern __typeof (gtk_widget_set_tooltip_window) IA__gtk_widget_set_tooltip_window __attribute((visibility("hidden")));
#define gtk_widget_set_tooltip_window IA__gtk_widget_set_tooltip_window

extern __typeof (gtk_widget_set_visible) IA__gtk_widget_set_visible __attribute((visibility("hidden")));
#define gtk_widget_set_visible IA__gtk_widget_set_visible

extern __typeof (gtk_widget_set_window) IA__gtk_widget_set_window __attribute((visibility("hidden")));
#define gtk_widget_set_window IA__gtk_widget_set_window

extern __typeof (gtk_widget_shape_combine_mask) IA__gtk_widget_shape_combine_mask __attribute((visibility("hidden")));
#define gtk_widget_shape_combine_mask IA__gtk_widget_shape_combine_mask

extern __typeof (gtk_widget_input_shape_combine_mask) IA__gtk_widget_input_shape_combine_mask __attribute((visibility("hidden")));
#define gtk_widget_input_shape_combine_mask IA__gtk_widget_input_shape_combine_mask

extern __typeof (gtk_widget_show) IA__gtk_widget_show __attribute((visibility("hidden")));
#define gtk_widget_show IA__gtk_widget_show

extern __typeof (gtk_widget_show_all) IA__gtk_widget_show_all __attribute((visibility("hidden")));
#define gtk_widget_show_all IA__gtk_widget_show_all

extern __typeof (gtk_widget_show_now) IA__gtk_widget_show_now __attribute((visibility("hidden")));
#define gtk_widget_show_now IA__gtk_widget_show_now

extern __typeof (gtk_widget_size_allocate) IA__gtk_widget_size_allocate __attribute((visibility("hidden")));
#define gtk_widget_size_allocate IA__gtk_widget_size_allocate

extern __typeof (gtk_widget_size_request) IA__gtk_widget_size_request __attribute((visibility("hidden")));
#define gtk_widget_size_request IA__gtk_widget_size_request

extern __typeof (gtk_widget_style_attach) IA__gtk_widget_style_attach __attribute((visibility("hidden")));
#define gtk_widget_style_attach IA__gtk_widget_style_attach

extern __typeof (gtk_widget_style_get) IA__gtk_widget_style_get __attribute((visibility("hidden"))) G_GNUC_NULL_TERMINATED;
#define gtk_widget_style_get IA__gtk_widget_style_get

extern __typeof (gtk_widget_style_get_property) IA__gtk_widget_style_get_property __attribute((visibility("hidden")));
#define gtk_widget_style_get_property IA__gtk_widget_style_get_property

extern __typeof (gtk_widget_style_get_valist) IA__gtk_widget_style_get_valist __attribute((visibility("hidden")));
#define gtk_widget_style_get_valist IA__gtk_widget_style_get_valist

extern __typeof (gtk_widget_thaw_child_notify) IA__gtk_widget_thaw_child_notify __attribute((visibility("hidden")));
#define gtk_widget_thaw_child_notify IA__gtk_widget_thaw_child_notify

extern __typeof (gtk_widget_translate_coordinates) IA__gtk_widget_translate_coordinates __attribute((visibility("hidden")));
#define gtk_widget_translate_coordinates IA__gtk_widget_translate_coordinates

extern __typeof (gtk_widget_trigger_tooltip_query) IA__gtk_widget_trigger_tooltip_query __attribute((visibility("hidden")));
#define gtk_widget_trigger_tooltip_query IA__gtk_widget_trigger_tooltip_query

extern __typeof (gtk_widget_unmap) IA__gtk_widget_unmap __attribute((visibility("hidden")));
#define gtk_widget_unmap IA__gtk_widget_unmap

extern __typeof (gtk_widget_unparent) IA__gtk_widget_unparent __attribute((visibility("hidden")));
#define gtk_widget_unparent IA__gtk_widget_unparent

extern __typeof (gtk_widget_unrealize) IA__gtk_widget_unrealize __attribute((visibility("hidden")));
#define gtk_widget_unrealize IA__gtk_widget_unrealize

extern __typeof (gtk_widget_get_window) IA__gtk_widget_get_window __attribute((visibility("hidden")));
#define gtk_widget_get_window IA__gtk_widget_get_window

extern __typeof (gtk_widget_set_has_window) IA__gtk_widget_set_has_window __attribute((visibility("hidden")));
#define gtk_widget_set_has_window IA__gtk_widget_set_has_window

extern __typeof (gtk_widget_get_has_window) IA__gtk_widget_get_has_window __attribute((visibility("hidden")));
#define gtk_widget_get_has_window IA__gtk_widget_get_has_window

extern __typeof (gtk_widget_set_can_default) IA__gtk_widget_set_can_default __attribute((visibility("hidden")));
#define gtk_widget_set_can_default IA__gtk_widget_set_can_default

extern __typeof (gtk_widget_get_can_default) IA__gtk_widget_get_can_default __attribute((visibility("hidden")));
#define gtk_widget_get_can_default IA__gtk_widget_get_can_default

extern __typeof (gtk_widget_set_can_focus) IA__gtk_widget_set_can_focus __attribute((visibility("hidden")));
#define gtk_widget_set_can_focus IA__gtk_widget_set_can_focus

extern __typeof (gtk_widget_get_can_focus) IA__gtk_widget_get_can_focus __attribute((visibility("hidden")));
#define gtk_widget_get_can_focus IA__gtk_widget_get_can_focus

extern __typeof (gtk_widget_set_realized) IA__gtk_widget_set_realized __attribute((visibility("hidden")));
#define gtk_widget_set_realized IA__gtk_widget_set_realized

extern __typeof (gtk_widget_get_realized) IA__gtk_widget_get_realized __attribute((visibility("hidden")));
#define gtk_widget_get_realized IA__gtk_widget_get_realized

extern __typeof (gtk_widget_set_mapped) IA__gtk_widget_set_mapped __attribute((visibility("hidden")));
#define gtk_widget_set_mapped IA__gtk_widget_set_mapped

extern __typeof (gtk_widget_get_mapped) IA__gtk_widget_get_mapped __attribute((visibility("hidden")));
#define gtk_widget_get_mapped IA__gtk_widget_get_mapped

#endif
#endif
#if IN_HEADER(__GTK_WINDOW_H__)
#if IN_FILE(__GTK_WINDOW_C__)
extern __typeof (gtk_window_activate_default) IA__gtk_window_activate_default __attribute((visibility("hidden")));
#define gtk_window_activate_default IA__gtk_window_activate_default

extern __typeof (gtk_window_activate_focus) IA__gtk_window_activate_focus __attribute((visibility("hidden")));
#define gtk_window_activate_focus IA__gtk_window_activate_focus

extern __typeof (gtk_window_activate_key) IA__gtk_window_activate_key __attribute((visibility("hidden")));
#define gtk_window_activate_key IA__gtk_window_activate_key

extern __typeof (gtk_window_add_accel_group) IA__gtk_window_add_accel_group __attribute((visibility("hidden")));
#define gtk_window_add_accel_group IA__gtk_window_add_accel_group

extern __typeof (gtk_window_add_embedded_xid) IA__gtk_window_add_embedded_xid __attribute((visibility("hidden")));
#define gtk_window_add_embedded_xid IA__gtk_window_add_embedded_xid

extern __typeof (gtk_window_add_mnemonic) IA__gtk_window_add_mnemonic __attribute((visibility("hidden")));
#define gtk_window_add_mnemonic IA__gtk_window_add_mnemonic

extern __typeof (gtk_window_begin_move_drag) IA__gtk_window_begin_move_drag __attribute((visibility("hidden")));
#define gtk_window_begin_move_drag IA__gtk_window_begin_move_drag

extern __typeof (gtk_window_begin_resize_drag) IA__gtk_window_begin_resize_drag __attribute((visibility("hidden")));
#define gtk_window_begin_resize_drag IA__gtk_window_begin_resize_drag

extern __typeof (gtk_window_deiconify) IA__gtk_window_deiconify __attribute((visibility("hidden")));
#define gtk_window_deiconify IA__gtk_window_deiconify

extern __typeof (gtk_window_fullscreen) IA__gtk_window_fullscreen __attribute((visibility("hidden")));
#define gtk_window_fullscreen IA__gtk_window_fullscreen

extern __typeof (gtk_window_get_accept_focus) IA__gtk_window_get_accept_focus __attribute((visibility("hidden")));
#define gtk_window_get_accept_focus IA__gtk_window_get_accept_focus

extern __typeof (gtk_window_get_decorated) IA__gtk_window_get_decorated __attribute((visibility("hidden")));
#define gtk_window_get_decorated IA__gtk_window_get_decorated

extern __typeof (gtk_window_get_deletable) IA__gtk_window_get_deletable __attribute((visibility("hidden")));
#define gtk_window_get_deletable IA__gtk_window_get_deletable

extern __typeof (gtk_window_get_default_widget) IA__gtk_window_get_default_widget __attribute((visibility("hidden")));
#define gtk_window_get_default_widget IA__gtk_window_get_default_widget

extern __typeof (gtk_window_get_default_icon_list) IA__gtk_window_get_default_icon_list __attribute((visibility("hidden")));
#define gtk_window_get_default_icon_list IA__gtk_window_get_default_icon_list

extern __typeof (gtk_window_get_default_icon_name) IA__gtk_window_get_default_icon_name __attribute((visibility("hidden")));
#define gtk_window_get_default_icon_name IA__gtk_window_get_default_icon_name

extern __typeof (gtk_window_get_default_size) IA__gtk_window_get_default_size __attribute((visibility("hidden")));
#define gtk_window_get_default_size IA__gtk_window_get_default_size

extern __typeof (gtk_window_get_destroy_with_parent) IA__gtk_window_get_destroy_with_parent __attribute((visibility("hidden")));
#define gtk_window_get_destroy_with_parent IA__gtk_window_get_destroy_with_parent

extern __typeof (gtk_window_get_focus) IA__gtk_window_get_focus __attribute((visibility("hidden")));
#define gtk_window_get_focus IA__gtk_window_get_focus

extern __typeof (gtk_window_get_focus_on_map) IA__gtk_window_get_focus_on_map __attribute((visibility("hidden")));
#define gtk_window_get_focus_on_map IA__gtk_window_get_focus_on_map

extern __typeof (gtk_window_get_gravity) IA__gtk_window_get_gravity __attribute((visibility("hidden")));
#define gtk_window_get_gravity IA__gtk_window_get_gravity

extern __typeof (gtk_window_get_group) IA__gtk_window_get_group __attribute((visibility("hidden")));
#define gtk_window_get_group IA__gtk_window_get_group

extern __typeof (gtk_window_get_icon) IA__gtk_window_get_icon __attribute((visibility("hidden")));
#define gtk_window_get_icon IA__gtk_window_get_icon

extern __typeof (gtk_window_get_icon_list) IA__gtk_window_get_icon_list __attribute((visibility("hidden")));
#define gtk_window_get_icon_list IA__gtk_window_get_icon_list

extern __typeof (gtk_window_get_icon_name) IA__gtk_window_get_icon_name __attribute((visibility("hidden")));
#define gtk_window_get_icon_name IA__gtk_window_get_icon_name

extern __typeof (gtk_window_get_mnemonic_modifier) IA__gtk_window_get_mnemonic_modifier __attribute((visibility("hidden")));
#define gtk_window_get_mnemonic_modifier IA__gtk_window_get_mnemonic_modifier

extern __typeof (gtk_window_get_mnemonics_visible) IA__gtk_window_get_mnemonics_visible __attribute((visibility("hidden")));
#define gtk_window_get_mnemonics_visible IA__gtk_window_get_mnemonics_visible

extern __typeof (gtk_window_get_modal) IA__gtk_window_get_modal __attribute((visibility("hidden")));
#define gtk_window_get_modal IA__gtk_window_get_modal

extern __typeof (gtk_window_get_position) IA__gtk_window_get_position __attribute((visibility("hidden")));
#define gtk_window_get_position IA__gtk_window_get_position

extern __typeof (gtk_window_get_resizable) IA__gtk_window_get_resizable __attribute((visibility("hidden")));
#define gtk_window_get_resizable IA__gtk_window_get_resizable

extern __typeof (gtk_window_get_role) IA__gtk_window_get_role __attribute((visibility("hidden")));
#define gtk_window_get_role IA__gtk_window_get_role

extern __typeof (gtk_window_get_screen) IA__gtk_window_get_screen __attribute((visibility("hidden")));
#define gtk_window_get_screen IA__gtk_window_get_screen

extern __typeof (gtk_window_get_size) IA__gtk_window_get_size __attribute((visibility("hidden")));
#define gtk_window_get_size IA__gtk_window_get_size

extern __typeof (gtk_window_get_skip_pager_hint) IA__gtk_window_get_skip_pager_hint __attribute((visibility("hidden")));
#define gtk_window_get_skip_pager_hint IA__gtk_window_get_skip_pager_hint

extern __typeof (gtk_window_get_skip_taskbar_hint) IA__gtk_window_get_skip_taskbar_hint __attribute((visibility("hidden")));
#define gtk_window_get_skip_taskbar_hint IA__gtk_window_get_skip_taskbar_hint

extern __typeof (gtk_window_get_urgency_hint) IA__gtk_window_get_urgency_hint __attribute((visibility("hidden")));
#define gtk_window_get_urgency_hint IA__gtk_window_get_urgency_hint

extern __typeof (gtk_window_get_title) IA__gtk_window_get_title __attribute((visibility("hidden")));
#define gtk_window_get_title IA__gtk_window_get_title

extern __typeof (gtk_window_get_transient_for) IA__gtk_window_get_transient_for __attribute((visibility("hidden")));
#define gtk_window_get_transient_for IA__gtk_window_get_transient_for

extern __typeof (gtk_window_get_type) IA__gtk_window_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_window_get_type IA__gtk_window_get_type

extern __typeof (gtk_window_get_type_hint) IA__gtk_window_get_type_hint __attribute((visibility("hidden")));
#define gtk_window_get_type_hint IA__gtk_window_get_type_hint

extern __typeof (gtk_window_get_window_type) IA__gtk_window_get_window_type __attribute((visibility("hidden")));
#define gtk_window_get_window_type IA__gtk_window_get_window_type

extern __typeof (gtk_window_group_add_window) IA__gtk_window_group_add_window __attribute((visibility("hidden")));
#define gtk_window_group_add_window IA__gtk_window_group_add_window

extern __typeof (gtk_window_group_get_type) IA__gtk_window_group_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_window_group_get_type IA__gtk_window_group_get_type

extern __typeof (gtk_window_group_new) IA__gtk_window_group_new __attribute((visibility("hidden")));
#define gtk_window_group_new IA__gtk_window_group_new

extern __typeof (gtk_window_group_remove_window) IA__gtk_window_group_remove_window __attribute((visibility("hidden")));
#define gtk_window_group_remove_window IA__gtk_window_group_remove_window

extern __typeof (gtk_window_group_list_windows) IA__gtk_window_group_list_windows __attribute((visibility("hidden")));
#define gtk_window_group_list_windows IA__gtk_window_group_list_windows

extern __typeof (gtk_window_group_get_current_grab) IA__gtk_window_group_get_current_grab __attribute((visibility("hidden")));
#define gtk_window_group_get_current_grab IA__gtk_window_group_get_current_grab

extern __typeof (gtk_window_has_group) IA__gtk_window_has_group __attribute((visibility("hidden")));
#define gtk_window_has_group IA__gtk_window_has_group

extern __typeof (gtk_window_has_toplevel_focus) IA__gtk_window_has_toplevel_focus __attribute((visibility("hidden")));
#define gtk_window_has_toplevel_focus IA__gtk_window_has_toplevel_focus

extern __typeof (gtk_window_iconify) IA__gtk_window_iconify __attribute((visibility("hidden")));
#define gtk_window_iconify IA__gtk_window_iconify

extern __typeof (gtk_window_is_active) IA__gtk_window_is_active __attribute((visibility("hidden")));
#define gtk_window_is_active IA__gtk_window_is_active

extern __typeof (gtk_window_list_toplevels) IA__gtk_window_list_toplevels __attribute((visibility("hidden")));
#define gtk_window_list_toplevels IA__gtk_window_list_toplevels

extern __typeof (gtk_window_maximize) IA__gtk_window_maximize __attribute((visibility("hidden")));
#define gtk_window_maximize IA__gtk_window_maximize

extern __typeof (gtk_window_mnemonic_activate) IA__gtk_window_mnemonic_activate __attribute((visibility("hidden")));
#define gtk_window_mnemonic_activate IA__gtk_window_mnemonic_activate

extern __typeof (gtk_window_move) IA__gtk_window_move __attribute((visibility("hidden")));
#define gtk_window_move IA__gtk_window_move

extern __typeof (gtk_window_new) IA__gtk_window_new __attribute((visibility("hidden")));
#define gtk_window_new IA__gtk_window_new

extern __typeof (gtk_window_parse_geometry) IA__gtk_window_parse_geometry __attribute((visibility("hidden")));
#define gtk_window_parse_geometry IA__gtk_window_parse_geometry

extern __typeof (gtk_window_present) IA__gtk_window_present __attribute((visibility("hidden")));
#define gtk_window_present IA__gtk_window_present

extern __typeof (gtk_window_present_with_time) IA__gtk_window_present_with_time __attribute((visibility("hidden")));
#define gtk_window_present_with_time IA__gtk_window_present_with_time

extern __typeof (gtk_window_propagate_key_event) IA__gtk_window_propagate_key_event __attribute((visibility("hidden")));
#define gtk_window_propagate_key_event IA__gtk_window_propagate_key_event

extern __typeof (gtk_window_remove_accel_group) IA__gtk_window_remove_accel_group __attribute((visibility("hidden")));
#define gtk_window_remove_accel_group IA__gtk_window_remove_accel_group

extern __typeof (gtk_window_remove_embedded_xid) IA__gtk_window_remove_embedded_xid __attribute((visibility("hidden")));
#define gtk_window_remove_embedded_xid IA__gtk_window_remove_embedded_xid

extern __typeof (gtk_window_remove_mnemonic) IA__gtk_window_remove_mnemonic __attribute((visibility("hidden")));
#define gtk_window_remove_mnemonic IA__gtk_window_remove_mnemonic

extern __typeof (gtk_window_reshow_with_initial_size) IA__gtk_window_reshow_with_initial_size __attribute((visibility("hidden")));
#define gtk_window_reshow_with_initial_size IA__gtk_window_reshow_with_initial_size

extern __typeof (gtk_window_resize) IA__gtk_window_resize __attribute((visibility("hidden")));
#define gtk_window_resize IA__gtk_window_resize

extern __typeof (gtk_window_set_accept_focus) IA__gtk_window_set_accept_focus __attribute((visibility("hidden")));
#define gtk_window_set_accept_focus IA__gtk_window_set_accept_focus

extern __typeof (gtk_window_set_auto_startup_notification) IA__gtk_window_set_auto_startup_notification __attribute((visibility("hidden")));
#define gtk_window_set_auto_startup_notification IA__gtk_window_set_auto_startup_notification

extern __typeof (gtk_window_set_decorated) IA__gtk_window_set_decorated __attribute((visibility("hidden")));
#define gtk_window_set_decorated IA__gtk_window_set_decorated

extern __typeof (gtk_window_set_deletable) IA__gtk_window_set_deletable __attribute((visibility("hidden")));
#define gtk_window_set_deletable IA__gtk_window_set_deletable

extern __typeof (gtk_window_set_default) IA__gtk_window_set_default __attribute((visibility("hidden")));
#define gtk_window_set_default IA__gtk_window_set_default

extern __typeof (gtk_window_set_default_icon) IA__gtk_window_set_default_icon __attribute((visibility("hidden")));
#define gtk_window_set_default_icon IA__gtk_window_set_default_icon

#ifndef _WIN64
extern __typeof (gtk_window_set_default_icon_from_file) IA__gtk_window_set_default_icon_from_file __attribute((visibility("hidden")));
#define gtk_window_set_default_icon_from_file IA__gtk_window_set_default_icon_from_file

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_window_set_default_icon_from_file_utf8) IA__gtk_window_set_default_icon_from_file_utf8 __attribute((visibility("hidden")));
#define gtk_window_set_default_icon_from_file_utf8 IA__gtk_window_set_default_icon_from_file_utf8

#endif
extern __typeof (gtk_window_set_default_icon_list) IA__gtk_window_set_default_icon_list __attribute((visibility("hidden")));
#define gtk_window_set_default_icon_list IA__gtk_window_set_default_icon_list

extern __typeof (gtk_window_set_default_icon_name) IA__gtk_window_set_default_icon_name __attribute((visibility("hidden")));
#define gtk_window_set_default_icon_name IA__gtk_window_set_default_icon_name

extern __typeof (gtk_window_set_default_size) IA__gtk_window_set_default_size __attribute((visibility("hidden")));
#define gtk_window_set_default_size IA__gtk_window_set_default_size

extern __typeof (gtk_window_set_destroy_with_parent) IA__gtk_window_set_destroy_with_parent __attribute((visibility("hidden")));
#define gtk_window_set_destroy_with_parent IA__gtk_window_set_destroy_with_parent

extern __typeof (gtk_window_set_focus) IA__gtk_window_set_focus __attribute((visibility("hidden")));
#define gtk_window_set_focus IA__gtk_window_set_focus

extern __typeof (gtk_window_set_focus_on_map) IA__gtk_window_set_focus_on_map __attribute((visibility("hidden")));
#define gtk_window_set_focus_on_map IA__gtk_window_set_focus_on_map

extern __typeof (gtk_window_set_startup_id) IA__gtk_window_set_startup_id __attribute((visibility("hidden")));
#define gtk_window_set_startup_id IA__gtk_window_set_startup_id

extern __typeof (gtk_window_set_geometry_hints) IA__gtk_window_set_geometry_hints __attribute((visibility("hidden")));
#define gtk_window_set_geometry_hints IA__gtk_window_set_geometry_hints

extern __typeof (gtk_window_set_gravity) IA__gtk_window_set_gravity __attribute((visibility("hidden")));
#define gtk_window_set_gravity IA__gtk_window_set_gravity

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_window_get_frame_dimensions) IA__gtk_window_get_frame_dimensions __attribute((visibility("hidden")));
#define gtk_window_get_frame_dimensions IA__gtk_window_get_frame_dimensions

extern __typeof (gtk_window_set_frame_dimensions) IA__gtk_window_set_frame_dimensions __attribute((visibility("hidden")));
#define gtk_window_set_frame_dimensions IA__gtk_window_set_frame_dimensions

extern __typeof (gtk_window_get_has_frame) IA__gtk_window_get_has_frame __attribute((visibility("hidden")));
#define gtk_window_get_has_frame IA__gtk_window_get_has_frame

extern __typeof (gtk_window_set_has_frame) IA__gtk_window_set_has_frame __attribute((visibility("hidden")));
#define gtk_window_set_has_frame IA__gtk_window_set_has_frame

#endif
extern __typeof (gtk_window_set_icon) IA__gtk_window_set_icon __attribute((visibility("hidden")));
#define gtk_window_set_icon IA__gtk_window_set_icon

#ifndef _WIN64
extern __typeof (gtk_window_set_icon_from_file) IA__gtk_window_set_icon_from_file __attribute((visibility("hidden")));
#define gtk_window_set_icon_from_file IA__gtk_window_set_icon_from_file

#endif
#ifdef G_OS_WIN32
extern __typeof (gtk_window_set_icon_from_file_utf8) IA__gtk_window_set_icon_from_file_utf8 __attribute((visibility("hidden")));
#define gtk_window_set_icon_from_file_utf8 IA__gtk_window_set_icon_from_file_utf8

#endif
extern __typeof (gtk_window_set_icon_list) IA__gtk_window_set_icon_list __attribute((visibility("hidden")));
#define gtk_window_set_icon_list IA__gtk_window_set_icon_list

extern __typeof (gtk_window_set_icon_name) IA__gtk_window_set_icon_name __attribute((visibility("hidden")));
#define gtk_window_set_icon_name IA__gtk_window_set_icon_name

extern __typeof (gtk_window_set_keep_above) IA__gtk_window_set_keep_above __attribute((visibility("hidden")));
#define gtk_window_set_keep_above IA__gtk_window_set_keep_above

extern __typeof (gtk_window_set_keep_below) IA__gtk_window_set_keep_below __attribute((visibility("hidden")));
#define gtk_window_set_keep_below IA__gtk_window_set_keep_below

extern __typeof (gtk_window_set_mnemonic_modifier) IA__gtk_window_set_mnemonic_modifier __attribute((visibility("hidden")));
#define gtk_window_set_mnemonic_modifier IA__gtk_window_set_mnemonic_modifier

extern __typeof (gtk_window_set_mnemonics_visible) IA__gtk_window_set_mnemonics_visible __attribute((visibility("hidden")));
#define gtk_window_set_mnemonics_visible IA__gtk_window_set_mnemonics_visible

extern __typeof (gtk_window_set_modal) IA__gtk_window_set_modal __attribute((visibility("hidden")));
#define gtk_window_set_modal IA__gtk_window_set_modal

extern __typeof (gtk_window_set_opacity) IA__gtk_window_set_opacity __attribute((visibility("hidden")));
#define gtk_window_set_opacity IA__gtk_window_set_opacity

extern __typeof (gtk_window_get_opacity) IA__gtk_window_get_opacity __attribute((visibility("hidden")));
#define gtk_window_get_opacity IA__gtk_window_get_opacity

#ifndef GTK_DISABLE_DEPRECATED
extern __typeof (gtk_window_set_policy) IA__gtk_window_set_policy __attribute((visibility("hidden")));
#define gtk_window_set_policy IA__gtk_window_set_policy

#endif
extern __typeof (gtk_window_set_position) IA__gtk_window_set_position __attribute((visibility("hidden")));
#define gtk_window_set_position IA__gtk_window_set_position

extern __typeof (gtk_window_set_resizable) IA__gtk_window_set_resizable __attribute((visibility("hidden")));
#define gtk_window_set_resizable IA__gtk_window_set_resizable

extern __typeof (gtk_window_set_role) IA__gtk_window_set_role __attribute((visibility("hidden")));
#define gtk_window_set_role IA__gtk_window_set_role

extern __typeof (gtk_window_set_screen) IA__gtk_window_set_screen __attribute((visibility("hidden")));
#define gtk_window_set_screen IA__gtk_window_set_screen

extern __typeof (gtk_window_set_skip_pager_hint) IA__gtk_window_set_skip_pager_hint __attribute((visibility("hidden")));
#define gtk_window_set_skip_pager_hint IA__gtk_window_set_skip_pager_hint

extern __typeof (gtk_window_set_skip_taskbar_hint) IA__gtk_window_set_skip_taskbar_hint __attribute((visibility("hidden")));
#define gtk_window_set_skip_taskbar_hint IA__gtk_window_set_skip_taskbar_hint

extern __typeof (gtk_window_set_urgency_hint) IA__gtk_window_set_urgency_hint __attribute((visibility("hidden")));
#define gtk_window_set_urgency_hint IA__gtk_window_set_urgency_hint

extern __typeof (gtk_window_set_title) IA__gtk_window_set_title __attribute((visibility("hidden")));
#define gtk_window_set_title IA__gtk_window_set_title

extern __typeof (gtk_window_set_transient_for) IA__gtk_window_set_transient_for __attribute((visibility("hidden")));
#define gtk_window_set_transient_for IA__gtk_window_set_transient_for

extern __typeof (gtk_window_set_type_hint) IA__gtk_window_set_type_hint __attribute((visibility("hidden")));
#define gtk_window_set_type_hint IA__gtk_window_set_type_hint

extern __typeof (gtk_window_set_wmclass) IA__gtk_window_set_wmclass __attribute((visibility("hidden")));
#define gtk_window_set_wmclass IA__gtk_window_set_wmclass

extern __typeof (gtk_window_stick) IA__gtk_window_stick __attribute((visibility("hidden")));
#define gtk_window_stick IA__gtk_window_stick

extern __typeof (gtk_window_unfullscreen) IA__gtk_window_unfullscreen __attribute((visibility("hidden")));
#define gtk_window_unfullscreen IA__gtk_window_unfullscreen

extern __typeof (gtk_window_unmaximize) IA__gtk_window_unmaximize __attribute((visibility("hidden")));
#define gtk_window_unmaximize IA__gtk_window_unmaximize

extern __typeof (gtk_window_unstick) IA__gtk_window_unstick __attribute((visibility("hidden")));
#define gtk_window_unstick IA__gtk_window_unstick

#endif
#endif
#if IN_HEADER(__GTK_WIN32_EMBED_WIDGET_H__)
#if IN_FILE(__GTK_WIN32_EMBED_WIDGET_C__)
#ifdef G_OS_WIN32
extern __typeof (gtk_win32_embed_widget_get_type) IA__gtk_win32_embed_widget_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_win32_embed_widget_get_type IA__gtk_win32_embed_widget_get_type

#endif
#endif
#endif
#if IN_HEADER(__GTK_INFO_BAR_H__)
#if IN_FILE(__GTK_INFO_BAR_C__)
extern __typeof (gtk_info_bar_get_type) IA__gtk_info_bar_get_type __attribute((visibility("hidden"))) G_GNUC_CONST;
#define gtk_info_bar_get_type IA__gtk_info_bar_get_type

extern __typeof (gtk_info_bar_new) IA__gtk_info_bar_new __attribute((visibility("hidden")));
#define gtk_info_bar_new IA__gtk_info_bar_new

extern __typeof (gtk_info_bar_new_with_buttons) IA__gtk_info_bar_new_with_buttons __attribute((visibility("hidden")));
#define gtk_info_bar_new_with_buttons IA__gtk_info_bar_new_with_buttons

extern __typeof (gtk_info_bar_get_action_area) IA__gtk_info_bar_get_action_area __attribute((visibility("hidden")));
#define gtk_info_bar_get_action_area IA__gtk_info_bar_get_action_area

extern __typeof (gtk_info_bar_get_content_area) IA__gtk_info_bar_get_content_area __attribute((visibility("hidden")));
#define gtk_info_bar_get_content_area IA__gtk_info_bar_get_content_area

extern __typeof (gtk_info_bar_add_action_widget) IA__gtk_info_bar_add_action_widget __attribute((visibility("hidden")));
#define gtk_info_bar_add_action_widget IA__gtk_info_bar_add_action_widget

extern __typeof (gtk_info_bar_add_button) IA__gtk_info_bar_add_button __attribute((visibility("hidden")));
#define gtk_info_bar_add_button IA__gtk_info_bar_add_button

extern __typeof (gtk_info_bar_add_buttons) IA__gtk_info_bar_add_buttons __attribute((visibility("hidden")));
#define gtk_info_bar_add_buttons IA__gtk_info_bar_add_buttons

extern __typeof (gtk_info_bar_set_response_sensitive) IA__gtk_info_bar_set_response_sensitive __attribute((visibility("hidden")));
#define gtk_info_bar_set_response_sensitive IA__gtk_info_bar_set_response_sensitive

extern __typeof (gtk_info_bar_set_default_response) IA__gtk_info_bar_set_default_response __attribute((visibility("hidden")));
#define gtk_info_bar_set_default_response IA__gtk_info_bar_set_default_response

extern __typeof (gtk_info_bar_response) IA__gtk_info_bar_response __attribute((visibility("hidden")));
#define gtk_info_bar_response IA__gtk_info_bar_response

extern __typeof (gtk_info_bar_set_message_type) IA__gtk_info_bar_set_message_type __attribute((visibility("hidden")));
#define gtk_info_bar_set_message_type IA__gtk_info_bar_set_message_type

extern __typeof (gtk_info_bar_get_message_type) IA__gtk_info_bar_get_message_type __attribute((visibility("hidden")));
#define gtk_info_bar_get_message_type IA__gtk_info_bar_get_message_type

#endif
#endif
#endif /* G_HAVE_GNUC_VISIBILITY */
#endif /* DISABLE_VISIBILITY */
